\ CSTRING.FTH - string class

((
Copyright (c) 2000
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
See TODO markers in the code

Change History
==============
20001103 SJC001	Fixed ResizeBuffer code to deal correctly with 
		case where the buffer size is being DECREASED
20030612 GMA001 added zplace and appendz
))


( *> classlib                                                           )
( **                                                                    )
( *S CString - Dynamic String Class                                     )
( **                                                                    )
( *P A CString object contains a variable-length sequence of characters.)
( ** It also provides functions and operators which allow for easy      )
( ** to Concatenation and comparison operators, together with automatic )
( ** memory management. CString objects are far easier to use than      )
( ** ordinary character arrays.                                         )
( **                                                                    )
( *P CString is based on the FORTH char data type.                      )
( **                                                                    )
( *P CString Objects have the following useful characteristics:         )
( **                                                                    )
( *B CString objects can grow as a result of concatenation operations.  )
( ** The memory management is automatic.                                )
( **                                                                    )
( *B You can easily substitute CString objects for const char* and      )
\ ** LPCTSTR function arguments by using the (LPCTSTR) operator or its  )
( ** associated method GetLPCTSTR.                                      )
( **                                                                    )
( *B You can lock a CString object to provide a character buffer at a   )
( ** fixed address and size for hacking.                                )
( **                                                                    )
( *H Variable Type Arguments                                            )
( **                                                                    )
( *P Some of the members of this class can take either a character or   )
( ** a zero-terminated string as a parameter. This is autodetected by   )
( ** the simple assumption that any value less than the maximum value   )
( ** storable in a char is an array pointer. For 8 bit character systems)
( ** this means a pointer cannot be in the range 0..255.                )
( **                                                                    )
\ =======================================================================

class CString

private:        \ *** The Memory Store (Internals)
                        LPVOID  m_buffer
                        int     m_length
                        meth:   ResizeBuffer

public:         \ *** The String as an Array

                        meth:   GetLength
                        meth:   IsEmpty
                        meth:   Empty
                        meth:   GetAt
                        meth:   SetAt
                        meth:   GetLPCTSTR
        []              oper:   GetAt
        []to            oper:   SetAt
        (LPCTSTR)       oper:   GetLPCTSTR

public:         \ *** Assignment / Contatenation

                        meth:   to
                        meth:   SetLPCTSTR
                        meth:   Add
                        meth:   AddLPCTSTR
				meth:   Zplace
				meth:   Appendz
        to              oper:   to
        (LPCTSTR)to     oper:   SetLPCTSTR
        +=              oper:   Add
        (LPCTSTR)+=     oper:   AddLPCTSTR

public:         \ *** Comparison

                        meth:   Compare
                        meth:   CompareNoCase
				meth:   Scan

public:         \ *** Extraction

                        meth:   Mid
                        meth:   Left
                        meth:   Right
                        meth:   SpanIncluding           \ TODO
                        meth:   SpanExcluding           \ TODO

public:         \ *** Other Conversions

                        meth:   MakeUpper
                        meth:   MakeLower
                        meth:   MakeReverse
                        meth:   Replace
                        meth:   Remove
                        meth:   Search
                        meth:   Insert
                        meth:   Delete
                        meth:   Format                  \ TODO
                        meth:   FormatV                 \ TODO
                        meth:   TrimLeft
                        meth:   TrimRight
                        meth:   FormatMessage           \ TODO

public:	\ utilities

				meth:   P$
end-class

\ -----------------------------------------------------------------------------
( **                                                                    )
( *H The CString Class Members                                          )
( **                                                                    )
\ -----------------------------------------------------------------------------

: CString::ResizeBuffer \ rsize --
( *G Resize the current string buffer memory to RSIZE chars.            )
  ::to m_length
  m_buffer
  if
    m_length allocate           ASSERT( dup 0= )
    drop >r
    m_buffer zcount 1 chars + m_length min r@ swap move			\ SJC003
    m_buffer free drop
    r> ::to m_buffer
  else
    m_length allocate           ASSERT( dup 0= )
    drop ::to m_buffer
  then
;

\ -----------------------------------------------------------------------------

: CString::Empty        \ --
( *G Release all string memory and return to init state.                )
  m_buffer ?dup if free drop then
  0 ::to m_buffer
  0 ::to m_length
;

: CString::GetAt        \ idx -- char
( *G Return the ascii character at IDX position in the string.          )
  ASSERT( m_buffer )
  ASSERT( dup m_length < )
  m_buffer + c@
;

: CString::GetLength    \ -- n
( *G Return the length of the current string.                           )
  ASSERT( m_buffer )
  m_length 1-
;

: CString::GetLPCTSTR   \ -- z$
( *G Return a pointer to the string as a zero-terminated. After         )
( ** obtaining this pointer, any operation which modifies the string may)
( ** destroy this pointer.                                              )
  ASSERT( m_buffer )
  m_buffer
;

: CString::IsEmpty      \ -- BOOL
( *G Return TRUE if there is no string information.                     )
  m_buffer 0=
;

: CString::SetAt        \ idx char --
( *G Place character CHAR at the IDX position in the string.            )
  ASSERT( m_buffer )
  ASSERT( over m_length 1- < )
  swap m_buffer + c!
;

\ -----------------------------------------------------------------------------

: CString::Add          \ *CString --
( *G Add the contents of the CString class pointed to onto the end of   )
( ** the current string.                                                )
  (CString).GetLPCTSTR AddLPCTSTR
;

: CString::AddLPCTSTR   \ z$ --
( *G Add the supplied zero terminated string to the end of the current. )
  zcount 1 chars +
  dup GetLength + ResizeBuffer
  m_buffer zcount + swap move
;

: CString::SetLPCTSTR   \ z$ --
( *G Replace the current string with the supplied zero terminated one.  )
  zcount
  dup 1 chars + ResizeBuffer            \ reallocate buffer
  m_buffer swap 1 chars + move          \ copy
;

: CString::Zplace   \ caddr len --
( *G Replace the current string with the supplied counted string, converting to z$ as we go.  )
	{ | Temp[ MAX_PATH ] }
	Temp[ ::zplace		\ counted string to buffer
	Temp[ SetLPCTSTR
;

: CString::Appendz	\ caddr len --
( *G Add the supplied counted string to the end of the current, converting to z$ as we go.  )
	{ | Temp[ MAX_PATH ] }
	Temp[ ::zplace		\ counted string to buffer
	Temp[ AddLPCTSTR
;

: CString::to           \ *CString --
( *G Replace the current string with the contents of the CString class  )
( ** whose pointer is supplied.                                         )
  (CString).GetLPCTSTR SetLPCTSTR
;

\ -----------------------------------------------------------------------------

: CString::Compare      \ z$ -- flag
( *G Compare the current string with the zero-terminated string supplied. )
  IsEmpty if drop 1 exit then
  zcount m_buffer zcount ::compare
;

: CString::CompareNoCase \ z$ -- flag
( *G As CString::Compare except character case is ignored.              )
  IsEmpty if drop 1 exit then
  zcount m_buffer zcount ::icompare
;

: CString::Scan 		\ char -- n flag
( *G scan z$ for the occurence of char, return character pos and true or false false )
	{ char }
	0 0 
	m_buffer zcount bounds
	Do
	 i c@ char =
	 If
	  2Drop i -1 Leave
	 Then
	Loop
;

\ -----------------------------------------------------------------------------

: CString::Mid          \ first count -- *CString(dynamic)
( *G Return a new dynamic instance pointer for a CString which contains )
( ** a substring of the current. COUNT characters from index FIRST are  )
( ** copied.                                                            )
  ASSERT( over GetLength <= )   \ first should be under length but allow
                                \ exact match for null string
  ASSERT( 2dup + 1- GetLength < )
  CString dnew >r               \ first count -- ; R: *CString --
  dup 1 chars + r@ (CString).ResizeBuffer
  r@ (CString).m_buffer r@ (CString).m_length erase
  m_buffer rot + r@ (CString).m_buffer rot move
  r>
;

: CString::Left         \ count -- *CString(dynamic)
( *G Return a new dynamic instance pointer for a CString which contains )
( ** a substring of the current. COUNT characters are copied from the   )
( ** start left of the string.                                        )
  0 swap Mid
;

: CString::Right        \ count -- *CString(dynamic)
( *G Return a new dynamic instance pointer for a CString which contains )
( ** a substring of the current. COUNT characters are copied from the   )
( ** end right of the string.                                         )
  GetLength over - swap Mid
;

\ -----------------------------------------------------------------------------

: CString::Delete       \ index count -- newlen
( *G Remove COUNT characters from the string starting at the INDEX      )
( ** position. If count+index exceeds the string length it is trucated. )
( ** Also returns the new length of the string after the delete.        )
  ASSERT( over 0< not )
  ASSERT( over GetLength < )
  over Left >r          \ index count -- ; R: *cs --
  + GetLength swap - 0 max      \ #right -- ; R: *cs --
  ?dup if
        Right
        dup r@ (CString).Add
        ::delete
  then
  r@ to
  r> ::delete
  GetLength
;

: CString::Insert       \ index z$ -- int | index char -- int
( *G Passed either an index and a z$ or an index and a character this   )
( ** will perform an insert operation. The string or character supplied )
( ** is inserted starting at the original offset INDEX. Returns the     )
( ** new length of the string.                                          )
  ASSERT( over GetLength 1+ < )         \ allow for 1 extra to insert o end
                                        \ (basically an Append)
  over Left tuck			\ -- index newl z$ newl
  (Cstring).addLPCTSTR			\ -- index newl
  GetLength rot - Right			\ -- newl newr
  2dup swap (Cstring).add		\ -- newl newr
  ::delete 				\ -- newl
  dup to
  ::delete
  GetLength
;

: CString::MakeUpper    \ --
( *G Convert the classes string data to upper case where possible.      )
  m_buffer m_length bounds
  ?do
    i c@
    dup [char] a [char] z within?
    if
      bl - i c!
    else
      drop
    then
  loop
;

: CString::MakeLower    \ --
( *G Convert the classes string data to lower case where possible.      )
  m_buffer m_length bounds
  ?do
    i c@
    dup [char] A [char] Z within?
    if
      bl + i c!
    else
      drop
    then
  loop
;

: CString::MakeReverse  \ --
;

: CString::Replace      \ oldchar newchar -- #replaced | oldstr newstr -- #replaced

                        { | nc p ss[ 256 chars ] rs[ 256 chars ] -- }

  ss[ 256 chars erase
  rs[ 256 chars erase
  dup 255 > if
    zcount rs[ swap move
    zcount ss[ swap move
  else
    rs[ c!
    ss[ c!
  then

  0 -> nc
                        
                        
  0 -> p

  begin
    p GetLength <
  while
    ss[ zcount m_buffer p + over ::compare
    0=
    if
        p ss[ zcount nip Delete drop    \ trash chars for replace
        p rs[ Insert drop               \ and insert new string
        nc 1+ -> nc
        p rs[ zcount nip + -> p
    else
        p 1+ -> p
    then
  repeat
  nc
;


: CString::Remove      \ z$ -- f
	{ | p res len ss[ 256 ] }
	GetLength -> Len
	ss[ 256 chars erase
	zcount ss[ swap move	\ move string
	0 -> p
	0 -> res

  	begin
	 ss[ zcount m_buffer p + over ::compare
	 0=
	  if
	    p ss[ zcount nip Delete drop    \ trash chars
	    -1 -> res
	   else
	     p 1+ -> p
	  then
	  p len <
	Until
	res
;

: CString::Search		\ z$ -- flag
	>r GetLPCTSTR zcount r> zcount ::SEARCH 
	Nip Nip
;

: CString::TrimLeft     \ char -- | z$ --
;

: CString::TrimRight    \ char -- | z$ --
;

: CString::P$		\  --
	GetLPCTSTR .z$
;


	
\ -----------------------------------------------------------------------------
( **                                                                    )
( *H Operator Associations                                              )
( **                                                                    )
( *P The following operators have been assigned to methods for this     )
( ** class.                                                             )
( **                                                                    )
( *D []              GetAt -- Get character at specified index.         )
( *D []to            SetAt -- Set chatacter at specified index.         )
\ *D (LPCTSTR)       GetLPCTSTR -- Return 0terminated string pointer.   )
( *D to              to -- Assign from another CString.                 )
\ *D (LPCTSTR)to     SetLPCTSTR -- Assign from a 0 terminated string.   )
( *D +=              Add -- Append from another CString.                )
\ *D (LPCTSTR)+=     AddLPCTSTR -- Append from a 0 terminated string.   )
( **                                                                    )
( *> ###                                                                )
\ -----------------------------------------------------------------------------

: z$ [char] ' emit .z$ [char] ' emit ;

: t1                    \ --
  CString new s1
  4 s1.left drop

  cr
  ."                                    111111111122222222223" cr
  ."                          0123456789012345678901234567890" cr

  ." Setting S1              "
  Z" Mary Had A Little Lamb" (lpctstr)to s1
  (lpctstr) s1 z$ cr

  ." Test Left(4)            "
  4 s1.left dup (CString).GetLPCTSTR z$ cr delete

  ." Test Right(6)           "
  6 s1.Right dup (CString).GetLPCTSTR z$ cr delete

  ." Test Mid(6,4)           "
  6 4 s1.Mid dup (CString).GetLPCTSTR z$ cr delete

  ." Test Delete(6,4)        "
  6 4 s1.Delete (lpctstr) s1 z$ 4 spaces . cr

  ." Test Delete(0,2)        "
  0 2 s1.Delete (lpctstr) s1 z$ 4 spaces . cr

  ." Test Delete(14,2)       "
  14 2 s1.Delete (lpctstr) s1 z$ 4 spaces . cr

  ." Test Delete(10,100)     "
  10 100 s1.Delete (lpctstr) s1 z$ 4 spaces . cr

  ." Test Insert(3,'dude')   " 
  3 Z" dude" s1.Insert (lpctstr) s1 z$ 4 spaces . cr

  ." Test Insert(14,'dude')  "
  14 Z" dude" s1.Insert (lpctstr) s1 z$ 4 spaces . cr

  ." Test Replace('d','D')   "
  [char] d [char] D s1.Replace (lpctstr) s1 z$ 4 spaces . cr

  ." Test Replace('DuDe','*')"
  Z" DuDe" Z" *" s1.Replace (lpctstr) s1 z$ 4 spaces . cr
;

