\ editor.fth - main portion of Glade text editor

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *> gtkbindings
\ *S A text editor in Glade
\ ===========
\ *P The original design and C code is by Micah Carrick, whose
\ ** tutorial is well worth studying. It is at:
\ *C   http://www.micahcarrick.com/gtk-glade-tutorial-part-1.html
\ *P The Forth code presented here is liberally derived from that
\ ** presentation and code.

\ *P To compile the text editor demo, CD to the directory containing
\ ** *\i{editor.fth} and then:
\ *C   include TextEdDemo.bld
\ *P To run the editor from the Forth console:
\ *C   runTextEd
\ *P The code will run unchanged on VFX Forth for Windows, Mac
\ ** and Linux.

struct /TextEd	\ -- len
\ *G Everything we need to know about the editor can be derived
\ ** from this structure.
   int te.Main			\ main window widget
   int te.About			\ about box widget
   int te.TextView		\ text view widget
   int te.StatusBar		\ status bar widget
   int te.SbContextId		\ status bar context ID
   int te.pFileName		\ address of current file name
   int te.QSaveDlg		\ Save? dialog
   int te.ErrorDlg		\ Error dialog
end-struct

0 value pTextEd	\ -- addr
\ *G Holds the address of a structure for the current text editor.

#1024 constant /NameBuffer	\ -- len
\ *G Largest file name.
/NameBuffer buffer: zFilenameBuffer	\ -- zaddr
\ *G Buffer to hold current file name.
#2048 constant /StatusBuffer	\ -- len
\ *G Size of the status buffer
/StatusBuffer buffer: zStatusBuffer	\ -- zaddr
\ *G Text buffer for status bar.


\ ********
\ *N Tools
\ ********

: EdErrBox	\ zmessage --
\ *G Displays an error message box.
  z" TextEd Error" pTextEd te.Main @ ErrorBox  ;



\ ************************
\ *N Status bar operations
\ ************************

: sbParams	\ -- sb context
\ *G REturn the status bar parameters
  pTextEd dup te.StatusBar @  swap te.SbContextId @
;

: setStatus	\ z$1 z$2 --
\ *G set the status buffer, merging the two strings. Then update
\ ** the status bar.
  swap zcount zStatusBuffer zplace	\ z1
  s"  " zStatusBuffer zAppend		\ separator
  zcount zStatusBuffer zAppend		\ z2
  sbParams 2dup gtk_statusbar_pop	\ discard previous
  zStatusBuffer gtk_statusbar_push drop	\ set new
  gtkPump				\ let it settle
;

: clearStatus	\ --
\ *G Clear the status bar.
  z" " z" " setStatus
;


\ ************************
\ *N TextViews and buffers
\ ************************

: CurrView	\ -- textview
  pTextEd te.TextView @
;

: CurrBuffer	\ -- buffer
  CurrView gtk_text_view_get_buffer	\ returns buffer
;

: modified?	\ -- flag
\ *G Return true if the current text has been modified and not saved.
  CurrBuffer gtk_text_buffer_get_modified  ;

: modified	\ --
\ *G Mark the current text buffer as modified.
  CurrBuffer 1 gtk_text_buffer_set_modified
;

: unmodified	\ --
\ *G Mark the current text buffer as unmodified.
  CurrBuffer 0 gtk_text_buffer_set_modified
;

: inactive	\ --
\ *G Set the current text window as inactive (unresponsive).
  CurrView 0 gtk_widget_set_sensitive	\ inactive
;

: active	\ --
\ *G Set the current text window as active (responsive).
  CurrView 1 gtk_widget_set_sensitive	\ active
;

: getCurrText	\ -- text
\ *G Get a copy of the current text. When you are finished with
\ ** it, you must release it with *\fo{g_free ( text -- )}.
  {: | start[ /GtkTextIter ] end[ /GtkTextIter ] -- text :}
  inactive
  CurrBuffer
  dup start[ end[ gtk_text_buffer_get_bounds
  start[ end[ 0 gtk_text_buffer_get_text
  unModified
  active
;


\ **************************
\ *N Loading and saving text
\ **************************

: CurrFilename	\ -- z$
\ *G REturn the current text file name.
  pTextEd te.pFileName @
;

: MustSave?	\ -- flag
\ *G Return true if the buffer has been modified and the user
\ ** wants to save it.
  Modified? 0=
  if  0 exit  endif
  pTextEd te.QSaveDlg @ runDlg GTK_RESPONSE_YES =
;

: getSaveFileName	\ --
\ *G Set the current text file name for saving.
  CurrFileName /NameBuffer pTextEd te.Main @ AskSaveFileNameBox drop
;

: getOpenFileName	\ --
\ *G Set the current text file name for loading.
  CurrFileName /NameBuffer pTextEd te.Main @ AskOpenFileNameBox drop
;

: sbSaving...	\ --
\ *G Show status bar as saving.
  z" Saving ..." CurrFileName setStatus
;

: sbLoading...	\ --
\ *G Show status bar as loading.
  z" Loading ..." CurrFileName setStatus
;

: fileStatus	\ --
\ *G Show filename on status bar.
  CurrFileName zNull setStatus
;

: loadCurrFile	\ -- *text 0 | -1
\ *G Load the contents of the current file. On success, return a
\ ** pointer to the text and zero. On failure, just return -1.
\ ** When finished with, the text pointer must be freed with
\ ** *\fo{g_free}.
  {: | text[ cell ] perr[ cell ] -- :}
  perr[ off
  CurrFileName text[ 0 perr[ g_file_get_contents 0= if
    perr[ @ ge.*message @ EdErrBox
    perr[ @ g_error_free
    -1 exit
  endif
  text[ @ 0
;

: loadCurrText	\ --
\ *G Load the text of the current window from the current file.
\ ** No action is taken if the filename is null.
  CurrFileName c@ if
    sbLoading...
    loadCurrFile 0= if
      >r  inactive
      CurrBuffer r@ -1 gtk_text_buffer_set_text
      r> g_free
      unmodified active
    endif
    fileStatus
  endif
;
: writeCurrText	\ --
\ *G Save the text of the current window to the current file.
\ ** No action is taken if the filename is null.
  {: | perr[ cell ] -- :}
  CurrFileName c@ if
    sbSaving...
    perr[ off
    CurrFileName  getCurrText dup >r
    -1 perr[ g_file_set_contents 0= if
      perr[ @ ge.*message @ EdErrBox
      perr[ @ g_error_free
    endif
    r> g_free
    fileStatus
  endif
;

: saveCurrText	\ --
\ *G As *\fo{writeCurrText} but asks for a file name if one has
\ ** not been set.
  CurrFileName c@ 0=
  if  getSaveFileName  endif
  writeCurrText
;

: saveAsCurrText	\ --
\ *G As *\fo{writeCurrText} but always asks for a file name.
  getSaveFileName  writeCurrText
;

: checkedSave	\ --
\ *G If text has been modified and the user wants saving,
\ ** write the text to a file.
  mustSave?
  if  saveCurrText  endif
;

: openCurrText	\ --
\ *G Open a new file.
  checkedSave
  getOpenFileName loadCurrText
;

: newCurrText	\ --
\ *G Start with an empty buffer.
  checkedSave
  CurrBuffer z" " -1 gtk_text_buffer_set_text
  unmodified
  0 CurrFileName c!
  clearStatus
;


\ ************
\ *N Clipboard
\ ************

((
#define GDK_SELECTION_PRIMARY 		_GDK_MAKE_ATOM (1)
#define GDK_SELECTION_SECONDARY 	_GDK_MAKE_ATOM (2)
#define GDK_SELECTION_CLIPBOARD 	_GDK_MAKE_ATOM (69)
#define GDK_TARGET_BITMAP 		_GDK_MAKE_ATOM (5)
#define GDK_TARGET_COLORMAP 		_GDK_MAKE_ATOM (7)
#define GDK_TARGET_DRAWABLE 		_GDK_MAKE_ATOM (17)
#define GDK_TARGET_PIXMAP 		_GDK_MAKE_ATOM (20)
#define GDK_TARGET_STRING 		_GDK_MAKE_ATOM (31)
#define GDK_SELECTION_TYPE_ATOM 	_GDK_MAKE_ATOM (4)
#define GDK_SELECTION_TYPE_BITMAP 	_GDK_MAKE_ATOM (5)
#define GDK_SELECTION_TYPE_COLORMAP 	_GDK_MAKE_ATOM (7)
#define GDK_SELECTION_TYPE_DRAWABLE 	_GDK_MAKE_ATOM (17)
#define GDK_SELECTION_TYPE_INTEGER 	_GDK_MAKE_ATOM (19)
#define GDK_SELECTION_TYPE_PIXMAP 	_GDK_MAKE_ATOM (20)
#define GDK_SELECTION_TYPE_WINDOW 	_GDK_MAKE_ATOM (33)
#define GDK_SELECTION_TYPE_STRING 	_GDK_MAKE_ATOM (31)
))

: CurrSelection	\ -- clipboard
\ *G Get the clipboard item for the current selection.
  69 gtk_clipboard_get
;

: doCut		\ --
\ *G Do the cut operation.
  CurrBuffer CurrSelection 1 gtk_text_buffer_cut_clipboard
;

: doCopy	\ --
\ *G Do the copy operation.
  CurrBuffer CurrSelection gtk_text_buffer_copy_clipboard
;

: doPaste	\ --
\ *G Do the paste operation.
  CurrBuffer CurrSelection 0 1 gtk_text_buffer_paste_clipboard
;

: doDelete	\ --
\ *G Do the delete operation.
  CurrBuffer 0 1 gtk_text_buffer_delete_selection drop
;


\ ************
\ *N Callbacks
\ ************

\ ============
\ *H File Menu
\ ============

2 0 CCBproc: on_new_MainFileMenu_activate	\ *widget *editor --
\ *G Callback for the "New" button.
  2drop newCurrText
;

2 0 CCBproc: on_Open_MainFileMenu_activate	\ *widget *editor --
\ *G Callback for the "Open" button.
  2drop openCurrText
;

2 0 CCBproc: on_save_MainFileMenu_activate	\ *widget *editor --
\ *G Callback for the "Save" button.
  2drop saveCurrText
;

2 0 CCBproc: on_SaveAs_MainFileMenu_activate	\ *widget *editor --
\ *G Callback for the "Save As" button.
  2drop saveAsCurrText
;

2 0 CCBproc: on_MainWindow_destroy		\ *wiget *editor --
\ *G Callback for final destroy of main window.
  2drop GtkAppQuit
;

3 1 CCBProc: on_MainWindow_delete_event		\ *widget *event *editor -- 0/1
\ *G When the window is requested to be closed, we need to check if
\ ** they have unsaved work. We use this callback to prompt the user to
\ ** save their work before they exit the application. From the
\ ** "delete-event" signal, we can choose to effectively cancel the
\ ** close based on the value we return.
  3drop checkedSave 0
;

2 0 CCBproc: on_quit_MainFileMenu_activate	\ *wiget *editor --
\ *G Callback for the "Quit" button.
  2drop checkedSave
  pTextEd te.Main @ gtk_widget_destroy
;


\ ============
\ *H Edit menu
\ ============

2 0 CCBproc: on_Cut_MainEditMenu_activate
\ *G Callback for the "Cut" button.
  2drop doCut
;

2 0 CCBproc: on_Copy_MainEditMenu_activate
\ *G Callback for the "Copy" button.
  2drop doCopy
;

2 0 CCBproc: on_Paste_MainEditMenu_activate
\ *G Callback for the "Paste" button.
  2drop doPaste
;

2 0 CCBproc: on_Delete_MainEditMenu_activate
\ *G Callback for the "Delete" button.
  2drop doDelete
;


\ ============
\ *H Help menu
\ ============

2 0 CCBproc: on_About_MainHelpMenu_activate	\ *widget *user -- ; -- entry
\ *G Callback to run the About dialog.
  2drop  pTextEd te.About @ runDlg drop
;

2 0 CCBproc: on_aboutdialog1_close	\ *widget *user -- ; -- entry
\ *G Callback when about box is closed.
  2drop
;


\ *********************************
\ *N Initialisation and termination
\ *********************************

: loadTeGUI	\ --
\ *G Load the text editor's GUI. After the file has been loaded,
\ ** the widgets we need to access are extracted and their object
\ ** pointers saved in a *\fo{/TextEd} structure. The design file
\ ** object is then released.
  /TextEd g_slice_alloc to pTextEd
  pTextEd /TextEd erase

  z" ./TextEd.ui" loadBuilderXML
  0= abort" Can't load GUI file"
  z" MainWindow" builderObject pTextEd te.Main !
  z" MainTextview" builderObject pTextEd te.TextView !
  z" MainStatusbar" builderObject pTextEd te.StatusBar !
  z" QSaveDlg" builderObject pTextEd te.QSaveDlg !
  z" ErrorDlg" builderObject pTextEd te.ErrorDlg !
  z" aboutdialog1" builderObject pTextEd te.About !
  freeBuilder
;

: initTextEd	\ --
\ *Initialise the application.
  pTextEd te.QSaveDlg @  pTextEd te.Main @ gtk_widget_set_parent_window
  pTextEd te.ErrorDlg @  pTextEd te.Main @ gtk_widget_set_parent_window

  \ Status bar connection ID
  pTextEd te.StatusBar @ z" TextEd"		\ the status bar's message stack
  gtk_statusbar_get_context_id			\ needs an identifier
  pTextEd te.SbContextID !			\ we only use one

  \ set the text view font
  z" monospace 10" pango_font_description_from_string
  pTextEd te.TextView @  over gtk_widget_modify_font
  pango_font_description_free

  z" gtk-edit" gtk_window_set_default_icon_name	\ default icons
  0 zFilenameBuffer c!				\ no filename yet
  zFilenameBuffer pTextEd te.pFileName !	\ set default filename
;

: termTextEd	\ --
\ *G free up the application data and perform termination actions.
  pTextEd te.Main @ gtk_widget_hide
  /TextEd pTextEd g_slice_free1
;

: RunTextEd	\ --
\ *G Run the text editor.
  do_gtk_init  installGTKhooks
  loadTeGUI
  initTextEd

  pTextEd te.Main @ gtk_widget_show_all
\  gtk_main

\  termTextEd
;


\ ======
\ *> ###
\ ======

decimal
