\ To interface with a single PCF8583 Real Time Clock

((
Copyright (c) 1999
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com
))


\ ------------------------------------------------------------------------------
\ 8583 address prefix constant
\ ------------------------------------------------------------------------------

 $0A0 constant 8583_id0		\ device address, including
 				\ 4 MSBs = Manufacturers prefix ( A0 for 8583 )
				\ bits 4,3,2 = device number, LSB = R/W - R=1, W=0

\ ------------------------------------------------------------------------------
\ 8583 addresses of internal registers
\ ------------------------------------------------------------------------------

$000 constant control
$001 constant hundreds
$002 constant seconds
$003 constant minutes
$004 constant hours
$005 constant year_date
$006 constant weekday_month
$010 constant yearoffset_address1
$011 constant yearoffset_address2
$012 constant rambase			\ 	 		;  base address for free RAM

\ ------------------------------------------------------------------------------
\ writes a data-byte on the 8583 using (w_address, d_address, data )
\ w_address: addresses the RTC registers & memory
\ data: The data you are sending
\ ------------------------------------------------------------------------------

: write_8583		\ data w_address devaddr --
  start_i2c           			\ -- data w_address devaddr	; send start condition
  $0FE and write_byte_i2c		\ -- data w_address 		; send device address
  write_byte_i2c			\ -- data 			; send word address
  write_byte_i2c               		\ -- 				; send data
  stop_i2c				\			        ; send stop condition
;

\ ------------------------------------------------------------------------------
\ sets word pointer address using d_address and w_address
\ w_address:  addresses the RTC registers & memory
\ ------------------------------------------------------------------------------

: set_read_address	\ w_address devaddr --
  start_i2c			        \ -- w_address devaddr		; send start condition
  $FE and write_byte_i2c		\ -- w_address 		        ; send device address
  write_byte_i2c			\ -- 				; send device address
;

\ ------------------------------------------------------------------------------
\ reads a byte from the 8583 using w_address, d_address and data
\ w_address:  addresses the RTC registers & memory
\ data: if any ack_bit fails an error flag is returned, otherwise the data and a false flag are returned
\ ------------------------------------------------------------------------------

: read_8583	\ devaddr -- data
  start_i2c		   	  	\ 	 	   	; send start condition
  $001 or write_byte_i2c		\ -- 			; set LSB ( read mode ) and send d_address
  read_byte_i2c				\ -- data 	  	; write w_address
  stop_i2c				\ -- data		; send stop condition
;

\ ----------------------------------------------------------------------------
\ Fetches a byte of data at a given address
\ ----------------------------------------------------------------------------

: rtc@		\ w_address devaddr -- byte
  tuck set_read_address			\			; set the word address
  read_8583				\			; read the data byte
;

\ ----------------------------------------------------------------------------
\ writes a byte of data to a given address
\ ----------------------------------------------------------------------------

: rtc!		\ byte w_address devaddr --
  write_8583 				\ write data at devaddr
;

\ ------------------------------------------------------------------------------
\ initialise control register, default is for clock is stopped
\ ------------------------------------------------------------------------------

: resetclock	\ devaddr --
  >r
  $083 control r@ rtc! 	\	; 50% timer duty = off, 50% alarm duty = off
  			\	; alarm enable = off, mask flag = off
			\	; function mode = 32.768Khz, hold last count = off
			\	; stop counting = on

  $000 hundreds r@ rtc!
  $000 seconds r@ rtc!
  $000 minutes r@ rtc!
  $000 hours r@ rtc!
  $000 year_date r@ rtc!
  $000 weekday_month r> rtc!
;

\ ------------------------------------------------------------------------------
\ stop the clock
\ bit 8 is set, all other bits are perseved
\ ------------------------------------------------------------------------------

: stopclock	\ devaddr --
  >r
  control r@ rtc@ $080 OR	\ --  data 	; read control register mask and set bit8
  control r> rtc!		\		; store new value in control register
;

\ ------------------------------------------------------------------------------
\  start the clock
\  bit 8 is cleared all, other bits are perserved
\ ------------------------------------------------------------------------------

: startclock	\ devaddr --
  >r
  control r@ rtc@ $080 OR	\ -- bit8 		; read control register mask and set bit8
  $080 XOR 	 	        \ -- bit8_XORed	  	; invert bit8 so it is set to 0
  control r> rtc!      		\ 			; write new value in control register
;

\ ------------------------------------------------------------------------------
\  latch the the clock, clock still counts but clock registers are frozen to enable a read
\  bit 7 is set, all other bits are perserved
\ ------------------------------------------------------------------------------

: latchclock	\ devaddr --
  >r
  control r@ rtc@ $040 OR	\ -- data 	; read control register mask and set bit 7
  control r> rtc!		\		; store new value in control register
;

\ ------------------------------------------------------------------------------
\  un-latch the the clock, clock registers are un-frozen and updated with latest count values
\  bit 7 is cleared, all other bits are perserved
\ ------------------------------------------------------------------------------

: unlatchclock	\ devaddr --
  >r
  control r@ rtc@ $040 OR	\ -- bit7 		; read control register mask and set bit7
  $040 XOR 	 	        \ -- bit7_XORed		; invert bit7 so it is set to 0
  control r> rtc!     		\ 			; write new value in control register
;

\ ------------------------------------------------------------------------------
\ store a value at an address and converts it from bin to bcd first
\ ------------------------------------------------------------------------------

: number!	\ byte address devaddr --
  rot bin2bcd -rot rtc!		\ convert byte to BCD before storing
;

\ ------------------------------------------------------------------------------
\ fetches a value from an address gievn and converts it from bcd to bin
\ ------------------------------------------------------------------------------

: number@	\ address devaddr -- bin_byte
  rtc@  		\ -- bcd_byte 		; fetch a byte from the address given
  bcd2bin	        \ -- bin_byte 	        ; convert byte from bcd to binary
;

\ ------------------------------------------------------------------------------
\ store the date
\ ------------------------------------------------------------------------------

 : date!	\ dd mm yyyy devaddr --
   >r
   r@ stopclock
   year_date $000 r@ rtc!	\ -- dd mm yyyy			; clear year_date register

   dup				\ -- dd mm yyyy yyyy 	        ; duplicate the year
   $0ff AND	        	\ -- dd mm yyyy low_byte 	; mask low byte
   swap 			\ -- dd mm low_byte yyyy 	; swap items
   8 rshift	        	\ -- dd mm low_byte high_byte 	; shift high byte into LS byte
   yearoffset_address1 r@ rtc!	\ -- dd mm low_byte 	        ; store high byte
   yearoffset_address2 r@ rtc!	\ -- dd mm 		        ; store low byte

   swap bin2bcd	        	\ -- mm dd 			; convert from bin to bcd
   year_date r@ rtc! 		\ -- mm 			; store new year_date value

   \ weekday_month register
   bin2bcd			\ -- mm				; convert from bin to bcd
   weekday_month r@ rtc!	\ -- mm				; store new weekday_month value
   r> startclock
 ;

\ ------------------------------------------------------------------------------
\ fetch the date
\ ------------------------------------------------------------------------------

: date@		\ devaddr -- dd mm yyyy
  >r
  \ year_date register
  year_date r@ rtc@	        \ -- year_date 				; fetch year_date
  dup				\ -- year_date year_date       		; duplicate top item
  $0C0 AND			\ -- year_date yy 			; mask off bits 6 & 7
  6  rshift		        \ -- year_date yy 			; shift right 6

  yearoffset_address1 r@ rtc@  	\ -- year_date yy yoffset1 		; fetch year offset 1
  8 lshift		        \ -- year_date yy yoffset1 	        ; shift year offset 1 into high byte
  yearoffset_address2 r@ rtc@   \ -- year_date yy yoffset1 yoffset2	; fetch year offset 2 and store in low byte
  or				\ -- year_date yy yoffset 		; or both bytes to give 1 double year offset
  +			        \ -- year_date yyyy 			; add year offset to current year count

  swap			        \ -- yyyy year_date 			; swap the top 2 items
  $03f AND			\ -- yyyy dd 				; mask off the 6 LSB's
  bcd2bin			\ -- yyyy dd 				; convert from bcd to bin

  \ weekday_month register
  weekday_month r> rtc@	        \ -- yyyy dd weekday_month 		; fetch weekday_month
  $01F AND			\ -- yyyy dd mm 			; mask off 4 LSB's
  bcd2bin			\ -- yyyy dd mm 	  		; convert from bcd to bin

  rot				\ -- dd mm yyyy				; swap top 2 items
;

\ ------------------------------------------------------------------------------
\ store the time
\ ------------------------------------------------------------------------------

: time!		\ secs mins hrs devaddr --
  >r
  r@ stopclock
  $03f AND			\				; mask off 6 LSB's to force 24hrs format
  hours r@ number!		\ -- secs mins 			; store hours value
  minutes r@ number!    	\ -- secs 			; store minutes value
  seconds r@ number!		\  				; store seconds value
  r> startclock
;

\ ------------------------------------------------------------------------------
\ fetch the time
\ ------------------------------------------------------------------------------

: time@		\ devaddr -- secs mins hrs
  >r
  seconds r@ number@		\ -- secs
  minutes r@ number@		\ -- secs mins
  hours r> number@  $03f AND	\ -- secs mins hrs		; remove am/pm flag & 24hr flag
;

\ ------------------------------------------------------------------------------
\ fetch time and date, but latch clock to ensure a stable reading
\ ------------------------------------------------------------------------------

: time&date	\ -- secs mins hrs dd mm yyyy
  8583_id0 >r			\ assume date comes from first RTC
  r@ latchclock			\	          			; stop register updates
  r@ time@		        \ -- secs mins hrs
  r@ date@		        \ -- secs mins hrs dd mm yyyy
  r> unlatchclock	        \ 					; restart register updates
;

\ ------------------------------------------------------------------------------
\ display the date
\ ------------------------------------------------------------------------------

: .date		\ day month year --
  swap rot			\ -- yyyy mm dd  	; re-order date
  s>d <# ascii / hold # # #>    \ -- yyyy mm hold 	; put day value in hold buffer and update hold pointer
  type			        \ -- yyyy mm		; display day values held in hold buffer
  s>d <# ascii / hold # # #>    \ -- yyyy hold		; put month value in hold buffer and update hold pointer
  type				\ -- yyyy		; display month values held in hold buffer
  s>d <# # # # # #> 		\ -- hold		; put year value in hold buffer and update hold pointer
  type			        \ -- 	 		; display year values held in hold buffer
;

\ ------------------------------------------------------------------------------
\ display in us format
\ ------------------------------------------------------------------------------

: .usdate	\ day month year --
  rot rot		        \ -- yyyy dd mm		; re-order date
  s>d <# ascii / hold # # #>    \ -- yyyy dd hold 	; put month value in hold buffer and update hold pointer
  type				\ -- yyyy dd     	; display month values held in hold buffer
  s>d <# ascii / hold # # #>    \ -- yyyy hold		; put day value in hold buffer and update hold pointer
  type			        \ -- yyyy       	; display day values held in hold buffer
  s>d <# # # # # #> 		\ -- hold        	; put year value in hold buffer and update hold pointer
  type			        \ -- 		     	; display year values held in hold buffer
;

\ ------------------------------------------------------------------------------
\ diplay the time
\ ------------------------------------------------------------------------------

 : .time 	\ -- --
   s>d <# ascii : hold # # #>   \ -- secs mins hold	; put hours value in hold buffer and update hold pointer
   type			        \ -- secs mins		; display hours values held in hold buffer
   s>d <# ascii : hold # # #>   \ -- secs hold		; put minutes value in hold buffer and update hold pointer
   type			        \ -- secs 	        ; display minutes values held in hold buffer
   s>d <# # # #> 		\ -- hold		; put seconds value in hold buffer and update hold pointer
   type  			\		       	; display seconds values held in hold buffer
 ;

\ ------------------------------------------------------------------------------
\ diplay the time & date
\ ------------------------------------------------------------------------------

: .time&date	\ secs mins hrs dd mm yyyy --
  cr   ." The date is "
  .date
  cr   ." The time is "
  .time
;

\ ------------------------------------------------------------------------------
\ print current time & date
\ ------------------------------------------------------------------------------

: .current	\ -- --
  time&date
  .time&date
;

\ ------------------------------------------------------------------------------
-docgen
