From gerry@jackson9000.fsnet.co.uk Sun Dec 23 22:09:40 2007
Newsgroups: comp.lang.forth
Subject: Interpreter mode loops
From: Gerry <gerry@jackson9000.fsnet.co.uk>
Date: Sun, 23 Dec 2007 14:09:40 -0800 (PST)

I'm not sure why but I started thinking about ANS Forth words
SAVE-INPUT and RESTORE-INPUT and realised that they could be
used to implement simple loops in a source file while in
interpreter mode. I would have thought that this must have done
before but a Google search of the internet and c.l.f
archives didn't turn up anything. The nearest I could find were
a couple of cases where 2 passes were made through some source
code. So I thought I would post this - my apologies if it is
already well known.

Simple loops can be implemented as follows:

: ndup  ( xn ... x1 n -- xn ... x1 xn ... x1 )
   dup 0 ?do dup pick swap loop drop
;

: ndrop  ( xn ... x1 n -- )
   0 ?do drop loop
;

: [begin] save-input ;

: [repeat]  ( xn ... x1 n -- xn ... x1 n )
   dup 1+ ndup restore-input
   abort" Cannot loop back"
;

: [until]  ( xn ... x1 n f -- xn ... x1 n | )
   if
      dup 1+ ndrop         ( -- )
   else
      [repeat]
   then
;

: [while]  ( xn ... x1 n f -- xn ... x1 n | )
   0=
   if
      dup 1+ ndrop
      postpone [else]
   then
;

\ Example of use
\ Display 0 to 99 with 10 integers per line using [begin] [until]

variable ct 0 ct !
100 constant limit

[begin]
   ct @ dup 10 mod 0=
   [if] cr [then]
   dup 3 .r
   1+ dup ct !
   limit =
[until]

cr .s

\ Display 0 to 99 with 10 integers per line using [begin] [while]
[repeat]

0 ct !

[begin]
   ct @ limit <
[while]
   ct @ dup 10 mod 0=
   [if] cr [then]
   dup 3 .r
   1+ ct !
[repeat] [then]

cr .s

Save the code in a file and then include it. It works with
GForth, VFX Forth but not with Win32 (6.10.05 build 2). I see
that GForth has its own version of [begin] etc but these seem to
only permit one liners.

A better implementation would move the input source state from
the stack to allow the user free use of the stack. Also [while]
should scan the source code for [repeat] to avoid the ugly need
for [then] after [repeat].

No doubt [do] [loops] could be coded but would be rather more
difficult, particularly if [i] [leave] etc were included.

Gerry

<snip>

Following the original post in this thread I've extended the
concept to include [do] [loop]s and friends. Whether any of this
is useful I'll leave others to judge, I've no particular need
for it and did it for a challenge and for fun. Particularly when
I realised that the data stack could be used to hold 2 stacks,
the usual one at the top and another, akin to the return stack,
at the bottom. Rather bizarre and horribly inefficient but it's
all in interpreter mode anyway. The return stack is not
available in interpreter mode of course. ROLL is indispensable
for using the datastack in this way (I think), hence the title
change.

I decided that the implementation:

1. would be in ANS Forth, hence portable
2. would not interfere with user use of the data stack and
   dataspace
3. would provide additional words such as [>r] etc for the user
4. eliminate the need for extra [THEN]s (see the original post)

When a word such as [begin] is executed the input state is moved
to the bottom of the data stack with the number of stacks cells
used right at the bottom. When this is needed it can be accessed
using definitions of [n>r] [nr>] [nr@] and so
on which make use of DEPTH, ROLL and PICK.

It can still do with some improvements e.g. to make nesting of
loops less restrictive as [begin] loops put stuff on the
"return" stack but basically its all there.

Adding [case] statements should be trivial.

The implementation is:

-----------------------

\ Some helpers

: ndup  ( xn ... x1 n -- xn ... x1 xn ... x1 )
   dup 0 ?do dup pick swap loop drop
;

: ndrop  ( xn ... x1 n -- )
   0 ?do drop loop
;

: dupnstrings  ( can un...ca1 u1 u -- can un...ca1 u1 u can un...ca1
u1 u )
   dup 2* 1+ ndup
;

: dropnstrings  ( can un ... ca1 u1 u -- )
   dup 2* 1+ ndrop
;

\ Look away here if you hate inefficiency!

: -roll  ( xu ... x0 u -- x0 xu ... x1 )
   dup 0
   ?do
      dup >r roll r>
   loop
   drop
;

\ Interpreter "return" stack operations

: [>r]  ( xu ... x0 -- x0 xu ... x1 )
   depth 1- -roll
;

: [r>]  ( xu ... x0 -- xu+1 ... x0 xu )
   depth 1- roll
;

: [r@]  ( xu ... x0 -- xu ... x0 xu )
   depth 1- pick
;

: [n>r]  ( i*x xu ... x0 u -- xu ... x0 i*x )
   depth 2 - tuck swap - -1
   ?do dup >r roll r> loop
   drop
;

: [nr>]  (  xu ... x0 i*x u -- i*x xu ... x0 )
   depth 2 - swap 0
   ?do dup >r roll r> loop
   drop
;

: [nr@]  (  xu ... x0 i*x u -- xu ... x0 i*x xu ... x0 )
   depth 2 - swap 0
   ?do dup >r pick r> loop
   drop
;

\ For skipping text

char Z char A - constant charZ-A
char a char A - constant chara-A

: char>lower  ( ch1 -- ch2 )
   dup [char] A - charZ-A u>
   if exit then
   chara-A +           ( -- ch2 )
;

: string>lower    ( caddr u -- )
   0
   ?do
      dup c@               ( -- caddr ch )
      char>lower over c!   ( -- caddr )
      char+
   loop
   drop
;

\ Compare string ca u with n strings, return true if any of them match

: comparestrings  ( can un ... ca1 u1 n ca u -- f )
   rot >r 0 -rot r> 0               ( -- can un ... ca1 u1 f ca u n
0 )
   ?do
      2>r -rot 2r@ compare          ( -- can un ... ca2 u2 f f1 )
      0= or 2r>                     ( -- can un ... ca2 u2 f1 ca u )
   loop
   2drop                            ( -- fn )
;

\ skip over the first of the strings given by can un ... ca1 u1
\ Note these strings must be in lower case to be recognised

: skipover  ( can un ... ca1 u1 n -- )
   begin
      begin
         bl word count 2dup string>lower  ( -- can un ... ca1 u1 n
caddr2 u2 )
         dup
      while
         2>r dupnstrings 2r>
         comparestrings                   ( -- can un ... ca1 u1 n f )
         if
            dropnstrings exit
         then
      repeat
      2drop refill 0=
   until
   cr -1 abort" End of file - string not found"
\   dropnstrings                           ( -- )
;

\ Interpreter mode control words

: [begin]    ( i*x -- n xn ... x1 i*x )
   save-input        ( i*x -- i*x xn ... x1 n )
   dup >r [n>r]      ( -- xn ... x1 i*x )
   r> [>r]           ( -- n xn ... x1 i*x )
;

: [repeat]  ( n xn ... x1 i*x -- n xn ... x1 i*x )
   [r@] 1+ [nr@] [r@] restore-input
   abort" Cannot loop back"
   drop
;

: dropinputstate  ( n xn ... x1 i*x -- i*x )
   [r@] 1+ dup >r [nr>] r> ndrop
;

: [until]  ( n xn ... x1 i*x f -- n xn ... x1 i*x | i*x )
   if
      dropinputstate
   else
      [repeat]
   then
;

: [while]  ( n xn ... x1 i*x f -- n xn ... x1 i*x | i*x )
   0=
   if
      dropinputstate
      s" [repeat]" 1 skipover
   then
;

: [do]  ( n1 n2 -- )
   [begin] 2 [n>r]
;

: exitloop  s" [+loop]" s" [loop]" 2 skipover ;

: [?do]  ( n1 n2 -- )
   2dup =
   if
      2drop
      exitloop
   else
      [do]
   then
;

-1 1 rshift invert constant minint  \ hex 80...0 ms bit set

: [+loop]  ( step -- )
   >r 2 [nr>]              ( -- lim i )
   2dup r@ + 2swap         ( -- lim i' lim i )
   swap - minint +         ( -- lim i' n1 )
   dup r@ +                ( -- lim i' n1 n2 )
   r@ 0>
   if
      < r> drop            ( -- lim i' f )
   else
      r> 0<
      if > else 2drop false then ( -- lim i' f )
   then
   if
      [repeat]
      2 [n>r]
   else
      2drop
      dropinputstate
   then
;

: [loop]  1 [+loop] ;

: [i]  2 [nr@] nip ;

\ It doesn't seem sensible to have [unloop] defined as we
\ have nowhere to exit to following an [unloop]

: [leave]
   2 [nr>] 2drop dropinputstate
   exitloop
;

\ --------------------------- end of definitions
\ Examples of use

\ See the effect of [n>r] etc
cr

-1 -2 -3 555 444 333 222 111 cr .s
5 [n>r] cr .s
5 [nr>] cr .s
3 [nr@] cr .s

depth ndrop cr

\ Display 0 to 39 with 10 integers per line using [begin] [until]

40 0
[begin]
   dup 10 mod 0=
   [if] cr [then]
   dup 3 .r
   1+ 2dup =
[until]
2drop cr

\ Display 0 to 39 with 10 integers per line using [begin] [while]
[repeat]

40 0
[begin]
  2dup >
[while]
   dup 10 mod 0=
   [if] cr [then]
   dup 3 .r
   1+
[repeat]
2drop cr

\ Display 0 to 39 with 10 integers per line using a [do] ... [loop]

40 0
[do]
   [i] dup 10 mod 0=
   [if] cr [then]
   3 .r
[loop] cr

\ Test various combinations of [do] [?do] [loop] [+loop]

100   0 [do] [i] 4 .r 10  [+loop]  cr
100   0 [?do] [i] 4 .r 10 [+loop] cr
100 100 [?do] ." never executed 1" 10 [+loop] cr
  0   0 [?do] ." never executed 2" [loop]  cr
-100  0 [do] [i] 4 .r -10 [+loop] cr
100   0 [do] [i] 4 .r  9  [+loop] cr
0  -100 [do] [i] 4 .r  9  [+loop] cr
50  -50 [do] [i] 4 .r  9  [+loop] cr
-100  0 [do] [i] 4 .r -9  [+loop] cr

\ Test [leave]

10 0
[do]
   [i] 5 >=
   [if] [leave] [then]
   [i] 4 .r
[loop]
cr

---------------------------------------------
Again it works with GForth and VFX Forth and the test
output from GForth is:

<8> -1 -2 -3 555 444 333 222 111
<8> 555 444 333 222 111 -1 -2 -3
<8> -1 -2 -3 555 444 333 222 111
<11> -3 555 444 333 222 111 -1 -2 -3

  0  1  2  3  4  5  6  7  8  9
 10 11 12 13 14 15 16 17 18 19
 20 21 22 23 24 25 26 27 28 29
 30 31 32 33 34 35 36 37 38 39

  0  1  2  3  4  5  6  7  8  9
 10 11 12 13 14 15 16 17 18 19
 20 21 22 23 24 25 26 27 28 29
 30 31 32 33 34 35 36 37 38 39

  0  1  2  3  4  5  6  7  8  9
 10 11 12 13 14 15 16 17 18 19
 20 21 22 23 24 25 26 27 28 29
 30 31 32 33 34 35 36 37 38 39
   0  10  20  30  40  50  60  70  80  90
   0  10  20  30  40  50  60  70  80  90


   0 -10 -20 -30 -40 -50 -60 -70 -80 -90-100
   0   9  18  27  36  45  54  63  72  81  90  99
-100 -91 -82 -73 -64 -55 -46 -37 -28 -19 -10  -1
 -50 -41 -32 -23 -14  -5   4  13  22  31  40  49
   0  -9 -18 -27 -36 -45 -54 -63 -72 -81 -90 -99
   0   1   2   3   4
 ok




