\ CALENDAR.FTH - Perpetual calendar for VFX Forth

((
Copyright (c) 1999
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

Original code written by:
  Christophe Lavarenne


A PERPETUAL CALENDAR
====================
by Christophe LAVARENNE

Abstact:
The calendar is a complex numbering system. Artihmetic on dates ask for their
conversion into numbers of days relative to an origin date. Here is a low cost
conversion algorithm in Forth using only integer arithmetic.

INTRODUCTION
Today's world-wide most used calendar is the Gregorian one, which reformed the
Julian calendar at the end of the XVIth century. Both are solar calendars, based
on the mean revolution period of the earth around the sun, which is nowadays
said to be around 365.2423 mean days (year and day durations are non perfectly
stables, this is why "mean" values are used).

JULIAN CALENDAR
From the name of the roman imperator Julius Caesar who set it up, this calendar
approximated the mean solar year duration in the following way :
- usual years have 365 days, spread over twelve months
- leap-years (multiples of 4) have 366 days (by adding a 29th of February)
which gives a mean of 365.25 days, involving an excess of about three fourth of
a day per century.

GREGORIAN CALENDAR
From the name of the Pope Gregory XIII who set it up to reform the Julian
calendar, this calendar improves the approximation in the following way :
- century years (multiple of 100) are not leap-years
- four-century years (multiple of 400) remain leap-years
which gives a mean of 365.2425 days, lowering the excess to about one day every
3000 years. Moreover, as the Julian calendar was 10 days late on the sun, the
Thursday 4th of October 1582 was directly followed by the Friday 15th of October
1582.

LIMITED CALENDAR
The year 2000 being a four-century one, exception to the century exception, it
remains a leap-year. Therefore, between the 1st of March 1900 and the 28th of
February 2100, the Gregorian calendar is equivalent to the Julian one, simpler.
Most computer calendars take advantage of this simplification, more especially
as the number of days that a 16 bits unsigned integer may code is less than two
centuries (179 years and 4 months).

FULL CALENDAR
In order to code more than two centuries, 32 bits integer arithmetic is needed
and the supplementary rules of the Gregorian calendar must be followed.
Nevertheless, care must be taken between the 4th and the 15th of October 1582.

COMPUTATION METHOD
The year origin is changed to the 1st of March in order to move the calendar
irregularity (the 29th of February) to the last day of each period (of 1, 4,
100 and 400 years).

Notice then that the year is divided into regular periods of 153 days over 5
months:
31 days		30 days		31 days		30 days		31 days
03:March	04:April	05:May		06:June		07:July
08:August	09:September	10:October	11:November	12:December
13:January	14:February

Numbering January=13 and February=14, the number of days between the 1st of
March and any other day in the year is obtained with the following formula
(evaluated using integer arithmetic) :

	n = (month+1)*153/5-122 + (day-1)

The number N of days between the 1st of March of the origin year (the first year
of a four-century period) and the 1st of March of any following year is obtained
with the following formula (evaluated using integer arithmetic) :

	N = 146097*F +36524*C +1461*L +365*Y
	(year = origin +400*F +100*C +4*L +Y) with (0<=C<4, 0<=L<25, 0<=Y<4)

N may be obtained more simply from the year :

	N = 365*y +y/4 -y/100 +y/400 (with y=year-origin)

The inverse conversion from (n+N) is obtained by modulo computation in five
steps, taking care of the exceptional 29th of February which adds a day
"in excess" at the period end
(1461=365*4+1 and 146097=36524*4+1, when 36524=1461*25-1) :

1) (n+N) = 146097*F+r1 (F = number of four-century periods, 0<=r1<146097)

2) if r1=146096, then 29th of February, last day of the four-century period,
   else r1=36524*C+r2 (C = number of century periods, 0<=C<4, 0<=r2<36524)

3) r2=1461*L+r3 (L = number of leap-year periods, 0<=L<25, 0<=r3<1461)

4) if r3=1460, then 29th of February, last day of the leap-year period,
   else r3=365*Y+n (Y = number of full-year periods, 0<=Y<4, 0<=n<365)

5) n=(month+1)*153/5 +(day-123)
   and year=origin +400*F +100*C +4*L +Y (+1 if month>12)

The file CALENDAR.FTH gives three perpetual calendar implementations :
- one limited  (16 bits 270 bytes) working between 2/28/1900 and 7/31/2079
- one extended (32 bits 366 bytes) working between 10/4/1582 and 2/28/32767
- one complete (32 bits 422 bytes) working between 2/28/0000 and 2/28/32767

Each define three words (with date being a unsigned integer, 16 or 32 bits):
- DATE> ( day month year -- date ) converts a compound date into an integer date
- >DATE ( date -- day month year ) converts an integer date into a compound date
- WEEKDAY ( date -- n ) converts an integer date into a code (0=Sun..6=Sat)

The following utility words are also provided
- .DATE ( date -- ) formats and prints an integer date
- FORMAT-DATE ( date -- addr len ) is a DEFERred word used by .DATE
  which converts a date into a string. See the code for more details.

A similar set of words is also provided for time handling
- TIME> ( secs mins hrs -- d-secs ) converts a compound time into an integer time
- >TIME ( d-secs -- secs mins hrs days ) converts an integer time into a compound
  time. By generating days as well, overflows when manipulating time can be
  detected and added to the date.
- .TIME ( d-secs -- ) formats and prints an integer time
- FORMAT-TIME ( d-secs -- addr len ) is a DEFERred word used by .TIME
  which converts a time into a string. See the code for more details.

Note that the default selection (Gregorian) keeps both date and time
as double numbers. This makes the word set and code portable to 16 bit
embedded systems.

Note that the Julian calendar has a restricted range on 16 bit systems,
and will not be supported for these by MPE.

For all applications that do not use BC dates, the Gregorian selection
is appropriate, and is the default selection.
))


\ *******
\ Loading
\ *******

only forth definitions

decimal

\ Three different versions are available :
\ For every day and embedded use, the Julian calendar is faster (16bits)
\ For dates before 1900 or after 2079 use the Gregorian calendar
\ For dates before 1582, use the Julian+Gregorian calendar.
\ set the relevant flag true to select

0 constant Julian?		\ Julian only calendar (for every day use)
1 constant Gregorian?		\ Gregorian only calendar (for dates before 1900)
0 constant Julian+Greporian?	\ Julian + Gregorian calendar (for historians)

\ output formats
\ weekday codes: 0=Sun 1=Mon ...  6=Sat
\ month   codes: 1=Jan 2=Feb ... 12=Dec
\ year is always output complete (0000 ... 1990 ... 32767)
\ year may be entered abreviated for XXth (0=1900 ... 99=1999)


\ *********
\ TIME&DATE
\ *********

: time		\ -- secs mins hrs
  time&date  drop 2drop
;

: date		\ -- day month year
  time&date  >r >r >r  3drop  r> r> r>
;


\ *********************************
\ Julian only calendar (using 16bits)              CLA 08/05/90
\ domain from 3/1/1900 to 7/31/2079
\ *********************************

Julian? [if]

: date>         \ day month year -- un ; see month codes
  dup 99 > if 1900 - then >r            \ year<100: XXth century
  dup 3 < if  12 + r> 1- >r  then       \ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 365 um* drop + r> 4 / +            \ and add 3/1/y - 3/1/0
;

: weekday       \ un -- n               \ see weekday codes
  4 + 0 7 um/mod drop
;

: >date         \ un -- day month year ; see month codes
  0 1461 um/mod                         \ no leap-year periods
  4 * >r                                \ convert to no years
  365 /mod                              \ no years and days
  dup 4 = if 2drop 365 3 then           \ correct if Febr.29th
  r> + >r                               \ -- days | years
  123 + dup 5 153 */                    \ compute month
  dup >r 153 5 */ - r>                  \ compute day
  1- dup 12 > if 12 - r> 1+ >r then     \ correct month and year
  r> 1900 +                             \ -- day month year
;

[then]


\ ************************************
\ Gregorian only calendar (using 32 bits)          CLA 08/05/90
\ domain from 10/15/1582 to 2/28/32767
\ ************************************

Gregorian? [if]

: date>         \ day month year -- ud ; see month codes
  dup 100 < if 1900 + then >r           \ year<100: XXth century
  dup 3 < if  12 + r> 1- >r  then       \ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 400 / + r@ 100 / - r@ 4 / +        \ leap-year corrections
  0 r> 365 um* d+                       \ and add 3/1/y - 3/1/0
;

: weekday       \ ud -- n ; see weekday codes
  2* 3 + 0 tuck d+ 7 um/mod drop        \ 65536=7n+2
;

: >date         \ ud -- day month year ; see month codes
  48699 um/mod 0 tuck 3 um/mod          \ no 4-century periods
  400 * >r  48699 um* d+                \ modulo 146097=48699*3
  36524 um/mod                          \ no of century periods
  dup 4 = if 2drop 36524 3 then         \ correct if Febr.29th
  100 * r> + >r                         \ accumulate years
  0 1461 um/mod                         \ no leap-year periods
  4 * r> + >r                           \ accumulate years
  365 /mod dup 4 = if 2drop 365 3 then  \ correct if Febr.29th
  r> + >r                               \ accumulate years
  123 + dup 5 153 */                    \ compute month
  dup >r 153 5 */ - r>                  \ compute day
  1- dup 12 > if 12 - r> 1+ >r then     \ correct month and year
  r>                                    \ -- day month year
;

[then]


\ *************************************************************
\ Julian + Gregorian calendar (using 32 bits)      CLA 08/05/90
\ domain from 3/1/0000 to 2/28/32767    (422 bytes of code)
\ 10/4/1582 +1 = 10/15/1582 (10/5/1582 to 10/14/1582 skipped)
\ *************************************************************

Julian+Greporian? [if]

: date>         \ day month year -- ud ; see month codes
  dup 100 < if 1900 + then >r           \ year<100: XXth century
  dup 3 < if  12 + r> 1- >r  then       \ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  dup r@ 228 1582 d- nip 0< not         \ after 10/4/1582 ?
  if 2+ r@ 400 / + r@ 100 / - then      \ gregorian - julian
  r@ 4 / + 0 r> 365 um* d+ ;            \ and add 3/1/y - 3/1/0

: weekday       \ ud -- n               \ see weekday codes
  2* 1+ 0 tuck d+ 7 um/mod drop ;       \ 65536=7n+2

: >date         \ ud -- day month year ; see month codes
  2dup 578043, d- nip 0<                \ if before 10/15/1582
  if
    0 >r                                \  Julian only calendar
  else
    2 s>d d-                            \ else Gregorian :
    48699 um/mod 0 tuck 3 um/mod        \  no 4-century periods
    400 * >r  48699 um* d+ 36524 um/mod \  no 1-century periods
    dup 4 = if 2drop 36524 3 then       \  correct if Febr.29th
    100 * r> + >r  0                    \  accumulate years
  then
  1461 um/mod  4 * r> + >r              \ no leap-year periods
  365 /mod dup 4 = if 2drop 365 3 then  \ correct if Febr.29th
  r> + >r                               \ accumulate years
  123 + dup 5 153 */                    \ compute month
  dup >r 153 5 */ - r>                  \ compute day
  1- dup 12 > if 12 - r> 1+ >r then     \ correct month and year
  r>                                    \ -- day month year
;

[then]


\ ****************************************************************
\ Time (of day) can also be handled in a similar way. By using
\ a double number, the code is portable to 16 bit embedded systems
\ ****************************************************************

: time>		\ secs mins hrs -- dsecs ; time to seconds
  >r
  #60 * + 0
  r> #3600 um* d+
;

: >time		\ dsecs -- secs mins hrs days
		\ allows for time calculation with day rollover
  0 >r					\ no days yet
  begin					\ clip to secs in day
    #86400, 2over d- d0<
   while
    #86400, d-				\ remove day
    r> 1+ >r				\ count day
  repeat
  #60 mu/mod  #60 um/mod		\ divide and order
  r>					\ days
;


\ **************
\ Output formats                                   CLA 08/05/90
\ **************

((  \ This code is smaller and displays directly with no buffering
\ -----
\ Dates
\ -----

: .dow		\ d-date -- ; display day of week
  date> weekday 3 *			\ compute weekday code
  c" SunMonTueWedThiFriSat"		\ weekday names
  1+ + 3 type  ." , "			\ print weekday
;

: .month	\ month -- ; display month by name
  1-  3 *				\ 3 chars per month
  c" JanFebMarAprMayJunJulAugSepOctNovDec" \ month names
  1+ + 3 type space			\ print month
;

defer .date	\ d-date -- ; format and print date

: .num-date	\ d-date -- ; display as "yyyy/mm/dd "
  base @ >r  decimal                    \ output in decimal only
  >date					\ -- day month year
  <#
    0 # # # # 2drop			\ yyyy
    ascii / hold  0 # # 2drop		\ mm
    ascii / hold  0 # # 		\ dd
  #> type space
  r> base !
;

: .eng-date	\ d-date -- ; display as "dow, dd Mon yyyy "
  base @ >r  decimal                    \ output in decimal only
  >date					\ -- day month year
  >r					\ save year
  2dup r@ .dow				\ print day of week
  swap .  .month  r> .			\ print monthday, month, year
  r> base !                             \ restore base
;

: .us-date	\ d-date -- ; display as "mm/dd/yyyy "
  base @ >r  decimal                    \ output in decimal only
  >date					\ -- day month year
  >r					\ save year
  2 .r ." /"				\ month
  2 .r ." /"				\ day
  r> 4 .r space				\ year
  r> base !                             \ restore base
;

assign .eng-date to-do .date


\ -----
\ Times
\ -----

defer .time	\ dsecs --

: .hh:mm:sss	\ dsecs --
  >time drop
  0 .r  ." :"  0 .r  ." :"  .
;

assign .hh:mm:ss to-do .time
))

\ ---------
\ Utilities
\ ---------

also system definitions

#32 buffer: datebuff

: +frag		\ addr len -- ; add fragment to buffer
  tuck					\ preserve len
  datebuff count + swap cmove		\ add new
  datebuff c+!				\ update length
;


\ -----
\ Dates - the date MUST be a double number; do not select Julian
\ -----

: [date		\ d-date -- base day month year
  base @ -rot
  >date  0 datebuff c!
;

: date]		\ base -- addr len
  base !
  datebuff count
;

: +bl		\ -- ; add " " to output
  S"  " +frag
;

: +/		\ -- ; add "/" to output
  S" /" +frag
;

: +,		\ -- ; add "," to output
  S" ," +frag
;

: +year#	\ yr --
  0 <# # # # # #> +frag
;

: +month#	\ month --
  0 <# # # #> +frag
;

: +month$	\ month -- ; add month as text
  1-  3 *				\ 3 chars per month
  c" JanFebMarAprMayJunJulAugSepOctNovDec" \ month names
  1+ + 3 +frag				\ add month
;

: +day#		\ day --
  0 <# # # #> +frag
;

: +dow$		\ day month year -- ; add day of week
  date> weekday 3 *			\ compute weekday code
  c" SunMonTueWedThuFriSat"		\ weekday names
  1+ + 3 +frag				\ add weekday
;

also forth definitions

defer format-date	\ d-date -- addr len ; format date

: .date		\ d-date --
  format-date type space
;

: num-date	\ d-date -- addr len ; display as "yyyy/mm/dd "
  [date					\ -- day month year
    +year# +/ +month# +/ +day#		\ yyyy/mm/dd
  date]
;

: eng-date	\ d-date -- addr len ; display as "dow, dd Mon yyyy "
  [date					\ -- day month year
    >r					\ save year
    2dup r@ +dow$ +, +bl		\ day of week
    swap +day# +bl +month$ +bl r> +year#  \ day, month, year
  date]
;

: us-date	\ d-date -- ; display as "mm/dd/yyyy "
  [date					\ -- day month year
    -rot +month# +/ +day# +/ +year#
  date]
;

assign eng-date to-do format-date

previous previous definitions


\ -----
\ Times
\ -----

also system definitions

: [time		\ d-secs -- base secs mins hrs
  base @ -rot
  >time drop  0 datebuff c!
;

: time]		\ base -- addr len
  base !
  datebuff count
;

: +n2		\ n -- ; add a 2 digit number
  0 <# # # #> +frag
;

: +:		\ -- ; add ":" to output
  S" :" +frag
;

also forth definitions

defer format-time	\ dsecs -- addr len

: .time		\ dsecs --
  format-time type space
;

: pc-time	\ dsecs -- addr len
  [time					\ -- secs mins hours
    +n2 +: +n2 +: +n2
  time]
;

assign pc-time to-do format-time

previous previous definitions

