(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : iForth Environments
 * DESCRIPTION : Adaptive speech filter using least mean square adaption
 * CATEGORY    : Sound utilities
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-08-01 17:46:27 better algoritms results in a >10% performance improvement
		: 2009-08-07 00:13:08 sse2 optimized kernels included
		: 2009-08-22 01:13:47 all filters sizes are allowed and supported by the filter kernel
		: 2009-12-26 17:35:49 added support for freezing adp filter coefficients
		: smallest adaptive filter is 32
		: timing offset is always ~1ms
 *)

\ *S The Adaptive Filter object
JACKAUDIO DEFINITIONS

HERE ," Adaptive Filter" CONSTANT #adp-id

#PRIVATE-DSP
	DSP-VALUE adp_data
	DSP-VALUE adp_muls
	DSP-VALUE adp_frozen?
	DSP-FVALUE adp_speed
	CONSTANT #adp_struct
: CHK-ADP	#adp-id plug_type <> ABORT" Invalid adaptive filter object" ;

: calc-adp-val	\ ( -- ) (F val -- val' )
	\ write new data ahead in the circular buffer
	adp_muls
	FDUP	#FJACK-SAMPLE NEGATE JACK-SAMPLE-RATE #10 RSHIFT - #FIR_BUFF 1- AND DFLOATS adp_data + DUP DF! FDUP #FIR_BUFF DFLOATS + DF!			\ (F Xr )
	FDUP	#FJACK-SAMPLE NEGATE JACK-SAMPLE-RATE #10 RSHIFT - 1- #FIR_BUFF 1- AND #FIR_BUFF 2* + DFLOATS adp_data + DUP DF! FDUP #FIR_BUFF DFLOATS + DF!	\ (F Xr )
	#FJACK-SAMPLE NEGATE #FIR_BUFF 1- AND DFLOATS adp_data +  DUP $08 AND IF #FIR_BUFF 2* 1- DFLOATS + THEN
	2DUP	\ both loops need the multipliers and data fields
	plug_taps (calc-fir-val)	EXECUTE F-			\ (F new )
	adp_frozen?
	IF	2DROP
	ELSE	FDUP adp_speed F* plug_taps S>F F/ 		\ (F r 2*r)
		plug_taps (calc-adp-muls) EXECUTE
	THEN ;
: (fill-adaptive-values)
	adp_data #FIR_BUFF 5 * DFLOATS ERASE ;

: (init-adaptive-filter)
	0.001e0 TO adp_speed
	plug_dynmem $10 #B-ALIGNED TO adp_data
	adp_data #FIR_BUFF 4 * DFLOATS + TO adp_muls
	FALSE TO adp_frozen?
	(fill-adaptive-values) ;
: (set-adp-quality)	( n -- )
	4 MAX #FIR_BUFF MIN $FFFFE AND DUP TO plug_taps 2/ JACK-SAMPLE-RATE #10 RSHIFT + TO plug_timeshift
	FALSE TO adp_frozen?
	(fill-adaptive-values) ;

FORTH DEFINITIONS
: MAKE-ADAPTIVE-FILTER	( size -- object )
\ *G Constructs an adaptive filter object and leave it's address, the 'size'
\ ** is the number of taps corrected to even.
\ ** Adaptive filters add a signal delay by half the number of taps.
\ ** This filter uses a simple 'least mean square' algorithm.
\ *P You may select a different number of taps later via *\forth{SET-OBJ-QUALITY ( taps object -- )}.
	['] (init-adaptive-filter)
	#FIR_BUFF DFLOATS 5 * $20 +
	#adp_struct MAKE&ALLOCATE-PLUGIN-OBJECT
	#32 MAX #FIR_BUFF MIN $FFFFFE AND DUP TO plug_taps 2/ JACK-SAMPLE-RATE #10 RSHIFT + TO plug_timeshift
	#adp-id TO plug_type  ADAPTIVE TO plug_filtype
	['] (set-adp-quality) TO plug_xt_quality
	['] calc-adp-val TO plug_xt @DSP ;
: ADAPTIVE-FILTER:	( n -- ) \ name
\ *G Defines a named adaptive plugin object, calling 'name' returns the object address.
	 MAKE-ADAPTIVE-FILTER CONSTANT ;
: SET-FREEZE-ADAPTIVE	( flag object -- )
\ *G You can freeze the adaptive filter coefficients - this is usefull when you have 'tuned' to a noise signal
\ ** and want to stay with this filter. *\forth{FALSE} will continue adapting otherwise it will be frozen.
\ ** As changing the filter qaulity means other coefficients *\forth{SET-OBJ-QUALITY} for an adaptive filter
\ ** always unfreezes it.
	-->DSP CHK-ADP 0<> TO adp_frozen? ;
: SET-ADAPTIVE-SPEED	( n object -- )
\ *G Sets the speed of the adapting algorithm, 0-100 are accepted.
	-->DSP CHK-ADP #101 SWAP - 1 MAX #101 MIN S>F 4e0 F/ 1/F TO adp_speed (fill-adaptive-values) ;
