(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Forth Environments
 * DESCRIPTION : Sinus wave generator for jack
 * CATEGORY    : Sound processing
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-04-28 07:53:51
 *)

\ *S The Sinus Tone Oszillator Object, a detailed example for writing plugin objects
\ *P When writing more plugins read this and the *\forth{types.frt} carefully, also you
\ ** can use the *\forth{prototype.frt} as a starting point.

\ *P Words that are used inside the object should be defined in the *\forth{JACKAUDIO} vocabulary.
\ *[
ALSO JACKAUDIO ALSO DEFINITIONS
\ *]

\ *P Now a pointer to a counted text should be defined, later this is set to the *\forth{plug_type}
\ ** object-element; it's there for a) making the plug_type defined and b) the *\forth{LIST-PLUGINS}
\ ** and *\forth{LIST-ALL-PLUGINS} tools gives more detailed information.
\ *[
HERE ," SINUS oscillator" CONSTANT #sin-id
#PRIVATE-DSP
	DSP-FVALUE sin_phase
	DSP-FVALUE sin_step
	CONSTANT #sinstruct
\ *]
\ *P Then the object elements are defined and the size of the whole object struct is made a *\forth{CONSTANT}

: CHK-SIN	#sin-id plug_type <> ABORT" Invalid SINUS-TONE object" ;
\ *G Check for a correct plug_type and aborts when this not ok

\ *P Each object type MUST have a specified signal processing funtion; the execution token of this function is
\ ** stored in the *\forth{plug_xt} element. The funcion MUST take one *\forth{FLOAT} as a parameter and MUST
\ ** leave one *\forth{FLOAT} as a result. So the depth of the floating stack must never be changed.
\ *[
: calc_sin_val	\ ( -- ) (F val -- val' )
	sin_step sin_phase F+ REDUCE.2PI FDUP TO sin_phase FSIN plug_level F* F+ ;
\ *]

\ *P Before you start writing your own audio-data-processing code remember that
\ *B the code is executed in a JACK demon callback
\ *B the code MUST exit with a correct stack and floating stack
\ *B the code MUST NEVER do anything that might give larger delays than a few usecs
\ *B the code MUST NEVER do text/graphics output, file writing, use semaphores
\ *B the JACK demon will halt the client as 'zombified' when delay is too large
\ *P Whenever you want to do things like that: FORGET ABOUT THEM in the process function.
\ ** What you could and should do: Write thread-safe buffering functions!
\ *P You may watch the jack performance with two included tools, *\forth{JACK-LOAD} and *\forth{JACK-XRUNS}.

\ *P Many - as this example plugin - can be tuned to a frequency; the io_procs module implements
\ ** a generic tuning word *\forth{SET-PLUGIN-FREQUENCY} which calls a function defined in the object
\ ** making phase.
\ *[
: (tune-sine-object)	( frequency -- )
	FDUP TO plug_frequency PI F* F2* JACK-SAMPLE-RATE S>F F/ TO sin_step ;
\ *]

\ *P *\bold{Every} object definer needs a function that initialises the plugin object when it is defined
\ ** and also in the turnkey bootstrap word *\forth{INIT-ALL-JACK-OBJECTS}. If - in some cases - there is
\ ** no such function needed *\forth{MAKE&ALLOCATE-PLUGIN-OBJECT} must be given 0.
\ *[
: (init-sine-object)
	 ; \ nothing to do; no allocation
\ *]

\ *P Now you can make an object-definer, as this should be public you should switch to *\forth{FORTH} before.

\ *[
FORTH DEFINITIONS
: MAKE-SINE-OSCILLATOR	( -- addr )	\ Create a sinus oszillator struct
	\ Make an object and make it the 'current' object
	['] (init-sine-object) 0 #sinstruct MAKE&ALLOCATE-PLUGIN-OBJECT
	\ Set the correct plug_type and set a plug_filtype mode
	#sin-id TO plug_type  SIN-TONE TO plug_filtype
	\ set the execution token of the signal processing function
	['] calc_sin_val TO plug_xt
	\ set the tuning function
	['] (tune-sine-object) TO plug_xt_tune
	\ these data can be permanent
	0e0 TO sin_step  0e0 TO sin_phase
	\ leave the 'current' object on the stack
	@DSP ;

\ A more simple definer makes the object and defines it as a constant
: SINE-OSCILLATOR:	( -- ) \ name
        MAKE-SINE-OSCILLATOR CONSTANT ;


PREVIOUS PREVIOUS DEFINITIONS
\ *]
