(*
 * LANGUAGE    | ANS Forth
 * PROJECT     | Audio processing
 * DESCRIPTION | FFT plugin for fjack using fftw3
 * CATEGORY    | Tool
 * AUTHOR      | Hanno Schwalm
 * LAST CHANGE | 2009-06-15 22:14:00, Hanno Schwalm
		2009-12-08 22:25:46 bypassing ports for speed
		2009-12-12 01:52:36 added a delay for analysis to compensate for filter times.
 *)
\ *S The FFT Analyser Plugin
\ *P Quite often you will have to do a signal analysis of your audio signals somewhere
\ ** in the input->output processing chain. You can define such an analyser plugin with
\ ** *\forth{MAKE-FFT-PORT} and can always do an ad-hoc real time analysis of the signal.
\ *P This plugin does an fft analysis, avaraging of the spectrum, finds maxima and the
\ ** average level in a defined frequency range. In the background this uses the fftw3 library.
JACKAUDIO DEFINITIONS

HERE ," FFT plugin" CONSTANT #fft-id
#PRIVATE-DSP
	DSP-VALUE fft_vector
	DSP-VALUE fft_intlist
	DSP-VALUE fft_maxlist
	DSP-VALUE fft_avlist
	DSP-VALUE fft_avcnt
	DSP-VALUE fft_bigmax
	DSP-VALUE fft_analow
	DSP-VALUE fft_anahigh
	DSP-VALUE fft_size
	DSP-VALUE fft_shift
	DSP-FVALUE fft_bigmaxgain
	DSP-FVALUE fft_avgain
	CONSTANT #fft_struct
8 CONSTANT #max-average
$4000 CONSTANT #maxpoints
: CHK-FFT		plug_type #fft-id <> ABORT" Invalid FFT-PORT object" ;

FORTH DEFINITIONS
: FFT-PORT-SIZE		( fftport -- size )
\ *G Tells the size of the port.
	-->DSP CHK-FFT fft_size ;
: FFT-HAMMING		( mode fftport )
\ *G Changes the 'windowing' function of the analyser port; 0 means no windowing, 1 is hamming and 2 hann.
	-->DSP CHK-FFT
 	CASE	0 OF	fft_vector WINDOWING-OFF	ENDOF
 		1 OF	fft_vector HAMMING-WINDOW	ENDOF
 		2 OF	fft_vector HANN-WINDOW		ENDOF
 	ENDCASE ;
: FFT-ANARANGE		( low-frequency high-frequency fftport -- )
\ *G Sets the frequency range in which the average signal level is calculated.
	-->DSP CHK-FFT 2DUP > IF SWAP THEN
	fft_size JACK-SAMPLE-RATE */ fft_size 2/ 3 - MIN  TO fft_anahigh
	fft_size JACK-SAMPLE-RATE */ 3 MAX TO fft_analow
	fft_analow fft_anahigh >= IF 3 TO fft_analow fft_size 2/ 3 - TO fft_anahigh THEN ;
: FFT-GET-ANARANGE	( fftport -- low-frequency-index high-frequency-index )
\ *G Gets the frequency range for average analysis.
	-->DSP CHK-FFT fft_analow fft_anahigh ;

JACKAUDIO DEFINITIONS
: @FFT-dB	( struct idx ) ( F: -- dB )
	OVER #FFT-SIZE S>F 1/F
	>FFT-COMPLEXDATA DF@+ FSQR DF@ FSQR F+ FSQRT  F* 1e-100 FMAX FLOG 20e0 F* ;

: (fft_analysis)	( -- )
	-100000e0 TO fft_bigmaxgain  0 TO fft_bigmax
	fft_size 2/ 0
	DO	fft_vector I @FFT-dB
		I #max-average * fft_avcnt #max-average 1- AND + SFLOATS fft_avlist + SF!
		fft_avlist I #max-average * SFLOATS +	( addr )
		0e0 #max-average 0 DO SF@+ F+ LOOP DROP #max-average S>F F/ fft_intlist I SFLOATS + SF!
	LOOP
	fft_size 2/ 4 - 4
	DO	I SFLOATS fft_intlist + DUP SF@	TRUE		\ ( addr flag )	( F centre )
		FDUP OVER   -1 SFLOATS + SF@       F> AND 	\ ( addr flag	( F centre )
 		FDUP OVER   1 SFLOATS + SF@        F> AND
  		FDUP OVER  -2 SFLOATS + SF@ 3e0 F+ F> AND
  		FDUP OVER   2 SFLOATS + SF@ 3e0 F+ F> AND
  		FDUP OVER  -3 SFLOATS + SF@ 7e0 F+ F> AND
  		FDUP OVER   3 SFLOATS + SF@ 7e0 F+ F> AND
  		FDUP OVER  -4 SFLOATS + SF@ 10e0 F+ F> AND
  		FDUP OVER   4 SFLOATS + SF@ 10e0 F+ F> AND
		0<> NIP FDROP ( flag ) I fft_maxlist + C!
	LOOP
	0e0 fft_anahigh fft_analow
	DO	fft_intlist I SFLOATS + SF@ F+	( average )
		fft_maxlist I + C@	\ maximum found?
		IF	fft_intlist I SFLOATS + SF@  FDUP fft_bigmaxgain F>
			IF TO fft_bigmaxgain  I TO fft_bigmax ELSE FDROP THEN
		THEN
	LOOP
	fft_anahigh fft_analow - S>F F/ TO fft_avgain ;
: (init-fft-filter)
	0 TO fft_avcnt  #fft-id TO plug_type 1 TO fft_shift
	3 TO fft_analow $FF0 TO fft_anahigh
	plug_dynmem TO fft_maxlist
	fft_maxlist #maxpoints CHARS + #64 #B-ALIGNED TO fft_intlist
	fft_intlist #maxpoints 2/ SFLOATS + TO fft_avlist
	#maxpoints 2/ 0 DO 0e0 I SFLOATS fft_intlist + SF! LOOP
	#maxpoints 0 DO 0 I CHARS fft_maxlist + C! LOOP
	#maxpoints 2/ #max-average * 0 DO 0e0 I SFLOATS fft_avlist + SF! LOOP ;
FORTH DEFINITIONS

: MAKE-FFT-PORT	( size -- fftport )
\ *G Define a fft-port, this can be used as a plugin in the processing chain. For the iForth system
\ ** there is a fft viewing module available *\forth{FFT-VIEWPORT:} that works perfectly with fft-ports.
	['] (init-fft-filter)
	#maxpoints 2/ SFLOATS
	#maxpoints 2/ #max-average * SFLOATS +
	#maxpoints CHARS +	$200 +
	#fft_struct MAKE&ALLOCATE-PLUGIN-OBJECT
	DUP TO fft_size FFT-STRUCT TO fft_vector
	TRUE TO plug_bypassed
	0 ( ['] NOOP ) TO plug_xt
	@DSP ;
\ *P *\image{"The FFT viewer" "fftplug" "png" 25 25}

: FFT-PORT:	\ ( size -- ) name
\ *G Defines a named fft-port.
	MAKE-FFT-PORT	CONSTANT ;
: FFT-ANALYSIS	( fftport -- )
\ *G Does an fft analysis of the last audio data in the stream plus avaraging and maxima finding.
	-->DSP CHK-FFT
	#frames-cnt fft_size - fft_shift ABS - 1- LOCALS| pos |
	fft_vector @	\ target
	fft_size 0
	DO	plug_output pos I + @IDX-HISTORY-DATA DF!+
	LOOP DROP
	1 +TO fft_avcnt
	fft_vector CALCULATE-FFT
	(fft_analysis) ;

: @FFT-dB-LAST		\ ( fftport idx -- ) (F -- level-dB )
\ *G gets the average level for the last 8 *\forth{FFT-ANALYSIS} for this port.
	SWAP -->DSP CHK-FFT #max-average * fft_avcnt #max-average 1- AND + SFLOATS fft_avlist + SF@ ;
: @FFT-dB-AV		\ ( fftport idx -- ) (F -- average-dB )
\ *G gets the average level for the last 8 *\forth{FFT-ANALYSIS} for this port.
	SWAP -->DSP CHK-FFT SFLOATS fft_intlist + SF@ ;
: FFT-MAXIMUM?		\ ( fftport idx -- flag )
\ *G returns *\forth{TRUE} when a maximum is detected at this point.
	SWAP -->DSP CHK-FFT fft_maxlist + C@ 0<> ;
: FFT-GREATEST-MAXIMUM	\ ( fftport -- index ) (F -- level )
\ *G Finds the greatest maximum for the port and returns the index.
	-->DSP CHK-FFT fft_bigmax fft_bigmaxgain ;
: FIND-AVERAGE-LEVEL	\ ( fftport -- ) (F -- avlevel )
\ *G returns the average signal level in the specified frequency range measured in dB
	-->DSP CHK-FFT fft_avgain ;
: FFT-IDX>FREQU		\ ( index fftport -- frequency )
\ *G Converts index for this port to frequency.
	-->DSP CHK-FFT JACK-SAMPLE-RATE fft_size */ ;
: FFT-FREQU>IDX		\ ( frequency fftport -- idx fftport )
\ *G Converts frequency for this port to index, the port kept on the stack as this is most often used later.
	-->DSP CHK-FFT fft_size JACK-SAMPLE-RATE */ @DSP ;
: SET-FFT-DELAY		( delay fftport -- )
\ *G The fft analyser allows a delay in the history to compensate in fft displays.
	-->DSP 1 MAX #HISTORY-SIZE 2/ MIN TO fft_shift ;
