(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Forth Environments
 * DESCRIPTION : FFTW3 frontend
 * CATEGORY    : Utility
 * AUTHOR      : Hanno Schwalm; 2009-06-15 22:13:47
 * change	2009-07-29 19:40:09; removed explicite access of data - address calculation instead, inverse fft added
		2009-08-10 12:53:10; safe AtCold
		2009-08-26 20:20:12; checking for an existing plan before fft execution, library binding extracted
		2009-12-07 18:30:40; added IMPORT-WISDOM - working fine

*)
\ *S The FFTW3 library interface
\ *P This library can be tuned for optimized performance, this implementation currently supports the import of
\ ** system wisdom files - see the fftw3 docs how to generate such a wisdom file.
\ *P Without such wisdom the best algorithm is just estimated, you can change this to one of the other methods in
\ ** fftw3lib.frt.
JACKAUDIO DEFINITIONS

0 VALUE LINK-FFT-STRUCTS
0 VALUE LINK-INVERSE-FFT-STRUCTS
: MAKE-FFTW-PLAN	( struct -- )
	FFTW3-LIB?
	IF	DUP 1 CELLS + @ OVER
		@ DUP #FFTW_PLAN_SPEED fftw_plan_dft_r2c_1d
	ELSE	0
	THEN	SWAP 2 CELLS + ! ;
: MAKE-INVERSE-FFTW-PLAN	( struct -- )
	FFTW3-LIB?
	IF	DUP 1 CELLS + @ OVER
		@ DUP #FFTW_PLAN_SPEED fftw_plan_dft_c2r_1d
	ELSE	0
	THEN	SWAP 2 CELLS + ! ;
: FREE-PLAN-STRUCT	( struct -- )
	DUP	2 CELLS + @ ?DUP IF fftw_destroy_plan	0 OVER 2 CELLS + ! THEN
		6 CELLS + @ ?DUP IF FREE DROP		0 OVER 6 CELLS + ! THEN	;

: FREE-FFTW3-PLANS	( addr -- )
	FFTW3-LIB? 0= IF DROP EXIT THEN
	BEGIN	DUP
	WHILE	DUP FREE-PLAN-STRUCT 3 CELLS + @
	REPEAT	DROP ;
: FREE-ALL-FFT-PLANS	( -- ) \ forget all ffts
	STOP-JACK-CLIENT
	LINK-FFT-STRUCTS		FREE-FFTW3-PLANS
	LINK-INVERSE-FFT-STRUCTS	FREE-FFTW3-PLANS ; ' FREE-ALL-FFT-PLANS AtExit
: GET-FFT-DYNMEM	( struct -- )
	DUP 1 CELLS + @  DUP DFLOATS SWAP SFLOATS + $200 + ALLOCATE ?ALLOCATE	( struct addr )
	2DUP SWAP 6 CELLS + !
	$10 #B-ALIGNED							( struct 'addr )
	2DUP SWAP 4 CELLS + !	\ window array
	OVER 1 CELLS + @ 2 + SFLOATS + #64 #B-ALIGNED SWAP ! ;

:NONAME	( -- ) \ reload all fft structs
	LINK-FFT-STRUCTS
	BEGIN	DUP
	WHILE	DUP GET-FFT-DYNMEM
		DUP 5 CELLS + OFF								\ windowing off
		DUP MAKE-FFTW-PLAN
		3 CELLS + @
	REPEAT	DROP
	LINK-INVERSE-FFT-STRUCTS
	BEGIN	DUP
	WHILE	DUP GET-FFT-DYNMEM
		DUP 5 CELLS + OFF								\ windowing off
		DUP MAKE-INVERSE-FFTW-PLAN
		3 CELLS + @
	REPEAT	DROP ; AtCold
: DO-INPUT-WINDOWING	( struct -- )
	DUP @ SWAP					\ data struct
	DUP 4 CELLS + @ SWAP 1 CELLS + @		\ data window size
	0 DO SF@+ SWAP DUP DF@ F* DF!+ SWAP LOOP 2DROP ;

FORTH DEFINITIONS
: CHECK-FFTW-SIZE	( n -- ok-mask )			\ sizes in fftw may be multiples of primes 2-13
\ *G Checks the desired size of a fftw3 plan. Principally all sizes are suported but many ready-compiled
\ ** libraries only support a limited number range. For all
\ *P n = 2*a * 3^b * 5^c * 7^d * 11^e * 13^f
\ *P the fftw3 library has algorithms available.
\ ** The 'ok' is a bit-mask about a-f being >= 1, for probably impossible sizes %000000 is returned
\ *C %000001-a, %000010-b, %000100-c, %001000-d, %010000-e, %100000-f
\ *C %000001 has best algorithms
\ *C %001110 still is good
\ *C %110000 has existing algorithms in probably all compiled libraries but will be slower
	DUP 4 < IF DROP 0 EXIT THEN  0 SWAP
	BEGIN	DUP 2 MOD 0= WHILE 2 /		SWAP $01 OR SWAP REPEAT
	BEGIN	DUP 3 MOD 0= WHILE 3 /		SWAP $02 OR SWAP REPEAT
	BEGIN	DUP 5 MOD 0= WHILE 5 /		SWAP $04 OR SWAP REPEAT
	BEGIN	DUP 7 MOD 0= WHILE 7 /		SWAP $08 OR SWAP REPEAT
	BEGIN	DUP #11 MOD 0= WHILE #11 /	SWAP $10 OR SWAP REPEAT
	BEGIN	DUP #13 MOD 0= WHILE #13 /	SWAP $20 OR SWAP REPEAT
	1 <> IF DROP 0 THEN ;

: WINDOWING-OFF	( struct )
\ *G The input parameters for the signal analysis may be 'windowed', this switches windowing off.
	5 CELLS + OFF ;
: HAMMING-WINDOW ( struct )
\ *G The input parameters for the signal analysis may be 'windowed', this switches to 'hamming'
	DUP 5 CELLS + ON  DUP 1 CELLS + @ 0
	DO	DUP 7 CELLS + SF@  I S>F F* FCOS -0.46e0 F* 0.54e0 F+ 1.85e0 F* DUP 4 CELLS + @ I SFLOATS + SF!
	LOOP	DROP ;
: HANN-WINDOW	( struct )
\ *G The input parameters for the signal analysis may be 'windowed', this switches to 'hann'
	DUP 5 CELLS + ON  DUP 1 CELLS + @ 0
	DO	DUP 7 CELLS + SF@  I S>F F* FCOS -0.50e0 F* 0.50e0 F+ 2.00e0 F* DUP 4 CELLS + @ I SFLOATS + SF!
	LOOP	DROP ;


\ 0 address of data	\ 1 size	\ 2 plan	\ 3 Link	\ 4 windowing buffer	\ 5 windowing on?
\ 6 allocated memory	\ 7 sfloat: 2pi/n -- used for windowing
: FFT-STRUCT	( n -- struct )
\ *G When you want to to signal analysis using the fftw3 library you must define a forth struct that
\ ** defines the interface to the fftw3 library. 'n' is the number of samples to be analysed, *\forth{FFT-STRUCT}
\ ** defines the struct and returns a pointer to it. This struct pointer will be used by the next words.
\ *P Internally there are arrays for real parameter data and complex results at place,
\ ** a fftw-plan is calculated.
\ ** See *\forth{CHECK-FFTW-SIZE} for a discussion about possible sizes.
	DUP CHECK-FFTW-SIZE 0= ABORT" Wrong size in libfftw3"
	ALIGN HERE >R 7 CELLS 1 SFLOATS + ALLOT
	DUP R@ 1 CELLS + !
	R@ GET-FFT-DYNMEM
	PI F2* ( size ) S>F F/ R@ 7 CELLS + SF!
	LINK-FFT-STRUCTS R@ 3 CELLS + !  R@ TO LINK-FFT-STRUCTS
	R> DUP MAKE-FFTW-PLAN DUP HANN-WINDOW ;
\ 0 address of data	\ 1 size	\ 2 plan	\ 3 Link	\ 4 windowing buffer	\ 5 windowing on?
\ 6 allocated memory	\ 7 sfloat: 2pi/n -- used for windowing
: INVERSE-FFT-STRUCT	( n -- struct )
\ *G When you want to to signal analysis using the fftw3 library you must define a forth struct that
\ ** defines the interface to the fftw3 library. 'n' is the number of samples to be analysed, *\forth{INVERSE-FFT-STRUCT}
\ ** defines the struct and returns a pointer to it. This struct pointer will be used by the next words.
\ *P Internally there are arrays for real parameter data and complex results at place,
\ ** a fftw-plan is calculated.
\ ** See *\forth{CHECK-FFTW-SIZE} for a discussion about possible sizes.
	DUP CHECK-FFTW-SIZE 0= ABORT" Wrong size in libfftw3"
	ALIGN HERE >R 7 CELLS 1 SFLOATS + ALLOT
	DUP R@ 1 CELLS + !
	R@ GET-FFT-DYNMEM
	PI F2* ( size ) S>F F/ R@ 7 CELLS + SF!
	LINK-INVERSE-FFT-STRUCTS R@ 3 CELLS + !  R@ TO LINK-INVERSE-FFT-STRUCTS
	R> DUP MAKE-INVERSE-FFTW-PLAN ;

: #FFT-SIZE	( struct -- #elements )
\ *G tells the number of samples for this struct.
	1 CELLS + @ ;

: >FFT-COMPLEXDATA	( struct idx -- addr )
\ *G Calculates the address in a *\forth{FFT-STRUCT} result complex datastructure or *\forth{INVERSE-FFT-STRUCT}
\ ** input parameter complex datastructure.
\ *P All FFT data are of *\forth{DFLOAT} type and so must be accessed with *\forth{DF@ DF!} and friends. In complex
\ ** number arrays the real-part is always at the lower address.
	2* DFLOATS SWAP @ + ;

: >FFT-DATA	( struct idx -- addr )
\ *G Calculates the address in a *\forth{FFT-STRUCT} input parameter datastructure or *\forth{INVERSE-FFT-STRUCT}
\ ** result datastructure.
	DFLOATS SWAP @ + ;


\ : !FFT-REAL	>FFT-DATA DF! ;
\ : @FFT-REAL	>FFT-COMPLEXDATA DF@ ;

: CALCULATE-FFT	( struct )
\ *G After storing the input data into the *\forth{FFT-STRUCT} parameter array you can do the signal analysis with *\forth{CALCULATE-FFT}.
	DUP 5 CELLS + @ IF DUP DO-INPUT-WINDOWING THEN 2 CELLS + @
	DUP 0= ABORT" needs fftw3 lib to CALCULATE-FFT"
	fftw_execute ;

: CALCULATE-INVERSE-FFT	( struct )
\ *G After storing the input data into the *\forth{INVERSE-FFT-STRUCT} complex parameter array you can do the signal
\ analysis with *\forth{CALCULATE-FFT}.
	2 CELLS + @
	DUP 0= ABORT" needs fftw3 lib to CALCULATE-INVERSE-FFT"
	fftw_execute ;
