(*
 * LANGUAGE    : ANS Forth
 * PROJECT     : Audio processing basics for in->out processing
 * DESCRIPTION : handling sound input/output for jack
 * CATEGORY    : Sound Utilities
 * AUTHOR      : Hanno Schwalm, 2009-04-28 07:56:59
		2009-05-09 00:16:03 plug_linking added for information about all defined objects
				listing of plugins changed
				fixes for object setting & removing jack plugins
		2009-06-03 01:58:53; AtCold hooks for turnkeys
		2009-06-05 20:00:32; Plugins all use dynamic memory; new MAKE&ALLOCATE-PLUGIN-OBJECT
		2009-08-01 11:34:29; added support for bandpass and bandstop filters
		2009-08-17 21:00:54; (.pluginfo) fixed
		2009-11-28 00:49:59; changed to new method of chaining, pluginfo fixed for chaining support
		2009-12-01 19:05:45; added bypassing support and reconnections
		2009-12-01 19:06:51; added some planned plugin types
		2009-12-18 06:57:29; added debugging information and also allowed some extra memory for dynmem
		2009-12-23 01:06:11; added support for calculation of phase shift; fir, iir and adaptives are supported to far
		2009-12-26 20:06:33; SET-PLUGIN-BYPASS instead of switching
		2009-12-29 17:57:24; added EFFECTS-FIR
		2009-12-29 19:23:48; taps and bypass information
		2010-01-21 19:01:48; added SET-OBJ-FLOAT-FREQUENCY
		2010-01-26 23:56:38; object removing for iForth to allow complete removal of apps,
					also changes in main - JACK-TMP-MEM is used as a flag
*)

\ *T The fJACK Input->Output Processing
\ *P You know that you can put audio processing plugin objects in any one of the slots, this module takes care of it.
\ ** The order of processing in the jack-callback is like this:
\ *P *\image{"Processing order" "order" "png" 50 50}
\ *P In effect all plugins can assume, that data in a lower slot number have been completely processed already.

\ *S Defining the plugin objects
\ *P A plugin object is a data struct, some parts of this struct are the same in all plugin objects - see the
\ ** client.frt source for details. These public parts are used by the plugin manager to take care about
\ ** safety, setting parameters like frequency, quality and more.
\ *P Also the structs are 'connected' to other plugin objects or the input/output jack ports via
\ ** plug_input and plug_output. basically a plugin object reads samples from the plug_input history buffer,
\ ** does it's specific processing and writes data to the plug_input history buffer - in the graphics it's the
\ ** red arrows. Also you may read data from other history buffers by words like *\forth{@HISTORY-DATA} and friends,
\ ** see the green arrow and docs in the client manual.
\ *P *\image{"Plugin objects" "plugin" "png" 50 50}


DECIMAL JACKAUDIO DEFINITIONS

\ *P The fjack system knows some predefined plugin object types, when you write other plugins
\ ** use numbers above #256 for your private applications or let me know to include them in here.
\ *[
$01 CONSTANT LOW-PASS
$02 CONSTANT HIGH-PASS
$03 CONSTANT BAND-PASS
$04 CONSTANT BAND-STOP
$05 CONSTANT ADAPTIVE
$06 CONSTANT DECODER
$07 CONSTANT SIN-TONE
$08 CONSTANT DELAYER
$09 CONSTANT AUTOLEVEL
$0A CONSTANT USER-FIR
$0B CONSTANT MIXER-PLUGIN
$0C CONSTANT LESLIE-PLUGIN
$0D CONSTANT PHASER-PLUGIN
$0E CONSTANT ECHO&HALL-PLUGIN
$0F CONSTANT VIBRATO-PLUGIN
$10 CONSTANT EQUALIZER
$11 CONSTANT EFFECTS-FIR
\ *]

HERE ," undefined object" CONSTANT #no-id 	\ This defines a string pointer used as a plug_type

\ CONSTANT #PRIVATE-DSP	( -- n )
\ *G All plugin objects have some data in common, the public part of the objects struct is used
\ ** by the plugin handler.
\ *P The common part of a plugin object holds predefined *\forth{DSP-VALUE}s, so when you are making a
\ ** plugin object definer you must leave this public part in peace. So when defining the struct you should 'start with *\forth{#PRIVATE-DSP}.

0 VALUE PLUGIN-LINK

: SET-OBJECT-COLD-VALUES
	0 TO plug_channel
	0 TO plug_slot
	-1 TO plug_#used
	-1e0 TO plug_frequency
	-1e0 TO plug_frequency2
	#no-id TO plug_type
	0 TO plug_filtype
	0 >inhistory  TO plug_input
	0 >inhistory  TO plug_output
	0 TO plug_dynmem
	0 TO plug_dynmemsize
	1e0 TO plug_level
	0 TO plug_static
	0 TO plug_timeshift
	0 TO plug_bypassed
	0 TO plug_xt_tune
	0 TO plug_xt_quality
	0 TO plug_taps ;

: *rg_PLUG	( n c -- a )	JACK-IO-SLOTS * + CELLS *io-plug-slots + ;
: *xt_LIST	( n -- a )	CELLS ACTIVE-PLUGINS + ;
:NONAME
	0 TO fJACK-MAXLAT
	0 LOCALS| #plug |
	JACK-CHANNELS 0
	DO	0 I TO >latency
		JACK-IO-SLOTS 0
		DO	I J *rg_PLUG @ ?DUP
			IF	DUP TO #plug -->DSP
				plug_#used -1 > plug_bypassed 0= AND IF J >latency plug_timeshift +  J TO >latency THEN
			THEN
		LOOP	fJACK-MAXLAT I >latency MAX TO fJACK-MAXLAT
	LOOP ; TO CALC-LATENCIES
: REORDER-PLUGINS		( -- )
\ *G It compresses the execution token list to a zero-terminated list for faster processing
\ ** in the input->output processing chain.
	0 0 LOCALS| #plug #used |
	INIT-PLUGINS
	JACK-IO-SLOTS 0
	DO	JACK-CHANNELS 0
		DO	J I *rg_PLUG @ ?DUP		\ a plugin registered for this slot?
			IF	TO #plug  #used #PLUGINS > ABORT" Too many active plugins"
				#plug -->DSP
				I >outhistory TO plug_input		\ the current last slot for this channels is input
				#used plughistory TO plug_output	\ define a history buffer as output
				AUTO-HISTORY-CLEAR? IF plug_output #HISTORY-SIZE SFLOATS ERASE THEN
				plug_output I TO >outhistory		\ outhistory is set
				#plug #used *xt_LIST !			\ write object to execution list
				#used TO plug_#used 1 +TO #used
			THEN
		LOOP
	LOOP CALC-LATENCIES ?DUP IF EXECUTE THEN ;
: .idx	S>D <# # # #> TYPE ;
: .hist-index	( addr -- idx )
	DUP *historybuffers - 1 SFLOATS / #HISTORY-SIZE / 0 LOCALS| flag idx addr |
	idx JACK-CHANNELS < IF ." <" idx .idx SPACE EXIT THEN
	JACK-CHANNELS 0
	DO	I >outhistory addr =
		IF	TRUE TO flag SPACE I .idx ." >" LEAVE
		THEN
	LOOP	flag IF EXIT THEN
	SPACE idx .idx SPACE ;
: (.pluginfo)	( -- )
	-->DSP CR @DSP H. plug_#used -1 >
	IF	#PLUGINS 0
		DO	I *xt_LIST @ @DSP =
			IF	."  X:" I .idx SPACE plug_input .hist-index ." -" plug_output .hist-index
				LEAVE
			THEN
		LOOP
		."  S:" plug_slot .idx ."  C:" plug_channel .idx
	ELSE	#25 SPACES
	THEN
	2 SPACES plug_type COUNT TYPE SPACE
	plug_filtype
	CASE	LOW-PASS	OF ." Lowpass "		ENDOF
		HIGH-PASS	OF ." Highpass "	ENDOF
		BAND-PASS	OF ." Bandpass "	ENDOF
		BAND-STOP	OF ." Bandstop "	ENDOF
		EQUALIZER	OF ." Equalizer "	ENDOF
	ENDCASE
	plug_frequency2 F>S plug_frequency F>S
	DUP 0<	IF DROP ELSE 1 .R ." Hz" THEN
	DUP 0<	IF DROP ELSE ."  -" 1 .R ." Hz" THEN
	plug_taps ?DUP IF ." , size " . THEN
	plug_bypassed IF ."  BYPASSED" THEN ;
:NONAME \ INIT-ALL-JACK-OBJECTS
	PLUGIN-LINK
	BEGIN	DUP
	WHILE	-->DSP
		plug_dynmemsize ?DUP IF	$100 + ALLOCATE ?ALLOCATE ELSE 0 THEN TO plug_dynmem
		plug_xt_initfilter ?DUP IF EXECUTE THEN
		plug_link
	REPEAT DROP ; AtCold

: FREE-PLUGIN-MEM	\ Free all memory of plugins
	STOP-JACK-CLIENT
	PLUGIN-LINK
	BEGIN	DUP
	WHILE	-->DSP
		plug_dynmem ?DUP IF FREE DROP THEN 0 TO plug_dynmem
		plug_link
	REPEAT	DROP 0 TO CALC-LATENCIES ; ' FREE-PLUGIN-MEM AtExit

FORTH DEFINITIONS
\ *S Using the plugin objects in the signal processing chain

: LIST-PLUGINS		( -- )
\ *G prints all objects plugged into the process chain.
	ACTIVE-PLUGINS
	BEGIN	@+ DUP
	WHILE	(.pluginfo)
	REPEAT	2DROP ;
: LIST-ALL-PLUGINS
\ *G prints information about all defined plugins.
	PLUGIN-LINK
	BEGIN	DUP
	WHILE	(.pluginfo) plug_link
	REPEAT DROP ;
: REMOVE-JACK-PLUGIN	( filterobject -- )
\ *G revoves the object from the process chain.
	-->DSP plug_#used -1 = IF EXIT THEN
	plug_slot plug_channel *rg_PLUG OFF
	0 TO plug_channel -1 TO plug_#used 0 TO plug_slot
	REORDER-PLUGINS ;
JACKAUDIO DEFINITIONS
: __plugin:
	CREATE	PLUGIN-LINK ,
IS-IFORTH?
	[IF]	FORGET>	@ JACK-TEMP-MEM 0= IF DROP EXIT THEN
		INIT-PLUGINS SAFE-WAITING? -->DSP
		plug_slot plug_channel *rg_PLUG OFF
		0 TO plug_channel -1 TO plug_#used 0 TO plug_slot
		plug_dynmem ?DUP IF FREE DROP THEN 0 TO plug_dynmem
		plug_link PLUGIN-LINK MIN TO PLUGIN-LINK
		REORDER-PLUGINS
	[THEN]
	DOES>	DROP ;
FORTH DEFINITIONS

: MAKE&ALLOCATE-PLUGIN-OBJECT	( initialize-xt dynmemsize objectsize -- object )
\ *G Whenever you define a plugin object you MUST use this function to initialize it.
\ ** 'size' is the objects size in bytes, the XT is the function to initialize the object.
\ *P Dynamic memory size is allocated and plug_dynmem is set.
\ *P This is done at compile time - so the object can be used at once -
\ ** also this allows turnkey applications, as when booting all defined objects are initialized.
\ *P When developing your own plugins you can and definitly should use the automatic memory allocation, this is done before
\ ** the initialize-xt is called.
 	HERE $10 #B-ALIGNED HERE - ALLOT	\ align to the next 16byte block
 	HERE SWAP ALLOT -->DSP
	SET-OBJECT-COLD-VALUES
	DUP TO plug_dynmemsize ?DUP IF $100 + ALLOCATE ?ALLOCATE TO plug_dynmem THEN
	DUP TO plug_xt_initfilter ?DUP IF EXECUTE THEN
	PLUGIN-LINK TO plug_link  @DSP TO PLUGIN-LINK
 	BASE @ >R HEX GET-JACK-STRING >R
 	S" ALSO JACKAUDIO __plugin: __" R@ PLACE+
 	PLUGIN-LINK (.) R@ PLACE+
 	S"  PREVIOUS" R@ PLACE+
 	R> COUNT EVALUATE R> BASE ! ;

: SET-PLUGIN-BYPASS	( flag filterobject -- )
\ *G You may set the objects bypass status. When the bypass flag is set - flag is *\forth{TRUE}, the plugin just passed the
\ ** audio data to it's history buffer but does no processing on the data. See comments in *\forth{SET-JACK-PLUGIN}.
	-->DSP 0<> TO plug_bypassed ;

: SET-JACK-PLUGIN	( filterobject slot channel -- )
\ *G Puts the object in the slot/channel. channel must be between 0 and *\forth{JACK-CHANNELS-1},
\ ** slot must be between 0 and *\forth{JACK-IO-SLOTS}. If there are wrong parameters an abort
\ ** message is displayed.
\ ** If an object is already in use at that slot it is removed.
\ *P Please remember two things, at the moment you may only use *\forth{#PLUGINS} at the same time, the history buffers
\ ** are allocated when starting fjack. *\forth{SET-JACK-PLUGIN} and *\forth{REMOVE-JACK-PLUGIN} change
\ ** the processing order and clear the history buffers used. On slow machines the clearing might take some time
\ ** resulting in audio drops when changing the plugin objects in use. To overcome this problems you could keep
\ ** a plugin object in the processing chain and just toggle it's bypass status.
	LOCALS| channel slot object |
	slot 0 JACK-IO-SLOTS 1- WITHIN 0= ABORT" slot for SET-JACK-PLUGIN out of range"
	channel 0 JACK-CHANNELS WITHIN 0= ABORT" channel for PLUGIN-AUDIO-PROCESS out of range"
	slot channel *rg_PLUG @ ?DUP IF REMOVE-JACK-PLUGIN THEN
	object -->DSP 			\ Leave if already in use
	object slot channel *rg_PLUG !
	channel TO plug_channel
	slot TO plug_slot REORDER-PLUGINS ;
: RECONNECT-PLUGIN	( history-buffer object/output-port -- )
\ *G Due to the new input->output sheduler every plugin object or output port can get it's data
\ ** from another history buffer. This function can be used to connect at runtime.
\ ** See the example where the two dotted red lines show reconnected plugins, the green line shows
\ ** a reconnected output port. You can get the address of the history buffer by *\forth{OBJECT-HISTORY}.
	DUP JACK-CHANNELS <
	IF	TO >outhistory
	ELSE	-->DSP TO plug_input
	THEN ;
\ *P *\image{"Reconnecting" "reconnect" "png" 50 50}

: LOCATE-JACK-PLUGIN	( object -- slot channel true | false )
\ *G Gives information about the slot/channel for a specified object and a *\forth{TRUE} flag
\ ** when the object was installed or *\forth{FALSE} when not found.
	-->DSP plug_#used -1 >
	IF plug_slot plug_channel TRUE ELSE FALSE THEN ;
: GET-PLUGIN-OBJECT		( slot channel -- 0|filterobject )
\ *G Gives the object found in slot/channel, when no object has been placed there with
\ ** *\forth{SET-JACK-PLUGIN} a *\forth{FALSE} flag is returned.
	2DUP 0 JACK-CHANNELS WITHIN SWAP 0 JACK-IO-SLOTS WITHIN AND
	IF *rg_PLUG @ ELSE 2DROP 0 THEN ;
: SET-OBJ-FREQUENCY	( frequency object -- )
\ *G Many plugins have a specified tune-to-frequency function available, you can use this to do so.
\ ** This sets the frequency of the specified object, the object itself known what do do,
\ ** f.e. filters will be tuned to the specified frequency. A few object types have to frequencies,
\ ** *\forth{BAND-PASS BAND-STOP} filters expect two integers as a special case.
	-->DSP S>F plug_xt_tune ?DUP IF EXECUTE ELSE FDROP THEN ;

: SET-OBJ-FLOAT-FREQUENCY	( object -- ) ( F: frequency -- )
\ *G In many situations the precision of *\forth{SET-OBJ-FREQUENCY} is not good enough, then you can set
\ ** the frequency by this. BEWARE: ONLY ONE parameter possible as float!
	-->DSP plug_xt_tune ?DUP IF EXECUTE ELSE FDROP THEN ;

: SET-OBJ-QUALITY	( quality object -- )
\ *G Many plugins have a quality - in notch filters this would be the sharpness in FIR and adaptive filters this will be the number
\ ** of taps. This functions handles the objects quality in a general way so look for the special plugin docs.
	-->DSP plug_xt_quality ?DUP IF EXECUTE ELSE DROP THEN ;
\ ** The effect of setting the quality can be viewed in the fft display plugin, here is a band filter
\ ** with a lowpass set to ~1.5kHz with 41 and 125 poles
\ *P *\image{"lowquality" "41poles" "png" 25 25} *\image{"high quality" "125poles" "png" 25 25}

: SET-OBJ-LEVEL	\ ( object -- ) (F level -- )
\ *G All plugin objects have a specific level value this word sets it.
        -->DSP TO plug_level ;
: SLOT/CHANNEL-HISTORY	( slot channel -- history )
\ *G Finds the history address of a slot/channel pair, if the slot is less than 0 it always means
\ ** the input port. If a slots object is undefined the result is 0.
	OVER 0<
	IF	NIP OBJECT-HISTORY
	ELSE	GET-PLUGIN-OBJECT DUP IF OBJECT-HISTORY THEN
	THEN ;
: CALCULATE-FILTER-DELAY	( slot1 slot2 channel -- n )
\ *G Calculates the phase delay in the processing chain between 2 slots for a given channel,
\ ** the result is the phase shift measured in samples.
	0 LOCALS| delay chan |
	2DUP < IF SWAP THEN
	?DO	I chan GET-PLUGIN-OBJECT ?DUP IF -->DSP plug_#used -1 > plug_bypassed 0= AND IF plug_timeshift +TO delay THEN THEN
	LOOP delay ;

JACKAUDIO DEFINITIONS
0 VALUE BENCH-DATA
0 VALUE BENCH-CNT
:NONAME	0 TO BENCH-DATA 0 TO BENCH-CNT ; AtCold
: ((bench-kern))
	BENCH-CNT #FIR_BUFF 1- AND DFLOATS BENCH-DATA + DF@ plug_xt EXECUTE FDROP 1 +TO BENCH-CNT ;
: (bench-kern)	\ ( -- ) (F -- )
	((bench-kern)) ((bench-kern)) ((bench-kern)) ((bench-kern)) ((bench-kern))
	((bench-kern)) ((bench-kern)) ((bench-kern)) ((bench-kern)) ((bench-kern)) ;
FORTH DEFINITIONS

: OBJ-BENCHMARK	( object -- )
\ *G You can measure the performance of any filter plugin with this, define any sort of filter object and
\ ** pass the object to this tool.
	#2000 0 LOCALS| time cnt |
	STOP-JACK-CLIENT
	BENCH-DATA 0=
	IF	#FIR_BUFF DFLOATS ALLOCATE ?ALLOCATE TO BENCH-DATA
		#FIR_BUFF 0 DO I S>F PI F* 2e0 F* 0.01e0 F* FSIN I DFLOATS BENCH-DATA + DF! LOOP
	THEN
	-->DSP
	?MS >R cnt 0 DO (bench-kern) LOOP ?MS R> - 1 MAX TO time
	#2000 time / cnt * TO cnt
	?MS >R cnt 0 DO (bench-kern) LOOP ?MS R> - TO time
	."  : " time S>F 10e6 F* cnt 10 * S>F F/ F>S . ." nsec/run" ;
