(*
 * LANGUAGE    : Forth; iForth and MPE Forth working
 * PROJECT     : Forth Tool
 * DESCRIPTION : Audio processing client using JACK
 * CATEGORY    : Realtime Audio tools
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2009-08-01 15:01:10; Hanno Schwalm hanno@schwalm-bremen.de
		2009-08-10 13:16:51 supporting FORGET on iforth
		2009-12-30 13:17:32 less dummy words for doc generation
		2010-02-06 02:20:26 moved headers to fjack.frt
 *)

\ *! fjack
\ *T fJACK - a JACK Audio Connection Kit client

\ *S Introduction
\ *P When doing audio processing in Forth systems we were dependent on a working audio handler
\ ** made available by the Forth vendor or had a lot of work to do.
\ *P Modern sound systems regularly make use of a callback driven model, ALSA for linux is just one example.
\ ** But ALSA is not portable and doesn't allow connecting applications together in a sound-flow environment.
\ ** One popular system allowing this is the 'jack-audio-connection-kit' JACK.
\ ** It is available for all major desktop platforms, Linux, OSX and Windows.
\ ** Jack makes use of the available hardware driver in a real time process. There are many audio applications
\ ** around using the JACK interface and Forth can now connect to *\bold{all} of them.
\ *P fJACK is written in Forth and uses some of the advanced techniques like threading,
\ ** binding of external libraries and callbacks. So it is *\bold{not} ANS Forth.
\ ** The good news: it's available for all iForth versions (Linux, Windows and OSX - 32/64bit) and VFX Forth (Linux, Windows and OSX).

\ *S Installation
\ *N Installing the jack audio connection kit
\ *P First you must install a JACK system and get it running. Read the HOWTO in the net
\ ** and specialized audio groups to find out how you can get the most out of it. Getting JACK running with low latency and no drops/xruns
\ ** is sometimes a bit difficult. Look out for hardware that is well supported for your environment.
\ ** It used to be necessary to use a real-time-kernel, with modern low-latency linux kernels and optimized settings in security.conf
\ ** The situations has become much better over the last years.
\ *N Downloading jack
\ *P This is a good place to find jack: http://jackaudio.org/download or go for compiled packages for your system.

\ *N What sort of jack?
\ *P Currently there is a change in the jack system, either the 'old' jackd or the newer jack2 system available for
\ ** all linux distributions, windows and OSX. Also the new systems supports multicore cpus much better, things like
\ ** netjack are reimplemented and a session management is integrated. So i strongly suggest to upgrade to jack2,
\ ** at the moment it's v1.9.7. Some advanced features of fjack are only available with the latest releases.
\ *P *\bold{All} newer versions (since 0.118 or 1.9.4) have the ability to auto-start
\ ** a jack server when a client wants to register and no running server is found.
\ ** Even so - this is not a good way to start a jack server at all. fJACK can still do so but it's much better to use
\ ** the session controlling applications like QjackCtl. On linux systems, this also means that other sound systems like
\ ** pulseaudio are easy to integrate.
\ *P So the user can choose between three options:
\ *B Start the jack server before using fJACK with a tool like qjackctl. This allows precise setups of
\ ** jack according to your system and is normally done at startup. I think this is the best way. Also this allows session management,
\ ** starting fjack turnkey applications and other audio application with one click and connecting them all.
\ *B Let fJACK start the server automatically. This needs a correct setup of jack and you will have to set
\ ** *\forth{JACK-AUTOSTARTING?} to true.
\ *B Write your special *\forth{KICKSTART} word before loading fjack. This method is the least preferable.

\ *N What about the fftw3 library?
\ *P All recent fJACK versions use the fftw3 library to define the filter response of FIR filters so you really should install it even
\ ** when you don't need the signal analysis fft plugin. *\forth{BAND-PASS BAND-STOP} and *\forth{USER-FIR} filters with a user-specified
\ ** frequency response and the *\forth{EQUALIZER} will not work without it.
\ ** Calculating the filter parameters is not a real performance problem. But doing signal analysis on large amount of data
\ ** or with sizes <> 2^n depends on a 'calibrated' fftw3 system. Currently creating fftw3 plans uses *\forth{#FFTW_ESTIMATE} because
\ ** in fjack only 2^n sizes are used and then there not much of a difference. See fftw3lib.frt and fftw3.frt

\ *N What about performance?
\ *P fJACK 0.122 or later uses optimized sse2 code filter kernels on all platforms. This is done by including the code generated
\ ** on iForth systems via very simple binary code snippets. These are included at compile time depending on the used architecture and
\ ** are also used in turnkeys. So for vfx this means the same performance as iforth systems.
\ *P For iForth this also means the assembler is only loaded when you want to change the optimized code snippets,
\ ** so probably this is only used by me. If you really want to go deep into sse2 assembler programming you will find
\ ** the compile-time flag save-sse2-overlay? at the fjack.frt preamble. But -- be warned!
\ *P The fjack client tests the running cpu, if your machine doesn't offer sse2 code there is a fallback to forth code
\ ** which gives roughly half performance in all FIR filters. This test just uses the cpuid instruction, all supported operating systems
\ ** make sure, the xmm registers are safe with threading.

\ *N What about memory?
\ *P fJACK needs quite a lot of ram allocated from the system. As this is a realtime application, all memory *\bold{must}
\ ** be permanent and *\bold{MUST NOT} be swapped in/out. When using the default config parameters fJACK needs ~10MB of
\ ** allocated memory. A rather precise calculation of allocated memory can be done by:
\ *B JACK-CHANNELS #PLUGINS + #HISTORY-SIZE * SFLOATS
\ *P See the 'What can be configured in fJACK?' chapter to find out more about changing such config options.

\ *N Contact, help & bug reports
\ *P Feel free to ask for help at hanno@schwalm-bremen.de
\ *P Whenever you plan to use fJACK in commercial applications read the Licence.txt and send me an email.
\ ** Using fJACK for private or educational use is free of charge on all platforms.

\ *N Release History
\ *W <BR><CODE>&nbsp;&nbsp;</CODE><A HREF="history.txt">History</A>

\ *N Where to put the fjack sources?
\ *P The fJACK system sources are in the *\bold{fjack} directory containing the client and extension module sources
\ ** plus the 'loader' *\bold{fjack.frt}. The fjack/sysfiles directory holds the connections
\ ** and some binary files. This subdirectory should be made writable.

\ *N iForth
\ *P *\bold{All} iForth versions are supported. The fJACK sources should be in - or go to -
\ ** the *\bold{include} directory. *\forth{IN fjack} will do.

\ *N VFX Forth
\ *P The fJACK sources coming with your vfx system are included in the *\bold{examples} directory. You should *\bold{cd} to this directory, then
\ ** *\forth{INCLUDE fjack.frt} will do. When you get an fjack update you could replace that version or have it somewhere
\ ** else.

\ *N Testing the client
\ *P Loading as above compiles the basic Forth interface to jack,
\ ** defines a JACK client with some ports and connects these ports.
\ ** There are little demos availabe, see fjack/examples/tests.frt or fjack/examples/board.frt.
\ *P iforth comes with some advanced tools using graphical frontends to control filters. Try
\ ** gui/jackdemo.frt to get an idea. Sorry - this is and won't be available for vfx.

\ *N Making fJACK Turnkey Applications
\ *P fJACK can also be used in turnkey applications, all plugin object structs
\ ** and the client data are in permanent memory space to support this.
\ *P fJACK inside a turnkey applications will start automatically, you will just have to put your application 'xt' in the
\ ** boot chain.
\ *C ' application AtCold
\ *P Stopping the client and system housekeeping is done in the *\forth{AtExit} chain.
\ *P If you want to use fJACK inside sessions as supported by qjackctl you'll have to make a turnkey binary. vfx
\ ** produces 'complete' binaries, iforth *\forth{SAVE-SESSION} writes an image file that must be loaded by the iforth
\ ** server application. When you have used the linux iforth installer there will already by a script 'fJACK' in your dfwforth
\ ** directory. This script must be copied to fjack/sysfiles/fJACK - this is a place every fJACK installation knows about and
\ ** the session callbacks use.


\ *T The fJACK Client User Manual

\ MARKER -fjack
\ *G The ANS *\forth{MARKER -fjack} can be safely used to remove fJACK from your forth system.
\ ** It's supported on all platforms.

[DEFINED] IFORTH!  CONSTANT IS-IFORTH? IMMEDIATE
[DEFINED] VFXFORTH CONSTANT IS-VFX? IMMEDIATE
[UNDEFINED] pCONSTANT [IF] : pCONSTANT CONSTANT ; [THEN]
VOCABULARY JACKAUDIO
\ *G The *\forth{JACKAUDIO} vocabulary holds internals of the fJACK system, it's words could be changed
\ ** in next releases and should be used in end-user applications with great care.

ALSO JACKAUDIO ALSO DEFINITIONS

: JACK-VERSION	( -- )	S" v1.13" ;
\  \ *G The current version number

\ *N A short processing overview
\ *P fJACK does real-time processing on a user defined number of audio channels (ports) *\forth{JACK-CHANNELS}, it controls and is controlled by
\ ** a jack server. You can connect any other jack-ready application via these ports with fJACK. To allow audio data processing
\ ** you can define audio plugins and put them in any of the slots. So the basic overview is like this
\ *P *\image{"The Signal Processing Graph" "io_slots" "png" 50 50}

\ *P Please recognize that the processing callback has changed in version 0.115, now every processing plugin has it's
\ ** history buffer. Later you will find out, that *\bold{every} plugin can indeed read data from *\bold{any} history buffer.
\ ** This new sheme allows much improved sound effects and mixing.

CREATE $FJACKSYSFILE $100 ALLOT
0 VALUE SSE2-CODE?	( -- flag )	\ All library checking flags are detected before starting the client
0 VALUE FFTW3-LIB?	( -- flag )
0 VALUE JACK-LIB?	( -- flag )
0 VALUE JACK-LIB+?	( -- flag )
0 VALUE JACK-SESSION?	( -- flag )
0 VALUE WISDOM-LOADED?	( -- flag )
0 VALUE JACK-LATENCY?	( -- flag )

	INCLUDE fjack/config.frt
	INCLUDE fjack/filkern.frt

IS-IFORTH?	[IF] INCLUDE fjack/I_osdeps.frt	[THEN]
IS-VFX?		[IF] INCLUDE fjack/V_osdeps.frt	[THEN]

	INCLUDE fjack/types.frt
	INCLUDE fjack/jacklib.frt
	INCLUDE fjack/fftw3lib.frt
	INCLUDE fjack/client.frt
	INCLUDE fjack/wavbase.frt
	INCLUDE fjack/wavfiles.frt

	' START-JACK-CLIENT DUP EXECUTE AtCold
	' STOP-JACK-CLIENT AtExit

JACK-MIDI? [DEFINED] MAKE-JACK-DOCS OR
[IF]	INCLUDE fjack/midi.frt
[THEN]

JACK-MODULES? [DEFINED] MAKE-JACK-DOCS OR
[IF]	INCLUDE fjack/fftw3.frt
	INCLUDE fjack/io_procs.frt
   	INCLUDE fjack/fir2.frt
    	INCLUDE fjack/adaptive.frt
   	INCLUDE fjack/iir.frt
 	INCLUDE fjack/fft.frt
 	INCLUDE fjack/mixer.frt
 	INCLUDE fjack/effects.frt
[ELSE]
	JACKAUDIO DEFINITIONS
	: FREE-ALL-FFT-PLANS ; : FREE-PLUGIN-MEM ;
[THEN]
	FORTH DEFINITIONS

[DEFINED] MAKE-JACK-DOCS
[IF]	INCLUDE fjack/examples/tests.frt
[THEN]

\ *T Have Fun!
\ *P fJACK is a pretty complex system -- don't be frustrated!
\ *P Read the examples sources and feel free to send me an email, hanno@schwalm-bremen.de

JACKAUDIO DEFINITIONS
IS-IFORTH?
[IF]	: FORGET-JACK-SYSTEM
		DROP STOP-JACK-CLIENT FREE-ALL-FFT-PLANS FREE-PLUGIN-MEM FREE-SNIPPETS FREE-JACK-MEM ;
		' FORGET-JACK-SYSTEM IS-FORGET FORGET-JACK-SYSTEM
	FORTH DEFINITIONS
	:ABOUT	??CR ." A jack server client. Look for the HTML docs. Load the demo with: IN fjack/examples/tests" ;
	: JACKTURNKEY
		CR CR ." Writing a fJACK application image, please copy to your images directory!" CR
		S" SAVE-SYSTEM iforthjack.img" EVALUATE ;
	NESTING @ 2 <=
	[IF]	.ABOUT -fjack
	[THEN]
[THEN]

IS-VFX?
[IF]	: FORGET-JACK-SYSTEM
		STOP-JACK-CLIENT FREE-ALL-FFT-PLANS FREE-PLUGIN-MEM FREE-SNIPPETS FREE-JACK-MEM ;

	FORTH DEFINITIONS
	[DEFINED] -docgen [IF] -docgen [THEN]
	CR .(  *** See full HTML docs. Load the demo with: INCLUDE fjack/examples/tests.frt ***)
	prune: FORGET-JACK-SYSTEM ;
[THEN]
PREVIOUS PREVIOUS DEFINITIONS

