(*
 * LANGUAGE    : ANS Forth with DFW extensions
 * PROJECT     : Forth Environments
 * DESCRIPTION : Play .WAV files in the background
 * CATEGORY    : Sound Utilities
 * AUTHOR      : Marcel Hendrix, Hanno Schwalm
 * LAST-CHANGE : moved from wavfiles.frt
*)

(*
	The interface to the fJACK system.

    /*\
    |*|----====< ".WAV" file format >====----
    |*|
 0  |*|     4 bytes 'RIFF'
 1  |*|     4 bytes <length1>	    ; length_file-8 (length of what follows)
 2  |*|     4 bytes 'WAVE'
 3  |*|     4 bytes 'fmt '	    ; block = name:4, size:4, bytes:size
 4  |*|     4 bytes <length2>	    ; length of 'fmt' block
 5  |*|     2 bytes  01 	    ; format tag 1: integer 3: IEEE float
    |*|     2 bytes  xx 	    ; channels (1=MONO, 2=STEREO)
 6  |*|     4 bytes  xxxx	    ; samples per second
 7  |*|     4 bytes  xxxx	    ; average samples per second
 8  |*|     2 bytes  01 	    ; block alignment (shown for MONO 8-bit)
    |*|     2 bytes  08 	    ; bits per sample (shown for 8-bit)
    |*|     ? bytes		    ; fill to length2 bytes
    |*|     4 bytes 'data'
   .....    <more blocks>	    ; their name is not 'data'
    |*|     4 bytes ?		    ; sometimes: actual sample data byte count, else 'ja  '
    |*|       bytes <sample data>
    |*|
    |*| 	Note: the sample data must end on an even byte boundary.
    |*| 	      All numeric data fields are in the Intel format
    |*| 	      of low-high byte ordering.
    |*|
    \*/

*)

DECIMAL JACKAUDIO DEFINITIONS
\ *T Using .wav Files
\ *P Some constants - in *\forth{JACKAUDIO} - for better readability
\ *[
    2 CONSTANT STEREO
    1 CONSTANT MONO
    8 CONSTANT 8BIT-AUDIO
  #16 CONSTANT 16BIT-AUDIO
  #32 CONSTANT 32BIT-AUDIO
 #-32 CONSTANT SFLOAT-AUDIO
\ *]
1e0 FVALUE SPEEDUP

0 VALUE recording?
0 VALUE playing?
0 VALUE reposition-playing?
0 VALUE reposition-relative?
0 VALUE current-file-position?

FORTH DEFINITIONS
\ *P As playing and recording of complete .wav files is done in a background thread is it sometimes
\ ** conveniant to have a hook when the recording/playing is actually started or stopped in the thread.
\ *[
DEFER BEGIN-OF-PLAYING		' NOOP IS BEGIN-OF-PLAYING
DEFER END-OF-PLAYING		' NOOP IS END-OF-PLAYING
DEFER BEGIN-OF-RECORDING	' NOOP IS BEGIN-OF-RECORDING
DEFER END-OF-RECORDING		' NOOP IS END-OF-RECORDING
\ *]
JACKAUDIO DEFINITIONS

: wav-params ( addr -- 'wav /wav )
	5 LOCALS| #recs 'head |
	'head #16 + 	\ length2
	BEGIN	32B@+ +	( name, count, bytes)
		-1 +TO #recs #recs 0= ABORT" unstructured .WAV file"
		DUP 4 + SWAP 4 S" data" COMPARE 0=
	UNTIL
	32B@+ >R
	DUP 'head #8 + -  'head 4 + 32B@  SWAP -
	R> UMIN ;

: @MINPARS	( header -- channels mode srate )
	>R	R@ #22 + C@
		R@ #34 + C@  R@ #20 + C@ 3 = IF NEGATE THEN
		R> #24 + 32B@ ;
: WAV-HEAD?	( header -- flag )
	DUP	 4 S" RIFF"	COMPARE 0=
	SWAP 8 + 8 S" WAVEfmt "	COMPARE 0= AND ;

: @WAVPARS	( header|0 -- 'wav size channels mode srate true | false )
	DUP 0= IF EXIT THEN
	DUP WAV-HEAD? 0= IF DROP FALSE EXIT THEN
	DUP wav-params ROT @MINPARS TRUE ;
: (WRITE-WAV-HEADER)	( header fd -- )
	LOCALS| fd head |
	0 S>D fd REPOSITION-FILE DROP
	head #44 fd WRITE-FILE DROP ;
: FIX-WAV-HEADER	( size mode channels srate buffer -- )
	LOCALS| head srate channels mode size |
	head #44 ERASE
   	S" RIFF"			head		SWAP CMOVE
  	size #36 +			head #4 +	32B!
   	S" WAVEfmt "			head #8 +	SWAP CMOVE
  	#16				head #16 +	32B!	\ format length
	mode 0> IF $01 ELSE $03 THEN	head #20 +	C!	\ format tag
  	channels ABS			head #22 +	C!
  	srate				head #24 +	32B!
  	srate channels * mode 8 / * ABS	head #28 +	32B!	\ bytes/second
  	mode 8 / channels * ABS		head #32 +	C!	\ size of a complete sample
  	mode ABS			head #34 +	C!
  	S" data"			head #36 +	SWAP CMOVE
  	size				head #40 +	32B! ;
: FIX-WAV-FILE	( size mode channels srate buffer fd -- )
	LOCALS| fd head |
	head FIX-WAV-HEADER
	head fd (WRITE-WAV-HEADER)
	fd CLOSE-FILE DROP ;

: WRITE-EMPTY-HEADER	( header fd -- )
	OVER #44 ERASE (WRITE-WAV-HEADER) ;

: .WHAT ( header -- )
	@WAVPARS 0= IF ." Not a simple .WAV file." EXIT THEN
	BASE @ >R DECIMAL
	1 .R ." /sec, "
	DUP 0> IF 1 .R ." bit, " ELSE DROP ." float, " THEN
	CASE 1 OF ." mono" ENDOF 2 OF ." stereo" ENDOF 4 OF ." quad" ENDOF
		DUP . ." channels"
	ENDCASE ." , "
	#1024 / 1 .R ." kB." DROP R> BASE ! ;

0 VALUE SNIPPETS-LINK	\ points to top snippet struct, each
			\ struct for snippets: 0:addr of data; 1:link to last; 2:nfa 3:snippet/wavblock 4:filename
: LOAD-SNIPPET	( struct -- )
	0 0 LOCALS| fd fsize struct |
	0 struct !	\ clear data address
	struct 4 CELLS + COUNT R/O BIN OPEN-FILE IF DROP EXIT THEN TO fd
	fd FILE-SIZE IF DROP fd CLOSE-FILE DROP EXIT THEN DROP TO fsize
	fsize ALLOCATE IF DROP fd CLOSE-FILE EXIT THEN DUP struct !
	fsize fd READ-FILE ?FILE DROP  fd CLOSE-FILE DROP ;
: INIT-WAVING
	0 TO recording?
	0 TO playing?
	0 TO reposition-playing?
	0 TO reposition-relative?
	0 TO current-file-position? ;


:NONAME	( -- ) \ load all snippets
	INIT-WAVING SNIPPETS-LINK BEGIN ?DUP WHILE DUP LOAD-SNIPPET CELL+ @ REPEAT ; DUP AtCold EXECUTE

: FREE-SNIPPET-STRUCT	( addr -- 'addr )
	DUP @ ?DUP IF FREE DROP 0 OVER ! THEN
	CELL+ @ DUP SNIPPETS-LINK MIN TO SNIPPETS-LINK ;
: MAKE-SNIPPET-STRUCT	( str len snippet? -- )
	HERE >R OVER 4 CELLS + 2 + ALLOT ALIGN
	R@ 3 CELLS + !
	R@ 4 CELLS + PLACE
	SNIPPETS-LINK R@ CELL+ !
	@LATEST-NAME R@ 2 CELLS + !
	R@ TO SNIPPETS-LINK
	R> LOAD-SNIPPET ;
: FREE-SNIPPETS
	SNIPPETS-LINK BEGIN DUP WHILE FREE-SNIPPET-STRUCT REPEAT DROP ; ' FREE-SNIPPETS AtExit





FORTH DEFINITIONS
