(*
 * LANGUAGE    : ANS Forth with DFW extensions
 * PROJECT     : Forth Environments
 * DESCRIPTION : Play .WAV files in the background
 * CATEGORY    : Sound Utilities
 * AUTHOR      : Marcel Hendrix, Hanno Schwalm
 * LAST-CHANGE :
 *	2011-10-14 19:11:48 major workup
 *	2011-10-05 02:45:32 checking for channels is corrected in recording/playing wav files
 *	2011-09-29 00:30:52 integrated WAVSOUND, like snippets but for the wavboard
 *	2010-03-03 21:48:33 check for correct modes and channels to make playing wav files safer
 *	2010-01-28 01:05:05 snippets have FORGET> field
 *	2009-06-03 02:16:27 also wavheader is turnkey safe
 *	2009-05-13 07:26:21 redefined snippets for turnkeys
 *	2009-05-04 01:24:08 VFX checked; speed changes different
 *	2009-01-18 00:44:00 fixed wav playing size calculation
 *	2009-01-13 10:51:45 Hanno Schwalm; 64bit supported, new CB( sheme
 *	2008-09-16 04:11:47 Hanno Schwalm; using background threads for playing and recording
 *	2008-07-30 15:02:09 Hanno Schwalm; implemented with jack interface
 *	Sunday, April 13, 2008, 18:43 PM, Marcel Hendrix, new asynchronous Linux audio.c interface
 *	Sunday, March 30, 2008, 1:04 AM, Marcel Hendrix, fixed race condition
 *	March 2008, Hanno Schwalm, using the old 151 SYSCALL parameters again and
 *                trying to find files in a better way
 *	May 24, 2003, Marcel Hendrix, new call 151 (also reading)
 *	November 7, 1996, Marcel Hendrix, added 'new' Win95 .WAV format
 *	February 2, 1996, Marcel Hendrix, added .SND .AU support
 *	January 20, 1996, Marcel Hendrix, Linux version
 *	July 18, 1995, Marcel Hendrix, DPMI update
 *	July 2, 1995, Marcel Hendrix, DPMI changes
 *	February 13, 1995, Marcel Hendrix

 *)

(*
	The interface to the sound devices under JACK.
	This assumes you have a JACK audio demon working with the jack.frt module.

    /*\
    |*|----====< ".WAV" file format >====----
    |*|
 0  |*|     4 bytes 'RIFF'
 1  |*|     4 bytes <length1>	    ; length_file-8 (length of what follows)
 2  |*|     4 bytes 'WAVE'
 3  |*|     4 bytes 'fmt '	    ; block = name:4, size:4, bytes:size
 4  |*|     4 bytes <length2>	    ; length of 'fmt' block
 5  |*|     2 bytes  01 	    ; format tag 1: integer 3: IEEE float
    |*|     2 bytes  xx 	    ; channels (1=MONO, 2=STEREO)
 6  |*|     4 bytes  xxxx	    ; samples per second
 7  |*|     4 bytes  xxxx	    ; average samples per second
 8  |*|     2 bytes  01 	    ; block alignment (shown for MONO 8-bit)
    |*|     2 bytes  08 	    ; bits per sample (shown for 8-bit)
    |*|     ? bytes		    ; fill to length2 bytes
    |*|     4 bytes 'data'
   .....    <more blocks>	    ; their name is not 'data'
    |*|     4 bytes ?		    ; sometimes: actual sample data byte count, else 'ja  '
    |*|       bytes <sample data>
    |*|
    |*| 	Note: the sample data must end on an even byte boundary.
    |*| 	      All numeric data fields are in the Intel format
    |*| 	      of low-high byte ordering.
    |*|
    \*/

*)

DECIMAL JACKAUDIO DEFINITIONS

0 VALUE ForceStereo?		\ overrule a file's STEREO/MONO indicator

: SET-PARMS	( addr n -- )	LOCALS| cnt addr |	cnt addr ! cnt 0 ?DO addr I 1+ CELLS + ! LOOP ;
: GET-PARMS	( addr -- )	@+ LOCALS| cnt addr |	cnt 0 ?DO cnt 1- I - CELLS addr + @ LOOP ;

: XT-BOARD-THREAD	( pars -- )
	DUP @ 0 0 0 LOCALS| ticket size buff fd pars |
	fd FILE-SIZE DROP D>S DUP TO size $100 < IF fd CLOSE-FILE DROP EXIT THEN
	size ALLOCATE IF DROP fd CLOSE-FILE DROP EXIT THEN TO buff
	buff size fd READ-FILE 2DROP fd CLOSE-FILE DROP
	0 0 0 buff @WAVPARS 0=
	IF	3DROP 3DROP 3DROP
	ELSE	S>F SPEEDUP F* JACK-SAMPLE-RATE S>F F/ START-WAVBOARD-JOB TO ticket
		ticket pars CELL+ !
		BEGIN	ticket WAVBOARD-PLAYING? #50 MS 0= UNTIL
	THEN
	buff FREE DROP ;

: XT-PLAY-THREAD ( pars -- )	\ play wavfile thread		\ size channels mode rate #block buff fd
	DUP GET-PARMS 0 0 0 0 LOCALS| wavstart done p-ids idx fd *mem #chunk rate mode channels size pars |
	0 TO reposition-playing?
	fd FILE-POSITION DROP D>S TO wavstart
	BEGIN-OF-PLAYING
	pars $200 + TO p-ids  p-ids 4 CELLS ERASE
	BEGIN	playing? 0>
	WHILE	reposition-playing?
		IF	reposition-playing? channels * mode ABS 8 / * reposition-relative? IF done + THEN
			0 MAX size MIN #chunk / TO idx
			idx #chunk * TO done  0 TO reposition-playing?
			done wavstart + S>D fd REPOSITION-FILE DROP
		THEN
		done channels /  mode ABS 8 / / TO current-file-position?
		BEGIN idx 3 AND CELLS p-ids + @ WAV-PLAYING? DUP 0 MAX #100 MIN MS 0= UNTIL
		*mem idx 3 AND #chunk * + DUP #chunk size done - MIN 0 MAX fd READ-FILE 0= OVER 0> AND  JACK-RUNNING? AND
		IF	DUP +TO done rate channels mode PLAY-WAV 0=
			IF idx 3 AND CELLS p-ids + ! ELSE DROP -1 TO playing? THEN
		ELSE	2DROP -1 TO playing?
		THEN
		1 +TO idx
	REPEAT
	4 0 DO BEGIN	p-ids I CELLS + @ WAV-PLAYING? DUP 0 MAX #100 MIN MS 0= UNTIL LOOP
	fd CLOSE-FILE DROP
	pars FREE DROP *mem FREE DROP
	END-OF-PLAYING 0 TO playing? ;
: XT-RECORD-THREAD	( pars -- )
	DUP GET-PARMS 0 0 0 0 LOCALS| size head r-id idx *mem #chunk rate channels mode time fd pars |
	BEGIN-OF-RECORDING
	pars $80 + TO r-id
	pars $A0 + TO head
	head fd WRITE-EMPTY-HEADER
  	BEGIN	recording? 0>  time ?MS - 0> AND
  	WHILE	idx 3 >
		IF	BEGIN	r-id idx 3 AND CELLS + @ WAV-READING? DUP MS 0= UNTIL
			idx 3 AND #chunk * *mem +  #chunk fd WRITE-FILE DROP
			#chunk +TO size
   		THEN
   		idx 3 AND  #chunk * *mem +   #chunk rate channels mode READ-WAV
   		IF DROP -1 TO recording? ELSE r-id idx 3 AND CELLS + ! THEN
		1 +TO idx
	REPEAT
	4 0 DO BEGIN	r-id I CELLS + @ WAV-READING? DUP 1 MAX MS 0= UNTIL LOOP
	END-OF-RECORDING *mem FREE DROP pars FREE DROP 0 TO recording? ;

IS-IFORTH?
[IF]	' XT-RECORD-THREAD	CB( _int )CB-int RECORD-THREAD
	' XT-PLAY-THREAD	CB( _int )CB-int PLAY-THREAD
	' XT-BOARD-THREAD	CB( _int )CB-INT BOARD-THREAD
[THEN]

IS-VFX?
[IF]	fromc 1 1 CALLBACK: RECORD-THREAD	ASSIGN XT-RECORD-THREAD to-callback RECORD-THREAD
	fromc 1 1 CALLBACK: PLAY-THREAD		ASSIGN XT-PLAY-THREAD to-callback PLAY-THREAD
	fromc 1 1 CALLBACK: BOARD-THREAD	ASSIGN XT-BOARD-THREAD to-callback BOARD-THREAD
[THEN]

: (PLAY-WAVFILE)	( c-addr len 0/1 -- ) ( samplerate channels mode c-addr len 2 -- )
	0 0 0 0 0 0 LOCALS| forced-rate forced-mode forced-channels head fd pars dismode len name |
	dismode 2 = IF TO forced-mode TO forced-channels TO forced-rate THEN
	JACK-RUNNING? 0=		IF dismode IF CR ." Jack not running" THEN EXIT THEN
	playing? 0<>			IF dismode IF CR ." already playing wav file" THEN EXIT THEN
	name len R/O BIN OPEN-FILE	IF dismode IF CR ." couldn't open file: " name len TYPE THEN DROP EXIT THEN TO fd
	$1000 ALLOCATE			IF CR ." Allocation problem" DROP fd CLOSE-FILE DROP EXIT THEN DUP $800 + TO head TO pars
	head $100 ERASE head $200 fd READ-FILE 2DROP
	dismode IF head CR .WHAT THEN
	head @WAVPARS 0= IF pars FREE ?ALLOCATE fd CLOSE-FILE ?FILE EXIT THEN	\ 'wav size channels mode srate
	2>R 2>R head - S>D fd REPOSITION-FILE ?FILE 2R> 2R>			\ size channels mode srate
	dismode 2 =
	IF	3DROP forced-channels forced-mode forced-rate
	ELSE	2>R DUP 1 = ForceStereo? AND IF DROP STEREO THEN 2R>
	THEN

	DUP S>F SPEEDUP F* F>S CALC-WAV-RATE	( old new )
	2DUP <> dismode 0<> AND IF BASE @ DECIMAL OVER ."  <" 1 .R ." >" BASE ! THEN NIP
	3DUP -ROT CALC-WAV-BLOCK #16 * DUP 4 * ALLOCATE ?ALLOCATE fd			\ size channels mode rate #block buff fd
	pars 7 SET-PARMS  1 TO playing? PLAY-THREAD pars DUP $300 + START-BACK-THREAD ;

FORTH DEFINITIONS
: .WAV	( c-addr u -- )	1 (PLAY-WAVFILE) ;
\ *G Plays the complete file defined by the parameters in a background thread, show some information about the file.

: .SILENTWAV ( c-addr u -- ) 0 (PLAY-WAVFILE) ;
\ *G Plays the complete file defined by the parameters in a background thread, no information or error messages are printed.

: .SOUND ( samplerate channels mode c-addr u -- ) 2 (PLAY-WAVFILE) ;
\ *G Play any wav file with specified parameters.

: STOP-PLAYING	( -- )
\ *G Any file already playing can be stopped by this command.
	JACK-RUNNING? 0= playing? 0= OR IF EXIT THEN
	-1 TO playing? BEGIN playing? WHILE #10 MS REPEAT ;
: REPOSITION-PLAY-FILE	( position relative-flag -- )
\ *G The played wav file may be repositioned during playing. If relative-flag is *\forth{TRUE}
\ ** the position is relative to the current position, otherwise it's the absolute position.
\ ** The 'position' always means a sample position, so you don't have to worry about channels and mode.
	TO reposition-relative? TO reposition-playing? ;
: PLAY-FILE-POSITION	( -- position|-1 )
\ *G This tells you the current position of the played wav file, it's also a sample postion
\ ** so don't care about channels and mode. If the position is negative, there is no file played at the moment.
	-1 TO current-file-position?
	BEGIN	current-file-position? -1 = playing? 0<> AND
	WHILE	1 MS
	REPEAT	current-file-position? ;

: START-RECORDING	( time samplerate channels mode name namelen -- )
\ *G Starts recording a .wav file with a given file name. Also the audio
\ ** parameters must be specified. time is in seconds. This word is the basis of
\ ** the other file recording commands.
	recording? 0<>  JACK-RUNNING? 0= OR IF 2DROP 2DROP 2DROP EXIT THEN
	R/W CREATE-FILE ?FILE
	$200 ALLOCATE ?ALLOCATE LOCALS| pars fd mode channels rate time |
	fd
	time #1000 * ?MS + #100 +
	mode DUP >R
	channels 1 MAX JACK-CHANNELS MIN DUP ABS >R
	rate CALC-WAV-RATE
	DUP R> R> CALC-WAV-BLOCK #16 * DUP 4 * ALLOCATE ?ALLOCATE
	pars 7 SET-PARMS
	1 TO recording?
	RECORD-THREAD pars pars $1E0 + START-BACK-THREAD ;
: STOP-RECORDING	( -- )
\ *G Stops .wav file recording.
	JACK-RUNNING? 0= recording? 0= OR IF EXIT THEN
	-1 TO recording? BEGIN recording? WHILE 10 MS REPEAT ;
: RECORD-WAV	( time name len -- )
\ *G Records a .wav file with the duration of time seconds and the given file name.
\ ** audio parameters are *\forth{JACK-SAMPLE-RATE STEREO 16BIT-AUDIO}
	2>R JACK-SAMPLE-RATE STEREO 16BIT-AUDIO 2R>	START-RECORDING ;
: RECORD-SIMPLE	( time name len -- )
\ *G Records a .wav file with the duration of time seconds and the given file name.
\ ** audio parameters are *\forth{16000 MONO 8BIT-AUDIO}
	2>R #16000 MONO 8BIT-AUDIO 2R>			START-RECORDING ;
: RECORD-FLOATS	( time name len -- )
\ *G Records a .wav file with the duration of time seconds and the given file name.
\ ** audio parameters are *\forth{JACK-SAMPLE-RATE STEREO SFLOAT-AUDIO}
	2>R JACK-SAMPLE-RATE STEREO SFLOAT-AUDIO 2R>	START-RECORDING ;
: RECORD-STUDIO ( time name len -- )
\ *G Records a 4 channel .wav file with *\forth{SFLOAT-AUDIO}. Best for later processing
\ ** in applications like audacity ...
	2>R JACK-SAMPLE-RATE JACK-CHANNELS SFLOAT-AUDIO 2R>	START-RECORDING ;
: WAV-PLAYING-ON?	( -- flag )
\ *G flag is true when fJACK is already playing a .wav file.
	playing? 0<> ;
: WAV-RECORDING-ON?	( -- flag )
\ *G flag is true when fJACK is already recording a .wav file.
	recording? 0<> ;

: SNIPPET ( c-addr u -- ) \ name
\ *G Expects a filename on stack and creates a word 'name'. The file must be a wavfile, it's
\ ** read into allocated memory.
\ ** The runtime behaviour of the defined word is: play the audio data immediatly via the fJACK client.
\ *P Snippets are defined in a linked list; audio data are also available in turnkey applications
\ ** as they are loaded in allocated memory automatically.
	CREATE	TRUE MAKE-SNIPPET-STRUCT
IS-IFORTH? [IF]	FORGET>	STOP-ALL-WAVBOARD FREE-SNIPPET-STRUCT DROP [THEN]
	DOES>	@ @WAVPARS 0= IF EXIT THEN
		S>F SPEEDUP F* F>S -ROT >R >R >R NEGATE R> R> R> PLAY-WAV 2DROP ;

: WAVSOUND ( c-addr u -- ) \ name
\ *G Expects a filename on stack and creates a word 'name'. The file must be a wavfile, it's read into allocated memory.
\ ** The runtime behaviour of the defined word is:
\ *C name ( delay-ms lowest-channel repeats -- id ) ( F: speedup/-1 -- )
\ *P ... playing the audio data via the wavboard. See *\forth{START-WAVBOARD-JOB} for more details. If speedup is -1 the samplerate conversion
\ ** is turned off.
\ *P wavsounds are defined in the snippets linked list; audio data are also available in turnkey applications
\ ** as they are loaded in allocated memory automatically.
	CREATE	FALSE MAKE-SNIPPET-STRUCT
IS-IFORTH? [IF]	FORGET> STOP-ALL-WAVBOARD FREE-SNIPPET-STRUCT DROP [THEN]
	DOES>	@ @WAVPARS 0= IF 2DROP DROP FDROP EXIT THEN
		S>F F* SPEEDUP F* JACK-SAMPLE-RATE S>F F/ START-WAVBOARD-JOB ;

: .SNIPPETS	( -- )
\ *G Prints a lift of all defined *\forth{SNIPPET}s and *\forth{WAVSOUND}s.
	CR SNIPPETS-LINK
	BEGIN	?DUP
	WHILE	DUP @ .WHAT #36 HTAB
		DUP 4 CELLS + COUNT TYPE
		DUP 3 CELLS + @ IF ."  SNIPPET " ELSE ."  WAVSOUND " THEN
		DUP 2 CELLS + @ .NAMEHEADER CR
		CELL+ @
	REPEAT CR ;
: .WAVBOARD	( name len -- 0|id )
\ *G Plays the file via the wavboard and returns the wavboard job id or 0 when playing could not
\ ** be started. This is done with a back thread making sure all memory is deallocated when playing
\ ** has finished
	0 0 LOCALS| ticket buff len name |
	JACK-RUNNING? 0=		IF 0 EXIT THEN
	$400 ALLOCATE			IF DROP 0 EXIT THEN TO buff
	name len R/O BIN OPEN-FILE	IF DROP buff FREE DROP 0 EXIT THEN buff !
	0 buff CELL+ !
	BOARD-THREAD buff DUP $200 + START-BACK-THREAD
	100 0 DO #10 MS  buff CELL+ @ ?DUP IF TO ticket LEAVE THEN
	LOOP
	buff FREE DROP ticket ;


JACKAUDIO DEFINITIONS
	JACK-LOGIN-WAVFILE WAVSOUND startjack-audio
FORTH DEFINITIONS
IS-IFORTH? [IF] WARNING @ WARNING OFF [THEN]

: START-JACK-CLIENT	START-JACK-CLIENT  0 0 3 1E0 startjack-audio DROP ;
: STOP-JACK-CLIENT	STOP-PLAYING STOP-RECORDING STOP-JACK-CLIENT ;

IS-IFORTH? [IF] WARNING ! [THEN]

