\ File: LinList.fth

((  =============================================================
                        Tectime Data Systems
    =============================================================
    File        :   LinList.fth
    Title       :
    Description :   A expandable linear list of fixed length records
    Created     :
    Requirements:
    Author      :   G. C. Smith

    Changes     :

    =============================================================
))

\ *! Docs\LinList
\ *T An expandable linear list of fixed length records

\ *S    Introduction

\ *P Code here provides words to create linear lists of
\ ** fixed-length records but which can grow in length
\ ** dynamically at run-time.

\ *P Two defining words are provided: *\fo{ValList:} which creates
\ ** a variable-length list of cell-sized records, and *\fo{RecList:}
\ ** which takes a record size and creates a list of the specified
\ ** sized records.

\ *P Lists created with *\fo{ValList:} are intended to operate rather
\ ** like Forth *\fo{Value}s in that list items values are set and
\ ** retreived without knowledge of storage addresses.

\ *P Use *\fo{val>List (n list -- )} to append an item to the list and
\ ** *\fo{lstVal (list idx -- n)} to obtain the value *\i{n} of
\ ** the *\i{idx}-th item. The address of a particular item can be
\ ** obtained using *\fo{lstRec  (list idx -- addr)} if required.

\ *P *\i{Value List}s can be created outside the Forth dictionary
\ ** using *\fo{Alloc-ValList   ( -- addr|0)} and for repeated actions
\ ** on a list's items *\fo{Each-lstVal  ( list xt -- ???)} is
\ ** provided.

\ *P A *\i{Record List} is created by *\fo{RecList: ( recsize "name" ++)}
\ ** the create a word in the dictionary and
\ ** *\fo{Alloc-RecList  ( recsize -- addr|0)} to create a list
\ ** which has been allocated from OS memory. See below for more
\ ** words which operate on *\i{Record List}s.

\ *P *\i{Record List}s and *\i{Value List}s have one advantage over
\ ** a much simpler array-like structure in that their lengths can
\ ** vary: appending items to lists with *\fo{val>List} and
\ ** *\fo{Rec>list} will increase the size of the lists when required.

\ *P The number of items that may be stored in a list is termed its
\ ** *\i{capacity} and this is increased by 64 whenever appending
\ ** an item to a list which is full. A list's capacity can be set
\ ** with *\fo{lstSetCapacity ( n list --)}. The numer of items
\ ** a stored in a list is returned by *\fo{lstCount ( list -- n)}.


\ ===============================================================
\ *L
\ *N Glossary
\ ===============================================================

\ *P All words are compiled into the vocabulary *\fo{LinearLists}
\ ** and those intended for use by applications are *\fo{Export}ed
\ ** to the Forth vocabulary. The words below are those exported
\ ** words.


only forth definitions

vocabulary LinearLists
also LinearLists definitions

struct /RecList
    cell    field   ll.Count    \ number of items in list
    cell    field   ll.Cap      \ capacity; number of records allocated
    cell    field   ll.RecLen   \ size of each record.
    cell    field   ll.Data     \ address of allocated data
end-struct

: ?ll   \ list -- ;
\ *G List the list's properties: its count, capacity, record length
\ ** and the address of its data.
    cr
    dup ll.Count ? ." items of " dup ll.RecLen ? ." bytes each." cr
    ." Total capacity is " dup ll.cap ? cr
    ." Data is at " dup ll.data @ . cr
    ." Data Length is " dup ll.cap @ swap ll.RecLen @ * . cr
;

: RecList:      \ recsize "name" ++
\ *G Define a *\i{Record List} with records which are *\i{recsize}
\ ** bytes long.
    /Reclist buffer:
    latest-xt execute ll.RecLen !
;

: ValList:  \ "name" ++
\ *G Creates a named dictionary item which is a *\i{Value List}. It
\ ** is a cell-sized *\i{Record List} for which special words are
\ ** available.
    cell RecList:
;

: Alloc-RecList  \ n -- addr|0 ; n = record size
\ *G Allocates a block of memory and format it as a *\i{Record List}
\ ** with a record size of n bytes. *\i{addr} is the address of the
\ ** allocated memory or zero if the allocation failed.

    /RecList allocate       \ n addr 0 | n x err#
    if 2drop 0              \ 0
    else tuck ll.RecLen !   \ addr
    then
;

: Alloc-ValList    \ -- addr|0
\ *G Allocates a cell-sized *\i{Record List} which can be used as
\ ** a *\i{Value List}.
    cell Alloc-RecList
;


also Environment
Max-N
previous
16 / constant MaxListsize       \ this is what Delphi 6 does


errdef err:allocList    "Cannot allocate memory for linear list"
errdef err:ListCap      "List capacity exceded"
errdef err:ListIndex    "List index out of bounds"



: SetCap    \ list newcap -- ; addr n ;
\ set size of data memory (its CAPacity) to n records
    dup MaxListSize > err:ListCap ?throw
                                    \ list cap
    swap >r                         \ cap               R: list
    r@ ll.RecLen @ over *           \ cap newsize       R: list
    r@ ll.cap @ 0=                  \ cap size flag ; first time?   R: list
    if  allocate                    \ cap addr1 ior1    R: list
    else r@ ll.Data @ swap resize   \ cap addr2 ior2    R: list
    then                            \ cap addr ior      R: list
    err:allocList ?throw            \ cap addr          R: list
    r@ ll.data !                    \ cap               R: list
    r> ll.cap !                     \ --                R: --
;

: GrowList  \ list -- ;
\ increase size of available data space. Increase it's capacity.
    dup ll.cap @ 64 + SetCap
;

: lstSetCapacity    \ n list --
\ *G Set the list's capacity to n records. Note that this might
\ ** cause a *\fo{Resize} of previously allocated memory and this
\ ** may/will result in memory having a different address.
    2dup ll.cap @ <>
    if  swap SetCap
    else 2drop then
;

: lstCapacity       \ list -- size
    ll.cap @
;


: ?grow     \ list -- ; grow list if full
    dup ll.count @ over ll.cap @ =
    if growlist else drop then
;

: ?index    \ list idx -- ; throw if invalid list index
    0 rot ll.count @ within 0= err:ListIndex ?throw
;

: (lstRec)      \ list idx -- add ; return the idx-th item address
    over ll.reclen @ * swap ll.data @ +
;

: lstNewRec     \ list -- addr
\ *G Adds an item to the list and returns its address. The lists's
\ ** capacity is increased if necesary. The item is set to all zeros.
    dup ?grow               \ list
    dup ll.reclen @ swap    \ len ist
    dup ll.count tuck       \ len 'ct list 'ct
    @ (lstRec)              \ len 'ct addr
    1 rot +!                \ len addr
    dup rot erase
;

: val>List      \ n list -- ;
\ *G Append the value *\i{n} to the given list. Uses *\fo{lstNewRec}.
   lstNewRec !
;

: Rec>list      \ addr list -- ;
\ *G Append a new record to the given list and copies the contents
\ ** of the record at *\i{addr} to the new list record. Uses
\ ** *\fo{lstNewRec}.
    dup lstNewRec           \ addr list 'rec
    swap ll.reclen @ move   \ --
;

: lstRec        \ list idx -- addr ;
\ *G Return the address of the *\i{idx}-th item of the list given.
    2dup ?index (lstRec)
;

: lstVal        \ list idx -- n ;
\ *G Return the given list's *\i{idx}-th item value.
    2dup ?index (lstRec) @
;

: lstCount  \ list -- n
\ *G Returns the number of items in the given list. A list's count
\ ** is not the same as its capacity, which is the number of items
\ ** it can contain before it needs to be resized.
    ll.count @
;

: lstClear  \ list -- ;
\ *G Mark a list as 'empty'. Any allocated memory is not *\fo{Free}d.
    ll.count off
;

: lstFree   \ list -- ;
\ *G Marks the given list as 'empty' and *\fo{Free} its allocated
\ ** memory.
    dup ll.data @ dup if free then drop \ list
    dup ll.RecLen @  swap       \ reclen list
    dup /Reclist erase          \ reclen list
    ll.RecLen !                 \ --
;

: Free-RecList  \ reclist -- ; free list allocated by ALLOC-RECLIST
    dup lstFree Free drop
;

: Free-ValList  \ reclist -- ; free list allocated by ALLOC-VALLIST
    dup lstFree Free drop
;

: .lst      \ list --
\ *G Lists a the values of a values list. Intended for debug/test
\ ** puposes.
    cr
    dup ll.count @ 0 ?do
    i 3 .r ." ) "
    dup i lstVal . cr
    loop
    drop
;

: Each-lstVal   \ list xt -- ???
\ *G *\i{list} is the address of a *\i{Value List}
\ ** and *\i{xt} is the execution of a Forth word. *\i{xt} is executed
\ ** with the value of each item in the list on top of the stack.
\ ** Normal use would be for the executed word to consume each value
\ ** and have no other stack effect.
            { list xt -- ??? }
    list ll.count @ 0 ?do
    list i (lstRec) @ xt execute
    loop
;

: Each-lstRec   \ list xt -- ???
\ *G *\i{list} is the address of a *\i{Record List) or *\i{Value List}
\ ** and *\i{xt} is the execution of a Forth word. *\i{xt} is executed
\ ** with the address of each item in the list on top of the stack.
\ ** Normal use would be for the executed word to consume each address
\ ** and have no other stack effect.
            { list xt -- ??? }
    list ll.count @ 0 ?do
    list i (lstRec) xt execute
    loop
;

: lstSetRecLen  \ n list --
\ *G Set the record length of the given list to *\i{n} bytes long.
\ ** Any previously allocated memory is *\fo{Free}d first.
    dup lstFree ll.RecLen !
;

: lstCopyRec    \ list srcIndex destIndex --

    rot >r                  \ srcIndex destIndex        R: list
    r@ swap (lstRec)        \ srcIndex destAddr         R: list
    r@ rot (lstRec)         \ destAddr srcAddr          R: list
    swap r> ll.RecLen @ move  \ --                        R: --
;

: (lstDelRec)    \ list idx --
    2dup ?index                 \ list idx
    over ll.Count @ 1- swap ?do \ list
    dup i 1+ i lstCopyRec       \ list
    loop                        \ list
    -1 swap ll.Count +!         \ --
;

: lstDelete     \ list idx --
\ *G Deletes the *\i{idx}-th item from the given list. The contents
\ ** of each record greater than *\i{idx} is copied to its record
\ ** below it in the list and the number of items in the list is
\ ** decreased by one.
    ['] (lstDelRec) catch if 2drop then
;

: lstExchange   \ list idx1 idx2 --
\ *G Swap the contents of items *\i{idx1} and *\i{idx2} in the
\ ** given list.
                { list idx1 idx2 | addr1 addr2 addr3 len }
    list idx1 lstRec -> addr1
    list idx2 lstRec -> addr2
    list lstNewRec -> addr3
    list ll.reclen @ -> len
    addr1 addr3 len move
    addr2 addr1 len move
    addr3 addr2 len move
    -1 list ll.count +!
;


: lstIndex      \ recn list -- n
    tuck 0 (lstrec) -   \ list (recn-rec0)
    swap ll.RecLen @ /  \ idx
;

: lstData/Len   \ list -- addr len
    dup ll.data @       \ list addr
    over ll.RecLen @    \ list addr rlen
    rot ll.count @ *    \ addr len
;

: (lstInsert)   \ list idx --
\ insert a record at the given position by moving all records
\ 'up' one position.
    2dup ?index

    over swap (lstRec)          \ list addrlo
    over lstNewRec over - >r    \ list addrlo       R: size
    swap ll.reclen @            \ addrlo reclen     R: size

    2dup over +                 \ addrlo reclen addrlo addrhi   R: size
    r> cmove>                   \ addrlo reclen
    erase                   \ --
;

: lstInsert     \ list idx --
    ['] (lstInsert) catch
    if 2drop then
;

\ ===============================================================
\ Searching for values
\ ===============================================================

: lstValIndex   \ n ValList -- idx|-1 ; return the index of value n
    dup ll.count @ 0 ?do        \ n vlist
        2dup i (lstRec) @ =     \ n vlist flag ; found it?
        if 2drop i unloop exit then
    loop
    2drop -1
;

: lstValExists? \ n ValList -- flag ; TRUE iff n is in list
    lstValIndex 0 >=
;

 ((
vallist: l1
000 l1 val>list
111 l1 val>list
222 l1 val>list
333 l1 val>list
444 l1 val>list
555 l1 val>list
666 l1 val>list
777 l1 val>list
888 l1 val>list
999 l1 val>list
???
\ ))


\ ===============================================================
\
\ ===============================================================
\ only forth definitions context @
\ also LinearLists definitions
' Forth voc>wid

Export ValList:     \ "name" ++
Export lstVal       \ list idx -- n ; return the idx-th item value
Export Alloc-ValList    \ -- addr|0
Export val>List      \ n list -- ; append n to list
Export Each-lstVal  \ list xt -- ???

Export Free-RecList  \ reclist -- ; free list allocated by ALLOC-RECLIST
Export Free-ValList  \ reclist -- ; free list allocated by ALLOC-VALLIST

Export /RecList
Export RecList:
Export Alloc-RecList    \ n -- addr|0
Export Rec>list         \ addr list -- ;
Export lstSetRecLen     \ n addr --
Export Each-lstRec      \ list xt -- ???
Export lstNewRec        \ list -- addr
Export lstDelete        \ list idx --
Export lstExchange      \ list idx1 idx2 --

Export lstCount         \ list -- n
Export lstClear         \ list -- ; mark list as 'empty'; memory not FREEd
Export lstFree          \ list -- ; mark list as 'empty'; and FREE memory
Export .lst             \ list --
Export lstSetCapacity   \ size list --
Export lstCapacity      \ list -- size
Export lstRec           \ list idx -- addr

\ Export lstStart     \ list -- ; start a traversal of list
\ Export lstMore?     \ list -- flag ; has traversal morwe to go?
\ Export lstNext      \ list -- item ; return next in list
\ Export lstCurr      \ list -- item ; return current in list

Export lstIndex         \ rec list -- idx
Export lstInsert        \ list idx --

Export ?ll              \ list --
Export lstData/Len      \ list -- addr len ; suitable for writing to file
Export lstValIndex      \ n ValList -- idx|-1 ; return the index of value n
Export lstValExists?    \ n ValList -- flag ; TRUE iff n is in list

drop
only forth definitions

\ ===============================================================
\ *L
\ *> ###
\ ===============================================================

((
: last-xt	\ -- xt
  latest ctrl>nfa name>
;

: VocModule	\ "<name>" -- ; -- prev-current
\ *G As *\fo{MODULE}, but uses a named wordlist in the
\ ** *\fo{CURRENT} wordlist/vocabulary.
  vocabulary
  current @
  also  last-xt execute  definitions
;
))
