\ PID.FTH - Three-term controller implementation

((
Copyright (c) 1994-2001
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441

PLEASE NOTE THAT OUR PHONE NUMBER CHANGED IN APRIL 2000

To do
=====

Change history
==============
20010424 SFP002 Updated for XC6 and VFX Forth
19940925 MWR001 Original release
))


\ ================================
\ *T PID Controller Implementation
\ *! pid
\ ================================

\ *S Introduction
\ *P The source code file PID.FTH contains an implementation of a classic
\ ** three-term (PID) controller algorithm in Forth. The was originally
\ ** developed for the Hitachi H8 microprocessor using the MPE Forth Cross
\ ** Compiler Version 5, and tested with the TDS 2020 single-board
\ ** computer. Since then the code has been updated to run with MPE's
\ ** VFX Forth for Windows, and for the MPE Forth 6 cross compilers.

\ *P All words are written in high-level Forth, and should be capable of
\ ** porting to any architecture processor with little or no adaptation.
\ ** To port to other processor architectures, it should only be necessary
\ ** to modify the defining word Controller  and the set of data access
\ ** words associated with it.

\ *P The implementation is for integer arithmetic, and is designed to
\ ** operate with control variables in ranges up to -10000..10000
\ ** (giving a maximum of 4 digits precision).


\ *S Using the PID system
\ *P The three-term control algorithm is implemented via a defining word
\ ** PID-Controller: which creates a three term controller object, and a
\ ** set of words for accessing and manipulating such a controller.
\ *E PID-Controller: MyPID
\ *P When MyPID executes, it returns the address of its data. The layout
\ ** of the data is defined by the structure /PID-DATA.

\ *N Setting Controller Parameters
\ *P Words are provided for setting the constants of the proportional,
\ ** integral and derivative terms of the control algorithm (SetKp,
\ ** SetKi, SetKd).  Each of these takes two integer parameters
\ ** representing numerator and denominator of a rational value.

\ *P When a controller is created, the coefficients are set at
\ ** Kp = 1, Ki = 0, Kd = 0.

\ *P SetDeadband sets a deadband for the computed error (i.e. setpoint
\ ** - process variable) which controls whether the output value is
\ ** refreshed and can therefore suppress output jitter on a process
\ ** which is effectively in control. The output is not recomputed if
\ ** the average of the last two error values is in the range:
\ ** (-deadband value) .. (deadband value).
\ ** When a controller is created, its deadband value is set at 0.

\ *P SetSP sets the setpoint for the controller. The value given must be
\ ** less than 10000. The initial setpoint assigned when a controller is
\ ** created is 0.

\ *P SetSlidingWindow sets the size of the sliding window on past error
\ ** readings. In this implementation, the integral term is computed
\ ** by multiplying the average of the last n error values by the integral
\ ** term coefficient, where n is the sliding window size. By default, a
\ ** sliding window size of 64 is set up when a controller object is created.
\ ** The user may reduce the size but in this implementation cannot increase
\ ** it beyond this size. Reducing the window size does not affect the storage
\ ** allocated to a controller object or the performance of the algorithm in
\ ** computing the integral term. The main purpose of changing the size is
\ ** to avoid an excessive bias effect at changeover where there is a short
\ ** term square wave profile to the setpoint.
\ ** When a controller is first run, and subsequently on each occasion when
\ ** SetSlidingWindow is invoked, the buffer of past error readings is cleared.

\ *N Reading the Controller State
\ *P A set of words (ReadOutput, ReadPV and ReadError) provides access to
\ ** the current settings and readings of a controller, returned as
\ ** signed values.

\ *P ReadLockedStatus returns a flag which, when True, inidcates that on
\ ** the last iteration the output was not recomputed because the average
\ ** of the last two errors was within the deadband.

\ *N User Hooks
\ *P A Controller object operates on output, measured (PV) and setpoint
\ ** values that are uniformly scaled and dimensionless. By default, a
\ ** controller feeds its output back to its process value via a simulation
\ ** (see the next section) and, given an opportunity, will free run as
\ ** often as it is invoked to carry out one iteration of the control
\ ** algorithm. Three hooks are provided to allow the user to specify
\ ** the handling of  the output, the supply of the process value, and
\ ** the timing of its next iteration.

\ *P OutputMethod allows the user to specify a word to call whenever
\ ** the controller has completed a single PID step. The word can be used
\ ** to scale or otherwise condition the output signal, send it to a
\ ** real-world output device, or whatever else is needed to make the
\ ** control useful. It should take a signed value on the stack
\ ** (the internal output value) and return nothing. It is called whether
\ ** or not the controller is locked through deadbanding, so that it can
\ ** correctly handle output methods that are time-dependent such as pulse
\ ** train generation.

\ *P InputMethod allows the user to specify a word to call whenever the
\ ** controller is about to execute a single PID step. The word can be
\ ** used to acquire data from the real world and to scale or otherwise
\ ** condition the input signal. It should take no parameters on the
\ ** stack and return a signed value (the process value) which
\ ** can be directly compared with the setpoint value.

\ *P SetSamplingTrigger allows the user to specify a word called at
\ ** each pass when the controller is run continuously with the word RunPID.
\ ** It should take no parameters and return a logical on the stack, set
\ ** True if the control algorithm should be single-stepped on this pass.
\ ** This allows the actual control to be synchronized with external events
\ ** such as timers. A controller object is created with a default trigger
\ ** routine installed which always returns True, and thus will free run
\ ** a controller.

\ *N Simulation
\ *P When a new Controller object is created, the InputMethod and
\ ** OutputMethod vectored user routines are plugged with default routines
\ ** which generate a process variable value by simulating a load on the
\ ** output. The recurrence relation used in this simulation is as follows:
\ *E Where 	Pi = output value at time i,
\ **   Vi = process value at time i,
\ **   i = 1,2,3...
\ **
\ **   z10 = 0
\ **   z20 = 0
\ **   z1i = 5Pi + z1i-1
\ **   Vi = z2i = (0.52z1i + z2i-1)/2
\ *P Note: this simulation as supplied works on a single set of internal
\ ** variables and therefore cannot be used as the default behaviour for
\ ** more than one Controller object at a time. It could be modified to
\ ** use variables specific to each Controller, but at the cost of increasing
\ ** the RAM requirement for each Controller, whether or not it is of
\ ** interest.

\ *N Running the Controller
\ *P Words are provided to single-step the controller, run it
\ ** continuously, and control continuous running in a multitasking
\ ** environment.

\ *P Kick will single-step the given Controller object as follows:
\ *B Invoke the user-supplied or default vectored routine to return
\ ** the current process value
\ *B Recalculate the error from (setpoint - process value)
\ *B Update the error history for integral and derivative term
\ ** calculations
\ *B If two latest (absolute) error terms show deviation from deadband,
\ ** recalculate the output
\ *B Invoke the user-supplied or default vectored routine to process
\ ** the output.

\ *P RunPID will continuously run the Controller. As supplied it will
\ ** run in a single-tasking environment and take over the single
\ ** Forth thread, but can be stopped by invoking StopPID for the
\ ** Controller (see below) within an interrupt service routine.
\ ** If the word PAUSE has already been defined in a multitasking
\ ** environment, RunPID will include a call to PAUSE and any
\ ** other thread can then invoke StopPID on the Controller to
\ ** terminate its run.

\ *P The action of RunPID is as follows:
\ *E Clear Stop Flag
\ ** BEGIN
\ **   Execute vectored routine to check whether to trigger next iteration;
\ **   IF next iteration to be triggered:
\ **     Single step Controller;
\ **   END;
\ **   ( Yield to other processes);
\ ** UNTIL signalled to stop;

\ ** Debug and Testing
\ *P The control algorithm has a simple debug mode which can be
\ ** switched on globally for all defined Controller objects by
\ ** setting the VALUE DebugPID? to True. In debug mode, each
\ ** time a controller is single-stepped through its control
\ ** algorithm it will report on the Forth console:
\ *B the number of iterations to date
\ *B the current setpoint, process variable and error values
\ *B newly calculated values for the porportional, integral
\ ** and derivative terms;
\ *B if deadbanding does not come into play, the new output value.

\ *P A couple of testing routines are provided. InitController
\ ** will install a set of Controller parameters that are stable
\ ** for the simulation provided. Test will switch the system
\ ** into debug mode and repeatedly single-step the specified
\ ** Controller each time the user presses a key, until he or
\ ** she presses Esc.

\ *S PID COntroller Glossary

decimal

\   System constants

10000 value UnitRange			\ Internal units range from 0-10000
64    value MaxWindowSize		\ Largest window for integration in CELLS

-700 constant pid_range			\ out of range error code


\ Define access words for controller

struct /pid-data	\ -- len ; defines the PID structure
\ *G Defines the structure used for a PID controller.
  2 cells field Kp		\ proportional scaling ratio
  2 cells field Ki		\ integral scaling ratio
  2 cells field Kd		\ derivative scaling ratio
  cell    field Deadband
  cell    field Setpoint
  cell    field PV
  cell    field PError
  cell    field Output
  cell    field Locked
  cell    field WindowSz
  cell    field Readings
  cell    field Newest
  cell    field PrevVal
  2 cells field Totalizer
  cell    field VTrigger
  cell    field VInput
  cell    field VOutput
  cell    field RunFlag
  cell    field IBuffer		\ size defined by WindowSz field
end-struct

: WindowReading	\ n addr1 -- addr2
\ *G Return address of a specified error reading in the integral window.
  IBuffer SWAP cells +
;


\ -------------
\ Utility words
\ -------------

[undefined] M/ [if]
: M/		\ d1 n2 -- quot
  sm/rem NIP
;
[then]


\ -------------
\ Debug support
\ -------------

true value DebugPID?	\ -- flag ; Set TRUE for interactive testing

VARIABLE Pass		\ -- addr ; Counts iterations of control loop


\ -------------------------------------------
\ Dummy user routines plugged into controller
\ -------------------------------------------

\ Variables to support dummy routines
VARIABLE SimPV	\ -- addr ; Holds process variable simulated from output
2VARIABLE Zi1	\ -- addr ; Holding registers for simulation
VARIABLE ZI2

\ Dummy routines for plugging vectored execution procedures
\ NOTE: provided to support simulation on ONE LOOP ONLY at a time
\       Running two loops with the simulation routines left in both
\       will produce very strange results!

: DefTrigger	\ -- flag ; Run loop continuously
  True
;

: DefInput	\ -- n ; Return computed PV
  SimPV @
;

\ Simulation:
\   z1(0) = 0
\   z2(0) = 0
\   z1(i) = 5 x OP(i) + z1(i-1)
\   PV(i) = z2(i) = (0.52 x z1(i) + z2(i-1)) / 2
\     for i = 1, 2, 3...
: DefOutput	\ n --
  5 M* Zi1 2@ D+                        \ Second-order load term 1
  2DUP Zi1 2!
  100 M/ 26 * Zi2 @ 2/ +                \ Second order load term 2
  DUP Zi2 !
  SimPV !
;


\ --------------------------------
\ Defining word for PID controller
\ --------------------------------

: PID-Controller:	\ -- addr ; Target and VFX Forth version
\ *G Creates a controller object with a default set of properties
\ ** and behaviour as described below under individual words.
  CREATE
    here >r
    /pid-data allot			\ reserve data space
    r@ /pid-data erase
    1 1 r@ Kp 2!			\ default constant initialisation
    0 1 r@ Ki 2!
    0 1 r@ Kd 2!
    MaxWindowSize r@ WindowSz !		\ Default window size for integral calc
    -1 r@ Newest !			\ Pointer to latest in window
    -2 r@ PrevVal !			\ Pointer to previous in window
    ['] DefTrigger r@ Vtrigger !	\ Vectored trigger routine for 'Run'
    ['] DefInput r@ VInput !		\ Vectored input preprocessing
    ['] DefOutput r> Voutput !		\ Vectored output postprocessing
    MaxWindowSize cells ALLOT		\ Window for integral calc
  DOES>
;


\ -----------------------------
\ Setting controller attributes
\ -----------------------------

: ?UnitRange	\ n -- ; error if value out of range
  UnitRange > pid_range ?throw		\ Check in range
;

: SetKp		\ numer denom pid --
\ *G Set constant for proportional term, where
\ ** n1/n2 represents a constant in the range 0..1 as
\ ** numerator/denominator
  Kp 2!					\ Store in Controller RAM
;

: SetKi		\ numer denom pid --
\ *G Set constant for integral term, where
\ ** n1/n2 represents a constant in the range 0..1 as
\ ** numerator/denominator.
  Ki 2!					\ Store in Controller RAM
;

: SetKd		\ numer denom pid --
\ *G Set constant for the derivative term, where
\ ** n1/n2 represents a constant in the range 0..1 as
\ ** numerator/denominator
  Kd 2!					\ Store in Controller RAM
;

: SetDeadband	\ n pid -- ; Set deadband limit
\ *G Sets the controller deadband (i.e. the absolute value for
\ ** the error below which no change will be made to the output).
  OVER ?unitRange			\ Check in range
  Deadband !				\ Store in Controller RAM
;

: SetSP		\ n addr -- ; Set controller Setpoint
  OVER ?UnitRange			\ Check in range
  Setpoint !				\ Store in Controller RAM
;

: SetSlidingWindow	\ size pid -- size2 ;
\ *G Sets the size of a 'sliding window' on past readings of the
\ ** error from which the integral term will be calculated.
\ ** A 'size' of 0 implies 'let the software decide on a suitable
\ ** buffer size', a positive 'n1' implies 'allocate a buffer of
\ ** this size'. 'size2' is the actual size of the sliding window
\ ** allocated. The default window size is 64.
  >r					\ stash pid address
  dup 1 < if				\ restrict range
    drop MaxWindowSize
  else
    MaxWindowSize MIN           	\ Limit to maximum buffer
  endif
  dup R@ WindowSz !               	\ Store in size
  0 R@ Readings !	             	\ Clear buffer and...
  0 S>D R@ Totalizer 2!
  -1 R@ Newest !	              	\ ...reset pointers
  -2 R> PrevVal !
;

\ Plug in user sampling trigger routine
: SetSamplingTrigger	\ xt pid --
\ *G Only required where 'Run' is used (see below) to specify
\ ** a user routine which controls whether it is time to run the
\ ** control algorithm again. The user routine should take no
\ ** parameters and should return a boolean on the stack
\ ** (TRUE implies time to single-shot the control algorithm again).
\ ** By default a routine is installed which always returns TRUE.
  VTrigger !				\ Store execution address
;


\ -------------------------
\ Reading controller values
\ -------------------------

: ReadOutput	\ pid -- n
\ *G Read current output setting for the controller.
  Output @                              \ Return current output setting
;

: ReadPV	\ pid -- n
\ *G Read current PV reading for the controller.
  PV @					\ Return current PV reading
;

: ReadError	\ pid -- n
\ *G Read current error value for the controller.
  PError @                              \ Return current error reading
;


: ReadLockedStatus	\ pid -- flag ; true if locked
\ *G Read current locked status for the controller.
  Locked @				\ Return current status
;


\ ---------------------
\ Data transfer methods
\ ---------------------

\ Set up output handling routine
: OutputMethod	\ xt pid -- flag
\ *G Sets XT as the output method action of controller PID. The stack
\ ** effect of XT should be "n --", where n is the value to be used
\ ** as the output.
\ *P Allows the user to define a routine to process the output
\ ** signal each time it is refreshed by the controller. This
\ ** could involve conditioning the signal and/or sending it to
\ ** a physical device. The routine should take an integer from
\ ** the stack and return nothing on the stack. By default a
\ ** controller installs a routine which calculates a new
\ ** simulated value for the process variable and stores it
\ ** internally.
  VOutput !                             \ Store execution address
;

: InputMethod	\ xt pid -- ; Set up input handling routine
\ *G Sets XT as the input method action of controller PID. The stack
\ ** effect of XT should be "-- n", where n is the value to be used
\ ** as the input.
\ *P This allows the user to define a routine to process the input signal
\ ** each time the controller's control algorithm is invoked and
\ ** return it as the process variable. This could involve reading
\ ** the signal from a physical device and/or conditioning it.
\ ** The routine should take no parameters and should return a
\ ** single integer on the stack. By default a controller
\ ** installs a routine which reads a value for the process variable
\ ** from that internally stored by the default (simulation)
\ ** output method.
  VInput !                              \ Store execution address
;


\ ---------------
\ Process Control
\ ---------------

\ Calculate error term from current setpoint and process variable value
: CalcError	\ pid -- pid
  DUP Setpoint @  OVER PV @ -		\ Calculate error
  OVER PError !                         \ Store it in controller
;

\ Update and report 'locked' status
: Locked?	\ pid -- pid flag
  DUP PError @ ABS                      \ Get current error magnitude
  OVER DUP PrevVal @                    \ Get previous reading
  SWAP WindowReading @ ABS +            \ ...and add it
  OVER Deadband @ 2* <=                 \ Get deadband setting, compare
  2DUP SWAP Locked !                    \ Store new status and return it
;

\ Calculate proportional term
: PropTerm	\ pid -- pid n
  DUP PError @                          \ Get current error
  OVER Kp 2@ */                         \ Scale by prop. constant
;

\ Calculate integral term using a 'sliding window'
: IntTerm	\ pid -- pid n
  >R                                    \ Controller address needed all the time
  \ Check whether the circular buffer is full
  R@ Readings DUP @                     \ Get current number of readings
  DUP R@ WindowSz @ < IF                \ If less than maximum..
    1+ SWAP !                           \ Update number of readings
  ELSE
    \ Number of readings stays the same, but oldest error reading
    \ must be removed from integral term
    2DROP
    R@ Totalizer DUP 2@                 \ Get current totalizer
    R@ Newest @                         \ 'Newest' currently points to oldest
    R@ WindowReading @                  \ Get it from the buffer
    S>D D-                              \ Subtract it...
    ROT 2!                              \ ...and store new totalizer value
  ENDIF
  \ Store the new error reading in the sliding window buffer
  \ and update the pointer to the 'last in'
  R@ Newest DUP @                       \ Get index to latest in buffer
  1+  R@ WindowSz @ MOD                 \ Update index to latest (with wrap)...
  R@ PError @                           \ Get new error reading...
  OVER R@ WindowReading !               \ ...and store it in buffer
  SWAP !                                \ and store updated index
  \ Add the current error to the totalizer
  R@ Totalizer DUP 2@                   \ Get current totalizer
  R@ PError @ S>D D+                    \ Add current error...
  ROT 2!                                \ ...and store new value
  \ Update the 'previous reading' pointer (for the benefit of
  \ the derivative term)
  R@ PrevVal DUP @
  1+  R@ WindowSz @ MOD                 \ Update index to previous (with wrap)...
  SWAP !                                \ ...and store
  \ Calculate the integral term to be returned (downweighted when
  \ less than the maximum readings in the sliding window)
  R@ Totalizer 2@                       \ Calculate sum(Readings)/Window size...
  R@ WindowSz @ M/
  R@ Ki 2@ */                           \ ...scale by integral constant...
  R> SWAP                               \ ...and return on TOS
;

\ Calculate derivative term using previous and latest
: DerivTerm	\ pid -- pid n
  DUP Readings @ 1 > IF                 \ Check enough readings to do it
    DUP DUP Newest @
    OVER WindowReading @                \ Get 'latest' error
    OVER PrevVal @
    ROT WindowReading @                 \ Get 'previous' error
    -                                   \ Return (latest-previous)
    OVER Kd 2@ */                       \ Scale by derivative constant
  ELSE
    0
  ENDIF ;

\ Recalculate output from P, I, D terms
: CalcOutput	\ pid -- pid
  PropTerm
  DebugPID? IF
    CR ."   P        : " DUP 5 .r
  ENDIF

  >R                                    \ Get proportional term
  IntTerm
  DebugPID? IF
    CR ."   I        : " DUP 5 .r
  ENDIF

  >R                                    \ Get integral term
  DerivTerm
  DebugPID? IF
    CR ."   D        : " DUP 5 .r
  ENDIF

  R> R>                                 \ Get derivative term
  + +                                   \ Sum terms

  \ Check whether output is to be updated
  OVER Locked? 0= IF
    DebugPID? IF
      CR ."   Output   : " OVER 5 .r
    ENDIF
    Output !                            \ Update output
  ELSE
    2DROP                               \ In deadband, don't update
  ENDIF
;

: Kick		\ pid -- ; Single-shot the control algorithm
\ *G Single-shots the control algorithm for the controller
\ ** specified by 'pid'.
\ ** alls the installed routine to input the process variable
\ ** (see 'InputMethod' above), recalculates the output,
\ ** and calls the installed routine to process the output
\ ** (see 'OutputMethod' above).
  DUP VInput @ EXECUTE                  \ Execute user-supplied input routine
  OVER PV !                             \ Store result
  CalcError                             \ Calculate error from PV and SP

  DebugPID? IF
    CR ." PASS: " Pass @ 4 .R
    1 Pass +!
    CR ."   SP       : " DUP Setpoint @ 5 .r
    CR ."   PV       : " DUP PV @ 5 .r
    CR ."   Error    : " DUP PError @ 5 .r
  ENDIF

  CalcOutput                          \ Recompute output
  DUP Output @
  SWAP VOutput @ EXECUTE              \ Execute user-supplied output routine
;

: StopPID	\ pid --
\ *G For use in multitasking environments and interrupt routines.
\ ** Signals the controller to terminate its run on the next
\ **scanning loop.
  RunFlag off                         \ Store execution address
;

: RunPID	\ pid --
\ *G Put a controller into a continuous processing loop.
\ ** Primarily for use in multitasking environments. Puts the
\ ** current thread into an indefinite loop which can only be
\ ** terminated by executing 'Stop' on the controller from another
\ ** thread or an interrupt routine. In each iteration it
\ ** invokes the installed routine to check the sampling
\ ** trigger (see 'SetSamplingTrigger' above) and if it is set,
\ ** calls 'Kick' to recompute the output. Yields control
\ ** after each iteration.
  True OVER RunFlag !                   \ Signal 'running'
  BEGIN
    DUP VTrigger @ EXECUTE              \ Check trigger set for next iteration
    IF  DUP Kick  ENDIF                 \ Run one iteration
    PAUSE
    DUP RunFlag @ 0=                    \ Check not externally terminated
  UNTIL
  DROP                                  \ Discard controller address
;


\ ----------
\ Test words
\ ----------

\ Priming values for a 'square wave' test
9070 CONSTANT SP1	\ range is +/- 10,000 setpoint
1480 CONSTANT SP2


: InitController	\ pid --
\ *G Set up a 'stable' set of control parameters for the simulation provided.
  >r
  SP1      r@ SetSP			\ initial setpoint
  50       r@ SetDeadband		\ initial deadband
  406 1000 r@ SetKp			\ initial proportional coeff.
  169 1000 r@ SetKi			\ initial integral coeff.
  370 1000 r@ SetKd			\ initial derivative coeff.
  r> drop
  1 Pass !                              \ initialise count for debug mode
;

: Test		\ addr --
\ *G Run the controller in 'single shot' mode and report all variables
\ ** until the operator presses the <Esc> key.
  True to DebugPID?
  BEGIN
    DUP Kick
    KEY 27 =
  UNTIL
  DROP
;


\ ------------
\ Task example
\ ------------

((
\
\ A task would look like:
\
pid-controller: fred

: init-fred
  5000 fred SetSp                       \ define setpoint
  50 fred SetDeadband                   \ define deadband
  406 1000 fred SetKp                   \ initial proportional coeff.
  169 1000 fred SetKi                   \ initial integral coeff.
  370 1000 fred SetKd                   \ initial derivative coeff.
  ['] MyInput fred InputMethod          \ data acquisition word
  ['] MyOutput fred OutputMethod	\ data output word
;

: fred-task     \ -- ; the loop itself
  sp@ s0 !  rp@ r0 !                    \ so depth can be tested
  init-fred                             \ initialise all controllers
  begin
    wait-kick-timer                     \ let period timer elapse
    fred kick                           \ the controller we defined
    bert kick                           \   another controller
    george kick                         \   yet another controller
  again
;
))


\ ======
\ *> ###
\ ======

