\ Quotations.fth - nested definitions for various Forth systems

0 [if]
This is the commentary for comp.lang.forth
==========================================
A revised summary, 18 July 2012

The following notes on quotations and closures incorporate
material from several people, including Anton Ertl, Andrew Haley,
Marcel Hendrix, Ward Mcfarland, Bernd Paysan, and Elizabeth
Rather. If you have been left out and want acknowledgement, please
let me know.

Quotations
==========

Discussion
----------
The essence of quotations is to provide nested colon definitions,
in which the inner definition(s) are nameless. The expression

: foo ... [: some words ;] ... ;

is equivalent to

:noname some words ; Constant #temp#
: foo ... #temp# ... ;

A simple quotation is an anonymous colon definition that is
defined inside a colon definition or quotation.  It has no access
to locals of the enclosing definitions.

If quotations are to have the same privileges as words defined
with : and :NONAME they should be able to use local variables and
use RECURSE. Quotations should not be able to access the locals of
the outer word because we have no knowledge of when the quotation
is executed and hence whether outer locals are still alive.

One example use of quotations is to provide a solution to the use
of CATCH in a form close to other languages' TRY ... EXCEPT
blocks.

: foo     \ i*x -- j*x
  setup
  [: fee fi fo fum ;] catch
  if ... then
  teardown
  ( throw again )
;

Under the term "Phrases", quotations have been in MacForth since
at least 1987. They are documented in the infamous "missing
chapter" of the MacForth manual.

Specification
-------------
The following notations have been seen in the wild, oldest first.
  :| ... |;
  :[ ... ]:
  [: ... ;]

[: ( -- nested-sys )
Compilation: suspends compiling to the current definition, starts a new
nested definition, and compilation continues with this nested
definition. Outer locals are not visible in the nested definition.
Locals may be defined in the nested definition. Inside the nested
definition RECURSE applies to the nested definition.

;] ( nested-sys -- )
Compilation: Ends the current nested definition, and resumes compilation
to the previous current definition. It appends the following run-time to
the current definition:

run-time: ( -- xt )
xt is the execution token to execute the nested definition.


Closures
========

Discussion
----------
A closure is like a quotation, except that it has (read and write)
access to the locals of the enclosing definitions.

Simple quotations are relatively easy to implement, closures are hard,
because they can extend the life of the enclosed locals beyond the end
if their definition; in general garbage collection is needed to manage
the enclosed locals of closures.  Therefore everyone up to now has
only seriously thought about implementing and standardizing simple
quotations, and any discussions about closures are just theoretical
exercises.

I have chosen to distinguish nameless words intended as closures
by using [[: ... ;]] to define them. I have also chosen to
distinguish the required persistence of data by declaring it using
{{ ... }} which is syntactically like { ... } locals but has very
different persistence requirements.

: foo {{ a b -- }} ... [[: ... a b ... ;]] execute ;

Knuth's Man-Or-Boy test creates a quotation that *modifies* the
outer locals and then recursively calls its host, accessing outer
locals that one might have expected to long gone out of
scope. It forces the quotation implementation to store copies of
*all imaginable* outer locals and make them permanently available.

This can get even more interesting when a closure is required to
remember not only locals, but things like BASE, >IN, arbitrary
variables, file handles, buffer addresses, etc..(*)

There is a Forth example by Gerry Jackson which provides an
example of lambda definitions. This example is in ANS Forth and
runs on a range of systems.
  www.qlikz.org/forth/lambda.zip

Comments
--------
Maybe we first need a convincing example of the usefulness of
quotations and closures.

Aside from looking obscure and clever and satisfying the "[x]
can do it, why not Forth?" test, what on earth does this buy you?

What does a closure achieve that cannot be achieved by other
scoping and persistence mechanisms such as a class and an object
with heap allocation?

Specification
-------------

TBD
==========================================
[then]


\ **********
\ SwiftForth
\ **********

[defined] enter-swift [if]
\ The size of the compile-time locals structure
lvar-comp cell- @ sizeof constant lvar-size

\ Push and pop the  compile-time locals
\ This will leak memory if the compiler aborts
: push-locals ( - a t | 0)
    lvar-comp any? dup if
        lvar-comp  lvar-size allocate throw  dup >r  lvar-size move
        lvar-comp clear  r>  then ;
: pop-locals ( a t | 0)
    if  dup  lvar-comp  lvar-size  move  free throw  then ;

icode (quotation)
   here $400 + call
   ret   end-code

: [:   push-locals \ Save our local variables
    last 2 cells + @ \ Save last code address
    postpone (quotation)  postpone begin
    here last 2 cells + ! \ Last code address points to our quotation
    ;  immediate

: ;]    postpone exit
    postpone then \ Branch over the quotation
    last 2 cells + ! \ Restore last code address
    postpone r>  postpone code>
    pop-locals \ Restore our local variables
    ;  immediate
[then]


\ ********
\ bigForth
\ ********

[defined] bigforth [if]
: [: ( compile-time: -- orig colon-sys )
     state @ IF  loffset @ last @  POSTPONE AHEAD  true  ELSE  false  THEN
     :noname ; immediate

: ;] ( compile-time: orig colon-sys -- ; run-time: -- xt )
    POSTPONE ; >r IF ]  POSTPONE THEN  r> POSTPONE ALiteral  last ! loffset !
    ELSE  r>  THEN ( xt ) ; immediate
[then]


\ ******
\ gforth
\ ******

s" gforth" environment? [if]
  2drop
:noname  false :noname ;
:noname  locals-wordlist last @ lastcfa @
    postpone AHEAD
    locals-list @  locals-list off
    locals-dp @ dup >r     postpone SCOPE
    true  :noname  r> locals-dp ! ;
interpret/compile: [: ( -- quotation-sys )
\G Starts a quotation

: ;] ( compile-time: quotation-sys -- ; run-time: -- xt )
\g ends a quotation
    POSTPONE ; >r IF
	]  postpone ENDSCOPE
	locals-dp !  locals-list !
	postpone THEN
	lastcfa ! last ! to locals-wordlist
	r> postpone ALiteral
    ELSE  r>  THEN ( xt ) ; immediate
[then]


\ *********
\ VFX Forth
\ *********
\ This version supports locals in the quotation. The code requires
\ a build of 17 July 2012 or later.

[defined] VFXForth [if]
' localvars voc>wid wid-threads	2constant localThreads \ -- addr len

also system

: saveLThreads	\ -- ... addr
\ preserve compilation and locals state
\ This will leak memory if the compiler aborts
  <headerless> @  last @  last-colon-xt @
  tdp @  lvcount @  flvcount @  TDPstart @  current @
  localThreads dup ProtAlloc swap 	\ -- threads addr len
  over >r  move  r>			\ -- addr
  localThreads erase			\ wipe locals voc
  local-nests incr			\ increase locals nesting
  lvcount off  flvcount off		\ necessary resets
;

: restoreLThreads \ addr ... --
\ restore previous compilation and locals state
  dup localThreads move  ProtFree
  current !  TDPstart !  flvcount !  lvcount !  tdp !
  last-colon-xt !  last !  <headerless> !
  local-nests decr
;

previous

: [:		\ comp: -- i*x orig colon-sys
  state @ IF
    saveLThreads			\ -- i*x
    POSTPONE AHEAD  true
  ELSE
    false
  THEN
  postpone [  :noname
; immediate

: ;]		\ comp: i*x orig colon-sys -- ; run-time: -- xt
  discard-sinline  POSTPONE ;  >r   IF
    ]  POSTPONE THEN  r> POSTPONE Literal
    restoreLThreads
    discard-sinline
  ELSE
    r>
  THEN ( xt )
; immediate

[then]


\ *********************
\ Compatibility for all
\ *********************

Synonym  :| [:
Synonym  |: ;]

Synonym  :[ [:
Synonym  ]: ;]


\ *********
\ Test code
\ *********

0 [if]

: t1		\ a b -- c
  {: a b -- :}
  [:
    {: | a b -- :}
    100 to a  200 to b
    cr ." Mul = " * .
    ."  a*b = " a b * .
  ;] >r
  a b r@ execute
  ."  a b " a . b .
  r> drop
;

[then]
