\ ASM8687.FTH - 386/387 Assembler

((
Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2009, 2012, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20200402 SFP068 Added section boundaries for x64 compatibility.
20120203 SFP016 Added CMPXCHG and friends.
20091113 SFP015	Added RDMSR and WRMSR instructions.
20071017 SFP014 Updated documentation. Added P4?
		Added compiler alternatives for speed/size.
		Renamed REG? to AREG? to avoid name conflict in optimiser.
20060823 SFP013 Fixed FDIV/R and FSUB/R for swapped operands.
20060804 SFP012	Allowed scaled index without base register.
		Corrected DR1? and DIR2? for SIB byte usage.
20050905 SFP011 Added CMOVcc instructions.
		Added CPU selection.
20040705 SFP010 Fixed ASM]
20040309 SFP009 Added locals protection to ;CODE
20030804 SFP008 Forced CODE definitions to start on 4 byte boundary
20020221 SFP007 Finally fixed FSUB/R reg, reg opcodes
20011228 SFP006 Fixed bug for immediate shift by 0
20010415 SFP005 Added DISCARD-SINLINE to ;CODE
20010321 SFP003 Converted for cross compiler
20000524 SFP002 Changed 3MODES/L for AND OR XOR to permit short forms
20000523 SFP001 Removed >ASM >FORTH which were DTC relics.
                Added MASM: ;MASM [ASM ASM] A[ ]A
99/05/25 SJC000 Changed default mode to INTEL PREFIX
))

( *! asm8687                                                            )
( *T Intel 386+ Assembler                                               )
( **                                                                    )
( *P VFX Forth has a built-in assembler. This is to enable you to write  )
( ** time-critical definitions - if time is a constraint - or to do     )
( ** things that might perhaps be more difficult in Forth - things such )
( ** as interrupt service routines. The assembler supports the 80386    )
( ** and the 80387 chips. Definitions written in assembler may use all  )
( ** the variables, constants, etc. used by the Forth system, and may   )
( ** be called from the keyboard or from other words just like any      )
( ** Forth high-level word. It is important when writing a code         )
( ** definition to remember which machine registers are used by the     )
( ** Forth system itself. These registers are documented later in this  )
( ** chapter. All other registers may be used freely. The reserved      )
( ** registers may also be used - but their contents must be preserved  )
( ** while they are in use and reset afterwards.                        )
( **                                                                    )
( *P The assembler mnemonics used in the Forth assembler are just the   )
( ** same as those documented in the Intel literature. The operand      )
( ** order is also the same. The only difference is the need for        )
( ** a space between each portion of the instruction. This is a         )
( ** requirement of the Forth interpreter.                              )
( **                                                                    )
( *P The assembler has certain defaults. These cover the order of the   )
( ** operands, the default addressing modes and the segment size. These )
( ** are described later in this chapter.                               )

\ **********************
\ *S Using the assembler
\ **********************
\ *P Normally the assembler will be used to create new Forth words
\ ** written in assembler. Such words use *\fo{CODE} and
\ ** *\fo{END-CODE} in place of *\fo{:} and *\fo{;} or *\fo{CREATE}
\ ** and *\fo{;CODE} in place of *\fo{CREATE} and *\fo{DOES>}.

\ *P The word *\fo{CODE} creates a new dictionary header and enables
\ ** the assembler.

\ *P As an example, study the definition of *\fo{0<} in assembly
\ ** language. The word *\fo{0<} takes one operand from the stack
\ ** and returns a true value, -1, if the operand was less than
\ ** zero, or a false value, 0, if the operand was greater than
\ ** or equal to zero.

( *E CODE 0< \ n - t/f ; define the word 0<                             )
( **   OR EBX, EBX   \ use OR to set flags                              )
( **   L,            \ less than zero ?                                 )
( **   IF,           \ y:                                               )
( **     MOV EBX, # -1       \   -1 is true flag                        )
( **   ELSE,         \ n:                                               )
( **     SUB EBX, EBX        \  dirty set to 0                          )
( **   ENDIF,                                                           )
( **   NEXT,         \ return to Forth                                  )
( ** END-CODE                                                           )

\ *P Notice how the word *\fo{NEXT,} is used. *\fo{NEXT,} is a
\ ** macro that assembles a return to the Forth inner interpreter.
\ ** All code words must end with a return to the inner interpreter.
\ ** The example also demonstrates the use of structuring words
\ ** within the assembler. These words are pre-defined macros
\ ** which implement the necessary branching instructions. The
\ ** next example shows the same word, but implemented using local
\ ** labels instead of assembler structures for the control
\ ** structures.

( *E CODE 0<         \ n - t/f ; define the word 0<                     )
( **   OR EBX, EBX   \ use OR to set flags                              )
( **   JGE L$1       \ skip if AX>=0                                    )
( **   MOV EBX, # -1         \ -1 is true flag                          )
( **   JMP L$2       \ this part done                                   )
( ** L$1:            \ do following otherwise                           )
( **   SUB EBX, EBX  \ dirty set to 0                                   )
( ** L$2:                                                               )
( **   NEXT,         \ return to Forth                                  )
( ** END-CODE                                                           )

\ ****************************
\ *S Assembler extension words
\ ****************************
( *P There are several useful words provided within VFX Forth to control )
( ** the use of the assembler.                                          )

\ *C ;code      \ --
\ *P Used in the form:
\ *C   : <namex> CREATE .... ;CODE ... END-CODE
\ *P Stops compilation, and enables the assembler. This word is
\ ** used with *\fo{CREATE} to produce defining words whose run-time
\ ** portion is written in code, in the same way that *\fo{CREATE ... DOES>}
\ ** is used to create high level defining words.

\ *P The data structure is defined between *\fo{CREATE} and
\ ** *\fo{;CODE} and the run-time action is defined between
\ ** *\fo{;CODE} and *\fo{END-CODE}. The current value of the
\ ** data stack pointer is saved by *\fo{;CODE} for later use by
\ ** *\fo{END-CODE} for error checking.

\ *P When *\fo{<namex>} executes the address of the data area will
\ ** be the top item of the CPU call stack. You can get the address
\ ** of the data area by *\fo{POP}ing it into a register.

\ *P A definition of *\fo{VARIABLE} might be as follows:
( *E : VARIABLE                                                         )
( **   CREATE 0 ,                                                       )
( ** ;CODE                                                              )
( **   sub     ebp, 4                                                   )
( **   mov     0 [ebp], ebx                                             )
( **   pop     ebx                                                      )
( **   next,                                                            )
( ** END-CODE                                                           )
( **                                                                    )
( ** VARIABLE TEST-VAR                                                  )

( *C CODE       \ --                                                    )
( *P A defining word used in the form:                                  )
( *C CODE <name> ... END-CODE                                           )
\ *P Creates a dictionary entry for *\fo{<name>} to be defined by
\ ** a following sequence of assembly language words. Words defined
\ ** in this way are called *\b{code} definitions. At compile-time,
\ ** *\fo{CODE} saves the data stack pointer for later error
\ ** checking by *\fo{END-CODE}.

\ *C END-CODE   \ --
\ *P Terminates a code definition and checks the data stack pointer
\ ** against the value stored when *\fo{;CODE} or *\fo{CODE} was
\ ** executed. The assembler is disabled. See: *\fo{CODE} and
\ ** *\fo{;CODE}.

( *C LBL:       \ --                                                    )
( *P A defining word that creates an assembler routine that can be      )
( ** called from other code routines as a subroutine. Use in the form:  )
( *C LBL: <name>                                                        )
( *C   ...code...                                                       )
( *C END-CODE                                                           )
\ *P When *\fo{<name>} executes it returns the address of the first
\ ** byte of executable code. Later on another code definition can
\ ** call *\fo{<name>} or jump to it.

\ ****************************
\ *S Dedicated Forth registers
\ ****************************
\ *P The Forth virtual machine is held within the processor register set.
\ ** Register usage is as follows:
\ *D EAX	scratch
\ *D EBX	cached top of data stack
\ *D ECX	scratch
\ *D EDX	scratch
\ *D ESI	Forth User area pointer
\ *D EDI	Forth local variable pointer
\ *D EBP	Forth data stack pointer - points to NOS
\ *D ESP	Forth return stack pointer

\ *P All unused registers may be freely used by assembler routines, but they may
\ ** be altered by the operating system or wrapper calls. Before calling the
\ ** operating system, all of the Forth registers should be preserved. Before
\ ** using a register that the Forth system uses, it should be preserved and then
\ ** restored on exit from the assembler routine. Be aware, in particular, that
\ ** callbacks will generally modify the EAX register since this is used to hold
\ ** the value returned from them.

\ ***********************
\ *S Default segment size
\ ***********************
\ ** The assembler is designed to be able to generate code for the 80386.
\ ** However, this processor is able to support both 32-bit segments and the
\ ** 16-bit segments associated with the older processors in the family. It is
\ ** therefore important to specify to the assembler which segment size code is
\ ** being generated for. The default segment size is 32-bit as VFX Forth runs as
\ ** a flat-model task. However, to modify the default segment size, the
\ ** following directives are supported:
\ *C USE32
\ *C USE16
\ *P The first of these specifies that code from that point onwards is for a
\ ** 32-bit segment. The second directive specifies that, from that point
\ ** onwards, code generated is for a 16-bit segment. The default is
\ ** *\fo{USE32}. These directives should be used outside a code
\ ** definition, not within a definition.

\ *P It is possible to override the default segment size on an
\ ** instruction-by-instruction basis. This is detailed later.

\ *******************
\ *S Assembler syntax
\ *******************
\ =============================
\ *N Default assembler notation
\ =============================
\ *P The assembler is designed to be very closely compatible with MASM and other
\ ** assemblers. To this end the assembler assembles code written in the
\ ** conventional prefix notation. However, because code may be converted from
\ ** other MPE Forth systems, the postfix notation is also supported. The default
\ ** mode is prefix. The directives to switch mode are as follows:
\ *C PREFIX
\ *C POSTFIX
\ *P These switch the assembler from then onwards into the new mode. The
\ ** directives should be used outside a code definition, not within one. Their
\ ** use within a code definition will lead to unpredictable results.

\ *P The assembler syntax follows very closely that of other 80386 assemblers.
\ ** The major difference being that the VFX Forth assembler needs white space
\ ** around everything. For example, where in MASM one might define:
\ *E MOV EAX,10[EBX]
\ *P we must write:
\ *E MOV EAX , 10 [EBX]
\ *P This distinction must be borne in mind when reading the following addressing
\ ** mode information.

\ =======================
\ *N Register to register
\ =======================
\ *P Many instructions have a register to register form. Both operands are
\ ** registers. Such an instruction is of the form:
\ *E   MOV EAX , EBX
\ *P This moves the contents of EBX into EAX. For compatibility with older MPE
\ ** assemblers the first operand may be merged with the comma thus:
\ *C   MOV EAX, EBX

\ *P This use of a register name with a 'built-in' comma also applies to other
\ ** addressing modes.

\ =================
\ *N Immediate mode
\ =================
\ *P If the assembler is set for direct-as-default (the *\fo{MPE}
\ ** directive has been used), immediate numbers must be defined
\ ** explicitly. This is done by the use of a hash (#) character:
\ *C MOV EAX , # 23
\ *P This example places the number 23 in EAX. The directives
\ ** *\fo{OFFSET} and *\fo{SEG} are synonyms for *\fo{#}.

\ *P By default, the assembler is set to immediate-as-default (the
\ ** *\fo{INTEL} directive has been used). In this case immediate
\ ** numbers do not have to be specifically defined:
\ *C   MOV EAX , 23
\ *P The above code also places the number 23 in EAX.

\ ==============
\ *N Direct mode
\ ==============
\ *P If the assembler is set for direct-as-default (the *\fo{MPE}
\ ** directive has been used), direct addresses need not be
\ ** defined explicitly:
\ *C   MOV EAX , 23
\ *P This example places the contents of address 23 in EAX. If the
\ ** assembler is  set for immediate-as-default (the *\fo{INTEL}
\ ** directive has been used), direct addresses have to be
\ ** specifically defined, using the *\fo{PTR} or *\fo{[]} directives:
\ *C   MOV EAX , PTR 23
\ *C   MOV EAX , [] 23
\ *P Both the above code fragments also place the contents of address 23 in EAX.

\ ======================
\ *N Base + displacement
\ ======================
\ *P Intel define an addressing mode using a base and a displacement. In this
\ ** mode, the effective address is calculated by adding the displacement to the
\ ** contents of the base register. An example:
\ *C   MOV EBX , # 0100
\ *C   MOV EAX , 10 [EBX]
\ *P In this example, EAX is filled with the contents of address 0100+10, or
\ ** address 110.

\ *P The assembler lays down different modes for displacements of 8-bit or 32-bit
\ ** size, but this is internal to the assembler.
\ ** The following registers may be used as base registers with a displacement:

\ *C [EAX] [ECX] [EDX] [EBX] [EBP] [ESI] [EDI]

\ *P If the displacement is zero then the assembler internally defines the mode
\ ** as Base only. However, the displacement of zero must be supplied to the
\ ** assembler:
\ *C MOV EBX , # 0100
\ *C MOV EAX , 0 [EBX]
\ *P This places in EAX the contents of address 100 (pointed to by EBX).

\ *P The following registers may be used as a base with no displacement:
\ *C [EAX] [ECX] [EDX] [EBX] [ESI] [EDI]

\ ==============================
\ *N Base + index + displacement
\ ==============================
\ *P The 80386 also allows two registers to be used to indirectly address memory.
\ ** These are known as the base and the index. Such instructions are of the
\ ** form:
\ *C MOV EAX , # 100
\ *C MOV EBX , # 200
\ *C MOV EDX , 10 [EAX] [EBX]

\ *P This will place in EDX the contents of address 100+200+10, or address 310.
\ ** EAX is the base and EBX is the index. Again, the displacement may be 8-bits,
\ ** 32-bits or have a value of zero. The assembler distinguishes between these
\ ** three cases. The base and index registers may be any of the following:
\ *C [EAX] [EBX] [ECX] [EDX] [ESI] [EDI]
\ *P In addition, [EBP] may be used as the index register, and [ESP] may be used
\ ** as the base register.

\ ====================================
\ *N Base + index*scale + displacement
\ ====================================
\ *P The 80386 further supports an addressing mode where the index register is
\ ** automatically scaled by a fixed amount - either 2, 4 or 8. This is designed
\ ** for indexing into two-dimensional arrays of elements of size greater than
\ ** byte-size. One register may be used as the first index, another for the
\ ** second index, and the word size becomes implicit in the instruction. The
\ ** form of this addressing mode is very similar to that outlined above, with
\ ** the exception that the index operand includes the number which is the scale:
\ *C   MOV EBX , # 100
\ *C   MOV ECX , # 2
\ *C   MOV EAX , 10 [EBX] [ECX*4]
\ *P This stores into EAX, the contents of address 100+(4*2)+10, or address 118.
\ ** The list of registers which may be used as base is the same as the above.
\ ** The list of scaled indexes is as follows:
\ *C   [EAX*2] [ECX*2] [EDX*2] [EBX*2] [EBP*2] [ESI*2] [EDI*2]
\ *C   [EAX*4] [ECX*4] [EDX*4] [EBX*4] [EBP*4] [ESI*4] [EDI*4]
\ *C   [EAX*8] [ECX*8] [EDX*8] [EBX*8] [EBP*8] [ESI*8] [EDI*8]

\ ====================
\ *N Segment overrides
\ ====================
\ *P Some instructions may be prefixed with a segment override. These force data
\ ** addresses to refer to a segment other than the data segment. The override
\ ** must precede the instruction to which it relates:
\ *C   MOV EBX , # 100
\ *C   ES: MOV EAX , 10 [EBX]
\ *P This will set EAX to the value contained in address 110 in the extra
\ ** segment. The list of segment overrides is:
\ *C   CS: DS: ES: FS: GS: SS:

\ ======================
\ *N Data size overrides
\ ======================
\ *P The default data size for a*\fo{ USE32} segment is 32-bit, but
\ ** the default data size for a *\fo{USE16} segment is 16-bit. These
\ ** are the default data sizes the assembler will use. If the data
\ ** is of a different size a data size override will have to be used.
\ ** To define the size of the data the following size specifiers
\ ** are used:
\ *C   BYTE or B.
\ *C   WORD or W.
\ *C   DWORD or D.
\ *C   QWORD
\ *C   TBYTE
\ *C   FLOAT
\ *C   DOUBLE
\ *C   EXTENDED

\ *P It is only necessary to specify size when ambiguity would
\ ** otherwise arise. For example:
\ *E    MOV   0 [EDX], # 10	\ can't tell
\ **    MOV   0 [EDX], EAX	\ EAX specifies
\ *P The *\fo{BYTE} size defines that a byte operation is required:
\ *C   MOVZX EAX , BYTE 10 [EBX]

\ *P The abbreviation *\fo{B.} may also be used in place of
\ ** *\fo{BYTE} to define a byte operation. The *\fo{WORD}
\ ** specifier defines that 16-bits are required:
\ *C   MOV AX , WORD 10 [EBX]

\ *P The abbreviation *\fo{W.} may also be used to define a word
\ ** operation. *\fo{DWORD} is the default for a *\fo{USE32}
\ ** segment, and indicates that 32-bit data is to be used:
\ *C MOV EAX , DWORD 10 [EBX]
\ *C FSTP DWORD 10 [EBX]
\ *P The abbreviation *\fo{D.} may also be used to specify a
\ ** *\fo{DWORD} operation. The remaining size specifiers define
\ ** data sizes for the floating point unit.

\ *P *\fo{QWORD} defines a 64-bit operation:
\ *C FSTP QWORD 10 [EBX]
\ *P TBYTE defines a 10-byte (80-bit) operation, such as:
\ *C FSTP TBYTE 10 [EBX]

\ *P *\fo{FLOAT}, *\fo{DOUBLE} and *\fo{EXTENDED} are synonyms for
\ ** *\fo{DWORD}, *\fo{QWORD} and *\fo{TBYTE} respectively.

\ *P The segment type defines the default data size and address size for the code
\ ** in the segment. If needed, it is possible to force the data size or the
\ ** address size laid down to be the other. There is a set of data and address
\ ** size overrides which work for one instruction only. These are:
\ *C   D16:
\ *C   D32:
\ *C   A16:
\ *C   A32:
\ *P and they would be used as follows:
\ *C   D16: MOV EAX , # 23
\ *C   A16: MOV EAX , 10 [EBX]

\ *P The first of these, in a *\fo{USE32} segment, would lay down
\ ** 16-bit data to be loaded into AX. The second would lay down
\ ** a 16-bit offset from [EBX] for the effective address in the
\ ** instruction. The situation would be reversed in a *\fo{USE16}
\ ** segment - the *\fo{A32:} and *\fo{D32:} directives would cause
\ ** 32-bit data or addresses to be laid.

\ ===============================
\ *N Near and far, long and short
\ ==============================
\ *P Jumps and branches may be either intra-segment or inter-segment. The former
\ ** is a short branch or call whilst the latter is a long branch or call. The
\ ** assembler is able to lay down either form. The default for a JMP or a CALL
\ ** is near, whilst the default for a conditional branch is short. RET follows
\ ** the same pattern as CALL. The directives supporting short/long and near/far
\ ** are:
\ *C   SHORT  LONG  NEAR  FAR

\ *P These would be used as follows:
\ *E 2 CONSTANT THAT         \ the segment number
\ ** LBL: THIS               \ the address
\ **
\ ** CALL THIS
\ ** CALL NEAR THIS
\ ** CALL FAR THAT THIS
\ ** JMP THIS
\ ** JMP NEAR THIS
\ ** JMP FAR THAT THIS
\ **
\ ** JCC THIS
\ ** JCC SHORT THIS
\ ** JCC LONG THIS
\ **
\ ** RET THIS
\ ** RET NEAR THIS
\ ** RET FAR THAT THIS

\ *P For compatibility with older MPE assemblers the mnemonics
\ ** *\fo{CALL/F}, *\fo{RET/F} and *\fo{JMP/F} are also provided.

\ ====================
\ *N Syntax exceptions
\ ====================
\ *P The assembler in VFX Forth follows both the syntax and the mnemonics defined
\ ** in the Intel Programmers Reference books, for both the 80386 and the 80387.
\ ** However, there are certain exceptions. These are listed below.

\ *P The zero operand forms of certain stack register instructions for the 80387
\ ** have been omitted. Their functionality is supported however. Such
\ ** instructions are listed below, with a form of the syntax which will support
\ ** the function:
\ *E FADD    FADDP ST(1) , ST
\ ** FCOM    FCOM ST(1)
\ ** FCOMP   FCOMP ST(1)
\ ** FDIV    FDIVP ST(1) , ST
\ ** FDIVR   FDIVRP ST(1) , ST
\ ** FMUL    FMULP ST(1) , ST
\ ** FSUB    FSUBP ST(1) , ST
\ ** FSUBR   FSUBRP ST(1) , ST

\ ** Certain 80386 instructions have either one operand or two operands, of which
\ ** only one is variable. These instructions are:
\ *C MUL DIV IDIV NEG NOT
\ *P These instructions take only one operand in the VFX Forth assembler.

\ ===============
\ *N Local labels
\ ===============
\ *P If you need to use labels within a code definition, you may use the local
\ ** labels provided. These are used just like labels in a normal assembler, but
\ ** some restrictions are applied.

\ *P Ten labels are pre-defined, and their names are fixed. Additional
\ ** labels can be defined up to a maximum of 32. There is a limit of 128 forward
\ ** references.

\ ** A reference to a label is valid until the next occurrence of
\ ** *\fo{LBL:}, *\fo{CODE} or *\fo{;CODE}, whereupon all the
\ ** labels are reset.

\ *P A reference to a label in a definition must be satisfied in that definition.
\ ** You cannot define a label in one code definition and refer to it from
\ ** another.

\ *P The local labels have the names *\fo{L$1 L$2 ... L$10} and
\ ** these names should be used when referring to them e.g.
\ *C   JNE L$5

\ *P A local label is defined by words of the same names, but with a colon as a
\ ** suffix:
\ *C   L$1: L$2: ... L$10:

\ *P Additional labels (up to a maximum of 32 altogether) may be referred to by:
\ *C   n L$
\ *P where n is in the range 11..32 (decimal), and they may be defined by:
\ *C n L$:
\ *P where n is again in the range 11..32 (decimal).

\ ================
\ *N CPU selection
\ ================
\ *P This assembler is designed to cope with CPUs from 80386
\ ** upwards. Some instructions are only available on later CPUs.
\ ** Note that CPU selection affects the assembler and the VFX code
\ ** code generator, *\b{not} the run time of your application. If
\ ** you select a higher CPU level than the application runs on,
\ ** incorrect operation will occur.
\ *C  CPU=386     \ -- ; select base instruction set
\ *C  CPU=PPro    \ -- ; Pentium Pro and above with CMOVcc
\ *C  CPU=P4      \ -- ; Pentium 4 and above
\ *C  PPro?       \ -- flag ; true if at least Pentium Pro
\ *C  P4?         \ -- flag ; true if at least Pentium 4

\ *P The VFX code generator also uses this information to enable
\ ** various code generation techniques. For VFX Forth for DOS,
\ ** the default selection is for 386 class CPUs, for all others
\ ** it is for the Pentium 4 instruction set.


only forth definitions

vocabulary asm-core
vocabulary asm-access

only forth definitions

\ the following words allow second compilation for overlay
((
: ,,            \ n -- ; stash n in dictionary
  ,
;
interpreter
: ,,            \ n -- ; stash n in dictionary
  ,
;
target
))

: f-and         \ n1 n2 -- n3 ;
  and
;

: f-or          \ n1 n2 -- n3 ;
  or
;

\ SFP014...
compiler
: f-and         \ n1 n2 -- n3 ;
  and
;

: f-or          \ n1 n2 -- n3 ;
  or
;
target
\ ...SFP014


((
: f-loop        \ sys
  [compile] loop
; immediate

: f-leave       \ sys
  [compile] leave
; immediate
))

only forth also asm-core definitions


\ ***********
\ Error codes
\ ***********

#-701 #ErrDef asm_invAddrMode	"ASM: Invalid addressing mode"
#-702 #ErrDef asm_inv8bitOffset	"ASM: Number/offset not in range -128..+127"
#-703 #ErrDef asm_invLabelRef	"ASM: Label reference number out of range"
#-704 #ErrDef asm_invLabelDef	"ASM: Label definition number out of range"
#-705 #ErrDef asm_invCPU	"ASM: Invalid for selected CPU"


\ *************
\ CPU selection
\ *************

\ SFP011...
also forth definitions

\ These MUST match the values in XASM386.FTH
0
  dup equ 386cpu	1+	\ base
  dup equ 486cpu	1+	\ BSWAP XADD CMPXCHG INVD INVLPG
  dup equ P1cpu		1+	\ CMPXCHG8B CPUID RDTSC RDMSR WRMSR
  dup equ PProcpu	1+	\ CMOVcc
  dup equ P2cpu		1+	\ MMX
  dup equ P3cpu		1+	\ SSE
  dup equ P4cpu		1+	\ SSE2
drop

\ p4cpu
CPUtype  value CPUtype	\ selects CPU level
: CPU=386	( -- )  386cpu to CPUtype  ;
: CPU=486	( -- )  486cpu to CPUtype  ;
: CPU=PPro	( -- )  PProCpu to CPUtype  ;
: CPU=P4	( -- )  P4cpu to CPUtype  ;

: PPro?		( -- flag )  CPUtype PProCPU >=  ;
: P4?		( -- flag )  CPUtype P4cpu >=  ;

previous definitions
\ ...SFP011


\ *********
\ Variables
\ *********

variable seg-type
  seg-type on                   \ force it to USE32

variable d-override             \ have specified a datasize override
variable a-override             \ have specified an addresssize override

variable far/near               \ far or near call
variable short/long             \ short or long branch

variable default-mode           \ mpe/intel defaults
  default-mode on                                                               \ SJC000

variable r1                     \ number of first register
variable r2                     \ number of second register
variable r3                     \ number of third register
variable mode1                  \ mode for operand one
variable mode2                  \ mode for operand two
variable mode3                  \ mode for operand three
variable b/w/d1                 \ size of register operand one
variable b/w/d2                 \ size of register operand two
variable b/w/d3                 \ size of register operand three
variable (b/w/d1)               \ size of register operand one, before forcing
variable w/d1                   \ size of memory operand one
variable w/d2                   \ size of memory operand two

variable adr-mod                \ which addressing mode to use
variable segreg                 \ which segement override is in use
variable sib-val                \ value of sib byte
variable sib-flag               \ says we have a sib for this instruction
variable 1-flag                 \ have we had a register for op1
variable 2-flag                 \ have we had a register for op2
variable #ops                   \ how many registers have we used
variable base-count             \ how many bases/indexes
variable special-reg            \ where a special register is used, op1 or op2
variable special-type           \ what type of special reg
variable memoryformat           \ datasize for '387 instructions
variable last-fwait             \ address of last FWAIT instruction
variable got-[]                 \ user has defined   [] nnn  as operand
variable co1                    \ user has specified a op1 with a ,
variable co2                    \ user has specified a op2 with a ,
variable ret-imm-flag           \ in case we have a ret # n instruction

defer (reset)

decimal

create cr-lf            \ -- addr ; data to make cr/lf in file
  13 c, 10 c,                   \ a cr by any other name

variable asm-debug              \ says whether debugging or not

: space-pad             \ -- ; move pile of spaces to here
  pad off                               \ turn off
  30 out @ - 0 max dup pad c!           \ generate count and count byte
  pad 1+ swap bl fill                   \ fill rest with spaces
;

: write-text            \ addr len -- ; write data to log
  2drop
;

hex

create sp               \ the space string, for writing to file
  bl c,                         \ is a counted string of one space

: .hex                  \ n -- ; used for log file only
  3 out +!                              \ increment counter
  base @ >r hex                         \ preserve base
  0ff f-and s>d <# # # #> write-text    \ write number
  sp 1 write-text                       \ write space
  r> base !                             \ restore base
;

: log-line              \ -- ; maybe log the i/p line to the file
  asm-debug @
  if
    space-pad                           \ work out how many spaces
    pad 1+ pad c@ write-text            \ pad to col.30
    tib #tib @ write-text               \ write text to file
    cr-lf 2 write-text                  \ and a new-line
    out off                             \ back to start of line
  endif
;

only forth also asm-core also forth definitions


only forth also asm-core definitions

variable <prefix>               \ prefix/postfix mode
         <prefix> on            \
create aprior  0 ,  0 ,         \ see ?prefix
create last-pos 0 ,  0 ,        \ holds BLK & >IN
variable pos-#tib               \ -- addr ; holds previous #TIB
create pos-buff  $0100 allot    \ -- addr ; holds previous TIB


\ save-buff save-pos restore-pos

: save-pos      \ -- ; preserves opcode position in text
  blk @  >in @  last-pos 2!
  blk @ 0=                              \ if not in screen file
  if  tib pos-buff #tib @ $100 min cmove \ save TIB
      #tib @ pos-#tib !                 \ save #TIB
  endif  ;

: restore-pos   \ -- ; restore text position
  last-pos 2@  >in !  dup blk !  0=     \ if in text file
  if  pos-buff tib pos-#tib @ $100 min cmove \ restore TIB
      pos-#tib @ #tib !
  endif  ;

: prefix?       \ -- t/f ; true if in prefix mode
  <prefix> @
;

: ?prefix       \ n -- n' ; executes previous opcode
  prefix? if
    r>
    aprior 2@  2swap  aprior 2!
    >r
  endif
;

: null-op       \ -- ; run dummy for labels etc
  0 ?prefix drop                \ execute nothing
  (reset)                       \ do a reset
;

: kickoff       \ -- ; dummy operator used as initial setting
  drop  (reset)
;

((
: set-null	\ -- ; initialise APRIOR
\ >BODY only needed to step over JMP DOCOLON in DTC systems
  0 ['] kickoff >body aprior 2!         \ set noop into last opcode
;
))
: set-null      \ -- ; initialise APRIOR ; SFP001
  0 ['] kickoff aprior 2!               \ set noop into last opcode
;

: asm-start     \ -- ; initialise assembler
  set-null save-pos
  (reset)
;

: asm-end       \ -- ; reset assembler
  null-op
;

: here(t)       \ -- addr ;
  here                          \ just here for host comp
;

: allot(t)      \ n -- ;
  allot
;

: !(t)          \ 32b addr -- ;
  !                             \ 32b store
;

: @(t)          \ addr -- 32b ;
  @                             \ 32b store
;

: w!(t)         \ 16b addr -- ;
  w!                            \ map to w! for 32b forth
;

: w@(t)         \ addr -- 16b ;
  w@                            \ maps to w@ for 32b forth
;

: c!(t)         \ 8b addr -- ;
  c!                            \ still a c! for 32b forth
;

: c@(t)         \ addr -- 8b ;
  c@                            \ still a c@ in 32 forth
;

hex

-short-branches
: ,(t)		\ 32b -- ;
  asm-debug @ if
    dup 0ff f-and .hex                  \
    dup 0ff00 f-and 100 / .hex          \
    dup 0ff0000 f-and 10000 / .hex      \
    dup 0ff000000 f-and 1000000 / .hex  \ .... debug .....
  endif
  ,                                     \ 32b forth
;
+short-branches

: w,(t)         \ 16b -- ;
  asm-debug @ if
    dup 0ff f-and .hex          \
    dup 0ff00 f-and 100 / .hex  \ ..... debug .....
  endif
  w,                            \ maps to w, for 32b forth
;

: c,(t)         \ 8b -- ;
  asm-debug @ if
    dup .hex                    \ debug only
  endif
  c,                            \ still a c, in 32b forth
;

hex

: a-error       \ n -- ; replacement for ERROR
  prefix?
  if  restore-pos  endif
  decimal
  throw				\ SFP012
\  cr ." Assembler error code: " .
\  cr true abort" Aborted"
;

: ?a-error      \ t/f n -- ; replacement for ?ERROR
  swap if
    a-error                     \ if an error, do a-error
  else
    drop                        \ otherwise just tidy up
  endif
;


\ ======================
\ Segment and data sizes
\ ======================

: short?        \ n -- t/f  ; true if in range -128..+127
  -080 07F within?
;

: ?short        \ n -- ; error if n not in -128..+127
  short? 0= asm_inv8bitOffset ?a-error
;


: use32?        \ -- t/f ; is this a use32 segment?
  seg-type @                    \ return the flag
;

: use16?        \ -- t/f ; is this a use16 segment?
  seg-type @ 0=                 \ return inverse of flag
;


\ ======================
\ Reference calculations
\ ======================

: absolute-data \ 32b -- ; stash data
  use32? if
    d-override @ if                     \ if override
      w,(t)                             \ stash 16b
    else
      ,(t)                              \ normal: stash 32b in dict.
    endif
  else
    d-override @ if                     \ if override
      ,(t)                              \ stash 32b
    else
      w,(t)                             \ normal: stash 16b in dict.
    endif
  endif
;

: absolute-addr	\ 32b -- ; stash addr
  use32? if
    a-override @ if                     \ if override
      w,(t)                             \ stash 16b
    else
      ,(t)                              \ normal: stash 32b in dict.
    endif
  else
    a-override @ if                     \ if override
      ,(t)                              \ stash 32b
    else
      w,(t)                             \ normal: stash 16b in dict.
    endif
  endif
;

: rel8          \ addr -- ;
  here(t)  1+ -                 \ offset from here to addr
  dup ?short                    \ test for 8b relative
  c,(t)				\ stash addr here
;

: rel16/32	\ addr -- ;
  here(t)                       \ addr of this byte
  use32?
  if    4+                      \ offset of 4 for 32bit addr
  else  2+                      \ or offset of 2 for 16bit addr
  endif
  -                             \ offset from here to addr
  absolute-addr
;

previous definitions


\ ===================================
\ Forward references and local labels
\ ===================================

also asm-core definitions

128 constant maxref      \ no. of forward references
create fwdrefs
  maxref  8 * allot
  fwdrefs maxref 8 * erase

\ fwdrefs structure:
\ ____ ____ ____ ____
\     |    |    |    |
\ l#  |addr|l#  |addr|     :: l# == label number, addr == address of label reference
\ ____|____|____|____|
\ 0    4    8              :: label#+`fwd' (see defs below)


32 constant maxlab      \ no. of local labels
create labtab
  maxlab cells allot
  labtab maxlab cells erase

\ labtab structure:
\ ____ ____ ____ ____
\     |    |    |    |
\ addr|addr|addr|addr|     :: addr == address of label definition
\ ____|____|____|____|
\ 0    4    8              :: label#+`label' (see defs below)

: fwd           \ label# -- addr ; `fwd'
  8 * fwdrefs + ;

: label         \ label# -- addr ; `label'
  cells labtab + ;

hex

: unres-refs?   \ -- t/f ; true if unres
  0  maxref 0                           \ for the number allowed
  do
    i fwd @                             \ get the reference
    if  drop -1 leave  endif            \ if unresolved, set true flag
  loop
;

: initfwd	\ -- ; initialize tables
  fwdrefs maxref 8 * erase               \ clear out both tables
  labtab  maxlab cells erase
;
\ initfwd

: (offset)      \ -- 3/5 ;
  use32? if                     \ if 32bit segment
    5                           \ need 5 to calculate 32b offset
  else                          \ if 16bit segment
    3                           \ need 3 to calculate 16b offset
  endif
;

: ?!(t)		\ 16/32b addr -- ;
  use32? if                     \ if 32b segment
    !(t)                        \ stash 32bits
  else                          \ if 16bit segment
    w!(t)                       \ stash 16bits
  endif
;

hex

-short-branches
: resolve       \ n -- n ; resolve forward references
  maxref  0                             \ for each forward reference
  do
    i fwd dup @                 \ get the label# of the unres
    2 pick =                    \ if it is this one
    if
      dup 4+ @ dup c@(t)                \ get byte at addr of label reference
      0e9 over = swap 0e8 = f-or        \ jmp/call opcodes, rel16/32
      if                                \ if a 16/32bit branch
        here(t) over (offset) + -       \ calculate 16 bit offset
        swap 1+ ?!(t)                   \ store rel branch in target
      else                              \ if an 8bit branch
        here(t) over 2+ -               \ calculate 8 bit offset
        dup ?short                      \ range check
        swap 1+ c!(t)                   \ store branch in target
      endif
      0 swap !                          \ remove from table
    else                        \ if it is not this one
      drop                              \ then tidy up
    endif
  loop
;
+short-branches

hex

: addfwd        \ label# -- label# ; add forward reference to table
  maxref 1+ 0                           \ for quantity of legal labels
  do
    maxref i = asm_invLabelRef ?a-error	\ if its too big, then error
    i fwd  dup @ 0=                     \ if there is a zero here
    if
      2dup !                            \ then stash label number
      here(t) swap 4+ !  leave          \ stash here in next four bytes
    else
      drop                              \ tidy up on failure
    endif
  loop
;

: !labtab       \ n -- ; set up addr of local label
  >r                            \ get label number out of way of last opcode
  null-op                       \ force to lay down last opcode - updates HERE
  r>                            \ restore label number for the resolver
  resolve                       \ resolve any that can be
  here(t) swap label !          \ and stash this address in the label#
;

: @labtab       \ n -- addr ;
  dup label @                   \ get the address of this label
  swap over 0=                  \ if it is still zero
  if
    addfwd 2drop  here(t)       \ then add it, and return this addr for code
  else
    drop
  endif
;

\ get local label addr, or if undefined
\ add current dict.position to forward refs

hex

also asm-access definitions

\ local label definition

: l$:           \ label# -- ;
  dup 0 maxlab 1-                       \ is it
  within? 0= asm_invLabelDef ?a-error	\ within the limits
  !labtab                               \ if so, then handle it
;

\ local label reference

: l$            \ label# -- ;
  dup 0 maxlab 1-
  within? 0= asm_invLabelRef ?a-error
  @labtab                               \ handle only if in range
;

: loc-def       \ label# -- ; -- ;
  create
    ,
  does>
    @  l$:
;

: loc-ref       \ label# -- ; -- addr ;
  create
    ,
  does>
    @  l$
;

previous definitions

decimal

also asm-access definitions

 1 loc-def L$1:          1 loc-ref L$1
 2 loc-def L$2:          2 loc-ref L$2
 3 loc-def L$3:          3 loc-ref L$3
 4 loc-def L$4:          4 loc-ref L$4
 5 loc-def L$5:          5 loc-ref L$5
 6 loc-def L$6:          6 loc-ref L$6
 7 loc-def L$7:          7 loc-ref L$7
 8 loc-def L$8:          8 loc-ref L$8
 9 loc-def L$9:          9 loc-ref L$9
10 loc-def L$10:        10 loc-ref L$10
11 loc-def L$11:        11 loc-ref L$11


\ ====================
\ Register definitions
\ ====================

hex

only forth also asm-core also asm-access definitions

: .info         \ -- ; used for debug only
  cr ." Sizes: " b/w/d1 ? b/w/d2 ?
  cr ." Width: " w/d1 ? w/d2 ?
  cr ." Modes: " mode1 ? mode2 ?
  cr ." #'s:   " r1 ? r2 ?
  cr
;

only forth also asm-access also asm-core definitions

: .info	        \ --
  .info
;

only forth also asm-core definitions

: ?PPro+	\ -- ; error if not PPro+ ; SFP011
  CPUtype PProcpu < asm_InvCPU ?a-error
;

: reg           \ comma n size mode -- ; -- ; defining a register
  create
    , , , ,                             \ stash aspects of register
  does>
    #ops @
    case
      0 of  dup @ mode1 !               \ set mode
            dup cell+ @ b/w/d1 !        \ set byte/word
            dup 3 cells + @ co1 !       \ set whether has comma or not
            2 cells + @ r1 !            \ set register number
            1-flag on                   \ we now have a base for reg 1
            1 #ops !                    \ and to prove it
        endof
      1 of  dup @ mode2 !               \ set mode
            dup cell+ @ b/w/d2 !        \ set byte/word
            dup 3 cells + @ co2 !       \ set whether has comma or not
            2 cells + @ r2 !            \ set register number
            2-flag on                   \ we now have a base for reg 2
            2 #ops !                    \ count says so
        endof
      2 of  dup @ mode3 !               \ set mode
            dup cell+ @ b/w/d3 !        \ set byte/word ; SFP016
            2 cells + @ r3 !            \ set register number ; SFP016
            3 #ops !                    \ count says so
        endof
          cr abort" #ops failure in REG"
    endcase
    base-count off                      \ we have now had a reg, and no base or index
    ret-imm-flag on                     \ in case we have a ret # n following
;

: spec-reg	\ type reg# -- ; -- ; defining a special register
  create
    , ,                                 \ save the relevant bit patterns
  does>
    #ops @                              \ what reg to define next
    case
      0 of  dup @ r1 !                  \ set the first operand
            cell+ @ special-type !      \ set the type of special reg
            1 special-reg !             \ say we have a special case in pos1
            1 #ops !                    \ increment to next reg
        endof
      1 of  dup @ r2 !                  \ must be setting the second operand
            cell+ @ special-type !      \ set the type of special reg
            2 special-reg !             \ say we have special in pos2
            2 #ops !                    \ increment to (next) reg
        endof
    endcase
;


also asm-access definitions

\ comma n size mode -- ;
1 0 2 3 reg EAX,   1 1 2 3 reg ECX,  1 2 2 3 reg EDX,  1 3 2 3 reg EBX,
1 0 1 3 reg AX,    1 1 1 3 reg CX,   1 2 1 3 reg DX,   1 3 1 3 reg BX,
1 4 2 3 reg ESP,   1 5 2 3 reg EBP,  1 6 2 3 reg ESI,  1 7 2 3 reg EDI,
1 4 1 3 reg SP,    1 5 1 3 reg BP,   1 6 1 3 reg SI,   1 7 1 3 reg DI,
1 0 0 3 reg AL,    1 1 0 3 reg CL,   1 2 0 3 reg DL,   1 3 0 3 reg BL,
1 4 0 3 reg AH,    1 5 0 3 reg CH,   1 6 0 3 reg DH,   1 7 0 3 reg BH,
1 0 1 4 reg ES,    1 1 1 4 reg CS,   1 2 1 4 reg SS,   1 3 1 4 reg DS,
1 4 1 4 reg FS,    1 5 1 4 reg GS,

0 0 2 3 reg EAX    0 1 2 3 reg ECX   0 2 2 3 reg EDX   0 3 2 3 reg EBX
0 0 1 3 reg AX     0 1 1 3 reg CX    0 2 1 3 reg DX    0 3 1 3 reg BX
0 4 2 3 reg ESP    0 5 2 3 reg EBP   0 6 2 3 reg ESI   0 7 2 3 reg EDI
0 4 1 3 reg SP     0 5 1 3 reg BP    0 6 1 3 reg SI    0 7 1 3 reg DI
0 0 0 3 reg AL     0 1 0 3 reg CL    0 2 0 3 reg DL    0 3 0 3 reg BL
						       0 3 0 3 reg BLA
0 4 0 3 reg AH     0 5 0 3 reg CH    0 6 0 3 reg DH    0 7 0 3 reg BH
0 0 1 4 reg ES     0 1 1 4 reg CS    0 2 1 4 reg SS    0 3 1 4 reg DS
0 4 1 4 reg FS     0 5 1 4 reg GS
0 0 0 6 reg ,PC    0 0 0 7 reg SO

0 0 0 5 reg #

: offset        \ -- ; synonym for # (as per spec.)
  #
;

: seg           \ -- ; another synonym for #
  #
;

: direct-mode   \ -- 5/6 ; direct mode number according to seg size
  use32?
  if   5                        \ set register as 5 for direct 32bit
  else 6                        \ set register number: 6 for direct 16 bit
  endif
;

: set-direct2   \ -- ; set direct mode for op2
  direct-mode r2 !
  0 mode2 !                     \ set mode
  0 mode3 !
;

: set-direct1   \ -- ; set direct mode for op1
  direct-mode r1 !
  0 MODE1 !                     \ set mode
;

: ptr           \ -- ; specifies direct rather that immediate
  #ops @ 0= if                  \ what reg to define next
    set-direct1                 \ set op1 to direct mode
  else
    set-direct2                 \ set op2 to direct mode
  endif
;

: []		\ -- ;
  ptr                           \ as above, but also
  got-[] on                     \ say we had one of these
;

\ type reg# -- ;
1 0 spec-reg DR0   2 0 spec-reg CR0
1 1 spec-reg DR1
1 2 spec-reg DR2   2 2 spec-reg CR2
1 3 spec-reg DR3   2 3 spec-reg CR3
1 6 spec-reg DR6                      3 6 spec-reg TR6
1 7 spec-reg DR7                      3 7 spec-reg TR7

1 0 spec-reg DR0,  2 0 spec-reg CR0,
1 1 spec-reg DR1,
1 2 spec-reg DR2,  2 2 spec-reg CR2,
1 3 spec-reg DR3,  2 3 spec-reg CR3,
1 6 spec-reg DR6,                      3 6 spec-reg TR6,
1 7 spec-reg DR7,                      3 7 spec-reg TR7,

previous definitions

((
: <<3           \ n -- n<<3 ;
  dup +
  dup +
  dup +
;

: <<6           \ n -- n<<6 ;
  <<3 <<3
;
))

\ SFP014...
: <<3		\ n -- n<<3 ;
  3 lshift
;

: <<6		\ n -- n<<6 ;
  6 lshift
;
compiler
: <<3		\ n -- n<<3 ;
  3 lshift
;

: <<6		\ n -- n<<6 ;
  6 lshift
;
target
\ ...SFP014

: [z]		\ -- ; indicates no base register ; SFP012
  #ops @ 0= if
    2 w/d1 !  0 mode1 !  5 r1 !
    1 base-count !
    1-flag on
  else
    2 w/d2 !  0 mode2 !  5 r2 !
    1 base-count !
    2-flag on
  endif
  sib-flag on
;

: sib           \ index scale -- ; -- ;
  create
    , ,                       \ lay down reg then scale
  does>
\    sib-flag @                 \ if there is already a sib
\    if 033 error endif         \ then raise an error
    base-count @ 0=		\ if no base register, force null base register ; SFP012
    if  [z]  endif		\ SFP012
    dup
    @ <<6                       \ -- scale<<6
    swap cell+ @ <<3 f-or       \ -- ssiii000
    sib-val c!                  \ stash it
    sib-flag on                 \ say so
;


also asm-access definitions

\ index scale -- ;
0 1 sib [eax*2]   0 2 sib [eax*4]   0 3 sib [eax*8]
1 1 sib [ecx*2]   1 2 sib [ecx*4]   1 3 sib [ecx*8]
2 1 sib [edx*2]   2 2 sib [edx*4]   2 3 sib [edx*8]
3 1 sib [ebx*2]   3 2 sib [ebx*4]   3 3 sib [ebx*8]
5 1 sib [ebp*2]   5 2 sib [ebp*4]   5 3 sib [ebp*8]
6 1 sib [esi*2]   6 2 sib [esi*4]   6 3 sib [esi*8]
7 1 sib [edi*2]   7 2 sib [edi*4]   7 3 sib [edi*8]

previous definitions



decimal

: direct1?      \ -- t/f ; is op1 direct?
  r1 @ 6 = use16? f-and                 \ if direct 16bit or
  r1 @ 5 = use32? f-and f-or    	\ if direct 32bit
;

: direct2?      \ -- t/f ; is op2 direct?
  r2 @ 6 = use16? f-and         \ if direct 16bit or
  r2 @ 5 = use32? f-and f-or    \ if direct 32bit
;

: (index)       \ reg-addr addr -- ; set an [index*1] register for op1/2
  @ <<3
  swap @ f-or
  sib-val !                     \ stash in buffer
  sib-flag on                   \ say so
;

: (base1)       \ offset addr -- ; set a [base] register for op1
  @ r1 !                        \ set register number
  dup 0= if
    direct1?                    \ is op1 direct, as offset=0
    if    1                     \ set mode
    else  0
    endif
  else
    dup short?                  \ or is offset 8b only?
    if    1
    else  2
    endif
  endif
  mode1 !
;

decimal

: (base2)       \ offset addr -- ; set a [base] reg for op2
  @ r2 !                        \ set register number
  dup 0= if                     \ is offset zero, therefore direct?
    direct2?                    \
    if    1                     \ set mode: no displacement, just a [reg]
    else  0
    endif
  else
    dup short?                  \ or is offset 8b only
    if    1
    else  2
    endif
  endif
  mode2 !
;

: ?esp		\ addr reg-addr -- ;
  dup @ 4 = if
    (index)
  else
    2drop
  endif
;

: (ind)         \ addr --
\ set state to indicate we have had (another) index
  dup >r                                \ stash spare address
  dup cell+ @                           \ get size
  #ops @                                \ get which operand
  if w/d2 ! else w/d1 ! endif           \ set effective address size
  #ops @ 0= if                          \ if no registers yet ...
    base-count @ 0= if                  \ if no base yet
      (base1)                           \ then this is the base
      1 base-count !
      1-flag on
      r@ r1 ?esp
    else
      r1 swap (index)                   \ else this is the index*1
      base-count off
    endif
  else
    base-count @ 0= if
      (base2)                           \ either the base
      2 #ops !
      2-flag on
      1 base-count !
      r@ r2 ?esp
    else
      r2 swap (index)                   \ or the index*1
    endif
  endif
  r> 2 cells + @ if                     \ ends in a comma?
    #ops @
    case
      0 of  co1 on  endof
      2 of  co2 on  endof
    endcase
    1 #ops +!
  endif
;

: ind           \ comma size n -- ;  n1 -- n1
  create
    , , ,
  does>
   (ind)
;

: ind(16/32)    \ comma size n32  comma size n16 -- ;  n1 -- n1
  create
    , , ,  , , ,
  does>
    use32?                      \ if 32b segment
    if 3 cells + endif          \ offset to 32b data
    (ind)
;

also asm-access definitions

\ comma size n --
\ just 16 bit
1 1 0 ind [BX+SI],  1 1 1 ind [BX+DI],  1 1 2 ind [BP+SI],  1 1 3 ind [BP+DI],
0 1 0 ind [BX+SI]   0 1 1 ind [BX+DI]   0 1 2 ind [BP+SI]   0 1 3 ind [BP+DI]

\ comma size n32 comma size n16 --
\ both 32 bit and 16 bit
1 1 6  1 1 4 ind(16/32) [SI],     0 1 6  0 1 4 ind(16/32) [SI]
1 1 7  1 1 5 ind(16/32) [DI],     0 1 7  0 1 5 ind(16/32) [DI]
1 1 5  1 1 6 ind(16/32) [BP],     0 1 5  0 1 6 ind(16/32) [BP]
1 1 3  1 1 7 ind(16/32) [BX],     0 1 3  0 1 7 ind(16/32) [BX]

\ comma size n --
\ just 32 bit
1 2 0 ind [EAX],    1 2 1 ind [ECX],    1 2 2 ind [EDX],    1 2 3 ind [EBX],
1 2 4 ind [ESP],    1 2 5 ind [EBP],    1 2 6 ind [ESI],    1 2 7 ind [EDI],
0 2 0 ind [EAX]     0 2 1 ind [ECX]     0 2 2 ind [EDX]     0 2 3 ind [EBX]
0 2 4 ind [ESP]     0 2 5 ind [EBP]     0 2 6 ind [ESI]     0 2 7 ind [EDI]

previous definitions


hex

also asm-access definitions

: byte          \ -- ;
  b/w/d1 @ (b/w/d1) !           \ preserve the present size
  0 b/w/d1 !                    \ set both byte/word flags
  0 b/w/d2 !
  0 b/w/d3 !
  0 memoryformat !              \ set this (unneccessary, but complete)
;

: worda         \ -- ;
  b/w/d1 @ (b/w/d1) !           \ preserve the present size
  1 b/w/d1 !
  1 b/w/d2 !                    \ set both
  1 b/w/d3 !                    \ set both
  1 memoryformat !              \ specifies m16 for 387 (int)
;

: dword         \ -- ;
  2 b/w/d1 !                    \ force type of operands
  2 b/w/d2 !
  2 b/w/d3 !
  2 memoryformat !              \ specifies m32 for 387 (int)
;

: qword         \ -- ;
  3 memoryformat !              \ specifies m64 for 387 (int)
;

: tbyte         \ -- ;
  4 memoryformat !              \ specifies m80 for 387 (bcd)
;

: float         \ -- ;
  2 memoryformat !              \ specifies m32 for 387 (real)
;

: double        \ -- ;
  3 memoryformat !              \ specifies m64 for 387 (real)
;

: extended      \ -- ;
  4 memoryformat !              \ specifies m80 for 387 (real)
;

: b.            \ -- ;
  byte                          \ synonym
;

: w.            \ -- ;
  worda                         \ synonym
;

: dw.           \ -- ;
  dword                         \ synonym
;

previous definitions


: intel?        \ -- t/f ; is default mode immediate?
  default-mode @                \ which is default mode?
;
\ SFP014
compiler
: intel?        \ -- t/f ; is default mode immediate?
  default-mode @                \ which is default mode?
;
target
\ ...SFP014

only forth also asm-access also asm-core definitions

: #),           \ -- ;
  intel?                        \ if we required immediate default
  if
    0 r1 !                      \ reset reg no as well
    5 mode1 !                   \ default to immediate addressing
  else                          \ if we require direct default
    set-direct1                 \ then set direct mode for op1
  endif
;

: #)            \ -- ;
  intel? if                     \ if we required immediate default
    0 r2 !                      \ reset reg number as well
    5 mode2 !                   \ set immediate as default
    5 mode3 !                   \ set immediate as default
  else                          \ if we require direct default
    set-direct2                 \ set direct for op2
  endif
;

: reset         \ -- ;
  #),  #)                       \ default register and modes
  0 adr-mod !
  use32? if                     \ reset according to default size
    dword
    2 (b/w/d1) !
    2 w/d1 !  2 w/d2 !
  else
    worda
    1 (b/w/d1) !
    1 w/d1 !  1 w/d2 !
  endif
  qword
  a-override off  d-override off        \ turn of overrides
  sib-flag off                          \ no sib yet
  #ops off                              \ no operands yet
  1-flag off  2-flag off                \ no base found yet on either side
  base-count off                        \ as above
  far/near off
  short/long off
  special-reg off                       \ reset this also
  got-[] off                            \ we have not had one of these yet
  co1 off  co2 off
  log-line                              \ generate list file
  save-pos                              \ save position for prefix use
  ret-imm-flag off                      \ in case we dont have a # with ret
;

assign reset to-do (reset)

only forth also asm-core also asm-access definitions

: reset         \ -- ; reset the assembler sees
  reset                         \ the above
;


only forth also asm-core definitions

: dir1?         \ -- t/f ; is register1 direct?
  direct1?
  mode1 @ 0= f-and			\ and mode 0
  sib-flag @ 0= f-and			\ SFP012
;

: dir2?         \ -- t/f ; is register2 direct?
  direct2?
  mode2 @ 0= f-and			\ and mode 0
  sib-flag @ 0= f-and			\ SFP012
;

: areg?          \ addr -- t/f ;
  @ 3 =
;

: seg?          \ addr -- t/f ;
  @ 4 =
;

: #?            \ -- t/f ;
  mode2 @ 5 =
;

: ac?           \ addr -- t/f ;
  @ 0=
;

\ SFP014...
compiler
: areg?          \ addr -- t/f ; name conflict with optimiser.
  @ 3 =
;
: seg?          \ addr -- t/f ;
  @ 4 =
;

: #?            \ -- t/f ;
  mode2 @ 5 =
;

: ac?           \ addr -- t/f ;
  @ 0=
;
target
\ ...SFP014

: word?         \ -- t/f ; used in XCHG only
  b/w/d1 @
  b/w/d2 @
  f-and                         \ are both operands word ones?
;

hex

: i/d?          \ addr -- t/f ;
  @ 0 2 within?         \ immediate or direct ?
;

: r/i/d?        \ addr -- t/f ;
  @ 0 3 within?         \ register, immediate or direct?
;

: must-be       \ n -- ;
  0= asm_invAddrMode ?a-error	\ error if false
;

: ?m1=i/d       \ -- ; is op.1 immediate or direct,
  mode1
  i/d? must-be                  \ error if not
;

: ?m2=i/d       \ -- ; is op.2 immediate or direct,
  mode2
  i/d? must-be                  \ error if not
;


hex

: 0op1          \ n -- ; -- ; single byte 0 operand instructions
  create
    ,                          \ lay down byte
  does>
    ?prefix                     \ do the last one
    @ c,(t)                     \ comma it
    reset                       \ reset for next instruction
;

: 0op1-32       \ n -- ; -- ; single byte 0 operand instructions
  create
    ,                           \ lay down byte
  does>
    ?prefix                     \ do the last one
    use16?                      \ if in 16bit segment
    if   66 c,(t)  endif        \ then lay down override
    @ c,(t)                     \ comma it
    reset                       \ reset for next instruction
;

: 0op1-16	\ n -- ; -- ; single byte 0 operand instructions
  create
    ,                          \ lay down byte
  does>
    ?prefix                     \ do the last one
    use32?              	\ if in 32bit segment
    if   66 c,(t)  endif        \ then lay down override
    @ c,(t)                     \ comma it
    reset                       \ reset for next instruction
;


: 0op2		\ n1 n2 -- ; two byte 0-op instructions
  create
    , ,                         \ lay down bytes
  does>
    ?prefix                     \ do the last one
    dup 4+
    @ c,(t)                     \ comma first
    @ c,(t)                     \ comma second
    reset                       \ reset for next instruction
;


hex

: ax/al,#       \ n opcode -- ; puts down opcode and data for ax or al
  b/w/d1 @ if
    1+ c,(t) absolute-data      \ 1/2: word, laid down accordingly
  else
    c,(t) c,(t)                 \ 0:byte
  endif
;

: +mod+r/m1     \ n1 -- n2 ; adds mod and r/m for register 1
  mode1 @
  <<6 +
  r1 @ +
;

: +mod+r/m2     \ n1 -- n2 ; adds mod and r/m for register 2
  mode2 @
  <<6 +
  r2 @ +
;

: +reg1         \ n1 -- n2 ; adds reg for register 1
  r1 @
  <<3 +
;

: +reg2         \ n1 -- n2 ; adds reg for register 2
  r2 @
  <<3 +
;

: opcode+w1     \ opcode -- opcode'
  b/w/d1 @
  if  1+  endif
;

: opcode+w2     \ opcode -- opcode'
  b/w/d2 @
  if  1+  endif
;

: opcode+sw,    \ imm-op opcode -- imm-op
  opcode+w1
  b/w/d1 @ if
    over short?
    if  2+  endif
  endif
  c,(t)
;


: disp,         \ displacement reg mode --
  case
    0  of
         dup 6 =  use16? f-and  \ if direct 16bit or
         swap 5 = use32? f-and f-or     \ direct 32bit
         if    absolute-addr
         else  drop
         endif
       endof
    1  of  drop  c,(t)  endof
    2  of  drop  absolute-addr   endof
           asm_invAddrMode a-error
  end-case
;

hex

: disp1,        \ disp -- ;
  r1 @  mode1 @  disp,
;

: disp2,        \ disp -- ;
  r2 @  mode2 @  disp,
;

: ?disp1,       \ disp -- ;
  mode1 areg? 0=
  if  disp1,  endif
;

: ?disp2,       \ disp -- ;
  mode2 areg? 0=
  if  disp2,  endif
;


hex

: lay-sib1      \ -- ;
  sib-flag @ if                 \ if a sib byte is required...
    sib-val c@                  \ get the ssiii000 part
    r1 @                        \ get base part
    07 f-and                    \ mask top bits
    f-or                        \ make sib byte
    c,(t)                       \ lay down
  endif
;

: lay-sib2      \ -- ;
  sib-flag @ if                 \ if a sib byte is required...
    sib-val c@                  \ get the ssiii000 part
    r2 @                        \ get base part
    07 f-and                    \ mask top bits
    f-or                        \ make sib byte
    c,(t)                       \ lay down
  endif
;

: ?sib,         \ 8b sib# -- ; maybe lay a sib
  swap  sib-flag @ if                   \ if sib required
    0f8 f-and 04 f-or  c,(t)            \ mask sib bit, lay r/m
    1 =                                 \ depending whether sib1 or sib2
    if   lay-sib1                       \ also lay sib1
    else lay-sib2                       \ or sib2
    endif
  else                                  \ if no sib required
    c,(t)                               \ lay straight
    drop                                \ ignore the sib number
  endif
;

: ?ext1,        \ disp 8b -- ; maybe lay sib and displacemnt
  1 ?sib,                               \ maybe lay a sib
  ?disp1,                               \ maybe lay a displacement
;

: ?ext2,        \ disp 8b -- ; maybe lay sib and displacemnt
  2 ?sib,                               \ maybe lay a sib
  ?disp2,                               \ maybe lay a displacement
;


: m/r,m/r?      \ -- t/f
  mode1 areg? if
    mode2 r/i/d?
  else
    mode2 areg? if
      mode1 r/i/d?
    else
      0
    endif
  endif
;

: mem/reg,#     \ [ displacement ] imm-op op-code adr-mod --
  >r opcode+sw, r>                      \ -- (disp) io a-m
  +mod+r/m1                             \ -- (disp) io a-m'
  1 ?sib,
  >r ?disp1, r>
  b/w/d1 @ if
    dup short?
    if    c,(t)
    else  absolute-data
    endif
  else
    c,(t)
  endif
;

: mem/reg,#l    \ [ displacement] imm-op opcode adr-mod --
  swap
  opcode+w1 c,(t)
  +mod+r/m1 1 ?sib,
  >r ?disp1, r>
  b/w/d1 @                              \ lay down the immediate data
  if
    absolute-data
  else
    c,(t)
  endif
;

: mem/reg,mem/reg       \ [ offset ] op-code --
  mode1 areg?                            \ if reg1 is r/m
  if
    2+ opcode+w1 c,(t)                  \ make opcode
    0 +mod+r/m2
    +reg1                               \ make basic modr/m
    ?ext2,
  else                                  \ if not...
    mode2 areg? must-be                  \ it must be reg2 is reg
    opcode+w2 c,(t)                     \ lay opcode
    0 +mod+r/m1
    +reg2                               \ make basic modr/m
    ?ext1,
  endif
;


hex

: lay-addr-override     \ -- ; lay the addr override
  067 c,(t)                     \ lay addr override
  a-override on                 \ say override set
;

: lay-data-override     \ -- ; lay the data override
  066 c,(t)                     \ lay it
  d-override on                 \ say override set
;

: ?lay-manual-override  \ -- t/f ; is a manual override requested?
  0                             \ dummy false flag
  d-override @                  \ if data override spec'd
  if
    lay-data-override           \ lay it
    drop -1                     \ say override laid
  endif
  a-override @                  \ if address override spec'd
  if
    lay-addr-override           \ lay it
    drop -1                     \ say override laid
  endif
;

: data32?       \ -- t/f ; is the data size 32b?
  b/w/d1 @ 2 =
  b/w/d2 @ 2 = f-or
;

: data16?       \ -- t/f ; is the data size 16b?
  b/w/d1 @ 1 =
  b/w/d2 @ 1 = f-or
;

: 2*data32?     \ -- t/f ; are both data sized 32b?
  b/w/d1 @ 2 =
  b/w/d2 @ 2 = f-and
;

: 2*data16?     \ -- t/f ; are both data sized 16b?
  b/w/d1 @ 1 =
  b/w/d2 @ 1 = f-and
;

: addr32?       \ -- t/f ; is the address size 32b?
  w/d1 @ 2 =
  w/d2 @ 2 = f-or
;

: addr16?       \ -- t/f ; is the address size 16b?
  w/d1 @ 1 =
  w/d2 @ 1 = f-or
;

: ?both         \ -- ; maybe lay both overrides
  data32? use16? f-and addr32? f-and
  data16? use32? f-and addr16? f-and f-or
  if
    lay-addr-override
    lay-data-override
  endif
;

: ?data-override        \ -- ; maybe lay data override
  data32? use16? f-and addr16? f-and
  data16? use32? f-and addr32? f-and f-or
  2*data16? use32? f-and f-or
  2*data32? use16? f-and f-or
  if  lay-data-override  endif
;

: ?addr-override        \ -- ; maybe lay an address override
  data32? use32? f-and addr16? f-and
  data16? use16? f-and addr32? f-and f-or
  if  lay-addr-override  endif
;

: ?lay-auto-override    \ -- ; maybe lay an override automatically
  ?addr-override
  ?data-override
  ?both
;

: ?overrides	\ -- ; maybe lay an override
  ?lay-manual-override 0=               \ if no manual override needed
  if  ?lay-auto-override  endif         \ then maybe lay one anyway
;

: ?in-override  \ -- ; maybe lay a data override for IN
  b/w/d1 @ 1 =  use32? f-and    \ if IN AX 32b seg
  b/w/d1 @ 2 =  use16? f-and f-or       \ or IN EAX 16b seg
  if  lay-data-override  endif
;

: ?out-override \ -- ; maybe lay a data override for IN
  b/w/d2 @ 1 =  use32? f-and    \ if IN AX 32b seg
  b/w/d2 @ 2 =  use16? f-and f-or       \ or IN EAX 16b seg
  if  lay-data-override  endif
;

: ?auto-movx-overrides          \ -- ; see if an auto one needed for movxx
  (b/w/d1) @ 1 =  b/w/d1 @ 1 = f-or use32? f-and
  (b/w/d1) @ 2 =  b/w/d1 @ 2 = f-or use16? f-and f-or
  w/d2 @ 0=  b/w/d2 @ 0= f-or   f-and
  if  lay-data-override  endif		\ then lay an override
;

: ?movx-overrides       \ -- ; maybe lay an override for MOVxX
  ?lay-manual-override 0=               \ if no manual override needed
  if  ?auto-movx-overrides  endif       \ then see if an auto one
;

: 3modes	\ o1 o2 o3 am -- ; [ displacement ] [ imm-op ] --
  create
    , , , ,
  does>
    ?prefix                     	\ do the last one
    ?overrides                          \ lay any data or address override
    dup @ adr-mod !  #? if
      mode1 areg?  r1 ac? f-and if
        2 cells + @ ax/al,#             \ lay immediate,ax form
      else
        cell+ @  adr-mod @  mem/reg,#   \ lay immediate,reg form
      endif
    else
      m/r,m/r? must-be                  \ must be r/m
      3 cells + @
      mem/reg,mem/reg                   \ then do it
    endif
    reset
;

: ac/reg,#	\ imm-op ^op -- ; SFP012
\ Select EAX,#n if r1=EAX and imm-op cannot be a byte value, or
\ reg,#n if it can (shorter opcode).
  mode1 areg?  r1 ac? f-and if		\ ac,#n?
    over short? 0=  use16? or if	\ if not byte, ACC form is shorter
      2 cells + @ ax/al,#  exit		\ ACC,immediate form
    endif
  endif
  cell+ @  adr-mod @  mem/reg,#		\ immediate, long/short form ; SFP002
;

(( \ original
: ac/reg,#	\ imm-op ^op -- ; SFP012
\ Select EAX,#n if r1=EAX and imm-op cannot be a byte value, or
\ reg,#n if it can (shorter opcode).
  mode1 areg?  r1 ac? f-and if
    2 cells + @ ax/al,#			\ immediate,ax form
  else
    cell+ @  adr-mod @  mem/reg,#	\ immediate, long/short form ; SFP002
  endif
;
))

: 3modes/l      \ o1 o2 o3 am -- ; [ displacement ] [ imm-op ] --
  create
    , , , ,
  does>
    ?prefix                     	\ do the last one
    ?overrides
    dup @ adr-mod !  #? if
      ac/reg,#
    else
      m/r,m/r? must-be
      3 cells + @ mem/reg,mem/reg	\ mem/reg form
    endif
    reset
;

only forth also asm-access also asm-core definitions

: ?[]           \ -- ;
  got-[] @
  if  set-direct2  endif
;

: ?[]1          \ -- ;
  got-[] @
  if  set-direct1  endif
;

: (jmp)         \ addr -- ;
  ?overrides
  dir1? got-[] @ 0= f-and       \ if op1 is direct; ie JMP xxxx
  intel? mode1 @ 5 = f-and f-or \ or if op1 is immediate and mode is intel
  if
    dup here(t) 2+ - short?     \ if short because of size,
    short/long @ 0= f-and       \ and LONG not selected
    if
      0eb c,(t)                 \ JMP REL8 opcode
      rel8                      \ and relative 8bit address
    else                        \ if long
      0e9 c,(t)                 \ JMP REL16/32 opcode
      rel16/32                  \ and relative 16/32 bit address
    endif
  else                          \ JMP RM16/32; if register specified
    ?[]1                        \ this mode
    mode1  r/i/d? must-be       \ must be these choices
    0ff c,(t)                   \ magic number
    020 +mod+r/m1               \ basic modr/m
    ?ext1,
  endif
  reset
;

: (jmp/f)       \ seg addr -- ;
  ?overrides
  dir1? got-[] @ 0= f-and       \ if no regs or [], then must be direct
  intel? mode1 @ 5 = f-and f-or \ or if op1 is immediate and mode is intel
  if
    0ea c,(t)                   \ JMP PTR far opcode
    absolute-addr  w,(t)        \ 32/16b offs 16b seg
  else                          \ if indirect, ie JMP [] xxxx
    ?[]1 ?m1=i/d                \ must be these modes
    0ff c,(t)                   \ JMP MEM far opcode
    028 +mod+r/m1               \ build modr/m
    ?ext1,
  endif
  reset
;

only forth also asm-core also asm-access definitions

: jmp/f         \ seg off -- ; for compatibility with old asm
  0 ?prefix drop                \ do the last one
  (jmp/f)                       \ lay the far jump
;

: jmp           \ addr | seg/off -- ;
  0 ?prefix drop                \ do the last one
  far/near @                    \ if far or near
  if    (jmp/f)                 \ lay far form
  else  (jmp)                   \ or near form
  endif
;

previous definitions


also asm-core definitions

: (call)        \ addr -- ;
  ?overrides
  dir1? got-[] @ 0= f-and       \ if no regs or [], then must be direct
  intel? mode1 @ 5 = f-and f-or \ or if op1 is immediate and mode is intel
  if
    0e8 c,(t)                   \ CALL REL16/32 opcode
    rel16/32                    \ lay relative address
  else                          \ if a reg or a [], then  must be indirect
    ?[]1                        \
    mode1 r/i/d? must-be        \ must be register/immediate/direct
    0ff c,(t)                   \ CALL RM16/32 opcode
    010 +mod+r/m1               \ make modr/m
    ?ext1,
  endif
  reset                         \  finished the opcode
;

: (call/f)      \ seg off -- ;
  ?overrides
  dir1? got-[] @ 0= f-and       \ if no regs or [], then must be direct
  intel? mode1 @ 5 = f-and f-or \ or if op1 is immediate and mode is intel
  if
    09a c,(t)                   \ CALL FAR ss oooo
    absolute-addr w,(t)         \ lay offset then segment - 16/32o:16s
  else                          \ with [], then must be indirect
    ?[]1 ?m1=i/d
    0ff c,(t)                   \ CALL FAR [] oooo
    018 +mod+r/m1
    ?ext1,
  endif
  reset
;

only forth also asm-core also asm-access definitions

: call/f        \ seg off -- ; for compatiblity with old asm
  0 ?prefix drop                \ do the last one
  (call/f)
;

: call          \ addr | seg/off -- ; general call
  0 ?prefix drop                \ do the last one
  far/near @
  if    (call/f)                \ do far
  else  (call)                  \ or do near
  endif
;

previous definitions


also asm-core definitions

: bit-scan      \ n1 n2 -- ; (reg mem) -- ;
  create
    , ,
  does>
    ?prefix                     \ do the last one
    ?overrides
    dup cell+ @ c,(t)           \ lay first byte
    @ c,(t)                     \ lay second byte
    mode1 areg? must-be          \ the only way to be
    0 +mod+r/m2
    +reg1                       \ generate modr/m
    ?ext2,
    reset
;

: bit-test      \ reg b1 b2a b2b -- ; (reg/mem reg/imm) -- ;
  create
    , , , ,
  does>
    ?prefix                     \ do the last one
    ?overrides
    dup 2 cells + @ c,(t)       \ lay first byte
    mode2 @ 5 <> if		\ if not immediate
      cell+ @ c,(t)             \ lay second byte for reg form
      0 +mod+r/m1               \
      +reg2                     \ generate modr/m
      ?ext1,
    else
      dup @ c,(t)               \ lay second byte for imm form
      0 +mod+r/m1               \ generate base modr/m
      swap 0c + @ <<3 f-or      \ add /4 bit from databook
      1 ?sib,
      mode1 areg? 0=
      if
        swap disp1,             \ lay down any displacement or imm
      endif
      c,(t)                     \ lay immediate data
    endif
    reset
;

: 2op-r,r/m     \ op1 op2 -- ; reg16,32 reg/mem16,32 -- ;
  create
    , ,
  does>
    ?prefix                     \ do the last one
    mode1 areg? must-be          \ r1 must be a register
    ?overrides                  \ overrides do apply
    dup cell+ @  c,(t)
    @  c,(t)                    \ lay two bytes of opcode
    0 +mod+r/m2 +reg1           \ basic modr/m
    ?ext2,
    reset
;

: 2op-r/m,r	\ opc -- ; CMPXCHG  r/m, r ; SFP016
  create
    ,
  does> ?prefix
    mode2 areg? must-be			\ it must be mode2 is reg
    ?overrides				\ overrides do apply
    $0F c,(t)				\ prefix opcode
    @ opcode+w2 c,(t)			\ lay primary opcode
    0 +mod+r/m1  +reg2  ?ext1,		\ make basic modr/m
    reset
;

: 1op-r32		\ opc -- ; BSWAP and friends ; SFP016
  create
    ,
  does> ?prefix
    mode1 areg? must-be			\ it must be mode is reg
    b/w/d1 @ 2 = must-be		\ size must be 32 bits
    ?overrides				\ overrides do apply
    $0F c,(t)				\ prefix opcode
    @ r1 @ f-or c,(t)			\ lay primary opcode
    reset
;

: load-table    \ op1 op2 reg -- ; mem -- ;
  create
    , , ,
  does>
    ?prefix                     \ do the last one
    mode1 areg? 0= must-be       \ must be a mem operand
    dup 2 cells + @ c,(t)       \ lay first byte
    dup cell+ @ c,(t)           \ lay second byte
    0 +mod+r/m1                 \ modr/m bits of that byte
    swap @                      \ get reg bit of modr/m
    <<3 +                       \ add reg bit
    ?ext1,
    reset                       \ as ever
;

: 2op-r/m       \ op1 op2 reg -- ; reg/mem16 -- ;
  create
    , , ,
  does>
    ?prefix                     \ do the last one
    b/w/d1 @ 1 = must-be        \ must be 16b register or effective addr
    dup 2 cells + @ c,(t)       \ lay first opcode byte
    dup cell+ @ c,(t)           \ lay second byte of opcode
    0 +mod+r/m1                 \ basic modr/m
    swap @ <<3 +                \ add ref bit to it
    ?ext1,
    reset
;

: mov(x)x       \ op1 op16 op32 -- ; reg reg/mem -- ;
  create
    , , ,
  does>
    ?prefix                             \ do the last one
    ?movx-overrides                     \ special case overrides
    dup 2 cells + @ c,(t)               \ always lay first byte of instruction
    b/w/d2 @ 0= w/d2 @ 0= f-or		\ if source is byte
    if    cell+ @ c,(t)                 \ lay first version
    else  @ c,(t)                       \ lay second version
    endif
    0 +mod+r/m2                         \ get basic modr/m
    +reg1                               \ make further modr/m
    ?ext2,
    reset                               \ finished
;

: 0op16/32      \ size op -- ; -- ;
  create
    , ,
  does>
    ?prefix                             \ do the last one
    dup                                 \ extra copy of address of data
    cell+ @ 1 =                         \ if 16b instruction
    if
      use32?                            \ if segment is 32bit
      if  lay-data-override  endif      \ lay override
    else                                \ else a 32b instruction
      use16?                            \ if segment is 16bit
      if  lay-data-override endif       \ lay override
    endif
    @ c,(t)                             \ get instruction and lay it
    reset
;

: set           \ op2 -- ; r/m8 -- ;
  create
    ,
  does>
    ?prefix                     \ do the last one
    0f c,(t)                    \ lay first byte
    @ c,(t)                     \ lay second byte of opcode
    0 +mod+r/m1
    ?ext1,
    reset                       \ reset for next instruction
;

: d-shift       \ op1 op2 -- ;
  create
    , ,
  does>
    ?prefix                     \ do the last one
    ?overrides                  \ in case its the wrong size
    dup cell+ @ c,(t)           \ lay first byte of opcode
    @                           \ get second byte of opcode
    mode3 @ 5 <>                \ if op3 is not immediate
    if                          \ if op3 is register
      1+                        \ add to shift to non-immediate form
    endif
    c,(t)                       \ lay second byte of opcode
    0 +mod+r/m1 +reg2           \ make basic modr/m
    1 ?sib,                     \ lay it and any sib byte
    mode1 areg? 0=               \ if there is a displacement to lay
    if
      mode3 @ 5 =               \ if there is also immediate data
      if swap endif             \ then swap to get the displacement
      disp1,                    \ then lay it
    endif
    mode3 @ 5 =                 \ if op3 is immediate data
    if  c,(t) endif             \ lay it down
    reset                       \ and now we have finished
;


also asm-access definitions

hex

: arpl          \ -- ;
  0 ?prefix drop                \ do the last one
\  ?overrides                   \ PharLap doesn't
  mode2 areg? must-be            \ the only way to be
  63 c,(t)                      \ lay opcode
  0 +mod+r/m1                   \
  +reg2                         \ generate modr/m
  ?ext1,
  reset
;

: bound         \ -- ;
  0 ?prefix drop                \ do the last one
  ?overrides
  mode1 areg? must-be              \ op1 must be a 16/32 reg
  mode2 areg? 0= must-be           \ op2 must not be
  62 c,(t)                        \ lay opcode
  0 +mod+r/m2                     \
  +reg1                           \ generate modr/m
  ?ext2,
  reset
;

: enter         \ imm16 imm8 -- ;
  0 ?prefix  drop                               \ do the last one
  mode1 @ 5 = mode2 @ 5 =  f-and must-be        \ must be imm, imm; (mode3 due to `,')
  0c8 c,(t)                                     \ lay opcode
  swap  w,(t)                                   \ lay imm16
  c,(t)                                         \ lay imm8
  reset                                         \ ready for next one
;


previous definitions

: imul(0)       \ [disp] -- ; IMUL R/M form
  0f6 opcode+w1 c,(t)           \ magic number
  028 +mod+r/m1                 \ make mod byte
  ?ext1,                        \ lay all necessary bytes
;

: imul(1a)      \ imm# -- ; IMUL R,I
  dup short?
  if                            \ if reg, imm8
    06b c,(t)                   \ lay byte of opcode
    0 +mod+r/m1 +reg1           \ generate basic modr/m
    c,(t)                       \ lay it
    c,(t)                       \ lay immediate data
  else                          \ if reg, imm16/32
    069 c,(t)                   \ lay byte of opcode
    0 +mod+r/m1 +reg1           \ generate basic modr/m
    c,(t)                       \ lay it
    b/w/d1 @ 1 =                \ is it 16 bit target register
    if    w,(t)                 \ lay immediate data (16b)
    else  ,(t)                  \ lay immediate data (32b)
    endif
  endif
;

: imul(1b)      \ [disp] -- ; IMUL R,R/M
  0f c,(t)  0af c,(t)           \ lay bytes of opcode
  0 +mod+r/m2 +reg1             \ make basic modr/m
  ?ext2,
;

: imul(1)       \ [disp] -- ; IMUL R,R/M/I forms
  #?                            \ have immediate?
  if                            \ mode2=imm; IMUL R,I
    imul(1a)
  else                          \ mode2=mem; IMUL R,R/M
    imul(1b)
  endif
;

: imul(2)       \ [disp] -- ; the IMUL R,R/M,I form
  dup >r                        \ immediate data out the way
  short?                        \ is immediate data 8b or not
  if                            \ if 8 bit data
    06b c,(t)                   \ lay opcode
    0 +mod+r/m2 +reg1           \ basic modr/m
    ?ext2,
    r> c,(t)                    \ lay 8b immediate data
  else                          \ if 16 or 32 bit data
    069 c,(t)                   \ lay opcode
    0 +mod+r/m2 +reg1           \ basic modr/m
    ?ext2,
    r>                          \ get immediate data back
    b/w/d1 @ 1 =                \ if reg1 is 16bit
    if    w,(t)                 \ .. lay 16b immediate
    else  ,(t)                  \ else lay 32b immediate
    endif
  endif
;

also asm-access definitions

: imul          \ ??? -- ;
  0 ?prefix drop                \ do the last one
  ?overrides
  co1 @                         \ if we have had at least op1,op2
  if
    co2 @
    if                          \ if we have had op1,op2,...
      imul(2)                   \ the IMUL R,R/M,I form
    else                        \ if we have had op1,op2
      imul(1)                   \ the IMUL R,R/M/I form
    endif
  else                          \ if we have had op1
    imul(0)                     \ the IMUL R/M form
  endif
  reset
;

previous definitions


hex

: inc/dec       \ o1 o2 am --
  create
    , , ,
  does>
    ?prefix                     \ do the last one
    ?overrides
    dup @ adr-mod !
    mode1 areg?
    b/w/d1 @  f-and
    if
      2 cells + @
      r1 @ f-or
      c,(t)
    else
      cell+ @
      b/w/d1 @ 1 min
      f-or c,(t)
      adr-mod @
      +mod+r/m1
      ?ext1,
    endif
    reset
;


hex

: mul/div       \ o1 am --
  create
    ,
  does>
    ?prefix                     \ do the last one
    ?overrides
    @ adr-mod !
    0f6 opcode+w1 c,(t)         \ magic number
    adr-mod @ +mod+r/m1
    ?ext1,
    reset
;

also asm-access definitions

: esc           \ -- ;
  0 ?prefix drop                \ do the last one
  ?overrides
  0d8 c,(t)                     \ ESC opcode
  0 +mod+r/m2 c,(t)
  ?disp2,
  reset
;

previous definitions


hex

also asm-access definitions

: in            \ ??? -- ;
  0 ?prefix drop                        \ do the last one
  ?in-override
  mode1 areg?
  r1 ac? f-and  must-be
  #?
  if
    0e4 opcode+w1 c,(t)  c,(t)          \ magic number
  else
    mode2 areg?
    r2 @ 2 = f-and  must-be
    0ec opcode+w1 c,(t)                 \ magic number
  endif
  reset
;

: out           \ ??? -- ;
  0 ?prefix drop                                \ do the last one
  ?out-override
  mode2 areg?
  r2 ac?  f-and must-be
  mode1 @ 5 =                                   \ if immediate data
  if
    0e6 opcode+w2  c,(t)                        \ OUT #,R
    c,(t)
  else
    mode1 areg?                                  \ OUT DX,R
    r1 @ 2 = f-and  must-be
    0ee opcode+w2 c,(t)                         \ magic number
  endif
  reset
;

previous definitions

hex

also asm-access definitions

: int           \ n --
  0 ?prefix drop                \ do the last one
  dup 3 =                       \ if int3, then cc
  if   drop 0cc c,(t)
  else
    0cd c,(t)                   \ INT opcode
    c,(t)                       \ and int#
  endif
  reset
;

previous definitions

: j/lp          \ o16 -- ; addr -- ;
  create
    ,
  does>
    ?prefix                     \ do the last one
    short/long @ 0=             \ if default/short
    if
      @ c,(t) rel8              \ lay byte and rel branch
    else                        \ if long
      0f c,(t)                  \ lay first byte
      @ 010 + c,(t)             \ lay second byte
      rel16/32                  \ lay rel branch as per seg size
    endif
    reset
;

hex

: (reg,mem)	\ [reg] [mem] -- ; SFP011
  mode1 areg? must-be          \ must be register
  mode2 r/i/d? must-be        \ no segs, or #
  0 +mod+r/m2 +reg1
  ?ext2,
;

: reg,mem       \ o1 -- ; reg mem -- ;
  create
    ,
  does>
    ?prefix                     \ do the last one
    ?overrides                  \ any overrides needed
    @ c,(t)                     \ comma in opcode
    (reg,mem)			\ lay rest ; SFP011
    reset
;

: reg,mem(2)    \ o1 o2 -- ; reg mem -- ;
  create
    , ,
  does>
    ?prefix                     \ do the last one
    ?overrides                  \ any overrides needed
    dup cell+ @ c,(t)           \ comma in first opcode
    @ c,(t)                     \ comma in second opcode
    (reg,mem)			\ lay rest ; SFP011
    reset
;

: reg,mem(3)	\ o1 o2 -- ; reg mem -- ; SFP011
  create
    , ,
  does>
    ?prefix                     \ do the last one
    ?PPro+			\ at least Pentium Pro required
    ?overrides                  \ any overrides needed
    dup cell+ @ c,(t)           \ comma in first opcode
    @ c,(t)                     \ comma in second opcode
    b/w/d1 @ must-be		\ no byte operations
    (reg,mem)			\ lay rest
    reset
;

: lay-spec      \ -- ;
  3 <<6                 	\ mod part is always `11'
  special-reg @ 1 = if		\ if special is in pos1
    r1 @ <<3 f-or       	\ mask in the special reg#
    r2 @ f-or           	\ mask in general reg#
  else                          \ if special is in pos2
    r2 @ <<3 f-or       	\ mask in the special reg#
    r1 @ f-or           	\ mask in general reg#
  endif
  c,(t)                 	\ lay down modr/m byte
;

-short-branches
: mov-spec      \ -- ;
  b/w/d1 @ 2 = b/w/d2 @ 2 = f-or must-be        \ must be 32b reg involved
  0f c,(t)                              \ lay down first byte of opcode
  special-reg @ 1 =
  if                                    \ if special is operand1
    special-type @                      \ is it DRn, CRn or TRn ?
    case
      1 of 023 c,(t) lay-spec endof     \ type DRn, lay opcode
      2 of 022 c,(t) lay-spec endof     \ type DRn, lay opcode
      3 of 026 c,(t) lay-spec endof     \ type DRn, lay opcode
           abort" Failure with type in mov-spec"
    endcase
  else                                  \ if special is operand2
    special-type @                      \ is it DRn, CRn or TRn ?
    case
      1 of 021 c,(t) lay-spec endof     \ type DRn, lay opcode
      2 of 020 c,(t) lay-spec endof     \ type DRn, lay opcode
      3 of 024 c,(t) lay-spec endof     \ type DRn, lay opcode
           abort" Failure with type in mov-spec"
    endcase
  endif
  reset
;
+short-branches

also asm-access definitions

-short-branches
: mov           \ ??? -- ;
  0 ?prefix drop                        \ do the last one
  special-reg @ 0<> if                  \ if a special reg has been used
    mov-spec  exit                      \ do the special case
  endif
  #? if                                 \ if immediate source
    mode1 areg? if
      ?overrides
      0b0 b/w/d1 @ 1 min <<3 + r1 @ + c,(t)     \ magic number
      b/w/d1 @
      if   absolute-data  else  c,(t)  endif
    else
       mode1 i/d? must-be
       ?overrides
       0c6 0 mem/reg,#l                 \ magic number
    endif
  else                                  \ seg or reg for source
    mode2 seg? if
      mode1 r/i/d? must-be
      08c c,(t) 0 +mod+r/m1             \ magic number
      +reg2
      ?ext1,
    else
      mode1 seg? if                     \ seg. reg. for dest
        mode2 r/i/d? must-be
        08e c,(t) 0 +mod+r/m2           \ magic number
        +reg1
        ?ext2,
      else
        mode2 areg?
        r2 ac? f-and
        dir1? f-and                     \ mem, ac
        if
          ?overrides
          0a2 opcode+w2 c,(t)  absolute-addr    \ magic number
        else
          mode1 areg?
          r1 ac? f-and
          dir2? f-and                   \ ac, mem
          if
            ?overrides
            0a0 opcode+w1 c,(t)  absolute-addr  \ magic number
          else
            m/r,m/r?                    \ mem/reg,mem/reg
            if
              ?overrides
              088 mem/reg,mem/reg       \ magic number
            else
              asm_invAddrMode a-error
            endif
          endif
        endif
      endif
    endif
  endif
  reset
;
+short-branches

previous definitions

-short-branches
: pop/push      \ fs gs o1 am1 o2 o3 --
  create
    , , , , , ,
  does>
    ?prefix				\ do the last one
    mode1 @ 5 = if			\ if immediate datafrom the PUSH intruction
      drop				\ drop the address of the data .. its fixed
      dup short?  b/w/d1 @ 0= f-and if	\ if data short and no forcing ; SFP015
        06a c,(t)  c,(t)		\ lay short form, lay immediate 8b data
      else
        ?overrides
        068 c,(t) absolute-data          \ lay opcode and immediate data
      endif
      reset  exit
    endif
    mode1 seg? if
      r1 @ 4 < if                       \ if DS, ES or SS
        @ r1 @ <<3 f-or c,(t)           \ seg reg operand
      else                              \ if FS or GS
        0f c,(t)                        \ lay first byte
        r1 @
        case
          4 of  5 cells + @ c,(t) endof \ the FS byte
          5 of  4 cells + @ c,(t) endof \ the GS byte
        endcase
      endif
    else
      ?overrides
      mode1 areg? if
        cell+ @ r1 @ f-or c,(t)		\ reg operand
      else
        mode1 r/i/d? must-be
        dup 3 cells + @ c,(t)
        2 cells + @ +mod+r/m1
        ?ext1,
      endif
    endif
    reset
;
+short-branches

hex

: shifts        \ am -- ;
  create
    ,
  does> @
    ?prefix                             \ do previous instruction
    ?overrides
    mode1 r/i/d? must-be                \ mem/reg,
    #?
    if
      swap dup >r  1 =                  \ shift by 1/n
      if
        0d0                             \ magic number
      else
        0c0                             \ magic number
      endif
      opcode+w1 c,(t)
      +mod+r/m1
      ?ext1,
      r> dup 1 <>			\ SFP006
      if
        c,(t)
      else
        drop
      endif                             \ n>1 lay n
    else
      0d2 opcode+w1 c,(t)               \ cl
      +mod+r/m1 c,(t)
      ?disp1,
    endif
    reset
;

hex

: rets          \ op -- ;
  mode1 @ 5 =                   \ is the operand immediate
  ret-imm-flag @ f-and          \ and been defined with #
  if
    c,(t) w,(t)                 \ always 16b immediate data
  else
    1+ c,(t)
  endif
  reset                         \ reset at end
;

: returns       \ o(near) o(far) --
  create
    , ,
  does>
    ?prefix                     \ do the last one
    far/near @ 0=
    if                          \ if near
      cell+                     \ point to the next opcode in word
    endif
    @ rets                      \ do the stuff
;


hex

also asm-access definitions

: test          \ ??? -- ;
  0 ?prefix drop                \ do the last one
  ?overrides
  #? if
    mode1 areg?
    r1 ac? f-and if
      0a8 ax/al,#                       \ magic number
    else
      mode1 r/i/d? must-be
      0f6 0 mem/reg,#l                  \ magic number
    endif
  else
    mode2 areg? must-be                  \ must be register
    mode1 r/i/d? must-be                \ no segs, or #
    084 opcode+w1 c,(t)
    0 +mod+r/m1 +reg2
    ?ext1,
  endif
  reset
;



hex

-short-branches
: xchg          \ ??? -- ;
  0 ?prefix drop                \ do the last one
  ?overrides
  r1 ac?
  mode2 areg? f-and
  word? f-and
  if                            \ XCHG AX,R
    090 r2 @ f-or c,(t)         \ magic number
  else
    r2 ac?
    mode1 areg? f-and
    word? f-and
    if                          \ XCHG R,AX
      090 r1 @ f-or c,(t)               \ magic number
    else
      mode1 areg?
      if                        \ XCHG R,M
        086 opcode+w1 c,(t)     \ magic number
        0 +mod+r/m2
        +reg1
        ?ext2,
      else                      \ XCHG M,R
        086 opcode+w2 c,(t)     \ magic number
        0 +mod+r/m1
        +reg2
        ?ext1,
      endif
    endif
  endif
  reset
;
+short-branches

decimal

previous definitions


hex

: lay-seg       \ seg -- ; -- ; set the segment# override
  create
    ,                           \ lay segment number
  does>
    ?prefix                     \ do the last one
    @ dup c,(t)                 \ lay the segment
    segreg !                    \ set reg to say which
    reset                       \ must do this, therefore xx: must be *before* instruction
;

also asm-access definitions

02e lay-seg cs:
03e lay-seg ds:
026 lay-seg es:
064 lay-seg fs:
065 lay-seg gs:
036 lay-seg ss:


hex

only forth also asm-core also asm-access definitions

: fwait         \ -- ; lays wait opcode
  here(t) 1-  last-fwait @  <>          \ previous byte <> fwait
  if
    here(t) last-fwait !                \ mark where laid
    09b c,(t)                           \ lay opcode: magic number
  endif
;


only forth also asm-access also asm-core definitions

(( unused
: segreg?       \ -- ;
  segreg @ ?dup                 \ if a segment override is used:
  if
    -1 allot(t)  ( fwait )      \ step back one byte and (not) lay fwait opcode
    c,(t)  segreg off           \ lay segreg byte, and reset
  else
\    fwait                      \ else just (not) lay fwait opcode
  endif
;
))

\ simple no operand instructions (no FWAIT): Fxxx

: f0op          \ op1 op2 -- ; -- ;
  create
    , ,                         \ lay down the two bytes of opcode
  does>
    ?prefix                     \ do the last one
    #ops @ 0= must-be           \ must be no operands
\   fwait                       \ (dont) lay an fwait
    dup                         \ spare copy of address
    cell+ @ c,(t)               \ fetch first byte, lay it
    @ c,(t)                     \ fetch second byte, lay it
\   fwait                       \ and (not) another fwait
    reset                       \ finished
;

\ simple no operand instructions (with FWAIT): Fxxx

: f0op(w)       \ op1 op2 -- ; -- ; simple no operand instructions
  create
    , ,                         \ lay down the two bytes of opcode
  does>
    ?prefix                     \ do the last one
    #ops @ 0= must-be           \ must be no operands
    fwait                       \ lay an fwait
    dup                         \ spare copy of address
    cell+ @ c,(t)               \ fetch first byte, lay it
    @ c,(t)                     \ fetch second byte, lay it
\   fwait                       \ (but not) another fwait
    reset                       \ finished
;


\ opcodes with either zero or one stack operand: Fxxx or Fxxx ST(n)

: fs1/0         \ op1 op2 -- ; -- ;
  create
    , ,                         \ lay opcodes
  does>
    ?prefix                     \ do the last one
    #ops @ 0=
    if                          \ if no registers
      dup cell+ @ c,(t)         \ lay first byte
      @ 1+ c,(t)                \ lay default second byte
    else                        \ if any registers
      mode1 @ 9 = must-be       \ must be a stack register
      dup cell+ @ c,(t)         \ lay first byte
      @ r1 @ f-or c,(t)         \ lay second byte with register number
    endif
    reset
;


\ simple one-operand instruction: Fxxx ST(n)

: fs1           \ op1 op2 -- ; -- ;
  create
    , ,                         \ lay opcodes
  does>
    ?prefix                     \ do the last one
    #ops @ 1 =                  \ must only have one operand
    mode1 @ 9 = f-and must-be   \ and it must be a st.
\   fwait                       \ (dont) lay an fwait
    dup cell+ @ c,(t)           \ lay first byte
    @ r1 @ f-or  c,(t)          \ get second op, add reg#, lay it
\   fwait                       \ (dont) lay an fwait
    reset                       \ finished
;


\ one operand (either m32 or m16): Fxxx m32 or Fxxx m16

: fm16/32       \ op32 op16 mask -- ; -- ;
  create
    , , ,                       \ lay the three required bytes
  does>
    ?prefix                     \ do the last one
\   fwait                       \ (dont) lay an fwait
    dup
    memoryformat @ 2 =
    if                          \ if m32
      2 cells +                 \ point to opcode
    else                        \ if m16
      cell+                     \ point to opcode
    endif
    @ c,(t)                     \ get opcode and lay it
    0 +mod+r/m1                 \ generate basic modr/m
    swap @                      \ get reg mask
    <<3 +                       \ generate whole modr/m
    ?ext1,
\   fwait                       \ (dont) lay an fwait
    reset
;


\ simple expression using two stack operators: Fxxx ST(n), ST

: fs1,s		\ op1 op2 -- ; -- ;
  create
    , ,                                 \ lay opcodes
  does>
    ?prefix                             \ do the last one
    mode1 @ 9 = mode2 @ 9 = f-and       \ must be st(n), st(n)
    r2 @ 0= f-and must-be                       \ and must be st(n), st0
\   fwait                               \ (dont) lay an fwait
    dup cell+ @ c,(t)                   \ get first op, lay it
    @ r1 @ f-or  c,(t)                  \ get second op, add reg#, lay it
\   fwait                               \ (dont) lay an fwait
    reset                               \ finished
;


\ one memory operand (indeterminate size): Fxxx mxx

: fmxx          \ op1 reg-mask -- ; -- ;
  create
    , ,                                 \ lay mask and opcode
  does>
    ?prefix                             \ do the last one
    mode1 @ 9 <> must-be                \ must not be a stack reg.
\   fwait                               \ (dont) lay fwait
    dup cell+ @ c,(t)                   \ lay byte of opcode
    0 +mod+r/m1                         \ generate modr/m byte
    swap @ <<3 +                        \ modify with reg-mask bits
    ?ext1,
\   fwait                               \ (dont) lay fwait
    reset
;

\ one memory operand (indeterminate size) (with FWAIT): Fxxx mxx

: fmxx(w)       \ op1 reg-mask -- ; -- ;
  create
    , ,                                 \ lay mask and opcode
  does>
    ?prefix                             \ do the last one
    mode1 @ 9 <> must-be                \ must not be a stack reg.
    fwait                               \ lay fwait
    dup cell+ @ c,(t)                   \ lay byte of opcode
    0 +mod+r/m1                         \ generate modr/m byte
    swap @ <<3 +                        \ modify with reg-mask bits
    ?ext1,
\   fwait                               \ (dont) lay fwait
    reset
;


\ either m16 or AX, with or without a FWAIT

: fm/ax         \ op1 op2 regmask wait? -- ; -- ;
  create
    , , , ,                             \ lay down bytes
  does>
    ?prefix                             \ do the last one
    mode1 @ 9 <> must-be                \ must not be a stack reg.
    dup @                               \ if a wait is needed
    if  fwait  endif                    \ lay it
    dup 3 cells + @ >r                  \ get the byte
    r1 @ 0= b/w/d1 @ 1 = f-and          \ if reg1== AX
    if
      r> 2 + c,(t)                      \ lay AX form of first opcode
      2 cells + @ c,(t)                 \ lay second byte of opcode
    else                                \ if mem16
      r> c,(t)                          \ lay mem form of first opcode
      0 +mod+r/m1                       \ generate basic modr/m
      swap cell+ @ <<3 +                \ mask in regmask bits
      ?ext1,
    endif
\   fwait                               \ dont lay fwait
    reset                               \ finished
;


: m16           \ addr -- reg-mask ;
  dup 0c + @ c,(t)              \ lay op1
  2 cells + @                   \ get reg-mask1
;

: m32           \ addr -- reg-mask ;
  dup cell+ @ c,(t)             \ lay op2
  2 cells + @                   \ get reg-mask1
;

: m64           \ addr -- reg-mask ;
  dup 3 cells + @ c,(t)         \ lay op1
  @                             \ get reg-mask2
;

\ either m16, m32 or m64: Fxxx m16 or Fxxx m32 or Fxxx m64

: fm16/32/64    \ op1 reg-mask1 op2 reg-mask2 -- ; -- ;
  create
    , , , ,                     \ lay opcode and reg mask
  does>
    ?prefix                     \ do the last one
    mode1 @ 9 <> must-be        \ must not be a stack reg
\   fwait                       \ (dont) lay fwait
    memoryformat @              \ depending on mxx size
    case
      1 of  m16 endof           \ do the 16bit form
      2 of  m32 endof           \ do the 32bit form
      3 of  m64 endof           \ do the 64bit form
        abort" Failure with memoryformat in FP opcode"
    endcase
    0 +mod+r/m1 swap <<3 +      \ generate modr/m
    ?ext1,
\   fwait                       \ (dont) lay fwait
    reset                       \ finished
;


\  either m32, m64 or st(n)

: fm32/64/s     \ op1 op2 op3 reg-mask -- ; -- ;
  create
    , , , ,                     \ lay down bytes
  does>
    ?prefix                     \ do the last one
\   fwait                       \ (dont) lay a fwait
    mode1 @ 9 =
    if                          \ if reg1 is a stack reg
      dup 2 cells + @ c,(t)     \ lay the first byte of this form
      cell+ @ r1 @ + c,(t)      \ get second byte and add reg#, then lay
    else                        \ if the operand is memory
      dup                       \ spare copy of address
      memoryformat @
      2 =
      if                        \ if the mem32 form
        3 cells +               \ point to first byte of this form
      else                      \ if the mem64 form
        2 cells +               \ point to first byte of this form
      endif
      @ c,(t)                   \ get first byte and lay it
      0 +mod+r/m1 swap @ <<3 +  \ make modr/m byte
      ?ext1,
    endif
\   fwait                       \ (dont) lay a fwait
    reset
;


\  either m32, m64, m80 or st(n)

-short-branches
: fm32/64/80/s  \ op1 op2 op3 op4 op5 reg-mask1 reg-mask2 -- ; -- ;
  create
    , , , , , , ,                       \ lay down bytes
  does>
    ?prefix                             \ do the last one
\   fwait                               \ (dont) lay a fwait
    mode1 @ 9 =
    if                                  \ if reg1 is a stack reg
      dup 3 cells + @ c,(t)             \ lay the first byte of this form
      2 cells + @ r1 @ + c,(t)          \ get second byte and add reg#, then lay
    else                                \ if the operand is memory
      memoryformat @ 1 > must-be        \ fail if wrong size
      dup                               \ spare copy of address
      memoryformat @
      case
        2 of 6 cells + @ c,(t) 4+ endof \ mem32: lay first byte of this form
        3 of 5 cells + @ c,(t) 4+ endof         \ mem64: lay first byte of this form
        4 of 4 cells + @ c,(t)    endof \ mem80: lay first byte of this form
      endcase
      0 +mod+r/m1 swap @ <<3 +          \ make modr/m byte
      ?ext1,
    endif
\   fwait                       \ (dont) lay a fwait
    reset
;
+short-branches


\ either m32, m64, st(n) or no operands
\ NB form below ignores zero-op case, user must do Fxxx ST(1)

: fm/s1/0       \ op32 mask32 op64 mask64 op1st op2st op10 op20 -- ; -- ;
  create
    , , , , , , , ,                     \ lay them all
  does>
    ?prefix                             \ do the last one
\   fwait                               \ (dont) lay a fwait
    mode1 @ 9 =
    if                                  \ if operand is stack(n)
      dup 3 cells + @ c,(t)             \ lay first opcode
      2 cells + @                       \ get second opcode
      r1 @ + c,(t)                      \ add reg# and lay opcode
    else
      memoryformat @ 1 > must-be        \ must be float or double
      memoryformat @ 2 =
      if                                \ if m32
        dup 7 cells + @ c,(t)           \ lay first opcode
        6 cells + @                     \ get reg mask
      else                              \ if m64
        dup 5 cells + @ c,(t)           \ lay first opcode
        4 cells + @                     \ get reg-mask
      endif
      <<3
      0 +mod+r/m1 +                     \ make modr/m byte
      ?ext1,
    endif
\   fwait                               \ (dont) lay a fwait
    reset                               \ finish up
;


\ either m32, m64 st,st1 or st1,st
\ NB form below ignores zero-op case, user must do Fxxx ST(1)

((
: fm/ss/0       \ reg-mask op2 -- ; -- ;
  create
    , ,                                 \ lay down these bytes
  does>
    ?prefix                             \ do the last one
\   fwait                               \ (dont) lay an fwait
    mode1 @ 9 =
    if                                  \ if either st1,st or st,st1
      r1 @ 0=
      if                                \ if st,st1
        0d8 c,(t)                       \ lay first byte
        @ r2 @ + c,(t)                  \ get second byte, add r#, and lay
      else                              \ if st1,st
        0dc c,(t)                       \ lay first byte
        @ r1 @ + c,(t)                  \ get second byte, add r#, and lay
      endif
    else                                \ if m32 or m64
      memoryformat @ 2 =
      if                                \ if m32
        0d8 c,(t)                       \ lay first byte
      else                              \ if m64
        0dc c,(t)                       \ lay first byte
      endif
      cell+ @ <<3 0 +mod+r/m1 +         \ generate modr/m
      ?ext1,
    endif
\   fwait                               \ (dont) lay an fwait
    reset                               \ finished
;
))
: fm/ss/0	\ rev-mask reg-mask op2 -- ; --
  create
    , , ,                               \ lay down these bytes
  does> ?prefix                         \ do the last one
    mode1 @ 9 = if                      \ if either st1,st or st,st1
      r1 @ 0= if                        \ if st,st1
        0d8 c,(t)                       \ lay first byte
        @ r2 @ + c,(t)                  \ get second byte, add r#, and lay
      else                              \ if st1,st
        0dc c,(t)                       \ lay first byte
\ SFP013...
        dup 2 cells + @ swap @ xor	\ xor mask with second byte for reversed form
         r1 @ + c,(t)			\ add reg#, and lay
\        @ r1 @ + c,(t)                  \ get second byte, add r#, and lay
\ ...SFP013
      endif
    else                                \ if m32 or m64
      memoryformat @ 2 = if             \ if m32
        0d8 c,(t)                       \ lay first byte
      else                              \ if m64
        0dc c,(t)                       \ lay first byte
      endif
      cell+ @ <<3 0 +mod+r/m1 +         \ generate modr/m
      ?ext1,
    endif
\   fwait                               \ (dont) lay an fwait
    reset                               \ finished
;

only forth also asm-core also asm-access definitions

\ comma n size mode -- ;
0 0 1 9 reg ST     1 0 1 9 reg ST,
0 0 1 9 reg ST(0)  1 0 1 9 reg ST(0),
0 1 1 9 reg ST(1)  1 1 1 9 reg ST(1),
0 2 1 9 reg ST(2)  1 2 1 9 reg ST(2),
0 3 1 9 reg ST(3)  1 3 1 9 reg ST(3),
0 4 1 9 reg ST(4)  1 4 1 9 reg ST(4),
0 5 1 9 reg ST(5)  1 5 1 9 reg ST(5),
0 6 1 9 reg ST(6)  1 6 1 9 reg ST(6),
0 7 1 9 reg ST(7)  1 7 1 9 reg ST(7),

\ byte1 byte2 -- ;
D9 0F0 f0op    F2XM1    D9 0E1 f0op FABS     D9 0E0 f0op FCHS
DB 0E2 f0op(w) FCLEX    DB 0E2 f0op FNCLEX   DE 0D9 f0op FCOMPP
D9 0FF f0op    FCOS     D9 0F6 f0op FDECSTP  D9 0F7 f0op FINCSTP
DB 0E3 f0op(w) FINIT    DB 0E3 f0op FNINIT   D9 0E8 f0op FLD1
D9 0E9 f0op    FLDL2T   D9 0EA f0op FLDL2E   D9 0EB f0op FLDPI
D9 0EC f0op    FLDLG2   D9 0ED f0op FLDLN2   D9 0EE f0op FLDZ
D9 0D0 f0op    FNOP     D9 0F3 f0op FPATAN   D9 0F8 f0op FPREM
D9 0F5 f0op    FPREM1   D9 0F2 f0op FPTAN    D9 0FC f0op FRNDINT
D9 0FD f0op    FSCALE   D9 0FE f0op FSIN     D9 0FB f0op FSINCOS
D9 0FA f0op    FSQRT    D9 0E4 f0op FTST     DA 0E9 f0op FUCOMPP
D9 0E5 f0op    FXAM     D9 0F4 f0op FXTRACT  D9 0F1 f0op FYL2X
D9 0F9 f0op    FYL2XP1

DB 0E0 f0op FENI     DB 0E1 f0op FDISI    \ not used in '387


\ byte1 byte2 -- ;
DD 0E0 fs1/0 FUCOM
DD 0E8 fs1/0 FUCOMP
D9 0C8 fs1/0 FXCH


\ byte1 byte2 -- ;
DD 0C0 fs1 FFREE


\ byte32 byte16 reg-mask -- ;
DA DE 00 fm16/32 FIADD
DA DE 02 fm16/32 FICOM
DA DE 03 fm16/32 FICOMP
DA DE 06 fm16/32 FIDIV
DA DE 07 fm16/32 FIDIVR
DA DE 01 fm16/32 FIMUL
DB DF 02 fm16/32 FIST
DA DE 04 fm16/32 FISUB
DA DE 05 fm16/32 FISUBR


\ byte1 byte2 -- ;
0DE 0C0 fs1,s FADDP
0DE 0F8 fs1,s FDIVP
0DE 0F0 fs1,s FDIVRP
0DE 0C8 fs1,s FMULP
0DE 0E8 fs1,s FSUBP
0DE 0E0 fs1,s FSUBRP


\ op1 reg-mask -- ;
0DF 04 fmxx FBLD                \ 10     bytes
0DF 06 fmxx FBSTP               \ 10     bytes
0D9 05 fmxx FLDCW               \  2     bytes
0D9 04 fmxx FLDENV              \ 14/28  bytes
0DD 04 fmxx FRSTOR              \ 94/108 bytes
0DD 06 fmxx FNSAVE              \ 94/108 bytes
0D9 07 fmxx FNSTCW              \  2     bytes
0D9 06 fmxx FNSTENV             \ 14/28  bytes

\ op1 reg-mask -- ;
0DD 06 fmxx(w) FSAVE            \ 94/108 bytes
0D9 07 fmxx(w) FSTCW            \  2     bytes
0D9 06 fmxx(w) FSTENV           \ 14/28  bytes

\ op1 op2 regmask wait? -- ;
DD 0E0 07 -1 fm/ax FSTSW                \ is either DD or DF depending on mode
DD 0E0 07  0 fm/ax FNSTSW

\ op1 reg-mask1 op2 reg-mask2 -- ;
DF 00 DB 05 fm16/32/64 FILD
DF 03 DB 07 fm16/32/64 FISTP

\ op1 op2 op3 op4 op5 reg-mask -- ;
D9 DD DB DD D8 03 07 fm32/64/80/s FSTP
D9 DD DB D9 C0 00 05 fm32/64/80/s FLD

\ op1 op2 op3 reg-mask -- ;
D9 DD D0 02 fm32/64/s FST

\ op32 mask32 op64 mask64 op1st op2st op10 op20 -- ;
D8 02  DC 02  D8 D0  D8 D1  fm/s1/0 FCOM
D8 03  DC 03  D8 D8  D8 D9  fm/s1/0 FCOMP


\ SFP013...
\ rev-mask reg-mask op2 --
00 00 C0 fm/ss/0 FADD
08 06 F0 fm/ss/0 FDIV
08 07 F8 fm/ss/0 FDIVR
00 01 C8 fm/ss/0 FMUL
08 04 E0 fm/ss/0 FSUB
08 05 E8 fm/ss/0 FSUBR
\ ...SFP013


only forth also asm-core definitions also asm-access definitions
hex

037 0op1 AAA     03f 0op1 AAS     0f8 0op1 CLC     09b 0op1 WAIT
0fc 0op1 CLD     0fa 0op1 CLI     0f5 0op1 CMC     0d7 0op1 XLATB
027 0op1 DAA     02f 0op1 DAS     06c 0op1 INSB    0d7 0op1 XLAT
0f4 0op1 HLT     0ce 0op1 INTO    0cf 0op1 IRET    09f 0op1 LAHF
0f0 0op1 LOCK
090 0op1 NOP
0f3 0op1 REP     0cf 0op1 IRETD   ( 0c9 0op1 LEAVE )
0f3 0op1 REPE    0f3 0op1 REPZ    0f2 0op1 REPNE   0f2 0op1 REPNZ
09e 0op1 SAHF    0f9 0op1 STC
0fd 0op1 STD     0fb 0op1 STI

0a6 0op1    CMPSB       0ac 0op1    LODSB   0a4 0op1    MOVSB
0a7 0op1-16 CMPSW       0ad 0op1-16 LODSW   0a5 0op1-16 MOVSW
0a7 0op1-32 CMPSD       0ad 0op1-32 LODSD   0a5 0op1-32 MOVSD

06e 0op1    OUTSB       0ae 0op1    SCASB   0aa 0op1    STOSB
06f 0op1-16 OUTSW       0af 0op1-16 SCASW   0ab 0op1-16 STOSW
06f 0op1-32 OUTSD       0af 0op1-32 SCASD   0ab 0op1-32 STOSD

06d 0op1-16 INSW        098 0op1-16 CBW     099 0op1-16 CWD
06d 0op1-32 INSD        098 0op1-32 CWDE    099 0op1-32 CDQ

0d5 0a 0op2 AAD
0d4 0a 0op2 AAM
0f  06 0op2 CLTS
0F  A2 0op2 CPUID
$0F $31 0op2 RDTSC
$0F $32 0op2 RDMSR	\ SFP015
$0F $30 0op2 WRMSR	\ SFP015
$0F $77 0op2 EMMS	\ SFP015
$0F $33 0op2 RDPMC	\ SFP015
$0F $AA 0op2 RSM	\ SFP015


hex

010 014 080 010 3modes   ADC      000 004 080 000 3modes   ADD
( 020 024 080 020 3modes/l AND )     038 03C 080 038 3modes   CMP
( 008 00C 080 008 3modes/l OR  )     018 01C 080 018 3modes   SBB
028 02C 080 028 3modes   SUB      ( 030 034 080 030 3modes/l XOR )

048 0FE 008 inc/dec DEC
040 0FE 000 inc/dec INC

030 mul/div DIV         038 mul/div IDIV
020 mul/div MUL
018 mul/div NEG         ( 010 mul/div NOT )

hex
\ o16 -- ;
077 j/lp JA     077 j/lp JNBE   073 j/lp JAE   073 j/lp JNB
072 j/lp JB     072 j/lp JNAE   076 j/lp JBE   076 j/lp JNA
0E3 j/lp JCXZ   074 j/lp JE     074 j/lp JZ    07F j/lp JG
07F j/lp JNLE   07D j/lp JGE    07D j/lp JNL   07C j/lp JL
07C j/lp JNGE   07E j/lp JLE    07E j/lp JNG   0e3 j/lp JECXZ
075 j/lp JNE    075 j/lp JNZ    071 j/lp JNO   07B j/lp JNP
07B j/lp JPO    079 j/lp JNS    070 j/lp JO    07A j/lp JP
07A j/lp JPE    078 j/lp JS     072 j/lp JC    073 j/lp JNC

( 0E2 j/lp LOOP )   0E1 j/lp LOOPZ  0E1 j/lp LOOPE 0E0 j/lp LOOPNZ
\ LOOP is defined very late to avoid problems with the Forth LOOP
\ and the assembler's LOOP
0E0 j/lp LOOPNE

hex

$00F $008 0op2 INVD			\ SFP016
$0C8 1op-r32 BSWAP			\ SFP016
$0B0 2op-r/m,r CMPXCHG			\ SFP016
$0C0 2op-r/m,r XADD			\ SFP016

0C5 reg,mem LDS         0f b2 reg,mem(2) LSS
0C4 reg,mem LES         0f b4 reg,mem(2) LFS
                        0f b5 reg,mem(2) LGS

08D reg,mem LEA

0a1 0a9 08F 000 058 007 pop/push POP
0a0 0a8 0FF 030 050 006 pop/push PUSH

010 shifts RCL          018 shifts RCR
( 000 shifts ROL          008 shifts ROR )

038 shifts SAR          020 shifts SAL
020 shifts SHL          028 shifts SHR

0c2 0CA returns RET

: ret/f         \ [imm] -- ;
  0CA rets
;

\ byte1 byte2 -- ;
0f bc bit-scan BSF
0f bd bit-scan BSR

\ /d byte1 byte2(reg) byte2(imm) -- ;
04 0f a3 ba bit-test BT
07 0f bb ba bit-test BTC
06 0f b3 ba bit-test BTR
05 0f ab ba bit-test BTS

0f 02 2op-r,r/m LAR
0f 03 2op-r,r/m LSL
0f 0be 0bf mov(x)x MOVSX
0f 0b6 0b7 mov(x)x MOVZX

\ SFP011...
0F 40 reg,mem(3) CMOVO
0F 41 reg,mem(3) CMOVNO
0F 42 reg,mem(3) CMOVB
0F 42 reg,mem(3) CMOVC
0F 42 reg,mem(3) CMOVNAE
0F 43 reg,mem(3) CMOVAE
0F 43 reg,mem(3) CMOVNB
0F 43 reg,mem(3) CMOVNC
0F 44 reg,mem(3) CMOVE
0F 44 reg,mem(3) CMOVZ
0F 45 reg,mem(3) CMOVNE
0F 45 reg,mem(3) CMOVNZ
0F 46 reg,mem(3) CMOVBE
0F 46 reg,mem(3) CMOVNA
0F 47 reg,mem(3) CMOVNBE
0F 47 reg,mem(3) CMOVA
0F 48 reg,mem(3) CMOVS
0F 49 reg,mem(3) CMOVNS
0F 4A reg,mem(3) CMOVP
0F 4A reg,mem(3) CMOVPE
0F 4B reg,mem(3) CMOVNP
0F 4B reg,mem(3) CMOVNPO
0F 4C reg,mem(3) CMOVL
0F 4C reg,mem(3) CMOVNGE
0F 4D reg,mem(3) CMOVNL
0F 4D reg,mem(3) CMOVGE
0F 4E reg,mem(3) CMOVLE
0F 4E reg,mem(3) CMOVNG
0F 4F reg,mem(3) CMOVNLE
0F 4F reg,mem(3) CMOVG
\ ...SFP011

1 061 0op16/32 POPA
2 061 0op16/32 POPAD
1 09d 0op16/32 POPF
2 09d 0op16/32 POPFD
1 060 0op16/32 PUSHA
2 060 0op16/32 PUSHAD
1 09c 0op16/32 PUSHF
2 09c 0op16/32 PUSHFD

090 set SETO
091 set SETNO
092 set SETB            092 set SETC            092 set SETNAE
093 set SETAE           093 set SETNB           093 set SETNC
094 set SETE            094 set SETZ
095 set SETNE           095 set SETNZ
096 set SETBE           096 set SETNA
097 set SETA            097 set SETNBE
098 set SETS
099 set SETNS
09a set SETP            09a set SETPE
09b set SETNP           09b set SETPO
09c set SETL            09c set SETNGE
09d set SETGE           09d set SETNL
09e set SETLE           09e set SETNG
09f set SETG            09f set SETNLE

0f 0a4 d-shift SHLD
0f 0ac d-shift SHRD

0f 01 02 load-table LGDT
0f 01 03 load-table LIDT
0f 01 00 load-table SGDT
0f 01 01 load-table SIDT

0f 00 02 2op-r/m LLDT
0f 01 06 2op-r/m LMSW
0f 00 03 2op-r/m LTR
0f 00 00 2op-r/m SLDT
0f 01 04 2op-r/m SMSW
0f 00 01 2op-r/m STR
0f 00 04 2op-r/m VERR
0f 00 05 2op-r/m VERW

only forth also asm-core definitions


hex
0 constant con-flg

: ?>mark        \ -- f addr
  con-flg  here(t)
  0 c,(t)
;

: ?>resolve     \ f addr --
  here(t) over 1+ -  dup ?short  swap c!(t)
  con-flg ?pairs
;

: ?<mark        \ -- f addr
  con-flg  here(t)
;

: ?<resolve     \ f addr --
  here(t) 1+ -  dup ?short  c,(t)
  con-flg  ?pairs
;

hex

also asm-access definitions

75 constant Z,     74 constant NZ,    79 constant S,
78 constant NS,    7D constant L,     7C constant GE,
7F constant LE,    7E constant G,     73 constant CY,
72 constant NC,    77 constant BE,    76 constant A,
71 constant O,     70 constant NO,    7A constant PO,
7B constant PE,    72 constant AE,    73 constant B,
E3 constant NCXZ,  72 constant NCY,

previous definitions

hex

also asm-access definitions

\ ***********************
\ *S Assembler structures
\ ***********************
( *P Structures like the Forth control structures have been added to the )
( ** assembler. They allow forward branches without the need for labels and )
( ** impose the strictures of structured programming to the assembler level. )
( ** Devotees of spaghetti programming are free to go their own way as the )
( ** copious supply of branch instructions are still available, and the local )
( ** label facility may be used with all branch instructions. )

( *P The status flag indicator required must prefix conditional structures. The )
( ** structure assembled will have a branch opcode that is the logical inverse of )
( ** the one specified. Thus for *\fo{EQ, IF,} a JNE will be assembled so that the code )
( ** after *\fo{IF,} is executed if the EQ status occurs. The assembler structure words )
( ** end in a comma e.g. *\fo{IF,} to differentiate them from the regular Forth )
( ** structures, and to indicate that code is being generated. )

\ *P The structures are described below, and the symbol cc condition code) may
\ ** be any one of the following:
( *D Z,		equal to 0 )
( *D NZ,	not equal to 0 )
( *D S,		less than 0 )
( *D NS,	greater than or equal to 0 )
( *D L,		less than )
( *D GE,	greater than or equal )
( *D LE,        less than or equal )
( *D G,         greater than )
( *D B,		unsigned less than - address compares )
( *D AE,        unsigned greater than or equal )
( *D BE,        unsigned less than or equal )
( *D A,         unsigned greater than )
( *D O,         overflow )
( *D NO,	no overflow )
( *D PE,	parity even )
( *D PO,        parity odd )
( *D CY,        carry flag set )
( *D NC,        carry flag not set )
( *D NCXZ,      ECX/CX register non-zero )

( *P The structure words build sets of assembler branches to perform functions )
( ** equivalent to their high-level namesakes, but the names end with a comma to )
( ** distinguish them from the high-level Forth words. Be sure you understand the )
( ** high level structures before using the assembler equivalents. )
( ** The structures are:

( *E cc IF, ... THEN,			)
( ** cc IF, ... ELSE, ... THEN,		)
( ** 					)
( ** BEGIN, ... AGAIN,			)
( ** BEGIN, ... cc UNTIL,		)
( ** BEGIN, ... cc WHILE, ... REPEAT,	)

( *P An additional structure allows a section of code to be performed 'n' times. )
( ** All it actually does is to load ECX with 'n'and mark the start of a backward )
( ** branch so that the mark may be used later. The structure is: )
( *E n TIMES, ... LOOP,			)

: if,	\ ??? -- ;
  >r null-op r>                 \ preserve condition, and do last opcode
  c,(t)  ?>mark                 \ sort this one out
;

: then,
  null-op  ?>resolve
;

: endif,
  null-op  ?>resolve
;

: else,                                 \ macro
  null-op  0eb                          \ magic number
  if,
    2swap
  then,
;

: begin,
  null-op  ?<mark
;

: until,                \ ??? -- ;
  >r  null-op  r> c,(t)  ?<resolve
;

: again,
  null-op  0eb until,                    \ macro with magic number
;

: while,
  if,                           \ macro
;

: repeat,
  null-op
  2swap
  again,  then,         \ macro
;

: times,	\ n -- ; macro
  prefix? if
    use32? if
      mov ecx, #
      here(t) 5 + swap
    else
      mov cx, #
      here(t) 3 + swap
    endif
  else
    use32? if
      ecx, # mov
    else
      cx, # mov
    endif
    here(t)
  endif
;

: loop,		\ orig --
  null-op  $0F2 until,                    \ macro with LOOP opcode
;


previous definitions


only forth also asm-access also asm-core definitions

: asm-init      \ cs dp --
  segreg off
  last-fwait off                     \ reset assembler
  reset  initfwd
;

decimal

only forth also asm-core definitions

also asm-core

: asm-on	\ -- ; turn assembler on
  asm-init                              \ initialise assembler
  also asm-access
;

: asm-off       \ -- ; turn assembler off
  previous
;

only forth  also asm-core  also asm-access  also forth definitions

decimal

target-only
: Proc		\ "name" --
  ?exec !csp  create
    initfwd  asm-start asm-on
;  immediate

: lbl:		\ "name" --
  ?exec !csp  create
    asm-start  asm-on
;  immediate

: code		\ "name" --
  ?exec !csp
\  create  0 >body negate allot
  bl word count header code-align	\ sort of CREATE ; SFP012
    initfwd  asm-start asm-on
    here last-colon-xt !
;  immediate
host&target

: ;code		\ --
  ?comp ?csp
  discard-sinline			\ SFP004
  drop-locals                           \ compiles in stack cleanup ; SFP009
  cleanup-locals                        \ cleans up local dictionary etc. ; SFP009
  compile (;code)
  state off
  !csp
  initfwd  asm-start asm-on
  r> drop				\ exit caller
;  immediate

: end-code	\ --
  asm-end  asm-off
  unres-refs? #-257 ?throw
  ?exec ?csp  reveal
;  immediate


only forth  also asm-core  also asm-access definitions

: next,		\ return to Forth
  ret
;

: ,             \ -- ; define comma separator for the assembler
  [ only forth also asm-core ]
  #ops @ 2 <  co1 @ 0= f-and if \ if: ...
    co1 on                      \ then force: ...,
    1 #ops !
    exit
  endif
  co1 @ if                      \ if: ...,...
    co2 on                      \ then force: ...,...,
    2 #ops !
  endif
  [ also asm-access ]
;

synonym ,a ,

: d16:          \ -- ;
  use32?                        \ if default is 32b
  if
    d-override on               \ then set override
  endif
;

: d32:          \ -- ;
  use16?                \ if default is 16b
  if
    d-override on               \ then set override
  endif
;

: a16:          \ -- ;
  use32?                        \ if default is 32b
  if
    a-override on               \ then set override
  endif
;

: a32:          \ -- ;
  use16?                        \ if default is 16b
  if
    a-override on               \ then set override
  endif
;

: far           \ -- ;
  far/near on
;

: near          \ -- ;
  far/near off
;

: short         \ -- ;
  short/long off
;

: long          \ -- ;
  short/long on
;

: word		\ --
  worda
;
hex

\ Conflict with Forth words of same name
020 024 080 020 3modes/l AND
008 00C 080 008 3modes/l OR
030 034 080 030 3modes/l XOR
010 mul/div NOT
000 shifts ROL          008 shifts ROR
0c9 0op1 LEAVE
0E2 j/lp LOOP

\ These have 'a' suffix to avoid name collision during cross compilation
020 024 080 020 3modes/l ANDa
008 00C 080 008 3modes/l ORa
030 034 080 030 3modes/l XORa
010 mul/div NOTa
000 shifts ROLa         008 shifts RORa
0c9 0op1 LEAVEa
0E2 j/lp LOOPa


\ ==========================
\ *S Assembler mode switches
\ ==========================

decimal

only forth also asm-core also forth definitions

: use16		\ -- ; force 16bit segment
  0 seg-type !  reset  ;

: use32         \ -- ; force 32bit segment
  -1 seg-type !  reset  ;

: mpe           \ -- ; force def # addressing
( *G Select the MPE default addressing mode, in which )
( ** the default addressing mode is direct addressing. )
( ** This is provided for compatibility with legacy MPE	)
( ** systems. )
( ** The indicators *\fo{[]} and *\fo{#} can be used for code	)
( ** which must be compiled in either condition.	)
  default-mode off  reset  ;

: intel         \ -- ; force def.direct addressing
( *G Select the INTEL default addressing mode, in which )
( ** the default addressing mode is immediate addressing. )
( ** The indicators *\fo{[]} and *\fo{#} can be used for code	)
( ** which must be compiled in either condition.	)
  default-mode on  reset  ;

: prefix	\ -- ; select prefix mode
( *G Set the assembler to use prefix notation with the opcode first. )
( ** This is the default condition. )
  <prefix> on  ;

: postfix	\ -- ; select postfix mode
( *G Set the assembler to use postfix notation with the opcode )
( ** after the operands. )
  <prefix> off  ;

only forth definitions


\ ******************************
\ *S Macros and Assembler access
\ ******************************

( *P Because of the performance of the VFX optimiser, use of assembler )
( ** is only necessary when defining new compilation structures. Otherwise )
( ** the use of assembler code should be avoided. )

( *P Assembler macros are defined as follows: )
( *E MASM: <name>
( **   <assembler code goes here>
( ** ;MASM )
( ** e.g. the following macro pops the top of the NDP stack to the external )
( ** floating point stack. )
( ** MASM: popFPU	\ -- ; pops FTOS to float stack              )
( **   mov   eax, FSP-OFFSET [esi]   \ get FP stack pointer          )
( **   lea   eax, -FPCELL [eax]      \ update stack pointer          )
( **   fstp  fword 0 [eax]           \ store and pop                 )
( **   mov   FSP-OFFSET [esi], eax   \ restore FP stack pointer      )
( ** ;MASM )

( *P In line assembler code may be compiled into the middle of a colon )
( ** definition by using the phrase: )
( *E : <name>      \ just another Forth word )
( **   ... )
( **   [ASM  <insert assembler here>  ASM] )
( **   ... )
( ** ; )

( *P A fragment of assembler for compilation when the containing )
( ** word is executed can be defined by using the following: )
( *E : a-compiler  \ will compile some assembler )
( **   ... )
( **   a[ <fragment to be compiled> ]a
( **   ... )
( ** ; )
( *P The following example compiles an in-line floating point literal. )
( *E : o_flit,	\ F: f -- ; F: -- f ; compile floating point literal   )
( **   a[ popFPU                        \ references a previous macro  )
( **      jmp   here 2+  FPCELL +       \ skip inline literal          )
( **   ]a                                                              )
( **   f,                                                              )
( **   a[ fld   fword ptr here FPCELL - ]a                             )
( ** ;                                                                 )

\ *P When using in-line code generators such as *\fo{[ASM ...ASM]}
\ ** you should flush the code generator contents with *\fo{[O/F]}.
( *E [O/F] [ASM ... ASM]	)
\ *P After *\fo{[ASM} the top of the data stack will be in EBX
\ ** with all other stack items pointed to by EBP. The code
\ ** generator expects this same state to exist after *\fo{ASM]}.

also asm-core
: MASM:		\ --
  ?exec
  also asm-access definitions
  :
  postpone asm-end  postpone asm-start
; immediate

: ;MASM		\ --
  ?comp
  previous definitions
  postpone asm-end  postpone ;
; immediate

: [ASM		\ -- ; start assembler sequence
  ?comp  asm-end asm-start
  also asm-access  state off
; immediate

: ASM]		\ -- ; end assembler sequence
  asm-end  ]  previous			\ SFP010
;

: a[            \ -- ; start assembler sequence to be compiled later, A[ ... ]A
  also asm-access
  postpone asm-init
;  immediate

: ]a            \ -- ; end of assembler sequence
  postpone asm-end
  previous  ;  immediate
previous


 \ ******************************
 \ *S Generating new instructions
 \ ******************************
 \ *P Some processors, especially those used for embedded
 \ ** applications, have processor-specific instructions that are
 \ ** extensions to the iA32 instruction set. In order to use these,
 \ ** a few facilities are available.

also asm-core also asm-access definitions
: dxb            \ b -- ; lay byte
\ *G Lay a byte into the instruction stream. Use in the form:
\ *C   dxb $55
  0 ?prefix drop  c,(t) reset  ;

: dxw            \ w -- ; lay 16 bits
\ *G Lay a 16-bit word into the instruction stream. Use in the form:
\ *C   dxw $55AA
  0 ?prefix drop  w,(t) reset  ;

: dxl            \ l -- ; lay 32 bit long
\ *G Lay a 32-bit dword into the instruction stream. Use in the form:
\ *C   dxl $11223344
  0 ?prefix drop  ,(t) reset  ;

: $             \ -- chere
\ *G Return the PC value of the start of the instruction.
  here(t)  ;
previous previous definitions


\ ========================
\ *S Assembler error codes
\ ========================
\ *C #-701 Invalid addressing mode
\ *C #-702 N not in range -128..+127
\ *C #-703 Label reference number out of range
\ *C #-704 Label definition number out of range
\ *C #-705 Invalid instruction for selected CPU type


\ ======
\ *> ###
\ ======


