\ Tokeniser.fth - tokeniser for VFX Forth

((
Copyright (c) 2009
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441

Design Notes
============
The CODE-LEN field of a word header contains either the length of the
word's binary code (<64kb) or the address of its compiler. The
difference is detected by binary code sizes always being less than
64kb, and word addresses always being above 64kb. The XT of the word
being compiled is always passed to its compiler.

The CODE-GEN field of a word header contains the address of a
/TOKENSTREAM structure which controls recompilation of the word.
It holds details such as search order state.

/TOKENSTREAM structures are currently allocated from the heap, and
therefore disappear when the program is closed. To prevent problems,
the tokeniser is disabled for xts below FENCE, set whenever a
program is SAVEd.

When the word is first compiled, a /TokenStream structure is built
into the TokenSpace array allocated from the heap. This is followed
by the token stream itself. If it is decided that the word cannot
be tokenised, e.g. contains EXIT, the /TokenStream structure is
invalidated.


To do
=====

Change history
==============
20200809 SFP004 Correctd for SSE64.
20200722 SFP004 Added DOCALL handler.
20180501 SFP003 Corrected for NDCS changes.
20110714 SFP002 Improved recovery after error.
20090910 SFP001 Improved DISCARD-SINLINE.
))


\ ===========
\ *> opcontrl
\ *S Tokeniser
\ ===========
\ *P From VFX Forth v4.3, build 2825, the tokeniser replaces the
\ ** previous source inliner. The change was made to improve ANS
\ ** and Forth200x standards compliance, and to reduce issues
\ ** with particularly "guru" code. To prevent breaking your
\ ** existing code, the tokeniser uses the same word names for
\ ** its control words.

\ *P The tokeniser keeps track of what is compiled for a word,
\ ** and reruns the compilation of short definitions rather than
\ ** copying the compiled code inline. This gives the VFX code
\ ** generator many more opportunities to remove stack operations
\ ** and produces smaller and faster code while encouraging users
\ ** to write short definitions. That having been said, the
\ ** relationship of code size with and without the tokeniser
\ ** enabled is obscure at best.

\ *P Under some rare conditions, usually those requiring tinkering
\ ** with internal structures of VFX Forth during compilation, it is
\ ** necessary to have a level of control over the tokeniser. This
\ ** section documents those words.

only forth  also system  also h-c definitions


\ **************
\ Error messages
\ **************

internal
#-399	#ErrDef	tok_Bad		"SIN: Invalid token"
#-398	#ErrDef tok_BadStart	"SIN: Failed memory allocation."
external

\ ************
\ Control data
\ ************

also forth definitions
true value debugSin?	\ -- flag ; true to debug ; SFP011
FALSE Value Sin?	\ -- flag ; true if SIN enabled
false value Sindoes?	\ -- flag ; true if SINDOES enabled
0 value SinActive?	\ -- flag ; true if SIN compiler active ; SFP014
previous definitions

(( now in optimiser, see opt386.fth
0 value GenTokens?	\ -- flag ; true to generate a token stream
0 Value pSourceSin	\ -- addr ; pointer to token stream for word
))
0 Value SinIsForDOES	\ -- flag ; true if in DOES> clause

0 value TokenSpace	\ -- addr ; TokenSpace allocated from heap
1 Mb value /TokenSpace	\ -- len ; size of token space
0 value TokenOffset	\ -- addr ; current offset in TokenSpace
0 value ExecTokens?	\ -- flag ; true to execute a token stream
0 value ExecDone?	\ -- flag ; true when token execution has finished.
0 value DPstart		\ -- addr ; holds the start address of a word.

: NoTokenSpace	\ --
\ Mark token space as unusable.
  1 Mb to /TokenSpace  0 to TokenSpace  0 to TokenOffset
  0 -> GenTokens?  0 -> ExecTokens?
  0 -> ExecDone?  0 -> DPstart
;
' NoTokenSpace AtCold

: setTokenMem	\ addr ior --
\ Enable or disable token space according to ior.
  if
    0 -> Sin?  0 -> SinDoes?
    drop noTokenSpace
  else
    to TokenSpace
  endif
;

: initTokenSpace	\ --
\ Initialise token space and stream pointers.
  /TokenSpace allocate setTokenMem
  #32 to TokenOffset			\ must not be zero
;
' initTokenSpace AtCold

: CheckTokenSpace	\ --
\ Resize the token space if required.
  /TokenSpace TokenOffset - 8 kb < if
    TokenSpace /TokenSpace 2* dup -> /TokenSpace
    resize setTokenMem
  endif
;

: !tok		\ x offset --
\ Store x at the given offset in the token space.
  TokenSpace + !
;

: @tok		\ offset -- x
\ Fetch x from the given offset in the token space.
  TokenSpace + @
;

: ,tok		\ x --
\ Add cell x to the token stream. This word is forward referenced by
\ COMPILE-REF and NDCS-REF in the optimisers (OPTxxx.fth) to generate
\ the token stream.
\ dup .dword space
  TokenOffset !tok  TokenOffset cell + -> TokenOffset
;

#12 constant /flitToken

: f!tok		\ F: f -- ; offset --
\ Store f at the given offset in the token space.
  TokenSpace + s" f!" assess
;

: f@tok		\ offset -- ; F: -- f
\ Fetch f from the given offset in the token space.
  TokenSpace + s" f@" assess
;

: f,tok		\ F: f --
\ Add float f to the token stream.
\ dup .dword space
  TokenOffset f!tok  TokenOffset /flitToken + -> TokenOffset
;

variable TIP	\ -- addr
\ Token interpreter pointer.

: NextToken	\ -- x
\ Fetch next token and update TIP.
  TIP @ @tok  cell TIP +!
;

: NextFToken	\ F: -- f
\ Fetch next float token and update TIP.
  TIP @ f@tok  /flitToken TIP +!
;

: execLitToken	\ --
\ Compile a literal.
  NextToken postpone literal
;

: execFlitToken	\ --
\ Compile a floating point literal. This is a kluge to cope with
\ the various floating point packs.
  NextFToken
  case  FpSystem
    1 of  -2 c#fliteral  endof		\ HFP387
    2 of  -2 c#fliteral  endof		\ NDP387
    3 of  -1 c#fliteral  endof		\ OpenGL32
    4 of  -1 c#fliteral  endof		\ SSE64 ; SFP005
      (fliteral)
  endcase
;

: execCallToken	\ --
\ Compile an absolute call.
  nextToken doLayAbsCall
;


\ ==================
\ *N Tokeniser state
\ ==================
\ The token stream is a threaded system containing the following
\ cell tokens:
\  0         illegal
\  1         start of tokenisation
\  2         end of tokenisation
\  3         32 bit literal
\  4         FP literal
\  xt        target xt to be compiled

0 constant BadToken
1 constant StartToken
2 constant EndToken
3 constant LitToken
4 constant FlitToken
5 constant callToken	\ xt laid by DOCALL,

#15 constant #Specials

also forth definitions
: discard-sinline       \ --
\ *G Stops the current definition from being handled by the
\ ** tokeniser. This is usually required by a compilation
\ ** word which generates inline data, and for which repetition
\ ** of the word containing the inline data would generate large
\ ** code with little speed advantage.
  pSourceSin if
    BadToken ,tok  0 to pSourceSin  0 to GenTokens?
  endif
;

#128 Value SinThreshold	\ -- u
\ *G If the binary size of a word is less than this value,
\ ** it can be tokenised. *\b{Subject to change}.
previous definitions


\ *************************
\ Compiling tokenised words
\ *************************

: canSin?	\ -- flag
\ True if we can tokenise.
  Sin? optimising @ and
;

: begin-sinline	\ --
\ +G Initialise the tokeniser during compilation, producing an
\ +* initialised token stream.
  canSin? 0= if  exit  then		\ checks
  discard-sinline			\ discard previously allocated tokens
  checkTokenSpace			\ check token memory
  1 to GenTokens?			\ enable token generation
  TokenOffset to pSourceSin		\ mark start
  here -> DPstart
  StartToken ,tok			\ lay start token
;

: end-sinline	\ --
\ +G Finish tokenisation and finalise the token stream.
  pSourceSin if
    here DPstart - SinThreshold u>	\ get out if too big
    if  discard-sinline  exit  then
    EndToken ,tok			\ exit token
    pSourceSin latest-xt tuck >code-gen !	\ set pointer to token stream
    \ don't set the new compiler if we are using the code-gen
    \ field to store token data for children
    SinIsForDOES 0= if			\ if not a defining word
      ['] c_sin swap >code-len !	\ set the word's compiler
    else
      drop
    then
    0 to pSourceSin			\ mark all done
    0 to GenTokens?			\ SFP002
  then
;

: genLitToken	\ x --
\ Generate a literal token for value x.
  GenTokens? if
    LitToken ,tok  ,tok  exit
  endif
  drop
;

: genFlitToken	\ F: f --
\ Generate a literal token for value x.
  GenTokens? if
    FlitToken ,tok  f,tok  exit
  endif
  s" fdrop" assess
;

also forth definitions
: .Tokens	\ xt --
\ *G Display the token stream for a word.
  TIP @ >r
  >code-gen @ dup TIP ! if
    0 begin
      case  NextToken cr
        BadToken of  ." Bad Token"  drop -1  endof
        StartToken of  ." StartToken"  endof
        EndToken of  ." End Token"  drop -1  endof
        LitToken of  ." Lit " nextToken .dword    endof
        FlitToken of  ." Flit " NextFToken s" f." assess  endof
        callToken of  ." ['] " nextToken ip>nfa .name  ." doCall,"  endof
          dup >name .name
      endcase
      dup
    until
    drop
  else
    cr ." No token stream"
  endif
  r> TIP !
;

: .Tokeniser	\ --
\ *G Display tokeniser state
  cr ." Optimising? " optimising? .
  cr ." Sin? " sin? .
  cr ." GenTokens? " GenTokens? .
  cr ." TokenOffset " TokenOffset .dword
  cr ." /TokenSpace " /TokenSpace .dword
  cr
;
previous definitions


\ ************************
\ Executing a token stream
\ ************************

: execTok	\ token -- res ; 0=continue
\ Perform the action of a token. Return 0 to do the next one,
  case
    LitToken of  execLitToken  endof	\ literal
    EndToken of  -1 -> ExecDone?  endof	\ end of stream
    FlitToken of  execFlitToken  endof	\ FP literal
    callToken of  execCallToken  endof	\ absolute call
      tok_Bad throw			\ catches 0, 1
  endcase
;

: execTokens	\ --
\ Token execution loop
  NextToken StartToken <> tok_BadStart ?throw
  begin
    NextToken dup #Specials u<
    if  execTok  else  compile-word  endif
    ExecDone?
  until
;

: execStream	\ offset --
\ Execute the token stream at the given offset.
  TIP @ >r  TIP !			\ TIP is token pointer
  GenTokens? >r  0 -> GenTokens?	\ don't generate tokens
  SinActive? >r  -1 -> SinActive?	\ tokeniser active
  ExecDone? >r  0 -> ExecDone?		\ not done yet
\  state @ >r  ]				\ necessary?
  ['] execTokens catch
\  r> state !
  r> -> ExecDone?
  r> -> SinActive?
  r> -> GenTokens?
  r> TIP !
  throw
;

: untokened,	\ xt --
\ Remove token info and compile a plain unoptimised reference.
  dup >code-len ZeroOptData  docall,
;

: c_sin		{ xt -- }
\ +G Compile the tokenised code for the given definition.
  canSin? 0=				\ don't if not enabled
  if  xt docall,  exit  then
  xt fence @ <				\ don't if below FENCE, why?
  if  xt untokened,  exit  then		\ because SAVE/BYE loses the token data
  xt >code-gen @ ?dup 0=		\ get token offset
  if  xt untokened,  exit  then		\ don't if no tokens
  execStream
;

: c_does_sin    { xt -- }
\ +G Code generator for DOES> portion.
  canSin? Sindoes? and 0=
  if  xt docall,  exit  then
  xt fence @ <				\ don't if below FENCE, why?
  if  xt untokened,  exit  endif	\ because SAVE/BYE loses the SINLINE data
  xt >code-gen @ ?dup 0=		\ don't if no token structure
  if  xt untokened,  exit  then
  xt >body clit,  execStream		\ lay literal for data address
;

: does_sin      \ -- ; executed by DOES> at compile time
\ +G The tokeniser version of (;CODE) and (DOES>).
\ +* Sets the code-gen and code-len fields of the child,
\ +* using the CODE-GEN address compiled inline. DOES_SIN
\ +* is followed by the original (;CODE) or (DOES>).
  r> dup @ @ swap cell+ >r        	\ get token offset, update return ; -- xt
  latest-xt tuck >code-gen !		\ set the two fields in the child ; SFP013
  ['] c_does_sin swap >code-len !
;


\ ============================
\ +N Redefinitions for Control
\ ============================

also forth definitions

: :             \ "name" --
  :  begin-sinline  false to SinIsForDOES
;

((
: :NNFAST	\ C: -- colon-sys ; Exec: i*x -- j*x ; R: -- nest-sys ; SFP013
  :noname  begin-sinline  false to SinIsForDOES
;
))

: :NONAME       \ C: -- colon-sys ; Exec: i*x -- j*x ; R: -- nest-sys ; SFP013
  :noname  not-inlined  false to SinIsForDOES
;

: ;             \ --
  UnSafe-RS?				\ not for return stack hackers
  if  discard-sinline  then
  postpone ;  end-sinline
; immediate

: does>         \ --
  discard-sinline			\ won't sinline the CREATE part
  canSin? sindoes? and if 		\ check conditions
    ['] does_sin docall,		\ this will set the child's pointers
    latest-xt >code-gen ,		\ from the CODE-GEN field of the parent
    postpone does>			\ use original DOES>
    true to SinIsForDOES		\ flag it
    begin-sinline			\ restart sinline
  else
    postpone does>
  endif
; immediate

: interp>	\ --
  discard-sinline  postpone interp>
; immediate

previous definitions


\ ====================
\ *N Tokeniser control
\ ====================

also forth definitions

(( for DocGen
FALSE Value Sin?	\ -- flag
\ *G A *\fo{VALUE} which enables tokenising when set. Using
\ ** *\fo{SIN?} enables you to determine the state of the
\ ** tokeniser

false value sindoes?	\ -- flag
\ *G A *\fo{VALUE} which enables tokenising of *\fo{DOES>}
\ ** clauses when set. Using this value enables you to determine
\ ** the state of the tokeniser.

false value SinActive?	\ -- flag
\ *G Returns true when the tokeniser is active. It is used
\ ** to inibit some immediate words which must not be rerun when
\ ** the word they are in is tokenised.
))

: +sin		\ --
\ *G Enable tokenising of following definitions.
  true to Sin?
;

: -sin          \ --
\ *G Disable tokenising of following definitions.
  false to sin?
  false to sindoes?
;

: +sindoes      \ --
\ *G Enable tokenising of the run time portions of defining words.
\ ** Many defining words produced with *\fo{CREATE ... DOES>} have short
\ ** run time actions. The address returned by *\fo{DOES>} is a literal and
\ ** provides many opportunities for both space and speed optimisation.
  true to sindoes?
;

: -sindoes      \ --
( *G Disable tokenising of the run time portions of defining words. )
  false to sindoes?
;

: [sin		\ -- i*x
\ *G *\fo{[SIN} and *\fo{SIN]} define a range of source code and
\ ** must be used interpretively, not during compilation.
\ ** *\fo{[SIN} saves the current tokeniser state and *\fo{SIN]}
\ ** restores it. Often used in the form:
\ *C   [SIN -SIN ... SIN]
  ?exec  sin? sindoes?
; immediate

: sin]		\ i*x --
\ *G See *\fo{[SIN} above.
  ?exec  -> sindoes?  -> sin?
; immediate

: [-sin		\ -- i*x
\ *G *\fo{[-SIN} saves the current tokeniser state, and turns
\ ** off the tokeniser. *\fo{SIN]} restores the saved tokeniser
\ ** state. Used in the form:
\ *C   [-SIN ... SIN]
  ?exec  sin? sindoes?  -sin -sindoes
; immediate

: [+sin		\ -- i*x
\ *G *\fo{[+SIN} saves the current tokeniser state, and turns
\ ** on the tokeniser. *\fo{SIN]} restores the saved tokeniser
\ ** state. Used in the form:
\ *C   [+SIN ... SIN]
  ?exec  sin? sindoes?  +sin +sindoes
; immediate

: Sinlined?	\ xt -- flag
\ *G Return true if the word defined by xt can be compiled by
\ ** the tokeniser.
  >code-gen @ 0<>
;

: RemoveSin	\ xt --
\ *G Remove tokeniser information from a word. If the
\ ** word has no tokeniser information it is unaffected.
  >code-len dup cell + @ if		\ -- cl ; code-gen field non-zero
    dup ZeroOptData			\ zero fields
  endif
  drop
;

: DoNotSin	\ --
\ *G If the last word with a dictionary header must not be
\ ** tokenised, place *\fo{DoNotSin} after its definition,
\ ** e.g.
\ *C   : foo ... ; DoNotSin
  latest-xt RemoveSin
;

: IMMEDIATE     \ --
\ *G Mark the last defined word as immediate. Immediate words will
\ ** execute whenever encountered regardless of *\fo{STATE}.
\ ** *\fo{IMMEDIATE} also disables tokenising of the last
\ ** defined word. In practice, this is not a performance issue
\ ** as *\fo{IMMEDIATE} words are executed at compile time.
  doNotSin immediate
;

: [opt		\ -- i*x
  postpone [sin  [opt
;

: [-opt		\ -- i*x
  postpone [-sin  [-opt
;

: opt]		\ i*x --
  postpone sin]  opt]
;

previous definitions


\ *****************
\ Shut down actions
\ *****************
\ The token space is allocated from the heap. At shut down it is
\ freed and token information removed.

2variable remove-range

: ((RemoveSinInRange))  \ #thread nfa -- true
  nip  dup remove-range 2@ within	\ if within range
  if  name> RemoveSin  else  drop  then	\ remove SIN info
  true					\ true to continue
;

: (RemoveSinInRange)    \ wid flag -- true
  drop
  ['] ((RemoveSinInRange)) swap WalkWordList
  true
;

also forth definitions
: RemoveSINinRange      \ start end --
\ *G Remove all tokeniser information for definitions within the
\ ** given range.
  remove-range 2!
  ['] (removesininrange) WalkAllWordLists
;

: RemoveAllSins	\ --
\ *G Remove all tokeniser data in the system.
\ ** *\fo{RemoveAllSins} is executed by the exit chain during
\ ** *\fo{BYE}.
   fence @ here RemoveSINinRange
   TokenSpace
   if  TokenSpace free drop  endif
;
' RemoveAllSins AtExit

previous definitions


\ ***********************
\ Test code for debugging
\ ***********************

((
: .BadSin	\ xt --
  cr ." Bad SIN info in " dup >name .name space
  ." CG = " dup >code-gen @ .dword  space
  ." CL = " dup >code-len @ .dword  space
  drop
;

: CheckSin	\ #thread nfa -- true
  nip					\ discard #thread
  name> dup >code-gen @ dup if		\ -- xt cg ; if code-gen field non-zero
    case  over >code-len @		\ inspect code-len field
      ['] c_does_sin of  endof
      ['] c_sin      of  endof
        2 pick .BadSin
    endcase
  endif
  2drop true				\ true to continue
;

: widCheckSin	\ wid flag -- true
  drop
  ['] CheckSin swap WalkWordList
  true
;

also forth definitions
: CheckAllSins	\ --
  ['] widCheckSin WalkAllWordLists
;
previous definitions
))


\ ***********
\ Cleaning up
\ ***********

only forth definitions

true to sin?
true to sindoes?


\ ==========
\ *N Gotchas
\ ==========
\ *P These gotchas are very rare conditions. They usually only
\ ** appear when you write words that affect the semantics
\ ** (meaning) of compilation. You can use *\fo{[-sin ... sin]}
\ ** to drill down to the words that are causing problems.
\ *E [-sin
\ ** : foo ... ;
\ ** : poo ... ;
\ ** sin]


\ -------------------------------
\ *H Immediate and defining words
\ -------------------------------
\ *P The tokeniser hooks into the guts of *\fo{COMPILE,} and
\ ** *\fo{LITERAL}. Compilation performed through these words
\ ** is unaffected by the tokeniser.

\ *P Tokenising of *\fo{IMMEDIATE} words is disabled to reduce
\ ** problems with "guru" code. In nearly all cases, these words
\ ** are only executed at compile time, so there is minimal
\ ** impact on application performance. If an immediate word
\ ** causes compilation using *\fo{COMPILE,} and *\fo{LITERAL},
\ ** the tokeniser will detect this and generate tokens, e.g.
\ *E : z1 postpone dup postpone over ; immediate
\ ** : z2 z1 ;
\ ** ' z2 .tokens
\ ** StartToken
\ ** DUP
\ ** OVER
\ ** End Token

\ *P In the majority of cases the tokeniser handles defining
\ ** words quite adequately. In a few cases, such as defining
\ ** new types of *\fo{xVALUE}, better code generation can be
\ ** obtained by performing some calculation at compile time.
\ ** Such defining words should set a compiler for their children.
\ *P To do this, use *\fo{SET-COMPILER} and *\fo{INTERP>} rather
\ ** than *\fo{DOES>}. *\fo{INTERP>} indicates to the compiler
\ ** that what follows is performed when the child is
\ ** interpreted and that a compiler for the child has been
\ ** defined. The following example is the kernel definition
\ ** of *\fo{VALUE}.
\ *E : value         \ n -- ; ??? -- ???
\ **   create
\ **     ,  ['] valComp, set-compiler
\ **   interp>
\ **     valInterp
\ ** ;
\ *P Note that the chidren of words using *\fo{INTERP>} are
\ ** *\b{not} immediate - they have separate interpretation and
\ ** compilation actions. *\fo{SET-COMPILER ( xt -- )} above sets
\ ** *\fo{valComp,} to be the compiler of the last word *\fo{CREATE}d.
\ ** *\fo{SET-COMPILER} takes the *\i{xt} of the word it is to
\ ** compile so that information can be extracted from the word.

\ *P There are rare occasions on which you may want to add a
\ ** compiler to a non-defining word. Rather than making the
\ ** word immediate and state-smart, which can lead to problems,
\ ** you can add the compiler yourself. This is especially
\ ** desirable when the compiler uses carnal knowledge of VFX Forth
\ ** rather than just *\fo{COMPILE,} and *\fo{LITERAL}, The example
\ ** is taken from the VFX Forth kernel.
\ *E : DO    \ Run: n1|u1 n2|u2 -- ; R: -- loop-sys
\ **   NoInterp  ;
\ ** comp: drop  s_do,  3  ;


\ -------------------------
\ *H Return stack modifiers
\ -------------------------
\ *P In nearly all cases, words that modify the return stack
\ ** will be detected and these words will not be tokenised.
\ ** However, in some cases words containing such words should
\ ** *\b{not} be tokenised because the flow of control has been
\ ** modified. The first example below fails, but the second
\ ** does not. Note that, according to the ANS and Forth200x
\ ** standards, these words are non-standard because they make
\ ** the assumption that, on entry to a word, the top item on
\ ** return stack is the return address. The example below is
\ ** taken from a third-party application ported to VFX Forth.

\ *P This example is correctly detected, but fails because the
\ ** code also requires the word containing *\fo{LIST>} not
\ ** to be tokenised.
\ *E : list> ( thread -- element )
\ **   BEGIN  @ dup  WHILE  dup r@ execute  REPEAT
\ **   drop r> drop ;
\ ** ...
\ ** : .fonts fonts LIST> .font ;
\ *P The example above makes two assumptions, one about the
\ ** return stack in the use of *\fo{R@} and *\fo{R>}, and
\ ** another about how colon definitions begin in *\fo{EXECUTE}.

\ *P The solution is to disable the tokeniser when the word
\ ** is compiled. The containing word is forced to be untokenised.
\ *E : (list>) ( thread -- element )
\ **   BEGIN  @ dup  WHILE  dup r@ execute  REPEAT
\ **   drop r> drop ;
\ ** : list> ( thread -- element )
\ **   postpone (list>) discard-sinline ; immediate
\ ** ...
\ ** : .fonts fonts LIST> .font ;
\ *P If you need to write words such as these, partitioning them
\ ** as above, plus careful use of *\fo{:NONAME} to create the
\ ** second part improves portability and maintainability.


\ ----------
\ *H Using :
\ ----------
\ *P If you build a new compiling word that uses colon, *\fo{:},
\ ** its children can themselves be tokenised. If your new word
\ ** saves and restores data from the return stack indirectly,
\ ** the tokeniser may not detect this, leading to obscure
\ ** runtime or compilation errors. This situation can be avoided
\ ** by adding *\fo{DISCARD-SINLINE} after the use of colon, e.g.
\ *E : MY:	\ --
\ **   :  postpone save-state  discard-sinline
\ ** ;
\ **
\ ** : MY;	\ --
\ **   postpone restore-state  postpone ;
\ ** ;

\ ------------
\ *H Code size
\ ------------
\ *P Some coding styles can lead to excessive expansion of code
\ ** size by the tokeniser. Apart from turning the tokeniser off,
\ ** you can try reducing the size set in the value
\ ** *\fo{SinThreshold}. Note that the relationship between
\ ** the compiled size of a word and its equivalent after
\ ** token expansion in another word is often obscure.


\ ======
\ *> ###
\ ======

