\ inline.fth - Inlining for VFX Forth

((
Copyright (c) 1998, 199, 2001, 2003, 2009
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

Change history
--------------
20090420 SFP009 Count underflow of return stack.
20090415 SFP008 More checks.
20030804 SFP007 Replaced LATEST CTRL>NFA NAME>
20010315 SFP006	Added DISCARD-SINLINE to return stack imbalance code
SJC005  99/11/12        Added UnSafe-RS? to FORTH for use in the source inliner
SFP004	12 Nov 1999	Added protection for : foo ... r> ... >r ... ;
SJC003  99/10/07        Changed to allow ALL code definitions to be inlined
                        (subject to other rules)
SFP002  19 Nov 1998	Converted to PFW/VFX
SFP001  29 Oct 1998     Added ABSURD INLINING
SFP000  28 Sep 1998     First version


Description
-----------
The length of a definition (including the final return) is stored
in the CODE-LEN field of the symbol table.

During compilation the start address of the definition may be found from
the value field of the symbol. At the end of the definition, the length
is calculated and stored in the CODE-LEN field.

When the definition is referenced and compiled by COMPILE-REF, the
CODE-LEN field is compared against the MAX-INLINE variable. If inlining
is permitted, the code is copied from the original to the new location.
Otherwise a CALL is compiled.

Some words contain code that cannot be inlined, e.g defining words.
The variable INLINE-OK? is reset by these words, and the CODE-LEN
field is then left at 0 to indicate that inlining should not be used.
Note that since the CODE-LEN field includes the final return opcode,
the length of the return opcode must be subtracted before inlining
takes place.

The size of the return opcode is defined in the cross compiler's
build file.

Inlining is inhibited by:
  :noname
  does> ;code
  exit
  >r r@ r>  2>r 2r@ 2r>
  COMPILE-REF generating a CALL

CODE <name> definitions may be inlined by placing INLINE after the
END-CODE
  CODE <name> ... END-CODE  INLINE
))


only forth  also h-c definitions

variable inline-ok?     \ -- addr ; true if inlining permitted
  inline-ok? off
variable max-inline     \ -- addr ; holds maximum size of code to be inlined
  max-inline off
variable rs-changed     \ -- addr ; hold change in the return stack
  rs-changed off
variable rs-negative	\ -- addr ; true if RS went negative, i.e return address used	SFP004
  rs-negative off
variable rs-under	\ -- addr ; undershoot on RS in cells.
  rs-under off
variable EarlyExit	\ -- addr ; set if EXIT before semi-colon.
  EarlyExit off

\ These four variables must be set up in the XOPTxxx/XGENxxx files
variable #:entry        \ -- addr ; size of entry code to COLON definition
variable #:exit         \ -- addr ; size of exit code for COLON definition
variable #code-entry    \ -- addr ; size of entry code for CODE definition
variable #code-exit     \ -- addr ; size of exit code for CODE definition


\ ********************************
\ Words used by the cross compiler
\ ********************************

: set-code-len  \ -- ; set the CODE-LEN field of any symbol
  latest if
    latest-xt chere(t) over -		\ length of code ; SFP007
    swap >code-len !			\ store it
  endif
;

: start-inline  \ -- ; tell compiler inlining of word is permitted
  inline-ok? on  rs-changed off
  rs-negative off  rs-under off		\ SFP004, SFP009
  EarlyExit off
;

[-short-branches
: end-inline    \ -- ; sets CODE-LEN field of last COLON definition
  EarlyExit @ 0= if
    \ check for unbalanced RS or RS underflow
    rs-changed @  rs-negative @ or	\ no inlining if imbalanced ; SFP008
    if  discard-sinline  endif		\ SFP006
((
    \ Check for undeflow
    rs-under @ if			\ check RS undeflow, SFP009
\ cr ." Word " latest-xt >name .name
      latest-xt >info dup c@ $0F and	\ -- addr b
\ ." , Old MPE info byte " dup .byte
      rs-under @ 4 lshift $F0 and or	\ merge underflow field
\ ." , Merged to " dup .byte
      swap c!  rs-under off
    endif
))
  endif
  inline-ok? @				\ inlining ok?
  rs-changed @ 0= and			\ and no change in R.S. depth?
  rs-negative @ 0= and			\ and R.S. never went negative ; SFP004
  if  set-code-len  endif               \ set up CODE-LEN field
  inline-ok? off  rs-changed off
  rs-negative off  EarlyExit off
;
short-branches]

: not-inlined   \ --
  inline-ok? off
;

: not-inlined-exit   \ -- ; SFP009
  inline-ok? off  EarlyExit on
;

: inlined       \ --
  inline-ok? on
;

: inline?       \ xt -- flag ; true if word can be inlined
  >code-len @  1 max-inline @ within?
;

: :inline,(t)		\ xt -- ; inline a colon definition
  dup #:entry @ +			\ start of code to be inlined
  swap >code-len @			\ size of code to be inlined
  #:entry @ -  #:exit @ -               \ less entry and exit code
  0 max  bounds
  ?do  i c@c(t) c,c(t)  loop
;

: code-inline,(t)       \ xt -- ; inline a code definition
  dup #code-entry @ +			\ start of code to be inlined
  swap >code-len @			\ size of code to be inlined
  #code-entry @ -  #code-exit @ -       \ less entry and exit code
  0 max  bounds
  ?do  i c@c(t) c,c(t)  loop
;

: code-word?	\ xt -- flag		\ for the moment **********
  drop true                             \ SJC003
;

: inline,(t)	\ xt -- ; inline a definition
  dup code-word?
  if  code-inline,(t)  else  :inline,(t)  endif
;

: ?inline-rs    \ n -- ; no inlining if <n items on return stack
  rs-changed @ - dup 0> if		\ check for RS underflow ; SFP009
\ cr ." RS underflow " dup . .source-line
    rs-under @ max rs-under !		\ save underflow
    rs-negative on			\ mark as undershot
    not-inlined  discard-sinline	\ SFP006
  else
    drop
  endif
;

: +rs-count     \ n -- ; update return stack change
  rs-changed +!  0 ?inline-rs		\ change return stack
;


\ ***********************************
\ User directives to control inlining
\ ***********************************

also forth definitions
\ These constants define the maximum size of a word that can be inlined
    0 constant no
  #12 constant normal
 #255 constant aggressive
#4095 constant absurd                   \ SFP001

: inlining      \ n -- ; set size of inlining
  max-inline !
;

: inlining?     \ -- n ; get size of inlining
  max-inline @
;

: unsafe-rs?    \ -- flag ; return TRUE if return stack depth changes           \ SJC005
                \           need not fail source level inlining
                \           ( IE the local return stack depth is never -ve )
  rs-negative @
;

: inline        \ -- ; mark latest CODE definition for inlining
  latest-xt Code-Word? 0= #-314 ?throw	\ SFP007
  set-code-len
;
previous definitions

((
\ test code
: .symx         \ -- ; display symbol data
  get-token current @ v-find(t) 0= 0 ?error
  $20 - $40 dump
;
))

