\ P4OPT.FTH - Data optimisation for Pentium 4 CPUs

((
Copyright (c) 2003, 2004, 2009, 2010, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20131115 SFP005 Corrected IARRAY.
20121129 SFP004 Added 2VARIABLE.
20101005 SFP003 Added 2VALUE.
20090409 SFP002 Updated for v4.3.
20041021 SFP001 Further tuning for 12% performance boost.
))

decimal

\ ===========
\ *> opcontrl
\ ===========
\ *S Code/Data separation
\ *P From VFX Forth v4.3 onwards, code/data separation is turned on
\ ** by default.

\ =======================
\ *N Problem and solution
\ =======================
\ *P CPUs from the Pentium 3 onwards have serious performance
\ ** problems when data is close to code, leading to
\ ** a wide variation in performance depending on data
\ ** location. Measurements on the random number generator
\ ** in the benchmark suite had a variation of 7:1.

\ *P The file *\i{Sources\Kernel\386Com\OPTIMISE\P4opt.fth} (with
\ ** Professional and Mission versions) contains code for data
\ ** space management for these processors. Results show that
\ ** performance is improved by a factor of 2.3 on *\i{BENCHMRK.FTH}
\ ** and that performance is now independent of location. There
\ ** is no degradation of performance on other CPUs. The code
\ ** generation switches are:
\ *E   +IDATA        \ -- ; enable code/data separation
\ **   -IDATA        \ -- ; dsable code/data separation
\ *P Note that when enabled, phrases such as
\ *C   VARIABLE <name> <size> ALLOT
\ *P will *\b{not} give the expected result. This is discussed
\ ** in more detail below.

\ *P The solution is to separate code and data. When the
\ ** optimisation is enabled, data is held in *\fo{IDATA} chunks
\ ** away from code. There is no change to *\fo{CREATE},
\ ** *\fo{ALLOT}, comma and friends, which still operate on
\ ** normal dictionary areas. The notation is derived from cross
\ ** compiler usage in embedded systems.

\ =====================================
\ *N Defining words and data allocation
\ =====================================
\ *P The following is a conventional definition of
\ ** a character/byte array defined in the dictionary.
\ *E : cCARRAY	\ n -- ; i -- c-addr
\ **   CREATE  ALLOT  DOES> + ;
\ *P The data space reserved by ALLOT is intermingled
\ ** with code, leading to bad performance. The second
\ ** implementation is for best performance with P4 CPUs.
\ ** *\fo{IRESERVE ( n -- c-addr )} reserves an n-byte block
\ ** in the *\fo{IDATA} area and returns its address. The
\ ** children of *\fo{ICARRAY} are made immediate in order
\ ** to emulate the effect of the source inliner on
\ ** children of *\fo{CCARRAY}. The implementation below is
\ ** is illustrative only. State-smart words (considered "evil" by
\ ** some) can be be avoided using *\fo{set-compiler} and
\ ** *\fo{interp>}.
\ *E : icarray	\ n -- ; i -- c-addr
\ **   dup ireserve  dup rot erase      \ reserve IDATA space
\ **   create  immediate                \ children are IMMEDIATE
\ **     ,                              \ address in IDATA
\ **   does>
\ **     @  state @ if                  \ compiling
\ **       postpone literal  postpone +
\ **     else                           \ interpreting
\ **       +
\ **     endif
\ ** ;
\ *P In order to make the array defining word *\fo{CARRAY}
\ ** independent of whether P4 optimisation is enabled
\ ** *\fo{CARRAY} simply selects which version to use.
\ *E : CARRAY	\ n -- ; i -- c-addr
\ **   idata?
\ **   if  icarray  else  ccarray  endif
\ ** ;

\ ==========
\ *N Gotchas
\ ==========
\ *P When *\fo{+IDATA} is in use, standard defining words such
\ ** as *\fo{VARIABLE} and *\fo{VALUE} will reserve space in the
\ ** *\fo{IDATA} areas, but *\fo{ALLOT} still reserves space in
\ ** the dictionary. Consequently code such as:
\ *C   VARIABLE <name> <size> ALLOT
\ *P will break when *\fo{+IDATA} is active. Use:
\ *C   <size> BUFFER: <name>
\ *P for all such allocations.

\ *P Words such as *\fo{>BODY} and *\fo{BODY>} will not work
\ ** correctly on words whose data area is in an IDATA region.

only forth definitions decimal


\ ===========
\ *N Glossary
\ ===========

variable iblock		\ -- addr
\ *G Holds the address of the current IDATA block.
variable iblock#	\ -- addr
\ *G Holds the size of the current IDATA block.
variable idp		\ -- addr
\ *G Holds the current location in the current IDATA block.
variable def-ialign	\ -- addr
\ Holds the default IDATA block alignment, by default 4 kbytes.
  4 kb def-ialign !	\ was 2kb
variable def-igap	\ -- addr
\ *G Holds the minimum code/data gap size, by default 8 kbytes.
  8 kb def-igap !
variable def-iblock#	\ -- addr
\ *G Holds the default IDATA block size, by default 64 kbytes.
  #64 kb def-iblock# !

: bin-align	\ n --
\ *G Force alignment to an N byte boundary where N is a
\ ** power of two. The space stepped over is set to 0.
  dp @
  over 1- +  swap negate and		\ -- dp'
  here over over - erase		\ wipe gap
  dp !
;

((
: align512	\ --
\ \ *G Align the dictionary to a 512 byte boundary.
  #512 bin-align
;

: align1024	\ --
\ \ *G Align the dictionary to a 1024 byte boundary.
  #1024 bin-align
;
))
: alignidef	\ --
\ *G Align the dictionary to the IDATA default boundary.
  def-ialign @ bin-align
;

: inoroom?	\ n -- flag
\ *G Retuns true if there is not enough room in the current
\ ** IDATA block.
  idp @ +  iblock @ iblock# @ + u< invert
;

[-sin
: make-iblock	\ n --
\ *G Make an IDATA block that is at least n bytes long. If n
\ ** is less than the default size in *\fo{DEF-IBLOCK#} the block
\ ** will be the default size.
  def-igap @ allot			\ force code/data gap
  alignidef				\ on default boundary
  here dup iblock ! idp !		\ block starts here
  def-iblock# @ max allot		\ reserve at least default
  alignidef				\ on default boundary
  here iblock @ - iblock# !		\ set block size
;
sin]

: ialign	\ --
\ *G Step the IDATA block pointer to the next 4 byte boundary
  idp @  3 + -4 and  idp !
;

: ialign16	\ --
\ *G Step the IDATA block pointer to the next 16 byte boundary
  idp @  15 + -16 and  idp !
;

[-sin
: ireserve	\ n -- a-addr
\ *G Reserve n bytes in the current IDATA block.
  ialign  dup 8 >
  if  ialign16  endif
  dup inoroom?
  if  dup make-iblock  endif
  idp @ tuck + idp !
;
sin]

0 value idata?		\ -- flag
\ *G Returns true if data is reserved in the IDATA block.

: +idata	\ --
\ *G Force data to be reserved in IDATA blocks.
  true -> idata?
;

: -idata	\ --
\ *G Data is reserved conventionally in the normal dictionary
\ ** space.
  false -> idata?
;

: 2variable	\ -- ; -- addr
\ *G If *\fo{IDATA?} is true data is reserved in an IDATA block,
\ ** otherwise it is reserved in the dictionary.
  idata? if
    2 cells ireserve  dup 2 cells erase  constant
  else
    2variable
  endif
;

: variable	\ -- ; -- addr
\ *G If *\fo{IDATA?} is true data is reserved in an IDATA block,
\ ** otherwise it is reserved in the dictionary.
  idata? if
    cell ireserve  dup off  constant
  else
    variable
  endif
;

: buffer:	\ size -- ; -- addr
\ *G If *\fo{IDATA?} is true data is reserved in an IDATA block,
\ ** otherwise it is reserved in the dictionary.
  idata? if
    dup ireserve  dup rot erase  constant
  else
    buffer:
  endif
;

: ivalCompChild	\ xt --
\ Compiler for child of IVALUE.
  >body @ s_val,			\ SFP002
;

: ivalue	\ n -- ; -- n
\ Creates an IDATA value.
  cell ireserve tuck !
  create
    ,  ['] ivalCompChild set-compiler
  interp>
    @ valInterp				\ SFP002
; doNotSin

: value		\ n -- ; -- n
\ *G If *\fo{IDATA?} is true data is reserved in an IDATA block,
\ ** otherwise it is reserved in the dictionary.
  idata?
  if  ivalue  else  value  endif
;

: i2valComp,	\ xt --
\ Given the xt, compile the run time action for a 2VALUE.
  >body @ 2val,  ;

: i2value	\ n -- ; -- n
\ Creates an IDATA 2value.
  2 cells ireserve dup >r 2! r>
  create
    ,  ['] i2valComp, set-compiler
  interp>
    @ 2valInterp
; doNotSin

: 2value		\ n -- ; -- n
\ *G If *\fo{IDATA?} is true data is reserved in an IDATA block,
\ ** otherwise it is reserved in the dictionary.
  idata?
  if  i2value  else  2value  endif
;

: carrayChildComp	\ xt -- ; compiler for children of array
  >body postpone literal  postpone +
;

: CARRAY	\ n -- ; i -- c-addr
\ *G Creates a byte array. When the child executes, the
\ ** address of the i'th byte in the array is returned.
\ ** The index is zero based.
\ ** If *\fo{IDATA?} is true data is reserved in an IDATA block,
\ ** otherwise it is reserved in the dictionary.
\ *E 10 CARRAY MYCARRAY		\ create 10 byte array
\ **   5 MYCARRAY .		\ display address of element 5
\  CREATE  ALLOT  DOES> ( n -- )  + ;
  create
    allot  ['] carrayChildComp set-compiler
  interp>
    +
; doNotSin

: arrayChildComp	\ xt -- ; compiler for children of array
  >body postpone literal  postpone swap  postpone cells  postpone +
;

: ARRAY		\ n -- ; i -- a-addr
\ *G Creates a cell size array. When the child executes, the
\ ** address of the i'th cell in the array is returned.
\ ** The index is zero based.
\ ** If *\fo{IDATA?} is true data is reserved in an IDATA block,
\ ** otherwise it is reserved in the dictionary.
\ *E 10 CARRAY MYARRAY		\ create 10 byte array
\ **   6 MYARRAY .		\ display address of element 6
\  create  CELLS ALLOT  DOES> ( n -- a ) SWAP CELLS +
  create
    cells allot  ['] arrayChildComp set-compiler
  interp>
    swap cells +
; doNotSin

: icarrayChildComp	\ xt -- ; compiler for children of array
  >body @ postpone literal  postpone +
;

: icarray	\ n -- ; i -- c-addr
\ Creates a byte array.
  dup ireserve  dup rot erase
  create
    ,  ['] icarrayChildComp set-compiler
  interp>
    @ +
; doNotSin

: iarrayChildComp	\ xt -- ; compiler for children of array
  >body @ postpone literal  postpone swap  postpone cells  postpone +
;

: iarray	\ n -- ; i -- c-addr
\ Creates a cell size array.
  cells dup ireserve  dup rot erase	\ SFP005
  create
    ,  ['] iarrayChildComp set-compiler
  interp>
    @ swap cells +
; doNotSin

: CARRAY	\ n -- ; i -- c-addr
\ Creates an array of byte elements.
  idata?
  if  icarray  else  carray  endif
;

: ARRAY		\ n -- ; i -- a-addr
\ Creates an array of cell elements.
  idata?
  if  iarray  else  array  endif
;

true -> idata?		\ turn on by default


\ ======
\ *> ###
\ ======

decimal

