\ openlib.fth - builds an Mach-O file that calls libraries
\ - no linker required

\ Copyright (c) 2009, 2010
\ MicroProcessor Engineering
\ 133 Hill Lane
\ Southampton SO15 5AF
\ England
\
\ tel: +44 (0)23 8063 1441
\ net: mpe@mpeforth.com
\      tech-support@mpeforth.com
\ web: www.mpeforth.com
\
\ From North America, our telephone and fax numbers are:
\   011 44 23 8063 1441
\
\
\ To do
\ =====
\
\ Change history
\ ==============
\ 20110108 SFP003 Corrected OSX/BSD system calls.
\ 20110103 SFP002 Conversion to Forth
\ 20090112 JOR001 first working version

interpreter
: bin-align	\ u --
\ *G Align on a *\i{u} byte boundary, where *\i{u} is a power
\ ** of two.
  here  over 1- +  swap negate and  org  ;
: z16,"		\ -- ; z16," foo bar"
\ *G Lay a zero terminated string in 16 byte slot.
  chere #16 + >r  z",  r> corg  ;
target

$1000 equ _file0		\ memory address of first byte of file
$4000 equ _start		\ file entry point
$0080:0000 equ TextSize		\ ask for 8 megabytes for my interpreter


cdata
here _file0 <> [if]
  cr primary section at wrong address" abort
[then]


\ *************
\ Mach-O header
\ *************
\ The header and command list are in the first segment declared with
\ data.

l: mhdr
   $FEED:FACE ,		\ magic
   7 ,			\ cputype
   3 ,			\ cpusubtype
   2 ,			\ filetype, executable
   10 ,			\ ncmds, number of commands that follow
l: ^sizeofcmds
   ( sizeofcmds ) 0 ,	\ sizeofcmds, sans this header
   $85 ,		\ flags

\ Segments have to be aligned to page size (4096)

\ First page of our program space where we land by referencing a
\ null pointer. Note that no space is actually allocated.
l: firstcmd		\ commands start here

l: pagezero
   1 ,			\ LC_SEGMENT
   0 ,			\ size
   z16," __PAGEZERO"	\ segname
   0 ,			\ vmaddr
   $1000 ,		\ vmsize
   0 ,			\ fileoff
   0 ,			\ filesize
   0 ,			\ maxprot
   0 ,			\ initprot
   0 ,			\ nsects
   0 ,			\ flags
here pagezero - pagezero cell + !


\ Commands and the header occupy part of the first
\ segment which happens to be text in our case.
\ Actual code (start) will go right after, although
\ there's nothing preventing us from wasting the rest
\ of the first text page and starting our code on the
\ next one.

\ This can be handy in the Forth cross-compiler
\ since the headers will likely be generated after
\ the kernel code. We can pre-allocate 3 pages
\ at the beginning of the file and use an offset
\ to make sure the kernel code goes after, starting
\ on a new page boundary.

\ Last but not least, segments and other load commands
\ don't have to go in this order. The first segment
\ can be a text segment.

\ =============================
\ Data segment with one section
\ =============================

l: dataseg
  1 ,			\ LC_SEGMENT
  0 ,			\ size, filled in after section 0
  z16," __DATA"		\ segname
  $1000 ,		\ vmaddr
  $1000 ,		\ vmsize
  0 ,			\ fileoff
  $1000 ,		\ filesize
  7 ,			\ maxprot
  3 ,			\ initprot
  1 ,			\ nsects
  0 ,			\ flags

l: datasec0		\ section 0
  z16," __const"	\ sectname
  z16," __DATA"		\ segname
  $1000 ,		\ addr
l: datasec0_ds
  ( datasize ) 0 ,	\ size, filled in later
  0 ,			\ offset
  0 ,			\ align on 2^0
  0 ,			\ reloff
  0 ,			\ nreloc
  0 ,			\ flags
  0 ,			\ reserved1
  0 ,			\ reserved2
here dataseg - dataseg cell + !


\ ================================
\ Stubs, e.g. dlopen, dlclose, etc. - segment with one section
\ ================================

l: stubseg
  1 ,			\ LC_SEGMENT
  ( _stubs_ ) 0 ,	\ size
  z16," __IMPORT"	\ segname
  $2000 ,		\ vmaddr
  $1000 ,		\ vmsize
  $1000 ,		\ fileoff
  $1000 ,		\ filesize
  7 ,			\ maxprot
  7 ,			\ initprot
  1 ,			\ nsects
  0 ,			\ flags

l: stubs0		\ section 0
  z16," __jump_table"	\ sectname
  z16," __IMPORT"	\ segname
  $2000 ,		\ addr
  20 ,			\ size, 4 5-byte stubs
  $1000 ,		\ offset
  6 ,			\ align on 2^6
  0 ,			\ reloff
  0 ,			\ nreloc
  $0400:0008 ,		\ flags
  0 ,			\ reserved1
  5 ,			\ reserved2, stub size
here stubseg - stubseg cell + !
\ _stubs_ equ $-stubs


\ =======================================
\ Data segment used by the dynamic linker
\ =======================================

l: linkage
  1 ,			\ LC_SEGMENT
  ( _linkage_ ) 0 ,	\ size
  z16," __LINKEDIT"	\ link table
  $3000 ,		\ vmaddr
  $1000 ,		\ vmsize
l: linkage_symoff/len
  ( symbols-$$ ) 0 ,	\ fileoff
  ( _symbols_ ) 0 ,	\ filesize
  7 ,			\ maxprot
  1 ,			\ initprot
  0 ,			\ nsects
  0 ,			\ flags
here linkage -  linkage cell + !
\ _linkage_ equ $-linkage


\ ============
\ Symbol table
\ ============

l: symtab
  2 ,			\ LC_SYMTAB
  ( _symtab_ ) 0 ,	\ size
  ( symbols-$$ ) 0 ,	\ symoff
  4 ,			\ nsyms
  ( strings-$$ ) 0 ,	\ stroff    +16
  ( _strings_ ) 0 ,	\ strsize   +20
here symtab -  symtab cell + !
\ _symtab_ equ $-symtab

\ Additional symbol table info used by the dynamic linker

l: dysymtab
  $0b ,			\ LC_DYSYMTAB
  ( _dysymtab_ ) 0 ,	\ size
  0 ,			\ ilocalsym
  0 ,			\ nlocalsym
  0 ,			\ iextdefsym
  0 ,			\ nextdefsym
  0 ,			\ iundefsym
  4 ,			\ nundefsym
  0 ,			\ tocoff
  0 ,			\ ntoc
  0 ,			\ modtaboff
  0 ,			\ nmodtab
  0 ,			\ extrefsymoff
  0 ,			\ nextrefsyms
l: ^indirectsymoff
  ( indirect-$$ ) 0 ,	\ indirectsymoff
  4 ,			\ nindirectsyms
  0 ,			\ extreloff
  0 ,			\ nextrel
  0 ,			\ locreloff
  0 ,			\ nlocrel
here dysymtab -  dysymtab cell + !
\ _dysymtab_ equ $-dysymtab

\ Our dynamic linker

l: dylinker
  $0e ,			\ LC_LOAD_DYLINKER
  ( _dylinker_ ) 0 ,	\ size
  12 ,			\ nameoff, of the string below
  z", /usr/lib/dyld"
  align
here dylinker -  dylinker cell + !
\ _dylinker_ equ $-dylinker


\ ===============================
\ Initial thread (register) state
\ ===============================

l: thrstate
  5 ,			\ LC_UNIXTHREAD
  ( _thrstate_ ) 0 ,	\ size
  $01 ,			\ i386_THREAD_STATE
  $10 ,			\ i386_THREAD_STATE_COUNT - 16 items
  #10 cells allot&erase
  _start ,		\ 11th cell is EIP
  #5 cells allot&erase
here thrstate -  thrstate cell + !
\ _thrstate_ equ $-thrstate


\ =======================================
\ Dynamic libraries that we will be using
\ =======================================

l: dylib
  $0c ,			\ LC_LOAD_DYLIB
  ( _dylib_ ) 0 ,	\ size
  $18 ,			\ nameoff
  $02 ,			\ timestamp
  $006F:0103 ,		\ currentver
  $0001:0000 ,		\ compatver
  z", /usr/lib/libSystem.B.dylib"
  align
here dylib -  dylib cell + !
\ _dylib_ equ $-dylib


\ =============================
\ Main text segment and section
\ =============================

l: textseg
  1 ,			\ LC_SEGMENT
  ( _text_ ) 0 ,	\ size
  z16," __TEXT"		\ segname
  $4000 ,		\ vmaddr
  TextSize ,		\ vmsize
  $3000 ,		\ fileoff
l: textseg_fs
  ( codesize ) 0 ,	\ filesize
  7 ,			\ maxprot
  7 ,			\ initprot
  1 ,			\ nsects
  0 ,			\ flags

l: textsec0		\ text segment section 0
  z16," __text"		\ sectname
  z16," __TEXT"		\ segname
  _start ,		\ addr
l: textsec0_fs
  ( codesize ) 0 ,	\ size
  _start _file0 - ,	\ offset
  0 ,			\ align on 2^0
  0 ,			\ reloff
  0 ,			\ nreloc
  $8000:0400 ,		\ flags
  0 ,			\ reserved1
  0 ,			\ reserved2
here textseg -  textseg cell + !
\ _text_ equ $-text

here firstcmd -  ^sizeofcmds !
\ sizeofcmds equ $-pagezero


\ ===============
\ Additional data
\ ===============
\ Not aligned to page size so part of the same first
\ virtual memory page as the load commands above.
\ Second page if you count PAGEZERO.

align l: signon	\ sign on string issued by Linux system calls
  z", openlibmo exampleX"		\ X replaced by LF
  $0a here 2- c!
here 1- signon - equ l_signon

\ parameter strings and numbers for printf call to shared library
align l: libname
  z", /usr/lib/libSystem.B.dylib"
align l: printfname
  z", printf"
align l: formatstring
  z", Hello from %s, my number is %dX"
  $0a here 2- c!
align l: myname
  z", Stephen"
align l: mynum
  12345678 ,

here signon -  datasec0_ds !		\ size of data in section
\ here _file0 -  datasec0_ds !		\ size of data in section
\ datasize equ $-signon

  #4096 bin-align			\ step to next 4k page

\ *****
\ Stubs
\ *****

asmcode
l: dlopen
  hlt  hlt  hlt  hlt  hlt
l: dlsym
  hlt  hlt  hlt  hlt  hlt
l: dlclose
  hlt  hlt  hlt  hlt  hlt
l: dlerror
  hlt  hlt  hlt  hlt  hlt
end-code

  #4096 bin-align			\ step to next 4k page


\ *******
\ Symbols
\ *******

l: symbols

l: sym_dlopen
  ( offdlopen ) 0 ,	\ nstrx
  $01 c,		\ type N_EXT
  $00 c,		\ sect
  $0101 w,		\ desc
  dlopen ,		\ value

l: sym_dlsym
  ( offdlsym ) 0 ,	\ nstrx
  $01 c,		\ type N_EXT
  $00 c,		\ sect
  $0101 w,		\ desc
  dlsym ,		\ value

l: sym_dlclose
  ( offdlclose ) 0 ,	\ nstrx
  $01 c,		\ type N_EXT
  $00 c,		\ sect
  $0101 w,		\ desc
  dlclose ,		\ value

l: sym_dlerror
  ( offdlerror ) 0 ,	\ nstrx
  $01 c,		\ type N_EXT
  $00 c,		\ sect
  $0101 w,		\ desc
  dlerror ,		\ value

l: indirect	\ indirect symbol table
  0 ,  1 ,  2 ,  3 ,	\ dlopen, dlsym, dlclose, dlerror

indirect _file0 -  ^indirectsymoff !


\ Symbol String table

l: strings
  $20 c,  0 c,

here strings -  sym_dlopen !	\ set offset
  z", _dlopen"

here strings -  sym_dlsym !	\ set offset
  z", _dlsym"

here strings -  sym_dlclose !	\ set offset
  z", _dlclose"

here strings -  sym_dlerror !	\ set offset
  z", _dlerror"

symbols _file0 -  linkage_symoff/len !		\ offset of symbols
here symbols -    linkage_symoff/len cell+ !	\ length of symbols

symbols _file0 -  symtab 2 cells + !		\ offset of symbols
strings _file0 -  symtab 4 cells + !
here strings -    symtab 5 cells + !

  #4096 bin-align		\ step to next page


\ ********************
\ Start of application
\ ********************
\ Mac x86 ABI requires 16 byte stack alignments.
\ This is required by SSE, apparently, and all
\ Intel Macs in production come with SSE.

((
%macro clib_prolog 1
   mov ebx, esp        ; remember current esp
   and esp, 0xFFFFFFF0 ; align to next 16 byte boundary (could be zero offset!)
   sub esp, 12         ; skip ahead 12 so we can store original esp
   push ebx            ; store esp (16 bytes aligned again)
   sub esp, %1         ; pad for arguments (make conditional?)
%endmacro

; arg must match most recent call to clib_prolog

%macro clib_epilog 1
   add esp, %1         ; remove arg pad
   pop ebx             ; get original esp
   mov esp, ebx        ; restore
%endmacro
))


\ By fixing the entry point we can make some parts of an interpreter
\ simpler.
proc _entry
   jmp     short $ #10 +		\ 2 byte jump, 2 cells
end-code
_entry _start <> [if]
  cr ." _ENTRY <> _START" abort
[then]


\ storage. N.B. proves write permission in code segment
l: plib		\ holds library handle
   -1 ,
l: pprintf	\ holds address of printf
   -1 ,

\ Extracts from the C header files
\ void *dlopen(const char *filename, int flag);
\ void *dlsym(void *handle, const char*symbol);
\ int dlclose(void *handle);
\ char *dlerror(void);

\ #define RTLD_LAZY 0x00001 /* Lazy function call binding.  */
\ #define RTLD_NOW  0x00002 /* Immediate function call binding.  */
\ #define RTLD_BINDING_MASK   0x3 /* Mask of binding time value.  */
\ #define RTLD_NOLOAD 0x00004 /* Do not load the object.  */
\ #define RTLD_GLOBAL 0x00100

asmcode
l: realstart
\ sign on
   mov     ecx, # signon
   mov     edx, # l_signon
   call    primtype
\   jmp     getout

\ Open the library
\ start prolog
   mov ebx, esp			\ remember current esp
   and esp, # $FFFFFFF0		\ align to next 16 byte boundary (could be zero offset!)
   sub esp, # #12		\ skip ahead 12 so we can store original esp
   push ebx			\ store esp (16 bytes aligned again)
   sub esp, # 8			\ pad for arguments (make conditional?)
\  clib_prolog 8
   push    # 1			\ RTLD_LAZY
   push    # libname		\ "libc.so.6"
   call    dlopen		\ open library
\ start epilog
   add esp, # #16		\ remove arg pad
   pop ebx			\ get original esp
   mov esp, ebx			\ restore
\  clib_epilog 16
   mov     plib [] , eax	\ EAX is library handle or error code
\ mov     ebx, eax		\ return value in EAX
\ call    showhex		\ display result
   mov     ebx, # 1		\ return code
   cmp     eax, # 0		\ test return value
   je      long getout

\ Find printf
\ start prolog
   mov ebx, esp			\ remember current esp
   and esp, # $FFFF:FFF0	\ align to next 16 byte boundary (could be zero offset!)
   sub esp, # #12		\ skip ahead 12 so we can store original esp
   push ebx			\ store esp (16 bytes aligned again)
   sub esp, # 8			\ pad for arguments (make conditional?)
\   clib_prolog 8
   push    # printfname		\ string "printf"
   push    dword [] plib	\ library handle
   call    dlsym		\ find function address
\ start epilog
   add esp, # #16		\ remove arg pad
   pop ebx			\ get original esp
   mov esp, ebx			\ restore
\   clib_epilog 16
   mov     pprintf [] , eax	\ EAX = function address or error code
\ mov     ebx, eax
\ call    showhex		\ display result
   mov     ebx, # 2		\ return code
   cmp     eax, # 0		\ test return value
   je      ( long ) getout
\ jmp     getout
end-code asmcode

\ call printf
\ start prolog
   mov ebx, esp			\ remember current esp
   and esp, # $FFFFFFF0		\ align to next 16 byte boundary (could be zero offset!)
   sub esp, # #12		\ skip ahead 12 so we can store original esp
   push ebx			\ store esp (16 bytes aligned again)
   sub esp, # 4			\ pad for arguments (make conditional?)
\   clib_prolog 4
   push    dword [] mynum	\ second paramter to format string
   push    # myname		\ first parameter to format string
   push    # formatstring	\ format string itself
   call    [] pprintf		\ print and expand format string
\ start epilog
   add esp, # #16		\ remove arg pad
   pop ebx			\ get original esp
   mov esp, ebx			\ restore
\   clib_epilog 16
\ mov     ebx, eax		\ number of characters displayed
\ call    showhex		\ display result
end-code asmcode

\ close library
\ start prolog
   mov ebx, esp			\ remember current esp
   and esp, # $FFFF:FFF0	\ align to next 16 byte boundary (could be zero offset!)
   sub esp, # #12		\ skip ahead 12 so we can store original esp
   push ebx			\ store esp (16 bytes aligned again)
   sub esp, # #12		\ pad for arguments (make conditional?)
\   clib_prolog 12
   push    dword [] plib	\ library handle
   call    dlclose
\ start epilog
   add esp, # #16		\ remove arg pad
   pop ebx			\ get original esp
   mov esp, ebx			\ restore
\   clib_epilog 16
\ mov     ebx, eax
\ call    showhex		\ display result
   mov     ebx, # 3		\ error return code
   cmp     eax, # 0		\ test result
   jne     getout
end-code asmcode

\ back to OSX
   mov     ebx, # 42		\ success return code
l: getout
   push    ebx			\ exit code
   sub     esp, # 4		\ extra space
   mov     eax, # 1		\ exit function number
   int     $80			\ Unix x86 system call

\ INT $080 type routine, EBX=handle ECX=caddr EDX=len
l: primtype
  push    edx			\ len
  push    ecx			\ addr
  push    # 1			\ STDOUT
  sub     esp, # 4		\ extra space needed!
  mov     eax, # 4		\ write call
  int     $80
  add     esp, # #16
  ret

\ emit single character, EAX=char
l: primemit
   push    eax
   mov     ecx, esp
   mov     edx, # 1
   call    primtype
   pop     eax
   ret

\ Display EBX as 8 digit hex number
l: showhex
   push    eax
   push    ebx

   mov     ecx, # 8
l: shx1
   push    ecx
   push    ebx
   mov     eax, ebx
   shr     eax, # 28
   cmp     eax, # 9
   jle     shx2
   add     eax, # 7
l: shx2
   add     eax, # $30
   call    primemit
   pop     ebx
   pop     ecx
   shl     ebx, # 4
   sub     ecx, # 1
   jne     shx1
   mov     eax, # $20
   call    primemit

   pop     ebx
   pop     eax
   ret
end-code

align
here _entry -
  dup textseg_fs !
  dup textsec0_fs !
drop
\ codesize equ $-_start
