\ BUILD.FTH - build numbering system

((
Copyright (c) 2004, 2008, 2010, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20200104 SFP005 Updated CREATE-BUILD to delete existing file.
                This helps with Windows network drives.
20100404 SFP004 Major unmarked overhaul.
20080506 SFP003 Added BUILD$, and time and date facilities.
20080502 SFP002 Added BUIDFILE for cross compiler compatibilty.
20040616 SFP001 Fixed MAKE-$, removed redundant $LENGTH.
))

decimal

\ ========
\ *> tools
\ *S Automatic build numbering
\ ========
\ *P The build numbering system allows you to generate a string
\ ** in the system which can be used for displaying version
\ ** information.

\ *P The system relies on a file (normally called *\i{BUILD.NO})
\ ** which holds the complete build version string. The string
\ ** can consist of any characters, e.g "Version 1.00.0034". The
\ ** contents of the file can be placed as a counted string in
\ ** the dictionary by *\fo{BUILD$,}. After successful
\ ** compilation of your application, *\fo{UPDATE-BUILD} will
\ ** update the build number file by treating *\b{all} the digits
\ ** in the build string as a single number to be incremented.			)


\ ==============
\ Internal words
\ ==============
\ The following words are private to the build system, and can
\  be found in the *\fo{System} vocabulary.

also system definitions

variable ^build-buffer	\ -- addr
\ Holds pointer to the application's counted version string.
\ variable build.no-size
\ variable $length
variable carry	\ -- addr
\ Holds carry flag when updating
#256 buffer: tmp-build-buffer	\ --
\ Scratch buffer for updating
#256 buffer: $Buildfile		\ -- addr
\ Version file name
  s" build.no" $Buildfile place
variable hBuildfile	\ -- addr
\ Holds build file handle.

: Create-Build  \ --
\ Create a new build number file and store the file handle
\ in *\fo{hBuildfile}.
  $Buildfile count fileExist? if	\ SFP005
    $Buildfile count delete-file drop
  then
  $Buildfile count r/w Create-File swap hBuildfile !
  abort" Can't create build file"
;

: Open-Build    \ --
\ Open an existing build number file and store the file handle
\ in *\fo{hBuildfile}.
  $Buildfile count r/w Open-File swap hBuildfile !
  abort" Can't open build file"
;

: Close-Build   \ --
\ Close the open build file.
  hBuildfile @ Close-File drop  ;

: Write-Build   \ caddr len --
\ Write a buffer into the open build file.
  hBuildfile @ write-file abort" Can't write build file"  ;

: Read-Build    \ Buffer -- #bytes
\ Read contents of the open build file into a memory buffer.
  4096 hBuildfile @ Read-File abort" Can't read build file"  ;

: Load-Build	\ addr -- len
\ Open a build file, read its contents and close it.
  open-build  read-build  close-build  ;

: Digit?        \ char -- t/f
\ Check that a character is a decimal digit.
  [char] 0 [char] 9 within?  ;

: ?Increment    \ addr --
\ Increment digit at ADDR as appropriate and store a carry flag.
  dup c@ carry @ +              \ add carry
  dup [char] 9 > if             \ step to next column?
    drop [char] 0 swap c!       \   wrap around to ASCII 0 and set digit
    1 carry !                   \   set carry for next digit
  else
    swap c!                     \   set this new ASCII digit
    0 carry !                   \   no carry for next digit
  endif
;

: ?Incr-Char    \ addr --
\ Increment char at ADDR if appropriate.
  dup c@ digit?                 \ is char a digit
  if  ?increment  else  drop  endif
;

: Build++	\ caddr len --
\ Increment build no. stored at buffer.
  1 Carry !                     \ bump build.no by 1
  Begin
    dup 0>                      \ still chars to do
   While
    1- 2dup + ?incr-char	\ increment char if necessary
  Repeat
  2drop
;

: White-Space?  \ char -- t/f
\ TRUE if char is a white space character.
  dup bl =  over 9 = OR		\   space or a tab
  over #13 = or  swap #10 = OR	\   cr or lf
;

: Make-$        \ addr --
\ Format counted string in buffer.
  0  1 2 pick c@ 1 max ?DO		\ -- caddr 0
    over i + c@ white-space? 0=
    IF  drop i  LEAVE  ENDIF		\ replace 0 with length
  -1 +LOOP
  swap c!
;

\ SFP003...
create Months	\ -- addr
  ", JanFebMarAprMayJunJulAugSepOctNovDec"

: Mon$		\ n -- caddr len ; n=1..12
  ( 1 max #12 min ) 1- 3 * Months + 1+ 3
;

: (date)	\ -- dow dd mm yyyy
  os_getlocaltime			\ -- secs min hrs day mon year dow
  swap >r swap >r swap >r
  nip nip nip
  r> r> r>
;

: (time)	\ -- hh mm ss
  os_getlocaltime 4drop  -rot swap  ;

: GetDate$	\ buff --
\ Get the current date as a counted string in a buffer.
  base @ >r  decimal
  0 over c!  >r				\ set empty string
  (date)				\ -- dow dd mm yyyy ; R: -- base buff
  rot (u.) r@ append			\ dd
  bl r@ addchar
  swap Mon$ r@ append			\ mm
  bl r@ addchar
  (u.) r> append			\ yyyy
  drop
  r> base !
;

: (2.r)		\ u -- caddr len
  0 <# # # #>
;

: GetTime$	\ buff --
\ Get time as counted string in buffer.
  base @ >r  decimal
  0 over c!  >r				\ set empty string
  (time)				\ -- hh mm ss ; R: -- base buff
  rot (2.r) r@ append			\ hh
  [char] : r@ addchar
  swap (2.r) r@ append			\ mm
  [char] : r@ addchar
  (2.r) r> append			\ ss
  r> base !
;
\ ...SFP003

previous definitions

\ ============
\ Public words
\ ============

also system

: Make-Build    \ buffer --
\ *G Read the contents of the build number file and place as a counted
\ ** string in the application defined buffer for later use.
  dup ^build-buffer !           \ set pointer to application buffer
  dup 1+ Load-Build		\ read in version string
  #255 min over c!		\ set length
  make-$                        \ format string
;

\ SFP003...
: Build$,	\ --
\ *G Read the contents of the build number file and place as a
\ ** counted string at *\fo{HERE}. *\fo{ALLOT} the required space.
  here dup Make-Build  c@ 1+ allot  ;

: Date$,	\ --
\ *G Compile date as counted string.
  pad GetDate$  pad count $,  ;

: Time$,	\ --
\ *G Compile time as counted string
  pad GetTime$  pad count $,  ;

: DateTime$,	\ --
\ *G Compile date and time as counted string
  { | temp[ #256 ] -- }
  0 temp[ c!
  pad GetDate$  pad count temp[ append
  s" , " temp[ append
  pad GetTime$  pad count temp[ append
  temp[ count $,
;
\ ...SFP003

: Set-BuildFile	\ c-addr u --
\ *G Set the build number file.
  #255 min  $Buildfile place  ;

: BuildFile	\ -- ; Buildfile <filename>
\ *G Use GetPathSpec to parse a filename from the input stream,
\ ** and make it the current build number file.
  GetPathSpec Set-BuildFile  ;

: Update-Build  \ --
\ *G Update the contents of the build number file ready for the next
\ ** build.
  ^build-buffer @ count tmp-build-buffer place
  create-build
  tmp-build-buffer count
  2dup Build++  write-build
  close-build
;

previous

\ *P The following example, defines which file to use, loads
\ ** the text into a buffer, and finally updates the build text.
\ ** By placing *\fo{Update-Build} last in your load file, your
\ ** build number file will only be updated for each successful
\ ** build.
\ *E s" MyBuild.no" Set-Buildfile    \ set file to use
\ ** #256 buffer: MyVersion$  \ -- caddr
\ **   MyVersion$ make-build         \ load version string
\ ** ...
\ ** update-build                    \ put this last in load file

\ ======
\ *> ###
\ ======

