\ Generic.fth - Generic I/O API interface.

((
Copyright (c) 1998-2004
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20061207 MPE002 Added ReadEx function 21
20041216 MPE001 Added more functions for device setting.
))

decimal

\ ========
\ *! genio
\ *T Generic IO
\ ========
( *P Generic IO is the name given to VFX Forth's entire input/output )
( ** architecture. This system allows for a "device-driver" to be       )
( ** written to a standard format such that the drivers are all         )
( ** interchangeable within the Forth System. As noted later you will   )
\ ** see that all standard Forth I/O words (such as *\fo{EMIT}) are passed
\ ** through Generic I/O.                                                )

( *P Under VFX each thread has it's own current input and output stream )
( ** and can be accessed via the standard Forth IO words and a general  )
( ** purpose wordset which acts upon current thread devices. See Later)
( ** In addition Generic IO also supports a wordset which can use a     )
( ** nominated device directly. This second wordset follows the same    )
( ** naming convention as the current-thread wordset.                   )

\ ===========================
\ *S Format of a GENIO Driver
\ ===========================
( *P An instance of a Generic Driver is described by the following      )
( ** structure, the address of such a structure is called a SID         )
\ ** (structure-identifier).
\ *E   CELL     Device Handle (interpretation depends on device),
\ **   CELL     Pointer to function Table (see below),
\ **   ????     Device Private Data.
( *P The function table is a list of execution tokens for words to      )
( ** perform various standard actions. Each action word will receive    )
( ** the SID to operate upon at the top of the data stack.              )
( *R @page                                                              )
( *P To be a "Generic Device" the vector table must hold valid entries  )
( ** for:                                                               )
( *E Index      Name            Description                             )
( **   0        OPEN            Open/Initialise a device.               )
( **   1        CLOSE           Close a device.                         )
( **   2        READ            Read to a block of memory.              )
( **   3        WRITE           Write a specified block of memory.      )
( **   4        KEY             Perform an action equivalent to Forths  )
( **                            KEY definition, i.e. a blocking character)
( **                            read.                                  )
( **   5        KEY?            Perform an action equivalent to Forths  )
( **                            KEY?, i.e. any-input-pending?         )
( **   6        EKEY            Supports EKEY )
( **   7        EKEY?           Supports EKEY? )
( **   8        ACCEPT          Added to support FORTH definition of the)
( **                            same name. Read a character stream into )
( **                            a memory buffer.                        )
( **   9        EMIT            Write a single character to a stream.   )
( **  10        EMIT?           Check that EMIT can work.               )
( **  11        TYPE            As with Forths TYPE. Write a string of  )
( **                            characters to a device.                 )
( **  12        CR              Perform nearest equivalent action of    )
( **                            "carriage return" on the device.        )
( **  13        LR              As with CR except for "line-feed"       )
( **  14        FF              As with CR except for "form-feed"       )
( **  15        BS              As with CR except for "backspace"       )
( **  16        BELL            Where applicable to the device emit an  )
( **                            audible beep.                           )
( **  17        SETPOS          Set current position. May reflect screen)
( **                            cursor/file position etc.               )
( **  18        GETPOS          Read current position.                  )
( **  19        IOCTL           Perform a special function. Each device )
( **                            may or may not support various IOCTL    )
( **                            codes. The currently assigned function  )
( **                            codes used by MPE are documented later. )
\ **  20        FLUSHOP         Flush any pending output for device.
\ **  21        RFU/READEX      As READ with additional return of count.
\ **                            Not available on all devices.
\ *P Devices that require additional functions may add these at the
\ ** end of the table. If additional functions are added the first
\ ** three must be as below. It is valid for these to perform no
\ ** action except to return a zero ior.
\ *E  22        initialise device    addr len sid -- ior
\ **  23        terminate device     sid -- ior
\ **  24        configure device     sid -- ior ; produces a dialog


\ ===============================
\ *S Current Thread Device Access
\ ===============================
( *P The following definitions act upon the nominated input or output   )
( ** stream for the calling thread. Definitions declared with *\fo{IPFUNC} act)
( ** upon the current input stream and definitions declared with *\fo{OPFUNC} )
( ** act upon the current output stream.                                )


struct gen-sid  \ -- len
\ *G Define the generic I/O structure known as a SID. This
\ ** structure does not include any private data.
\ *[
  cell  field   gen-handle
  cell  field   gen-vector
  0     field gen-private
end-struct
\ *]


\ *******************************
\ Constants that define functions
\ *******************************

decimal
\ -- n
 0 equ  OpenFnid
 1 equ  CloseFnid
 2 equ  ReadFnid
 3 equ  WriteFnid
 4 equ  KeyFnid
 5 equ  Key?Fnid
 6 equ  EkeyFnid
 7 equ  EKey?Fnid
 8 equ  AcceptFnid
 9 equ  EmitFnid
10 equ  Emit?Fnid
11 equ  TypeFnid
12 equ  CRFnid
13 equ  LFFnid
14 equ  FFFnid
15 equ  BSFnid
16 equ  BellFnid
17 equ  SetposFnid
18 equ  GetposFnid
19 equ  IoctlFnid
20 equ  FlushOPFnid
21 equ  RFUFnid		21 equ  ReadExFnid
22 equ  InitFnid
23 equ  TermFnid
24 equ  ConfigFnid


\ *************************
\ The main despatch routine
\ *************************

: DoIO          \ i*x sid fnid -- j*x ; Perform I/O routine n for given       \ SJC008
  cells                                 \ convert fnid to offset              \ SJC008
  over cell+ @ +                        \ add function table base address     \ SJC008
  @ execute                             \ get function and do it              \ SJC008
;


\ ***********************************************************************
\ The functions that can be assigned to Forth deferred words such as EMIT
\ ***********************************************************************

: GIOFunc       \ fnid -- ; i*x -- j*y
  create
    ,
  does>
    @ DoIo
;

: IPFunc        \ fnid -- ; i*x -- j*y
  create
    ,
  does>
    ip-handle @  swap @ DoIo
;

: OPFunc        \ fnid -- ; i*x -- j*y
  create
    ,
  does>
    op-handle @
    swap @ DoIo
;

OpenFnid    OPFunc open-gen     \ addr len attribs -- handle/sid ior
( *G Perform the Generic IO "OPEN" action for current output device.    )

CloseFnid   OPFunc close-gen    \ -- ior
( *G Perform the Generic IO "CLOSE" action for current output device.   )

ReadFnid    IPFunc read-gen     \ addr len -- ior
( *G Perform the Generic IO "READ" action for current input device.     )

WriteFnid   OPFunc write-gen    \ addr len -- ior
( *G Perform the Generic IO "WRITE" action for current output device.   )

KeyFnid     IPFunc key-gen      \ -- char
( *G Perform the Generic IO "KEY" action for current input device.      )
( ** This operation is identical to the Forth word *\fo{KEY}.           )

Key?Fnid    IPFunc key?-gen     \ -- flag
( *G Perform the Generic IO "KEY?" action for current input device.     )
\ ** This operation is identical to the Forth word *\fo{KEY?}.

EKeyFnid    IPFunc ekey-gen     \ -- echar
( *G Perform the Generic IO "EKEY" action for current input device.     )
\ ** This operation is identical to the Forth word *\fo{EKEY}.

EKey?Fnid   IPFunc ekey?-gen    \ -- flag
( *G Perform the Generic IO "EKEY?" action for current input device.    )
\ ** This operation is identical to the Forth word *\fo{EKEY?}.

AcceptFnid  IPFUnc accept-gen   \ addr len -- len'
( *G Perform the Generic IO "ACCEPT" action for current input device.   )
\ ** This operation is identical to the Forth word *\fo{ACCEPT}.

EmitFnid    OPFunc emit-gen     \ char --
( *G Perform the Generic IO "EMIT" action for current output device.    )
\ ** This operation is identical to the Forth word *\fo{EMIT}.

Emit?Fnid   OPFunc emit?-gen    \ -- flag
( *G Perform the Generic IO "EMIT?" action for current output device.   )
\ ** This operation is identical to the Forth word *\fo{EMIT?}.

TypeFnid    OPFunc type-gen     \ addr len --
( *G Perform the Generic IO "TYPE" action for current output device.    )
\ ** This operation is identical to the Forth word *\fo{TYPE}.

CRFnid      OPFunc cr-gen       \ --
( *G Perform the Generic IO "CR" action for current output device.      )
\ ** This operation is identical to the Forth word *\fo{CR}.

LFFnid      OPFunc lf-gen       \ --
( *G Perform the Generic IO "LF" action for current output device.      )

FFFnid      OPFunc ff-gen       \ --
( *G Perform the Generic IO "FF" action for current output device.      )
\ ** This operation is identical to the Forth word *\fo{PAGE}.

BSFnid      OPFunc bs-gen       \ --
( *G Perform the Generic IO "BS" action for current output device.      )

BellFnid    OPFunc bell-gen     \ --
( *G Perform the Generic IO "BELL" action for current output device.    )

SetposFnid  OPFunc setpos-gen   \ d mode -- ior ; x y mode -- ior
( *G Perform the Generic IO "SETPOS" action for current output device.  )

GetposFnid  OPFunc getpos-gen   \ mode -- d ior ; mode -- x y ior
( *G Perform the Generic IO "GETPOS" action for current output device.  )

IoctlFnid   OPFunc ioctl-gen    \ addr len fn# -- ior
( *G Perform the Generic IO "IOCTL" action for current output device.   )

FlushOPFnid OPFunc FlushOP-gen  \ -- ior
( *G Perform the Generic IO "FLUSH" action for current output device.   )

RFUFnid     OPFunc RFU-gen      \ --
ReadExFnid  IPFunc ReadEx-gen	\ addr len -- #read ior
\ *G Perform the Generic IO "READEX" action for current input device.
InitFnid    OPFunc init-gen     \ addr len -- ior
TermFnid    OPFunc term-gen     \ -- ior
ConfigFnid  OPFunc config-gen   \ -- ior

: gotoxy-gen                    \ x y --
  -1 setpos-gen throw  ;

: getxy-gen                     \ -- x y
  -1 getpos-gen throw  ;


\ =================================
\ *S IO based on a Nominated Device
\ =================================
( *P Generic IO allows you to perform an action on any device without   )
( ** changing the thread's current Input or Output Channel. All the      )
( ** definitions listed as *\fo{xxx-GEN} also have an equivalent definition   )
( ** called *\fo{xxx-GIO} which has as top of stack an additional parameter   )
( ** which is the *\i{SID} of the nominated device.                     )

\ *[
OpenFnid   GIOFunc open-gio     \ addr len attribs sid -- handle/sid ior
CloseFnid  GIOFunc close-gio    \ sid -- ior
ReadFnid   GIOFunc read-gio     \ addr len sid -- ior
WriteFnid  GIOFunc write-gio    \ addr len sid -- ior
KeyFnid    GIOFunc key-gio      \ sid -- char
Key?Fnid   GIOFunc key?-gio     \ sid -- flag
EKeyFnid   GIOFunc ekey-gio     \ sid -- echar
EKey?Fnid  GIOFunc ekey?-gio    \ sid -- flag
AcceptFnid GIOFunc accept-gio   \ addr len sid -- len'
EmitFnid   GIOFunc emit-gio     \ char sid --
Emit?Fnid  GIOFunc emit?-gio    \ -- flag
TypeFnid   GIOFunc type-gio     \ addr len sid --
CRFnid     GIOFunc cr-gio       \ sid --
LFFnid     GIOFunc lf-gio       \ sid --
FFFnid     GIOFunc ff-gio       \ sid --
BSFnid     GIOFunc bs-gio       \ sid --
BellFnid   GIOFunc bell-gio     \ sid --
SetposFnid GIOFunc setpos-gio   \ d mode sid -- ior ; x y mode sid -- ior
GetposFnid GIOFunc getpos-gio   \ mode sid -- d ior ; mode -- x y ior
IoctlFnid  GIOFunc ioctl-gio    \ addr len fn# sid -- ior
FlushOPFnid GIOFunc FlushOP-gio \ sid -- ior
ReadExFnid GIOFunc ReadEx-gio   \ addr len sid -- #read ior
RFUFnid    GIOFunc RFU-gio      \ sid --
InitFnid   GIOFunc init-gio     \ addr len sid -- ior
TermFnid   GIOFunc term-gio     \ sid -- ior
ConfigFnid GIOFunc config-gio   \ sid -- ior
\ *]


\ =======================================
\ *S Standard Forth words using GenericIO
\ =======================================
( *P The following standard Forth definitions are already vectored      )
( ** through their generic IO equivalents. The SID device handle used   )
( ** comes from the *\fo{USER} variables *\fo{OP-HANDLE} and *\fo{IP-HANDLE}. )

\ *E ACCEPT KEY KEY? EKEY EKEY? EMIT EMIT? TYPE CR
\ *P Also affected are any I/O words within the ANS Core wordset
\ ** that use these primitives, such as:
\ *E . .S SEE U. ) $. F. DUMP EXPECT QUERY

assign accept-gen to-do accept
assign key-gen    to-do key
assign key?-gen   to-do key?
assign ekey-gen   to-do ekey
assign ekey?-gen  to-do ekey?
assign emit-gen   to-do emit
assign emit?-gen  to-do emit?
assign cr-gen     to-do cr				\ SJC004
assign type-gen   to-do type

: generic-drivers
  assign accept-gen to-do accept
  assign key-gen    to-do key
  assign key?-gen   to-do key?
  assign ekey-gen   to-do ekey
  assign ekey?-gen  to-do ekey?
  assign emit-gen   to-do emit
  assign emit?-gen  to-do emit?
  assign cr-gen     to-do cr				\ SJC004
  assign type-gen   to-do type
;

\ ==========================
\ *S Miscellaneous I/O Words
\ ==========================
( *P The following IO words are defined along with Generic IO and will  )
( ** use the standard Generic IO vectors.                               )

: page		\ --
( *G Performs a FORM-FEED operation. The effect of this differs from    )
( ** device to device. Binary devices will simply output the #12        )
( ** character, screen devices will clear the screen and printer devices)
( ** will move on to the next page.                                     )
  ff-gen
;

: cls           \ --
( *G An alias for *\fo{PAGE} which reads more clearly when using a screen     )
( ** based device.                                                      )
  ff-gen
;

: at-xy         \ x y --
\ *G The ANS cursor relocation definition. Attempt to move the cursor
\ ** to relative position X Y. The actual translation of this varies
\ ** from device to device since it is implemented with the *\fo{SETPOS}
\ ** generic IO vector.
  0 setpos-gen drop
;

\ MPE001...
: SetIO	        \ sid --
\ *G Set the given device as the current I/O device.
  dup ip-handle !  op-handle !
;

: [IO		\ -- ; R: -- ip-handle op-handle
\ *G Used inside a colon definition *\b{only} to preserve the
\ ** the current I/O devices before switching them temporarily.
\ ** Usually used in the form:
\ *E [io  SomeDev SetIO
\ **   ...
\ ** io]
  ?comp
  postpone ip-handle  postpone @  postpone >r
  postpone op-handle  postpone @  postpone >r
; immediate

: IO]		\ -- ; R: ip-handle op-handle --
\ *G Used inside a colon definition *\b{only} to restore the
\ ** the current I/O devices after switching them temporarily.
\ ** See *\fo{[IO} for more details.
  ?comp
  postpone r>  postpone op-handle  postpone !
  postpone r>  postpone ip-handle  postpone !
; immediate
\ ...MPE001

[undefined] szSID [if]
create szSID	\ -- addr ; used as a property string
\ *G Within a winproc controlling a device, it is often useful
\ ** to be able to reference the SID of the device. This is best
\ ** done by using the *\b{SetProp} Windows call to set a property for the
\ ** Window - see the *\i{STUDIO} directories for examples. MPE code uses the
\ ** property string "SID" for this, and *\fo{szSID} is the address of the
\ ** zero terminated property string.
  z", SID"
[then]


\ ===================
\ *S Supplied Devices
\ ===================
( *P The following Generic IO device implementations can all be found   )
( ** in the supplied source library folder *\i{Lib/Genio}.              )

\ ======
\ *> ###
\ ======

