\ MPE Character Queue Support

((
Released for use with the MPE Forth Cross Compiler by:

Copyright (c) 2001, 2002, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com

To do
=====


Change history
==============
20120430 SFP003 Added INIT-DCQUEUE.
20021023 SFP002 Ported to VFX Forth
20021009 SFP001 Refactored to enhance interrupt locking in
		interrupt driven drivers.
))


\ ==========
\ *! cqueues
\ *T Character Queues
\ ==========
\ *P The file *\i{Lib\CQueues.fth} provides circular character
\ ** (byte) queues. If the equate *\fo{TASKING?} is non-zero,
\ ** the blocking routines will use *\fo{PAUSE}. Interrupts are
\ ** disabled for the queue empty/full checks.

\ *P This code has been ported from the cross compiler standard
\ ** target library. In the cross compiler, initialised and
\ ** unitialised data are is held in different areas of memory.
\ ** In VFX Forth, there is no distinction between *\fo{CDATA},
\ ** *\fo{IDATA} and *\fo{UDATA} areas. The size of a queue must
\ ** be a power of 2. The equate *\fo{TASKING?} does not exist
\ ** for VFX Forth. Interrupt locking does not exist in VFX
\ ** Forth for Windows/Mac/Linux, and if required must be
\ ** implemented using critical sections or mutexes.


\ ***************
\ *S Misc support
\ ***************

[undefined] [i [if]
: [i		\ -- ; R: -- intsys
\ *G Prevent interrupts, saving the current interrupt state on the return stack.
\ ** For most modern operating systems, this word is an *\fo{IMMEDIATE NOOP}.
; immediate

: i]		\ -- ; R: intsys --
\ *G Restore the interrupt state previus saved by *\fo{[I}.
\ ** For most modern operating systems, this word is an *\fo{IMMEDIATE NOOP}.
; immediate
[then]


\ ***********************
\ *S Queue data structure
\ ***********************

struct /cqueue	\ -- size ; does not include data buffer
\ *G Circular queue data structure.
\ *[
  int >qhead		\ Offset of head
  int >qtail		\ Offset of tail
  int >qchars		\ Number of characters in the queue
  int >qmask		\ Mask to apply to pointers (size-1)
  int >qbuffer		\ Base address of character buffer
end-struct
\ *]

: cqueue:	\ size -- ; -- cqueue ; size CQUEUE: <name>
\ *G An interpreter definition to build a character queue of the
\ ** specified size.
\ ** Executing <name> returns the address of the queue data structure.
\ ** N.B. The size of a queue must be a power of two, e.g. 32, 64 ...
  create
    here  /cqueue allot		\ -- size cqueue
    dup >qhead off
    dup >qtail off
    dup >qchars off
    over 1- over >qmask !
    here swap >qbuffer !  allot
;

: init-cqueue	\ cqueue -- ; initialise queue
\ *G Initialise the specified queue created by *\fo{CQUEUE:}.
  dup >qhead off
  dup >qtail off
  >qchars off
;

: init-hcqueue	\ size cqueue --
\ *G Initialise the specified queue created by *\fo{ALLOCATE}.
\ ** When a queue is allocated from the heap by a phrase
\ ** of the form *\fo{/CQUEUE <size> + ALLOCATE}, this word
\ ** must be used.
  swap 1- over >qmask !			\ set queue size
  dup  dup /cqueue + swap >qbuffer !	\ set buffer pointer
  init-cqueue				\ perform normal initialisation
;

: init-dcqueue	\ size cqueue --
\ *G Initialise the given queue, which does not have a buffer
\ ** yet. The buffer is *\fo{ALLOCATE}d. If the heap allocation
\ ** fails, the buffer address is set to zero.
  over allocate
  if  drop 0  endif			\ -- size cqueue addr
  over >qbuffer !			\ set buffer pointer
  swap 1- over >qmask !			\ set queue size
  init-cqueue				\ perform normal initialisation
;

: term-dcqueue	\ cqueue --
\ *G Shut down the given queue, *\fo{FREE}ing the buffer.
  dup >qbuffer @ free drop
  dup >qbuffer off  init-cqueue
;


\ *******************
\ *S Queue primitives
\ *******************

: (>cqueue)	\ char cqueue -- ; put character on cqueue
\ *G Put char into the queue with no checks.
  dup >qbuffer @ over >qtail @ +	\ -- char queue q*
  rot swap c! 				\ -- queue
  dup >qchars incr
  dup >qtail @ 1+
  over >qmask @ and
  swap >qtail !
;

: (cqueue>)	\ cqueue -- char ; get next character from queue
\ *G Get the next character from the queue with no checks.
  dup >qbuffer @ over >qhead @ +
  c@ swap				\ -- char queue
  dup >qchars decr
  dup >qhead @ 1+
  over >qmask @ and
  swap >qhead !
;

: (cqfull?)	\ queue -- flag ; TRUE if queue full
\ *G Return true if the queue is full. No interrupt
\ ** protection is provided.
  dup >qchars @  swap >qmask @ u>
;

: cqfull?	\ queue -- flag ; TRUE if queue full
\ *G Return true if the queue is full. Interrupt
\ ** protection is provided.
  [i  dup >qchars @  swap >qmask @ u>  i]
;

: cqchars	\ queue -- #chars
\ *G Return the number of characters in the queue.
  >qchars @
;

: cqempty?	\ queue -- flag ; TRUE if queue empty
\ *G Return true if the queue is empty.
  >qchars @ 0=
;

: cqnotempty?	\ queue -- flag ; TRUE if queue not empty
\ *G Return true if the queue is not empty, i.e. if it contains any
\ ** characters.
  >qchars @ 0<>
;

: >cqueue	\ char cqueue -- ; spins if full
\ *G Put a character into the queue. If the queue is full, the
\ ** system waits (blocks) until there is enough space.
  begin
    [I dup (cqfull?)
   while
    I]  pause
  repeat
  (>cqueue) I]
;

: ?>cqueue	\ char cqueue -- ; discards if full
\ *G Put a character into the queue. If the queue is full, the
\ ** character is discarded.
  dup cqFull?
  if  2drop  else  [I (>cqueue) I]  endif
;

: cqueue>	\ queue -- char ; spins while queue empty
\ *G Remove the next character, waiting if the queue is empty.
  begin
    [I dup cqempty?
   while
    I]  pause
  repeat
  (cqueue>) I]
;


\ ======
\ *> ###
\ ======

