\ SHA-1 Secure Hash Algorithm for VFX Forth
\ Original code by Wil Baden, converted to PFW/VFX by MPE

char . dp-char !

0 [IF] ******************************************************

             SHA-1 Secure Hash Algorithm
             ---------------------------

                   Portable Forth
                   --------------

SHA-1-INIT  initializes the secure hash algorithm.

SHA-1-UPDATE  updates the algorithm for each text string.
                                     ( str len -- )

SHA-1-FINAL  completes the algorithm and displays the result.

WARNING: Bit streams are not implemented yet -- I'll do it
as soon as I need it.

****************************************************** [THEN]
0 [IF] ------------------------------------------------------

        Common Definitions

Environmental dependency on:  1 CHARS  is an 8-bit address
unit;  1 CELLS  is 4; two's complement arithmetic.

    Comment out what you already have. 

LROTATE  does a circular left shift of a 32-bit cell.

Flip-Endian  converts between big-endian and little-endian
    cells.  Used here to convert to big-endian.

HTYPE  displays a byte array in hex.        ( str len -- )

CUT  is convenient when scanning a string to split it.
                            ( s n s+i n-i -- s+i n-i s i )

------------------------------------------------------ [THEN]

: LROTATE  ( x n -- x' )
    2DUP LSHIFT >R 32 SWAP - RSHIFT R> OR ;

: Flip-Endian  ( 01020304 -- 04030201 )
    DUP 24 LROTATE $FF00FF00 AND
    SWAP 8 LROTATE $00FF00FF AND OR ;

((
: HTYPE  ( str len -- )
    HEX 0 ?DO                   ( str)
        COUNT 0 <# # # #> TYPE
        I 1+ 4 MOD 0= IF SPACE THEN
    LOOP DROP DECIMAL ;
))
: HTYPE  ( str len -- )
    HEX 0 ?DO                   ( str)
        DUP I [ BASE C@ [IF] ] 3 XOR [ [THEN] ] + C@
        0 <# # # #> TYPE
        I 1+ 4 MOD 0= IF SPACE THEN
    LOOP DROP DECIMAL ;

: CUT  ( a m b n -- b n a m-n )  DUP >R 2SWAP R> - ;

0 [IF] ------------------------------------------------------

BUFFER:  names a data space area of a given size
                                     ( n "name" -- )

THIRD  copies the third element of the stack onto the
    top of the stack.                 ( x y z -- x y z x )

3DUP  copies the three top elements of the stack onto the
    top of the stack.             ( x y z -- x y z x y z )

H# xxxxxxxx  is used for hex literals.  Source not shown.
    Use it, code it, or change it.  Code on request.

NOT  should be equivalent to  0=  or  INVERT  .

------------------------------------------------------ [THEN]

0 [if]
: BUFFER:  ( n "name" -- )  CREATE  ALLOT ;

: 3DUP  ( x y z -- x y z x y z )  THIRD THIRD THIRD ;
[then]

: THIRD  ( x y z -- x y z x )  2 PICK ;

0 [IF] ------------------------------------------------------

        IMPLEMENTATION and INTERFACE

For encapsulation, the subordinate functions of the Secure
Hash Algorithm are defined in the word list of a vocabulary.
Rudimentary definitions of VOCABULARY words are given
(commented out) for anyone who somehow does not have them.

INTERFACE  introduces the public words of the algorithm.

------------------------------------------------------ [THEN]

FALSE [IF]
: VOCABULARY  ( "name" -- )  CREATE  WORDLIST ,  DOES>
    @ >R  GET-ORDER  NIP  R> SWAP  SET-ORDER ;
: ALSO    ( -- )  GET-ORDER OVER SWAP 1+ SET-ORDER ;
: PREVIOUS  ( -- )  GET-ORDER  NIP 1-  SET-ORDER ;
[THEN]

: INTERFACE  ( -- )
    GET-ORDER  THIRD SET-CURRENT  SET-ORDER ;

0 [IF] ------------------------------------------------------

        Secure Hash Algorithm

------------------------------------------------------ [THEN]

VOCABULARY SECURE-HASH-ALGORITHM

ALSO SECURE-HASH-ALGORITHM

DEFINITIONS

0 [IF] ------------------------------------------------------
Secure-Hash  contents is computed as the secure hash.

Message-Block  is used as buffer for intermediate calculations.

SIZE  has the intermediate size in bits of the message

Final-Count  has the final size in bits of the message.

Single-Byte  is used to pad the message to a multiple
    of 512 bits.
------------------------------------------------------ [THEN]

5 CELLS BUFFER: Secure-Hash

16 CELLS BUFFER: Message-Block

2 CELLS BUFFER: SIZE

2 CELLS BUFFER: Final-Count

1 BUFFER: Single-Byte

0 [IF] ------------------------------------------------------

BLK0  and  BLK  treat the 16 cells of Message-Block as
    though they were the 80 cell expansion.     ( i -- x )

------------------------------------------------------ [THEN]

BASE C@ [IF]              \  Little Endian

: BLK0  ( i -- x )
    CELLS Message-Block + DUP >R @ Flip-Endian DUP R> ! ;

[ELSE]                     \  Big Endian

: BLK0  ( i -- x )
    CELLS Message-Block + @ ;

[THEN]

: BLK  ( i -- x )
    DUP 13 + 15 AND CELLS Message-Block + @ >R
    DUP  8 + 15 AND CELLS Message-Block + @ R> XOR >R
    DUP  2 + 15 AND CELLS Message-Block + @ R> XOR >R
    DUP      15 AND CELLS Message-Block + @ R> XOR
    1 LROTATE
    DUP ROT  15 AND CELLS Message-Block + ! ;

0 [IF] ------------------------------------------------------

F  G  H  are the nonlinear functions for scrambling the
    data.  The names are taken from Menezes _Handbook
    of Applied Cryptography_.  Used in  TRANSFORM  .

MIX  is the invariant part of the scrambling.  Used in
    TRANSFORM  .

TRANSFORM  hashes 512 bits of input.  Used in  UPDATE  .

------------------------------------------------------ [THEN]

: F  ( d c b -- bc |or| b'd )
    DUP >R AND SWAP R> INVERT AND OR ;

: G  ( d c b -- bc |or| bd |or| cd )
    2DUP AND >R  OR AND R>  OR ;

: H  ( d c b -- d |xor| c |xor| b )
    XOR XOR ;

: MIX                 ( e d c b sha a m -- e d c b a )
    SWAP DUP >R  5 LROTATE + +           ( e d c b sha)( R: a)
    SWAP >R  SWAP >R  SWAP >R            ( e sha)( R: a b c d)
    +                                    ( sha)
    R> SWAP R> SWAP R> 30 LROTATE SWAP   ( e d c sha)( R: a)
    R> SWAP                              ( e d c b a)( R: )
    ;

: TRANSFORM                               ( -- )
    0 4 DO  I CELLS Secure-Hash + @  -1 +LOOP     ( e d c b a)

    16 0 DO  >R 3DUP F $5A827999 + R> I BLK0  MIX  LOOP
    20 16 DO  >R 3DUP F $5A827999 + R> I BLK  MIX  LOOP
    40 20 DO  >R 3DUP H $6ED9EBA1 + R> I BLK  MIX  LOOP
    60 40 DO  >R 3DUP G $8F1BBCDC + R> I BLK  MIX  LOOP
    80 60 DO  >R 3DUP H $CA62C1D6 + R> I BLK  MIX  LOOP

    5 0 DO  I CELLS Secure-Hash + +!  LOOP ;

0 [IF] ******************************************************

        SHA-1-INIT   SHA-1-UPDATE   SHA-1-FINAL

****************************************************** [THEN]

INTERFACE

: SHA-1-INIT  ( -- )
    $67452301  $EFCDAB89  $98BADCFE  $10325476
    $C3D2E1F0
    0 4 DO  I CELLS Secure-Hash + !  -1 +LOOP
    0. SIZE 2! ;

: SHA-1-UPDATE    ( str len --  )
    BEGIN  SIZE CELL+ @ 511 AND 3 RSHIFT >R 64 R@ - OVER > NOT
    WHILE
        2DUP 64 R@ - /STRING CUT
            Message-Block R@ + SWAP MOVE
        TRANSFORM
        SIZE 2@  64 R> - 3 LSHIFT M+ SIZE 2!
    REPEAT
    Message-Block R> +  SWAP  DUP >R  MOVE  ( )
    SIZE 2@ R> 3 LSHIFT M+ SIZE 2! ;

: SHA-1-FINAL  ( -- )
    SIZE 2@
    [ BASE C@ [IF] ]       \  Little-endian to big-endian.
        Flip-Endian SWAP Flip-Endian SWAP
    [ [THEN] ]
    Final-Count 2!
    Single-Byte $80 OVER C! 1 SHA-1-UPDATE
    BEGIN  SIZE CELL+ @ 511 AND 448 = NOT WHILE
        Single-Byte 0 OVER C! 1 SHA-1-UPDATE
    REPEAT
    Final-Count 8 SHA-1-UPDATE
    ;

: .sha		\ -- ; display result
  Secure-Hash 20 HTYPE
;

PREVIOUS DEFINITIONS

0 [IF] ------------------------------------------------------

        Test Vectors from Menezes

------------------------------------------------------ [THEN]

CR SHA-1-INIT
   SHA-1-FINAL
cr ." Result   : " .sha
cr ." Should be: DA39A3EE 5E6B4B0D 3255BFEF 95601890 AFD80709"

CR SHA-1-INIT  S" a" SHA-1-UPDATE
   SHA-1-FINAL
cr ." Result   : " .sha
cr ." Should be: 86F7E437 FAA5A7FC E15D1DDC B9EAEAEA 377667B8"

CR SHA-1-INIT  S" abc" SHA-1-UPDATE
   SHA-1-FINAL
cr ." Result   : " .sha
cr ." Should be: A9993E36 4706816A BA3E2571 7850C26C 9CD0D89D"

CR SHA-1-INIT  S" abcdefghijklmnopqrstuvwxyz" SHA-1-UPDATE
   SHA-1-FINAL
cr ." Result   : " .sha
cr ." Should be: 32D10C7B 8CF96570 CA04CE37 F2A19D84 240D3A89"

(
-- 
Wil Baden   Costa Mesa, California   WilBaden@Netcom.com
)

