\ Doubles.fth - VFX Forth double and quad number support

((
Copyright (c) 1990..2004, 2006, 2008, 2012, 2013
Construction Computer Software
Cape Town
South Africa

Copyright (c) 2005, 2006, 2012, 2013
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441

DOUBLES.FTH contains double, triple and quadruple precision
arithmetic operations.

No word may be removed or have changed effects without permission
from CCS.


To do
=====

Change history
==============
20171002 CCS008 Precision updates.
20130924 MPE007 Corrected UTT/MOD.
20130520 CCS006 Added (TT+).
		Added and corrected QN/MOD and UQN/MOD.
		Added more CCSisms.
20120917 CCS005 Added T@ T! Q@ Q!.
		Renamed T<<1 to T2*.
		Added 3SWAP and 3OVER.
20120512 MPE004 Added UD/MOD and UDS/MOD
20120430 MPE003 DocGen and code tidy.
20060816 MPE002 Overhauled to separate generic and CCS extensions.
20041201 MPE001 Modified to use ANS Forth naming conventions.
                Optimised for VFX code generators.
))

decimal

\ ==========
\ *! doubles
\ *T Double, Triple and Quad number support
\ ==========
\ *P The file *\i{Lib\Doubles.fth} contains double, triple and
\ ** quadruple precision arithmetic operations. These words are
\ ** in regular use by *\i{Construction Computer Software} (CCS).
\ ** No word will be removed or changed without permission from
\ ** CCS.

[FORTH]

[undefined] CCSisms? [if]
1 constant CCSisms?	\ -- 0/1
\ *G Change to zero not to compile CCS specific code. This
\ ** version is used only if the word has not already been
\ ** defined. This allows you to define *\fo{CCSisms?} as you
\ ** want before compiling this file.
[then]


\ ***********
\ *S Notation
\ ***********
\ *P The following stack notation is used:
\ *D d  double (64 bit) signed value.
\ *D t  triple (96 bit) signed value.
\ *D b  d.dpl, double (64 bit) signed value and single (8 bit) signed dpl.
\ *D q  quad (128 bit) signed value.
\ *D s  s.dpl single (32 bit) signed value and single (8 bit) signed dpl.
\ *D n  single (32 bit) signed value.
\ *D u  unsigned single (32 bit) unsigned value.
\ *D c  character (8 bit) unsigned value.
\ *D w  word (16 bit) unsigned value.
\ *D r  real (floating point) on floating point stack
\ *D f  real (floating point) on floating point stack
\ *D d.pl double (64 bit) signed value and single (8 bit) signed dpl.
\ *D x.dpl first d.dpl parameter.
\ *D y.dpl 2nd d.dpl parameter.
\ *D z.dpl result d.dpl parameter.

\ *********************
\ *S Storage Byte Sizes
\ *********************
\ *P The following notation is used for storage sizes:
\ *D kilobyte  (kB/KB) 10^3
\ *D megabyte  (MB)    10^6
\ *D gigabyte  (GB)    10^9
\ *D terabyte  (TB)    10^12
\ *D petabyte  (PB)    10^15
\ *D exabyte   (EB)    10^18
\ *D zettabyte (ZB)    10^21
\ *D yottabyte (YB)    10^24

\ ****************
\ *S Double values
\ ****************
\ *P The Maximum values for doubles:
\ *D Quintillion  1,000,000,000,000,000,000      10^18
\ *D Quadrillion      1,000,000,000,000,000      10^15
\ *D Trillion             1,000,000,000,000      10^12
\ *D Billion/Milliard         1,000,000,000      10^9
\ *D Million                      1,000,000      10^6
\ *D Lakh                           100,000      10^5
\ *D Myriad                          10,000      10^4
\ *D Zero dpl     9,223,372,036,854,775,807
\ *D one dpl        922,337,203,685,477,580.7
\ *D Two  dpl        92,233,720,368,547,758.07
\ *D Four dpl           922,337,203,685,477.5807
\ *D Six  dpl             9,223,372,036,854.775807
\ *D Seven dpl              922,337,203,685.4775807
\ *D Eight dpl               92,233,720,368.54775807

\ *S Extreme constants
  $80000000 CONSTANT NaN
\ *G 32bit no value
: D-NaN                                                       ( --- 0 nan )
   \ *G 64bit NaN
   $00000000 Nan
  ;

: D.DPL-NaN                                                       ( --- 0 nan )
   \ *G 64bit NaN and DPL
   D-NaN #128 ;

: D.DPL-NaN?                                                 ( d.dpl --- bool )
   DROP D-NaN D= ;

: D.DPL?                                                       ( d.dpl -- flag )
   \ *G True if d.dpl has a value
   DROP D0<> ;

: D.DPL-ZERO         ( --- #0 #0 #0 )
   #0 #0 #0
;

\ *****************
\ *S Coded routines
\ *****************

code t+		\ t1l t1m t1h t2l t2m t2h -- t3
\ *G Add two triple precision (96 bit) integers.
                \ 16  12  8   4   0   EBX
   mov   eax, #16 [ebp]			\ low items
   add   eax, 4 [ebp]
   mov   #16 [ebp], eax
   mov   eax, #12 [ebp]			\ middle items
   adc   eax, 0 [ebp]
   mov   #12 [ebp], eax
   adc   ebx, 8 [ebp]			\ high items
   lea   ebp, #12 [ebp]			\ update stack
   next,
end-code

code t-		\ t1l t1m t1h t2l t2m t2h -- t1-t2
\ *G Subtract two triple precision (96 bit) integers.
                \ 16  12  8   4   0   EBX
   mov   eax, #16 [ebp]			\ low items
   sub   eax, 4 [ebp]
   mov   #16 [ebp], eax
   mov   eax, #12 [ebp]			\ middle items
   sbb   eax, 0 [ebp]
   mov   #12 [ebp], eax
   mov   eax, 8 [ebp]			\ high items
   sbb   eax, ebx
   mov   ebx, eax
   lea   ebp, #12 [ebp]			\ update stack
   next,
end-code

code tnegate	\ t1 -- t2
\ *G Negate triple precision (96 bit) integer.
  not   4 [ebp]
  not   0 [ebp]
  not   ebx
  xor   edx, edx
  add   4 [ebp], # 1
  adc   0 [ebp], edx
  adc   ebx, edx
  next,
end-code

code t2*	\ t -- t<<1
\ *G Shift triple number 1 bit to left.
  shl   4 [ebp], # 1
  rcl   0 [ebp], # 1
  rcl   ebx, # 1
  next,
end-code

code q+		\ q10 q11 q12 q13 q20 q21 q22 q23 --
\ *G Add two quadruple precision (128 bit) integers.
                \ 24  20  16  12  8   4   0   EBX
   mov   eax, #24 [ebp]			\ lowest items
   add   eax, 8 [ebp]
   mov   #24 [ebp], eax
   mov   eax, #20 [ebp]			\ ls middle items
   adc   eax, 4 [ebp]
   mov   #20 [ebp], eax
   mov   eax, #16 [ebp]			\ ms middle items
   adc   eax, 0 [ebp]
   mov   #16 [ebp], eax
   adc   ebx, #12 [ebp]			\ high items
   lea   ebp, #16 [ebp]			\ update stack
   next,
end-code

code q-		\ q10 q11 q12 q13 q20 q21 q22 q23 -- q1-q2
\ *G Subtract two quadruple precision (128 bit) integers.
                \ 24  20  16  12  8   4   0   EBX
   mov   eax, #24 [ebp]			\ low items
   sub   eax, 8 [ebp]
   mov   #24 [ebp], eax
   mov   eax, #20 [ebp]			\ ls middle items
   sbb   eax, 4 [ebp]
   mov   #20 [ebp], eax
   mov   eax, #16 [ebp]			\ ms middle items
   sbb   eax, 0 [ebp]
   mov   #16 [ebp], eax
   mov   eax, #12 [ebp]			\ high items
   sbb   eax, ebx
   mov   ebx, eax
   lea   ebp, #16 [ebp]			\ update stack
   next,
end-code

code qnegate	\ q1 -- q2
\ *G Negate quadruple precision (128 bit) integer.
  not   8 [ebp]
  not   4 [ebp]
  not   0 [ebp]
  not   ebx
  xor   edx, edx
  add   8 [ebp], # 1
  adc   4 [ebp], edx
  adc   0 [ebp], edx
  adc   ebx, edx
  next,
end-code

code q2*	\ q -- q<<1
\ *G Shift quad number 1 bit to left.
  shl   8 [ebp], # 1
  rcl   4 [ebp], # 1
  rcl   0 [ebp], # 1
  rcl   ebx, # 1
  next,
end-code

: 3SWAP                                 \ x1 x2 x3 x4 x5 x6 -- x4 x5 x6 x1 x2 x3
   >R >R SWAP 2SWAP                                             \ 4 3 1 2 -- 6 5
   R> SWAP 2SWAP                                                \ 4 5 2 3 1 -- 6
   R> SWAP 2SWAP ;                                                 \ 4 5 6 1 2 3

: 3OVER                        \ x1 x2 x3 x4 x5 x6 -- x1 x2 x3 x4 x5 x6 x1 x2 x3
   5 PICK 5 PICK 5 PICK ;

\ ********************
\ *S Memory operations
\ ********************

: 3@		\ addr -- t
\ *G Fetch a triple. The most significant word is at the low address.
  dup 2 cells + @ swap			\ -- t.lo addr
  dup 1 cells + @ swap			\ -- t.lo t.mid addr
  @
;

: 3!		\ t addr --
\ *G store a triple. The most significant word is at the low address.
  tuck !  tuck cell + !  2 cells + !
;

: 4@		\ addr -- q
\ *G Fetch a quad. The most significant word is at the low address.
  dup 3 cells + @ swap			\ -- q.lo addr
  dup 2 cells + @ swap			\ -- q.lo q.ml addr
  dup 1 cells + @ swap			\ -- q.lo q.ml q.mh addr
  @					\ -- q
;

: 4!		\ q addr --
\ *G Store a quad. The most significant word is at the low address.
  tuck !  tuck cell + !  tuck 2 cells + !  3 cells + !
;


: D@		\ addr -- d
\ *G Fetch a double in native form. For the x86 ISA, the most
\ ** significant word is at the high address.
  dup @ swap cell + @
;

: D!		\ d addr --
\ *G store a double in native form. The most significant word is
\ ** at the high address.
  tuck cell + ! !
;

: T@		\ addr -- t
\ *G Fetch a triple in native form. For the x86 ISA, the most
\ ** significant word is at the high address.
  dup @ swap				\ -- t.lo addr
  dup cell + @ swap			\ -- t.lo t.mid addr
  2 cells + @
;

: T!		\ t addr --
\ *G store a triple in native form. The most significant word is
\ ** at the high address.
  tuck 2 cells + !  tuck cell + !  !
;

: Q@		\ addr -- q
\ *G Fetch a quad in native form. The most significant word is at
\ ** the high address.
  dup           @ swap			\ -- q.lo addr
  dup 1 cells + @ swap			\ -- q.lo q.ml addr
  dup 2 cells + @ swap			\ -- q.lo q.ml q.mh addr
      3 cells + @			\ -- q
;

: Q!		\ q addr --
\ *G Store a quad in native form. The most significant word is at
\ ** the high address.
  tuck 3 cells + !  tuck 2 cells + !  tuck 1 cells + !  !
;

: D.DPL!                                            ( d dpl addr --- )
   \ *G Store 64bit signed value and 8bit sign dpl at address.
   DUP >R 8 + C! R> D!
;

: D.DPL@                                            ( addr --- d dpl )
   \ *G Fetch 64bit signed value and 8bit sign dpl at address.
   DUP >R D@ R> 8 + C@S
;

\ ************************
\ *S High level primitives
\ ************************

: UP-PRECISION                                                  ( n1 --- n1 n2 )
   \ *G Increase the precision of n1 with the data stack width
   DUP 0< ;

[UNDEFINED] DU<	[IF]
: DU<		\ ud1 ud2 -- flag
\ *G True if ud1<ud2.
  ROT SWAP 2DUP U< IF
    4DROP True
  ELSE
    <>
    IF  2DROP False  ELSE  U<  THEN
  THEN
;
[THEN]

: TU<		\ ut1 ut2 -- flag
\ *G True if ut1<ut2.
  >r rot r>  2dup u< if			\ -- ud1 ud2 uh1 uh2
    2drop 2drop 2drop  true  exit
  endif
  <> if					\ -- ud1 ud2
    2drop 2drop  false  exit
  endif
  du<
;

: QU<		\ uq1 uq2 -- flag
\ *G True if uq1<uq2
  { | q1[ #4 cells ] q2[ #4 cells ] }
  q2[ Q! q1[ Q!
  q1[ #3 CELLS+ @ q2[ #3 CELLS+ @ 2DUP U<
  IF 2DROP True EXIT THEN
  <> IF   False EXIT THEN
  q1[ #2 CELLS+ @ q2[ #2 CELLS+ @ 2DUP U<
  IF 2DROP True EXIT THEN
  <> IF   False EXIT THEN
  q1[ #1 CELLS+ @ q2[ #1 CELLS+ @ 2DUP U<
  IF 2DROP True EXIT THEN
  <> IF False EXIT THEN
  q1[ CELL+ @ q2[ CELL+ @ U<
  IF True EXIT THEN
  False
; \ WJB - 2015/08/14 12:32:20 PM

: tabs		\ t -- |t|
\ *G Return the absolute value of a triple.
  dup 0<
  if  tnegate  endif
;

: qabs		\ q -- |q|
\ *G Return the absolute value of a quad.
  dup 0<
  if  qnegate  endif
;

SYNONYM N>D UP-PRECISION   \ n --- d
\ *G Convert signed single value to signed double.

SYNONYM D>T UP-PRECISION   \ d -- t
\ *G Convert signed double to signed triple.

SYNONYM T>Q UP-PRECISION   \ t -- q
\ *G Convert signed triple to signed quad.

: D>Q		\ d -- q
\ *G Convert signed double to signed quad.
  UP-PRECISION dup
;

: N-SIGN	\ n -- |n| flag
\ *G Checks the sign of a single number and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
  DUP ABS SWAP 0<
;

: D-SIGN	\ d -- |d| flag
\ *G Checks the sign of a double number and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
  DUP 0< DUP >R
  IF  DNEGATE  THEN
  R>
;

: T-SIGN	\ t -- |t| flag
\ *G Checks the sign of a triple number and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
  DUP 0<  dup >R
  IF  TNEGATE  THEN
  R>
;

: Q-SIGN	\ q -- |q| flag
\ *G Checks the sign of a quadruple number and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
  DUP 0<  dup >R
  IF  QNEGATE  THEN
  R>
;

: t0=		\ t -- flag
\ *G Return true if t=0.
  or or 0=
;

: t0<>		\ t -- flag
\ *G Return true if t<>0.
  or or 0<>
;

: T?                                                          ( 96bit --- bool )
   \ *G Test 96bit value for 0=
   DUP NaN = IF 3DROP False EXIT THEN
   T-SIGN DROP                                      \ positive value for 0= test
   T0<>                                                      \ not zero 96bit ??
; \ WJB - 2015/08/18 07:19:14 AM

: TT=                                                         ( t1 t2 --- bool )
   { %t2.x0 %t2.x1 %t2.x2 --- }
   %t2.x2 =
   SWAP %t2.x1 = AND
   SWAP %t2.x0 = AND
; \ WJB - 2015/08/18 07:20:41 AM

: T-BIT-ROLL>>                                                ( t1 t2 --- t3 t4 )
   \ *G Roll t1 and t2 1bit right. Carry bit for t1 saved in LSB of t2
   { | %t2[ #3 CELLS ] --- }
   T2* %t2[ T!
   DUP 0< #1 AND %t2[ OR!
   T2* %t2[ T@
; \ WJB - 2015/08/18 07:59:14 AM

\ *H Extend Object Type T-VAR:
: T-VAR-<ADD>                                                     ( t0 obj --- )
   \ *G Add 96bit value t0 to 96bit value at obj address. <add> method
   { %obj --- }
   %obj T@
   T+ %obj T!
;

: T-VAR-<INC>                                                        ( obj --- )
   \ *G INCR 96bit value at obj address. <inc> method
   { %obj --- }
   #1 #0 #0 %obj T-VAR-<ADD>
;

: T-VAR-<SUB>                                                     ( t0 obj --- )
   \ *G SUBTRACT t0 from 96 value at obj address. <sub> method
   \ -4 -1 - = -3
   \ -4 1  - = -5
   \ 4 -1  - =  5
   { %obj --- }
   %obj #2 CELLS+ @ 0<
   IF TNEGATE THEN
   %obj T-VAR-<ADD>
;

: T-VAR-<DEC>                                                        ( obj --- )
   \ *G DECR 96bit value at obj address. <dec> method
   { %obj --- }
   #1 #0 #0 T-VAR-<SUB>
;

: T-VAR-<NEGATE>                                                       \ obj ---
   \ *G NEGATE 96bit value at obj address. <negate> method
   DUP >R T@ TNEGATE R> T!
;

: T-VAR-<NEG>                                                        ( obj --- )
   \ *G NEGATE 96bit value at obj address if not already negative. Value at
   \ ** obj address will be negative.
   \ ** <neg> method
   { %obj --- }
   %obj #2 CELLS+ @ 0< 0=
   IF %obj T-VAR-<NEGATE> THEN
;

: T-VAR-<ABS>                                                        ( obj --- )
   \ *G ABS 96bit value at obj address. Value at obj address will be positive.
   \ ** <abs> method
   { %obj --- }
   %obj #2 CELLS+ @ 0<
   IF %obj T-VAR-<NEGATE> THEN
;

: T-VAR-<SIGN?>                                                 ( obj --- flag )
   \ *G ABS 96bit value at obj address. Return tru if value was negative or
   \ false. Value at obj address will be positive. <sign?> method
   { %obj --- }
   %obj #2 CELLS+ @ 0< DUP
   IF %obj T-VAR-<NEGATE> THEN
;

: q0=		\ q -- flag
\ *G Return true if q=0.
  or or or 0=
;

: q0<>		\ q -- flag
\ *G Return true if q<>0.
  or or or 0<>
;


\ ******************
\ *S Multiplications
\ ******************

: umds*		\ ud u -- ut
\ *G Unsigned mixed multiply.
\ ** Multiply double by single producing a triple.
\ long mul :  x1 x0
\              * y0
\         -----------
\  x0*y0 =    z1 z0
\  x1*y0 = z2 z1
\         -----------
\          z2 z1 z0
\         ===========
\         ===========
  DUP 0= IF 3DROP #0 #0 #0 EXIT THEN
  OVER 0= IF NIP UM* #0 EXIT THEN       \ x0*y0=z1.z0
  ROT OVER UM*                          \ x0*y0=z1.z0
  2SWAP UM*                             \ x1*y0=z2.z1
  -ROT 0 ROT 0 D+                       \ z1+z1 = z1.z2'
  ROT +
;

: mds*		\ d n -- t
\ *G Signed multiply of a double by a single to
\ ** return a triple result.
  dup if
    2dup xor >r			\ sign of result
    abs -rot  dabs rot umds*
    r> 0<
    if  tnegate  endif
  else
    drop 2drop  0 0 0		\ multiplier 0, return zeroes
  then
;

: umdd*		\ ud1 ud2 -- uq
\ *G Unsigned multiply, two unsigned doubles to produce a quad
\ long mul :       d1.x1 d1.x0
\                * d2.x1 d2.x0
\          -------------------
\  1               q1.x1 q1.x0 = d1.x0 * d2.x0
\  2         q1.x2 q1.x1       = d1.x1 * d2.x0
\  3         q1.x2 q1.x1       = d1.x0 * d2.x1
\  4   q1.x3 q1.x2             = d1.x1 * d2.x1
\     ------------------------
\      q1.x3 q1.x2 q1.x1 q1.x0
\     ========================
  { | ud1[ 2 cells ] ud2[ 2 cells ] tmp[ cell ] -- uq }
  ud2[ 2!  ud1[ 2!
  ud1[ cell+ @  ud2[ cell + @ um* tmp[ !		\ ud1.lo * ud2.lo
  ud1[ @ ud2[ cell + @ um* 0				\ ud1.hi * ud2.lo
  tmp[ @ 0 0 t+
  ud1[ cell + @ ud2[ @ um* 0 t+				\ ud1.lo * ud2.hi
  0  ud1[ @ ud2[ @ um* t+ 				\ ud1.hi * ud2.hi
;

: umts*		\ ut u -- uq
\ *G Unsigned mixed multiply of triple by single to produce a quad.
\ long mul :  x2 x1 x0
\                 * y0
\          -----------
\ x0*y0 =        z1 z0
\ x1*y0 =     z2 z1
\ x2*y0 =  z3 z2
\          -----------
\          z3 z2 z1 z0
\          ===========
  swap >r  dup >r		\ -- x0 x1 y0 ; R: -- x2 y0
  rot um*			\ -- x1 (z1.z0)l (z1.z0)h ; R: -- x2 y0
  rot 0 swap r@ um*		\ -- (z1.z0)l (z1.z0)l 0 (z2.z1)l (z2.z1)h ; R: -- x2 y0
  d+				\ -- (z1.z0)l (z2.z1')l (z2.z1')h ; R: -- x2 y0
  0 r> r> um* d+		\ -- uq
;

: mts*		\ t n -- q
\ *G Signed mixed multiply of triple by single to produce a quad.
  dup if                                \ check multiplier for 0=
    2dup xor >r				\ sign of result
    abs >r tabs r> umts*
    r> 0<
    if  qnegate  endif
  else
    drop 3drop 0 0 0 0                  \ multiplier 0= return zeroes
  then
;

: UDD*                                                       ( ud1 ud2 --- uq1 )
\ is this the same as UMDD* ?
   \ Unsigned multiply
   \ long mul :       d1.x1 d1.x0
   \                * d2.x1 d2.x0
   \          -------------------
   \  1               q1.x1 q1.x0 = d1.x0 * d2.x0
   \  2         q1.x2 q1.x1       = d1.x1 * d2.x0
   \  3         q1.x2 q1.x1       = d1.x0 * d2.x1
   \  4   q1.x3 q1.x2             = d1.x1 * d2.x1
   \     ------------------------
   \      q1.x3 q1.x2 q1.x1 q1.x0
   \     ========================
   { | %d1[ #2 CELLS ] %d2[ #2 CELLS ] %q1[ #4 CELLS ] --- }
   %d2[ D! %d1[ D! %q1[ #4 CELLS ERASE
   %d2[ CELL+ @ 0=
   IF %d1[ D@ %d2[ @ umds* #0 EXIT THEN
   %d1[ CELL+ @ 0=
   IF %d2[ D@ %d1[ @ umds* #0 EXIT THEN
   \
   %d1[ D@ %d2[ @ umds* %q1[ T!
   #0 %d1[ D@ %d2[ CELL+ @  umds*
   %q1[ Q@ Q+
; \ WJB - 2015/08/27 08:32:07 AM


\ ************
\ *S Divisions
\ ************

synonym uds/mod	mu/mod \ ud1 u2 -- urem udquot
\ *G Unsigned divide of double dividend (64 bit) by a single
\ ** divisor (32 bit), returning a single remainder and a
\ ** double quotient.
synonym ud/mod mu/mod	\ ud1 u2 -- urem udquot
\ *G A synonym for *\fo{mu/mod) that appears in some other Forth
\ ** systems.

: mds/rem	\  d n -- rem dquot
\ *S Signed 64 bit divided by 32 bit divisor, returning a single
\ ** remainder and a double quotient. No protection is
\ ** provided. Symmetric division. A signed version of *\fo{MU/MOD}.
  over >r				\ sign of remainder=sgn(dividend)
  2dup xor >r				\ sign of quotient
  abs -rot dabs rot mu/mod		\ unsigned divide
  r> 0<
  if  dnegate  endif
  r> 0<
  if  rot negate -rot  endif
;

: umts/mod	\ ut u -- urem utquot
\ *G Divide unsigned triple by unsigned single, returning unsigned
\ ** single remainder and quotient.
  >R                                    \ -- t0 t1 t2 ; R: -- u
  0 R@ UM/MOD                           \ -- t0 t1 r(t2/u) q(t2/u) ; R: -- u
  R> SWAP >R                            \ -- t0 t1 r(t2/u) u ; R: -- q(t2/u)
  DUP >R                                \ -- t0 t1 r(t2/u) u ; R: -- q(t2/u) u
  UM/MOD                                \  rem2.x1 = rem1 & quot1
  R> SWAP >R                            \ get divisor & save quot1 on return stack
  UM/MOD                                \ rem1.x0 = rem0 & quot0
  R> R>                                 \ retrieve quot 1 & quot2 from return stack
;

: umts/		\ ut u -- utquot
\ *G Divide unsigned triple by unsigned single, returning unsigned
\ ** single quotient.
  >R                                    \ -- t0 t1 t2 ; R: -- u
  0 R@ UM/MOD                           \ -- t0 t1 r(t2/u) q(t2/u) ; R: -- u
  R> SWAP >R                            \ -- t0 t1 r(t2/u) u ; R: -- q(t2/u)
  DUP >R                                \ -- t0 t1 r(t2/u) u ; R: -- q(t2/u) u
  UM/MOD                                \  rem2.x1 = rem1 & quot1
  R> SWAP >R                            \ get divisor & save quot1 on return stack
  UM/MOD nip                            \ rem1.x0 = rem0 & quot0
  R> R>                                 \ retrieve quot 1 & quot2 from return stack
;

: mts/rem	\ t n -- rem tquot
\ *S Signed divide a triple by a single, returning a single
\ ** remainder and a triple quotient. No protection is
\ ** provided. Symmetric division.
  over >r				\ sign of remainder=sgn(dividend)
  2dup xor >r				\ sign of quotient
  abs >r tabs r> umts/mod		\ unsigned divide
  r> 0<
  if  tnegate  endif
  r> 0<
  if  >r >r >r  negate  r> r> r>  endif
;

: mts/		\ t n -- tquot
\ *S Signed divide a triple by a single, returning a triple
\ ** quotient. No protection is provided. Symmetric division.
  2dup xor >r				\ sign of quotient
  abs >r tabs r> umts/			\ unsigned divide
  r> 0<
  if  tnegate  endif
;

code dt2*	\ t1 t2 -- t1' t2'
\ *G Shift left two triples by one bit. This is a factor used below.
\ ** The carry out from *\i{t1} is shifted into *\i{t2}.
  shl   #16 [ebp], # 1
  rcl   #12 [ebp], # 1
  rcl   #8 [ebp], # 1
  rcl   #4 [ebp], # 1
  rcl   #0 [ebp], # 1
  rcl   ebx, # 1
  next,
end-code

: utt/mod	\ ut1 ut2 -- utrem utquot
\ *G Unsigned divide of two triple numbers producing a triple
\ ** remainder and a triple quotient.
  {: | tdiv[ 3 cells ] trem[ 3 cells ] tquot[ 3 cells ] tmp -- :}
  tdiv[ T!  tquot[ T!  trem[ #3 CELLS ERASE
  96 0 do
    trem[ 2 cells + @ 0< -> tmp		\ preserve carry out
    tquot[ T@ trem[ T@ dt2* trem[ T! tquot[ T!	\ double triple left shift
    trem[ T@ tdiv[ T@ tu< invert  	\ check trial subtract
    tmp or if
      trem[ T@ tdiv[ T@ t- trem[ T!	\ subtract
      1 tquot[ or!			\ update quotient l.s. bit
    endif
  loop
  trem[ T@  tquot[ T@
;

: umqs/mod                                              ( uq u --- urem uqquot )
\ *G Unsigned divide of quad value by unsigned single producing a single
\ ** reainder and a quad quotient.
  { | %q1[ #4 CELLS ] %qquot[ #4 CELLS ] }
  >R
  %q1[ Q!
  %qquot[ #4 CELLS ERASE
  %q1[ #3 CELLS + @ 0                                              \ q1.x3 0 ---
  R@ UM/MOD                                                     \ rem.x3 quot.x3
  %qquot[ #3 CELLS + !                                            \ save quot.x3
  %q1[ #2 CELLS + @ SWAP                                      \ q1.x2 rem.x3 ---
  R@ UM/MOD                                                     \ rem.x2 quot.x2
  %qquot[ #2 CELLS + !                                            \ save quot.x2
  %q1[ CELL+ @ SWAP                                           \ q1.x1 rem.x2 ---
  R@ UM/MOD                                                     \ rem.x1 quot.x1
  %qquot[ CELL+ !                                                 \ save quot.x1
  %q1[ @ SWAP                                                 \ q1.x0 rem.x1 ---
  R> UM/MOD                                                     \ rem.x0 quot.x0
  %qquot[ !                                                       \ save quot.x0
  %qquot[ Q@
; \ WJB - 2015/08/12 02:47:04 PM

: uqd/mod                                             ( uq1 ud2 -- uqem uqquot )
\ *G Unsigned divide of a quad by a double.
  {: | ddiv[ 2 cells ] qrem[ 4 cells ] qquot[ 4 cells ] -- :}
  ddiv[ D! qquot[ Q! qrem[ #4 CELLS ERASE
  #128 #0
  DO
     qrem[  Q@ DUP 0< >R Q2*                   \ preserve carry out bit from rem
     qrem[  Q!
     qquot[ Q@ DUP 0< #1 AND qrem[ OR! Q2*               \ save carry out in rem
     qquot[ Q!
     R>                                                            \ cary bit...
     qrem[ Q@ ddiv[ D@ #0 #0 QU< INVERT                   \ check trial subtract
     OR
     IF qrem[ Q@ ddiv[ D@ #0 #0 Q- qrem[ Q!                            \ subtract
        #1 qquot[ or!                                  \ update quotient l.s. bit
     THEN
  LOOP
  qrem[ Q@ qquot[ Q@
; \ WJB - 2015/08/14 12:52:15 PM

((
: D/MOD  D>F D>F FSWAP F2DUP F/ F>D 2DUP D>F F* F- F>D 2SWAP ;  ok
))


\ **********
\ *S CCSisms
\ **********

CCSisms? [if]		\ nz for CCS specifics
\ *G Change to zero not to compile CCS specific code.
 #DEFINE bvar>.x0 @                                             ( bvar> --- x0 )
 #DEFINE bvar>.x1 4+ @                                          ( bvar> --- x1 )
 #DEFINE bvar>.dpl #8 + C@S                                    ( bvar> --- dpl )
 #DEFINE bvar>.val D@                                        ( bvar> --- x0 x1 )
 #DEFINE bvar>.b@  D.DPL@                                ( bvar> --- x0 x1 dpl )

 #DEFINE bvar>.x0! !                                            ( x0 bvar> --- )
 #DEFINE bvar>.x1! 4+ !                                         ( x1 bvar> --- )
 #DEFINE bvar>.dpl! #8 + C!                                    ( dpl bvar> --- )
 #DEFINE bvar>.val! D!                                       ( x0 x1 bvar> --- )
 #DEFINE bvar>.b!   D.DPL!                               ( x0 x1 dpl bvar> --- )

: DPL-VALID?                                                    ( dpl --- bool )
   { %dpl --- }
   %dpl #128 < %dpl -128 > AND
;

: BVAR>.DPL-SUB                                                  ( n bvar> --- )
	 \ *G Subtract n from the dpl at bvar>. N must be a valid DPL and the result DPL
	 \ ** must be valid. If invalid DPL then bvar> will be modified to D.DPL-Nan
   { %adj %bvar> --- }
   %adj DPL-VALID?
   IF   %bvar> bvar>.dpl %adj - DUP DPL-VALID?
        IF   %bvar> bvar>.dpl! EXIT THEN
        DROP
   THEN
   D.DPL-NaN %bvar> D.DPL!
;

: BVAR>.DPL-ADD                                                  ( n bvar> --- )
	 \ *G Add n to the dpl at bvar>. N must be a valid DPL and the result DPL
	 \ ** must be valid. If invalid DPL then bvar> will be modified to D.DPL-Nan
   { %adj %bvar> --- }
   %adj DPL-VALID?
   IF   %bvar> bvar>.dpl %adj + DUP DPL-VALID?
        IF   %bvar> bvar>.dpl! EXIT THEN
        DROP
   THEN
   D.DPL-NaN %bvar> D.DPL!
  ;

: (dd+)		\ d1 d2 -- t
\ *G Add two signed doubles and return a signed triple.
  2 pick 0< -rot  dup 0<  t+
;

: (TT+)		\ t1 t2 --- q
\ *G Add two signed triad vakues and return a signed quad value.
  >r >r >r  dup 0<		\ t1 -> q1
  r> r> r>  dup 0<		\ t2 -> q2
  Q+
;

: (dd*)		\ d1 d2 -- t
\ *G Signed multiply of two doubles to produce a triple.
  2dup d0<> if
    d-sign >r
    2swap d-sign >r
    umdd* drop
    r> r> xor
    if  tnegate  then
  else
    2drop 2drop  0 0 0
  endif
;

: L.DPL>D.DPL	\ l.dpl -- d.dpl
\ *G Convert L.DPL to D.DPL
  OVER 0< SWAP
;

\ *N Scale down helper words
: ULONG>10^N         ( u --- n )
   \ *G given unsigned long return possible dpl adjustment towards zero value
	 { %u --- }
	 \  012345678
	 %u #99999999 U> IF #9 EXIT THEN
	 \  01234567
	 %u #9999999 >   IF #8 EXIT THEN
	 \  0123456
	 %u #999999 >    IF #7 EXIT THEN
	 \  012345
	 %u #99999 >     IF #6 EXIT THEN
	 \  01234
	 %u #9999 >      IF #5 EXIT THEN
	 \  0123
	 %u #999 >       IF #4 EXIT THEN
	 \  012
	 %u #99 >        IF #3 EXIT THEN
	 \  01
	 %u #9 >         IF #2 EXIT THEN
	 \  0
	 %u #0 >         IF #1 EXIT THEN
   #0
;

CREATE 10(*)	\ -- addr
\ *G Array of powers of ten.
\ 98765432109876543210
                     1 ,	\ 10^0
                    10 ,	\ 10^1
                   100 ,	\ 10^2
                  1000 ,	\ 10^3
                 10000 ,	\ 10^4
                100000 ,	\ 10^5
               1000000 ,	\ 10^6
              10000000 ,	\ 10^7
             100000000 ,	\ 10^8
            1000000000 ,	\ 10^9

: 10^N		\ N -- 10^N
\ *G Return value 10 to the power n, where n is in the range 0..9.
  N-SIGN >R   DUP #0 #9 WITHIN?
  IF  cells 10(*) + @  ELSE  DROP #0  THEN
  R> ?NEGATE				\ retrieve sign flag & adjust value
;

: N.10^		\ N -- 10exponent
\ *G Calculate positive 10exponent greater than N
  DUP IF
    abs 0				\ init exponent
    BEGIN
      1+ DUP 10^N
      2 PICK >				\ 10th exponent >= value
      OVER 9 >= OR			\ protect against overflow
    UNTIL
    NIP
  THEN
;

  CREATE  D-10(*)
                 1 , 0 ,                                                  \ 10^0
                10 , 0 ,                                                  \ 10^1
               100 , 0 ,                                                  \ 10^2
              1000 , 0 ,                                                  \ 10^3
             10000 , 0 ,                                                  \ 10^4
            100000 , 0 ,                                                  \ 10^5
           1000000 , 0 ,                                                  \ 10^6
          10000000 , 0 ,                                                  \ 10^7
         100000000 , 0 ,                                                  \ 10^8
        1000000000 , 0 ,                                                  \ 10^9
        1410065408 , 2 ,                                                 \ 10^10
        1215752192 , 23 ,                                                \ 10^11
       $D4A51000   , $000000E8 ,                                         \ 10^12
       $4E72A000   , $00000918 ,                                         \ 10^13
       $107A4000   , $00005AF3 ,                                         \ 10^14
       $A4C68000   , $00038D7E ,                                         \ 10^15
       $6FC10000   , $002386F2 ,                                         \ 10^16
       $5D8A0000   , $01634578 ,                                         \ 10^17
       $A7640000   , $0DE0B6B3 ,                                         \ 10^18

: D-10^N                                                          ( N --- d10^N )
   \ Return 64bit value 10 to the power n
   N-SIGN >R                     \ check sign of exponent & save on return stack
   DUP #0 #18 WITHIN?
   IF   #8 * D-10(*) + D@ ELSE DROP #0 #0 THEN
   R> IF DNEGATE THEN                        \ retreive sign flag & adjust value
; \ WJB - 2016/08/30 05:51:59 PM

: D?DUP			\ D --- D\-1 or 0
  2DUP D0<> dup 0=
  if  nip nip  endif
;

: dn/mod	\  d n -- rem dquot
\ *S 64 bit divided by 32 bit divisor, returning a single
\ ** remainder and a double quotient.
\ ** N.B. If n is zero, rem and dquot are returned as zero.
  ?dup 0= if
    2drop 0 0 0
  else
    mds/rem  rot abs -rot
  endif
;

: DN/		\ d n -- d'
\ *G 64 bit signed value divide by 32 bit signed value.
\ ** If divisor=0 return 0 result.
\ ** N.B. The value will be rounded to -infinity or +infinity.
  { | %sign? %divisor %d1[ #2 CELLS ] --- }
  N-SIGN to %sign? to %divisor
  D-SIGN %sign? XOR to %sign?
  2DUP D0= IF EXIT THEN
  2DUP D-NaN D= IF EXIT THEN
  %d1[ D!
  %divisor   0= IF #0 #0 EXIT THEN
  %divisor #1 = IF %d1[ D@ %sign? ?DNEGATE EXIT THEN
  %d1[ D@ %divisor DN/MOD                        \ calc rem0 quot0 quot1
  ROT %divisor #0 #1 #0 D+ D2/ DROP >=
  IF #1 #0 D+ THEN                                 \ increment value
  %sign? ?DNEGATE                               \ correct the sign
; \ WJB - 2016/05/10 01:01:13 PM

((
: DN/		\ d n -- d'
\ *G 64 bit signed value divide by 32 bit signed value.
\ ** If divisor=0 return 0 result.
\ ** N.B. The value will be rounded to -infinity or +infinity.
\ note the mod is increment and compared with divisor divided by 2
\ 27 0 5 dn/ = 6
  DUP IF				\ check divisor for 0=
    OVER 0< OVER 0< XOR >R		\ calc and save result sign flag
    DUP >R                              \ save divisor
    DN/MOD                              \ calc rem0 quot0 quot1
    ROT 1+ R> ABS 2/ > IF               \ rem0 greater than 1/2 divisor
      DABS                              \ absolute value
      1 0 D+                            \ increment value
      R@ ?DNEGATE                       \ correct the sign
    THEN
    r> drop                             \ drop saved sign flag
  ELSE \ yes!
    DROP 2DROP 0 0
  THEN
;
))

Synonym D?	D0<>	\ d -- flag
Synonym D=D	D=	\ d1 d2 -- flag
Synonym D<D	D<	\ d1 d2 -- flag
Synonym D-ABS	dabs    \ d -- |d|

Synonym udn*	umds*	\ ud u -- ut
\ *G Unsigned mixed multiply. Multiply double by single to produce
\ ** a triple.
synonym DN*	mds*	\ d n -- t
\ *G Multiply double signed value by single signed value and return
\ ** triple signed result.
Synonym utn*	umts*	\ ut u -- uq
\ *G Unsigned mixed multiply. Multiply triple by single to produce
\ ** a quad.
Synonym tn*	mts*	\ t n -- q
\ *G Signed mixed multiply. Multiply triple by single to produce
\ ** a quad.
((
Synonym UTN/MOD umts/mod	\ ut n -- urem utquot
\ *G Unsigned divide of triple by single.
))

((
: UTN/MOD                                                 ( ut1 u --- urem ut2 )
   \ *G Unsigned 96bit value divide by unsigned 32bit value and 96bit result
   \ ** and 32bit remainder.
   >R
   #0 R@ UM/MOD
   -ROT R@ UM/MOD
   4 NREV ROT R> UM/MOD
   2SWAP
; \ WJB - 2015/08/20 11:42:16 AM
))

: D>=                                                          ( d1 d2 --- bool )
   \ *G Return true if d1 64bit signed value is greater or equal to d2
   4DUP ROT = -ROT = AND IF 4DROP True EXIT THEN
   D>
; \ WJB - 2015/08/12 10:16:58 AM

: UTN/                                                       ( ut1 u1 --- ut2 )
   \ *G Unsigned 96bit value divide by unsigned 32bit and 96bit result.
   DUP >R UMTS/MOD
   #3 ROLL
   #0
   R> #0 #1 #0 D+ D2/ D>=
   IF #1 #0 #0 T+ THEN
; \ WJB - 2015/08/20 11:42:00 AM

: UT10/MOD                                                   ( ut1 --- urem ut2 )
   \ *G Unsigned 96bit value divide by 10 and 96bit result and 32bit remainder.
   #10 UMTS/MOD
; \ WJB - 2015/08/20 11:42:16 AM

: UT10/                                                          ( ut1 --- ut2 )
   \ *G Unsigned 96bit value divide by 10 and 96bit result.
   UT10/MOD
   #3 ROLL #4 >
   IF #1 #0 #0 T+ THEN
; \ WJB - 2015/08/20 11:42:00 AM

: TN/MOD	( t n -- rem tquot )
\ *G Protected divide: signed triple by single.
\ ** Check for zero divisor and return 0 answer.
  DUP 0=   				\ divide by 0 check
  IF 4DROP #0 #0 #0 #0  exit  THEN
  N-SIGN >R				\ ? sign & abs divisor save flag
  >R T-SIGN R> SWAP >R                  \ ? sign of BD & save flag
  DUP 1 = IF                            \ check divisor for 1=
    DROP  >R 0 -ROT R>                  \ put remainder of zero on stack
  ELSE
    umts/mod
  THEN
  R> R> XOR                             \ retrieve sign flags & calc new sign
  IF  TNEGATE  THEN                     \ adjust BD to sign flag
;

: TN/	\ t n -- tquot
\ *G Protected divide: signed triple by single.
\ ** Check for zero divisor and return 0 answer.
\ ** NOTE: The value will be rounded to -infinity or +infinity
 { | %n1 %t1[ 3 cells ] %sign? -- }
  DUP 0=                   	\ divide by 0 check
  IF 4DROP #0 #0 #0 EXIT THEN
  N-SIGN to %sign? to %n1       			\ ? sign & abs divisor save flag
  T-SIGN %sign? XOR to %sign?
  3DUP T0= IF EXIT THEN
  %t1[ T!       \ ? sign of BD & save flag
  %n1 #1 >                                      \ check divisor for 1=
  IF %t1[ T@ %n1 UTN/ %t1[ T! THEN
  %t1[ T@
  %sign?                   \ retrieve sign flags & calc new sign
  IF TNEGATE  THEN                     \ adjust BD to sign flag
;

: udd/mod	\ ud1 ud2 -- udrem udquot
\ *G Unsigned divide of two double numbers producing a double
\ ** remainder and a double quotient.
  {: | ddiv[ 2 cells ] drem[ 2 cells ] dquot[ 2 cells ] tmp -- :}
  ddiv[ D!  dquot[ D!  0 0 drem[ D!
  64 0 do
    drem[ 1 cells + @ 0< -> tmp		\ preserve carry out
    dquot[ D@ drem[ D@ q2* drem[ D! dquot[ D!	\ double double left shift
    drem[ D@ ddiv[ D@ du< invert  	\ check trial subtract
    tmp or if
      drem[ D@ ddiv[ D@ d- drem[ D!	\ subtract
      1 dquot[ or!			\ update quotient l.s. bit
    endif
  loop
  drem[ D@  dquot[ D@
;

: utd/mod                                            ( ut1 ud2 -- udrem utquot )
\ *G Unsigned divide of a triple by a double.
  {: | ddiv[ 3 cells ] trem[ 3 cells ] tquot[ 3 cells ] -- :}
  #0 ddiv[ T! tquot[ T! trem[ #3 CELLS ERASE
  ddiv[ D@ #1 #0 D= IF trem[ D@ tquot[ T@ EXIT THEN
  ddiv[ D@ D0= IF  trem[ D@ #0 #0 #0 EXIT THEN
  #96 #0
  do
     trem[  T@ #0 Q2* >R                                 \ preserve carry out
     trem[  T!
     tquot[ T@ #0 Q2* trem[ OR!
     tquot[ T!
     R>                                                            \ cary bit...
     trem[ T@ ddiv[ T@ TU< INVERT                   \ check trial subtract
     OR
     if
       trem[ T@ ddiv[ T@ T- trem[ T!                         \ subtract
       #1 tquot[ or!                                  \ update quotient l.s. bit
     then
  loop
  trem[ D@ tquot[ T@
; \ WJB - 2015/08/13 04:50:41 PM

: TD/		\ t1 d2 -- t2
\ *G Protected divide: signed triple by double producing a triple.
\ ** Check for zero divisor and return 0 answer.
\ ** NOTE: The value will be rounded to -infinity or +infinity
  { | %d1[ 2 cells ] %t1[ 3 cells ] %sign? -- }
  d-sign to %sign? %d1[ D!
  t-sign %sign? XOR to %sign?
  3DUP T0= IF EXIT THEN
  %t1[ T!
  %d1[ D@ D-NAN D=
  IF #0 #0 #0 EXIT THEN
  %d1[ D@ D0=
  IF #0 #0 #0 EXIT THEN
  %d1[ D@ #1 #0 D=
  IF %t1[ T@ %sign? IF TNEGATE THEN
     EXIT
  THEN
  %d1[ CELL+ @ 0=
  IF %t1[ T@
     %d1[ @ #1 U>
     IF %d1[ @ UTN/ THEN
     %sign? IF TNEGATE THEN
     EXIT
  THEN
  %t1[ T@ %d1[ D@ UTD/MOD %t1[ T!
  %d1[ D@ #1 #0 D+ d2/ D>=    \ check for rounding
  IF	 %t1[ T@ 1 0 0 t+
  ELSE %t1[ T@
  THEN
  %sign? IF TNEGATE THEN
; \ WJB - 2015/08/25 07:40:23 AM


: TD>RND                                                       ( t1 --- d1 dpl )
   \ *G Down round t1 value to d1 value and return dpl adjustment
   T-SIGN >R
   #0                          \ prime dpl adjustment t1.x0 t1.x1 t1.x2 dpl+ ...
   BEGIN
        OVER N.10^       \ calculate 10 exponent t1.x0 t1.x1 t1.x2 dpl+ dpl2 ...
        TUCK      \ save current adjustment t1.x0 t1.x1 t1.x2 dpl2 dpl+ dpl2 ...
        +           \ add to previous adjustment t1.x0 t1.x1 t1.x2 dpl2 dpl+ ...
        >R                    \ save total adjustment t1.x0 t1.x1 t1.x2 dpl2 ...
        DUP 0<>                                     \ any adjustment required ??
        IF   10^N UTN/ False                   \ yes! adjust and go another round
        ELSE DROP True
        THEN
        R>
        SWAP
   UNTIL
   NIP                                                    \ drop zero t.x2 value
   OVER 0<                                                     \ d.x1 msb set ??
   IF 1+ >R #0 UT10/ DROP R> THEN             \ inc adjustment and shift value
   R>                                                 \ sign of the value set ??
   IF -ROT DNEGATE ROT THEN
;

: UQN/MOD	                                             ( uq u -- urem uqquot )
\ *G Protected divide: signed 128 bit quad by 32 bit single producing a single
\ ** remainder and a quad quotient.
\ ** Check for zero divisor and return 0 answer.
 { | %divisor %q1[ 4 CELLS ] -- }
  to %divisor
  4DUP Q0= IF #0 EXIT THEN
  %q1[ Q!
  %divisor 0=
  IF #0 #0 #0 #0 #0 EXIT THEN
  %divisor #1 =
  IF #0 %q1[ Q@ EXIT THEN
  %q1[ Q@ %divisor umqs/mod
; \ WJB - 2015/08/12 03:24:10 PM

: UQN/                                                       ( uq1 u1 --- uq2 )
   \ *G Unsigned 128bit value divide by unsigned 32bit and 128bit result.
   DUP >R UQN/MOD
   #4 ROLL R> #0 #1 #0 D+ D2/ DROP U>=
   IF #1 #0 #0 #0 Q+ THEN
; \ WJB - 2015/08/26 10:56:42 AM

: UQ10/MOD                                                   ( uq1 --- urem uq2 )
   \ *G Unsigned 128bit value divide by 10 and 128bit result and 32bit remainder.
   #10 UQN/MOD
; \ WJB - 2015/08/26 10:24:00 AM

: UQ10/                                                          ( uq1 --- uq2 )
   \ *G Unsigned 128bit value divide by 10 and 128bit result.
   UQ10/MOD
   #4 ROLL #4 >
   IF #1 #0 #0 #0 Q+ THEN
; \ WJB - 2015/08/26 10:44:44 AM

: QN/MOD	\ q n -- rem qquot
\ *G Protected divide: signed quad by single producing a single
\ ** remainder and a quad quotient.
\ ** Check for zero divisor and return 0 answer.
   { | %divisor %sflag -- }
   N-SIGN to %sflag to %divisor
   Q-SIGN %sflag XOR to %sflag                              \ ? sign of Q & save flag
   %divisor UQN/MOD
   %sflag IF QNEGATE THEN                                 \ adjust BD to sign flag
; \ WJB - 2015/08/12 03:24:10 PM

: QN/		\ q n -- qquot
\ *G 128bit signed value divide with 32bit signed value.
\ ** check for 0= divisor and return 0 answer
  { | %divisor %sflag %q1[ 4 CELLS ] -- }
  N-SIGN to %sflag to %divisor
  Q-SIGN %sflag XOR to %sflag
  4DUP Q0= IF EXIT THEN
  %q1[ Q!
  %divisor 0=
  IF #0 #0 #0 #0 #0 EXIT THEN
  %divisor #1 >
  IF   %q1[ Q@ %divisor QN/MOD
       4 roll
       %divisor 1+ 2/ >=
       IF #1 #0 #0 #0 Q+ THEN
  ELSE %q1[ Q@
  THEN
  %sflag IF QNEGATE THEN                                 \ adjust BD to sign flag
; \ WJB 2015/08/14 04:51:06 PM

\ *N Quad value helper words
: UQ>T-ADJUST                                                   ( q1 --- t dpl )
   \ UQ10/
	 { | %adj --- }
	 #0 to %adj \ Prime dpl adjustment
	 BEGIN
	 CASE
	     DUP #99999999  > ?OF #1000000000 UQN/ %adj #9 + to %adj ENDOF
	     DUP #9999999   > ?OF #100000000  UQN/ %adj #8 + to %adj ENDOF
	     DUP #999999    > ?OF #10000000   UQN/ %adj #7 + to %adj ENDOF
	     DUP #99999     > ?OF #1000000    UQN/ %adj #6 + to %adj ENDOF
	     DUP #9999      > ?OF #100000     UQN/ %adj #5 + to %adj ENDOF
	     DUP #999       > ?OF #10000      UQN/ %adj #4 + to %adj ENDOF
 	     DUP #99        > ?OF #1000       UQN/ %adj #3 + to %adj ENDOF
 	     DUP #9         > ?OF #100        UQN/ %adj #2 + to %adj ENDOF
 	     DUP #0         > ?OF #10         UQN/ %adj #1 + to %adj ENDOF
	 END-CASE
	 DUP 0=
   UNTIL
   %adj                                           \ q1.x0 q1.x1 q1.x2 q1.x3 dpl+ ...
   NIP                                                   \ drop zero q1.x3 value
   OVER 0<                                                    \ t1.x2 msb set ??
   IF 1+ >R UT10/ R> THEN                       \ inc adjustment and shift value
; \ WJB - 2015/08/26 11:15:40 AM

: Q>T-ADJUST                                                      ( q1 --- t1 dpl )
   \ *G Down round q1 value to t1 value and return dpl adjustment
   Q-SIGN >R
   UQ>T-ADJUST
   R>                                                 \ sign of the value set ??
   IF >R TNEGATE R> THEN
; \ WJB - 2015/08/26 08:55:39 AM

: UDN/MOD   ( ud1 u --- rem ud2 )
	MU/MOD
; \ WJB - 2015/08/25 11:59:04 AM

: UD10/MOD  ( ud1 --- rem ud2 )
   #0 #10 UM/MOD -ROT #10 UM/MOD ROT
; \ WJB - 2015/08/25 09:11:44 AM

: UD10/                                                          ( ud1 --- ud2 )
   #0 #10 UM/MOD -ROT #10 UM/MOD -ROT
   #4 U>
   IF #1 #0 D+ THEN
; \ WJB - 2015/08/25 09:05:44 AM

: UDN/  ( %ud1 u --- %ud2 )
	DUP >R UDN/MOD ROT #0
	R> #0 #1 #0 D+ D2/ D>=
	IF #1 #0 D+ THEN
; \ WJB - 2015/08/25 12:16:47 PM

: UD>N-ADJUST                                                  ( ud1 --- u dpl )
   \ *G Unsigned 64 bit down round to 32 bit
   DUP                                                    \ any hi word value ??
   IF   \ yes!
        DUP N.10^                                        \ calculate 10 exponent
        DUP >R 10^N UDN/ R>                      \ calc start value for down rnd
        OVER                                          \ check hi dword for value
        IF
             BEGIN
                  1+ >R UD10/ R>                              \ rnd BD down to D
                  OVER 0=                       \ check hi dword of BD for value
             UNTIL
        THEN
   ELSE #0                                        \ init dpl adjustment on stack
   THEN
   NIP                                                     \ drop hi dword of BD
   OVER 0<                                   \ check hi dword of D for bit31 set
   IF 1+ >R #0 UD10/ DROP R> THEN                             \ rnd BD down to D
; \ WJB - 2015/08/26 09:17:38 AM

: D>N-ADJUST                                                    ( d1 --- n dpl )
   \
   D-SIGN  >R                                           \ save sign flag & abs D
   UD>N-ADJUST                                            \ rnd D down to 32bits
   SWAP R> ?NEGATE SWAP                  \ retrieve sign flag from return stack
; \ WJB - 2015/08/26 09:19:38 AM

: UD-SCALE-UP                                                  ( d1 --- d2 dpl )
   { | %d1[ #2 CELLS ] %dpl --- }
   2DUP D0= IF #0 EXIT THEN
   %d1[ D! #0 to %dpl
   %d1[ CELL+ @ 0<
   IF %d1[ D@ #0 EXIT THEN
   %d1[ CELL+ @ 0=
   IF %d1[ @ #1000000000 UM* %d1[ D! #9 to %dpl THEN             \ x0*y0 = z0.z1
   %d1[ CELL+ @ 0=
   IF  %d1[ @ #1000000000 UM* %dpl #9 + EXIT THEN
   %d1[ CELL+ @
   DUP #10 U<
   IF  #100000000 UM* DROP                                        \ x1*y0= z1.z2
       %d1[ @ #100000000 UM*                                      \ x0*y0= z0.z1
       ROT +                                                             \ z1+z1
       %dpl #8 + EXIT
   THEN
   DUP #100 U<
   IF  #10000000 UM* DROP                                         \ x1*y0= z1.z2
       %d1[ @ #10000000 UM*                                       \ x0*y0= z0.z1
       ROT +                                                             \ z1+z1
       %dpl #7 + EXIT
   THEN
   DUP #1000 U<
   IF  #1000000 UM* DROP                                          \ x1*y0= z1.z2
       %d1[ @ #1000000 UM*                                        \ x0*y0= z0.z1
       ROT +                                                             \ z1+z1
       %dpl #6 + EXIT
   THEN
   DUP #10000 U<
   IF  #100000 UM* DROP                                           \ x1*y0= z1.z2
       %d1[ @ #100000 UM*                                         \ x0*y0= z0.z1
       ROT +                                                             \ z1+z1
       %dpl #5 + EXIT
   THEN
   DUP #100000 U<
   IF  #10000 UM* DROP                                            \ x1*y0= z1.z2
       %d1[ @ #10000 UM*                                          \ x0*y0= z0.z1
       ROT +                                                             \ z1+z1
       %dpl #4 + EXIT
   THEN
   DUP #1000000 U<
   IF  #1000 UM* DROP                                             \ x1*y0= z1.z2
       %d1[ @ #1000 UM*                                           \ x0*y0= z0.z1
       ROT +                                                             \ z1+z1
       %dpl #3 + EXIT
   THEN
   DUP #10000000 U<
   IF  #100 UM* DROP                                              \ x1*y0= z1.z2
       %d1[ @ #100 UM*                                            \ x0*y0= z0.z1
       ROT +                                                             \ z1+z1
       %dpl #2 + EXIT
   THEN
   DUP #100000000 U<
   IF  #10 UM* DROP                                               \ x1*y0= z1.z2
       %d1[ @ #10 UM*                                             \ x0*y0= z0.z1
       ROT +                                                             \ z1+z1
       %dpl #1 + EXIT
   THEN
   DROP
   %d1[ D@ %dpl
; \ WJB - 2015/08/19 10:32:49 AM

: UD-SCALE-DOWN                                                ( d1 --- d2 dpl )
   { | %d1[ #2 CELLS ] %d2[ #2 CELLS ] %dpl --- }
   %d1[ D!
   #0 to %dpl
   BEGIN
       %d1[ D@ 2DUP %d2[ D!
       2DUP D0= IF %dpl EXIT THEN
       UD10/MOD %d1[ D! 0=
   WHILE
       %dpl 1+ to %dpl
   REPEAT
   %d2[ D@ %dpl
; \ WJB - 2015/08/19 10:34:38 AM

: DD/                                                    ( d1 d2 --- d3 )
   \ *G 64 signed value divide by 64bit signed value
   \ NOTE: The value will be rounded to -infinity or +infinity
   { | %d1[ #2 CELLS ] %d2[ #2 CELLS ] %sign? --- }
   D-SIGN to %sign? %d2[ D!
   D-SIGN %sign? XOR to %sign? %d1[ D!
   %d1[ D@ D0= IF #0 #0 EXIT THEN
   %d2[ D@ D0= IF #0 #0 EXIT THEN
   %d2[ D@ D-NaN D= IF D-NaN EXIT THEN
   %d1[ D@ D-NaN D= IF D-NaN EXIT THEN
   %d2[ D@ #1 #0 D= IF %d1[ D@ %sign? IF DNEGATE THEN EXIT THEN
   %d2[ CELL+ @ 0=
   IF  %d1[ D@ %d2[ @ UDN/
       %sign? IF DNEGATE THEN
       EXIT
   THEN
   %d1[ D@ %d2[ D@ UDD/MOD 2SWAP
   %d2[ D@ #1 #0 D+ D2/ D>=
   IF 1 0 D+ THEN                  \ yes! apply rounding
   %sign? IF DNEGATE THEN          \ retreive sign flag and adjust result
; \ WJB - 2016/05/11 12:53:25 PM

: UT>D-ADJUST                                                    ( t --- d dpl )
   \ *G Round a positive 96bit value to 64bit value & dpl adjustment
   DUP                                           \ check T hi dword for value ??
   IF   \ yes!
        DUP 0<
        IF UT10/ #1 ELSE #0 THEN
        OVER N.10^                                       \ calculate 10 exponent
        TUCK + >R 10^N UTN/ R>                       \ calc start value for down rnd
        OVER                                    \ check BD hi dword for value ??
        IF  BEGIN
              1+ >R UT10/ R>                                \ rnd BD down to D
              OVER 0=                           \ check hi dword of BD for value
            UNTIL
        THEN
   ELSE #0                                         \ init dpl adjustment on stack
   THEN
   NIP                                                     \ drop hi dword of BD
   OVER 0<                                   \ check hi dword of D for bit31 set
   IF 1+ >R UD10/ R> THEN                             \ rnd BD down to D
; \ WJB - 2015/08/18 04:43:46 PM

: T>D-ADJUST                                ( 96 bit --- 64 bit\DPL adjustment )
   \ *G Round 96bit value to 64bit value & dpl adjustment.
   T-SIGN >R                                 \ check sign & save on return stack
   UT>D-ADJUST                                                \ rnd BD down to D
   R> IF >R DNEGATE R> THEN                        \ retrieve sign flag & adjust
; \ WJB - 2016/05/16 04:49:55 PM

: Q>D-ADJUST                                                     ( q --- d dpl )
  Q-SIGN >R
  UQ>T-ADJUST >R
  UT>D-ADJUST R> +
  R> IF -ROT DNEGATE ROT THEN
; \ WJB - 2015/08/26 11:53:06 AM

((
: FP-DD/-DPL                                                   ( d1 d2 --- d3 dpl )
   { | %d1[ #2 CELLS ] %d2[ #2 CELLS ] %sign? %dpl --- }
   2DUP D-NaN D= IF 4DROP #0 #0 #0 EXIT THEN
   D-SIGN to %sign? UD-SCALE-DOWN to %dpl %d2[ D!
   2DUP D-NaN D= IF 2DROP #0 #0 #0 EXIT THEN
   D-SIGN %sign? XOR to %sign?
   2DUP D0= IF #0 EXIT THEN
   %d1[ D!
   %d2[ D@ D0= IF #0 #0 #0 EXIT THEN
   %d2[ D@ #1 #0 D= IF %d1[ D@ %sign? IF DNEGATE THEN %dpl EXIT THEN
   %d1[ D@ UD-SCALE-UP #9 + %dpl + to %dpl
   #1000000000 DN* %d2[ D@ TD/ UT>D-ADJUST %dpl SWAP - to %dpl
   %sign? IF DNEGATE THEN
   %dpl
; \ WJB - 2016/05/11 12:56:28 PM
))

((
: DD/                                                           ( d1 d2 --- d3 )
   \ *G 64 signed value divide by 64bit signed value
   \ NOTE: The value will be rounded to -infinity or +infinity
   { | %d1[ #2 CELLS ] %d2[ #2 CELLS ] %sign? --- }
   2DUP  D0= IF 4DROP #0 #0 EXIT THEN
   2OVER D0= IF 4DROP #0 #0 EXIT THEN
   D-SIGN to %sign? %d2[ D!
   D-SIGN %sign? XOR to %sign? %d1[ D!
   %d2[ 4+ @ 0=                                            \ d1.x1 zero value ??
   IF   %d1[ D@
        %d2[ @
        #1 U>                                          \ test for divide by 1 ??
        IF  %d2[ @ UDN/MOD ROT
            %d2[ @ 2/ u>=
            IF #1 #0 D+ THEN
        THEN
        %sign? IF DNEGATE THEN
        EXIT
   THEN
   %d1[ D@ %d2[ D@ UDD/MOD
   2SWAP
   %d2[ D@ D2/
   D>= IF #1 #0 D+ THEN
   \ retreive sign flag and adjust result
   %sign? IF DNEGATE THEN
; \ WJB - 2015/08/26 07:15:36 AM
))

: DD/-DPL                                                   ( d1 d2 --- d3 dpl )
   { | %d1[ #2 CELLS ] %d2[ #2 CELLS ] %sign? %dpl --- }
   D-SIGN to %sign? UD-SCALE-DOWN to %dpl %d2[ D!
   D-SIGN %sign? XOR to %sign?
   2DUP D0= IF #0 EXIT THEN
   %d1[ D!
   %d2[ D@ D0= IF #0 #0 #0 EXIT THEN
   %d2[ D@ #1 #0 D=
   IF %d1[ D@ %sign? IF DNEGATE THEN
      %dpl
      EXIT
   THEN
   %d1[ D@ UD-SCALE-UP #9 + %dpl + to %dpl
   #1000000000 DN* %d2[ D@ TD/ UT>D-ADJUST %dpl SWAP - to %dpl
   %sign? IF DNEGATE THEN
   %dpl
; \ WJB - 2015/08/18 04:42:39 PM

: D.DPL/                                         ( d1 dpl1 d2 dpl2 --- d3 dpl3 )
   \ Divide a b-value with a b-value.
   { | %dpl --- }
   DUP DPL-VALID? 0= IF 3DROP 3DROP #0 #0 #0 EXIT THEN
   to %dpl
   ROT DUP DPL-VALID? 0= IF 3DROP 2DROP #0 #0 #0 EXIT THEN
   %dpl - to %dpl
   DD/-DPL %dpl +
; \ WJB - 2015/08/26 07:22:53 AM


Synonym TVAR@ T@
Synonym TVAR! T!

' T@ >name hidename
' T! >name hidename

: DPL-WITHIN-DRANGE?                                      ( dpl1 dpl2 --- bool )
   { %dpl1 %dpl2 --- }
   %dpl1 DPL-VALID? %dpl2 DPL-VALID? AND
   IF   %dpl1 %dpl2 2DUP MAX -ROT MIN - #19 <
   ELSE False
   THEN
; \ WJB 2016/05/05 11:25:34 AM

: DPL-WITHIN-LRANGE?                                      ( dpl1 dpl2 --- bool )
   { %dpl1 %dpl2 --- }
   %dpl1 DPL-VALID? %dpl2 DPL-VALID? AND
   IF   %dpl1 %dpl2 2DUP MAX -ROT MIN - #9 <
   ELSE False
   THEN
; \ WJB 2016/05/05 11:26:22 AM

: D.DPL-NEGATE                                                     ( b1 --- b2 )
\ *G Negate a double signed number with a dpl
   >R DNEGATE R> ;

: L.DPL-NEGATE                                              ( N\DPL --- -N\DPL )
\ *G Negate a single signed number with a dpl
   SWAP NEGATE SWAP ;

: L.DPL-SIGN                                                ( S --- S abs\flag )
\ *G Checks the sign of a single number with dpl and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
   >R N-SIGN R> SWAP ;

: D.DPL-SIGN                                        ( D.DPL --- ABS_D.DPL\flag )
\ *G Checks the sign of a double number with dpl and returns the abs value
\ ** and a flag indicating the sign.
\ ** Flag is true (-1) for negative or false (0) for positive.
   >R D-SIGN R> SWAP
  ;

: D.DPL-ABS                                              ( D.DPL --- abs D.DPL )
\ *G Return the absolute of a double number with dpl
   >R DABS R> ;

: D.DPL>T.DPL                                                ( d.dpl --- t.dpl )
\ *G Convert a D.DPL to a T.DPL
   OVER 0< SWAP ;

: D.DPL*                                            ( d1.dpl d2.dpl --- d3.dpl )
   { | %dpl %sign? %d1[ #2 CELLS ] %d2[ #2 CELLS ] --- }
   to %dpl D-SIGN to %sign?
   UD-SCALE-DOWN %dpl SWAP - to %dpl
   %d2[ D!
   %dpl + to %dpl D-SIGN %sign? XOR to %sign?
   UD-SCALE-DOWN %dpl SWAP - to %dpl
   %d1[ D!
   %d2[ D@ D0= IF #0 #0 #0 EXIT THEN
   %d1[ D@ D0= IF #0 #0 #0 EXIT THEN
   %d1[ D@ %d2[ D@ UDD* Q>D-ADJUST %dpl SWAP - to %dpl
   %sign? IF DNEGATE THEN
   %dpl
; \ WJB - 2015/08/27 08:32:27 AM

: D.DPL>$                                                 ( d.dpl --- addr cnt )
   \ *G Convert d.dpl to decimal string
   BASE @ >R                                                 \ save current base
   DECIMAL                                              \ set base 10 conversion
   D.DPL-SIGN >R                                                \ save sign flag
   <#                                         \ start picture numeric conversion
   DUP 0>
   IF   #0 DO # LOOP
        [CHAR] . HOLD
   ELSE ABS #0 #0 ROT #0 ?DO # LOOP
        2DROP
   THEN
   #S                                    \ process balance of numeric conversion
   R> SIGN #>                                        \ get sign flag and process
   R> BASE ! ;                                            \ restore current base

: T#   ( ut --- ut' )
\ *G Given a tripple number on the stack this will add the next digit to
\ ** the pictured numeric output buffer and return the next tripple
\ ** number to work with. N.B. the output string is built from
\ ** right (lsd) to left (msd).
   BASE @ UMTS/MOD 2SWAP SWAP
   9 OVER < IF  7 +  THEN
   [CHAR] 0 + HOLD
   -ROT
;

: T>$                                                  ( x0 x1 x3 --- addr cnt )
   T-SIGN >R
    <#                                        \ start picture numeric conversion
   BEGIN
       #10 UMTS/MOD 2SWAP SWAP
       [CHAR] 0 + HOLD
       -ROT
       3DUP OR OR 0<>                                            \ zero 96bit ??
       IF   False
       ELSE True
       THEN
   UNTIL
   3DROP
   R> SIGN                                           \ get sign flag and process
   HLD @ PAD OVER -
  ;

: Q>$                                                       ( q1 --- addr ucnt )
\ *G Covert quad value to decimal string
  { | %q1[ 4 CELLS ] }
   Q-SIGN >R
   %q1[ Q!
   <#                                         \ start picture numeric conversion
   BEGIN
       %q1[ Q@ #10 umqs/mod %q1[ Q!
       [CHAR] 0 + HOLD
       %q1[ Q@
       OR OR OR 0=                                             \ zero 128 bit ??
   UNTIL
   R> SIGN                                           \ get sign flag and process
   HLD @ PAD OVER -
; \ WJB - 2015/08/12 02:52:57 PM

: D.DPL>D                                                    ( D\DPL\DPL --- D )
   \ Dpl valid range -127 ~ +127
   2DUP =                                                       \ dpl's equal ??
   IF 2DROP EXIT THEN                                                     \ yes!
   - DUP DPL-VALID? 0=                    \ check dpl adjustment within range ??
   IF 3DROP #0 #0 EXIT THEN                    \ no! dpl adjustment out of range
   DUP ABS #19 >
   IF 3DROP #0 #0 EXIT THEN                    \ no! dpl adjustment out of range
   DUP 0>                               \ check direction of magnitude to adjust
   IF   BEGIN
             DUP 0>                          \ check adjustment for any value ??
        WHILE
             DUP #18 MIN TUCK - >R >R                    \ calc integer adjustment
             2DUP D0<>                             \ any value left to adjust ??
             IF   \ yes!
                  R>
                  D-10^N DD/                                  \ apply adjustment
                  R>
             ELSE R> R> 2DROP #0
             THEN
        REPEAT
        DROP                                                \ copy of adjustment
   ELSE ABS
        BEGIN
             DUP 0>                          \ check adjustment for any value ??
        WHILE
             DUP #9 MIN TUCK - >R >R
             2DUP D0<>                             \ any value left to adjust ??
             IF   \ yes!
                  R>
                  10^N DN* DROP                               \ apply adjustment
                  R>
             ELSE R> R> 2DROP #0
             THEN
        REPEAT
        DROP                                                \ copy of adjustment
   THEN
; \ WJB - 2015/08/12 09:37:38 AM

: D.DPL>D.DPL                                            ( d.dpl\dpl --- d.dpl )
   DUP >R D.DPL>D R>
; \ WJB - 2016/05/16 11:34:54 AM

: D.DPL-INT                                                    ( d1 dpl --- d2 0 )
   \ Returns integer value of D.DPL
   \ MU/MOD
   { | %d1[ #2 CELLS ] %dpl %sign? --- }
   DUP 0= IF EXIT THEN
   to %dpl
   D-SIGN to %sign? %d1[ D!
   %dpl 0>
   IF
      BEGIN
            %d1[ D@ D0= IF #0 #0 #0 EXIT THEN
            %dpl #9 >                     \ check adjustment for greater then 9 digits
      WHILE %d1[ D@
            #9 10^N UDN/MOD %d1[ D!
            DROP
            %dpl #9 - to %dpl
      REPEAT
      \ final adjustment
      %d1[ D@
      %dpl 10^N UDN/MOD %d1[ D!
      DROP
      %d1[ D@
      %sign? IF DNEGATE THEN #0
      EXIT
   THEN
   \ -ve dpl
   %dpl ABS to %dpl                                                            \ yes!
   BEGIN
         %d1[ D@ D0= IF #0 #0 #0 EXIT THEN
         %dpl #9 >                 \ check adjustment for greater then 9 digits
   WHILE %d1[ D@ #9 10^N
         DN* DROP %d1[ D!
         %dpl #9 - to %dpl
   REPEAT
   \ final adjustment
   %d1[ D@ %dpl 10^N DN* DROP
   %sign? IF DNEGATE THEN #0
; \ WJB - 2016/05/16 12:14:42 PM

: UQ10* **                                  ( q1 --- q1x10 )
   { %x0 %x1 %x2 %x3 | %z4 %z3 %z2 %z1 %z0 --- }
   \ long mul :     x3 x2 x1 x0
   \                       * y0
   \               ------------
   \ 1. x0*y0 =           z1 z0
   \ 2. x1*y0 =        z2 z1
   \ 3. x2*y0 =     z3 z2
   \ 4. x3*y0 =  z4 z3
   \           -------------
   \               z3 z2 z1 z0
   \           =============
   \           =============
   \ Step 1 - final z0
   %x0 #10 UM*  to %z1 to %z0
   \ Step 2 - final z1
   %x1 #10 UM* to %z2
   #0  %z1 #0 D+   SWAP to %z1
   #0  %z2 #0 D+   SWAP to %z2
   to %z3
   \ Step 3 - final z2
   %x2 #10 UM*  #0 %z3 #0 D+ to %z4 to %z3
   #0  %z2 #0 D+   SWAP to %z2
   #0  %z3 #0 D+ #0 %z4 #0 D+ DROP to %z4
   to %z3
   \ Step 4 - final z3
   %x3 #10 UM*  #0 %z4 #0 D+ DROP to %z4
   #0  %z3 #0 D+  #0 %z4 #0 D+ DROP to %z4 to %z3
   %z0 %z1 %z2 %z3
; \ WJB - 2016/05/26 11:06:03 AM

: SCALE>Q-D>DPL+                                   ( b1> b2> bresult> --- bool )
	 \ Scale signed double variable to an unsigned quad can have 20 significant
	 \ decimal digits
   { %b1> %b2> %b3> | %mindpl %maxdpl %diff --- bool }
   %b1> bvar>.dpl %b2> bvar>.dpl 2DUP MAX to %maxdpl MIN to %mindpl
   %maxdpl %mindpl - to %diff
   %diff #20 > IF False EXIT THEN
   \
   %b1> bvar>.dpl %mindpl =
   IF   %b1> bvar>.val %diff #9 MIN %diff OVER - to %diff 10^N DN*
        %diff #9 MIN %diff OVER - to %diff 10^N TN*
        %diff 0>
        IF Q-SIGN >R
           %diff 1- to %diff UQ10*
           %diff 0>
           IF %diff 1- to %diff UQ10* THEN
           R> IF QNEGATE THEN
        THEN
        %b2> bvar>.val UP-PRECISION UP-PRECISION
   ELSE %b2> bvar>.val %diff #9 MIN %diff OVER - to %diff 10^N DN*
        %diff #9 MIN %diff OVER - to %diff 10^N TN*
        %diff 0>
        IF Q-SIGN >R
           %diff 1- to %diff UQ10*
           %diff 0>
           IF %diff 1- to %diff UQ10* THEN
           R> IF QNEGATE THEN
        THEN
        %b1> bvar>.val UP-PRECISION UP-PRECISION
   THEN
   Q+ Q>D-ADJUST
   %maxdpl SWAP - %b3> D.DPL!
   True
; \ WJB - 2016/05/18 12:39:53 PM

: BVAR>-D.DPL+ **                                 ( b1> b2> bresult> --- bool )
   { %b1> %b2> %b3> | %mindpl %maxdpl %dpldiff %sign? %scale --- bool }
   %b1> bvar>.val D0=
   IF %b2> %b3> #9 MOVE True EXIT THEN
   %b2> bvar>.val D0=
   IF %b1> %b3> #9 MOVE True EXIT THEN
   %b1> bvar>.dpl %b2> bvar>.dpl =
   IF %b1> bvar>.val %b2> bvar>.val (DD+) T>D-ADJUST
      %b1> bvar>.dpl SWAP - %b3> D.DPL!
      True
      EXIT
   THEN
   \
   %b1> bvar>.dpl %b2> bvar>.dpl 2DUP MAX to %maxdpl MIN to %mindpl
   %maxdpl %mindpl - to %dpldiff
   %dpldiff #1 #9 WITHIN?
   IF  %b1> bvar>.dpl %mindpl =
       IF   %b2> bvar>.val UP-PRECISION %b1> bvar>.val
       ELSE %b1> bvar>.val UP-PRECISION %b2> bvar>.val
       THEN
       %dpldiff 10^N DN*
       T+ T>D-ADJUST %maxdpl SWAP - %b3> D.DPL!
       True
       EXIT
   THEN
   \
   %dpldiff #1 #18 WITHIN?
   IF  %b1> bvar>.dpl %mindpl =
       IF   %b2> bvar>.val UP-PRECISION UP-PRECISION %b1> bvar>.val
       ELSE %b1> bvar>.val UP-PRECISION UP-PRECISION %b2> bvar>.val
       THEN
       D-SIGN to %sign?
       %dpldiff #9 MIN to %scale
       %scale 10^N DN*
       %dpldiff %scale - #9 MIN to %scale
       %scale 10^N UTN*
       Q+ Q>D-ADJUST %maxdpl SWAP - %b3> bvar>.dpl!
       %sign? ?DNEGATE
       %b3> bvar>.val!
       True
       EXIT
   THEN
   \
   %b1> %b2> %b3> SCALE>Q-D>DPL+
   IF True EXIT THEN
   \
   D.DPL-NaN %b3> D.DPL!
   False
; \ WJB - 2016/05/17 01:05:25 PM

: BVAR>-INC-DPL?                            ( bvar> --- bool )
   \ *G Can the dpl of bvar> be increased without loosing value
   \ ** Return true and modified bvar> or false
   { %bvar> | %sign? --- }
   %bvar> bvar>.dpl DPL-VALID? 0=
   IF D.DPL-NaN %bvar> D.DPL! False EXIT THEN
   %bvar> bvar>.val D-SIGN to %sign?
   #10 UMDS* 0=
   IF   DUP 0<
        IF   2DROP
             False
        ELSE %sign? ?DNEGATE %bvar> bvar>.val!
             #1 %bvar> BVAR>.DPL-ADD
             True
        THEN
   ELSE 2DROP
        False
   THEN
; \ WJB - 2016/05/17 02:19:33 PM

: BVAR>-DEC-DPL?                                             ( bvar> --- bool  )
   \ *G Can the dpl of d.dpl value at bvar> be decreased without loosing
   \ ** value. True and adjusted z.dpl in bvar> or false
   { %bvar> | %sign? --- }
   %bvar> bvar>.dpl DPL-VALID? 0=
   IF D.DPL-NaN %bvar> D.DPL! False EXIT THEN
   %bvar> bvar>.val D-Nan D= IF #128 %bvar> bvar>.dpl! False EXIT THEN
   %bvar> bvar>.val D-SIGN to %sign?
   #10 MU/MOD ROT 0=
   IF   %sign? ?DNEGATE %bvar> bvar>.val!
        #1 %bvar> BVAR>.DPL-SUB
        True
   ELSE 2DROP False
   THEN
; \ WJB - 2016/05/18 04:08:36 PM

: SCALE>MAX-D>DPL+ **                              ( b1> b2> bresult> --- bool )
   { %b1> %b2> %b3> --- bool }
   BEGIN
        %b1> %b2> %b3> BVAR>-D.DPL+
        IF True EXIT THEN
        \
        %b1> bvar>.dpl DUP %b2> bvar>.dpl MAX =
        IF   %b1> BVAR>-DEC-DPL? 0=
        ELSE %b2> BVAR>-DEC-DPL? 0=
        THEN
   UNTIL
   False
; \ WJB - 2016/05/18 04:17:09 PM

: SCALE>MIN-D>DPL+ **                              ( b1> b2> bresult> --- bool )
   { %b1> %b2> %b3> --- bool }
   BEGIN
        %b1> %b2> %b3> BVAR>-D.DPL+
        IF True EXIT THEN
        %b1> bvar>.dpl DUP %b2> bvar>.dpl MIN =
        IF   %b1> BVAR>-INC-DPL? 0=
        ELSE %b2> BVAR>-INC-DPL? 0=
        THEN
   UNTIL
   False
; \ WJB - 2016/05/18 10:55:38 AM

: BVAR>-NORMALISE- **                                         ( bvar> --- )
   \ *G Normalise d.dpl at bcar> to negative infinity without loosing value
   \ ** Invalid dpl will returm d.dpl-Nan.
   \ ** d-Nan will returm d.dpl-Nan.
   { %bvar> --- }
   %bvar> bvar>.dpl DPL-VALID? 0=
   IF D.DPL-NaN %bvar> D.DPL! EXIT THEN
   %bvar> bvar>.val D-Nan D= IF #128 %bvar> bvar>.dpl! EXIT THEN
   %bvar> bvar>.dpl 0>
   IF   %bvar> bvar>.val D-SIGN >R                       \ abs value and keep sign
        UD-SCALE-DOWN %bvar> BVAR>.DPL-SUB
        R> ?DNEGATE %bvar> bvar>.val!
   THEN
; \ WJB - 2016/05/18 11:26:10 AM

: SCALE>T-D>DPL+ **                                ( b1> b2> bresult> --- bool )
   { %b1> %b2> %b3> | %mindpl %maxdpl --- bool }
   %b1> %b2> %b3> BVAR>-D.DPL+
   IF True EXIT THEN
   %b1> bvar>.dpl %b2> bvar>.dpl 2DUP MAX to %maxdpl MIN to %mindpl
   %maxdpl %mindpl - #9 > IF False EXIT THEN
   \
   %b1> bvar>.dpl %mindpl =
   IF   %b1> bvar>.val %maxdpl %mindpl - 10^N DN*
        %b2> bvar>.val UP-PRECISION
   ELSE %b2> bvar>.val %maxdpl %mindpl - 10^N DN*
        %b1> bvar>.val UP-PRECISION
   THEN
   T+ T>D-ADJUST
   %maxdpl SWAP - %b3> D.DPL!
   True
; \ WJB - 2016/05/18 12:39:53 PM

: FP-D.DPL+                                            ( x.dpl y.dpl --- z.dpl )
  \ *G Add two d.dpl values tohether. z.dpl=x.dpl+y.dpl.
  \ ** d.dpl=64bit signed value and 8bit signed dpl.
   { | %b1[ #9 ] %b2[ #9 ] %b3[ #9 ] %mindpl %maxdpl %dpldiff %btmp> --- }
   %b2[ D.DPL! %b1[ D.DPL!
   %b1[ %b2[ %b3[ BVAR>-D.DPL+
   IF %b3[ D.DPL@ EXIT THEN
   \
   %b1[ BVAR>-NORMALISE-
   %b1[ %b2[ %b3[ BVAR>-D.DPL+
   IF %b3[ D.DPL@ EXIT THEN
   \
   %b2[ BVAR>-NORMALISE-
   %b1[ %b2[ %b3[ BVAR>-D.DPL+
   IF %b3[ D.DPL@ EXIT THEN
   \
   %b1[ %b2[ %b3[ SCALE>MIN-D>DPL+
   IF %b3[ D.DPL@ EXIT THEN
   \
   %b1[ %b2[ %b3[ SCALE>MAX-D>DPL+
   IF %b3[ D.DPL@ EXIT THEN
   \
   %b1[ %b2[ %b3[ SCALE>T-D>DPL+
   IF %b3[ D.DPL@ EXIT THEN
   \
   %b1[ %b2[ %b3[ SCALE>Q-D>DPL+
   IF %b3[ D.DPL@ EXIT THEN
   \
   %b1[ bvar>.dpl %b2[ bvar>.dpl 2DUP MAX to %maxdpl MIN to %mindpl
   %maxdpl %mindpl - to %dpldiff
   \ Case both dpl negative and value exceeds 64bit
   \ maxdpl negative or zero and mindpl negative and dpl difference exceeds
   \ 64bits
   %maxdpl #1 < %mindpl 0< AND
   %dpldiff #18 > AND
   IF  %b1[ bvar>.dpl %mindpl =
       IF %b1[ D.DPL@ ELSE %b2[ D.DPL@ THEN
       EXIT
   THEN
   \ Minimum dpl negative or zero and maxdpl greater than zero
   %mindpl #1 < %maxdpl 0> AND
   %dpldiff #18 > AND
   IF  %b1[ bvar>.dpl %maxdpl =
       IF  %b1[ ELSE %b2[ THEN
       DUP >R D.DPL@
       %maxdpl %dpldiff #18 - -
       DUP >R D.DPL>D R>
       R> D.DPL!
       %b1[ %b2[ %b3[ BVAR>-D.DPL+
       IF %b3[ D.DPL@ EXIT THEN
   THEN
   \
   BEGIN
          %b1[ bvar>.dpl %b2[ bvar>.dpl 2DUP MAX to %maxdpl MIN to %mindpl
          %b1[ bvar>.val D0<> %b2[ bvar>.val D0<> AND
          %maxdpl %mindpl - #18 > AND
   WHILE  %b1[ bvar>.dpl %maxdpl =
          IF   %b1[ to %btmp>
          ELSE %b2[ to %btmp>
          THEN
          %btmp> bvar>.val #10 DN/ 2DUP %btmp> bvar>.val!
          D0<>
          IF #1 %btmp> BVAR>.DPL-SUB ELSE #0 %btmp> bvar>.dpl! THEN
          %b1[ %b2[ %b3[ BVAR>-D.DPL+
          IF %b3[ D.DPL@ EXIT THEN
   REPEAT
   \
   %b1[ %b2[ %b3[ BVAR>-D.DPL+ DROP
   %b3[ D.DPL@
; \ WJB - 2016/05/19 05:13:21 PM

: D.DPL+                                                        ( b1 b2 --- b3 )
   FP-D.DPL+
;

[then]
\ *H D.DPL Sign?
: BVAR>-<SIGN?>                                              ( bvar> --- bool )
   \
   DUP >R D.DPL@ D.DPL-SIGN R> SWAP >R D.DPL! R> ;

\ *H D.DPL Round
: BVAR>-(DPL-ROUND) **                                        ( dpl brem> bvar> --- )
   \ #G Unsigned dpl round to bvar> and brem>
   { %dpl %brem> %bvar> | %diff --- }
   %brem> #9 ERASE
   %bvar> bvar>.dpl %dpl - ABS #9 MIN to %diff
   BEGIN
         %diff 0>
   WHILE %bvar> bvar>.val %diff 10^N UDN/MOD %bvar> bvar>.val!
         #0 %bvar> bvar>.dpl %brem> D.DPL@ D.DPL+ %brem> D.DPL!
         \
         %bvar> bvar>.dpl %diff - %bvar> bvar>.dpl!
         %bvar> bvar>.dpl %dpl  - #9 MIN to %diff
   REPEAT
   %dpl %bvar> bvar>.dpl!
;

: BVAR>.DPL-ROUND                                              ( dpl bvar> --- )
   \ *G Roud d.dpl up using dpl + 1 digit. Truncate the rest of the value
   \ ** Excel ROUND(argument,param) function 4/5 rule. Param can be positive or
   \ ** negative. Argument must a valid d.dpl. Param must be a valid dpl.
   \ Case 1 = bvar>.dpl 0<  dpl 0>= AND
   \ Case 2 = bvar>.dpl 0>= dpl 0>= AND
   \ Case 3 = bvar>.dpl 0>= dpl 0<  AND
   \ Case 4 = bvar>.dpl 0<  dpl 0<  AND
	 \
   { %dpl %bvar> | %sign? %diff %brem[ #9 ] --- }
   %dpl DPL-VALID? 0=              IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> D.DPL@ D.DPL-NaN?            IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> bvar>.dpl DPL-VALID? 0=  IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> bvar>.val D0=            IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> BVAR>-NORMALISE-
   %dpl %bvar> bvar>.dpl = IF EXIT THEN
   \ Case 1
   \ Negative dpl for value and positive for dpl
   %bvar> bvar>.dpl 0< %dpl #0 >= AND
   IF EXIT THEN
   \ Case 2
   \ both required dpl and value dpl greater or equal to zero
   \ %bval.dpl greater than dpl
   %dpl #0 >= %bvar> bvar>.dpl #0 >= AND
   IF %dpl %bvar> bvar>.dpl > IF EXIT THEN
      %bvar> bvar>.dpl %dpl - ABS #19 > IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN  \ cannot have any value
      %bvar> BVAR>-<SIGN?> to %sign?
      %dpl %brem[ %bvar> BVAR>-(DPL-ROUND)
      %brem[ D.DPL@ %dpl D.DPL>D.DPL %brem[ D.DPL!
      %brem[ D.DPL@ %bvar> D.DPL@ D.DPL+ %bvar> D.DPL!
      %sign?
      IF %bvar> bvar>.val DNEGATE %bvar> bvar>.val! THEN
      EXIT
   THEN
   \ Case 3
   \ Negative param for round and positive dpl for value
   %dpl 0< %bvar> bvar>.dpl #0 >= AND
   IF %bvar> BVAR>-<SIGN?> to %sign?
      #0 %brem[ %bvar> BVAR>-(DPL-ROUND) \ get integer
      %bvar> bvar>.dpl %dpl - ABS #19 > IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN  \ cannot have any value
      %bvar> bvar>.val D0= IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
      %dpl %brem[ %bvar> BVAR>-(DPL-ROUND)
      %brem[ D.DPL@ %dpl D.DPL>D.DPL %brem[ D.DPL!
      %brem[ D.DPL@ %bvar> D.DPL@ D.DPL+ %bvar> D.DPL!
      %sign?
      IF %bvar> bvar>.val DNEGATE %bvar> bvar>.val! THEN
      EXIT
   THEN
   \ Case 4
   \ Negative param for round and negative value dpl
   %dpl 0< %bvar> bvar>.dpl 0< AND
   IF %bvar> bvar>.dpl ABS %dpl ABS >=  IF EXIT THEN
      %bvar> bvar>.dpl %dpl - ABS #19 > IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN  \ cannot have any value
      %bvar> BVAR>-<SIGN?> to %sign?
      %dpl %brem[ %bvar> BVAR>-(DPL-ROUND)
      %brem[ D.DPL@ %dpl D.DPL>D.DPL %brem[ D.DPL!
      %brem[ D.DPL@ %bvar> D.DPL@ D.DPL+ %bvar> D.DPL!
      %sign?
      IF %bvar> bvar>.val DNEGATE %bvar> bvar>.val! THEN
   THEN
; \ WJB 2016/07/19 10:46:21 AM

: BVAR>.DPL-ROUNDUP                                            ( dpl bvar> --- )
   \ Case 1 = bvar>.dpl 0<  dpl 0>= AND
   \ Case 2 = bvar>.dpl 0>= dpl 0>= AND
   \ Case 3 = bvar>.dpl 0>= dpl 0<  AND
   \ Case 4 = bvar>.dpl 0<  dpl 0<  AND
   { %dpl %bvar> | %sign? %diff %brem[ #9 ] --- }
   %dpl DPL-VALID? 0=              IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> D.DPL@ D.DPL-NaN?            IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> bvar>.dpl DPL-VALID? 0=  IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> bvar>.val D0=            IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> BVAR>-NORMALISE-
   %dpl %bvar> bvar>.dpl = IF EXIT THEN
   %dpl #0 >= %bvar> bvar>.dpl #0 >= AND
   IF %dpl %bvar> bvar>.dpl > IF EXIT THEN THEN
   \ Case 1
   \ Negative dpl for value and positive for dpl then use unchanged value
   %bvar> bvar>.dpl 0< %dpl #0 >= AND
   IF EXIT THEN
   \ Case 2
   \ both required dpl and value dpl greater or equal to zero
   \ %bval.dpl greater than dpl
   %dpl #0 >= %bvar> bvar>.dpl #0 >= AND
   IF %dpl %bvar> bvar>.dpl > IF EXIT THEN
      %bvar> bvar>.dpl %dpl - ABS #19 > IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN  \ cannot have any value
      %bvar> BVAR>-<SIGN?> to %sign?
      %dpl %brem[ %bvar> BVAR>-(DPL-ROUND)
      %brem[ D.DPL@ D.DPL?
      IF #1 #0 %dpl %bvar> D.DPL@ D.DPL+ %bvar> D.DPL! THEN
      %sign?
      IF %bvar> bvar>.val DNEGATE %bvar> bvar>.val! THEN
      EXIT
   THEN
   \ Case 3
   \ Negative param for round and positive dpl for value
   %dpl 0< %bvar> bvar>.dpl #0 >= AND
   IF %bvar> bvar>.dpl %dpl - ABS #19 > IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN  \ cannot have any value
      %bvar> BVAR>-<SIGN?> to %sign?
      %dpl %brem[ %bvar> BVAR>-(DPL-ROUND)
      %brem[ D.DPL@ D.DPL?
      IF #1 #0 %dpl %bvar> D.DPL@ D.DPL+ %bvar> D.DPL! THEN
      %sign?
      IF %bvar> bvar>.val DNEGATE %bvar> bvar>.val! THEN
      EXIT
   THEN
   \ Case 4
   \ Negative param for round and negative value dpl
   %dpl 0< %bvar> bvar>.dpl 0< AND
   IF %bvar> bvar>.dpl ABS %dpl ABS >=  IF EXIT THEN
      %bvar> bvar>.dpl %dpl - ABS #19 > IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN  \ cannot have any value
      %bvar> BVAR>-<SIGN?> to %sign?
      %dpl %brem[ %bvar> BVAR>-(DPL-ROUND)
      %brem[ D.DPL@ D.DPL?
      IF #1 #0 %dpl %bvar> D.DPL@ D.DPL+ %bvar> D.DPL! THEN
      %sign?
      IF %bvar> bvar>.val DNEGATE %bvar> bvar>.val! THEN
   THEN

; \ WJB 2016/07/19 10:47:13 AM

: BVAR>.DPL-ROUNDDOWN                                            ( dpl bvar> --- )
   \ *G Roud down d.dpl up using dpl. Truncate the rest of the value.
   \ ** Excel ROUNDDOWN(argument,param) function. Param can be positive or
   \ ** negative. Argument must a valid d.dpl
   \ Case 1 = %bval.dpl 0<  %dpl 0>= AND
   \ Case 2 = %bval.dpl 0>= %dpl 0>= AND
   \ Case 3 = %bval.dpl 0>= %dpl 0<  AND
   \ Case 4 = %bval.dpl 0<  %dpl 0<  AND
   { %dpl %bvar> | %sign? %diff %brem[ #9 ] --- }
   \ Firewall incoming parameters
   %dpl DPL-VALID? 0=              IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> D.DPL@ D.DPL-NaN?            IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> bvar>.dpl DPL-VALID? 0=  IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> bvar>.val D0=            IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
   %bvar> BVAR>-NORMALISE-
   %dpl %bvar> bvar>.dpl = IF EXIT THEN
   \ Case 1
   \ Negative dpl for value and positive for dpl
   %bvar> bvar>.dpl 0< %dpl #0 >= AND
   IF EXIT THEN
   %bvar> bvar>.dpl %dpl - ABS #19 > IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN  \ cannot have any value
   \ Case 2
   \ both required dpl and value dpl greater or equal to zero
   \ %bval.dpl greater than dpl
   %dpl #0 >= %bvar> bvar>.dpl #0 >= AND
   IF %dpl %bvar> bvar>.dpl > IF EXIT THEN
      %bvar> BVAR>-<SIGN?> to %sign?
      %dpl %brem[ %bvar> BVAR>-(DPL-ROUND)
      %sign?
      IF %bvar> bvar>.val DNEGATE %bvar> bvar>.val! THEN
      EXIT
   THEN
   \ Case 3
   \ Negative param for round and positive dpl for value
   %dpl 0< %bvar> bvar>.dpl #0 >= AND
   IF %bvar> BVAR>-<SIGN?> to %sign?
      #0 %brem[ %bvar> BVAR>-(DPL-ROUND) \ get integer
      %bvar> bvar>.val D0= IF #0 #0 %dpl %bvar> D.DPL! EXIT THEN
      %dpl %brem[ %bvar> BVAR>-(DPL-ROUND)
      %sign?
      IF %bvar> bvar>.val DNEGATE %bvar> bvar>.val! THEN
      EXIT
   THEN
   \ Case 4
   \ Negative param for round and negative value dpl
   %dpl 0< %bvar> bvar>.dpl 0< AND
   IF %bvar> bvar>.dpl ABS %dpl ABS >= IF EXIT THEN
      %bvar> BVAR>-<SIGN?> to %sign?
      %dpl %brem[ %bvar> BVAR>-(DPL-ROUND)
      %sign?
      IF %bvar> bvar>.val DNEGATE %bvar> bvar>.val! THEN
   THEN
; \ WJB - 2016/07/19 12:32:40 PM

\ *********************
\ +S Protected division
\ *********************
\ +P Protected division can be used for two reasons
\ +( 1
\ +B Avoiding problems caused by unused data cells
\ +* being set to zero.
\ +B Avoiding overflow exceptions
\ +)

\ +P In many applications, the first  case can be handled by
\ +* testing for a zero divisor and returning zero for quotient
\ +* and/or remainder. The second case is almost always an error.

((
: */modP  \ n1 n2 n3 -- rem quot ; protected version
  -rot m* rot                   \ -- d n3
  over abs
  over abs 1 rshift  >= if      \ |dh| >= |n3| on error
    drop 2drop 0 0
  else
    sm/rem
  endif
;
))

\ ======
\ *> ###
\ ======

decimal

