\ BUFFER.FTH - memory buffer I/O device

((
Copyright (c) 2005, 2006
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20071106 SFP004 Fixed TB-#FREE.
20061207 MPE003 Overhauled. Added ReadEx and checked more actions.
20050823 MPE002 Added SizedTextBuff.
20050801 MPE001 Changed and documented SETPOS action.
))


\ ========
\ *> GENIO
\ ========
\ *N Memory Buffer Device
\ *P This Generic IO Device uses blocks of memory for input
\ ** and/or output. The source code is in *\i{LIB\GENIO\Buffer.fth}.
\ ** This is *\b{not} a circular buffer system. After a buffer has
\ ** been used, the read and write pointers are *\b{not} reset. You
\ ** must reset the buffer pointers using the IOCTL functions.

\ *P As of VFX Forth build 2380, this code has been overhauled. If
\ ** this causes you problems, the file *\i{BUFFER.old.FTH} contains
\ ** the previous (but now unsupported) version. The major changes
\ ** are:
\ *( 1
\ *B More error checking.
\ *B The *\fo{KEY} operation (and hence *\fo{ACCEPT} blocks if no
\ ** data is available.
\ *B The *\fo{KEY?} operation returns the number of unread bytes.
\ *B The *\fo{EMIT?} operation returns the number of unwritten bytes
\ ** in the buffer..
\ *B The close operation is protected if the device is already
\ ** closed.
\ *B The GENIO *\fo{ReadEx} function is implemented.
\ *B The *\fo{Textbuff-sid} GENIO structure has been documented.
\ *B IOCTL functions have been added. See later for details.
\ *)

\ *P Note that the *\fo{ACCEPT} operation does *\b{not} echo. It is designed
\ ** for extracting lines from saved input.

\ *P In order to create a device use the *\fo{TEXTBUFF:} definition given
\ ** later. *\fo{TEXTBUFF:} is compatible with ProForth 2.

\ *P When opening a memory device the parameters to *\fo{OPEN-GEN}
\ ** have the following meaning:
\ *D ADDR       Address of memory to use as buffer or ignored if
\ **            dynamic allocation is required.
\ *D LEN        The maximum length of the memory image.
\ *D ATTRIBS    When zero the ADDR parameter is ignored and LEN
\ **            bytes of memory are allocated from the heap.

MODULE GENIO-BUFFER

struct textbuff-sid     \ -- len
\ *G Defines the length of a SID for a text buffer device.
\ *[
  gen-sid +		\ handle=buffer, reuse field names of GEN-SID
  int tb-len            \ length of buffer
  int tb-attribs        \ attributes
  int tb-wr             \ address to write to, set by SETPOS, WRITE
  int tb-rd             \ address to read from set by SETPOS, READ
end-struct
\ *]

: reset-tb	\ buffer sid -- ; set pointers, clear buffer
  2dup gen-handle !               \ handle=address of buffer
  2dup tb-wr !                    \ write pointer
  2dup tb-rd !                    \ read pointer
  tb-len @ erase                  \ wipe buffer
;

: tb-#read	\ sid -- #bytes ; number of bytes read
  dup tb-rd @  swap gen-handle @ - \ number of bytes read from buffer
;

: tb-#written	\ sid -- #bytes ; number of bytes written
  dup tb-wr @  swap gen-handle @ - \ number of bytes written to buffer
;

: tb-#avail	\ sid -- #avail
\ number of bytes available to be read
  dup tb-wr @ swap tb-rd @ - 0 max
;

: tb-#free	\ sid -- #free
\ number of bytes which can be written
  dup gen-handle @ over tb-len @ +	\ end of buffer ; SFP004
  swap tb-wr @ - 0 max			\ number left
;

: open-tb	\ addr len attribs sid -- buffer ior
  >r					\ save SID
  -1 r@ gen-handle !			\ assume failure
  dup r@ tb-attribs !			\ set required attribs
  over r@ tb-len !			\ set required size
  case					\ -- addr len attribs ; look at attribs
    0 of  nip  allocate dup 0=		\ get required memory, check result
          if  over r@ reset-tb  endif
      endof
          2drop				\ discard attribs and len
          dup r@ reset-tb		\ set buffer addresses
          0				\ success
  end-case
  r> drop
;

: close-tb	\ sid -- ior
  dup gen-handle @ -1 =			\ check if already closed
  if  drop 0  exit  endif
  dup tb-attribs @ 0= if		\ if attributes are 0, free off buffer
    gen-handle @on free			\ free off the buffer, handle := -1
  else
    -1 swap gen-handle !		\ mark unused
    0					\ success
  endif
;

: (read-tb)	\ addr len sid -- ; read len bytes to addr
  dup >r tb-rd @ -rot			\ -- src dest len ; R: -- sid ; start of read
  dup r> tb-rd +!			\ new SEEK position = len+POS
  move					\ copy data
;

: read-tb	\ addr len sid -- ior
  (read-tb) 0
;

: ReadEx-tb	\ addr len sid -- #read ior
  tuck tb-#avail min dup >r swap	\ -- addr len' sid ; R: -- len'
  (read-tb)
   r> 0
;

: write-tb	\ addr len sid -- ior
  >r
  r@ tb-wr @ swap			\ -- addr dest len ; start of write
  dup r@ tb-wr +!			\ update write pointer
  move					\ copy data
  0 r> tb-wr @ c!			\ write 0 terminator
  0
;

: key-tb	\ sid -- char
  begin  dup tb-#avail 0= while  pause  repeat
  dup tb-rd @ c@  swap tb-rd incr
;

: key?-tb	\ sid -- u
  tb-#avail  ;

: ekey-tb	\ sid -- echar
  key-tb  ;

: ekey?-tb	\ sid -- flag
  key?-tb  ;

: accept-tb	\ addr len sid -- len'
  0 swap 2swap                          \ -- 0 sid addr len
  bounds ?do				\ -- cnt sid
    dup key?-tb 0= ?leave		\ get out if no character
    dup key-tb dup $0D = if		\ if carriage return
      drop  over key-tb drop		\ discard CR and following LF
      leave                             \ leave
    endif
    i c!  swap 1+ swap                  \ save character and count it
  loop
  drop                                  \ discard SID
;

: (emit-tb)	\ char sid --
  dup >r tb-wr @ tuck c!		\ write character
  0 swap 1 + c!				\ write terminator
  r> tb-wr incr				\ increment pointer
;

: bs-tb		\ sid -- ; destructive on display devices
  out @ if
    dup tb-wr decr  0 swap tb-wr @ c!	\ back up
    out decr
  else
    drop
  endif
;

: emit-tb	\ char sid --
  >r
  case
    $08 of  r@ bs-tb  endof		\ destructive backspace
    $0A of  $0A r@ (emit-tb)		\ LF
            op-line# incr
        endof
    $0D of  $0D r@ (emit-tb)		\ CR
            out off
        endof
          r@ (emit-tb)
          out incr
  end-case
  r> drop
;

((
: emit?-tb	\ sid -- flag
  dup tb-#read                    \ number of bytes in buffer
  swap tb-len @ u<  ;
))
: emit?-tb	\ sid -- ior
  tb-#free
;

: type-tb	\ addr len sid --
  -rot  bounds
  ?do  i c@ over emit-tb  loop
  drop  ;

: cr-tb		\ sid --
  $0D over emit-tb  $0A swap emit-tb  ;

: lf-tb		\ sid --
  $0A swap emit-tb  ;

: ff-tb		\ sid -- ; page/cls on display devices
\ *G This word is run by *\fo{PAGE}, *\fo{CLS} and *\fo{FF-GEN}. It resets
\ ** (empties) the buffer.
  dup gen-handle @ swap reset-tb  ;

: bell-tb	\ sid -- ; audible beeper
  drop  ;

: xypos-tb	\ x y sid -- addr ; return address in buffer of x,y
  { x y sid -- addr }
  sid gen-handle @  sid tb-#written	\ buffer addr and length
  y 0 ?do				\ number of lines
    $0D scan dup                        \ look for CR
    if  2 /string  endif                \ if found, step over
  loop
  drop  x +                             \ discard len, add x
;

: setpos-tb	\ x y mode sid -- ior
\ *G This word is run by *\fo{SETPOS-GEN}. Mode controls the x and y
\ ** input values as follows.
\ *D 0   x = #bytes written, y is ignored
\ *D -1  x=col, y=line for next character to be written
\ *D -2  x = #bytes read, y is ignored
\ *D -3  x=col, y=line for next character to be read
  >r
  case
     0 of  drop  r@ gen-handle @ + tb-wr !  0  endof
    -1 of  r@ xypos-tb r@ tb-wr !     0  endof
    -2 of  drop  r@ gen-handle @ + tb-rd !  0  endof
    -3 of  r@ xypos-tb r@ tb-rd !     0  endof
           drop  2drop -1               \ error
  end-case
  r> drop
;

[-sin
: xypos		\ buffer len -- x y
  0 0  2swap bounds ?do			\ -- x y
    i c@ #13 = if			\ look for CR
      nip 0 swap			\ reset X to 0
      1+				\ increment Y
      i 1 + c@ #10 =  1 and 1+		\ check for LF, return 2 for LF, 1 for not
    else
      swap 1+ swap			\ increment X
      1					\ step over one character
    endif
  +loop
;
sin]

: getpos-tb	\ mode sid -- x y ior
\ *G This word is run by *\fo{GETPOS-GEN}. Mode controls the x and y
\ ** return values as follows.
\ *D 0   x = #bytes written, y = 0
\ *D -1  x,y for next character to be written
\ *D -2  x = #bytes read, y = 0
\ *D -3  x,y for next character to be read
\ *D -4  x = addr, y = len of unread data
\ *D -5  x = base address, y = size of data area
  >r
  case
     0 of  r@ tb-#written 0  0  endof
    -1 of  r@ gen-handle @ r@ tb-#written xypos  0  endof
    -2 of  r@ tb-#read 0  0  endof
    -3 of  r@ gen-handle @ r@ tb-#read xypos  0  endof
    -4 of  r@ tb-rd @  r@ tb-wr @ over -  0  endof
    -5 of  r@ gen-handle @  r@ tb-len @  0  endof
           drop  0 0  -1
  end-case
  r> drop  ;

: ioctl-tb	\ addr len fn sid -- ior
\ *G This word is run by *\fo{IOCTL-GEN} and *\fo{IOCTL-GIO}.
\ ** *\i{Fn} controls the meaning of *\i{addr}, *\i{len} and the
\ ** *\i{ior} return value as follows:
\ *D 0   Get buffer address: *\i{addr}=0, *\i{len}=0, *\i{ior}=addr.
\ *D -1  Set write pointer: *\i{addr}=0, *\i{len}=offset, *\i{ior}=0.
\ *D -2  Get write pointer: *\i{addr}=0, *\i{len}=0, *\i{ior}=offset.
\ *D -3  Set read pointer: *\i{addr}=0, *\i{len}=offset, *\i{ior}=0.
\ *D -4  Get read pointer: *\i{addr}=0, *\i{len}=0, *\i{ior}=offset.
  >r
  case
     0 of  2drop  r@ gen-handle @  endof
    -1 of  nip  r@ gen-handle @ + r@ tb-wr !  0  endof
    -2 of  2drop  r@ tb-wr @ r@ gen-handle @ -  endof
    -3 of  nip  r@ gen-handle @ + r@ tb-rd !  0  endof
    -4 of  2drop  r@ tb-rd @ r@ gen-handle @ -  endof
           drop 2drop  -1
  end-case
  r> drop
;

: flushOP-tb	\ sid -- ior
  drop 0  ;

: RFU-tb	\ --
  1 abort" RFU not supported"  ;

: init-tb	\ addr len sid -- ior
  3drop  0  ;

: term-tb	\ sid -- ior
  drop  0  ;

: config-tb	\ sid -- ior ; produces a dialog
  drop  0  ;

create tb-vectors	\ -- addr
  ' Open-tb ,
  ' Close-tb ,
  ' Read-tb ,
  ' Write-tb ,
  ' Key-tb ,
  ' Key?-tb ,
  ' Ekey-tb ,
  ' EKey?-tb ,
  ' Accept-tb ,
  ' Emit-tb ,
  ' Emit?-tb ,
  ' Type-tb ,
  ' CR-tb ,
  ' LF-tb ,
  ' FF-tb ,
  ' BS-tb ,
  ' Bell-tb ,
  ' setpos-tb ,
  ' getpos-tb ,
  ' Ioctl-tb ,
  ' FlushOP-tb  ,
  ' ReadEx-tb ,
  ' Init-tb ,
  ' Term-tb ,
  ' Config-tb ,


\ ==================
\ *H Device Creation
\ ==================

: initTextBuffSid	\ addr --
\ *G Initialise a previously allocated SID for
\ ** a text buffer to the default values.
  dup textbuff-sid erase
  -1 over gen-handle !
  tb-vectors over gen-vector !
  $4000 swap tb-len !
; doNotSin

: textbuff:	\ "name" -- ; Exec: -- sid
\ *G Create a memory buffer called *\i{name}.
  create
    here  textbuff-sid allot  initTextBuffSid
\    -1 ,  textbuff-vectors ,  $04000 ,  \ handle, vector, default size
\    0 ,  0 ,  0 ,                       \ attribs, write pointer, read pointer
;

\ MPE002...
: SizedTextBuff	\ size -- sid|0
\ *G Allocates and opens a SID with a buffer of size *\i{size}
\ ** bytes, and returns the SID on success or 0 on failure.
  textbuff-sid allocate			\ -- size sid flag
  if  2drop 0  exit  endif
  dup >r initTextBuffSid		\ -- size ; R: -- sid
  0 swap 0 r@ open-tb nip		\ -- flag ; R: -- sid
  if  drop 0  r> free drop  exit  endif
  r>					\ -- sid
; doNotSin

: AllocTextBuff	\ -- sid|0
\ *G Allocates and opens a SID with a default 16kb text buffer
\ ** and returns the SID on success or 0 on failure.
  $4000 SizedTextBuff
; doNotSin
\ ...MPE002

: FreeTextBuff	\ sid --
\ *G Closes the SID and frees memory allocated by *\fo{AllocTextBuff}
\ ** or *\fo{SizedTextBuff}.
  dup close-tb drop  free drop
; doNotSin

export textbuff-sid
export initTextBuffSid
EXPORT textbuff:
export AllocTextBuff
export SizedTextBuff
export FreeTextBuff

END-MODULE


\ ======
\ *> ###
\ ======

