\ Lib\Genio\FileBuff.fth - Buffered Generic I/O file device

((
Copyright (c) 2003, 2006, 2007, 2008, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====
AddIOCTL to control OUT handling, CREATE/OPEN behaviour, and buffering.

Change history
==============
20100816 SFP005 Converted from file.fth. Buffered the output.
20080416 SFP004 Replaced RFU function with READEX function.
20071106 SFP003 Added handling of OUT to EMIT/TYPE.
20060118 SFP002 Removed scratch cells from SID.
		Used common "not implemented" abort.
		Added initFileBuff ( sid -- ).
		Added /FileDev ( -- len ).
20030402 SFP001 Use generic I/O handle location.
))

\ =======================================================================
\ *> GENIO
\ *N Buffered File Device
\ =======================
\ *P This Generic IO Device operates on a disk file for input
\ ** and/or output. The source code can be found in
\ ** *\i{Lib\Genio\FileBuff.fth}. Output is buffered.
\ ** This device should not be used when the speed
\ ** of small reads is important.

\ *P In order to create a device use the *\fo{FileBuff:} definition
\ ** given later.

\ *P When opening a file device the parameters to OPEN-GEN have
\ ** the following meaning:
( *E    ADDR            Address of string for filename.                 )
( **    LEN             Length of string at ADDR.                       )
( **    ATTRIBS         Open flags. These match the ANS r/o r/w etc.    )

\ *P The ReadEx function is now implemented.
\ =======================================================================

MODULE GIO-FILEBUFF

struct /FileBuff	\ -- len ; defines the SID of a file device
  gen-sid +		\ reuse field names of GEN-SID
  int fb./Obuff				\ size of output buffer
  int fb.#oBuff				\ number of chars in output buffer
  int fb.oBuff				\ address of output buffer
  int fb./Ibuff				\ size of input buffer
  int fb.#IBuff				\ number of chars in input buffer
  int fb.IBuff				\ address of input buffer
end-struct

: FileBuff-flush	\ sid -- ior
  dup fb.#oBuff 2@ dup if		\ -- sid caddr len
    rot dup >r gen-handle @ write-file	\ -- ior ; R: -- sid
    r> fb.#oBuff off
  else
    nip nip
  endif
; doNotSin

: >zaddr                \ addr len zaddr -- ; Copy string as 0 terminated
  2dup + 0 swap c!                      \ store 0 terminator
  swap move                             \ and move string data
;

: FileBuff-alloc	\ sid ior -- sid ior'
\ Allocate buffers if sid=0.
  dup 0= if
    over >r
    #512 dup r@ fb./OBuff !		\ 512 byte output buffer
    0 r@ fb.#Obuff !			\ nothing in buffer
    allocate if				\ -- sid ior x
      2drop -2				\ -- sid -2 ; allocation failed
      0 r@ fb.oBuff !
      r@ gen-handle @ close-file drop
    else
      r@ fb.oBuff !			\ buffer address
    endif
    r> drop
  endif
; doNotSin

: FileBuff-free	\ sid --
\ Delete I/O buffers
  dup fb.Obuff @ ?dup
  if  free drop  dup fb.Obuff off  endif
  drop
; doNotSin

[defined] Target_386_Windows [if]
: FileBuff-open          { caddr len fam sid | zname[ MAX_PATH ] -- sid ior }
  caddr len zname[ >zaddr               \ Make 0 terminated name string
  zname[ fam NULL NULL                  \ open or create file
  OPEN_ALWAYS NULL NULL CreateFile
  dup sid gen-handle !			\ store handle ; SFP001
  INVALID_HANDLE_VALUE =                \ set ior
  sid swap
  FileBuff-alloc			\ allocate I/O buffer(s)
;
[else]
: FileBuff-open	\ caddr len fam sid -- sid ior
  >r  3dup open-file if			\ -- caddr len fam handle : R: -- sid
    drop  create-file if		\ -- handle ; R: -- sid
      drop  r> -1
    else
      r> tuck gen-handle !  0
    endif
  else
    nip nip nip				\ -- handle ; R: -- sid
    r> tuck gen-handle !  0
  endif
  FileBuff-alloc				\ allocate I/O buffer(s)
;
[then]

: FileBuff-close         \ sid -- ior
  dup FileBuff-flush drop
  dup FileBuff-free			\ delete I/O buffers
  dup gen-handle @ >r			\ get handle and keep
  INVALID_HANDLE_VALUE swap             \ write default (INVALID) back
  gen-handle !				\ SFP001
  r> Close-File
;

: FileBuff-write         \ addr len sid -- ior
  over >r
  dup fb.#oBuff @ r> + over fb./oBuff @ > if
    dup FileBuff-flush drop
    gen-handle @ write-file
  else
    >r tuck				\ -- len addr len ; -- sid
    r@ fb.#Obuff 2@ + swap cmove	\ add to buffer
    r> fb.#oBuff +!			\ update content length
    0					\ success
  endif
; doNotSin

: filebuff-emit	\ char sid --
  { | char[ cell ] -- }
  swap char[ c!
  char[ 1 rot filebuff-write drop
  1 out +!
; doNotSin

: FileBuff-read          \ addr len sid -- ior
  dup FileBuff-flush drop
  gen-handle @ read-file nip
;

: FileBuff-key	\ sid -- char
  { | temp[ cell ] -- }
  temp[ 1 rot FileBuff-read drop
  temp[ c@
;

: FileBuff-key?          \ sid -- flag
  dup FileBuff-flush drop
  gen-handle @ dup >r			\ SFP001
  file-size drop			\ -- ud1
  r> file-position drop			\ -- ud1 ud2
  d= 0=
;

: NotFileBuff	\ --
\ Issue a FileBuff error message
  1 abort" operation not supported for FILE DEVICE"
;

: FileBuff-ekey          \ sid -- echar
  NotFileBuff	\ 1 abort" EKEY not supported for FileBuff"
;

: FileBuff-ekey?         \ sid -- flag
  NotFileBuff	\ 1 abort" EKEY? not supported for FileBuff"
;

: FileBuff-accept        \ addr len sid -- len'
  NotFileBuff	\ 1 abort" ACCEPT not supported for FileBuff"
;

: FileBuff-emit?         \ sid -- flag
  drop TRUE
;

: FileBuff-type          \ addr len sid --
  over out +!
  FileBuff-write drop
;

: FileBuff-cr            \ sid --
  #13 over FileBuff-emit
  #10 over FileBuff-emit
  drop
  out off  op-line# incr
;

: FileBuff-lf            \ sid --
  #10 swap FileBuff-emit  -1 out +!	\ SFP003
;

: FileBuff-ff            \ sid -- ; page/cls on display devices
  #12 swap FileBuff-emit
;

: FileBuff-bs            \ sid -- ; destructive on display devices
  #8 swap FileBuff-emit
;

: FileBuff-bell          \ sid -- ; audible beeper
  NotFileBuff	\ 1 abort" BELL not supported for FileBuff"
;

: FileBuff-setpos        \ x y mode sid -- ior
  NotFileBuff	\ 1 abort" SETPOS not supported for FileBuff"
;

: FileBuff-getpos        \ mode sid -- x y ior
  NotFileBuff	\ 1 abort" GETPOS not supported for FileBuff"
;

: FileBuff-ioctl         \ addr len fn sid -- ior
  NotFileBuff	\ 1 abort" IOCTL not supported for FileBuff"
;

: FileBuff-readex	\ addr len sid -- u2 ior ; SFP004
  gen-handle @ read-file
;

: FileBuff-init          \ addr len sid -- ior
  3drop 0	\ 1 abort" INIT not supported for FileBuff"
;

: FileBuff-term          \ sid -- ior
  drop 0	\ 1 abort" TERM not supported for FileBuff"
;

: FileBuff-config        \ sid -- ior ; produces a dialog
  drop 0	\ 1 abort" CONFIG not supported for FileBuff"
;

create FileBuff-vectors
  ' FileBuff-open ,
  ' FileBuff-close ,
  ' FileBuff-read ,
  ' FileBuff-write ,
  ' FileBuff-Key ,
  ' FileBuff-Key? ,
  ' FileBuff-Ekey ,
  ' FileBuff-EKey? ,
  ' FileBuff-Accept ,
  ' FileBuff-Emit ,
  ' FileBuff-Emit? ,
  ' FileBuff-Type ,
  ' FileBuff-CR ,
  ' FileBuff-LF ,
  ' FileBuff-FF ,
  ' FileBuff-BS ,
  ' FileBuff-Bell ,
  ' FileBuff-setpos ,
  ' FileBuff-Getpos ,
  ' FileBuff-Ioctl ,
  ' FileBuff-Flush  ,
  ' FileBuff-readex ,
  ' FileBuff-Init ,
  ' FileBuff-Term ,
  ' FileBuff-Config ,


\ ==================
\ *H Device Creation
\ ==================

(( for DocGen
struct /FileBuff	\ -- len
\ *G Returns the size of the sid structure for a file device.
))

: initFileBuff	\ sid --
\ *G Initialise the sid for a file device. Mostly used when
\ ** the structure has been allocated from the heap.
  dup /FileBuff erase
  FileBuff-vectors -1 rot 2!
;

: FileBuff:	\ "name" -- ; Exec: -- sid
\ *G Create a File based Generic IO device in the dictionary.
  create
    here  /FileBuff allot  initFileBuff
;

Export /FileBuff
Export initFileBuff
EXPORT FileBuff:

END-MODULE


\ ======
\ *> ###
\ ======

