\ Lib/HeapBuff.fth - dynamically sized buffers

((
Copyright (c) 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *> library
\ *S Dynamically sized buffers
\ ===========
\ *P The file *\i{Lib/HeapBuff.fth} provides facilities for buffers
\ ** that can be resized as required.

variable HeapBuffChain	\ -- addr
\ *G Holds the address of the last dynamic buffer.
  0 HeapBuffChain !

struct /HeapBuff		\ -- len
\ *G Structure controlling a dynamic buffer.
\ *[
  int db.link		\ pointer to previous buffer
  int db.req		\ initial requested size
  int db.size		\ current buffer size
  int db.used		\ offset to next free data
  int db.addr		\ address of data buffer - MUST BE AFTER db.used
end-struct
\ *]

: (HeapBuff)	\ u -- struct
\ *G Create a buffer of the given size and its controlling structure.
  /HeapBuff ireserve 			\ -- u db
  dup HeapBuffChain addLink		\ add to chain
  2dup db.req !				\ requested size
  tuck db.size !			\ current size
  0 over db.used !			\ none used
  dup db.size @ allocate		\ -- db addr ; reserve space
  if  drop 0  then			\ on error, force addr=0
  swap db.addr !			\ save address
;

: initHeapBuffs	\ --
\ *G Initialise all buffers.
  HeapBuffChain @ begin
    dup while
    dup initHeapBuff  db.link @
  repeat
  drop
;
' initHeapBuffs atCold

: termHeapBuffs	\ --
\ *G Release all buffers.
  HeapBuffChain @ begin
    dup while
    dup db.addr @ free drop  db.link @
  repeat
  drop
;
' termHeapBuffs atExit


\ ===========================
\ *N Statically sized buffers
\ ===========================
\ *P These buffers are created with a fixed size that is allocated
\ ** in the cold chain. They can be resized, but it has to be
\ ** done manually.

: HeapBuff:	\ u -- ; -- addr
\ *G Create a buffer of the given size. At runtime, the address
\ ** of the data buffer is returned.
  create  (HeapBuff) ,  does> @ db.addr @  ;


\ ============================
\ *N Dynamically sized buffers
\ ============================
\ *P These buffers are created with an initial size that is
\ ** allocated in the cold chain. They are normally resized
\ ** automatically when there is less than 512 bytes of free
\ ** space in the buffer.

: DynamicBuff:	\ u -- ; -- struct
\ *G Create a buffer of the given size. At runtime, the address
\ ** of the controlling structure is is returned.
  create  (HeapBuff) ,  does> @  ;

: here(db)	\ db -- addr
\ *G Return the next free location in the buffer as an absolute
\ ** address.
  db.used 2@ +  ;

: allot(db)	\ n db --
\ *G Reserve n bytes of space in the buffer.
  db.used +!  ;

: c,(db)	\ char db --
\ *G Store one byte in the buffer and step on.
  tuck here(db) c!  1 allot(db)  ;
: w,(db)	\ char db --
\ *G Store two bytes in the buffer and step on.
  tuck here(db) w!  2 allot(db)  ;
: ,(db)	\ char db --
\ *G Store a four-byte cell in the buffer and step on.
  tuck here(db) !  cell allot(db)  ;

: offs(db)	\ db -- offset
\ *G Return the next free location in the buffer as an offset
\ ** from the start of the buffer.
  db.used 2@ +  ;

: c@(db)	\ offset db -- b
\ *G Fetch a byte at the given offset in the buffer.
  db.addr @ + c@  ;
: w@(db)	\ offset db -- b
\ *G Fetch a 16 bit at the given offset in the buffer.
  db.addr @ + w@  ;
: @(db)		\ offset db -- b
\ *G Fetch a cell at the given offset in the buffer.
  db.addr @ + @  ;

: c!(db)	\ b offset db --
\ *G Store a byte at the given offset in the buffer.
  db.addr @ + c!  ;
: w!(db)	\ w offset db --
\ *G Store a 16 bit word at the given offset in the buffer.
  db.addr @ + w!  ;
: !(db)		\ x offset db --
\ *G Store a cell at the given offset in the buffer.
  db.addr @ + !  ;

: initHeapBuff	\ db --
\ *G Initialise a dynamic buffer. On error, no action is taken.
  tuck db.req @ over db.size !
  0 over db.used !
  dup db.size @ allocate drop swap db.addr ! \ reserve space
;


\ ======
\ *> ###
\ ======

decimal
