\ VFX Forth multitasker for 32 bit Linux

((
Copyright (c) 2000, 2002, 2003, 2004, 2008, 2012, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To Do
=====

Change History
==============
20200919 SFP003 Treat pthread_attr_t as opaque type.
		Use default attributes.
20120420 MPE002 Documentation tidy up
20080110 MPE001 Converted to Linux from Windows code.

The Forth TCB address is the taskid of the task for the Forth words.

At creation time, a task is passed its TCB as the activating parameter.
The first action of a task is to place its USER POINTER into the TCB
structure, and change TOS to 0 so that it is the default return value
if/when the task is terminated.
))


\ =========
\ *! tasker
\ *T Multitasker
\ =========


\ ===============
\ *S Introduction
\ ===============
\ *P The multitasker supplied with VFX Forth is derived from the
\ ** multitasker provided with the MPE Forth cross compilers, v6.1
\ ** onwards. Using a multitasking system can greatly simplify
\ ** complex tasks by breaking them down into manageable chunks.
\ ** This chapter leads you through:
\ *(
\ *B initialising the multitasker
\ *B writing a task
\ *B communicating between tasks
\ *B handling events
\ *)

\ *P The multitasker source code is in the file
\ ** *\i{Lib/Lin32/MultiLin32}. Note that the full version of
\ ** this file with all switches set except for test code is
\ ** compiled as part of the second stage build, but is not
\ ** present by default in the kernel version of VFX Forth.

\ ================
\ *S Configuration
\ ================
\ *P The configuration of the multitasker is controlled by
\ ** constants that control what facilities are compiled:
\ *E   0 constant event-handler?   \ true for event handler
\ **   0 constant message-handler? \ true for message handler
\ **   0 constant semaphores?      \ true for semaphores
\ **   0 constant test-code?       \ true for test code

\ ===============================
\ *S Initialising the multitasker
\ ===============================
\ *P The multitasker needs to be initialised before use. At
\ ** compile time you must define the tasks that your system
\ ** requires and at run-time, all the tasks must be initialised.

\ *P Before use the multitasker must be initialised by the word
\ ** *\fo{INIT-MULTI}, which initialises the primary task *\fo{MAIN},
\ ** and enables the multi-tasker.

\ *P To disable the multitasker, use *\fo{SINGLE}.

\ *P To enable the multitasker, use *\fo{MULTI}, which starts the
\ ** scheduler so new tasks can be added.

\ =================
\ *S Writing a task
\ =================
\ *P Tasks are very straightforward to write, but the way tasks
\ ** are scheduled needs to be understood. This implementation
\ ** uses the Linux *\b{pthreads} API, and so tasks are pre-emptively
\ ** scheduled. This is different from the cooperative scheduler used
\ ** by embedded systems. Despite this, the word *\fo{PAUSE} which
\ ** yields a timeslice is retained for compatibility, and
\ ** *\fo{PAUSE} is where the MPE event handling is incorporated.
\ *E : ACT1ON1     \ -- ; An example task
\ **   TASK0-IO                 \ select the console as the I/O device
\ **   DUP IP-HANDLE !  OP-HANDLE !
\ **   BEGIN                    \ Start an endless loop
\ **     [CHAR] * EMIT          \ Produce a character  )
\ **     1000 ms                \ Wait 1 second
\ **     PAUSE                  \ Needed!
\ **   AGAIN                    \ Go round again
\ ** ;
\ ** TASK TASK1    \ -- tcb ; name task, get space for it

\ *P The task name created by *\fo{TASK} is used as the task
\ ** identifier by all words that control tasks.

\ *N Task dependent variables
\ *P An area of memory known as the *\fo{USER} area is set aside
\ ** for each task. This is often called thread local
\ ** storage. This memory contains user variables which contain
\ ** task specific data. For example, the current number conversion
\ ** radix *\fo{BASE} is normally a user variable as it can vary
\ ** from task to task.

\ *P A user variable is defined in the form:
\ *E n USER <name>
\ *P where n is the nth byte in the user area. The word
\ ** *\fo{+USER} can be used to add a user variable of a given
\ ** size:
\ *E   <size> +USER <name>
\ *P The use of *\fo{+USER} avoids any need to know the offset at
\ ** which the variable starts.

\ *P A user variable is used in the same way as a normal variable.
\ ** By stating its name, its address is placed on the stack,
\ ** which can then be fetched using *\fo{@} and stored by *\fo{!}.

\ ====================
\ *S Controlling tasks
\ ====================
\ *P Tasks can be controlled in the following ways:
\ *(
\ *B activated
\ *B halted
\ *B restarted after it has been halted
\ *B terminated.
\ *)

\ *N Activating a task
\ *P A task is started by activating it. To activate a task, use
\ ** *\fo{INITIATE},
\ *E ' <action> <task> INITIATE
\ *P where *\fo{' <action>} gives the xt of the word to be run
\ ** and *\fo{<task>} is the task identifier. The task identifier
\ ** is used to control the task. Tasks defined by
\ ** *\fo{TASK <name>} return a task identifier when *\fo{<name>}
\ ** is executed.

\ *N Stopping a task
\ *P A task may be temporarily suspended. A task may also halt
\ ** itself. To temporarily stop a task, use *\fo{HALT}.
\ ** *\fo{HALT} is used in the form:
\ *E <task> HALT
\ *P where *\fo{<task>} is the task to be stopped. To restart a
\ ** halted task, use *\fo{RESTART} which is used in the form:
\ *E <task> RESTART
\ *P where *\fo{<task>} is the task to restart.

\ *P To stop the current task (i.e. stop itself) use *\fo{STOP( -- ).

\ *N Terminating a task
\ *P Terminating a task halts it, performs an optional clean up
\ ** action, and calls the operating system thread end function.
\ ** A thread must terminate itself, which leads
\ ** to some complexities. However, it does give the task an
\ ** opportunity to release any resources it may have allocated
\ ** (especially memory) at start up or during its execution. To
\ ** terminate a task use:
\ *E   <task> TERMINATE

\ *P Before the operating system thread end function is called,
\ ** the terminating task will execute its clean up code. The XT
\ ** of the clean up code is held in the task control block. If no
\ ** clean up action is required, zero is used.
\ *E   ... ['] CleanUp MyTask AtTaskExit ...

\ *P If you want a task to be a *\fo{BEGIN ... UNTIL} loop rather
\ ** than an endless loop, this is perfectly legal,
\ ** as returning from a thread will call the clean up code and
\ ** then the *\f{ExitThread} function. However, you must define
\ ** an exit code before you return from the task. Note that on
\ ** entry to the task there will already be a 0 on the stack.
\ *E : MyTask	\ 0 -- exitcode
\ **   <initialisation>         \ initialise task resources
\ **   begin                    \ round and round until done
\ **     <actions>  MyDone @
\ **   until
\ **   drop 0                   \ paranoid, return 0 as success
\ ** ;

\ *P Unlike MPE's embedded systems, under Linux you cannot
\ ** predict how long a task will take to start after *\fo{INITIATE}
\ ** or shut down after *\fo{TERMINATE}.


\ ====================
\ *S Handling messages
\ ====================
\ *P An essential feature of the multitasker is the ability to
\ ** send and receive messages between tasks. For cross compiler
\ ** compatiblility, the operating system mechanisms are not used.

\ *N Sending a message
\ *P To send a message to another task, use the word *\fo{SEND-MESSAGE},
\ ** used in the form:
\ *E message task SEND-MESSAGE
\ *P where *\i{message} is a 32-bit message and *\i{task} is
\ ** the identifier of the receiving task. The message can be data,
\ ** an address or any other type of information but its meaning
\ ** must be known to the receiving task.

\ *N Receiving a message
\ *P To receive a message, use *\fo{GET-MESSAGE}. *\fo{GET-MESSAGE}
\ ** suspends the task until a message arrives. When a message is
\ ** received the task is re-activated and the sending task and
\ ** the data are returned.

\ =========
\ *S Events
\ =========
\ *P Events are analogous to interrupts. Whereas interrupts happen
\ ** on hardware signals, events happen under software control.

\ *N Writing an event
\ *P An event is a normal Forth word. An event is associated to a
\ ** task so that when the event is triggered, the task is resumed.
\ ** Therefore, an event is usually used as initialisation for a
\ ** task. Note that an event handler must have NO net stack effect.

\ *P Events are initialised in a similiar way to tasks. They are
\ ** assigned in the form:
\ *E ASSIGN EVENT1 task TO-EVENT
\ *P where *\fo{EVENT1} is your event handler and *\fo{task} is
\ ** the task that it is to be associated with.

\ *P There are two ways of triggering an event:
\ *(
\ *B using *\fo{SET-EVENT}
\ *B setting a bit in the task status word.
\ *)

\ *P *\fo{SET-EVENT} is a word that sets an event flag for a task. Once
\ ** the event flag is set, the tasker will execute the event
\ ** before it switches to the task's main-line code. The task is
\ ** also restarted.

\ *P A bit can be set in a task's status word that indicates to
\ ** the multitasker that an event has taken place. This method
\ ** can be used to trigger an event from a hardware interrupt or
\ ** a device driver. Refer to `The multitasker internals' section
\ ** later in the chapter for details on the status cell.
\ ** This mechanism can be used to signal that some event has
\ ** taken place, and that consequent processing should start.

\ *P To stop an event handler being run, use *\fo{CLEAR-EVENT}.

\ ====================
\ *S Critical sections
\ ====================
\ *P Sometimes the multitasker has to be inhibited so that other
\ ** tasks are not run during critical operations that would
\ ** otherwise cause the scheduler to operate. This is achieved
\ ** using the words *\fo{SINGLE} and *\fo{MULTI}. Note that these
\ ** do *\{not} stop the Linux scheduler, only the MPE extensions.
\ ** If a full critical section is required, see the semaphore
\ ** source to find out how to use the Windows critical section API.
\ *E SINGLE   -- ; inhibit tasker
\ ** MULTI    -- ; restart tasker

\ *P The following words provided for embedded systems have no
\ ** equivalent because application programs have
\ ** no direct access to the interrupt control mechanisms:
\ *E   DI  EI  SAVE-INT  RESTORE-INT  [I  I]

\ *N Semaphores
\ *P A *\fo{SEMAPHORE} is a structure used for signalling between
\ ** tasks, and for controlling resource usage. It has two fields,
\ ** a counter (cell) and an owner (taskid, cell). The counter
\ ** field is used as a count of the number of times the resource
\ ** may be used, and the owner field contains the TCB of the task
\ ** that last gained access. This field can be used for priority
\ ** arbitration and deadlock detection/arbitration.

\ *P This design of a semaphore can be used either to lock a
\ ** resource such as a comms channel or disc drive during access
\ ** by one task, or as a counted semaphore controlling access to
\ ** a buffer. In the second case the counter field contains the
\ ** number of times the resource can be used. Semaphores are
\ ** accessed using *\fo{SIGNAL} and *\fo{REQUEST}.

\ *P *\fo{SIGNAL} increments the counter field of a semaphore,
\ ** indicating either that another item has been allocated to
\ ** the resource, or it is available for use again, 0 indicating
\ ** that it is in use by a task.

\ *P *\fo{REQUEST} waits until the counter field of a semaphore
\ ** is non-zero, and then decrements the counter field by one.
\ ** This allows the semaphore to be used as a COUNTED semaphore.
\ ** For example a character buffer may be used where the semaphore
\ ** counter shows the number of available characters. Alternatively
\ ** the semaphore may be used purely to share resources. The
\ ** semaphore is initialised to one. The first task to *\fo{REQUEST}
\ ** it gains access, and all other tasks must wait until the
\ ** accessing task *\fo{SIGNAL}s that it has finished with the
\ ** resource.

\ ========================
\ *S Multitasker internals
\ ========================
\ *P A multitasker tries to simulate many processors with just one
\ ** processor. It works by rapidly switching between each task.
\ ** On each task switch it saves the current state of the
\ ** processor, and restores the state that the next task needs.
\ ** The Forth multitasker creates a task control block for each
\ ** task. The task control block (TCB) is a data structure which
\ ** contains information relevant to a task.

\ *S A simple example
\ *P The following example is a simple demonstration of the
\ ** multitasker. Its role is to display a hash `#' every so
\ ** often, but leaving the foreground Forth console running. To
\ ** use the multitasker you must compile the file *\i{LIB\MULTIWIN32.FTH}
\ ** into your system. Note that the file has already been compiled
\ ** by the Studio IDE in *\i{VfxForth.exe}, but is not present in
\ ** *\i{VfxBase.exe}.

\ *P The following code defines a simple task called *\fo{TASK1}.
\ ** It displays a '$' character every second.

\ *E VARIABLE DELAY     \ time delay between #'s in milliseconds
\ **   1000 DELAY !             \ initialise time delay
\ ** : ACTION1          \ -- ; task to display #'s
\ **   TASK0-IO                 \ select the console as the I/O device
\ **   DUP IP-HANDLE !  OP-HANDLE !
\ **   [CHAR] $ EMIT            \ Display a dollar
\ **   BEGIN                    \ Start continuous loop
\ **     [CHAR] # EMIT          \ Display a hash
\ **     DELAY @ MS             \ Reschedule Delay times
\ **     PAUSE                  \ At least one per loop
\ **   AGAIN                    \ Back to the start ...
\ ** ;
\ *P The use of *\fo{PAUSE} in this example is not actually required
\ ** as *\fo{MS} periodically calls *\fo{PAUSE}.

\ *P Before any tasks can be activated, the multitasker must be
\ ** initialised. This is done with the following code:
\ *E INIT-MULTI
\ *P The word *\fo{INIT-MULTI} initialises all the multitasker's
\ ** data structures and starts multitasking. This word need only
\ ** be executed once in a multitasking system and is usually
\ ** executed at start up.

\ *P Note that on entry to a task, the stack depth will be 1. This
\ ** happens because Linux requires a return value when a task
\ ** terminates, and a value of zero is provided by the task
\ ** initialisation code.

\ *P To run the example task, type:
\ *E TASK TASK1
\ ** ASSIGN ACTION1 TASK1 INITIATE

\ *P This will activate *\fo{ACTION1} as the action of task
\ ** *\fo{TASK1}. Immediately you will see a dollar and a hash
\ ** displayed. If you press <return> a few times, you notice
\ ** that the Forth interpreter is still running. After a few
\ ** seconds another hash character will appear. This is the
\ ** example task working in the background.

\ *P The example task can be controlled in several ways:
\ *(
\ *B the rate of generation of hashes can be changed
\ *B it can be halted
\ *B once halted it can be restarted
\ *B it can be started from scratch
\ *)

\ *P Changing the variable *\fo{DELAY} can change the rate of
\ ** production of hashes. Try:
\ *E 2000 DELAY !
\ *P This changes the number of milliseconds between displaying
\ ** hashes to 2000 milliseconds.
\ ** Therefore the rate of displaying hashes halves.

\ *P Typing the task name followed by *\fo{HALT} halts the task:
\ *E TASK1 HALT
\ *P You notice that the hashes are not displayed any more.

\ *P The task is restarted by *\fo{RESTART}. Type:
\ *E TASK1 RESTART
\ *P You notice that the hashes are displayed again.

\ *P To restart the task from scratch, just kill it and activate
\ ** it again:
\ *E TASK1 TERMINATE
\ ** ASSIGN ACTION1 TASK1 INITIATE
\ *P You notice the dollar and the hash are displayed, followed by
\ ** more hashes.


\ ===========
\ *S Glossary
\ ===========

\ ****************
\ *N Configuration
\ ****************

1 constant event-handler?	\ -- n
\ *G The event handling code will be compiled if this constant is true.
1 constant message-handler?	\ -- n
\ *G The message handling code will be compiled if this constant is true.
1 constant semaphores?		\ -- n
\ *G The semaphore code will be compiled if this constant is true.
1 constant test-code?		\ -- n
\ *G The test code will be compiled if this constant is true.


\ *************************
\ *N Structures and support
\ *************************

56 constant /pthread_attr_t	\ -- len
\ *G The structure is now an opaque type - see
\ ** *\i{Kernel/x64Lin/Tests/lintest64.c}.

[defined] Target_x64_Linux [if]
#50 constant /tcb.callback	\ -- len
\ *G Size of the task callback data and code. Used for error checks.
\ ** X64 version.
[then]

[defined] Target_ARM64_Linux [if]
xxx constant /tcb.callback	\ -- len
\ *G Size of the task callback data and code. Used for error checks.
\ ** ARM version.
[then]

struct /TCB	\ -- size
\ *G Returns the size of a TCB structure, which controls the task.
\ *[
  int tcb.link		\ link to next task ; MUST BE FIRST	 0 offset
  int tcb.hthread	\ task handle			         8
  int tcb.up		\ user pointer			        16
  int tcb.pumpxt	\ xt of message pump or 0 for none      24
  int tcb.status	\ status bits			        32
  int tcb.mesg		\ message from another task	        40
  int tcb.msrc		\ TCB of task from which message came   48
  int tcb.event		\ xt of event handler		        56
  int tcb.clean		\ xt of clean up handler		64
  /sem_t field tcb.haltsem \ sem_t (32) for halt/suspend	72
  /tcb.callback field tcb.callback \ task callback structure    104
  aligned		\ force to cell boundary		154
end-struct		\					160
\ *]
\ *P The task status cell reserves the low 8 bits for use by VFX
\ ** Forth. The other bits may be used by your application.
\ *E Bit    When set                       When Reset
\ ** 0      RFU                            RFU
\ ** 1	    Message pending but not read   No messages
\ ** 2	    Event triggered                No events
\ ** 3	    Event handler has been run     No events - reset by user
\ ** 4..7   RFU                            RFU
\ ** 8..31  User defined                   User defined

\ bit masks for the status cell
$0001 constant run-mask		\ running
  run-mask invert constant ~run-mask
  0 constant run-bit#
$0002 constant msg-mask		\ message available
  msg-mask invert constant ~msg-mask
  1 constant msg-bit#
$0004 constant trg-mask		\ event triggered
  trg-mask invert constant ~trg-mask
  2 constant trg-bit#
$0008 constant evt-mask		\ event run
  evt-mask invert constant ~evt-mask
  3 constant evt-bit#

(( left here for DOCGEN use
cell +USER ThreadExit?	\ -- addr
\ *G Holds a non-zero value to cause the thread to exit.
  0 ThreadExit? !			\ for main

cell +USER ThreadTCB	\ -- addr
\ *G Holds a pointer to the thread's TCB.
  0 ThreadTCB !				\ for safety

cell +USER ThreadSync	\ -- addr
\ *G Holds bit patterns used for intertask synchronisation.
\ ** See later section.
  0 ThreadSync !
))

: AtTaskExit	\ xt tcb -- ; set task exit action
\ *G Sets the given task's cleanup action. Use in the form:
\ *C   ' <action> <task> AtTaskExit
  tcb.clean !
;


\ *****************************
\ *N Task definition and access
\ *****************************

create main	\ -- addr ; tcb of main task
\ *G The task structure for the first task run, usually the console
\ ** or the main application.
  /TCB allot
  main ThreadTCB !

: add-task      \ task -- ; insert into list
  main tcb.link @			\ save task currently pointed to
  over tcb.link !			\ and make new task point to it
  main tcb.link !			\ current task points to new task
;

: sub-task      \ task -- ; remove task from chain
\ look for task which points to required task
  main					\ head of chain ; -- target current
  begin
    2dup tcb.link @ <>			\ while curr. does not point to target
   while
    tcb.link @  dup main =		\ get next task, check back at self
    if  2drop  exit  endif		\   and get out if so
  repeat                                \ -- target points-to-it
\ set task that points to target to point to task pointed to by the target
  swap tcb.link @			\ task pointed to by target
  swap tcb.link !			\ set into link field
;

\ variable LastTCBrun

: initThread	\ tcb -- 0 ; set up new task - TCB provided by task creator
\ breakpoint
\  dup tcb.hThread @ 0= if		\ if no handle defined yet
    pthread_self over tcb.hThread !	\   use a temporary thread handle
\  endif
  up@ over tcb.up !			\ user pointer
  dup ThreadTCB !			\ pointer to this task's TCB
  ThreadExit? off			\ don't exit task
  ThreadSync off			\ no synchronisation yet
  dup tcb.haltsem 0 1 sem_init drop	\ Set up haltsem for running
\ dup LastTCBrun !
  drop 0
;

: termThread	\ stat -- stat ; perform task termination actions
  ThreadTCB @ tcb.clean perform
  ThreadTCB @ dup tcb.up off		\ mark TCB as inactive
  dup tcb.haltsem sem_destroy drop	\ Remove haltsem
  sub-task				\ remove from active list
;

: !++		\ addr n -- addr'
  over !  cell +
;

: !call++	\ addr target -- addr'
  over !call skipCall
;

[defined] Target_x64_Linux [if]
: InitTCB	\ tcb --
\ *G Initialise a task control block at addr. X64 version.
  dup /tcb erase
  tcb.callback			\ -- addr' ; start of callback area
  dup 3 cells + !++		\ pointer to callback entry		8   0
  1 !++  1 !++			\ #out #in				16  8
  ['] callback-entry !call++	\ call to entry code		        5   24
  ['] initThread !call++	\ call thread initialiser		5   29
  ['] callback-crash !call++	\ call action word		        5   34
  ['] termThread !call++	\ call to thread terminator	        5   39
  ['] callback-exit !call++	\ call to exit code		        5   44
  $C3 swap c!			\ ret					1   49
;				\                                           50
0 tcb.callback #34 + constant >task.callAction	\ offset to CALL action
main InitTCB
[then]

[defined] Target_ARM64_Linux [if]
: InitTCB	\ addr --
\ *G Initialise a task control block at addr. ARM version.
  dup /tcb erase
  tcb.callback			\ -- addr' ; start of callback area
  dup 3 cells + !++		\ pointer to callback entry		4  0
  1 !++  1 !++			\ #out #in				8  4
  RSPpush-ins !++		\ push 10 regs				4  12
  ['] callback-entry !call++	\ call to entry code		        4  16
  ['] initThread !call++	\ call thread initialiser		4  20
  ['] callback-crash !call++	\ call action word		        4  24
  ['] termThread !call++	\ call to thread terminator	        4  28
  ['] callback-exit !call++	\ call to exit code		        4  32
  RSPpop-ins !++		\ pop 10 regs				4  36
  movPC,LR-ins swap !		\ ret					4  40
  FlushCache			\                                          44
;
0 tcb.callback #24 + constant >task.callAction	\ offset to CALL action
main InitTCB
[then]

: task		\ -- ; -- addr ; define task, returns TCB address
\ *G Use in the form *\fo{TASK <name>}, creates a new task data
\ ** structure called *\fo{<name>} which returns the address of
\ ** the data structure when executed.
  create
    /TCB allot
;

: Self		\ -- tcb|0 ; returns TCB of current task
\ *G Returns the task id (TCB) of the current task. If called
\ ** outside a task, zero is returned.
  ThreadTCB @
;

: his		\ task uservar -- addr
\ *G Given a task id and a *\fo{USER} variable, returns the
\ ** address of that variable in the given task. This word is
\ ** used to set up *\fo{USER} variables in other tasks. Note
\ ** that the task must be running.
  up@ -  swap tcb.up @ +
;


\ ***************************
\ *N Task handling primitives
\ ***************************

0 value multi?	\ -- flag
\ *G Returns true if the tasker is enabled.

: single	\ -- ; disable scheduler
\ *G Disables the Forth portions of the scheduler, but does not
\ ** disable Linux scheduling.
  0 to multi?
;

: multi		\ -- ; enable scheduler
\ *G Enables the Forth portions of the scheduler, but does not
\ ** disable Linux scheduling.
  -1 to multi?
;

(( for DocGen
defer pause	\ --
\ *G *\fo{PAUSE} is the software entry to the pre-emptive scheduler,
\ ** and should be called regularly by all tasks.
\ ** The phrase *\fo{sched_yield drop} occurs at the end of the
\ ** default action *\fo{(PAUSE)}. If the task needs more than
\ ** this and does not use one of the existing message loop words
\ ** such as *\fo{IDLE}, place the XT of the message pump word
\ ** in offset *\fo{TCB.PUMPXT} of the Task Control Block and that XT
\ ** will be called once every time *\fo{PAUSE} is called.
\ ** Because of the way Linux works, *\fo{PAUSE} also controls
\ ** task closure. A task that does not call *\fo{PAUSE} cannot
\ ** be safely terminated except by the task itself, or by a call
\ ** to the API function *\f{kill()}. A task that calls *\fo{PAUSE}
\ ** in a loop without calling any delay mechanism will cause
\ ** CPU hogging.
))

[-short-branches
: (pause)	\ -- ; the scheduler itself
\ *G The action of *\fo{PAUSE} after the multitasker has been
\ ** compiled. If *\fo{SINGLE} has been set, no action is taken.
\ ** If *\fo{PAUSE} was not called from a task and *\fo{MULTI} is
\ ** set, the action is *\b{sched_yield}.
  multi? if				\ Do nothing if MPE portion disabled
    ThreadTCB @ if			\ Do almost nothing if not in a task
      ThreadExit? @ -1 = if
        termThread			\ run clean up action
        0 pthread_exit			\ should never come back!
      endif
      ThreadTCB @
      dup tcb.pumpxt perform		\ not actually used in Linux
      dup tcb.haltsem
        dup sem_wait drop		\ Wait here if task has been halted
        sem_post drop			\ Reset haltsem so others can halt us
[ event-handler? ] [if]
      dup tcb.status @ trg-mask and if	\ event triggered?
        dup tcb.status @		\   clear trigger bit, set event bit
        ~trg-mask and  evt-mask or
        over tcb.status !
        tcb.event perform		\ run event handler
      else
        drop
      endif
[then]
    endif
    sched_yield drop
  endif
;
assign (pause) to-do pause
short-branches]

((
A semaphore is an integer whose value is never allowed to fall
below zero. Two operations can be performed on semaphores:
increment the semaphore value by one using sem_post(3); and
decrement the semaphore value by one using sem_wait(3). If the
value of a semaphore is currently zero, then a sem_wait(3)
operation will block until the value becomes greater than zero.
))

: restart	\ tcb -- ; mark task TCB as running
\ *G If the task has been initiated but is now *\fo{HALT}ed or
\ ** *\fo{STOP}ped, it will be restarted.
  tcb.haltsem sem_post drop
;

: halt		\ tcb -- ; mark thread as halted
\ *G Stops an *\fo{INITIATE}d task from running until
\ ** *\fo{RESTART} is used.
  tcb.haltsem sem_wait drop
;

: stop          \ -- ; halt oneself
\ *G *\fo{HALT}s the current task.
  self halt  pause
;

: running?	\ tcb -- u
\ *G Returns the task's semaphore value, where non-zero indicates
\ ** that it is running. Returns 0 on error.
  { | temp[ cell ] -- }
  temp[ off
  tcb.haltsem temp[ sem_getvalue drop
  temp[ @
;


\ **************
\ Event handling
\ **************

\ *N Event handling

event-handler? [if]

: set-event	\ task -- ; set event trigger in task TCB
\ *G Sets the event trigger bit in the task. When *\fo{PAUSE} is
\ ** next executed by that task, its event handler will be run.
  dup halt				\ interlock this
  dup tcb.status dup @  trg-mask or  swap !
  restart				\ release interlock, wake it up
;

: event?	\ task -- flag ; true if task had event
( *G Returns true if the task has received an event trigger, )
( ** but has not yet run the event handler. )
  tcb.status @  trg-mask and 0<>
;

: clr-event-run	\ -- ; reset own EVENT_RUN flag
\ *G Clear the *\fo{EVENT_RUN} flag of the current task. This is
\ ** usually done if the task has to be put back to sleep after
\ ** the event handler has been run.
  ThreadTCB @ tcb.status dup @  ~evt-mask and  swap !
;

: to-event      \ xt task -- ; define action of a task
\ *G Used in the form below to define a task's event handler:
\ *C   assign <action> <task> to-event
  tcb.event !				\ set into TCB of task
;
[then]


\ *******************
\ *N Message handling
\ *******************

message-handler? [if]
: msg?		\ task -- flag ; true if task has message
\ *G Returns true if the given task has received a message.
  tcb.status @ msg-mask and 0<>
;

: (send-msg)	\ mesg task -- ; store message in task
  self over tcb.msrc !			\ set source
  dup
  tcb.status dup @  msg-mask or  swap !	\ set message bit
  tcb.mesg !				\ set message
;

: (get-message)	\ -- mesg task
  ThreadTCB @ >r
  r@ tcb.mesg @				\ message
  r@ tcb.msrc @				\ source task
  r> tcb.status
  dup @  ~msg-mask and  swap !		\ reset message bit
;

: send-message  \ mesg task -- ; send message to task (wakes it up)
\ *G Sends a message to a task, waking it up if it was asleep.
\ ** Interpretation of a message is the resposibility of the
\ ** receiving task. If the receiving task has unprocessed
\ ** messages, the sending task blocks.
  begin  dup msg?               	\ if task already has a message
    while  pause                	\ wait
  repeat                        	\ and try again
  tuck (send-msg)			\ send the message
  restart
;

: get-message   \ -- mesg task ; wait for any message
\ *G Waits until a message has been received from another task.
\ ** Interpretation of a message is the resposibility of the
\ ** receiving task. See *\fo{MSG?} which tells you if a message
\ ** is available.
  begin  self msg? 0=           	\ check for a pending message
    while  pause                	\ wait for it
  repeat                        	\ try again
  (get-message)                 	\ read it
;
[then]

event-handler? message-handler? and [if]
: wait-event/msg        \ -- ; wait for message or event trigger
\ *G Wait until a message or event occurs.
  begin
    self msg?  self event?  or 0=
   while
    pause
  repeat
;
[then]


\ ******************
\ *N Task management
\ ******************

: MakeThread	\ tcb --
  0 over tcb.hThread !			\ mark as no handle yet
  >r
  r@ tcb.hThread			\ receives thread handle
  0					\ use default attributes
  r@ tcb.callback @			\ entry point
  r>					\ parameter is TCB
  pthread_create drop
." started" cr
;

: to-task	\ xt task -- ; set action of task
\ *G Used in the form below to define a task's action:
\ *C   assign <action> <task> to-task
  >task.callAction + !call
;

: to-pump	\ xt task -- ; set message loop of task
( *G Used in the form below to define the action of the message pump:	)
( *C   assign <action> <task> to-pump					)
  tcb.pumpxt !
;

: initiate	\ xt task -- ; start task from scratch
\ *G Initialises a task running the given xt. All required O/S
\ ** resources are allocated by this word.
  dup tcb.up off			\ no USER area yet
  dup initTCB				\ set up TCB
  tuck to-task				\ set action of task
  dup MakeThread			\ start task
  dup add-task                          \ add to chain
  begin					\ wait until user area set
    pause				\ let it run
    dup tcb.up @
  until
  drop
;

: terminate	\ task -- ; stop task, and remove from list
\ *G Causes the specified task to die. You should not make
\ ** assumptions as to how long this will take. Unlike the
\ ** embedded systems implementations, this word is very
\ ** operating system dependent. The task may still be alive
\ ** on return from this call.
  dup tcb.up @ 0=			\ check if already terminated
  if  drop  exit  endif
  ThreadExit? his on
  pause					\ halt task, and remove from chain
;
\ *P *\b{N.B.} Do not use *\fo{self terminate} to cause a task to end.
\ ** Use the following instead:
\ *E : Suicide    \ -- ; terminate current task
\ **   pause  termThread  0 pthread_exit
\ ** ;
\ **
\ ** ... Suicide

: start:	\ task -- ; exits from caller
\ *G *\fo{START:} is used inside a colon definition. The code
\ ** before *\fo{START:} is the task's initialisation, performed
\ ** by the current task. The code after *\fo{START:} up to the
\ ** closing *\fo{;} is the action of the task. For example:
\ *E   TASK FOO
\ **   : RUN-FOO
\ **     ...
\ **     FOO START:
\ **     ...
\ **     begin ... pause again  )
\ **   ;

\ *P All tasks must run in an endless loop, except for
\ ** initialisation code. There are exceptions to
\ ** this, and these are discussed in the section on terminating
\ ** a task. When *\fo{RUN-FOO} is executed, the code after
\ ** *\fo{START:} is set up as the action of task *\fo{FOO} and
\ ** started. *\fo{RUN-FOO} then exits. If you want to perform
\ ** additional actions after starting the task, you should use
\ ** I*\fo{INITIATE} to start the task.
  r> swap initiate
;

: TaskState	\ task -- state
\ *G Returns true if the task has started and zero if the thread
\ ** has finished.
  tcb.up @ 0<>
;

\ \ : WaitTaskState	\ task state --
\ \ *G Wait until the task is in a given state.
\ \ *E   task -1 WAITTASKSTATE \ wait until task starts
\ \ **   task 0 WAITTASKSTATE  \ wait until task closes
\   begin
\     pause
\     2dup swap TaskState <>
\   until
\ ;

: init-multi    \ -- ; initialisation of multi-tasking
\ *G Initialise the Forth multitasker to a state where only the
\ ** task *\fo{MAIN} is known to be running. *\fo{INIT-MULTI} is
\ ** added to the cold chain and is also called during
\ ** compilation of *\i{MultiLin64.fth}. This word *\b{must}
\ ** be run from *\fo{MAIN}.
  main initTCB				\ must setup data structures
  main ThreadTCB !
  ThreadExit? off			\ don't exit task
  ThreadSync off			\ no synchronisation yet
  pthread_self main tcb.hThread !
  main tcb.haltsem 0 1 sem_init drop	\ Set up haltsem for running
  single				\ tasker disabled
  main main tcb.link !			\ initialise MAIN to point to itself
  multi					\ mark tasker enabled
;
' init-multi AtCold  init-multi
((
: initThread	\ tcb -- 0 ; set up new task - TCB provided by task creator
breakpoint
\  dup tcb.hThread @ 0= if		\ if no handle defined yet
    pthread_self over tcb.hThread !	\   use a temporary thread handle
\  endif
  up@ over tcb.up !			\ user pointer
  dup ThreadTCB !			\ pointer to this task's TCB
  ThreadExit? off			\ don't exit task
  ThreadSync off			\ no synchronisation yet
  dup tcb.haltsem 0 1 sem_init drop	\ Set up haltsem for running
dup LastTCBrun !
  drop 0
;
))

: term-multi	\ --
\ *G Performed in the exit chain when the program terminates.
\ ** closes all active tasks except *\fo{SELF}. This allows
\ ** all task clean-up actions to be performed before the
\ ** program itself finishes.
\ tell all tasks to finish
  main begin
    dup self <> if
      dup tcb.link @  swap terminate
    else
      tcb.link @
    endif
    dup main =
  until
  drop  #50 ms				\ let them finish
;
\ ' term-multi AtExit

: .task		\ task -- ; display task name
\ *G Given a task, e.g. as returned by *\fo{SELF}, display its
\ ** name or address.
  dup InForth? if
    ip>nfa .name
  else
    ." $" .dword
  endif
;

: .tasks	\ -- ; display active tasks
\ *G Display a list of all the active Forth tasks.
  cr  main
  begin
    dup .task  2 spaces  ?cr
    tcb.link @ dup main =
  until
  drop
;


\ ***********************
\ *N Task synchronisation
\ ***********************

$AAAA:5555:AAAA:5555 constant TaskReady	\ -- n
\ *G At task initiation, *\fo{USER} variable *\fo{THREADSYNC} is
\ ** set to zero. Set *\fo{THREADSYNC} to this value to indicate
\ ** that the task is willing to synchronise with another task.

$5555:AAAA:5555:AAAA constant TaskReadied	\ -- n
\ *G A synchronising task sets another task's *\fo{THREADSYNC}
\ ** to this value to indicate that synchronisation is
\ ** complete.

also system
: WaitForSync	\ --
\ *G Perform the slave synchronisation sequence.
  TaskReady ThreadSync !
  begin
    1 (OS_ms) pause
    ThreadSync @ TaskReadied =
  until
  ThreadSync off
;

: [Sync		\ task -- task
\ *G Used by a master task in the form:
\ *C   [Sync ... Sync]
\ *P to synchronise and pass data to another task, usually
\ ** when *\fo{USER} variables must be initialised. The slave
\ ** task must execute *\fo{WAITFORSYNC}.
  begin		\ wait for slave task to come ready
    1 (OS_ms) pause
    dup ThreadSync his @ TaskReady =
  until
;

: Sync]		\ task --
\ *G Used by a master task in the form:
\ *C   [Sync ... Sync]
\ *P to indicate the end of synchronisation.
  TaskReadied over ThreadSync his !	\ mark completion
  begin					\ wait for task to zero
    1 (OS_ms) pause  dup ThreadSync his @ 0=
  until
  drop
;
previous


\ *************
\ *N Semaphores
\ *************

semaphores? [if]

struct /semaphore	\ -- len
\ *G Structure used for Linux x64 semaphores.
  int sm.count		\ compatibility with XC6 and Win32
  int sm.owner		\ compatibility with XC6 and Win32
  /sem_t field sm.sem_t	\ Linux semaphore
end-struct

: semaphore	\ -- ; -- addr [child]
\ *G A *\fo{SEMAPHORE} is an extended variable used for signalling
\ ** between tasks and for resource allocation. The counter field
\ ** is used as a count of the number of times the resource may be
\ ** used, and the arbiter field contains the TCB of the task
\ ** that last gained access. This field can be used for priority
\ ** arbitration and deadlock detection/arbitration.
  create
    /semaphore allot&erase
;

: InitSem	\ semaphore --
\ *G Initialise the semaphore. This *\b{must} be done before using it.
  dup 0 1 rot sm.count 2!
  sm.sem_t InitCritSec
;

: ShutSem	\ semaphore --
\ *G Delete the critical section associated with the smaphore.
  sm.sem_t TermCritSec
;

: LockSem	\ semaphore --
\ *G Lock the semaphore.
  sm.sem_t [CritSec
;

: UnlockSem	\ semaphore --
\ *G Unlock the semaphore.
  sm.sem_t CritSec]
;

: signal        \ sem -- ; increment counter field of semaphore,
\ *G *\fo{SIGNAL} increments the counter field of a semaphore,
\ ** indicating either that another item has been allocated to
\ ** the resource, or that it is available for use again, 0
\ ** indicating in use by a task.
  dup sm.count incr  dup sm.owner off	\ inc. counter, release
  UnlockSem
;

: request       \ sem -- ; get access to resource, wait if count = 0
\ *G *\fo{REQUEST} waits until the counter field of a semaphore
\ ** is non-zero, and then decrements the counter field by one.
\ ** This allows the semaphore to be used as a *\b{counted}
\ ** semaphore. For example a character buffer may be used where
\ ** the semaphore counter shows the number of available
\ ** characters. Alternatively the semaphore may be used purely
\ ** to share resources. The semaphore is initialised to one.
\ ** The first task to *\fo{REQUEST} it gains access, and all
\ ** other tasks must wait until the accessing task *\fo{SIGNAL}s
\ ** that it has finished with the resource.
  dup LockSem
  dup sm.count decr			\ got it, decrement counter
  self swap sm.owner !			\ mark resource as mine
;

[then]


\ ********
\ all done
\ ********

decimal


\ *********
\ Test code
\ *********

test-code? [if]

op-handle @ value task0-io

task task1

#2200 value rate1

: default-io	\ -- ; set up default I/O
  task0-io dup op-handle ! ip-handle !
;

: action1	\ --
\ This code prints a '*' character every RATE1 milliseconds.
  default-io decimal			\ must initialise USER vars
  begin
    [char] * emit
    rate1 ms  pause			\ must PAUSE
  again
;
\ ' action1 task1 initiate		\ start TASK1

task task2

#1800 value rate2

: run-task2	\ --
\ This code prints a '/' character every RATE2 milliseconds.
  task2 start:
    default-io decimal			\ must initialise USER vars
    begin
      [char] / emit
      rate2 ms  pause			\ must PAUSE
    again
;

event-handler? [if]
: task2-event	\ -- ; MUST have null stack action
\ This code displays a message and clears the the event run flag.
  cr ." TASK2 event" cr			\ tell user
  clr-event-run				\ clear event run flag in this task
;
' task2-event task2 to-event
[then]

[then]


\ ======
\ *> ###
\ ======

decimal
