(* ====================================================================
text window

copyright 2007 Roelf Toxopeus

VFX version
Implements a text scrollview in a window
Last: 7 August 2020 at 11:30:47 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Two necessary classes and their relevant methods are defined:
NSScrollView and NSTextView.
You can have several scroll- and text-views in a window, so save the
current views in 'scrollview and 'textview if you plan to add others.
A textview is connected to a scrollview which becomes a window's
contentview. It's all about views.

SCROLLVIEW -- create and initialise a scrollview instance. Use some
default attributes, default scrolls only verticaly. Main thread required,
use PASS.

TEXTVIEW -- create and initialise a textview instance. Use some
default attributes. Main thread required, use PASS.

SCROLLS.HORIZONTAL -- set scrollview from given textview to allow
horizontal scrolling.

WINDOWFORTEXT -- initialise given window for text I/O. Return an instance
from the textview. This is executed on the main thread, because all
the messages are required to run on the main thread.

FONT -- return NSFont instance from given font name and size.

VIEWFONT -- set font for given textview instance, using fontname and size.

demo usage:

new.window mytvwin
s" A Text View" mytvwin w.title

variable mytextview
: /mywindow
mytvwin add.window
10 ms ( el capitan quirk)
mytvwin windowfortext
dup mytextview !
dup scrolls.horizontal
0" Monaco" 16 rot viewfont ;

-------------------------------------------------------------------- ))


/forth
decimal

\ --------------------------------------------------------------------
\ the scrollview

cocoaclass NSScrollView

cocoa: @setBorderType: ( NSBorderType:atype -- ret )      \ NSScrollView          
cocoa: @setHasVerticalScroller: ( BOOL:flag -- ret )      \ NSScrollView 
cocoa: @setHasHorizontalScroller: ( BOOL:flag -- ret )    \ NSScrollView 
cocoa: @setAutoresizingMask: ( NSUInteger:mask -- ret )   \ NSScrollView 

 0 constant NSNoBorder
 2 constant NSViewWidthSizable
16 constant NSViewHeightSizable

: SCROLLVIEW ( -- NSScrollView:ref )
   NSScrollView @alloc @init dup >r
   NSNoBorder r@ @setBorderType: drop
   yes        r@ @setHasVerticalScroller: drop
   no         r@ @setHasHorizontalScroller: drop
   NSViewWidthSizable NSViewHeightSizable or r> @setAutoresizingMask: drop
;

\ --------------------------------------------------------------------
\ the textview

COCOACLASS NSTextView

cocoa: @setVerticallyResizable: ( BOOL:flag -- ret )                        \ NSTextView
cocoa: @setHorizontallyResizable: ( BOOL:flag -- ret )                      \ NSTextView
cocoa: @textContainer ( -- NSTextContainer:ref )                            \ NSTextView
cocoa: @setWidthTracksTextView: ( BOOL:flag -- ret )                        \ NSTextContainer
cocoa: @isRichText ( -- bool )
cocoa: @setRichText: ( bool -- ret )

: TEXTVIEW ( -- NSTextView:ref )
   NSTextView @alloc @init dup >r
   no  r@ @setRichText: drop
   yes r@ @setVerticallyResizable: drop
   no  r@ @setHorizontallyResizable: drop
   NSViewWidthSizable r@ @setAutoresizingMask: drop
   yes r> @textContainer @setWidthTracksTextView: drop
;

\ --------------------------------------------------------------------
\ horizontal scroller change

cocoa: @enclosingScrollView ( -- NSScrollView:ref )                          \ NSTextView
cocoa: @setContainerSize: ( NSSize:widthfloat NSSize:heightfloat -- ret )    \ NSTextContainer

3.40282347e+38 fconstant FLT_MAX   \ see: /usr/include/float.h

: SCROLLS.HORIZONTAL ( textview -- )
   >r
   yes r@ @enclosingScrollView @setHasHorizontalScroller: drop
   yes r@ @setHorizontallyResizable: drop
   NSViewWidthSizable NSViewHeightSizable or r@ @setAutoresizingMask: drop
   flt_max fdup fpushs fpushs r@ @textContainer @setContainerSize: drop
   no r> @textContainer @setWidthTracksTextView: drop
;
   
\ --------------------------------------------------------------------
\ put the lot together:
\ all of these want the main thread!

cocoa: @setDocumentView: ( NSView:aView -- ret )                \ NSScrollView
cocoa: @makeKeyAndOrderFront: ( id:sender -- ret )              \ NSWindow
cocoa: @makeFirstResponder: ( NSResponder:responder -- BOOL )   \ NSWindow

: (WINDOWFORTEXT) ( wptr4 -- )
   +w.ref @ >r
   textview scrollview 2dup                  \ new text- and scrollview
   @setDocumentView: drop                    \ connect textview to scrollview
   r@ @setContentView: drop                  \ set scrollview to window's contentview
   0 r@ @makeKeyAndOrderFront: drop          \ make window the event receiver
   dup r> @makeFirstResponder: drop          \ make textview the text receiver
   pocket !
;

: WINDOWFORTEXT ( wptr4 -- textview )   1 0 ['] (windowfortext) pass pocket @ ;

\ --------------------------------------------------------------------
\ last set font:

cocoaclass NSFont

cocoa: @fontWithName:size: ( NSString:fontName CGFloat:fontSize -- NSFont:ref )  \ NSFont
cocoa: @setFont: ( NSFont:ref -- ret )                                           \ NSTextStorage/NSTextView

: FONT ( 0string:fontname fontsize -- nsfont:ref )
   swap >nsstring swap s>sf NSFont @fontWithName:size: ;

: VIEWFONT ( 0string:fontname fontsize textview -- )
   >r font ?dup
         if r> @setFont: drop
      else r> drop
   then ;

cr .( TextView for Cocoa window loaded)

\\ ( oef )
