(* ====================================================================
Timers using GCD dispatch queues

copyright 2014 Roelf Toxopeus

VFX version
Allows many independant timers, simple interface!
Last: 4 February 2019 at 13:12:00 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
For all the imported functions, see:
  usr/include/dispatch/*.h
  Apple's GCD_libdispatch_Ref
  Apple's Concurrency Programming Guide

NoTE: example of a GLOBAL: defined reference without using @ on it.
The external name of C objects in the UNIX world is the name with an underscore
prepended, here _dispatch_source_type_timer.
Thanks Mitch Bradley!

DISPATCH_SOURCE_TYPE_TIME a pointer to a constant (&_dispatch_source_type_timer)
defining a timer dispatch source type. Used for creating timers.

DISPATCH_TIME_NOW a 64b time constant indicating 'now', used by GCD dispatch functions.

Because the dispatch routines use nanoseconds, there are some conversion words:
S>NANO converts seconds to nanoseconds
MS>NAN converts milliseconds to nanoseconds
US>NAN converts microseconds to nanoseconds
These nanoseconds are 2 cell values on 32b Forth systems!
-------------------------------------------------------------------- ))

/forth
decimal

function: dispatch_source_create ( *type handle mask queue -- source )
function: dispatch_time ( when:lo when:hi delta:lo delta:hi -- dispatch_time_t:lo dispatch_time_t:hi )
function: dispatch_source_set_timer ( source start:lo start:hi interval:lo interval:hi leeway:lo leeway:hi -- ret )
function: dispatch_source_set_event_handler_f ( source handler -- ret )
function: dispatch_resume ( object -- ret )
function: dispatch_source_cancel ( source -- ret )
function: dispatch_release ( object -- ret )

\ don't fetch this one, we need a pointer
global: _dispatch_source_type_timer

: DISPATCH_SOURCE_TYPE_TIMER ( -- addr )  _dispatch_source_type_timer ;

0 dup 2constant DISPATCH_TIME_NOW

: S>NANO  ( n -- d )   1000000000 m* ;
: MS>NANO ( n -- d )   1000000 m* ;
: US>NANO ( n -- d )   1000 m* ;

(( --------------------------------------------------------------------
CREATE-TIMER creates a GCD dispatch timer. You need to initialise it before
it can be dispatched. A Timer type has 0 for the handle and mask parameters,
until further notice from Apple.
Could try parallelqueue for hi-priority.

TIME-NOW returns dispatch time in nanoseconds, meaning do it now.
The default clock is based on mach_absolute_time.
Using 0 0 delta parameter for now, start immediately.

FROM-NOW returns dispatch time in nanoseconds _after_ given ms from TIME-NOW

LEEWAY is a 64b variable containing the allowed leeway for the timer. Assuming
most of the time you want it as accurate as possible, a default value of 1 nanosecond
is used. This so you won't be bothered to pass a leeway parameter along. You can
set it to a preferred value with  <double:nanoseconds> LEEWAY 2!
Do this before you execute NEW-TIMER or LOAD-TIMER.

LOAD-TIMER will prepare the given timer for duty. Interval and eventhandler
are set. Afterwards the leeway parameter is reset to a default 1 nanosecond,
just in case a custom value was used.

NEW-TIMER creates a GCD dispatch timer, initialised and ready to run. It starts
in suspended state. The interval parameter is in milliseconds. The callback is
a Forth callback, run when the timer fires.

START-TIMER will start a suspended timer. Room for additions.

STOP-TIMER stops and removes a timer. Room for additions.
-------------------------------------------------------------------- ))

: CREATE-TIMER ( -- timer )   dispatch_source_type_timer 0 0 defaultqueue dispatch_source_create ;

: TIME-NOW ( -- dispatch-time:lo dispatch-time:hi )   dispatch_time_now 0 dup dispatch_time ;

: FROM-NOW ( ms -- dispatch-time:lo dispatch-time:hi )   time-now rot ms>nano dispatch_time ;

2variable leeway
: /LEEWAY ( -- )   1  ms>nano leeway 2! ;
/leeway

: LOAD-TIMER ( cb interval timer -- )
   swap >r
   dup r@ from-now  r> ms>nano  leeway 2@  dispatch_source_set_timer drop
   /leeway
   swap dispatch_source_set_event_handler_f drop ;

: NEW-TIMER ( cb interval -- timer )
   create-timer dup 0= abort" No Timer Source !"
   dup >r load-timer r> ;

: START-TIMER ( timer -- )  dispatch_resume drop ;

: STOP-TIMER ( timer -- )  dup dispatch_source_cancel drop  dispatch_release drop ;

\\

\ --- examples:

forthcall: *greet ( -- )   impostor's cr ." hoi" ;
*greet 5000 new-timer dup value mt start-timer

mt stop-timer

\ ---

variable accu
: bump ( -- )  1 accu +! ;
' bump 0 cb: *bump

variable mytimer

: tt ( -- )   *bump 250 new-timer dup mytimer ! start-timer ;

: ss ( -- )   mytimer @ stop-timer ;

: xx ( -- )  accu off tt counter 4999 ms ss timer accu ? ;

: bump2 ( -- )  10 accu +! ;
' bump2 0 cb: *bump2

variable mytimer2

: tt2 ( -- )   *bump2 1000 new-timer dup mytimer2 ! start-timer ;

: ss2 ( -- )   mytimer2 @ stop-timer ;

: xx2 ( -- )  accu off tt2 counter 4999 ms ss2 timer accu ? ;

: yy ( -- )   accu off tt tt2 counter 4999 ms ss ss2 timer accu ? ;

: bump3 ( -- )  7 accu +! ;
' bump3 0 cb: *bump3

variable mytimer3

: tt3 ( -- )   *bump3 700 new-timer dup mytimer3 ! start-timer ;

: ss3 ( -- )   mytimer3 @ stop-timer ;

: xx3 ( -- )  accu off tt3 counter 4999 ms ss3 timer accu ? ;

: zz ( -- )   accu off tt tt2 tt3 counter 4999 ms ss ss2 ss3 timer accu ? ;

cr .( examples: )
cr .( 1 timer:  xx   -> expect 5000  20)
cr .( 2 timers: yy   -> expect 5000  70)
cr .( 3 timers: zz   -> expect 5000 126)

\\ ( eof )
