(* ====================================================================
Parallel execution

copyright 2013 Roelf Toxopeus

VFX version.
Using Grand Central Dispatch concurrent queue's and the GCD group
calls to implement a suggested FORK JOIN CONTINUE.
Intention is execution of code on separate processor cores in parallel.
Last: 4 February 2019 at 12:28:19 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
PARALLELQUEUE -- returns the highest priority dispatch queue. This queue
is used for parallel processing.
DISPATCH-GROUP -- dispatch n cb's on a concurrent queue as a group.
Return the group for further usage.
JOIN-GROUP -- wait for given group to finish.

FORK JOIN CONTINUE -- process words between FORK and JOIN/CONTINUE to
run in parallel when definition under construction is run.
JOIN waits till forked group has finished.
CONTINUE does not wait.

FORK? -- flags if given xt is allowed to be forked. In essence a stopper
for the FORK JOIN/CONTINUE pairs.

NoTE: in VFX ' doesn't work over linebreaks.
     Use NEXTWORD FIND ?UNDEF to compensate in FORK.

The FORK JOIN pair is similar to tForth's PAR END-PAR

Example usage:

: FOTO-FUN ( -- )
   get-image
   fork
    do-red do-green do-blue do-alpha
   join
   show-image ;
-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ GCD group calls

function: dispatch_group_create ( -- group )

function: dispatch_group_async_f ( group queue *context 'cb -- ret )   

\ NoTE: dispatch_time_t is 64bit, so pass it as a Forth double in 32bit SF !!!!!!!!!!!!!!!!!
function: dispatch_group_wait ( group time:lo time:hi -- ret:long )

function: dispatch_release ( dispatch-object -- ret )

: PARALLELQUEUE ( -- queue )  dispatch_queue_priority_high 0 dispatch_get_global_queue ;

0 invert s>d 2constant DISPATCH_TIME_FOREVER   \   #define DISPATCH_TIME_FOREVER (~0ull) 

: (DISPATCH-GROUP) ( cb group -- )   swap >r parallelqueue 0 r> dispatch_group_async_f drop ;

: DISPATCH-GROUP ( cb1 ... cbn n -- group )   
   dispatch_group_create swap 0 ?do  tuck (dispatch-group) loop ;
      
: JOIN-GROUP ( group -- )  dup dispatch_time_forever dispatch_group_wait drop dispatch_release drop ;

\ --------------------------------------------------------------------
\ FORK API

: COMPILE-CBS ( cb1 ... cbn n -- )   0 ?do postpone literal loop ;

: JOIN ( cb1 ... cbn n -- )  dup >r compile-cbs r> postpone literal postpone dispatch-group postpone join-group ; immediate

: CONTINUE ( cb1 ... cbn n -- )  dup >r compile-cbs r> postpone literal postpone dispatch-group postpone dispatch_release postpone drop ; immediate

: FORK? ( xt -- flag )   dup ['] join <> swap ['] continue <> and ;

: FORK ( <text> -- cb1 ... cbn n ) \ should use compile only, can't be bothered now.
   0
   begin
      nextword find ?undef  dup fork?
   while
      postpone ahead >r     \ create callback pointers and skip over them
      0 1
      callback, @ tuck set-callback
      r> postpone then
      swap 1+   
   repeat
   execute ; immediate

\ --------------------------------------------------------------------
\ PAR API

: END-PAR ( cb1 ... cbn n -- )   postpone join ; immediate

: PAR ( <text> -- ) 
   0
   begin
      nextword find ?undef  dup ['] end-par <>    ( differs here from FORK )
   while
      postpone ahead >r     \ create callback pointers and skip over them
      callback, @ tuck set-callback
      r> postpone then
      swap 1+   
   repeat execute ; immediate

\\ ( eof )

\ simple test:

variable accu
: bump1 ( -- )   1 accu +!  500 ms cr ." nr.1 done" ;
: bump2 ( -- )   1 accu +!  869 ms cr ." nr.2 done" ;
: bump3 ( -- )   1 accu +! 1200 ms cr ." nr.3 done" ;
: bump4 ( -- )   1 accu +! 1700 ms cr ." nr.4 done" ;

: pp ( -- )
   accu off
   fork
      bump1
      bump2
      bump3
      bump4
   join
   cr ." accu ? " accu ? ;

: ff ( -- )
   accu off
   FORK
      bump1
      bump2
      bump3
      bump4
   CONTINUE
   cr ." accu ? " accu ? ;

(* --------------------------------------------------------------------
More interesting is filling a memory range.
Do it serialy in one go, or divide it up and hand them over to different
tasks/threads/cores running in parallel.

NoTE: run the first executed test twice, this causes mem to be cached.
Improves bench time considerately. I need to do something about this
cache business. Specialy for short bursts of memory access.
-------------------------------------------------------------------- *)

\ --------------------------------------------------------------------
\ One task does all the work in one go.

6000000 CONSTANT /MEM   \ buffer for dfloats, depends on UNUSED

/MEM BUFFER: MEM
: .MEM ( -- )   MEM 100 DUMP  MEM /MEM + 80 - 100 DUMP ;

variable accu
\ : WORK ( addr u -- )   2DUP ERASE BLANK ;
\ : WORK ( addr u -- )   2DUP ERASE  0 DO RANDOM OVER I + C! LOOP DROP ;
: WORK ( addr u -- )   0 DO  I S>F FDUP F* DUP I + DF!  [ 1 DFLOATS ] LITERAL ( 1 accu +!) +LOOP DROP ; 
: FILLMEM ( -- )   MEM /MEM  WORK ;

: FF ( -- )
   CR ." ff not dispatching, just 1 task, "
   COUNTER FILLMEM TIMER ." MS elapsed" ;

\ --------------------------------------------------------------------
\ Using dispatchgroup for 4 parallel execution jobs

/MEM 4 / CONSTANT MEM/4
MEM          CONSTANT MEM1
MEM1 MEM/4 + CONSTANT MEM2
MEM2 MEM/4 + CONSTANT MEM3
MEM3 MEM/4 + CONSTANT MEM4

: FILLMEM1 ( -- )   ( impostor's ." 1 starts ... " 8 FSTACK ) MEM1 MEM/4 WORK ( ." 1 done!") ;
: FILLMEM2 ( -- )   ( impostor's ." 2 starts ... " 8 FSTACK ) MEM2 MEM/4 WORK ( ." 2 done!") ;
: FILLMEM3 ( -- )   ( impostor's ." 3 starts ... " 8 FSTACK ) MEM3 MEM/4 WORK ( ." 3 done!") ;
: FILLMEM4 ( -- )   ( impostor's ." 4 starts ... " 8 FSTACK ) MEM4 MEM/4 WORK ( ." 4 done!") ;

: FF2 ( -- )
   CR ." ff2, using dispatch_group for 4 jobs, "
   COUNTER
   FORK FILLMEM1 FILLMEM2 FILLMEM3 FILLMEM4 JOIN
   TIMER ." MS elapsed" ;

\ --------------------------------------------------------------------
\ View cores, use Activity Monitor with CPU History and CPU Usage
\ windows open.

: DD ( -- )
   CR ." dd not dispatching, just 1 task."
   CR ." any key stops..."
   BEGIN  FILLMEM  KEY? UNTIL ;

: DD2 ( -- )
   CR ." dd2, using dispatch_group for 4 jobs."
   CR ." any key stops..."
   BEGIN
      FORK  FILLMEM1 FILLMEM2 FILLMEM3 FILLMEM4  JOIN
   KEY? UNTIL ;
