\ Lib/Osx32/Genio/Serial.fth - serial devices

((
Copyright (c) 2008, 2011, 2013
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  (901) 313 4312 - rings in UK


To do
=====

Change history
==============
20100116 SFP003 Improved string interpreter.
20080704 SFP002 Enhanced raw mode setting.
		Forced SID handle to -1 on close.
20080208 MPE001 First cut.
))

only forth definitions decimal

\ ========
\ *> genio
\ ========
\ *N Serial Device

\ *P This Generic IO Device operates on a serial port for input
\ ** and/or output.

\ *P In order to create a device use the *\fo{SERDEV:} definition
\ ** given later.

\ *P When opening a serial device the parameters to *\fo{OPEN-GEN}
\ ** and *\fo{open-gio} have the following meaning:
\ *E    ADDR            Address of configuration string.
\ **    LEN             Length of string at ADDR.
\ **    ATTRIBS         file fam, usually R/W.
\ *P The configuration string takes the form:
\ *E    /dev/ttyS0 9600 baud  no parity  8 data  1 stop
\ *P for Serial Port 0 at 9600 baud, 8 data bits, no parity, 1 stop bit.
\ ** Only the device name is mandatory. Words and IOCTL functions are
\ ** available to modify the port setting later. Split baud rates are
\ ** not supported - you will have to set these yourself. If only
\ ** the device name is given, the line will be set to 115200 baud,
\ ** N81 in raw mode. Additional configuration comands are documented
\ ** later in this section, e.g. for setting the DTR and RTS lines.
\ ** The configuration string is processed with *\fo{BASE} set to
\ ** *\fo{DECIMAL}. USB serial devices are discussed at the end of
\ ** this section.

\ *P A good description of Mac OS X serial ports is at:
\ *C   http://pbxbook.com/other/mac-tty.html


\ ********************
\ *H Serial primitives
\ ********************

struct /serial-sid	\ -- len
\ *G Defines the SID of a serial device.
  gen-sid +			\ reuse field names of GEN-SID
  int ss./le			\ length of line end string for output
  int ss.le$			\ address of line end string
  int ss.lechar			\ line end character for ACCEPT
  int ss.ignchar		\ ignore character for ACCEPT
end-struct

Extern: int tcgetattr( int fildes, void * termios_p );
\ 0=good
Extern: int tcsetattr( int fildes, int options, void * termios_p );
\ 0=good
Extern: int cfsetospeed( void * termios_p, int speed );
Extern: int cfsetispeed( void * termios_p, int speed );
Extern: int cfgetospeed( void * termios_p );
Extern: int cfgetispeed( void * termios_p );

struct /termios \ -- size
\ *G A structure corresponding to the termios structure used by
\ ** *\b{tcgetattr} and *\b{tcsetattr}. 32 bit specific.
\ *[
  4 field termios.c_iflag	\ input mode flags
  4 field termios.c_oflag	\ output mode flags
  4 field termios.c_cflag	\ control mode flags
  4 field termios.c_lflag	\ local mode flags
  1 field termios.c_line	\ line discipline
  NCCS field termios.c_cc	\ control characters
  3 field termios.padding	\ C aligns everything to 32-bits
  4 field termios.c_ispeed	\ input speed
  4 field termios.c_ospeed	\ output speed
end-struct
\ *]

also system definitions

create baudtab	\ -- addr
  0 ,		B0 ,
  50 ,		B50 ,
  75 ,		B75 ,
  110 ,		B110 ,
  134 ,		B134 ,
  150 ,		B150 ,
  200 ,		B200 ,
  300 ,		B300 ,
  600 ,		B600 ,
  1200 ,	B1200 ,
  1800 ,	B1800 ,
  2400 ,	B2400 ,
  4800 ,	B4800 ,
  9600 ,	B9600 ,
  19200 ,	B19200 ,
  38400 ,	B38400 ,
  57600 ,	B57600 ,
  115200 ,	B115200 ,
  230400 ,	B230400 ,
((
  460800 ,	B460800 ,
  500000 ,	B500000 ,
  576000 ,	B576000 ,
  921600 ,	B921600 ,
  1000000 ,	B1000000 ,
  1152000 ,	B1152000 ,
  1500000 ,	B1500000 ,
  2000000 ,	B2000000 ,
  2500000 ,	B2500000 ,
  3000000 ,	B3000000 ,
  3500000 ,	B3500000 ,
  4000000 ,	B4000000 ,
))
  -1 ,		-1 ,

: getBaud#	\ hertz -- u|-1
\ translate a baud rate in hertz to the OS X baud rate number.
  baudtab begin				\ -- hertz addr
    dup @ -1 <>
   while
    2dup @ <>
   while
    2 cells +
  repeat then
  nip cell+ @
;

: setAttr	\ fildes termios -- ior
  TCSADRAIN swap tcsetattr dup
  if  drop  errno @  endif
;

: (setBaud)	\ hertz termios --
  swap getBaud#
  2dup cfsetispeed drop  cfsetospeed drop
;

: (setParity)	\ fildes bits -- ior ; 0=success
  { | ti[ /termios ] -- }
  over ti[ tcgetattr
  if  2drop errno @ exit  endif
  ti[ termios.c_cflag @
  PARENB PARODD or invert and or	\ merge parity bits
  ti[ termios.c_cflag !
  ti[ setAttr
;

: (setData)	\ fildes bits -- ior ; 0=success
  { | ti[ /termios ] -- }
  over ti[ tcgetattr
  if  2drop errno @ exit  endif
  ti[ termios.c_cflag @
  CSIZE invert and or		\ merge data bits
  ti[ termios.c_cflag !
  ti[ setAttr
;

: (setStop)	\ fildes bits -- ior ; 0=success
  { | ti[ /termios ] -- }
  over ti[ tcgetattr
  if  2drop errno @ exit  endif
  ti[ termios.c_cflag @
  CSTOPB invert and or		\ merge data bits
  ti[ termios.c_cflag !
  ti[ setAttr
;

also forth definitions
: setBaud	\ hertz fildes -- ior ; 0=success
\ *G Set the baud rate for an opened file descriptor.
  { | ti[ /termios ] -- }
  tuck ti[ tcgetattr
  if  2drop errno @ exit  endif
  ti[ (setBaud)				\ -- fildes
  ti[ setAttr
;

: setParity	\ char fildes -- ior ; 0=success
\ *G Set the parity for an opened file descriptor.
\ ** The character must be one of N,E,O.
  case  swap
    [char] N of  0 (setParity)  endof
    [char] E of  PARENB (setParity)  endof
    [char] O of  PARENB PARODD or (setParity)  endof
       2drop -1
  end-case
;

: setData	\ u fildes -- ior ; 0=success
\ *G Set the data size for an opened file descriptor.
\ ** The data size *\i{u} must be one of 5,6,7,8.
  case  swap
    5 of  CS5 (setData)  endof
    6 of  CS6 (setData)  endof
    7 of  CS7 (setData)  endof
    8 of  CS8 (setData)  endof
       2drop -1
  end-case
;

: setStop	\ u fildes -- ior ; 0=success
\ *G Set the number of stop bits for an opened file descriptor.
\ ** The value of *i{u} must be one of 1 or 2.
  case  swap
    1 of  0 (setStop)  endof
    2 of  CSTOPB (setStop)  endof
       2drop -1
  end-case
;

: setDTR	\ flag fildes -- ior ; 0=success
\ *G Set DTR inactive if *\i{flag} is zero, otherwise set it active.
  { | temp[ cell ] -- }
  TIOCM_DTR temp[ !
  swap if  TIOCMBIS  else  TIOCMBIC  endif
  temp[ ioctl
;

: setRTS	\ flag fildes -- ior ; 0=success
\ *G Set RTS inactive if *\i{flag} is zero, otherwise set it active.
  { | temp[ cell ] -- }
  TIOCM_RTS temp[ !
  swap if  TIOCMBIS  else  TIOCMBIC  endif
  temp[ ioctl
;

: setUnix	\ sid --
\ *G Set the line to have Unix line handling.
  #10 over ss.lechar !			\ line end character
  #13 over ss.ignchar !			\ ignore character
  eol$ count rot ss./le 2!		\ native mode CR
;

: setDOS	\ sid --
\ *G Set the line to have Windows/DOS line handling.
  #13 over ss.lechar !			\ line end character
  #10 over ss.ignchar !			\ ignore character
  crlf$ count rot ss./le 2!		\ native mode CR
;
previous definitions

vocabulary SerialCfg
' SerialCfg voc>wid constant widSerialCfg

: SerCfg:	\ xt -- ; sid ior x -- sid ior'
\ Define a serial configuration word whose action is given by xt.
\ At runtime, the word is executed if ior is zero, otherwise the
\ data value is discarded. The effect of this is that a chain of
\ words takes no further action after the first failure.
  create  ,  does> @ >r
  over if
    drop  r> drop				\ do nothing if already failed
  else
    nip  over gen-handle @ r> execute		\ perform action
  endif
;

also SerialCfg definitions

' setBaud SerCfg: baud		\ sid ior baud -- sid ior'
\ *G Used in the configuration string to set the baud rate, e.g.
\ *C   9600 baud
' setData SerCfg: data		\ sid ior u -- sid ior'
\ *G Used in the configuration string to set the number of data bits, e.g.
\ *C   8 data
char N constant no
char E constant even
char O constant odd
' setParity SerCfg: parity	\ sid ior u -- sid ior'
\ *G Used in the configuration string to set parity, where *\i{u}
\ ** is one of the characters N, E, or O. Constants are defined, e.g.
\ *C   no parity
\ *C   even parity
\ *C   odd parity
' setStop SerCfg: stop		\ sid ior u -- sid ior'
\ *G Used in the configuration string to set the number of stop bits, e.g.
\ *C   1 stop
\ *C   2 stop
: 8n1		\ sid ior -- sid ior'
\ *G Used in the configuration string to set the most common case,
\ ** 8 data bits, no parity, 1 stop bit, e.g.
\ *C   8n1
  8 data  no parity  1 stop
;
' setDTR SerCfg: DTR		\ sid ior flag -- sid ior'
\ *G Used in the configuration string to set the DTR line, where
\ ** *\i{flag} is non-zero for active and zero for inactive.
\ *C   1 DTR
' setRTS SerCfg: RTS		\ sid ior flag -- sid ior'
\ *G Used in the configuration string to set the RTS line, where
\ ** *\i{flag} is non-zero for active and zero for inactive.
\ *C   1 RTS
: Unix		\ sid ior -- sid ior
\ *G Used in the configuration string.
\ ** Set the serial line to use LF as the line terminator sequence.
\ ** CR characters will be ignored by *\fo{ACCEPT}.
  over setUnix  ;
: DOS		\ sid ior -- sid ior
\ *G Used in the configuration string.
\ ** Set the serial line to use CR/LF as the line terminator sequence.
\ ** This can also be used for Macs before OS X, but LF characters will be
\ ** ignored by *\fo{ACCEPT}.
  over setDOS  ;
previous definitions

: (openSer)	\ addr len attribs sid -- sid ior
\ The basic open function.
  >r >r -leading -trailing r> open-file		\ -- handle ior ; R: -- sid
  tuck if  drop -1 endif			\ -- ior handle ; R: -- sid
  r@ gen-handle !  r> swap
;

[undefined] csplit [if]
: csplit	\ addr len char -- raddr rlen laddr llen
\ Extract a substring at the start of addr/len, returning
\ the string raddr/rlen which includes char (if found) and
\ the string laddr/llen which contains the text to left of char.
\ If the string does not contain the character, raddr is
\ addr+len and rlen=0.
  >r 2dup r> scan			\ -- addr len addr' len'
  2swap 2 pick -			\ -- addr' len' addr len-len'
;
[then]

\ SFP002...
IGNBRK BRKINT or ISTRIP or INLCR or IGNCR or ICRNL or ( IUCLC or )
IXON or IXOFF or
  constant ~iflags
OPOST ( OLCUC or )  ONLCR or  ONOCR or  ONLRET or  OFILL or  OFDEL or  OXTABS or
  constant ~oflags
CRTSCTS
  constant ~cflags
CREAD
  constant cflags
ISIG ICANON or ECHO or ( DEFECHO or ) PENDIN or
  constant ~lflags

: setRAW	\ sid --
  { | ti[ /termios ] -- }
  gen-handle @
  dup ti[ tcgetattr drop
  ~iflags ti[ termios.c_iflag bic!
  ~oflags ti[ termios.c_oflag bic!
  ~cflags ti[ termios.c_cflag bic!
  cflags ti[ termios.c_cflag or!
  ~lflags ti[ termios.c_lflag bic!
  1 ti[ termios.c_cc VMIN + c!
  0 ti[ termios.c_cc VTIME + c!
  TCSANOW ti[ tcsetattr drop
;
\ ...SFP002

: processTail	\ caddr len sid ior -- sid ior'
\ Intepret the string using the serial configuration
\ vocabulary.
  get-order n>r  widSerialCfg 1 set-order
  base @ >r decimal
  2swap ['] assess catch if		\ string EVALUATEd ( sid 0 -- sid ior )
    2drop drop -1
  endif
  r> base !
  nr> set-order
;

also SerialCfg
: open-Ser    \ addr len attribs sid -- sid ior
\ *G The string *\i{caddr/len} is split into two. The space delimited
\ ** left hand side is used as the device, e.g. "/dev/ttyS4" which is
\ ** opened in raw mode. A default set up of 115200 baud, n81 and Unix
\ ** line handling is applied, and then the right hand side of the string
\ ** is parsed. Only the words documented as available in the serial
\ ** configuration string may be used.
  2>r bl csplit 2r> (openSer) 		\ -- raddr rlen sid ior
  dup if  2swap 2drop  exit  endif
  over setRAW				\ SFP002
  #115200 baud  8n1  Unix
  processTail
;
previous

: close-Ser	\ sid -- ior
  dup gen-handle @ close-file  -1 rot gen-handle !
;

: key?-ser	\ sid -- n
  { | n[ cell ] -- }
  n[ off
  gen-handle @ FIONREAD n[ ioctl -1 =
  if  0  else  n[ @  endif
;

: key-ser	\ sid -- char
  { | buff[ cell ] -- }
  begin  dup key?-Ser 0= while  1 ms  repeat
  buff[ 1 rot gen-handle @ read-file 2drop
  buff[ c@
;

: read-ser	\ addr len sid -- ior
  -rot bounds
  ?do  dup key-ser i c!  loop
  drop  0
;

: write-ser	\ addr len sid -- ior
  gen-handle @ write-file
;

: ekey?-ser	\ sid -- flag
  key?-ser
;

: ekey-ser	\ sid -- echar
  key-ser
;

: (type-ser)	\ caddr len sid --
  gen-handle @ write-file drop
;

: emit-ser	\ char sid --
  { | buff[ cell ] -- }
  swap buff[ c!  1 out +!
  buff[ 1 rot (type-ser)
;

: emit?-ser	\ sid -- flag
  drop  true  ;

: type-ser	\ caddr len sid --
  over 0> if
    over out +!  (type-ser)
  else
    2drop drop
  then
;

: bs-ser	\ sid -- ; destructive on display devices
  #8 swap emit-ser
;

\ for native & Linux use EOL$, for MAC and DOS use CRLF$
: cr-ser	\ sid --
  dup ss./le 2@ rot (type-ser)
  out off  op-line# incr
;

: lf-ser	\ sid --
  cr-ser
;

: ff-ser		\ sid -- ; page/cls on display devices
  cr-ser
;

: accept-ser	{ address maxlen sid | len -- len' }
  0 -> len
  address maxlen bounds ?do
    case  sid key-ser
      dup #127 = if  drop #8  then	\ treat <DEL> as <BS>
      dup #9 = if  drop bl  then	\ treat <TAB> as space

      #8 of                             \ <BACKSPACE>
        len if				\ if not at left edge
          1 sub len			\ decrement length
          out @
          #8 emit  #32 emit  #8 emit
          1- out !
          -1                            \ +LOOP modification -1
        else
          0                             \ Don't increment Loop
        then
      endof

      sid ss.lechar @ of		\ <CR> for DOS/Win, <LF> for Unices ; SFP005
        bl dup emit i c!  leave		\ output a <SPACE>
      endof

      sid ss.ignchar @ of  0  endof	\ <LF> for DOS/Win, <CR> for Unices ; SFP005

        dup emit  i c!  1 add len  1	\ DEFAULT

    end-case
  +loop

  len
;

: bell-ser	\ sid -- ; audible beeper
  drop  ;

: setpos-ser	\ x y mode sid -- ior
  2drop 2drop 0  ;

: getpos-ser	\ mode sid -- x y ior
  2drop  0 0 0  ;

: ioctl-ser	\ addr len fn sid -- ior
\ *G The serial *\fo{ioctl} functions provide control over the
\ ** serial line outputs and Unix/DOS mode handling. Where
\ ** parameters are shown as *\i{??}, their value is ignored.
\ *E \ ?? ?? #50 sid -- ior ; Unix mode, LF
\ ** \ ?? ?? #51 sid -- ior ; DOS/Windows mode, CR/LF
\ ** \ ?? ?? #52 sid -- ior ; Mac mode, CR
\ ** \ ?? ?? #53 sid -- ior ; native mode, LF for Unices
\ ** \ caddr len #55 sid -- ior ; set string for CR.
\ ** \ linechar ignchar #56 sid -- ior ; set input chars for ACCEPT
\ ** \ ?? flag #60 sid -- ior ; set DTR, nz=active
\ ** \ ?? flag #61 sid -- ior ; set RTS, nz=active
  case  swap
    #50 of	\ -- ?? ?? sid ; Unix mode, LF
      nip nip setUnix 0
    endof
    #51 of	\ -- ?? ?? sid ; DOS mode, CR/LF
      nip nip setDOS 0
    endof
    #52 of	\ -- ?? ?? sid ; MAC mode, CR
      nip nip setDOS 0
    endof
    #53 of	\ -- ?? ?? sid ; native mode, LF
      nip nip setUnix 0
    endof
    #55 of	\ -- caddr len sid ; set CR string
      ss./le 2!  0
    endof
    #56 of	\ -- linechar ignchar sid ; set input terms
      tuck ss.ignchar !  ss.lechar !  0
    endof
    #60 of	\ -- ?? flag sid ; set DTR
      rot drop  gen-handle @ setDTR  0
    endof
    #61 of	\ -- ?? flag sid ; set RTS
      rot drop  gen-handle @ setRTS  0
    endof
      2drop 2drop  0
  end-case
;

: flushOP-ser	\ sid -- ior
  drop 0  ;

: RFU-ser	\ --
  1 abort" RFU not supported"  ;

: init-ser	\ addr len sid -- ior
  drop 2drop  0  ;

: term-ser	\ sid -- ior
  drop  0  ;

: config-ser	\ sid -- ior ; produces a dialog
  drop  0  ;

create serdev-vectors
  ' Open-ser ,
  ' Close-ser ,
  ' Read-ser ,
  ' Write-ser ,
  ' Key-ser ,
  ' Key?-ser ,
  ' Ekey-ser ,
  ' EKey?-ser ,
  ' Accept-ser ,
  ' Emit-ser ,
  ' Emit?-ser ,
  ' Type-ser ,
  ' CR-ser ,
  ' LF-ser ,
  ' FF-ser ,
  ' BS-ser ,
  ' Bell-ser ,
  ' setpos-ser ,
  ' Getpos-ser ,
  ' Ioctl-ser ,
  ' FlushOP-ser ,
  ' RFU-ser ,
  ' Init-ser ,
  ' Term-ser ,
  ' config-ser ,

previous definitions


\ ==================
\ *H Device Creation
\ ==================

also system

: initSerDev	\ sid --
\ *G Initialise the sid for a serial device.
  dup /serial-sid erase
  -1 over gen-handle !			\ handle=unopened
  serdev-vectors over gen-vector !	\ vectors
  setUnix				\ line mode is Unix
;

: serdev:        \ "name" -- ; Exec: -- sid
\ *G Create a Serial Port based Generic IO device in the dictionary.
\ *C  serdev: <name>
  create
    here  /serial-sid allot  initSerDev
;

previous


\ ======================
\ *H OS X serial devices
\ ======================
\ *P When using USB serial devices, the name used varies according
\ ** to the function. The names are:
\ *C   /dev/tty.usb??????
\ *C   /dev/cu.usb?????
\ *P You can list them at a Terminal prompt with
\ *C   ls /dev/tty*
\ *C   ls /dev/cu*

\ *P The *\i{/dev/tty*} devices are for modems waiting for a call into
\ ** the OS X machine. It is assumed that the DCD line is active.
\ ** Hence these are of little use for the three wire connections
\ ** (RX, TX, Gnd) typically used to connect to embedded systems.

\ *P The *\i{/dev/cu*} devices are much better suited for
\ ** connecting out (calling up) to other systems.


\ ==================================
\ *H Mac serial terminal emulators
\ ==================================
\ *P The ones listed here are just ones recommended by others.

\ *P screen - on every Mac. For hardcore Unix buffs.
\ *C   screen /dev/cu.usbserial 19200
\ *C   http://hints.macworld.com/article.php?story=20061109133825654

\ *P Coolterm - GUI app.
\ *C   http://freeware.the-meiers.org

\ *P goSerial - GUI app.
\ *C   http://www.furrysoft.de/?page=goserial

\ *P The most widely used equivalent to Windows' HyperTerm
\ ** appears to be *\i{minicom}. It isn't pretty, but it works
\ ** and is easy to use.
\ *C   http://pbxbook.com/other/mac-tty.html#minicom


\ ======
\ *> ###
\ ======

1 [if]		\ Test code
SerDev: sd	\ -- sid

create sercu$	\ -- addr
  ", /dev/cu.usbserial-FTFWRZP1  115200 baud  no parity 8 data 1 stop  1 dtr 1 rts DOS"

: tcu
  sercu$ count r/w sd open-gio . .
;

create sertty$	\ -- addr
  ", /dev/tty.usbserial-FTFWRZP1  115200 baud  no parity 8 data 1 stop  1 dtr 1 rts DOS"

: ttty
  sertty$ count r/w sd open-gio . .
;
[then]

0 [if]		\ Test code
SerDev: sd	\ -- sid
create ser$	\ -- addr
  ", /dev/ttyS0  115200 baud  no parity 8 data 1 stop  1 dtr 1 rts DOS"
[then]

decimal
