\ VFX Forth multitasker for 32 bit OS X

((
Copyright (c) 2000, 2002, 2003, 2004, 2008, 2011, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To Do
=====

Change History
==============
20120525 ROT006 Converted to use dispatch tasking - code from
		Roelf Toxopeus
20120523 MPE005 Added Ward's GET and RELEASE.
20120420 MPE004 Documentation tidy up
20120204 MPE003 Changed semaphore behaviours for OS X specifics.
20110204 MPE002 Converted to OSX from Linux code.
20080110 MPE001 Converted to Linux from Windows code.

The Forth TCB address is the taskid of the task for the Forth words.

At creation time, a task is passed its TCB as the activating parameter.
The first action of a task is to place its USER POINTER into the TCB
structure, and change TOS to 0 so that it is the default return value
if/when the task is terminated.
))


\ =========
\ *! tasker
\ *T Multitasker
\ =========


\ ===============
\ *S Introduction
\ ===============
\ *P As of VFX Forth for Mac v4.6, 25 May 2012, there has been
\ ** an overhaul of the multitasker. All attempts have been
\ ** abandoned to maintain compatibility with the cooperative
\ ** multitasker used in MPE embedded systems. The effects of
\ ** all retained words have not changed. The new tasker does not
\ ** affect most desktop applications, but does allow us to reduce
\ ** CPU utilisation and hence power consumption. The original
\ ** code is still present and may be found as *\i{Lib/Osx32/MultiOsx32.trad.fth}.

\ *P The multitasker supplied with VFX Forth is derived from the
\ ** multitasker provided with the MPE Forth cross compilers, v6.1
\ ** onwards. Using a multitasking system can greatly simplify
\ ** complex tasks by breaking them down into manageable chunks.
\ ** This chapter leads you through:
\ *(
\ *B initialising the multitasker
\ *B writing a task
\ *B communicating between tasks
\ *B handling events
\ *)

\ *P The multitasker source code is in the file
\ ** *\i{Lib/Osx32/MultiOsx32.fth}. Note that the full version of
\ ** this file with all switches set except for test code is
\ ** compiled as part of the second stage build, but is not
\ ** present by default in the kernel version of VFX Forth.

\ ================
\ *S Configuration
\ ================
\ *P The configuration of the multitasker is controlled by
\ ** constants that control what facilities are compiled:

[undefined] test-code? [if]
0 constant test-code?		\ -- n
\ *G The test code will be compiled if this constant is true
\ ** and has not already been defined.
[then]



\ ===============================
\ *S Initialising the multitasker
\ ===============================
\ *P The multitasker needs to be initialised before use. At
\ ** compile time you must define the tasks that your system
\ ** requires and at run-time, all the tasks must be initialised.

\ *P Before use the multitasker must be initialised by the word
\ ** *\fo{INIT-MULTI}, which initialises the primary task *\fo{MAIN},
\ ** and enables the multi-tasker.

\ *P To disable the multitasker, use *\fo{SINGLE}.

\ *P To enable the multitasker, use *\fo{MULTI}, which starts the
\ ** scheduler so new tasks can be added.

\ =================
\ *S Writing a task
\ =================
\ *P Tasks are very straightforward to write, but the way tasks
\ ** are scheduled needs to be understood. This implementation
\ ** uses the OS X *\b{pthreads} API, and so tasks are pre-emptively
\ ** scheduled. This is different from the cooperative scheduler used
\ ** by embedded systems. Despite this, the word *\fo{PAUSE} which
\ ** yields a timeslice is retained for compatibility, and
\ ** *\fo{PAUSE} is where the MPE event handling is incorporated.
\ *E : ACTION1     \ -- ; An example task
\ **   TASK0-IO                 \ select the console as the I/O device
\ **   DUP IP-HANDLE !  OP-HANDLE !
\ **   BEGIN                    \ Start an endless loop
\ **     [CHAR] * EMIT          \ Produce a character  )
\ **     1000 MS                \ Wait 1 second
\ **     PAUSE                  \ Needed!
\ **   AGAIN                    \ Go round again
\ ** ;
\ ** TASK TASK1    \ -- tcb ; name task, get space for it

\ *P The task name created by *\fo{TASK} is used as the task
\ ** identifier by all words that control tasks.

\ *N Task dependent variables
\ *P An area of memory known as the *\fo{USER} area is set aside
\ ** for each task. This is often called thread local
\ ** storage. This memory contains user variables which contain
\ ** task specific data. For example, the current number conversion
\ ** radix *\fo{BASE} is normally a user variable as it can vary
\ ** from task to task.

\ *P A user variable is defined in the form:
\ *E n USER <name>
\ *P where n is the nth byte in the user area. The word
\ ** *\fo{+USER} can be used to add a user variable of a given
\ ** size:
\ *E   <size> +USER <name>
\ *P The use of *\fo{+USER} avoids any need to know the offset at
\ ** which the variable starts.

\ *P A user variable is used in the same way as a normal variable.
\ ** By stating its name, its address is placed on the stack,
\ ** which can then be fetched using *\fo{@} and stored by *\fo{!}.

\ ====================
\ *S Controlling tasks
\ ====================
\ *P Tasks can be controlled in the following ways:
\ *(
\ *B activated
\ *B halted
\ *B restarted after it has been halted
\ *B terminated.
\ *)

\ *N Activating a task
\ *P A task is started by activating it. To activate a task, use
\ ** *\fo{INITIATE},
\ *E ' <action> <task> INITIATE
\ *P where *\fo{' <action>} gives the xt of the word to be run
\ ** and *\fo{<task>} is the task identifier. The task identifier
\ ** is used to control the task. Tasks defined by
\ ** *\fo{TASK <name>} return a task identifier when *\fo{<name>}
\ ** is executed.

\ *N Stopping a task
\ *P A task may be temporarily suspended. A task may also halt
\ ** itself. To temporarily stop a task, use *\fo{HALT}.
\ ** *\fo{HALT} is used in the form:
\ *E <task> HALT
\ *P where *\fo{<task>} is the task to be stopped. To restart a
\ ** halted task, use *\fo{RESTART} which is used in the form:
\ *E <task> RESTART
\ *P where *\fo{<task>} is the task to restart.

\ *P To stop the current task (i.e. stop itself) use *\fo{STOP( -- ).

\ *N Terminating a task
\ *P Terminating a task halts it, performs an optional clean up
\ ** action, and calls the operating system thread end function.
\ ** A thread must terminate itself, which leads
\ ** to some complexities. However, it does give the task an
\ ** opportunity to release any resources it may have allocated
\ ** (especially memory) at start up or during its execution. To
\ ** terminate a task use:
\ *E   <task> TERMINATE

\ *P Before the operating system thread end function is called,
\ ** the terminating task will execute its clean up code. The XT
\ ** of the clean up code is held in the task control block. If no
\ ** clean up action is required, zero is used.
\ *E   ... ['] CleanUp MyTask AtTaskExit ...

\ *P If you want a task to be a *\fo{BEGIN ... UNTIL} loop rather
\ ** than an endless loop, this is perfectly legal,
\ ** as returning from a thread will call the clean up code and
\ ** then the *\f{ExitThread} function. However, you must define
\ ** an exit code before you return from the task. Note that on
\ ** entry to the task there will already be a 0 on the stack.
\ *E : MyTask	\ 0 -- exitcode
\ **   <initialisation>         \ initialise task resources
\ **   begin                    \ round and round until done
\ **     <actions>  MyDone @
\ **   until
\ **   drop 0                   \ paranoid, return 0 as success
\ ** ;

\ *P Unlike MPE's embedded systems, under OS X you cannot
\ ** predict how long a task will take to start after *\fo{INITIATE}
\ ** or shut down after *\fo{TERMINATE}.


\ ====================
\ *S Critical sections
\ ====================
\ *P Sometimes the multitasker has to be inhibited so that other
\ ** tasks are not run during critical operations that would
\ ** otherwise cause the scheduler to operate. This is achieved
\ ** using the words *\fo{SINGLE} and *\fo{MULTI}. Note that these
\ ** do *\{not} stop the OS X scheduler, only the MPE extensions.
\ ** If a full critical section is required, see the semaphore
\ ** source to find out how to use the critical section API.
\ *E SINGLE   -- ; inhibit tasker
\ ** MULTI    -- ; restart tasker

\ *P A *\fo{SEMAPHORE} is a structure used for signalling between
\ ** tasks, and for controlling resource usage. It has two fields,
\ ** a counter (cell) and an owner (taskid, cell). The counter
\ ** field is used as a count of the number of times the resource
\ ** may be used, and the owner field contains the TCB of the task
\ ** that last gained access. This field can be used for priority
\ ** arbitration and deadlock detection/arbitration.

\ *P This design of a semaphore can be used either to lock a
\ ** resource such as a comms channel or disc drive during access
\ ** by one task, or as a counted semaphore controlling access to
\ ** a buffer. In the second case the counter field contains the
\ ** number of times the resource can be used. Semaphores are
\ ** accessed using *\fo{SIGNAL} and *\fo{REQUEST}.

\ *P *\fo{SIGNAL} increments the counter field of a semaphore,
\ ** indicating either that another item has been allocated to
\ ** the resource, or it is available for use again, 0 indicating
\ ** that it is in use by a task.

\ *P *\fo{REQUEST} waits until the counter field of a semaphore
\ ** is non-zero, and then decrements the counter field by one.
\ ** This allows the semaphore to be used as a COUNTED semaphore.
\ ** For example a character buffer may be used where the semaphore
\ ** counter shows the number of available characters. Alternatively
\ ** the semaphore may be used purely to share resources. The
\ ** semaphore is initialised to one. The first task to *\fo{REQUEST}
\ ** it gains access, and all other tasks must wait until the
\ ** accessing task *\fo{SIGNAL}s that it has finished with the
\ ** resource.

\ ========================
\ *S Multitasker internals
\ ========================
\ *P A multitasker tries to simulate many processors with just one
\ ** processor. It works by rapidly switching between each task.
\ ** On each task switch it saves the current state of the
\ ** processor, and restores the state that the next task needs.
\ ** The Forth multitasker creates a task control block for each
\ ** task. The task control block (TCB) is a data structure which
\ ** contains information relevant to a task.

\ *S A simple example
\ *P The following example is a simple demonstration of the
\ ** multitasker. Its role is to display a hash `#' every so
\ ** often, but leaving the foreground Forth console running. To
\ ** use the multitasker you must compile the file *\i{LIB\MULTIWIN32.FTH}
\ ** into your system. Note that the file has already been compiled
\ ** by the Studio IDE in *\i{VfxForth.exe}, but is not present in
\ ** *\i{VfxBase.exe}.

\ *P The following code defines a simple task called *\fo{TASK1}.
\ ** It displays a '$' character every second.

\ *E VARIABLE DELAY     \ time delay between #'s in milliseconds
\ **   1000 DELAY !             \ initialise time delay
\ ** : ACTION1          \ -- ; task to display #'s
\ **   TASK0-IO                 \ select the console as the I/O device
\ **   DUP IP-HANDLE !  OP-HANDLE !
\ **   [CHAR] $ EMIT            \ Display a dollar
\ **   BEGIN                    \ Start continuous loop
\ **     [CHAR] # EMIT          \ Display a hash
\ **     DELAY @ MS             \ Reschedule Delay times
\ **     PAUSE                  \ At least one per loop
\ **   AGAIN                    \ Back to the start ...
\ ** ;
\ *P The use of *\fo{PAUSE} in this example is not actually required
\ ** as *\fo{MS} periodically calls *\fo{PAUSE}.

\ *P Before any tasks can be activated, the multitasker must be
\ ** initialised. This is done with the following code:
\ *E INIT-MULTI
\ *P The word *\fo{INIT-MULTI} initialises all the multitasker's
\ ** data structures and starts multitasking. This word need only
\ ** be executed once in a multitasking system and is usually
\ ** executed at start up.

\ *P Note that on entry to a task, the stack depth will be 1. This
\ ** happens because OS X requires a return value when a task
\ ** terminates, and a value of zero is provided by the task
\ ** initialisation code.

\ *P To run the example task, type:
\ *E TASK TASK1
\ ** ASSIGN ACTION1 TASK1 INITIATE

\ *P This will activate *\fo{ACTION1} as the action of task
\ ** *\fo{TASK1}. Immediately you will see a dollar and a hash
\ ** displayed. If you press <return> a few times, you notice
\ ** that the Forth interpreter is still running. After a few
\ ** seconds another hash character will appear. This is the
\ ** example task working in the background.

\ *P The example task can be controlled in several ways:
\ *(
\ *B the rate of generation of hashes can be changed
\ *B it can be halted
\ *B once halted it can be restarted
\ *B it can be started from scratch
\ *)

\ *P Changing the variable *\fo{DELAY} can change the rate of
\ ** production of hashes. Try:
\ *E 2000 DELAY !
\ *P This changes the number of milliseconds between displaying
\ ** hashes to 2000 milliseconds.
\ ** Therefore the rate of displaying hashes halves.

\ *P Typing the task name followed by *\fo{HALT} halts the task:
\ *E TASK1 HALT
\ *P You notice that the hashes are not displayed any more.

\ *P The task is restarted by *\fo{RESTART}. Type:
\ *E TASK1 RESTART
\ *P You notice that the hashes are displayed again.

\ *P To restart the task from scratch, just kill it and activate
\ ** it again:
\ *E TASK1 TERMINATE
\ ** ASSIGN ACTION1 TASK1 INITIATE
\ *P You notice the dollar and the hash are displayed, followed by
\ ** more hashes.


\ *************************
\ *S Structures and support
\ *************************

struct /pthread_attr_t	\ -- len
\ *G Equivalent of *\fo{pthread_attr_t} structure.
  int pat.detachstate
  int pat.policy
  int pat.priority
  int pat.inheritsched
  int pat.scope
  int pat./guard
  int pat.stackset
  int pat.stackaddr
  int pat./stack
end-struct

#38 constant /tcb.callback	\ -- len
\ *G Size of the task callback data and code. Used for error checks.

struct /TCB	\ -- size
\ *G Returns the size of a TCB structure, which controls the task.
\ *[
  int tcb.link		\ link to next task ; MUST BE FIRST	 0
  int tcb.hthread	\ task handle			         4
  int tcb.up		\ user pointer			         8
  int tcb.pumpxt	\ user hook				12
  int tcb.status	\ status bits			        16
  int tcb.clean		\ xt of clean up handler		20
  /tcb.callback field tcb.callback \ task callback structure	24
  aligned		\ force to cell boundary
end-struct
\ *]
\ *P The task status cell reserves the low 8 bits for use by VFX
\ ** Forth. The other bits may be used by your application.
\ *E Bit    When set                       When Reset
\ ** 0      task running                   task halted
\ ** 1..7   RFU                            RFU
\ ** 8..31  User defined                   User defined

\ bit masks for the status cell
$0001 constant run-mask		\ running
  run-mask invert constant ~run-mask
  0 constant run-bit#

(( left here for DOCGEN use
cell +USER ThreadExit?	\ -- addr
\ *G Holds a non-zero value to cause the thread to exit.
  0 ThreadExit? !			\ for main

cell +USER ThreadTCB	\ -- addr
\ *G Holds a pointer to the thread's TCB.
  0 ThreadTCB !				\ for safety

cell +USER ThreadSync	\ -- addr
\ *G Holds bit patterns used for intertask synchronisation.
\ ** See later section.
  0 ThreadSync !
))

: AtTaskExit	\ xt tcb -- ; set task exit action
\ *G Sets the given task's cleanup action. Use in the form:
\ *C   ' <action> <task> AtTaskExit
  tcb.clean !
;

: perform	\ addr --
\ *G Execute contents of addr if non-zero. The non-zero contents of
\ ** *\i{addr} are *\fo{EXECUTE}d.
  @ ?dup
  if  execute  endif
;


\ *****************************
\ *S Task definition and access
\ *****************************

create main	\ -- addr ; tcb of main task
\ *G The task structure for the first task run, usually the console
\ ** or the main application.
  /TCB allot
  main ThreadTCB !

: add-task      \ task -- ; insert into list
  main tcb.link @			\ save task currently pointed to
  over tcb.link !			\ and make new task point to it
  main tcb.link !			\ current task points to new task
;

: sub-task      \ task -- ; remove task from chain
\ look for task which points to required task
  main					\ head of chain ; -- target current
  begin
    2dup tcb.link @ <>			\ while curr. does not point to target
   while
    tcb.link @  dup main =		\ get next task, check back at self
    if  2drop  exit  endif		\   and get out if so
  repeat                                \ -- target points-to-it
\ set task that points to target to point to task pointed to by the target
  swap tcb.link @			\ task pointed to by target
  swap tcb.link !			\ set into link field
;

: initThread	\ tcb -- 0 ; set up new task - TCB provided by task creator
  dup tcb.hThread @ 0= if		\ if no handle defined yet
    pthread_self over tcb.hThread !	\   use a temporary thread handle
  endif
  up@ over tcb.up !			\ user pointer
  dup ThreadTCB !			\ pointer to this task's TCB
  ThreadExit? off			\ don't exit task
  ThreadSync off			\ no synchronisation yet
  1 over tcb.status !			\ task is running
  drop 0
;

: termThread	\ stat -- stat ; perform task termination actions
  ThreadTCB @ tcb.clean perform
  ThreadTCB @ dup tcb.up off		\ mark TCB as inactive
  sub-task				\ remove from active list
;

: !++		\ addr n -- addr'
  over !  cell +
;

: !call++	\ addr target -- addr'
  over !call skipCall
;

: InitTCB	\ addr --
\ *G Initialise a task control block at addr.
  dup /tcb erase
  tcb.callback			\ -- addr' ; start of callback area
  dup 3 cells + !++		\ pointer to callback entry		4  0
  1 !++  1 !++			\ #out #in				8  4
  ['] callback-entry !call++	\ call to entry code		        5  12
  ['] initThread !call++	\ call thread initialiser		5  17
  ['] callback-crash !call++	\ call action word		        5  22
  ['] termThread !call++	\ call to thread terminator	        5  27
  ['] callback-exit !call++	\ call to exit code		        5  32
  $C3 swap c!			\ ret					1  37
;				\                                          38
0 tcb.callback #22 + constant >task.callAction	\ offset to CALL action
main InitTCB
$0001 main tcb.status or!
pthread_self main tcb.hthread !
main ThreadTCB !

: task		\ -- ; -- addr ; define task, returns TCB address
\ *G Use in the form *\fo{TASK <name>}, creates a new task data
\ ** structure called *\fo{<name>} which returns the address of
\ ** the data structure when executed.
  create
    here  /TCB allot  initTCB
;				\ 					   41

: Self		\ -- tcb|0 ; returns TCB of current task
\ *G Returns the task id (TCB) of the current task. If called
\ ** outside a task, zero is returned.
  ThreadTCB @
;

: his		\ task uservar -- addr
\ *G Given a task id and a *\fo{USER} variable, returns the
\ ** address of that variable in the given task. This word is
\ ** used to set up *\fo{USER} variables in other tasks. Note
\ ** that the task must be running.
  up@ -  swap tcb.up @ +
;


\ ***************************
\ *S Task handling primitives
\ ***************************

0 value multi?	\ -- flag
\ *G Returns true if the tasker is enabled.

: single	\ -- ; disable scheduler
\ *G Disables the Forth portions of the scheduler, but does not
\ ** disable OS X scheduling.
  0 to multi?
;

: multi		\ -- ; enable scheduler
\ *G Enables the Forth portions of the scheduler, but does not
\ ** disable OS X scheduling.
  -1 to multi?
;

(( for DocGen
defer pause	\ --
\ *G *\fo{PAUSE} is the software entry to the pre-emptive scheduler,
\ ** and should be called regularly by all tasks.
\ ** The phrase *\fo{sched_yield drop} occurs at the end of the
\ ** default action *\fo{(PAUSE)}. If the task needs more than
\ ** this and does not use one of the existing message loop words
\ ** such as *\fo{IDLE}, place the XT of the message pump word
\ ** in offset *\fo{TCB.PUMPXT} of the Task Control Block and that XT
\ ** will be called once every time *\fo{PAUSE} is called.
\ ** Because of the way OS X works, *\fo{PAUSE} also controls
\ ** task closure. A task that does not call *\fo{PAUSE} cannot
\ ** be safely terminated except by the task itself, or by a call
\ ** to the API function *\f{kill()}. A task that calls *\fo{PAUSE}
\ ** in a loop without calling any delay mechanism will cause
\ ** CPU hogging.
))

also system
[-short-branches
: (pause)	\ -- ; the scheduler itself
\ *G The action of *\fo{PAUSE} after the multitasker has been
\ ** compiled. If not called from a task, this is a *\fo{NOOP}.
\ ** If *\fo{SINGLE} has been set, no action is taken.
\ ** If *\fo{MULTI} is set, the action is *\b{sched_yield}
\ ** until the task status is non-zero.
  multi? if				\ Do nothing if MPE portion disabled
    ThreadTCB @ if			\ Do almost nothing if not in a task
      ThreadExit? @ -1 = if
        termThread			\ run clean up action
        pthread_self pthread_mach_thread_np thread_terminate drop
        \ 0 pthread_exit			\ should never come back!
      endif
      sched_yield drop
      ThreadTCB @ tcb.pumpxt perform	\ not actually used in OS X
      ThreadTCB @ tcb.status @ if	\ task active
        sched_yield drop
      else
        ThreadTCB @ tcb.hthread @ pthread_mach_thread_np thread_suspend drop
      endif
    endif
  endif
;
assign (pause) to-do pause
short-branches]
previous

: restart	\ tcb -- ; mark task TCB as running
\ *G If the task has been initiated but is now *\fo{HALT}ed or
\ ** *\fo{STOP}ped, it will be restarted.
  dup tcb.status @ $0001 and if		\ already running
    drop pause
  else
    $0001 over tcb.status or!		\ set the run bit in the TCB
    tcb.hthread @ pthread_mach_thread_np thread_resume drop
  endif
;

: halt		\ tcb -- ; mark thread as halted
\ *G Stops an *\fo{INITIATE}d task from running until
\ ** *\fo{RESTART} is used.
  dup tcb.status @ $0001 and if
    $0001 over tcb.status bic!		\ clear the run bit in the TCB
    tcb.hthread @ pthread_mach_thread_np thread_suspend drop
  else
    drop pause
  endif
;

: stop          \ -- ; halt oneself
\ *G *\fo{HALT}s the current task.
  self halt  pause
;

: running?	\ tcb -- u
\ *G Returns the task's run status, where non-zero indicates
\ ** that it is running.
  tcb.status @
;


\ ******************
\ *S Task management
\ ******************

/pthread_attr_t buffer: defThreadAttribs
\ Holds structure retrieved at initialisation.

\ 3 global concurrent queues
2 constant DISPATCH_QUEUE_PRIORITY_HIGH
0 constant DISPATCH_QUEUE_PRIORITY_DEFAULT
-2 constant DISPATCH_QUEUE_PRIORITY_LOW

: defaultqueue	\ -- queue ; return default dispatch queue
  DISPATCH_QUEUE_PRIORITY_DEFAULT 0 dispatch_get_global_queue
;

: MakeThread	\ tcb -- ; start thread
  0 over tcb.hThread !			\ mark as no handle yet
  defaultqueue				\ dispatch queue
  swap					\ param is TCB
  dup tcb.callback @			\ entry point
  dispatch_async_f
;

: to-task	\ xt task -- ; set action of task
( *G Used in the form below to define a task's action:			)
( *C   assign <action> <task> to-task					)
  >task.callAction + !call
;

: to-pump	\ xt task -- ; set message loop of task
( *G Used in the form below to define rhe action of the message pump:	)
( *C   assign <action> <task> to-pump					)
  tcb.pumpxt !
;

: initiate	\ xt task -- ; start task from scratch
\ *G Initialises a task running the given xt. All required O/S
\ ** resources are allocated by this word.
  dup tcb.up off			\ no USER area yet
  tuck to-task				\ set action of task
  dup MakeThread			\ start task
  dup add-task                          \ add to chain
  begin					\ wait until user area set
    pause				\ let it run
    dup tcb.up @
  until
  drop
; doNotSin

: terminate	\ task -- ; stop task, and remove from list
\ *G Causes the specified task to die. You should not make
\ ** assumptions as to how long this will take. Unlike the
\ ** embedded systems implementations, this word is very
\ ** operating system dependent. The task may still be alive
\ ** on return from this call.
  dup tcb.up @ 0=			\ check if already terminated
  if  drop  exit  endif
  ThreadExit? his on
  pause					\ halt task, and remove from chain
;
\ *P *\b{N.B.} Do not use *\fo{self terminate} to cause a task to end.
\ ** Use the following instead:
\ *E : Suicide    \ -- ; terminate current task
\ **   ThreadExit? on  begin pause again
\ ** ;
\ **
\ ** ... Suicide

: start:	\ task -- ; exits from caller
\ *G *\fo{START:} is used inside a colon definition. The code
\ ** before *\fo{START:} is the task's initialisation, performed
\ ** by the current task. The code after *\fo{START:} up to the
\ ** closing *\fo{;} is the action of the task. For example:
\ *E   TASK FOO
\ **   : RUN-FOO
\ **     ...
\ **     FOO START:
\ **     ...
\ **     begin  ... pause  again
\ **   ;

\ *P All tasks must run in an endless loop, except for
\ ** initialisation code. There are exceptions to
\ ** this, and these are discussed in the section on terminating
\ ** a task. When *\fo{RUN-FOO} is executed, the code after
\ ** *\fo{START:} is set up as the action of task *\fo{FOO} and
\ ** started. *\fo{RUN-FOO} then exits. If you want to perform
\ ** additional actions after starting the task, you should use
\ ** I*\fo{INITIATE} to start the task.
  r> swap initiate
;

: TaskState	\ task -- state
\ *G Returns true if the task has started and zero if the thread
\ ** has finished.
  tcb.up @ 0<>
;

\ \ : WaitTaskState	\ task state --
\ \ *G Wait until the task is in a given state.
\ \ *E   task -1 WAITTASKSTATE \ wait until task starts
\ \ **   task 0 WAITTASKSTATE  \ wait until task closes
\   begin
\     pause
\     2dup swap TaskState <>
\   until
\ ;

: init-multi    \ -- ; initialisation of multi-tasking
\ *G Initialise the Forth multitasker to a state where only the
\ ** task *\fo{MAIN} is known to be running. *\fo{INIT-MULTI} is
\ ** added to the cold chain and is also called during
\ ** compilation of *\i{MultiOsx32.fth}. This word *\b{must}
\ ** be run from *\fo{MAIN}.
  defThreadAttribs pthread_attr_init drop	\ get defaults
  #512 kb defThreadattribs pat./stack !	\ set what we need
  main initThread drop			\ must setup semaphore
  main ThreadTCB !
  pthread_self main tcb.hThread !
  single				\ tasker disabled
  main main tcb.link !			\ initialise MAIN to point to itself
  multi					\ mark tasker enabled
;
' init-multi AtCold  init-multi

: term-multi	\ --
\ *G Performed in the exit chain when the program terminates.
\ ** closes all active tasks except *\fo{SELF}. This allows
\ ** all task clean-up actions to be performed before the
\ ** program itself finishes.
\ tell all tasks to finish
  main begin
    dup self <> if
      dup tcb.link @  swap terminate
    else
      tcb.link @
    endif
    dup main =
  until
  drop  #50 ms				\ let them finish
;
\ ' term-multi AtExit

: .task		\ task -- ; display task name
\ *G Given a task, e.g. as returned by *\fo{SELF}, display its
\ ** name or address.
  dup InForth? if
    ip>nfa .name
  else
    ." $" .dword
  endif
;

: .tasks	\ -- ; display active tasks
\ *G Display a list of all the active Forth tasks.
  cr  main
  begin
    dup .task  2 spaces  ?cr
    tcb.link @ dup main =
  until
  drop
;


\ ***********************
\ *S Task synchronisation
\ ***********************

$AAAA5555 constant TaskReady	\ -- n
\ *G At task initiation, *\fo{USER} variable *\fo{THREADSYNC} is
\ ** set to zero. Set *\fo{THREADSYNC} to this value to indicate
\ ** that the task is willing to synchronise with another task.

$5555AAAA constant TaskReadied	\ -- n
\ *G A synchronising task sets another task's *\fo{THREADSYNC}
\ ** to this value to indicate that synchronisation is
\ ** complete.

also system
: WaitForSync	\ --
\ *G Perform the slave synchronisation sequence.
  TaskReady ThreadSync !
  begin
    1 (OS_ms) pause
    ThreadSync @ TaskReadied =
  until
  ThreadSync off
;

: [Sync		\ task -- task
\ *G Used by a master task in the form:
\ *C   [Sync ... Sync]
\ *P to synchronise and pass data to another task, usually
\ ** when *\fo{USER} variables must be initialised. The slave
\ ** task must execute *\fo{WAITFORSYNC}.
  begin		\ wait for slave task to come ready
    1 (OS_ms) pause
    dup ThreadSync his @ TaskReady =
  until
;

: Sync]		\ task --
\ *G Used by a master task in the form:
\ *C   [Sync ... Sync]
\ *P to indicate the end of synchronisation.
  TaskReadied over ThreadSync his !	\ mark completion
  begin					\ wait for task to zero
    1 (OS_ms) pause  dup ThreadSync his @ 0=
  until
  drop
;
previous


\ *************
\ *S Semaphores
\ *************

struct /semaphore	\ -- len
\ *G Structure used for OS X i32 semaphores.
  int sm.count			\ compatibility with XC6 and Win32
  int sm.owner			\ compatibility with XC6 and Win32
  /semosx field sm.semosx	\ OS X semaphore
end-struct

: semaphore	\ -- ; -- addr [child]
\ *G A *\fo{SEMAPHORE} is an extended variable used for signalling
\ ** between tasks and for resource allocation. The counter field
\ ** is used as a count of the number of times the resource may be
\ ** used, and the arbiter field contains the TCB of the task
\ ** that last gained access. This field can be used for priority
\ ** arbitration and deadlock detection/arbitration.
  create
    /semaphore allot&erase
;

: InitSem	\ semaphore --
\ *G Initialise the semaphore. This *\b{must} be done before using it.
  dup 0 1 rot sm.count 2!
  sm.semosx InitCritSec
;

: ShutSem	\ semaphore --
\ *G Delete the critical section associated with the smaphore.
  sm.semosx TermCritSec
;

: LockSem	\ semaphore --
\ *G Lock the semaphore.
  sm.semosx [CritSec
;

: UnlockSem	\ semaphore --
\ *G Unlock the semaphore.
  sm.semosx CritSec]
;

: signal        \ sem -- ; increment counter field of semaphore,
\ *G *\fo{SIGNAL} increments the counter field of a semaphore,
\ ** indicating either that another item has been allocated to
\ ** the resource, or that it is available for use again, 0
\ ** indicating in use by a task.
  dup sm.count incr  dup sm.owner off	\ inc. counter, release
  UnlockSem
;

: request       \ sem -- ; get access to resource, wait if count = 0
\ *G *\fo{REQUEST} waits until the counter field of a semaphore
\ ** is non-zero, and then decrements the counter field by one.
\ ** This allows the semaphore to be used as a *\b{counted}
\ ** semaphore. For example a character buffer may be used where
\ ** the semaphore counter shows the number of available
\ ** characters. Alternatively the semaphore may be used purely
\ ** to share resources. The semaphore is initialised to one.
\ ** The first task to *\fo{REQUEST} it gains access, and all
\ ** other tasks must wait until the accessing task *\fo{SIGNAL}s
\ ** that it has finished with the resource.
  dup LockSem
  dup sm.count decr			\ got it, decrement counter
  self swap sm.owner !			\ mark resource as mine
;


\ *******************
\ *S Exclusive access
\ *******************
\ *P The words *\fo{GET} and *\fo{RELEASE} access a resource
\ ** variable to control access to a resource.

Extern: Uint32 OSAtomicCompareAndSwapInt(
  UInt32 Old, UInt32 new, UInt32 * addr
);

: GET		\ varaddr --
\ *G If the resource variable is not available, wait until we
\ ** have set it.
  >r  begin
    0 ThreadTCB @ r@ OSAtomicCompareAndSwapInt 0=	\ addr does not contain 0
    ThreadTCB @ dup r@ OSAtomicCompareAndSwapInt 0=	\ addr does not contain taskid
    and
   while
    pause
  repeat
  r> drop
;

: RELEASE	\ varaddr --
\ *G Release the resource variable if it belongs to this task.
  ThreadTCB @ 0 rot OSAtomicCompareAndSwapInt drop
;


\ *********
\ Test code
\ *********

test-code? [if]

op-handle @ value task0-io		\ device used at compile time

task task1				\ first task

#2200 value rate1			\ delay in ms between iterations

: default-io	\ -- ; set up default I/O
  task0-io dup op-handle ! ip-handle !
;

: action1	\ --
\ This code prints a '*' character every RATE1 milliseconds.
  default-io decimal			\ must initialise USER vars
  begin
    [char] * emit
    rate1 ms  pause			\ must PAUSE
  again
;
\ ' action1 task1 initiate		\ start TASK1

task task2

#1800 value rate2

: run-task2	\ --
\ This code prints a '/' character every RATE2 milliseconds.
  task2 start:
    default-io decimal			\ must initialise USER vars
    begin
      [char] / emit
      rate2 ms  pause			\ must PAUSE
    again
;

[then]


\ ======
\ *> ###
\ ======

decimal
