(* ====================================================================
Cocoa runners utilities

Copyright 2008 Roelf Toxopeus

VFX64 Version.
Part of the interface to ObjC 2 Runtime.
Last: 24 August 2020 at 10:14:27 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Before running the method, we need to change the Forth stackorder
to ObjC stackorder. An 'argument' stack is used to temporarely store
the arguments while re-shuffling the stack.

Reorder the arguments on the data stack for the ObjC calling mechanism:
Normal:
( n1 ... nn receiver selector n -- receiver selector n1 ... nn )
STRET, passing an extra address used for returning structures.
( n1 ... nn a receiver selector n+1 -- a receiver selector n1 .. nn )

These words could be optimised, allthough they work since the PPC days
without any noticable speed implications!
This is also used in coco-i4-64, which is not slower than coco-vfx which
uses an optimized version.
ObjC is in general not used for computing intensive and strict time cases.

-ARGP -- clears argument stack.
PUSHES --  pushes n parameters from data stack to the argument stack.
POPS -- pops n parameters from argument stack to data stack.
REORDER -- puts all parameters in the required ObjC parameter order.
REORDER-STRET -- same, but deals with the STructure RET pointer as well.
-------------------------------------------------------------------- ))

/forth
decimal

32 constant maxargs
create argstack here maxargs cells dup allot erase   \ argument stack
argstack maxargs cells + constant argbp   \ arguments stack base pointer
argbp value argp                          \ arguments stack pointer, top arg stack

: >arg ( n -- )   argp cell- dup to argp ! ;

: arg> ( -- n )   argp dup @ swap cell+ to argp ;

: -argp ( -- )   argbp to argp ;

: argdepth ( -- n ) argbp argp - 8 / ;

: ?argp ( -- )   argdepth maxargs > if -argp true abort" argument stack overflow!" then ;

: pushes ( x1 ... xn n -- )   0 ?do >arg loop ;

: pops ( n -- x1 ... xn )    0 ?do arg> loop ;

0 value #args

\ different from original one!!!!!!!!!!!!!!!!!!
\ only reorder the _int parameters!
: REORDER ( x1 ... xn receiver selector n -- receiver selector x1 ... xn )
   to #args -argp
\ clear stack
   2>r  #args pushes
\ fill stack
   2r>  #args pops
;

\ allthough stret is a parameter included in the #args, it gets a special treatment.
\ it has to go go on the bottom of the stack under the object and selector.
: REORDER-STRET ( x1 ... xn stret receiver selector n+1 -- stret receiver selector x1 ... xn )
   1- to #args      \ stret reshuffled by 'hand' 
   -argp
\ clear stack
   >r 2>r
   #args pushes
\ fill stack
   2r> r>  #args pops
;

(( --------------------------------------------------------------------
The code generator for the ObjC message sending mechanism.
The following words use parts of the VFX EXTERN: interface.

FORCOCOA -- parses the parameter list(s) and together with the given
member of the objc_senMsg family, initializes the structures needed
for the codegenerator COCOA,
COCOA, --  generates code for the cocoa message sending.
-------------------------------------------------------------------- ))

also extern.voc

\ a n is the name of a member of the objc_msgXXX family
\ check if it's defined, searching the import-func-link
\ return it's structure address or 0
: find-importstructure ( addr n -- struc )
   2>r
   import-func-link @
   begin
     dup
   while
     dup >funcname zcount 2r@ compare 0= if 2r> 2drop exit then
     >funclink @
   repeat
   drop 2r> 2drop 0
;

: /offsets ( -- )   initOffsets  CalcForthOff  CalcFrameOff ;

: FORCOCOA ( addr n <parameters> -- #ints )
   find-importstructure dup 0= abort" Not in import structure link !"   \ should cleanup, FORGET last
   to LastImport                       \ for ^efn called by generateCall
   parseArgs 2>r dup 2 + 2r>           \ the parameter parsing, add rec and sel to #ints,
                                       \ but leave original for REORDER and REORDER-STRET
   setArgs                             \ the parameter setup for copyFrame
   /offsets                            \ borrowed from GenerateImport
                                       \ FUNCTION: doesn't need this becaus it calls GenerateImport.
;

: COCOA, ( -- )                        \ borrowed from MakeNormal
   prologue,                           \ we want this
   copyFrame  generateCall  returnData \ this
   epilogue,                           \ and this
   $C3 c,                              \ compile the RET/NEXT
;

previous

\\
