(* ====================================================================
Cocoa functions

Copyright 2008 Roelf Toxopeus

VFX64 version.
Part of the interface to ObjC 2 Runtime.
Declare message as if it is a function
Together with cocoacore.fth and cocoabridge.fth this file defines the
interface to ObjC message sending.
Last: 28 September 2020 at 14:31:52 CEST -rt
==================================================================== *)

(( --------------------------------------------------------------------
Turnkey proof:
A linked list is created, which will hold the message selectors.
Upon booting coco-sf, the list is traversed and all our defined Cocoa
messages are initialised.

@SELECTOR -- return selector for messagename. It tests the selector name
for a leading '@' character, and skips it in the name when found.
This leading '@' is used as a Cocoa identifier in coco-vfx64, it's optional.
Likewise the underscore _ prepended to foreign function names in some
Forth systems. The '@' is borrowed from ObjC where it serves as compiler
hint.
COCOASELECTORS -- holds head linked list for Cocoa selectors.
>COCOASELECTORS -- adds a selector to the list.
/COCOASELECTORS -- initialise the messages in the list to their proper
selectors, execute this somewhere in the startup from a turnkeyed coco-i4.
.COCOASELECTORS -- show the selectors in the list by their name.

COCOASELECTOR -- parse following name and add to the selectors list.

COCOA-MESSAGE, -- CREATE part, setting up the local structure:
codepointer, #ints, payload: selector name, node-link, node-payload, code.
The a n input represents one of the objc_msgXXX family, consumed by FORCOCOA
and later used by COCOA,

SEND-COCOA-MESSAGE -- DOES> part. Reorders the parameters in Objc format,
and calls the appropriate objc_Msg sender.
SEND-COCOA-STRET-MESSAGE -- DOES> part. Reorders the parameters in Objc
structure return format, and calls the appropriate objc_Msg sender.

COCOA-MESSAGE -- compile code to send message to ObjC runtime.
The objc_Msg sender is given as a string. The message  it self and the
parameters need to be parsed from the following text input.
This is the ordinary Cocoa message sending creation.
a n is either objc_msgSend, objc_msgSendSuper or objc_msgSend_fpret

COCOA-STRET-MESSAGE -- compile code to send message with a structure
for return values to ObjC runtime.
The objc_Msg sender is given as a string. The message  it self and the
parameters need to be parsed from the following text input.
This is the STRET Cocoa message sending creation.
a n is either objc_msgSend_stret or objc_msgSendSuper_stret

-- Top words

COCOA: -- create FUNCTION: like code to send an ordinary Cocoa message
to a class or instance as represented by the receiver.
   
SUPERCOCOA: -- create FUNCTION: like code to send a Cocoa message to
the parent of the receiver: the super.

COCOA-STRET: -- create FUNCTION: like code to send an Cocoa message
to a class or instance as represented by the receiver. An extra parameter
must be passed a long the rest: a pointer to a structure for the return
values.

SUPERCOCOA-STRET:  -- create FUNCTION: like code to send an Cocoa message
to the parent of a class or instance as represented by the receiver. An
extra parameter must be passed a long the rest: a pointer to a structure
for the return values.

COCOA-FPRET: -- create FUNCTION: like code to send an ordinary Cocoa
message to a class or instance as represented by the receiver. The
returned value is a fp value found on the fp stack.

Important:
The parameter parsing follows the same rules as used by FUNCTION:
See mac/system/function.fth and VfxForthOsx64/VFXBase/ExternSysVx64.fth
First ( ... -- ... ) for integers, a second ( ... -- ... ) for doubles.
-------------------------------------------------------------------- ))


/forth
decimal

\ ---------------------------------------------------------------------
\ --- relink selectors ------------------------------------------------

\ return selector for methodname
: @selector ( 0string -- sel )
	dup c@ [char] @  = IF char+ THEN		\ ignore leading 'identifier'
	sel_getUid dup 0= abort" @selector failed" ;

\ Note: for 'STRET' functions, add the stret pointer as the last rightmost input parameter
\ i.e. CocoaSelector @frame ( frame-rect -- ret )

variable cocoaselectors

: >cocoaselectors ( a -- )  cocoaselectors node node, ;

: relinkselector ( a -- )  dup cell+ @selector swap ! ;

: /COCOASELECTORS   ['] relinkselector cocoaselectors allnodes ;

: .selectorname ( a -- )  cell+ zcount cr type ;

: .COCOASELECTORS   ['] .selectorname cocoaselectors allnodes ;

\ ---------------------------------------------------------------------------------------
\ --- cocoa: family ---------------------------------------------------------------------

: cocoaselector ( <name-sel> -- sel )
	>in @ >r  bl word count pocket zplace   r> >in !
	pocket @selector ;

\ --- primitives

: cocoa-message, ( a n sel <parameterlist> -- )
   here >r 0 ,                                      \ place holder for codepointer
   -rot forcocoa ,                                  \ #ints
   here swap , pocket zcount z, >cocoaselectors     \ node-payload: selector, name  to cocoaselectorlist
   here r> ! cocoa,                                 \ code and resolving the codepointer
;

: send-cocoa-message ( xi a -- xj )
   dup @ >r                 \ the code to execute now as return address 
   cell+ 2@ reorder         \ add selector and reorder stack and ...
;

: send-cocoa-stret-message ( xi a -- xj )
   dup @ >r                 \ the code to execute now as return address 
   cell+ 2@ reorder-stret   \ add selector and reorder stack and ...
;

: cocoa-message ( a n <name> <parameterlist> -- )
   cocoaselector
   CREATE   cocoa-message,
   DOES>    send-cocoa-message ;

: cocoa-stret-message ( a n <name> <parameterlist> -- )
   cocoaselector
   CREATE   cocoa-message,                                 \ code
   DOES>    send-cocoa-stret-message ;

\ --- COCOA: family

: COCOA: ( <name> <parameterlist> -- )   s" objc_msgSend" cocoa-message ;
   
: SUPERCOCOA: ( <name> <parameterlist> -- )   s" objc_msgSendSuper" cocoa-message ;

: COCOA-STRET: ( <name> <parameterlist> -- )   s" objc_msgSend_stret" cocoa-stret-message ;

: SUPERCOCOA-STRET: ( <name> <parameterlist> -- )   s" objc_msgSendSuper_stret" cocoa-stret-message ;

: COCOA-FPRET: ( <name> <parameterlist> -- )   s" objc_msgSend_fpret" cocoa-message ;

\\
