(* ====================================================================
Window interface

Copyright 2008 Roelf Toxopeus

Part of GUI stuff.
VFX64 version.
A minimal window creating interface
Compatible where possible with MacForth and Mach2.
Last: 7 August 2020 at 17:49:47 CEST    -rt
==================================================================== *)

(( --------------------------------------------------------------------
About this window interface.

The OS does the housekeeping. Forth keeps track of things in case of a
turnkey, so windows can be restored with last settings.

Forth creates a named window record with NEW.WINDOW. 
This record is initialy used to create the new window. The first entrees
will be set with the all important windowref and possible CGRef during adding
the window or during CoreGraphic initialisation. The other fields are initialy
filled with default values. The bounds field is used for scratch later on.
The record is also used after a turnkey/snapshot to re-add the window.
All the W.xxx and +W.xxx words affect this record (only) and don't change
the window (directly).
For real changes use the other words like the SET.Wxxx and xxx.WINDOW words.
NoTE: there is a variant on NEW.WINDOW, (NEW.WINDOW) which allows creating
nameless windows.

Windows are added to the environment with ADD.WINDOW.
If an existing window is re-added, it will be closed first.

To update the current window settings in the window record, use SAVE.WINDOW.
Useful before a snapshot or turnkey.

When closing a window, the window structure in Forth is not forgotten,
it's just pointing to 0: no window
You can reuse the window structure for another window, with ADD.WINDOW.

Most words defined are in MacForth style:

/WINDOW.RECORD -- size in bytes for a default window record.
Field pointers in to the record:
+W.REF -- points to windowref field, contains NSWindow ref.
+W.GPORT -- points to windowgport field, contains gport ref.
+W.BOUNDS -- points to window Content bounds field.
+W.BACKING -- points to windowbacking field, do we need this?
+W.DEFER -- points to windowdefer field, do we need this?
+W.STYLE -- points to windowstyle field
+W.TITLE -- points to windowtitle field, contains pointer to counted string.
+W.TRANSPARENCY -- points to windowtransparency field, contains transparency CGFloat.
+W.BACKGROUND -- points to windowbackground field, contains NSColor ref.

Affects the window record only, execute these before ADD.WINDOW.
W.BOUNDS -- set initial window Content bounds, without eventual titlebar 22e0 width 
W.STYLE -- set initial window style
W.TITLE -- set initial window title
W.TRANSPARENCY--  set initial window transparency
W.BACKGROUND -- set initial window background colour

Change window after ADD.WINDOW, window record is used for scratch.
RESIZE.WINDOW -- resize window Content size, with given width and height.
MOVE.WINDOW -- move window to given x and y.
FIX.WINDOW -- place window with given CGRect.
SHOW.WINDOW -- show window.
HIDE.WINDOW -- hide window, doesn't close it.
CLOSE.WINDOW -- close window, reset window record.
SET.WTITLE -- set/change the window title.
SET.WTRANSPARENCY -- set/change the window transparency.
SET.WBACKGROUND -- set/change the window background colour.
SET.WSTYLE -- set/change the window stylemask.

Get window properties
WINDOW.TITLE -- retrieve window title
WINDOW.TRANSPARENCY -- get window transparency
WINDOW.STYLE -- return window style
WINDOW.BACKGROUND - return background color as rgba
WINDOW.FRAME -- get window frame size CGRect, including titlebar 22e0 width
SAVE.WINDOW -- save window info. Content bounds is saved. Use before doing a turnkey/snapshot.

/WINDOW -- initialises a window record to default values.
(NEW.WINDOW) -- allocates an initialised window record.
NEW.WINDOW -- creates a new named window record, initialised to defaults.
WINDOW.ARGS -- fetches the necessary parameters for creating a NSWindow
instance, from a window record.
CREATE.WINDOW -- creates new NSWindow instance. Values stored in window
record can be used as parameters.
ADD.WINDOW -- create NSWindow instance and show it on screen. Uses values
from given window record.

Some color utillities:
RGBA>NSCOLOR -- create a NSColorRef from r g b a CGFloats.
quartz.framework -- the Quartz framework, needed for:
@RGBA -- the reverse of RGBA>NSCOLOR, returns the r g b a components/values
from a NSColorRef. Uses CGColor functions. NSColor itself can't do this,
with all colors!

Note: Apple compilers do things differently wrt Rectangle structures: NSRect and
      thus CGRect as you might expect.
      According the 64bit AMD ABI one can pass the parameters in their fp registers.
      In stead Apple passes them in memory through the stack.
      This means we'll have to use a little trick in VFX to force the calling
      system to push the rectangle members up the C stack.

Trick: make sure the stackpicture for the integers ( ... )  has enough parameters to
       fill the registers. Use dummies if necessary. Then add the rect members. This
       will ensure that the rect mambers will be passed through the stack.
       The ABI says there are 6 registers for the 6 first calling parameters, the
       rest follows on the C stack.
       Remember that COCOA: and family will initiate the ( ... ) with the RECEIVER and
       SELECTOR parameters added. So there are 4 registers left to fill before the C
       stack is used.

As an example:
setFrame:

One expects it to be:
cocoa: @setFrame: ( -- ret ) ( F: x y w h -- )

But it will be:
cocoa: @setFrame: ( dummie1 dummie2 dummie3 dummie4 x y w h -- ret )

Remember for Cocoa we have a hidden/implied receiver which should be on top before
the message.
: @setFrame: ( -- ret ) ( F: x y w h -- ) >r 0 0 0 0 4fpushdf r> setFrame: ;

The word COCOA.DFIX will sort this out, given the number  of dummies you have to
pass: 
: @setFrame: ( -- ret ) ( F: x y w h -- ) 4 cocoa.dfix setFrame: ;

BTW no problems/issues with 2 DFLOATS as part of structures like NSPoint/CGPoint
and NSSize/CGSize, or 4 DFLOATS for a colorstructure: r g b alpha. These are passed
in their XMM registers like any other fp parameter.

A nice puzzle is @contentRectForFrameRect: both a stret and rect passing.
-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ NSWindow class

cocoaclass NSWindow

cocoa: initWithContentRect:styleMask:backing:defer: ( stylemask backing defer dummy x y width height -- id )
: @initWithContentRect:styleMask:backing:defer:  ( stylemask backing defer -- id ) ( F: x y width height  -- )  1 COCOA.DFIX initWithContentRect:styleMask:backing:defer: ;

cocoa: @setOpaque: ( flag -- ret )
cocoa: @orderFront: ( sender -- ret )
cocoa: @orderOut: ( sender -- ret )
cocoa: @close ( -- ret )
cocoa: @setTitle: ( NSString:title -- ret )
cocoa: @title ( -- NSStringRef )
cocoa: @setAlphaValue: ( -- ret ) ( F: viewAlpha -- )
cocoa-fpret: @alphaValue ( -- ) ( F: -- alpha )
cocoa: @setBackgroundColor: ( NSColor:backgroundColor -- ret )
cocoa: @backgroundColor ( -- ref )
\ cocoa: @display ( -- ret )
cocoa: @setStyleMask: ( nsuinteger:stylemask -- ret )
cocoa: @styleMask ( -- stylemask )
cocoa: @setContentSize: ( -- ret ) ( F: w h -- )
cocoa: @setFrameOrigin: ( -- ret ) ( F: x y -- )
cocoa: setFrame:display: ( flag dummy dummy dummy x y w h -- ret )
: @setFrame:display: ( flag -- ret ) ( F: x y w h -- )  3 cocoa.dfix setFrame:display: ;
cocoa-stret: @frame ( *rect -- ret )
cocoa-stret: contentRectForFrameRect: ( dummy dummy dummy x y w h *rect -- *rect )
: @contentRectForFrameRect: ( *rect -- *rect ) ( F: x y w h -- )  swap >r 3 cocoa.dfix r> swap contentRectForFrameRect: ;

\ --------------------------------------------------------------------
\ NSWindow parameters

cocoaclass NSColor

cocoa: @CGColor ( -- CGColorRef )
cocoa: @colorWithCalibratedRed:green:blue:alpha: ( -- NSColor:refrgba ) ( F: red green blue alpha -- )

: RGBA>NSCOLOR ( -- NSColorRef ) ( F: r g b a -- )   NSColor @colorWithCalibratedRed:green:blue:alpha: ;

0 constant NSBorderlessWindowMask
1 constant NSTitledWindowMask
2 constant NSClosableWindowMask
4 constant NSMiniaturizableWindowMask
8 constant NSResizableWindowMask
256 constant NSTexturedBackgroundWindowMask

NSTitledWindowMask NSClosableWindowMask or NSMiniaturizableWindowMask or
NSResizableWindowMask or constant ALLWIDGETS

NSTexturedBackgroundWindowMask constant NOWIDGETS

0 constant NSBackingStoreRetained
1 constant NSBackingStoreNonretained
2 constant NSBackingStoreBuffered

\ --------------------------------------------------------------------
\ necessary CoreGraphics color handling

framework Quartz.framework
function: CGColorGetNumberOfComponents ( color -- #components )
function: CGColorGetComponents ( color -- *components )

: @RGBA ( NSColorRef -- ) ( F: -- r g b a )
   @cgcolor dup
   CGColorGetComponents
   swap CGColorGetNumberOfComponents 2 > if
      cg4@   ( r g b a )
   else
      dup df@ fdup fdup  ( r g b )
      dfloat+ df@    ( a )
   then ;

\ --------------------------------------------------------------------
\ Forth NSWindow record and parameters

6 cells 9 dfloats + constant /WINDOW.RECORD

: +W.REF ( wptr4 -- 'windowref )  ;

: +W.GPORT ( wptr4 -- 'windowgport )  cell + ;

: +W.BOUNDS ( wptr4 -- 'windowbounds ) [ 0 +w.gport cell+ ] literal + ;

: +W.BACKING ( wptr4 -- 'windowbacking ) [ 0 +w.bounds 4 dfloats + ] literal + ;

: +W.DEFER ( wptr4 -- 'windowbacking ) [ 0 +w.backing cell+ ] literal +  ;

: +W.STYLE ( wptr4 -- 'windowstyle ) [ 0 +w.defer cell+ ] literal + ;

: +W.TITLE ( wptr4 -- 'windowtitle ) [ 0 +w.style cell+ ] literal + ;

: +W.TRANSPARENCY ( wptr4 -- 'windowtransparency ) [ 0 +w.title cell+ ] literal + ;

: +W.BACKGROUND ( wptr4 -- 'windowbackground ) [ 0 +w.transparency dfloat+ ] literal + ;

\ --------------------------------------------------------------------
\ set/change stuff before window is added with ADD.WINDOW

: W.BOUNDS ( wptr4 -- ) ( F: x y width height -- )   +w.bounds cg4! ;

: W.STYLE ( n wptr4 -- )   +w.style ! ;

: W.TITLE ( a n wptr4 -- )   -rot here >r string, r> swap +w.title ! ;

: W.TRANSPARENCY ( wptr4 -- ) ( F: alpha -- )  +w.transparency df! ;

: W.BACKGROUND ( wptr4 -- ) ( F: r g b alpha -- )  +w.background cg4! ;

\ --------------------------------------------------------------------
\ User interface words, interact with Cocoa
\ -- move, size etc.

: RESIZE.WINDOW ( wptr4 -- ) ( F: w h -- )   +w.ref @ @setContentsize: drop ;

: MOVE.WINDOW ( wptr4 -- ) ( F: x y -- )   +w.ref @ @setFrameOrigin: drop  ;

: FIX.WINDOW ( wptr4 -- ) ( F: x y w h -- )   yes swap +w.ref @ @setFrame:display: drop ;

\ : SHOW.WINDOW ( wptr4 -- )   +w.ref @ dup dup @orderFront: drop @display drop ;

: SHOW.WINDOW ( wptr4 -- )   +w.ref @ 0 swap @orderFront: drop ;

: HIDE.WINDOW ( wptr4 -- )   +w.ref @ dup @orderOut: drop ;
   
: CLOSE.WINDOW ( wptr4 -- )   dup +w.ref @ dup @close drop @release drop off ;

\ -- set window properties
: SET.WTITLE ( 0string wptr4 -- )   swap >nsstring swap +w.ref @  @setTitle: drop ;

: SET.WTRANSPARENCY ( wptr4 -- ) ( F: alpha -- )   +w.ref @ @setAlphaValue: drop ;

: SET.WSTYLE ( stylemask wptr4 -- )   +w.ref @ @setStyleMask: drop ;

: SET.WBACKGROUND ( wptr4 -- ) ( F: red green blue alpha -- )
   NSColor @colorWithCalibratedRed:green:blue:alpha:
   swap +w.ref @ @setBackgroundColor: drop ;

\ -- retrieve window properties
: WINDOW.FRAME ( wptr4 -- ) ( F: -- x y w h )
   dup +w.bounds dup >r swap +w.ref @ @frame drop r>  cg4@ ;

: CONTENT.FRAME ( wptr4 -- ) ( F: -- x y w h )
   dup window.frame dup +w.bounds swap +w.ref @ @contentRectForFrameRect: cg4@ ;
  
: WINDOW.TITLE ( wptr4 -- a n )  +w.ref @ @title >4thstring ;

: WINDOW.TRANSPARENCY ( wptr4 -- ) ( F: -- alpha )  +w.ref @ @alphaValue ;

: WINDOW.STYLE ( wptr4 -- stylemask )   +w.ref @ @styleMask ;
      
: WINDOW.BACKGROUND ( wptr4 -- ) ( F: -- red green blue alpha )   +w.ref @ @backgroundColor @rgba ;

\ -- save window properties
: SAVE.WINDOW ( wptr4 -- )
   \ dup window.frame 22e0 f- dup w.bounds   \ better is CONTENT.FRAME
   dup content.frame fdrop fdrop fdrop fdrop
   dup dup window.title rot w.title
   dup window.transparency dup w.transparency
   dup window.style over w.style
   dup window.background w.background ;

\ --------------------------------------------------------------------
\ Window creating words

\ NoTE: was z" for s" !!
: /WINDOW ( a -- )
   0                         over +w.ref !
   0                         over +w.gport !
   100e0 575e0 340e0 180e0   dup  w.bounds     \ was 600, 575 leaves room for titlebar and menubar
   allwidgets                over w.style
   NSBackingStoreBuffered    over +w.backing !
   no                        over +w.defer !
   dup s" untitled"          rot  w.title         \ 'saves' as a counted string!
   1e0                       dup  w.transparency
   1e0 fdup f2dup            w.background
;

: (NEW.WINDOW) (  -- addr )   here /window.record allot dup /window ;

\ : NEW.WINDOW ( <name > -- )  (new.window) constant ;

: NEW.WINDOW ( <name > -- )  create (new.window) drop ;

: WINDOW.ARGS ( wptr4 -- style backing defer ) ( F: -- x y w h )
   >r
   r@ +w.bounds cgrect@
   r@ +w.style @
   r@ +w.backing @
   r> +w.defer @
;

: CREATE.WINDOW ( style backing defer -- instance ) ( F: x y w h -- )
   NSWindow @alloc @initWithContentRect:styleMask:backing:defer: ;

: ADD.WINDOW ( wptr4 -- )
   dup +w.ref @ if dup close.window then
   dup window.args create.window
\ default opaque setting is YES, so need to chage it to change transparency!
   no over @setOpaque: drop
   over +w.ref !
   dup +w.title @ count drop over set.wtitle         \ this is ugly because of counted vs zero string
   dup +w.transparency df@ dup set.wtransparency
   dup +w.background cg4@ dup set.wbackground
   show.window
;

\\ ( eof )