(* ====================================================================
ObjC method creation in ObjC 2 Runtime

copyright 2006 Roelf Toxopeus

Part of adding new classes with Forth methods.
VFX version.
Adding methods while building new class in ObjC 2 Runtime.
Last: 15 October 2014 09:36:52 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
Forth interface to method creation for the ObjC 2 Runtime.
   
Turnkey proof:
Two linked lists are created, one holds the instance method info and one
holds the class method info.
Upon booting coco-vfx, the lists are traversed and all our defined instance
and class methods will be recreated, and added to their classes.

(ADD.METHOD) -- register and add method for given (meta)class.

FORTH-COCOAMETHODS -- holds head linked list for in Forth created ObjC
instance methods.
>FORTH-COCOAMETHODS -- adds an instance method to the list.
/FORTH-COCOAMETHODS -- recreate and add the instance methods in the list,
execute this somewhere in the startup from a turnkeyed coco-vfx.
.FORTH-COCOAMETHODS -- show the instance methods in the list by their name.

FORTH-COCOACLASSMETHODS -- holds head linked list for in Forth created ObjC
class methods.
>FORTH-COCOACLASSMETHODS -- adds a class method to the list.
/FORTH-COCOACLASSMETHODS -- recreate and add the class methods in the list,
execute this somewhere in the startup from a turnkeyed coco-vfx.
.FORTH-COCOACLASSMETHODS -- show the class methods in the list by their name.

Working towards the following syntax:

ADD.METHOD -- creates an internal structure containing the implementation
pointer, parameter types pointer, method-name and the class-name.
Adds the implementation as instance method to class.

ADD.CLASSMETHOD -- same as ADD.METHOD but adds the implementation as
a class method to the class. This is done by using the metaclass rather
than the class itself.
The metaclass from a class under construction can be obtained by calling
object_getClass on the class, before the class is added.
NoTE: object_getClass always returns the metaclass, before and after the
class is added. Passing a metaclass to object_getClass will return the
metaclass of its superclass.

The ADD words usually come between NEW.CLASS and ADD.CLASS, but can
also be used after ADD.CLASS is executed.
-------------------------------------------------------------------- ))

/forth
decimal

\ --------------------------------------------------------------------
\ relink forth-cocoamethods

function: class_addMethod ( class SEL:name imp *types -- bool )
function: sel_registerName ( 0string -- sel )

\ syntactic sugar, tradition causing the stack order
: (ADD.METHOD) ( imp *types name class -- )
   swap sel_registerName 2swap
   class_addMethod 0= abort" Can't add method !" ;
   
variable FORTH-COCOAMETHODS

: >FORTH-COCOAMETHODS ( a -- )   forth-cocoamethods node node, ;

: RELINKMETHOD   ( a -- )
   dup @                                    \ imp
   swap cell+ dup @                        \ *types
   swap cell+                              \ name
   dup zcount + 1+ ( aligned )
   find drop execute                        \ class id, is updated by /forth-cocoaclasses !!
   (add.method) ;                           \ relink

: /FORTH-COCOAMETHODS ( -- )  ['] relinkmethod forth-cocoamethods allnodes ;

: .FORTH-COCOAMETHODNAME ( a -- )  cell+ cell+ zcount cr type ;

: .FORTH-COCOAMETHODS ( -- )   ['] .forth-cocoamethodname forth-cocoamethods allnodes ;

\ --------------------------------------------------------------------
\ ADD.METHOD

: BUILD.METHOD ( imp *types name class -- address )
   2over 2over (add.method)
   align here >r
   2swap swap ,                  \ imp
   ,                              \ *types
   swap zcount z,                  \ method name
   class_getname zcount string,   \ class name (also Forth dict name!), don't save class id, might change
   r> ;                           \ start of structure
   
\ creates an internal structure: imp *types method-name class-name
: ADD.METHOD ( imp *types name class -- )  build.method >forth-cocoamethods ;

\ --------------------------------------------------------------------
\ relink forth-cocoaclassmethods

function: object_getClass ( object -- class )

variable FORTH-COCOACLASSMETHODS

: >FORTH-COCOACLASSMETHODS ( a -- )   forth-cocoaclassmethods node node, ;

: RELINKCLASSMETHOD   ( a -- )
   dup @                                 \ imp
   swap cell+ dup @                     \ *types
   swap cell+                           \ name
   dup zcount + 1+ ( aligned )
   find drop execute                     \ class id, is updated by /forth-cocoaclasses !!
   object_getclass                     \ get metaclass
   (add.method) ;                        \ relink

: /FORTH-COCOACLASSMETHODS ( -- )  ['] relinkclassmethod forth-cocoaclassmethods allnodes ;

: .FORTH-COCOACLASSMETHODS ( -- )   ['] .forth-cocoamethodname forth-cocoaclassmethods allnodes ;

\ --------------------------------------------------------------------
\ ADD.CLASSMETHOD

: ADD.CLASSMETHOD ( imp *types name class -- )   object_getClass build.method >forth-cocoaclassmethods ;
   
\\ ( eof )