(* ====================================================================
Adding a menu plus menu items to existing menubar.

Copyright 2017 Roelf Toxopeus

VFX version.
Forth ObjC Runtime interface usage example
Last: 9 September 2020 at 22:02:11 CEST   -rt
==================================================================== *)

(( --------------------------------------------------------------------
This example shows how to add a menuitem to the existing menubar programma-
tically. This menu is inspired by MacForth's -Project- menu. It contains
a list of all included files as its submenu menuitems. When clicked on,
the file will open in your editor.
In this example the menubar will be provided by vfxmenu.fth and a menuitem
name Included will be added to the menubar.
One new method will be added to VFXMenuClass, the class running VFXMenu.

A menu structure is as follows:
menu -> menuitem -> submenu -> menuitem -> ... -> action -> target

Which translates to:
NSApp mainMenu (the menubar) -> menuitems (Application Name, File, Edit, ... Help)
each menuitem -> submenu (the dropdown menu)
each submenu -> menuitems (clickable items)
each menuitem -> defaultaction or action with target set by us.
See simplemenu.fth

Note that in general speak we say MENU when a MENUITEM is meant.

Utillities for main menubar:
BARITEM -- returns menuitem for menuname in mainmenu
BARINDEX -- returns index for menuname in mainmenu
NEWBARITEM -- return new menuitem with name for the menubar. A submenu
is attached to the new menuitem.
NEWBARMENU -- insert menuitem with name, in front of Help menu.

The special added menu(item)s:
*edit: -- callback is implementation for the edit: method, which is
added to VFXMenuClass. Its sender parameter is the selected menuitem,
who's representedObject is a fullname, which is opened in our editor.

FILE-ITEM -- return new filename-item for submenu, the dropdown part.
It expects the FULLNAME zbuffer, which will be set as representedObject.
The filename without path will be the menuitem title.

ADD.FILE -- insert given fullpath filename at index in -Included- menu
LASTFILE-ITEM -- return last included file as new filename-item.
>INCLUDED -- add last included file at top -Included- list
INCLUDED>MENU -- add all included files to the -Include- menu.

Possible redefinitions to update the Included menu:
: LOAD-FILE ( -- )    load-file >included ;
: LOAD-EDIT ( -- )    load-edit >included ;
: INCLUDE-FILE ( fid -- )   include-file >included ;
: INCLUDED ( c-addr u -- )   included >included ; 
: INCLUDE ( <spaces>filepath -- )   include >included ;

Alternative, refresh by reinstalling Included menu:
COCOA: @removeItem: ( itemref -- ret )

: RE/IM ( -- )
   Z" Included" baritem NSApp @ @mainMenu @removeItem: drop
   included>menu ;
   
-------------------------------------------------------------------- ))

\ vfxmenu specifics, first add menubar to coco-sf
[undefined] vfxmenu [if]
include %cocoalib%/hotcoco/vfxmenu.fth
vfxmenu cr .( VFXMENU installed)
[then]

/forth
decimal

\ cocoa: @mainMenu ( -- menuref )   \ NSApplication

cocoaclass NSMenu
cocoa: @initWithTitle: ( nsstringref -- menuref )   \ NSMenu
cocoa: @itemWithTitle: ( nsstringref -- itemref )   \ NSMenu
cocoa: @indexOfItemWithTitle: ( nsstring -- index )   \ NSMenu
cocoa: @insertItem:atIndex: ( newItem index -- ret )   \ NSMenu
cocoa: @addItem: ( newitem -- ret )   \ NSMenu

cocoaclass NSMenuItem
cocoa: @submenu ( -- menuref )   \ NSMenuItem
cocoa: @setSubmenu: ( newmenu -- ret )   \ NSMenuItem
cocoa: @initWithTitle:action:keyEquivalent: ( nsstring action nsstring -- NSMenuItemRef )      \ NSMenuItem
cocoa: @setTarget: ( obj -- ret )   \ NSMenuItem
cocoa: @representedObject ( -- obj )   \ NSMenuItem
cocoa: @setRepresentedObject: ( obj -- ret )   \ NSMenuItem

: BARITEM ( za -- item )   >NSSTRING NSApp @ @mainMenu @itemWithTitle: ;

: BARINDEX ( za -- index )   >NSSTRING NSApp @ @mainMenu @indexOfItemWithTitle: ;

: (NEWITEM) ( name selector key -- item )   NSMenuItem @alloc @initWithTitle:action:keyEquivalent: ;

: (NEWMENU) ( name -- menu )   NSMenu @alloc @initWithTitle: ;

: NEWBARITEM ( zname -- item )
   >nsstring dup 0 @" " (newitem)    \ DUP 0= ABORT" NO NEW ITEM !"
   swap (newmenu)
   over @setSubmenu: drop ;         \ create the submenu (the dropdown) for later to be added items

: NEWBARMENU ( za -- )  
   newbaritem
   z" Help" barindex
   NSApp @ @mainMenu @insertItem:atIndex: drop ;

\ harness
: (EDIT-FILE) ( a n line -- )   locateinfo .locate ;

forthcall: *EDIT: ( rec sel sender -- n )  nip nip 0 swap @representedObject >4thstring 0 (edit-file)  0 ;

*edit: menuMethodTypes z" edit:"    VFXMenuClass add.method

\ harness
: -PATH ( a1 n2 -- a2 n2 )
         begin 2dup s" /" search
      while drop char+ >r over r> swap - /string
   repeat
   drop >r over r> swap - /string ;

: FILE-ITEM ( za -- NSMenuItemRef )
   dup >nsstring >r                        \ the fullname nsrtringref will be representedobject
   zcount -path drop >nsstring         dup >r
   z" edit:" @selector @" " (newitem)  R> @release drop ( NSStringRef)
   VFXMenuClassREF @ over @setTarget: drop
   r> over @setRepresentedObject: drop ;  \ don't release it !

\ harness
\ the actual filepath from the sourcefilename is a zstring in the body. no need for pocket.
: FULLNAME ( nfa -- za )   name> >body [ 2 cells ] literal + ;
variable 'lastfile
: (LAST-FILE)   ( #thread nfa -- flag )  nip fullname 'lastfile ! false ;
: LASTFILE ( -- a )   ['] (last-file) ['] sourcefiles >body @ WalkWordList 'lastfile ;
  
: LASTFILE-ITEM ( -- NSMenuItemRef )   lastfile @ file-item ;

: ADD.FILE ( za index -- )   swap file-item swap z" Included" baritem  @submenu @insertItem:atIndex: drop ;

: >INCLUDED ( -- )   lastfile @ 0 add.file ;

variable menuIndex

: FILE>MENU ( #thread nfa -- flag )   nip fullname menuIndex @ add.file 1 menuIndex +! true ;
   
: INCLUDED>MENU ( -- )
   Z" Included" newbarmenu
   menuIndex off  ['] file>menu  ['] sourcefiles >body @  WalkWordList ;

cr .( INCLUDED>MENU to add menu to main menubar)

\\ ( eof)
   
   
   