(* --------------------------------------------------------------------
FUNCTION: is the implemetation of a foreign function interface
as found in SwiftForth, VFX32 and iForth64.
The passing and returning of parameters is different in 64b Darwin Forths.
According the ABI, integers and doubles are passed and returned in their own
dedicated registers. Only when the number of registers is exceded, the C-stack
is used.
So the doubles and the integers are separated, they are pushed and popped from
their stacks in and out of the dedicated registers. This also means the order
of the integer and double parameters is only relevant wrt their own type.
This eases the passing and returning of doubles enormously compared with the
32b ABI, no more pushing from fp stack to data stack and vv when doing calls.

The 64bit FUNCTION: version makes use of a common used stack notating style,
( ... - ... ) ( F: ... -- ... )
... stands for whatever number of passed and returned values.

Explain by example:   
FUNCTION: myfunc ( n1 n2 nn -- ret )
items between the first found  (  and  --  will be parsed as int input parameters
items between the consecutive  --  and  )  will be parsed as int output parameters
items between a second found  (  and  --  will be parsed as double input parameters
items between the consecutive  --  and  )  will be parsed as double output parameters
FUNCTION: acos ( -- ) ( r1 -- r2 )
The use of S: and F: is mandatory.
FUNCTION: acos ( -- ) ( F: r1 -- r2 ) is the same as the former.

Note: the data stack picture always needs to be used together with the
fp stack picture, to force fp passing. Even if the data stack is and
stays empty. If the fp stack isn't used it can be omited (usual practice).

The following parameter names have special behavior regarding the in-
and out-put parameter handling:
   F:                in/out   noop ! Just for clarity.
   S:                in/out   noop !
   N:                in/out   noop !
   R:                in/out   noop !
   
Any other name will be treated as an integer in the first stack picture.
Any other name in the second stack picture will be treated as a double.
Leaving input and/or return empty ( -- ) means no input and or return
parameters.

The COCOA: family of words uses the same parameter parsing words.

AS or ASCALL: -- as FUNCTION: but extern will be defined as name.
Example:
AS _dup dup ( fd -- n )

GLOBAL: -- as FUNCTION: but only compiles code to return a pointer
to the found extern. This is either a function pointer or an exported
constant or a global variable.
Example:
GLOBAL: NSApp ( -- a )
-------------------------------------------------------------------- *)

(* --------------------------------------------------------------------
The following code is part of the ExternSysVx64.fth file.
But originates in the original coco-vfx 32bit package.
So when porting to another Forth which runs the cocoa distro,
FUNCTION: GLOBAL: ASCALL: or AS: should be expected to be available.
These words should be Swift proof, making no assumptions about the way
the prototypes are declared. No fuzzing with potential different styles
in C header files, not understood by a C parser implementation.
They show the Forth side of things, not unimportant.
They also solve the problem, when the system language isn't C derived,
and the C parser used for Extern: can fail.
Point in case: Apple's Swift. Main documentation is Swift based.
EXTERN: static func beep() 
Err# -1001 Invalid type in EXTERN: import
 -> EXTERN: static func beep()
                            ^
No copy/paste possible...
Every coco-XXX distribution has mac/doc/c(-rationale.txt

The comments, documentation and stackpictures are MPE style, not my
own.

Last: 25 August 2020 at 07:36:25 CEST  -rt
-------------------------------------------------------------------- *)

/forth
decimal

\\  ( don't include the rest, it's already in the kernel)
StopIncluding
EXIT

\ **********************
\ *S Compatibility words
\ **********************
\ *P These words are mainly for users converting code from other
\ ** Forth systems. They provide a more "Forth-like" interface that
\ ** does not permit copy/paste from the operating system header files
\ ** or documentation.

: skipTo(	\ -- ; )
\ Step forward to a space-delimited '('. )
  begin
    parse-name dup 0= abort" missing ("	\ error at end of line ; )
    s" (" compare 0=			\ look for '('; )
  until
;

: param? \ caddr len -- flag
\ Should ignore S: F: N: R: stack indicators as parameters.
\ Leave a true flag when genuine parameter, false otherwise.
        2dup upper
        2dup s" S:" compare dup 0= if nip nip  exit then
   drop 2dup s" F:" compare dup 0= if nip nip  exit then
   drop 2dup s" N:" compare dup 0= if nip nip  exit then
   drop      s" R:" compare dup 0= if nip nip  exit then
   -1 or ;

: parse-params	\ -- #in #out
\ Parse a parameter list of the form:
\   ( a b c d -- e f )
  skipTo(				\ scan up to '('    )
  0 begin				\ -- u ; begin parsing
    parse-name  2dup s" --" compare	\ compare leaves zero flag when equal !
   while				\ -- u caddr len
    2dup param? if                      \ do we have a genuine param?
      s" ..." compare if		\ -- u
        dup 0 >=			\ if not varargs (* not ...  -rt *)
        if  1+  endif			\  count up
      else
        drop -1				\ #in=-1 for varargs  (* is ...  -rt *)
      endif
    else  2drop  then			\ ignore invalid param, try again
  repeat
   2drop                                \ get rid of dupped word+count
   0 begin                              \ begin parsing outputs
    parse-name s" )" compare
   while
     1+
  repeat
;

: doubles?  \ -- f
\ check for fp stackpicture
   >in @ parse-name rot >in ! s" (" compare 0= ;	\ ) for syntax highlighter

: outMask \ #double-rets #int-rets -- outmask
\ Create a mask from the given number of double and int returns.
   2dup 0 0 d= if 2drop 0 exit then          \ void
   2dup 0 1 d= if 2drop 1 exit then          \ int64
   2dup 0 2 d= if 2drop 2 exit then          \ 2int64s
   2dup 1 0 d= if 2drop 3 exit then          \ double
   2dup 2 0 d= if 2drop 6 exit then          \ 2doubles
   cr ." Unknown return possibility: " swap . .
   abort ;

: parseArgs  \ <text> -- #in-ints #in-doubles outmask
\ Parse a parameter list of the form:
\   ( a b c d -- e ) and an optional second ( r1 r2 r3 r4 - r5 r6 )
   parse-params           ( #ints #int-rets )
   doubles? 0= if 0 dup   ( #ints #int-rets 0 0 )
       else parse-params  ( #ints #int-rets #doubles #double-rets )
   then
   rot outmask ;

: setVoid	\ struct --
  0 over ai.Size !
  UINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.PSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: setInt64	\ struct --
  8 over ai.Size !
  SINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.PSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: set2Ints	\ struct -- ; also for LRANGE
  16 over ai.Size !
  lRANGEtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.PSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: setDouble	\ struct --
  8 over ai.Size !
  FLOATtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.FSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: set2Doubles	\ struct --
  16 over ai.Size !
  dPOINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.FSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: setReturn	\ outmask --
\ Set the return type as indicated by the mask
  case
    0 of  ReturnRes setVoid     endof
    1 of  ReturnRes setInt64    endof
    2 of  ReturnRes set2Ints    endof
    3 of  ReturnRes setDouble   endof
    6 of  ReturnRes set2Doubles endof
      -1 abort" Bad return data"          \ unknown mask
  endcase
;

-short-branches
: setArgs    \ #in-ints #in-doubles out-mask --
\ Generate the data handling structures. Varargs are indicated
\ by #in=-1.
   setReturn            \ set the return type
   >r ( *)                           \ dealing with the doubles later
   dup 0< if
      r> 2drop  1 to varargs?
   else
      0 to varargs?
      dup numArgs ! dup NumInts ! 0       \ set the int args first
      ?do  i ArgList[] setInt64  loop
      numArgs @   ( start index)
      r> ( *) dup numArgs +! NumFloats !  \ continue adding the doubles
      numArgs @ swap
      ?do  i ArgList[] setDouble  loop
   then
;
+short-branches

\ *P The following section provides shared library imports in the form:
\ *C   function: foo  ( a b c d -- x ) ( F: r1 r2 -- rx )
\ *P where the brackets *\b{must} be space delimited. The first argument
\ ** list represents the arguments seen by Forth as integers. The second
\ ** arguments list is optional, and is used when F.P. doubles are involved.
\ ** Imports use the default calling convention for the operating system.

also forth definitions

: FUNCTION:	\ "<name>" "<parameter list>" --
\ *G Generate a reference to an external function. The Forth name
\ ** is the same as the name of the external function.
\ ** The first parameter list is a Forth integers list plus an optional
\ ** second Forth doubles list. There should always be an integer list,
\ ** the first list, even when not used.
\ ** Use in the form:
\ *C  function: foo1 ( a b c d -- )
\ *C  function: foo2 ( a b c d -- e )
\ *C  function: foo3 ( a b c d -- e ) ( r1 r2 r3 -- )
\ *C  function: foo4 ( a b c d -- ) ( r1 r2 r3 -- r4 )
\ *C  function: foo5 ( a b c d -- ) ( r1 r2 r3 -- r4 r5 )
\ *C  function: foo6 ( -- ) ( r1 r2 r3 -- r4 )
\ *C  function: foo7 ( -- ) ( r1 r2 r3 -- r4 r5 )
\ *C  function: foo7 ( -- a ) ( r1 r2 r3 -- )
\ *P The inclusion of Forth stack indicators like S: F: N: R: is optional.
\ *P They will be ignored while parsing and are only informative on which
\ *P Forth stack you must place the parameters.
\ *P The returned value may be 0, 1 integer or 0, 1, 2 doubles
\ ** corresponding to void, int/long and 64 bit double
  initArgData  setFirstLine		\ SFP022
  singleName				\ name of Forth word and DLL func ; SFP020
  setDefExt				\ set default declaraction type
  parseArgs setArgs			\ the parameter parsing and setup for...
  GenerateImport
;

: ASCALL:	\ "<synonym-name>" "<name>" "<parameter list>" --
\ *G Generate a reference to an external function. The Forth name
\ ** is not the same as the name of the external function.
\ ** Use in the form:
\ *C  ascall: forthname funcname ( a b c d -- e )
  initArgData  setFirstLine		\ SFP022
  parse-name ExternName place		\ set name of Forth word
  parse-name ImportName place		\ set name of DLL function
  setDefExt				\ set default declaraction type
  parseArgs setArgs                     \ the parameter parsing and setup for...
  GenerateImport
;

: GLOBAL:	\ "<name>" --
\ *G Generate a reference to an external variable.
\ ** Use in the form:
\ *C  global: varname
  singleName genExternVar		\ generate variable ; SFP020
;

previous definitions
