\ SOCKETIO.FTH - Sockets interface for VFX Forth for Linux

((
Copyright (c) 2003, 2005, 2008, 2013, 2021
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20210512 SFP007 Changed ERRNO @ to ERRNO l@ cos int is 32 bits.
                Updated SOCKREADLEN to use MSG_WAITALL.
20130702 SFP006 Changed MSG_NOSIGNAL to SO_NOSIGPIPE for OS X.
20080212 VRW005 Initial port from Windows version
20061214 MPE004 Corrected network problem.
20061206 MPE003 Update (CONNECT) to handle numeric IP addresses.
20060407 MPE002 Corrected SD-OPEN and SD-CLOSEONERR.
20050820 MPE001 Added PollSock.
))

decimal

\ ========
\ *> genio
\ ========
\ *N Sockets
\ *P This Generic IO Device operates on Mac OS X socket
\ ** for input or output. General socket programming words
\ ** are made available in the Forth vocabulary.

\ *P In order to create a device use the *\fo{SOCKDEV:} definition
\ ** given later.

\ *P When opening a socket device the parameters to *\fo{OPEN-GEN}
\ ** have the following meaning:
\ *E    ADDR            Address of configuration data structure
\ **    LEN             connection name zstring
\ **    ATTRIBS         0=socket, 1=connect, 2=listen.


\ **************
\ *H Sockets API
\ **************
\ *P Many of the BSD socket functions are defined. Note that the
\ ** *\b{accept} function is accessed by *\fo{SACCEPT} to avoid a
\ ** name clash with the ANS word *\fo{ACCEPT}.


only forth definitions  decimal

\ Generic sockets & BSD functions
extern: int OSCALL socket( int, int, int );
extern: int OSCALL close( int );
extern: int OSCALL shutdown( int s, int how );
extern: uint16 OSCALL htons( uint16 );
extern: uint32 OSCALL htonl( uint32 );
extern: uint16 OSCALL ntohs( uint16 );
extern: uint32 OSCALL ntohl( uint32 );
extern: int OSCALL bind( int, void *, unsigned int );
AliasedExtern: saccept int OSCALL accept( int, void *, int *);
\ *G Because the sockets *\b{accept} function has a name clash
\ ** with the Forth word *\fo{ACCEPT} it is made available as
\ ** *\fo{SACCEPT}.
extern: int OSCALL listen( int, int );
extern: int OSCALL connect( int s, void * name, int namelen );
[undefined] ioctl [if]
\ *\fo{ioctl} will almost certainly have been defined before now
extern: int OSCALL ioctl( DWORD s, long cmd, DWORD * argp );
[then]
extern: int OSCALL recv( HANDLE, char *, int, int );
extern: int OSCALL send( HANDLE, char *, int, int );
extern: int OSCALL getsockname( int s, void * name, int * namelen );
extern: void * OSCALL gethostbyname( char * name );
extern: void * OSCALL gethostbyaddr( void * addr, int len, int type );
extern: int OSCALL gethostname( char * name, int namelen );
extern: int OSCALL inet_addr( char * cp );
extern: char * OSCALL inet_ntoa( int in );

extern: unsigned int OSCALL sleep( unsigned int seconds);
extern: int OSCALL poll(void * , unsigned int , int );

extern: int OSCALL getaddrinfo( char * znode, char * zserv, void * hints, void * res );
Extern: void OSCALL freeaddrinfo( void * res );
extern: int OSCALL getifaddrs( void ** ifap );
extern: void OSCALL freeifaddrs( void * ifp );

struct /sockaddr	\ -- len
  1  field sa.len	\ structure
  1  field sa.family	\ type
  14 field sa.data	\ see /sockaddr_in
end-struct

struct /sockaddr_in	\ -- len
  2 field sin_family
  2 field sin_port
  4 field sin_addr
  8 field sin_reserved
end-struct
/sockaddr_in constant sockaddr_in

struct HOSTENT	\ -- len
        cell    field h_name
        cell    field h_aliases
        cell    field h_addrtype
        cell    field h_length
        cell    field h_addr_list
end-struct

struct /ifreq	\ -- len
  IFNAMSIZ field ifr.name
  #16 field ifr.result
end-struct

((
\ These are in the VFXsupp.so library
$4004667F constant FIONREAD
$8004667E constant FIONBIO
))

[undefined] SOCKET_ERROR [if]
-1 constant SOCKET_ERROR        \ Might not be defined on non-Windows platforms
[then]


\ ****************************************
\ *H Network order (big-endian) operations
\ ****************************************
\ *P TCP/IP protocols usually send data in what is called
\ ** network order, which just means most-significant byte
\ ** first. In memory, numbers are thus stored in big-endian
\ ** form. The following words provide memory operations
\ ** for this. These functions have to be capable of fetching
\ ** 32 bit cells from 16 bit aligned addresses, not just from
\ ** 32 bit aligned addresses.

\ Note also that these routines assume a byte-addressed CPU.

\ little-endian? [if]
: w@(n)         \ addr -- u16
\ *G Network order 16 bit fetch.
  dup c@ 8 lshift swap 1 + c@ or
;

: w!(n)         \ u16 addr --
\ *G Network order 16 bit store.
  over ( $FF and ) over 1 + c!
  swap 8 rshift swap c!
;
\ [else]
\ : w@(n)		w@  ;
\ : w!(n)		w!  ;
\ [then]

: @(n)          \ addr -- u32
\ *G Network order 32 bit fetch.
  dup c@ 8 lshift
  over 1 + c@ or 8 lshift
  over 2 + c@ or 8 lshift
  swap 3 + c@ or
;

: !(n)		\ u32 addr --
\ *G Network order 32 bit store.
  2dup 3 + c!  swap 8 rshift swap
  2dup 2 + c!  swap 8 rshift swap
  2dup 1 + c!  swap 8 rshift swap
  c!
;

: w,(n)		\ w --
\ *G Network order W,
  here w!(n)  2 allot
;

: ,(n)		\ x --
\ *G Network order version of , (comma).
  here !(n)  cell allot
;


\ ************************************
\ *H General socket functions in Forth
\ ************************************
\ *P These words are available in the *\fo{FORTH} vocabulary
\ ** for general socket programming.

max_path buffer: IPname	\ -- addr
\ *G Holds the local computer's name as a zero terminated string.

[undefined] IPaddress [if]
2 cells buffer: IPaddress	\ -- addr
\ *G Holds the local computer's IP address as a four byte
\ ** IPv4 number in network order. A value of 0 indicates
\ ** that the address is unknown.
  IPaddress 2 cells erase
[then]

: findLinkIP	\ caddr len addr --
\ *G Find the IP address assigned to the given link, e.g. eth0,
\ ** and place the link address at *\i{addr}.
  {: | resaddr sd ifr[ /ifreq ] -- :}
  ifr[ /ifreq erase
  to resaddr
  #15 min ifr[ ifr.name zplace
  AF_INET ifr[ ifr.result sin_family w!
  AF_INET SOCK_STREAM 0 socket to sd
  sd SIOCGIFADDR ifr[ ioctl drop
  sd close drop
  ifr[ ifr.result sin_addr @ resaddr !
\  ifr[ /ifreq dump
;

((
struct ifaddrs {
	struct ifaddrs  *ifa_next;
	char		*ifa_name;
	unsigned int	 ifa_flags;
	struct sockaddr	*ifa_addr;
	struct sockaddr	*ifa_netmask;
	struct sockaddr	*ifa_dstaddr;
	void		*ifa_data;
};
))
struct /ifaddrs	\ -- len
  int ifa.*next		\ pointer to next ifaddrs structure
  int ifa.*name		\ pointer to interface name
  int ifa.flags
  int ifa.*addr		\ pointer to interface address sockaddr
  int ifa.*netmask	\ pointer to interface netmask sockaddr
  int ifa.*dstaddr	\ pointer to interface P2P destination sockaddr
  int ifa.*data		\ pointer to address specific data
end-struct

: findCurrIP	\ addr --
\ *G Place the IP address of the current network at *\i{addr}.
\ ** This word assumes that only one link is active at a time.
\ ** The assumption is usually true.
  {: | interfaces[ cell ] -- :}
  interfaces[ off
  interfaces[ getifaddrs
  if  drop exit  endif
  interfaces[ @ begin
    dup
   while
    dup ifa.*addr @ sa.family c@ AF_INET = if
      dup ifa.*addr @ 16 dump key drop
    endif
    ifa.*next @
  repeat
  drop
  interfaces[ @ freeifaddrs
;

#256 buffer: NetIF$	\ -- addr
\ *G Holds the name of the default network device, usually *\b{en0}.
\ ** This is a counted string.
  s" en0" NetIF$ place

: InitOSXSockets	\ --
\ *G Initialise sockets; called by the cold chain and during
\ ** compilation.
  IPname MAX_PATH gethostname		\ returns 0 on success
  if  exit  endif
  IPname gethostbyname ?dup if		\ try for generic IP address
    h_addr_list @ @ @ IPaddress !
  then
  IPaddress c@ #127 =			\ if it is local
  IPaddress @ 0= or			\ or unknown
  if
    NetIF$ count IPaddress findLinkIP	\ find IP address of eth0
  endif
;
' InitOSXSockets AtCold  InitOSXSockets

: ?sockerr	\ serr -- ior
\ *G If *\i{serr} is -1, the actual errno value is
\ ** returned, otherwise zero is returned.
  -1 =
  if  errno l@  else  0  endif
;

: writesock	\ c-addr u hsock -- len ior
\ *G Write the buffer to a socket, returning the length actually
\ ** written and 0 on success, otherwise returning SOCKET_ERROR
\ ** and the Linux error code.
  -rot SO_NOSIGPIPE send dup ?sockerr
;

: readsock	\ c-addr u hsock -- len ior
\ *G Read into a buffer from a socket, returning the length actually
\ ** read and 0 on success, otherwise returning SOCKET_ERROR
\ ** and the Linux error code.
  -rot read dup ?sockerr
;

: pollsock	\ hsock -- #bytes|-1
\ *G Poll a socket and return the number of bytes available to be
\ ** read.
  { | #bytes[ cell ] -- }
  #bytes[ off
  FIONREAD #bytes[ ioctl -1 =
  if  -1  exit  endif
  #bytes[ @ 0 max
;

((
: sockReadLenOld	\ caddr len hsock -- ior
\ *G Read *\i{len} bytes of input from a socket to the buffer
\ ** at *\i{caddr}, returning ior=0 if all bytes have been read.
\ ** This is a blocking function which will not return until
\ ** *\i{len} bytes have been read or an error occurs.
  >r
  begin					\ -- caddr len ; R: -- hsock
    dup
   while
    begin				\ wait for input
      r@ pollsock ?dup 0=
     while
      #10 ms
    repeat
    -1 =
    if  2drop r> drop  -1 exit  endif
    2dup r@ readsock
    if  drop 2drop r> drop  -1  exit endif
    /string
  repeat
  2drop r> drop  0
;
))

: sockReadLen  \ caddr len hsock -- ior
\ *G Read *\i{len} bytes of input from a socket to the buffer
\ ** at *\i{caddr}, returning ior=0 if all bytes have been read.
\ ** This is a blocking function which will not return until
\ ** *\i{len} bytes have been read or an error occurs
\ ** or a signal is caught or a different type is received.
    -rot dup >r       \ hsock caddr len ; R: -- len
    \ ssize_t recv(int sockfd, void *buf, size_t len, int flags);
    MSG_WAITALL recv  \ res             ; R: -- len
    dup 0> if
      r> = 0=         \ compare against requested length
    else
      r> drop
      ?sockerr        \ check for additional error
    then
;

: bindTo	\ hs af port ipaddr -- res
\ *G A non-BSD function that binds a socket to the given set of
\ ** address family (af, usually AF_INET), port (port) and
\ ** IP address (ipaddr). The returned result (res) is 0 for
\ ** success, otherwise -1. See *\fo{BIND}.
  { | sai[ SOCKADDR_IN ] sailen[ cell ] -- }
  sockaddr_in sailen[ !
  sai[ SOCKADDR_IN erase		\ wipe buffer
  sai[ sin_addr !(n)			\ set ip address
  sai[ sin_port w!(n)			\ set port
  sai[ sin_family w!			\ set af
  sai[ SOCKADDR_IN bind			\ bind it
;

-1 constant INVALID_SOCKET
0 constant INADDR_NONE

: (Connect)	\ caddr u port# socket -- socket ior
\ *G Attempt to connect to a server. The socket has already been
\ ** created in the appropriate mode.
\ ** *\i{Caddr/u} describes the server address either as a name
\ ** or an IPaddress string and *\i{port#} is the requested port.
\ ** If *\i{u} is zero, *\i{caddr} is treated as a 32 bit number
\ ** representing an IPv4 address.
\ ** On success, the socket and zero are returned, otherwise
\ ** SOCKET_ERROR and the Linux error code are returned.
  { c-addr len p# hsock | sin[ SOCKADDR_IN ] z[ MAX_PATH ] -- }

  hsock INVALID_SOCKET = if
    errno l@  exit
  then

  sin[ SOCKADDR_IN erase                \ fill in address type
  AF_INET  sin[ sin_family w!		\ and port number for
  p# sin[ sin_port w!(n)		\ server address struct

  len if				\ if non-zero, treat as string
    c-addr len z[ zplace		\ z version of the address string.
    z[ inet_addr dup -1 = if		\ Is it a name or IP?
      drop  z[ gethostbyname ?dup if	\ Use DNS to resolve the supplied name
        h_addr_list @ @ @ sin[ sin_addr !
      else
        -1 errno l@			\ -- s_err ior
        hsock close drop		\ DNS lookup failed!
        exit
      then
    else				\ numeric IP address
      sin[ sin_addr !			\ store it ; MPE004
    then
  else					\ numeric IP address
    c-addr sin[ sin_addr !(n)		\ store it
  then

  hsock sin[ SOCKADDR_IN connect if	\ Attempt to connect socket using server
    -1 errno l@
    hsock close drop
    exit
  then

  hsock 0
;

: TCPConnect	\ c-addr u port# -- socket ior
\ *G Attempt to create a TCP socket and connect to a server.
\ ** */i{Caddr/u} describes the server address either as a name
\ ** or an IPaddress string and *\i{port#} is the requested port.
\ ** If *\i{u} is zero, *\i{caddr} is treated as a 32 bit number
\ ** representing an IPv4 address.
\ ** On success, the socket and zero are returned, otherwise
\ ** SOCKET_ERROR and the Linux error code are returned.
  AF_INET SOCK_STREAM IPPROTO_TCP socket	\ Create a socket
  (connect)
;

: UDPConnect	\ c-addr u port# -- socket ior
\ *G Attempt to create a TCP socket and connect to a server.
\ ** */i{Caddr/u} describes the server address either as a name
\ ** or an IPaddress string and *\i{port#} is the requested port.
\ ** If *\i{u} is zero, *\i{caddr} is treated as a 32 bit number
\ ** representing an IPv4 address.
\ ** On success, the socket and zero are returned, otherwise
\ ** SOCKET_ERROR and the Linux error code are returned.
  AF_INET SOCK_DGRAM IPPROTO_UDP socket	\ Create a socket
  (connect)
;

: closesocket
\ This appears to be used, even though I don't think it's part of the
\ official API...
  close
;

: ioctlsocket
\ Another alias
  ioctl
;


\ ****************
\ *H Socket device
\ ****************

\ *P A socket device is created by *\fo{SOCKETDEV: <name>}.
\ *E SocketDev: SDsid  \ -- addr

\ *P When opening a socket device the parameters to *\fo{OPEN-GEN}
\ ** have the following meaning:
\ *E    ADDR            Address of an /SDopen data structure.
\ **    LEN             Address of Windows IP address zstring
\ **                    If 0, /SDopen contains the IP adress.
\ **    ATTRIBS         mode: 0=socket, 1=connect,
\ **			2=listen
\ *P The following constants define the modes used to open
\ ** socket:
\ *C   SD_SOCKET SD_CONNECT SD_LISTEN
0 constant sd_socket
1 constant sd_connect
2 constant sd_listen

struct /SDopen	\ -- len
\ *G The structure required for opening a socket Generic I/O
\ ** device. Not all fields are used by all modes.
\ ** The */fo{/SDopen} structure is defined as follows:
\ *[
  int SDO.af                \ address family, usually AF_INET
  int SDO.type              \ socket type, e.g. SOCK_STREAM
  int SDO.protocol	    \ IPPROTO_TCP ...
  sockaddr_in field SDO.sa  \ SOCKADDR_IN structure
end-struct
\ *]

\ *P The *\fo{SDO.af} field is AF_INET for all TCP/IP operations.
\ ** The *\fo{SDO.TYPE} field is SOCK_STREAM for TCP or SOCK_DGRAM
\ ** for UDP. The *\fo{SDO.protocol} field is IPPROTO_TCP for TCP
\ ** or IPPROTO_UDP for UDP. The *\fo{SDO.sa} field is a SOCKADDR or
\ ** SOCKADDR_IN structure (same sizes), defined as follows:
\ *E struct sockaddr_in	\ -- len
\ **   2 field sin_family        \ address family, usually AF_INET
\ **   2 field sin_port          \ port ; in network order
\ **   4 field sin_addr          \ IP address ; in network order
\ **   8 field sin_reserved      \ RFU
\ ** end-struct
\ *P Note that only the first field is stored in native
\ ** (little-endian for Intel i32) order. The other fields
\ ** contain data in network (big-endian) form.

\ *P To open a socket, fill in a *\fo{/SDopen} structure, and
\ ** call *\fo{OPEN-GEN}. The following example connects to a
\ ** server.
\ *E SocketDev: SDsid  \ -- addr ; device
\ **
\ ** create zserver$   \ -- z$addr ; server name
\ **   z", www.mpeforth.com"
\ **
\ ** create MySDopen   \ -- addr
\ **   AF_INET ,       \ internet family
\ **   SOCK_STREAM ,   \ connection type
\ **   IPPROTO_TCP ,   \ TCP protocol
\ **   AF_INET w,      \ server family, start of SOCKADDR_IN
\ **   #80 w,(n)       \ server port
\ **   #0 ,(n)         \ server IP address if known
\ **   8 allot&erase   \ reserved
\ ** ...
\ ** MySDopen zserver$ SD_connect SDsid open-gio
\ *P This will return the sid again and a result code
\ ** (0=success).

\ *P The socket can then be used as the current I/O device.
\ *E : UseSDsid        \ --
\ **   SDsid dup op-handle !  ip-handle !
\ ** ;

#1440 constant /sock.opbuff	\ -- n ; size of o/p buffer
\ #1440 constant /sock.ipbuff	\ -- n ; size of i/p buffer

struct /socket-sid	\ -- len
\ *G Defines the SID of a socket device.
  gen-sid +				\ reuse field names of GEN-SID
  /SDopen field sock.sdo		\ /SDopen structure
  int sock.err				\ true when error has occurred
  int sock.#				\ debug counter
\ This next one is no longer used - but included to keep compatability
  int sock.hevent			\ WSAEVENT handle
  int sock.flags			\ flags for device control
  	\ bit 0 = no-echo
  int sock.#op				\ number of bytes buffered
  /sock.opbuff field sock.opbuff	\ output buffer
  aligned
end-struct

module genio-socket

[-sin
: sd-closeOnErr	\ sid -- sid ior
  errno l@				\ -- sid ior
  over gen-handle @ close drop	\ -- sid ior
  over INVALID_SOCKET swap gen-handle !	\ clean sid structure ; MPE002
;

: sd-open	\ sdo zname mode sid -- sid ior
  { sdo zname mode sid -- sid ior }

\ create the socket
  sdo dup SDO.af @
  over SDO.type @
  rot SDO.protocol @			\ MPE002
  socket dup sid gen-handle !
  INVALID_SOCKET = if
    sid  errno l@  exit
  then

\ if zname exists, get the IP address
  zname if				\ ignore zname=0
    zname c@ if				\ ignore zNULL
      zname inet_addr		        \ try to convert to IP address
      dup INADDR_NONE = if		\ name or IP?
        drop				\ name
        zname gethostbyname ?dup 0= if	\ Use DNS to resolve the supplied name
          sid sd-closeOnErr  exit
        then
        h_addr_list @ @ @		\ get IP address
      then
      sdo sdo.sa sin_addr !
    then
  then
  sdo sid sock.sdo /SDopen cmove	\ copy /SDopen structure

  case  mode
    sd_socket of  sid 0  endof
    sd_connect of
      sid gen-handle @ sdo SDO.sa SOCKADDR_IN connect
      if  sid sd-closeOnErr  else  sid 0  endif
    endof
    sd_listen of
      sid gen-handle @
      sid sock.sdo sdo.sa
      sockaddr_in bind if
        sid sd-closeOnErr
      else
        sid gen-handle @ SOMAXCONN listen
	if  sid sd-closeOnErr  else  sid 0  endif
      endif
    endof
      sid -1  rot
  endcase
;

: ?MarkErr	\ ior sid -- ior ; set sock.err if ior <> 0
  over if  over swap sock.err !  else  drop  endif
;

: sd-read	\ caddr len sid -- ior
  dup >r
  gen-handle @ -rot read
  ?sockerr r> ?MarkErr
;

: sd-readex	\ caddr len sid -- len ior
  dup >r
  gen-handle @ -rot read dup
  ?sockerr r> ?MarkErr
;

: (sd-xmit)	\ caddr len sid -- skerr
\  gen-handle @ -rot write
  gen-handle @ -rot SO_NOSIGPIPE send
;

: sd-flush	\ sid -- ior
\ *G Output to the socket is buffered to avoid running out
\ ** of Linux buffers. Call *\fo{FLUSHOP-GEN ( -- ior )} or
\ ** *\fo{KEY?} to transmit the buffered output.
  dup sock.#op @ if
    dup >r
    dup sock.opbuff  over sock.#op @  rot (sd-xmit)
    r@ sock.#op off
    ?sockerr r> ?MarkErr
  else
    drop  0
  endif
;

: sd-close	\ sid -- ior
\ *G The close function flushes pending output, closes the
\ ** event object if used, performs shutdown with how=1, and
\ ** closes the socket.
  dup sock.hevent @ if
    dup sock.hevent off
  endif
  dup >r
\ What does "how=1" actually do?
  gen-handle @  dup 1 shutdown drop  close
  ?sockerr r@ ?MarkErr
  INVALID_SOCKET r> gen-handle !	\ clean sid structure
;

: sd-buffop	\ caddr len sid --
  2dup 2>r
  dup sock.opbuff  swap sock.#op @ +  swap cmove
  2r> sock.#op +!
;

: (sd-send)	\ caddr len sid -- skerr
  2dup sock.#op @ + /sock.opbuff u> if
    dup sd-flush ?dup
    if  nip nip nip  exit  endif
    begin
      over /sock.opbuff u< not
     while
      >r
      over /sock.opbuff r@ (sd-xmit)	\ -- caddr len skerr ; R: -- sid
      -1 = if
        2drop  r> drop  -1  exit
      endif
      /sock.opbuff /string
      r>
    repeat
  then
  over if  sd-buffop  else  3drop  endif
  0
;

: ?socket_error	\ sid ior --
  -1 =
  if  errno l@ swap sock.err !  else  drop  endif
;

: sd-type	\ caddr len sid --
\ *G Buffered output.
  over out +!
  dup >r
  (sd-send)
  r> swap ?socket_error
;

: sd-write	\ caddr len sid -- ior
\ *G Buffered output.
  dup >r
  (sd-send)
  ?sockerr r> ?MarkErr
;

: sd-emit?	\ sid -- flag
  { | #bytes[ cell ]  -- }
  #bytes[ off
  dup gen-handle @ FIONREAD #bytes[ ioctl
  if  errno l@ swap sock.err !  0  else  drop -1  endif
;

: sd-emit	\ char sid --
\ *G Buffered output.
  { | ch[ cell ] -- }
  swap ch[ c!
  >r  ch[ 1 r@ (sd-send)  r> swap ?socket_error
  1 out +!
;

: sd-cr		\ sid --
\ *G Buffered output.
  { | ch[ cell ] -- }
  #13 ch[ c!  #10 ch[ 1 + c!
  ch[ 2 rot sd-type
  out off  op-line# incr
  pause
;

: sd-lf		( sid -- )	#10 swap sd-emit  ;
: sd-ff		( sid -- )	#12 swap sd-emit  ;
: sd-bell	( sid -- )	#7 swap sd-emit  ;

: sd-bs		\ sid --
  out @ dup if
    swap				\ -- out sid
    #8 over sd-emit  bl over sd-emit  #8 swap sd-emit
    1- out !
  else
    2drop
  then
;

: (sd-key?)	\ sid -- #bytes
  { | #bytes[ cell ]  -- }
  dup sock.# incr
  dup sock.err @
  if  drop -1  exit  endif
  #bytes[ off
  dup gen-handle @ FIONREAD #bytes[ ioctl
  if  errno l@ swap sock.err !  -1  exit  endif
  drop
  #bytes[ @
;

: sd-key?	\ sid -- #bytes|-1
\ *G The *\fo{KEY?} primitive for a socket returns the number
\ ** of bytes available. If an error occurs, -1 is returned
\ ** and *\fo{KEY} returns CR (ASCII code 13) so that *\fo{KEY}
\ ** and *\fo{ACCEPT} do not block. Use the *\fo{IOCTL} function
\ ** if you want to test for a specific error return code.
\ ** Any buffered output is sent first.
  dup sd-flush drop  (sd-key?)
;

: sd-key	\ sid -- char
\ *G If an error occurs, CR (ASCII code 13) is returned.
\ ** Any buffered output is sent first.
  { | ch[ cell ] -- }
  dup sock.err @
  if  drop $0D  exit  endif
  dup sd-flush drop
  begin
    dup (sd-key?) 0=
   while
    1 ms
  repeat
  dup gen-handle @ ch[ 1 read -1 =
  if  errno l@ swap sock.err !  $0D  exit  endif
  drop
  ch[ c@
;

: sd-peek	\ sid -- char
\ Look at the next character without removing it from the socket.
  { | ch[ cell ] -- }
  gen-handle @ ch[ 1 MSG_PEEK recv drop
  ch[ c@
;

: sd-echo?		\ sid -- flag
\ Returns non-zero if echoing is required.
  sock.flags @ 1 and 0=
;

: sd-savech	\ char addr sid -- ; save as required
\ Save char at addr, and output the character if
\ ECHOING is needed.
  >r
  over swap c!				\ stash character
  r> sd-echo?                           \ echo it?
  if  emit  else  drop  then            \ process accordingly
;

: sd-?bs	\ pos sid -- pos' step
\ If pos is non-zero and echo is required, perform BS and
\ return the size of the step, 0 or -1.
  sd-echo? over 0<> and if		\ echo and pos non-zero?
    #8 emit  #32 emit  #8 emit
    1- -1
  else
    0                                   \   no
  then
;

: sd-accept	\ caddr maxlen sid -- len
  { | sid spos -- len }
  -> sid
  out @ -> spos

  0 -rot  bounds ?do			\ -- len
    case  sid sd-key
      #8 of  sid sd-?bs  endof          \ <BACKSPACE>
      #127 of  sid sd-?bs  endof	\ treat <DEL> as <BACKSPACE>
      #9 of                             \ <TAB>
        bl i sid sd-savech  1+ 1        \ treat as space
      endof
      #10 of  0  endof			\ <LF>, ignore
      #13 of                            \ <CR>
        bl i sid sd-savech		\ stash as space
	sid sd-key? if			\ attempt to remove LF
	  sid sd-peek #10 =
	  if  sid sd-key drop  endif
	endif
        leave                           \ and leave
      endof
      #12 of                            \ treat <FF> as <CR>
        bl i sid sd-savech		\ stash as space
        leave                           \ and leave
      endof
        i sid sd-savech  1+ 1		\ DEFAULT
    end-case
  +loop
  dup spos + out !
;

: sd-setpos	( d mode sid -- ior )  2drop 2drop  -1  ;
: sd-Getpos	( mode sid -- d ior )  2drop  0 0 -1  ;

struct /pollfd
  cell field pollfd.fd
  2 field pollfd.events
  2 field pollfd.revents
end-struct

: sd-closed?	\ x x sid -- flag ; OBSOLETE
  { | buff[ /pollfd ] -- }
  buff[ /pollfd erase				\ Need events field to be zeroed
  nip nip				\ -- sid
  gen-handle @ buff[ !
  buff[ 1 1 poll drop
  buff[ pollfd.revents w@
  POLLNVAL and 0= 0=
;

: sd-ioctl	( addr len fn# sid -- ior )
\ *G The IOCTL primitive for a socket is used to get or set
\ ** socket status. The following functions
\ ** are supported by *\fo{IOCTL-GEN} for sockets.
\ *C   addr 0 #10 sid -- ior
\ *P Place the number of bytes available to be read by *\fo{recv}
\ ** at addr.
\ *C   0 0 #11 sid -- ior
\ *P Set the socket to notify when closed.
\ ** N.B. This is not currently implemented
\ *C   0 0 #12 sid -- ior
\ *P Ior is returned non-zero if the socket has been closed.
\ ** Ior is is returned false (zero) if the socket is still
\ ** open or notification has not been requested. The socket
\ ** must be open.
\ *C   state FD_xxx #20 sid -- ior
\ *P Set the created socket to notify on the FD_xxx flags. If
\ ** state is zero the socket is set/restored to blocking mode
\ ** otherwise it is set to blocking mode.
\ ** N.B. This is not currently implemented
\ *C   state FD_xxx #21 sid -- flags
\ *P Flags contains FD_xxx bits which indicate what events have
\ ** occurred from the set rquested by the call above.
\ ** Flags is is returned false (zero) if no events have been
\ ** reported or notification has not been requested. The
\ ** socket must be open.
\ ** N.B. This is not currently implemented
\ *C   0 0 #22 sid -- ior
\ *P Reset any notifications returned by function 21 above.
\ ** Ior is zero for success or the Linux error code.
\ ** N.B. This is not currently implemented
\ *C   0 0 #23 sid -- ior
\ *P Stop notification.
\ ** Ior is zero for success or the Linux error code.
\ ** N.B. This is not currently implemented
\ *C   0 flags #30 sid -- 0  ; set the device flags
\ *C   0 0     #31 sid -- flags ; get the device flags
\ *P The device flags control how some operations behave.
\ ** Flags is a set of bits as follows:
\ *E Bit 0 - set to stop echoing during *\fo{ACCEPT}.
  case  swap
    #10 of				\ -- addr len sid
      nip				\ -- addr sid
      over off
      gen-handle @ FIONREAD rot ioctl ?sockerr
    endof
    #12 of  sd-closed?  endof

    #30 of  sock.flags !  drop 0  endof
    #31 of  nip nip  sock.flags @  endof
      2drop 2drop  -1
  end-case
;

: sd-init	\ addr len sid -- ior
  3drop 0
;

: sd-term	\ sid -- ior
  drop 0
;

: sd-config	\ sid -- ior
  drop 0
;
sin]

create sd-vectors
  ' sd-open ,			\ addr len mode sid -- sid ior
  ' sd-close ,			\ sid -- ior
  ' sd-read ,			\ caddr len sid -- ior
  ' sd-write ,			\ caddr len sid -- ior
  ' sd-Key ,			\ sid -- char
  ' sd-Key? ,			\ sid -- ior
  ' sd-Key ,			\ sid -- char ; EKEY
  ' sd-Key? ,			\ sid -- ior ; EKEY?
  ' sd-accept ,			\ caddr len sid -- len2
  ' sd-Emit ,			\ char sid --
  ' sd-Emit? ,			\ sid -- ior
  ' sd-Type ,			\ caddr len sid --
  ' sd-CR ,			\ sid --
  ' sd-LF ,			\ sid --
  ' sd-FF ,			\ sid --
  ' sd-BS ,			\ sid --
  ' sd-bell ,			\ sid --
  ' sd-setpos ,			\ d mode sid -- ior
  ' sd-Getpos ,			\ mode sid -- d ior
  ' sd-Ioctl ,			\ addr len fn# sid -- ior
  ' sd-Flush ,			\ sid -- ior
  ' sd-readEx ,			\ caddr len sid -- u ior
  ' sd-Init ,			\ addr len sid --ior
  ' sd-Term ,			\ sid -- ior
  ' sd-Config ,			\ sid -- ior


\ ******************
\ *H Device Creation
\ ******************

: InitSD	\ addr --
\ *G Initialise the data required for a socket device at addr.
  dup /socket-sid erase
  -1 over gen-handle !
  sd-vectors swap gen-vector !
;

: SocketDev:	\ "name" -- ; Exec: -- sid
\ *G Create a new socket device called *\fo{name} in the dictionary.
  create
    here  /socket-sid allot  InitSD
;

export InitSD
export SocketDev:

end-module


\ *********
\ Test code
\ *********


\ ======
\ *> ###
\ ======

only forth definitions  decimal

