\ TimeBase.fth - OSX64 high level timers

((
Copyright (c) 1998, 1999, 2005, 2008, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com


Change history
==============
20120528 SFP006 Updated for changed critical sections.
20110201 SFP005 OS X conversion.
20100717 SFP004 Converted to use timerIDs.
20080114 SFP003 Converted to Linux use.
20050907 MPE002 Moved TIMERID to the SYSTEM vocabulary.
20050722 MPE001 Preserved old version of MS and start timers
                in compilation and at cold start.
))


\ ===========
\ *! timebase
\ *T Periodic Timers
\ ===========
\ *P This code provides a timer system that allows many timers to
\ ** be defined, all slaved from a single periodic interrupt. The
\ ** Forth words in the user accessible group documented below are
\ ** compatible with the token definitions for the PRACTICAL virtual
\ ** machine and with the code supplied with MPE's embedded targets.
\ ** This code assumes the presence of *\fo{TICKS}
\ ** which returns a time value incremented in milliseconds.

\ *P The timebase is approximate, and granularity and jitter are
\ ** affected by OS X and the time taken to run your own code.
\ ** By default, the timer is set to run every 100ms. The source
\ ** code is in the the file *\i{Lib/Osx64/TimeBase.fth}.
\ ** If you are using the multitasker, you *\b{must} compile
\ ** *\i{TimeBase.fth} after *\i{MultiOsx64.fth}.

\ *P The timer chain is built using a buffer area, and two chain
\ ** pointers. Each timer is linked into either the free timer
\ ** chain, or into the active timer chain.

\ *P All time periods are in milliseconds. Note that on a 32 bit
\ ** system such as VFX Forth, these time periods must be less
\ ** than 2^31-1 milliseconds, say 596 hours or 24 days, whereas
\ ** if the code is ported to a 16 bit system, time periods must
\ ** be less than 2^15-1 milliseconds, say 32 seconds.

\ ***********************
\ *S The basics of timers
\ ***********************
\ *P These basic words are defined for applications to use the
\ **timer system. Other words are detailed elswhere in this chapter.

\ *E START-TIMERS	\ -- ; must do this first
\ ** STOP-TIMERS	\ -- ; closes timers
\ ** AFTER		\ xt ms -- timerid/0 ; runs xt once after ms
\ ** EVERY		\ xt ms -- timerid/0 ; runs xt every ms
\ ** TSTOP		\ timerid -- ; stops the timer
\ ** MS			\ period -- ; wait for ms

\ *P After the timers have been started, actions can be added. The
\ ** example below starts a timer which puts a character on the debug
\ ** console every two seconds.
\ *E start-timers
\ ** : t	\ -- ; will run every 2 seconds
\ **   [char] * emit
\ ** ;
\ ** ' t 2 seconds every	\ returns timerid, use TSTOP to stop it
\ *P The item on stack is a timer handle, use *\fo{TSTOP} to halt
\ ** this timer.

\ *P *\fo{AFTER} is useful for creating timeouts, such as is required
\ ** to determine if something has happened in time. *\fo{AFTER}
\ ** returns a timerid. If the action you are protecting happens
\ ** in time, just use *\fo{TSTOP} when the action happens, and
\ ** the timer will never trigger. If the action does not happen,
\ ** the timer event will be triggered.

\ ***********************************
\ *S Considerations when using timers
\ ***********************************
\ *P All timers are executed within a single callback, and so all
\ ** timer action words share a common user area. This has some
\ ** impact on timer action words. Since you do not know in which
\ ** order timer action words are executed, you must set up any
\ ** *\fo{USER} variables such as *\fo{BASE} that you may use,
\ ** either directly or indirectly.

\ *P The callback that handles all the timers sets *\fo{IP-HANDLE}
\ ** and *\fo{OP-HANDLE} to a default that corresponds to the
\ ** interactive Forth console. If you use Forth I/O words such
\ ** as *\fo{EMIT} and *\fo{TYPE} within a timer action, you
\ ** *\b{must} set *\fo{IP-HANDLE} and *\fo{OP-HANDLE} before
\ ** using the I/O. For the sake of other timer action routines
\ ** that may still be using default I/O, it is polite to save and
\ ** restore *\fo{IP-HANDLE} and *\fo{OP-HANDLE} in your timer
\ ** action words.

\ *P Do not worry about calling *\fo{TSTOP} with a timerid that
\ ** has already been executed and removed from the active timer
\ ** chain; if *\fo{TSTOP} cannot find the timer, it will ignore
\ ** the request.

\ *P Be sure to use *\fo{START-TIMERS} in your main task, so that
\ ** the timer is not destroyed if a thread terminates.

\ ************************
\ *S Implementation issues
\ ************************
\ *P The following discussion is relevant if you want to modify
\ ** this code or port it to an embedded target. Functionally
\ ** equivalent code is provided with MPE's Forth VFX cross
\ ** compilers. In the OS X environment, timer interrupts are
\ ** implemented by signals, callbacks and critical sections.

\ *P By default, the word *\fo{DO-TIMERS} is run from within
\ ** the periodic timer callback. You may have latency issues if
\ ** a large number of timers is used, or if some timer routines
\ ** take a considerable time. In this case, it may be be better
\ ** to set up the timer routine to *\fo{RESTART} a task which
\ ** calls *\fo{DO-TIMERS}, e.g.
\ *E : TIMER-TASK    \ --
\ **   <initialise>
\ **   BEGIN
\ **     DO-TIMERS STOP
\ **   AGAIN
\ ** ;
\ *P Such a strategy also permits you to use a fast timer, say 1ms,
\ ** for a clock, and to trigger *\fo{TIMER-TASK} every say 32 ms.

\ ********************
\ *S Timebase glossary
\ ********************

\ ***********************************
\ Structures, variables and constants
\ ***********************************

/semosx buffer: lpcs	\ -- sem
\ *G Semaphore used for critical sections.

#32 constant #timers	\ -- n ; maximum number of timers
\ *G A constant used at compile time to set the maximum number
\ ** of timers required. Each timer requires RAM as defined by
\ ** the *\fo{ITIMER} structure below.

struct itimer   \ -- len
\ *G Interval timer structure.
\ *[
  cell field itlink                     \ link to next timer ; MUST be first
  cell field itTimerId			\ timer ID
  cell field itinterval                 \ period of timer in MS
  cell field ittimeout                  \ next timeout
  cell field itmode                     \ mode/flags, 0=periodic, 1=one shot
  cell field itxt                       \ word to execute
end-struct
\ *]

#timers itimer * constant /itimers	\ size of required buffer
/itimers buffer: timer-chain            \ for all timer chains

variable active-timers                  \ pointer to head of active timers chain
variable free-timers                    \ pointer to head of free timers chain
variable nextTimerID			\ next timer ID to be allocated

#8192                                   \ timer errors
  dup constant err_no_timers  1+
  dup constant err_bad_timer  1+
drop

0 constant periodic                     \ timer modes
1 constant single-shot


\ *******************
\ Timer ID management
\ *******************

: FormID	\ struct/0 -- ID/0
  dup if
    NextTimerID @ 0=			\ 0 is bad!
    if  1 NextTimerID +!  then
    NextTimerID @ dup rot itTimerID !
    1 NextTimerID +!
  endif
;

: FindID	\ id -- struct/0
  active-timers @
  begin
    dup
   while				\ get out at end of chain
    2dup itTimerID @ =			\ if this timer is the one
    if  nip  exit  endif		\   get out returning structure
    itlink @
  repeat
  nip					\ discard ID, return 0
;


\ **********************
\ Timer chain management
\ **********************

: init-timers   \ -- ; initialises timer system
  timer-chain /itimers erase            \ wipe buffer area
  0 active-timers !                     \ no active timers
  0                                     \ no free timers yet
  timer-chain /itimers bounds do
    i itlink !                          \ set old value
    i                                   \ this becomes new value
  itimer +loop
  free-timers !
;

: get-timer     \ -- struct ; throw if no free timer
  free-timers @ 0= err_no_timers ?throw
  lpcs [CritSec				\ make this safe!
  free-timers @                         \ new structure
  dup @ free-timers !                   \ unhook from free chain
  active-timers @ over itlink !         \ add to active timer chain
  dup active-timers !
  lpcs CritSec]
;

: release-timer \ struct -- ; exit quietly if timer can't be found
  lpcs [CritSec				\ make this safe!
  active-timers
  begin                                 \ -- struct s2
    2dup itlink @ <>                    \ itlink MUST be first
   while
     itlink @ dup 0= if			\ if at end of chain
       lpcs CritSec]
       2drop  exit			\ get out quietly
     endif
  repeat
\ struct prev -- ; take the next link, and place it in the previous link
  over itlink @  swap itlink !
  free-timers @ over itlink !           \ add to free chain
  free-timers !
  lpcs CritSec]
;

: (set-timer)   \ xt period mode -- struct ; set up timer
  get-timer
  tuck itmode !                         \ operating mode
  tuck  2dup itinterval !               \ period
  swap ticks +  swap ittimeout !        \ period+now = trigger
  tuck itxt !                           \ word to execute
;

: set-timer     \ xt period mode -- struct/0
  ['] (set-timer) catch
  if  2drop drop  0  endif
;

decimal

: do-timers     \ -- ; process all the timers in the chain
  active-timers @
  begin
    dup
   while                                \ -- struct
    dup itlink @ swap                   \ get next in chain
    dup ittimeout @  ticks - 0< if      \ if the interval has expired
      dup itxt @ execute                \ run the word
      dup itmode @ single-shot = if	\ if a single shot timer
        release-timer                   \ dispose of the timer
      else
        dup itinterval @                \ update the timeout point
        swap ittimeout +!
      endif
    else
      drop
    endif
  repeat
  drop                                  \ discard the end of chain
;


\ ****************
\ Diagnostic tools
\ ****************

: .t-chain      \ chain -- ; display a chain
  begin
    @ ?dup
   while
    dup .dword space
  repeat
;

: .t-active     \ -- ; display active timer chain
  cr ." Active timers:"
  cr
  active-timers .t-chain
;

: .t-free       \ -- ; display free timer chain
  cr ." Free timers:"
  cr
  free-timers .t-chain
;



\ **********************
\ User level timer words
\ **********************

: after		\ xt period -- timerid/0 ; xt is executed once,
\ *G Starts a timer that executes once after the given period.
\ ** A timer ID is returned if the timer could be started,
\ ** otherwise 0 is returned.
  single-shot set-timer formID
;

: every         \ xt period -- timerid/0 ; periodically
\ *G Starts a timer that executes every given period. A timer
\ ** ID is returned if the timer could be started, otherwise 0
\ ** is returned. The returned timerID can be used by *\fo{TSTOP}
\ ** to stop the timer.
  periodic set-timer formID
;

: tstop         \ timerid --
\ *G Removes the given timer from the active list.
  findID ?dup
  if  release-timer  endif
;

: seconds       \ n -- n'
\ *G Converts seconds to milliseconds.
  #1000 *
;


\ ***********************
\ Clock hardware handling
\ ***********************

decimal

Extern: int getitimer( int which, void * itimerval );
Extern: int setitimer( int which, void * nitimerval, void * otimerval );
0 constant ITIMER_REAL
1 constant ITIMER_VIRTUAL
2 constant ITIMER_PROF

struct /itimerval	\ -- len
\ *G Corresponds to the OS X *\b{itimerval} structure, which is
\ ** two *\b{timeval} structures.
\ *[
  int it.currsecs	\ current period
  int it.currusecs	\ in seconds & microseconds
  int it.nextsecs	\ next period (0 to stop)
  int it.nextusecs	\ in seconds & microseconds
end-struct
\ *]

/itimerval buffer: TBtimer	\ -- addr
\ *G Structure controlling the main timer.

: SetTimerData	\ ms timer --
\ *G Set a OS X *\b{itimerval} structure to run every *\i{ms}
\ ** milliseconds. Setting 0 ms will stop the timer.
  >r  1000 1000000 */mod		\ -- us secs ; R: -- timer
  2dup r@ it.currsecs 2!		\ current interval
  r> it.nextsecs 2!			\ next interval
;

#100 constant /period	\ -- ms
\ *G Main timer period in milliseconds.

3 0 callback: SIGALRMhandler	\ -- entrypoint
\ *G SIGALRM callback for main timer.

\ 0 value tbticks	\ for testing

: doSIGALRM	\ signum *siginfo *ucontext --
\ *G Action of main timer.
  3drop  ( 1 add tbticks )  do-timers
;  assign doSIGALRM to-callback SIGALRMhandler

also system definitions
0 value timers-running	\ -- x ; non-zero if timers running
previous definitions

also system
: start-timers  \ -- ; Start internal time clock
\ *G Initialises the timebase system, and starts the timebase
\ ** system. Note that all timer actions are cleared.
\ ** Performed at start up.
\  0 -> TBticks				\ for testing
  lpcs InitCritSec
  init-timers                           \ initialise time base system
  SIGALRMhandler SIGALRM setSignal	\ install handler
  /period TBtimer setTimerData		\ set timer period
  ITIMER_REAL TBtimer 0 setitimer drop	\ start timer
  true to timers-running
;
' start-timers AtCold  Start-Timers

: Stop-Timers   \ -- ; disable timer system
\ *G Halts the timebase interrupt.
  timers-running if
    false to timers-running
    0 TBtimer setTimerData		\ set timer period
    ITIMER_REAL TBtimer 0 setitimer drop \ start timer
    lpcs TermCritSec
  then
;
' Stop-Timers AtExit
previous


\ *********
\ Test code
\ *********

((
Start-timers

: t	\ -- ; will run every 2 seconds
  [char] * emit
;

' t 2 seconds every		\ returns a timer id, use TSTOP to stop it

cr
cr ." Item on stack is a timer handle, use TSTOP to halt this timer"
cr
))


\ ======
\ *> ###
\ ======

