\ XML.fth - simple XML support

((
Copyright (c) 2005, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====
Make SPAD structure and buffer contiguous?
Autoclean dynamic structures?

Change history
==============
20161021 SFP002 Overhauled DocGen for VFX Forth manual.
20161018 SFP001 Added control of display of unknown tags and attributes.
))

only forth definitions  decimal

\ ======
\ *> library
\ *S XML support
\ ======
\ *P The code in *\i{Lib\XML.fth} contains support for parsing XML input
\ ** and outputting XML using *\fo{TYPE} and friends. The parser
\ ** is derived from Jenny Brien's JenX parser published at EuroForth
\ ** and in the magazine ForthWrite. Additional code was taken
\ ** from a a modified JenX parser by Leo Wong. The generic XML
\ ** description is by permission of Willem Botha of Construction
\ ** Computer Software (http://www.ccssa.com).

\ *P Additional tools required for XML handling are contained
\ ** in this file. These may be moved to *\i{Lib\Win32\Helpers.fth}
\ ** in the future.


\ ==========
\ *N Why XML
\ ==========
\ *P Since XML is non-proprietary and easy to read and write, its an excellent
\ ** format for the interchange of data among different applications.

\ *P XML is a non-proprietary format, not encumbered by copyright,
\ ** patent, trade secret, or any other sort of intellectual property
\ ** restriction. It has been designed to be extremely powerful, while at the
\ ** same time being easy for both human beings and computer programs to read
\ ** and write. Thus its an obvious choice for exchange languages.

\ *P By using XML instead of a proprietary data format, you can use any tool
\ ** that understands XML to work with your data.

\ *P XML is ideal for large and complex documents because the data is
\ ** structured. It not only lets you specify a vocabulary that defines the
\ ** elements in the document; it also lets you specify the relations between
\ ** elements.

\ *P XML also provides a client-side include mechanism that integrates data
\ ** from multiple sources and displays it as a single document.

\ *P XML doesnt operate in a vacuum. Using XML as more than a data format
\ ** requires interaction with a number of related technologies. These
\ ** technologies include HTML for backward compatibility with legacy browsers,
\ ** the CSS and XSL stylesheet languages, URLs and URIs, the XLL linking
\ ** language, and the Unicode character set.

\ -------------------------
\ *H Cascading Style Sheets
\ -------------------------
\ *P Since XML allows arbitrary tags to be included in a document, there isnt
\ ** any way for the browser to know in advance how each element should be
\ ** displayed. When you send a document to a user you also need to send along
\ ** a style sheet that tells the browser how to format individual elements.
\ ** One kind of style sheet you can use is a Cascading Style Sheet (CSS).

\ *P CSS, initially designed for HTML, defines formatting properties like font
\ ** size, font family, font weight, paragraph indentation, paragraph
\ ** alignment, and other styles that can be applied to particular elements.
\ *P Its easy to apply CSS rules to XML documents. You simply change the names
\ ** of the tags youre applying the rules to.

\ ----------------------------
\ *H Extensible Style Language
\ ----------------------------
\ *P The Extensible Style Language (XSL) is a more advanced style-sheet
\ ** language specifically designed for use with XML documents. XSL documents
\ ** are themselves well-formed XML documents.

\ *P XSL documents contain a series of rules that apply to particular patterns
\ ** of XML elements. An XSL processor reads an XML document and compares what
\ ** it sees to the patterns in a style sheet. When a pattern from the XSL
\ ** style sheet is recognized in the XML document, the rule outputs some
\ ** combination of text.

\ *P XSL style sheets can rearrange and reorder elements. They can hide some
\ ** elements and display others. Furthermore, they can choose the style to use
\ ** not just based on the tag, but also on the contents and attributes of the
\ ** tag, on the position of the tag in the document relative to other
\ ** elements, and on a variety of other criteria.

\ ----------------
\ *H URLs and URIs
\ ----------------
\ *P XML documents can live on the Web, just like HTML and other documents.
\ ** When they do, they are referred to by Uniform Resource Locators (URLs),
\ ** just like HTML files.

\ *P Although URLs are well understood and well supported, the XML
\ ** specification uses the more general Uniform Resource Identifier (URI).
\ ** URIs are a more general architecture for locating resources on the
\ ** Internet, that focus a little more on the resource and a little less on
\ ** the location. In theory, a URI can find the closest copy of a mirrored
\ ** document or locate a document that has been moved from one site to
\ ** another.

\ -----------------------
\ *H XLinks and XPointers
\ -----------------------
\ *P As long as XML documents are posted on the Internet, youre going to want
\ ** to be able to address them and hot link between them. Standard HTML link
\ ** tags can be used in XML documents, and HTML documents can link to XML
\ ** documents.

\ *P XML lets you go further with XLinks for linking to documents and
\ ** XPointers for addressing individual parts of a document.

\ *P XLinks enable any element to become a link, not just an A element.
\ ** Furthermore, links can be bi-directional, multidirectional, or even point
\ ** to multiple mirror sites from which the nearest is selected. XLinks use
\ ** normal URLs to identify the site theyre linking to.

\ *P XPointers enable links to point not just to a particular document at a
\ ** particular location, but to a particular part of a particular document. An
\ ** XPointer can refer to a particular element of a document, to the first,
\ ** the second, or the 17th such element, to the first element thats a child
\ ** of a given element, and so on. XPointers provide extremely powerful
\ ** connections between documents that do not require the targeted document to
\ ** contain additional markup just so its individual pieces can be linked to
\ ** it. XPointers dont just refer to a point in a document. They can point to
\ ** ranges or spans.

\ ------------------------------------
\ *H How the Technologies Fit Together
\ ------------------------------------
\ *P XML defines a grammar for tags you can use to mark up a document. An XML
\ ** document is marked up with XML tags. The default encoding for XML
\ ** documents is Unicode.

\ *P Among other things, an XML document may contain hypertext links to other
\ ** documents and resources. These links are created according to the XLink
\ ** specification. XLinks identify the documents theyre linking to with URIs
\ ** (in theory) or URLs (in practice). An XLink may further specify the
\ ** individual part of a document its linking to. These parts are addressed
\ ** via XPointers.

\ *P If an XML document is intended to be read by human beingsand not all XML
\ ** documents arethen a style sheet provides instructions about how
\ ** individual elements are formatted. The style sheet may be written in any
\ ** of several style-sheet languages. CSS and XSL are the two most popular
\ ** style-sheet languages, though there are others including DSSSLthe
\ ** Document Style Semantics and Specification Languageon which XSL is based.


\ ***********************
\ *N Using the XML Parser
\ ***********************
\ *P All parsing is processed using the input stream. This allows
\ ** XML files to be parsed by *\fo{INCLUDE}, and strings from
\ ** sockets to be processed by *\fo{EVALUATE}.

\ *P The XML parser parses tags "<...>" and the text between them,
\ ** called the contents. Inside a tag the text is separated into
\ ** the tag name and the attribute name/value pairs 'name="value"'.
\ ** Everything is held as text. Nested tags are supported.
\ ** Three *\fo{DEFER}red words,
\ ** *\fo{doTags ( -- )},
\ ** *\fo{doContents ( -- )} and
\ ** *\fo{doAttribute ( val vlen name nlen -- )}
\ ** must be supplied by the application to handle the data. These
\ ** words are documented later. Their default action is to display
\ ** the data so that you can see what has been processed.

\ *P The parser just generates and isolates the text. It is up
\ ** to your application how the data is processed by the three
\ ** words above. When a tag is processed, the tag handling routine
\ ** can find the current tag name, the tag type, any attributes
\ **  and the preceeding contents. The most common way to process
\ ** tags and data is to ignore the contents before an opening
\ ** tag, but to handle attributes. At the closing tag, the
\ ** contents represent the data to be processed. Closing tag
\ ** names include the leading '/' character so that opening
\ ** and closing tags can be distinguished by name as well as
\ ** status.


\ ************************
\ *N Generating XML output
\ ************************
\ *P Simple facilities are provided for generating XML text
\ ** and tags from various types of data. These are designed
\ ** to allow other scripting tools to generate XML output.


\ ********
\ *N Tools
\ ********
\ *P This section contains general-purpose tools which may be useful
\ ** in other applications.

[undefined] equ [if]
: equ	constant  ;
[then]

1 value .UnknownXML?	\ -- flag
\ *G If non-zero (default), show unknown XML tags and attributes.


\ ==========
\ *H Strings
\ ==========

[undefined] movex [if]
: movex		\ src dest len --
\ *G An optimised version of *\fo{MOVE}.
  move
;
[then]

[undefined] csplit [if]
: csplit	\ addr len char -- raddr rlen laddr llen
\ *G Extract a substring at the start of addr/len, returning
\ ** the string raddr/rlen which includes char (if found) and
\ ** the string laddr/llen which contains the text to left of char.
\ ** If the string does not contain the character, raddr is
\ ** addr+len and rlen=0.
  >r 2dup r> scan                       \ -- addr len addr' len'
  2swap 2 pick -                        \ -- addr' len' addr len-len'
;
[then]

: #>c           \ caddr u -- char
\ *G Converts a decimal or hexadecimal number to a single integer.
  base @ >r
  over c@ upc [char] X =
  if  1 /string hex  else  decimal  endif
  case  (integer?)                      \ -- d 2 | n 1 | 0
    0 of  0  endof
    2 of  drop  endof
  endcase
  r> base !
;

\ *P In XML white space is defined by tab and CR. Under some
\ ** circumstances LF may also be treated as white space.

: white?	\ char -- flag
\ Return true if the character is to be treated as white space.
  dup 9 = over 13 = or swap bl = or
;

: skip-white	\ caddr u -- caddr' u'
\ *G Remove leading white space.
  begin  dup while  over c@ white? while  1 /string  repeat then
;

: scan-black	\ caddr u -- caddr' u.
\ *G Remove leading spaces and control characters.
  begin  dup while  over c@ bl 1+ u< while  1 /string  repeat then
;

: scan-quote	\ caddr u -- caddr' u'
\ *G Step forward until either a single or a double quote
\ ** character is found. The returned string includes the
\ ** quote character.
  begin
    dup
   while
    over c@ [char] ' <>
   while
    over c@ [char] " <>
   while
    1 /string
  repeat then then
;

: scan-white	\ caddr u -- caddr' u'
\ *G Step to next white space character.
  begin  dup while  over c@ white? 0= while  1 /string  repeat then
;

: -trailing-white	\ caddr u -- caddr' u'
\ *G Remove trailing white space.
  begin  dup while  1- 2dup chars + c@ white? 0= until  1+  then
;

: -leading-white	\ caddr u -- caddr' u'
\ *G Remove leading white space. A synonym for *\fo{skip-white}.
  begin  dup while  over c@ white? while  1 /string  repeat then
;

: -white	\ caddr u -- caddr' u'
\ *G Remove leading and trailing white space.
  -leading-white -trailing-white
; doNotSin

: >bl		\ addr u -- addr u
\ *G Convert control characters to spaces.
  2dup bounds ?do
    i c@ BL u<
    if  BL i c!  endif
  loop
;


\ =====================
\ *H Gregorian calendar
\ =====================
\ *P The output formats are:
\ weekday codes: 0=Sun 1=Mon ...  6=Sat
\ month   codes: 1=Jan 2=Feb ... 12=Dec

[-sin
: date>         \ day month year -- ud ; see month codes
\ *G Convert a day/month/year into a Gregorian day number.
  dup 100 < if 2000 + then >r           \ year<100: XXIth century
  dup 3 < if  12 + r> 1- >r  then       \ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 400 / + r@ 100 / - r@ 4 / +        \ leap-year corrections
  0 r> 365 um* d+                       \ and add 3/1/y - 3/1/0
;

: weekday       \ ud -- n ; see weekday codes
  2* 3 + 0 tuck d+ 7 um/mod drop        \ 65536=7n+2
;

: >date         \ ud -- day month year ; see month codes
  48699 um/mod 0 tuck 3 um/mod          \ no 4-century periods
  400 * >r  48699 um* d+                \ modulo 146097=48699*3
  36524 um/mod                          \ no of century periods
  dup 4 = if 2drop 36524 3 then         \ correct if Febr.29th
  100 * r> + >r                         \ accumulate years
  0 1461 um/mod                         \ no leap-year periods
  4 * r> + >r                           \ accumulate years
  365 /mod dup 4 = if 2drop 365 3 then  \ correct if Febr.29th
  r> + >r                               \ accumulate years
  123 + dup 5 153 */                    \ compute month
  dup >r 153 5 */ - r>                  \ compute day
  1- dup 12 > if 12 - r> 1+ >r then     \ correct month and year
  r>                                    \ -- day month year
;
sin]

1 1 1980 date> 2constant date0	\ -- ud
\ *G Defines day 0 as 1 Jan 1980 for dates.

: sdate>	\ day month year -- u
\ *G Convert a day/month/year to a single day integer
\ ** based as above.
  date> date0 d- drop
;

: >sdate	\ u -- day month year
\ *G Convert a single day integer to day/month/year
  0 date0 d+ >date
;


\ ===========
\ *H Day time
\ ===========
\ *P Time of day may be stored as a single integer count of
\ ** seconds. These routine provide conversion into secs/mins/hours
\ ** format.

#24 #60 * #60 * constant secs/day	\ -- 86400
\ *G Seconds per day.
#60 #60 * constant secs/hr	\ -- 3600
\ *G Seconds per hour.
#60 constant secs/min		\ -- 60
\ *G Seconds per minute.
#60 constant mins/hr		\ -- 60
\ *G Minutes per hour.
#24 constant hrs/day		\ -- 24
\ *G Hours per day.

: tod>		\ ss mm hh -- secs
\ *G Convert a time of day in ss/mm/hh form to a single
\ ** integer.
  secs/hr * swap  secs/min *  + +
;

: >tod		\ secs -- ss mm hh
\ *G Convert a seconds integer to ss/mm/hh form.
  #60 /mod  #60 /mod
;


\ ============
\ *H Stackpads
\ ============
\ *P Stackpads are effectively string stacks. String lengths are
\ ** kept as cells. Stackpads can be in statically (*\fo{ALLOT}ed)
\ ** or dynamically (*\fo{ALLOCATE}d) memory. A stackpad must be
\ ** initialised by *\fo{SINIT} before use and terminated by *\fo{STERM}
\ ** after use. In this implementation, defined stackpads are
\ ** initialised at *\fo{COLD} and terminated at *\fo{BYE}.

\ *P Strings on a stackpad are held in the following format,
\ ** where u is the length of the string in bytes:
\ *E len   contents
\ **  u    string text
\ **  ?    padding to cell boundary
\ **  cell u
\ *P The stackpad's top of stack pointer points to the length
\ ** cell of the top item. To provide a valid cell, a zero length item
\ ** is always created when the stackpad is initialised. Because
\ ** the length cell is after the text, it is easy to manipulate the
\ ** end of a string, to find the start address and to discard a
\ ** string.

\ *P The requirement to align the length cell adds a little
\ ** complexity, but permits portability to processors which
\ ** require data alignment, e.g. ARM, and improves speed on
\ ** PCs. Stackpads are controlled using the *\fo{/stackpad}
\ ** structure below. The *\fo{sp.ptos} field contains the stack
\ ** pointer. The *\fo{sp.buff} field permits underflow checks.
\ ** The *\fo{sp.len} field permits overflow checks.
\ ** The other fields allow for automatic instantiation
\ ** and termination of dynamically allocated stackpads.
\ ** Implementations without error checking only need the stack
\ ** pointer and could use the first cell of the buffer as the
\ ** stack pointer.

struct /stackpad	\ -- len
\ *G Structure defining a stackpad.
  int sp.prev		\ pointer to previous stackpad
  int sp.type		\ 0=static, 1=static/dynamic, 2=dynamic
  int sp.len		\ size of stackpad buffer
  int sp.buff		\ pointer to stackpad buffer
  int sp.ptos		\ pointer to top's length cell
end-struct

variable spChain	\ -- addr
\ *G Anchors the linked list of defined stackpads.
  0 spChain !

[-sin

: sSpad:	\ len -- ; -- spad
\ *G Create a static stackpad with *\fo{ALLOT}ed control area
\ ** and data buffer.
  create
    spChain link,			\ sp.prev
    0 ,  dup ,  here 2 cells + dup , , 	\ type, len, ^buff, ^tos
    0 ,  cell - allot			\ buffer
;

: mSpad:	\ len -- ; -- spad
\ *G Create a mixed stackpad with an *\fo{ALLOT}ed control area
\ ** and an *\fo{ALLOCATE}d buffer.
  create
    spChain link,			\ sp.prev
    1 ,  ,  0 , 0 ,			\ type, len, ^buff, ^tos
;

: newSpad	\ len -- spad
\ *G Create a dynamic stackpad with *\fo{ALLOCATE}d control area
\ ** and data buffer. A *\fo{THROW} occurs if the memory cannot
\ ** be allocated.
  dup /stackpad + allocate #-406 ?throw	\ request memory ; -- len spad
  2 over sp.type !			\ set type
  tuck sp.len !				\ set length ; -- spad
  dup /stackpad + tuck off		\ zero first data cell ; -- buffer spad
  2dup sp.buff !			\ set pointer to data buffer
  tuck sp.ptos !			\ set pointer to top item
  dup spChain AddLink
;

: sinit		\ spad --
\ *G Initialise a stackpad. A *\fo{THROW} occurs on error.
  dup sp.type @ 1 = if			\ mixed
    dup sp.len @ allocate #-406 ?throw
    over sp.buff !
  endif
  dup sp.buff @  over sp.ptos !		\ set pointer to top string
  0 swap sp.buff @ !			\ set null top string
;

: sterm		\ spad --
\ *G Release dynamic memory if the given stackpad has it.
  case dup sp.type @
    1 of				\ mixed
      sp.buff @off free #-407 ?throw
    endof
    2 of  				\ dynamic
      dup spChain DelLink
      free #-407 ?throw
    endof
          nip
  endcase
;

: initSpads	\ --
\ *G Initialise all defined stackpads. Performed at *\fo{COLD}.
  spChain @
  begin  dup while  dup sinit  sp.prev @  repeat
  drop
;
' initSpads AtCold

: termSpads	\ --
\ *G Clean up all defined stackpads, releasing any dynamically
\ ** allocated memory. Performed at *\fo{BYE}.
  spChain @
  begin  dup while  dup sterm  sp.prev @  repeat
  drop
;
' termSpads AtExit

sin]

: -align	\ caddr -- addr'
\ *G Align a byte address to the previous cell boundary.
\ ** N.B. This word assumes a byte addressed 32 bit Forth.
  -4 and
;

: >spstr	\ lp -- caddr u
\ *G Given a pointer to a length cell, return the string.
  dup @ tuck  - -align  swap
;

: >sps		\ lp -- caddr
\ *G Given a pointer to a length cell, find the start of the
\ ** string.
  dup @ - -align
;

: >spe		\ lp -- caddr
\ *G Given a pointer to the length cell, find the address of the
\ ** character after the string.
  >spstr +
;

: spush		\ caddr u spad --
\ *G Push a string onto the stackpad.
  swap >r				\ -- caddr spad ; R: -- u
  tuck sp.ptos @			\ -- spad caddr lp ; R: -- u
  cell+ r@ movex			\ -- spad ; R: -- u
  r@ cell+ aligned over sp.ptos +!	\ -- spad ; R: -- u
  r> swap sp.ptos @ !			\ --
;

: stos		\ spad -- caddr u
\ *G Return the address and length of the top string. The string
\ ** is *\b{not} popped.
  sp.ptos @ >spstr
;

: sdrop		\ spad --
\ *G Discard top string from stackpad.
  dup sp.ptos @  >sps cell -  swap sp.ptos !
;

: spop		\ spad -- caddr u
\ *G Return the address and length of the top string. The string
\ ** is popped. Note that the stackpad cannot safely be used until
\ ** all further processing of the string has been performed.
  dup stos rot sdrop
;

: snew		\ spad --
\ *G Add a zero-length string.
  cell over sp.ptos +!  sp.ptos @ off
;

: sappend	\ caddr u spad --
\ *G Add the given string to the top stackpad string.
  dup >r sp.ptos @	    		\ caddr u lp ; R: -- spad
  2dup @ + >r				\ caddr u lp ; R: -- spad len'
  >spe  2dup + aligned >r 		\ caddr u spe ; R: -- spad len' lp'
  swap movex				\ -- ; R: -- spad len' lp'
  r> r> over !				\ -- lp' ; R: -- spad
  r> sp.ptos !
;

: s+char	\ char spad --
\ *G Add the given character to the top stackpad string.
  dup >r sp.ptos @ 			\ -- char lp ; R: -- spad
  dup @ 1+ >r				\ -- char lp ; R: -- spad len'
  >spe tuck c!		  		\ -- spe ; R: -- spad len'
  1 + aligned  r> over !		\ -- lp' ; R: -- spad
  r> sp.ptos !
;

: .spad		\ spad --
\ *G Display the strings on a stackpad.
  dup sp.buff @  swap sp.ptos @		\ limit lp
  begin
    2dup u> not
   while
    halt? 0=
   while
    dup >spstr
    cr  over .dword space dup 4 .r
    ."  :" type ." :"
    >sps cell-
  repeat then
  2drop
;


\ ===========
\ *H Servants
\ ===========
\ *P Servants are a solution to *\fo{CASE} statements involving
\ ** strings. A wordlist is defined to hold the actions required
\ ** when a string is matched, the word names forming the strings
\ ** to be matched. A default action must be specified. Note that
\ ** in MPE Forths, the name search is case insensitive. Note also
\ ** that without extensions to the word creation mechanism, the
\ ** Because the strings are isolated in wordlists, calls may be
\ ** nested.

: (Servant)	\ i*x caddr u wid xt -- j*x
\ *G Looks up *\i{caddr/u} in the *\i{wid} wordlist. If the
\ ** word is found, it is executed. If the word is not found,
\ ** the *\i{caddr/u} string is passed to the default action
\ ** *\i{xt} which is executed.
  >r
  >r 2dup r> search-wordlist if		\ -- i*x caddr u xt2 ; R: -- xt
    r> drop  nip nip execute		\ -- j*x
  else					\ -- i*x caddr u ; R: -- xt
    r> execute				\ -- j*x
  then
;

: servant	\ wid xt -- ; i*x caddr u -- j*x
\ *G *\fo{Servant} creates a word that looks up *\i{caddr/u}
\ ** in a given wordlist and executes the matching word if found
\ ** or a default word if not found. *\fo{Servant} is supplied
\ ** with the *\i{wid} of the wordlist and the *\i{xt} of the
\ ** default action.
  create
    , ,					\ lay xt and wid
  does>					\ -- i*x caddr u ^data
   2@ (servant)
;

: creation	\ wid --
\ *G Perform *\fo{CREATE}, but define the word in the specified
\ ** wordlist.
  get-current swap set-current  create  set-current
;

: def:		\ wid --
\ *G Perform *\fo{:}, but define the word in the specified
\ ** wordlist.
  get-current swap set-current  :  set-current
;


\ *******************
\ *N XML input parser
\ *******************

\ ===============================
\ *H Required data and structures
\ ===============================

cell +user CurrSpad	\ -- addr
\ *G Holds the address of the stackpad being used for output.

cell +user RefillStatus	\ -- addr
\ *G Holds non-zero when *\fo{REFILL} has failed.

#32 kb mSpad: TagText	\ -- spad
\ *G Stackpad for tag text <tag ....>.
#32 kb mSpad: Contents	\ -- spad
\ *G Stackpad for everything not in a tag.
#32 kb mSpad: Attribs	\ -- spad
\ *G Stackpad for attribute handling in tags.


\ ===============
\ *H XML entities
\ ===============
\ *P In XML code the special characters and numbers are encoded
\ ** in the form:
\ *C   &xxx;
\ *P This code allows substitution of the original character.

: UnknownEntity \ caddr u --
\ *G The default action is to check for a number, and if that
\ ** fails just to pass the string to the output buffer.
\ ** Note that the string includes the leading '&' but
\ ** not the trailing ';'.
\ cr ." Unknown entity:"  2dup type  ." : " key drop
  1 /string				\ discard leading '&'
  dup 1 > if
    over c@ [char] # = if
      1 /string #>c CurrSpad @ s+char	\ convert to number
      exit
    endif
  endif
  CurrSpad @ sappend
;

wordlist constant entity?       \ -- wid
\ *G The private wordlist used to contain action words for
\ ** known entities.

: centity       \ char -- ; --
\ *G Children of this defining word add a character to the
\ ** current stackpad. The words are used by the servant
\ ** *\fo{DENT} below.
  entity? creation
    c,
  does>
    c@ CurrSpad @ s+char
;

\ *P The following standard XML entities are predefined:
\ *[
char < centity &LT
char > centity &GT
char ' centity &APOS
char " centity &QUOT
char & centity &AMP
\ *]

entity? ' UnknownEntity servant dent    \ caddr u --
\ *G A servant which converts known entities and XML numbers
\ ** of the form *\f{&#xxx;} to characters or just copies the
\ ** string to the current stackpad.

: dents+        \ caddr u --
\ *G Add the string to the top of the current stackpad,
\ ** decoding and translating any entities.
  begin
    [char] & csplit CurrSpad @ sappend	\ -- caddr' len'
    dup
   while                                \ string is "&xxx"
    [char] ; csplit dent  1 /string     \ process, skip ';'
  repeat
  2drop
;


\ ************
\ *H Tag input
\ ************

: .Tag		\ --
\ *G Default action of *\fo{doTags} below.
  cr ." TagText :" tagText stos type ." :"
  contents dup sdrop snew
;

: .Contents	\ --
\ *G Default action of *\fo{doContents} below.
  cr ." Contents:" contents stos type ." :"
;

: .Attribute	\ val vlen name nlen --
\ *G Display the attribute name and value strings.
  .UnknownXML? if
    cr ."   Attribute=" type ."  Value=" type
  else
    2drop 2drop
  endif
;

defer doTags		\ --
\ *G User defined action (default *\fo{.Tag}) that handles
\ ** tag strings. The tag handlers are responsible for all
\ ** handling of the *\fo{contents} stackpad. The top string
\ ** on the *fo{TagText} stackpad is discarded after processing
\ **  the tag text.
  assign .tag to-do doTags

defer doContents	\ --
\ *G User defined action (default *\fo{.Contents}) that handles
\ ** content strings. The *\fo{contents} stackpad is not
\ ** discarded by *\fo{doContents}.
  assign .Contents to-do doContents

defer doAttribute	\ val vlen name nlen --
\ *G Process an attribute given strings for the value and name.
\ ** The default action is to display the attrubte.
  assign .Attribute to-do doAttribute

: DefXML	\ --
\ *G Set the default XML handlers.
  assign .tag to-do doTags
  assign .Contents to-do doContents
  assign .Attribute to-do doAttribute
;

vocabulary inputTags	\ --
\ *G Vocabulary containing tag actions on input.
' inputTags voc>wid constant widInputs	\ -- wid
\ *G Wordlist containing tag actions on input.

((
vocabulary outputTags	\ --
\ *G Vocabulary containing tag actions on output.
' outputTags voc>wid constant widOutputs	\ -- wid
\ *G Wordlist containing tag actions on output.
))

#256 buffer: CurrName	\ -- addr
\ *G Buffer for the current tag name.
\ ** Held as a counted string.
\ ** For multi-threaded use this should be redefined
\ ** as thread-local storage.
#256 buffer: LastName
\ *G Buffer for the previous tag name.
\ ** Held as a counted string.
\ ** For multi-threaded use this should be redefined
\ ** as thread-local storage.
variable TagStatus
\ *G Status indicator for the current tag.
\ ** For multi-threaded use this should be redefined
\ ** as thread-local storage. The tag status is a bit mask in the
\ ** bottom 16 bits of a cell The upper 16 bits are reserved for
\ ** application use.
  0 TagStatus !
\ *[
  $0000 equ OPENING_TAG
  $0001 equ CLOSING_TAG
  $0002 equ EMPTY_TAG
  $0100 equ PI_TAG
  $0200 equ SPECIAL_TAG
\ *]

variable LastStatus
\ *G Status indicator for the previous tag.
\ ** For multi-threaded use this should be redefined
\ ** as thread-local storage.
  0 LastStatus !

: defInputTag	\ caddr u --
\ *G The default action for an unknown tag is to display the
\ ** content and tag strings.
  .UnknownXML? if
    cr
    cr ." Unknown :" type ." :  Status " TagStatus @ .dword
    cr ." Contents:" contents stos type ." :"
    cr ." Tag Text:" tagText stos type  ." :"
  else
    2drop  emptyIdle
  endif
  contents dup sdrop snew
;

widInputs ' defInputTag servant doInputTag	\ caddr u --
\ *G Processes input tags given a tag name string.

: getTagName	\ caddr u -- caddr' u' name nlen
\ *G From the given string, return the remaing string
\ ** and the tag name, which is the first whitespace
\ ** delimited token. Note that tag names include leading
\ ** '?' and '!' characters.
  scan-black 2dup 2>r
  scan-white dup 2r> rot -
  scan-black 2dup 1- chars + c@ [char] > =
  if  1-  then
;

: getAttribName	\ caddr u -- caddr' u' name nlen
\ *G From the given string, return the remaing string
\ ** and the attribute name, which is the first whitespace
\ ** delimited token before an '=' character.
  scan-black 2dup 2>r
  [char] = scan dup 2r> rot -
  -white
;

: getAttribValue	\ caddr u -- caddr' u' value vlen
\ *G From the given string, return the remaing string
\ ** and the attribute value string, which is enclosed
\ ** by quotation marks ' or ".
\  cr ." Attributes from:" 2dup type ." :"
  scan-quote
  over c@ -rot			\ preserve separator
  1 /string  over >r		\ -- char caddr u ; R: -- start ; skip separator
  rot scan			\ -- caddr len ; R: -- start
  over >r			\ -- caddr len ; R: -- start end
  1 /string			\ skip separator
  r> r> tuck -
;

: getAttribute	\ caddr u -- caddr' u'
\ *G From the given string extract an attribute name/value pair,
\ ** pass it to the deferred word *\fo{doAttribute} and return
\ ** the remaining string. Attributes are of the form:
\ *C   name = "value"
  dup 0=
  if  exit  endif
  2dup [char] = scan dup 0=		\ form is name="text", check '='
  if  2drop  drop 0  exit  endif
  [char] " scan dup 0=			\ check first '"'
  if  2drop  drop 0  exit  endif
  [char] " scan nip 0=			\ check second '"'
  if  exit  then
  getAttribName 2>r
  getAttribValue
  2r> doAttribute
;

: SetTagStatus	\ --
\ *G Set the tag status for opening/closing/empty, and
\ ** for processing instruction and specials (the !xxx tags).
  TagStatus off				\ assume plain opening tag
  case  CurrName 1+ c@			\ inspect first character
    [char] / of  $0001 TagStatus or!  endof
    [char] ? of  $0100 TagStatus or!  endof
    [char] ! of  $0200 TagStatus or!  endof
  endcase
  tagText stos + 1- c@ [char] / =	\ last char, empty tag?
  if  $0002 TagStatus or!  endif
;

: doTagText	\ caddr u --
\ *G Parse the tag text <text...> excluding the brackets,
\ ** extracting the tag name and the attributes.
  getTagName CurrName place
  begin
    skip-white dup
   while
    over c@ dup [char] > <> swap [char] / <> and
   while
    getAttribute
  repeat then
  2drop
;

: RunInputTag	\ --
\ *G The tag handler action of *\fo{doTags} for active processing
\ ** of XML tags.
  tagText stos doTagText		\ separate the tag text elements
  SetTagStatus
  CurrName count doInputTag
  CurrName count Lastname place
  TagStatus @ LastStatus !
;

: ActiveXML	\ --
\ *G Set the active XML handlers, so that known tags are
\ ** processed.
  assign runInputTag to-do doTags
  assign noop to-do doContents
  assign .Attribute to-do doAttribute
;


\ ==================
\ *H XML Parser core
\ ==================

: AsFarAs          \ char -- flag caddr u
\ *G Parse input stream up to *\i{char}, returning the extracted
\ ** string.
  SOURCE NIP >IN @ - >R
  PARSE
  DUP R> = -rot
; doNotSin

: withText	\ newspad -- oldspad
\ *G Start a new string on the given stackpad for a block of
\ ** processings and make it the current stackpad. Return
\ ** the previous current stackpad
  CurrSpad @ swap  dup CurrSpad ! snew
; doNotSin

: doneText	\ oldspad --
\ *G Discard the current stackpad string and restore
\ ** the previous stackpad.
  CurrSpad @ sdrop  CurrSpad !
; doNotSin

: doXMLblock	\ char --
\ *G Collect input text up to the terminating character
\ ** into the current stackpad, and expand entities.
  RefillStatus @			\ end of file check
  if  drop  exit  endif
  >r
  begin
    r@ AsFarAs dents+			\ scan for char, decode input
   while
    refill 0=				\ get more input
  until
  true RefillStatus !			\ mark done
  then
  r> drop
; doNotSin

: skipPast    \ c-addr u --
\ *G Step through the input stream for a string (not space
\ **  delimited), *\fo{REFILL}ing as necessary until the string
\ ** is found or input is exhausted.
  2>r
  begin					\ -- ; R: -- caddr len
    source >in @ /string 2r@ search 0=
   while				\ -- caddr' len' ; R: -- caddr len
    2drop  refill 0=
  until					\ -- ; R: -- caddr len
\ -- ; refill failed
  1 RefillStatus !  2r> 2drop  exit  then
\ caddr len -- ; string found
  nip					\ -- len ; discard start address
  source nip swap - r> + >in !		\ move >IN past string
  r> drop
; doNotSin

: doTagBlock	\ x --
\ *G Process a tag block "<name ... >" starting immediately after
\ ** the leading '<' character. The tag text is discarded after
\ ** the tag has been processed. If *\b{x} is non-zero, the tag
\ ** is initialised to "?xml"
  source >in @ /string 3 >= if		\ check for comment tag
    s" !--" s= if
      s" -->" skipPast  drop  exit
    endif
  else
    drop
  endif
  tagText withText >r			\ select stackpad
  if  s" ?xml " currSpad @ sappend  endif
  [char] > doXMLblock doTags
  r> doneText				\ discard stackpad restore old
; doNotSin

: doContentBlock	\ --
\ *G Process a content block up to but not including
\ ** the trailing '<' character.
  contents currSpad !			\ select stackpad
  [char] < doXMLblock doContents
; doNotSin

: ReadXML	\ --
\ *G Read XML from the current input stream.
  0 RefillStatus !
  1 doTagBlock				\ Skip the rest of the <?XML tag
  contents withText >r
\ Process contents and tags until end of input.
  begin
    RefillStatus @ 0=
   while
    doContentBlock  0 doTagBlock
  repeat
  r> doneText
; doNotSin

: <?xml		\ --
\ *G After *\fo{<?xml} has been executed, all further input is
\ ** treated as XML source and handled by the XML parser.
  RefillStatus @ >r
  ['] ReadXML catch
  r> RefillStatus !
  throw
; doNotSin


\ ********************************
\ *N Data content input and output
\ ********************************
\ *P These words are factors that can be used when constructing
\ ** systems that extract and produce data in XML files.
\ ** When producing an XML file, data is output by primitives
\ ** that take the address of the data.
\ ** When reading an XML file, data is set by primitives
\ ** that take a string and the address of the data.

\ ==================
\ *H XML text output
\ ==================
\ *P XML text output of tag or content data must not contain
\ ** the special characters which must be converted to the
\ ** standard entity format "&xxx;".

: XMLemit	\ char --
\ *G Output a character translating the special characters.
  case
    [char] < of  ." &lt;"  endof
    [char] > of  ." &gt;"  endof
    [char] & of  ." &amp;"  endof
    [char] ' of  ." &apos;"  endof
    [char] " of  ." &quot;"  endof
      emit
  end-case
;

: XMLtype	\ caddr len --
\ *G Output a string translating the special characters.
  bounds ?do  i c@ XMLemit  loop
; doNotSin


\ =============================
\ *H Single and double integers
\ =============================

: ud#>cl	\ ud -- caddr len
\ *G Convert an unsigned double to a decimal text string.
  base @ >r  decimal
  <# #s #>
  r> base !
; doNotSin

: d#>cl	\ ud -- caddr len
\ *G Convert a signed double to a decimal text string.
  base @ >r  decimal
  tuck dabs <# #s rot sign #>
  r> base !
; doNotSin

: cl>d#		\ caddr len -- d
\ *G Convert the string to a double number.
  base @ >r decimal
  -leading -trailing skip-sign >r
  0 0 2swap >number 2drop
  r> ?dnegate
  r> base !
; doNotSin

: cl>ud#	\ caddr len -- ud
\ *G Convert the string to an unsigned double number.
  base @ >r decimal
  -leading -trailing
  0 0 2swap >number 2drop
  r> base !
; doNotSin

: ?i		\ addr --
\ *G Display the contents of a signed 32 bit integer.
  @ s>d d#>cl XMLtype
;

: !i		\ caddr len dest --
\ *G Set the contents of a signed 32 bit integer.
  -rot  cl>d# drop  swap !
;

: ?ui		\ addr --
\ *G Display the contents of an unsigned 32 bit integer.
  @ 0 ud#>cl XMLtype
;

: !ui		\ caddr len dest --
\ *G Set the contents of an unsigned 32 bit integer.
  -rot  cl>ud#  rot !
;

: ?d		\ addr --
\ *G Display the contents of a signed 64 bit integer in Forth format
\ ** (high cell at low address).
  2@ d#>cl XMLtype
;

: !d		\ caddr len dest --
\ *G Set the contents of a signed 64 bit integer in Forth format
\ ** (high cell at low address).
  -rot  cl>d#  rot 2!
;

: ?ud		\ addr --
\ *G Display the contents of an unsigned 64 bit integer in Forth format
\ ** (high cell at low address).
  2@ ud#>cl XMLtype
;

: !ud		\ caddr len dest --
\ *G Set the contents of an unsigned 64 bit integer in Forth format
\ ** (high cell at low address).
  -rot  cl>ud#  rot 2!
;

: ?dI		\ addr --
\ *G Display the contents of a signed 64 bit integer in Intel format
\ ** (low cell at low address).
  2@ swap d#>cl XMLtype
;

: !dI		\ caddr len dest --
\ *G Set the contents of a signed 64 bit integer in Intel format
\ ** (low cell at low address).
  -rot  cl>d# swap  rot 2!
;

: ?udI		\ addr --
\ *G Display the contents of an unsigned 64 bit integer in Intel format
\ ** (low cell at low address).
  2@ swap ud#>cl XMLtype
;

: !udI		\ caddr len dest --
\ *G Set the contents of a signed 64 bit integer in Intel format
\ ** (low cell at low address).
  -rot  cl>ud# swap  rot 2!
;


\ =========================
\ *H Floating point numbers
\ =========================

[defined] sf@ [if]

: cl>f#		\ caddr u -- ; F: -- f
\ *G Convert a string to a floating point number. If a conversion
\ ** fault occurs, *\i{f} is set to zero.
  >float 0=
  if  %0  endif
;

: ?fs		\ addr --
\ *G Display the contents of 32 bit float.
  sf@ f.
;

: !fs		\ caddr u dest --
\ *G Set the contents of a 32 bit float.
  >r  cl>f#  r> sf!
;

: ?fd		\ addr --
\ *G Display the contents of 64 bit float.
  df@ f.
;

: !fd		\ caddr u dest --
\ *G Set the contents of a 64 bit float.
  >r  cl>f#  r> df!
;

: ?ft		\ addr --
\ *G Display the contents of an 80 bit float.
  tf@ f.
;

: !ft		\ caddr u dest --
\ *G Set the contents of an 80 bit float.
  >r  cl>f#  r> tf!
;
[then]


\ ==========
\ *H Strings
\ ==========

: .string	\ caddr len --
\ *G Output the given string in XML format.
  XMLtype
;

: ?cstring	\ caddr --
\ *G Output a Forth counted string.
  count .string
;

: !cstring	\ caddr len dest --
\ *G Set a Forth counted string.
  place
;

: ?wstring	\ caddr --
\ *G Output a word (16 bits) counted string
  dup w@  swap 2 + swap .string
;

: !wstring	\ caddr len dest --
\ *G Set a word (16 bits) counted string
  2dup 2>r			\ write count last
  2+ swap movex
  2r> w!			\ to avoid in-place problems
;

: ?lstring	\ caddr --
\ *G Output a cell (32 bits) counted string
  dup @  swap cell + swap .string
;

: !lstring	\ caddr len dest --
\ *G Set a cell (32 bits) counted string
  2dup 2>r			\ write count last
  cell+ swap movex
  2r> !				\ to avoid in-place problems
;


\ ================
\ *H Time and date
\ ================

: .xuw		\ u w --
\ *G Display the unsigned number u as w digits.
  swap 0 <# rot 0 ?do # loop #> XMLtype
; doNotSin

: .xdate	\ day month year --
\ *G Output a date in XML format "CCYY-MM-DD".
  4 .xuw ." -"  2 .xuw ." -" 2 .xuw
;

: .xtime	\ secs mins hours --
\ *G Output a time in XML format "HH-MM-SS".
  2 .xuw ." :"  2 .xuw ." :" 2 .xuw
;

: .xdateTime	\ secs mins hours day month year --
\ *G Output a date/time in XML format. No time zone is output.
  .xdate ." T" .xtime
;

: .tz		\ mins --
\ *G Output a time zone indicator as an offset from UTC
\ ** in minutes.
  dup if
    dup 0<
    if  ." -" abs  else  ." +"  endif
    0 <# # # [char] : hold # # #> XMLtype
  else
    drop  ." Z"
  endif
;

: xdt-utc	\  secs mins hours day month year --
\ *G Output a date/time in XML format. UTC is indicated.
  .xdateTime ." Z"
;

: xdt-zone	\  secs mins hours day month year zmins --
\ *G Output a date/time in XML format. The time zone is indicated
\ **  by a signed offset in minutes.
  >r  .xdateTime  r> .tz
;


\ =============
\ *H Tag output
\ =============

: .GenTag	\ caddr len --
\ *G Display the text as a tag "<...>".
\ ** Standard entities are encoded.
  ." <" XMLtype ." >"
;

: .GenTag+	\ attr alen name nlen --
\ *G Display attribute and tag name text as "<name attr>".
\ ** Standard entities are encoded.
  ." <" XMLtype space XMLtype ." >"
;

: .ClosingTag	\ caddr len --
\ *G Display the text as a closing tag "</...>".
\ ** Standard entities are encoded.
  ." </" XMLtype ." >"
;

: .EmptyTag	\ caddr len --
\ *G Display the text as an empty tag "<.../>".
\ ** Standard entities are encoded.
  ." <" XMLtype ." />"
;


\ ============
\ *N Test code
\ ============
\ *[
initSpads
ActiveXML
\ *]


\ ======
\ *> ###
\ ======

decimal
