\ VFX Forth OpenGL 32 bit floating point support

((
Copyright:
----------
Copyright (c) 2000-2002, 2004, 2006, 2007, 2014, 2018 to:
MicroProcessor Engineering
133 Hill lane
Southampton SO15 5AF
England
tel: +44 (0)23 8063 1441
net: tech-support@mpeforth.com
web: http://www.mpeforth.com

N.B. The bug in the FSUB/P and FSUBR/P assembler instructions has
     been fixed as of build 0785. Use them with care!

To do
=====
Remove all uses of PAD.

Change history
==============
20180628 SFP013 Updated FNUMBER? for VFX v5
20141023 GEK012 Fixed FTANH.
20070103 EDA011 Updated REPRESENT.
		Added ANS float environment variables as they were missing.
		Corrected NAN? and INFINITY?.
20060830 EDA010 Updated REPRESENT.
20040707 SFP009 Restored checking to F**
20021023 SFP008 Added FR>S and FR>D after much discussion on comp.lang.forth
20021007 SFP007 Corrected F>S and F>D to ANS spec.
20020320 SFP006 Corrected F>ASCII and >FLOAT.
20020216 SFP005 More rounding corrections.
		Corrected REPRESENT and reduced >BCD.
		Corrected exponent handling in numeric output.
20010313 SFP004 Added F# for PFW2.x compatibility
20010102 SFP003 Improved rounding in >BCD.
		Added REALS and INTEGERS for user control of number input.
20000927 SFP002	For performance, removed error checking from F** as this
		is consistent with the rest of the package.
20000919 SFP001	First release
))

( *! HfpGL32     )
( *T Open GL 32 bit Floating Point )

( *S Introduction )
( *P The package uses the FPU instruction set. The source code can be found )
( ** in the file LIB\HFPGL32.FTH. This package is specifically designed for )
( ** use with OpenGL. Unlike the other MPE floating point packs, HfpGL32 )
( ** keeps floating point numbers as 32 bit IEEE floats on the Forth data )
( ** stack. This is referred to in the Forth literature as a COMBINED stack )
( ** implementation. In normal operation, HfpGL32 floats are single cell )
( ** values on the data stack. The code here is a subset of the ANS Forth )
( ** floating point wordsets. )

( *P Similarly, HfpGL32.FTH defines literal and default floats using )
( ** F@ F! and friends to be in 32 bit format. )

( *P Because of the use of 32 bit floats throughout, number conversion, )
( ** especially on input, can have roundoff errors. For example: )
( *E   1000e-4 f. gives 9.9999994e-2 )
( **   1e-1 f.    gives 1.0000000e-1 )

( *P Please note that the trig functions are calculated in )
( ** radians, for calculations in degrees use DEG>RAD beforehand, )
( ** RAD>DEG afterwards. )

( *S Number formats and ANS Forth )
( *P The ANS Forth standard specifies that floating point numbers must )
( ** be entered in the form 1.234e5 and must contain a point '.' and )
( ** 'e' or 'E', and that double integers are terminated by a point '.'. )

( *P This situation prevents the use of the standard conversion words )
( ** in international applications because of the interchangable use of )
( ** the '.' and ',' characters in numbers. Because of this, VFX Forth )
( ** uses two system variables, FP-CHAR and DP-CHAR, to hold the character )
( ** used as the floating point and double integer indicator characters. )
( ** By default, FP-CHAR is initialised to '.' and DP-CHAR is initialised )
( ** to ',', as has been MPE practice for many years. For ANS compliance, )
( ** you should set them as follows: )

( *E \ ANS standard setting     )
( **   char . dp-char !		)
( **   char . fp-char !	        )
( ** : ans-floats	\ -- ; for strict ANS compliance        )
( **   [char] . dp-char !	)
( **   [char] . fp-char !	)
( ** ;				)
( ** \ MPE defaults	        )
( **   char , dp-char !		)
( **   char . fp-char !		)
( ** : mpe-floats	\ -- ; for existing and legacy code      )
( **   [char] , dp-char !	)
( **   [char] . fp-char !	)
( ** ;				)

( *P You can of course set these variables to any value which suits you )
( ** for your language and locale. Note that integer conversion is always )
( ** attempted before floating point conversion. This means that if FP-CHAR )
( ** and DP-CHAR are the same, floating point numbers must contain 'e' or )
( ** 'E'. If they are different, a number containg the FP-CHAR will be )
( ** successfully converted as a floating point number, even if it does )
( ** not contain 'e' or 'E'. )

( *S Floating point exceptions )
( *P Exception handling is determined by the operating system. On current )
( ** Windows platforms, floating point exceptions are not generated by )
( ** The NDP. This can be changed by altering the bottom 6 bits of the )
( ** NDP control word using CW@ and CW!. )

( *P By default, the system prompt will report exception status, and clear )
( ** the pending exception status. Exception status reporting does not mean that )
( ** the Windows exception handler has been triggered, it only means that )
( ** the status flag has been set. )

\ *S Standards compliance and F>D.
\ *P After much discussion on the comp.lang.forth newsgroup, a consensus
\ ** was reached that F>D and F>S must truncate to zero. This is also
\ ** the behaviour required by the Forth Scientific Library (FSL). Historically,
\ ** MPE floating point packs permit the integer rounding mode to be set
\ ** by the user. In order to support both camps, VFX Forth now behaves as
\ ** follows:
\ *B F>D and F>S truncate to zero,
\ *B FR>D and FR>S follow the current rounding mode.

( *S Glossary )


\ *******************************
\ Floating point type definitions
\ *******************************

[undefined] FpSystem [if]
0 value FpSystem
( *G The value FPSYSTEM defines which floating point pack is installed and )
( ** active. Each floating point pack defines its own type as follows: )
\ *(
( *B 0 constant NoFPSystem )
( *B 1 constant HFP387System )
( *B 2 constant NDP387System )
( *B 3 constant OpenGL32System )
\ *)
[then]

\ 0 constant NoFPSystem
\ 1 constant HFP387System
\ 2 constant NDP387System
  3 constant OpenGL32System
OpenGL32System -> FpSystem

\ Define size of items on the floating point stack, 4, 8, or 10 bytes
#4 constant FPCELL	\ -- n
( *G Defines the size of literals and floating point numbers in memory )
( ** and on the data stack. Always 4 for HfpGL32. )
FPCELL negate constant -FPCELL

7 constant #fdigits	\ -- u
( *G Returns the largest number of usable digits available from )
\ ** *\fo{REPRESENT}. Equivalent to environment variable *\fo{MAX-FLOAT-DIGITS}.

\ ****************************
\ Floating point debug support
\ ****************************

false constant [fpdebug] immediate
( *G Set this CONSTANT true when compiling HfpGL32.FTH, and a debug )
( ** build will be constructed. In this state, the state of the )
( ** FPU is checked after each word. If a floating point exception )
( ** has been generated, a diagnostic is issued, and the system )
( ** aborts. Set this only when testing. )

[fpdebug] [if]
  defer fpcheck				\ see later for real action
  assign noop to-do fpcheck
[then]


\ *********************************
\ Variable and frames for FP stacks
\ *********************************

\ memory copies of the staus word, control word, and rounding control word
\ are held on a stack frame built as an offset on the return stack pointed
\ to by BP

0 constant #sw          \ offset of status word
2 constant #cw          \ offset of control word
4 constant #rc          \ offset of rounding control word
6 constant #spare       \ dummy to bring up to 8 bytes

8 constant #fpframe     \ size of frame required for all four items


\ ****************
\ Assembler macros
\ ****************

also asm-access definitions

fpcell 4 = [if]
: fword		\ --
( *G Selects appropriate floating point size for the assembler. )
( ** Note that this is defined by the constant FPCELL. )
( ** FWORD will be a synonym for FLOAT, DOUBLE or TBYTE. )
  float  ;
[then]

: >FPU       \ f -- ; push TOS on data stack to FPU
  mov   -4 [ebp], ebx		\ get TOS into memory
  fld   fword -4 [ebp]		\ push TOS to FPU
  mov   ebx, 0 [ebp]		\ restore TOS
  lea	ebp, 4 [ebp]
;

: FPU>        \ -- f ; pop FTOS on FPU to data stack
  lea	ebp, -4 [ebp]		\ make space
  fstp  fword 0 [ebp]           \ store and pop
  mov	eax, 0 [ebp]		\ exchange old TOS and new
  mov   0 [ebp], ebx
  mov	ebx, eax
;

: fnext,	\ -- ; can be changed for debugging
[fpdebug] [if]
  fwait				\ forces last instruction flush
  call 	['] fpcheck
[then]
  ret
;

previous definitions


\ ************
\ FP constants
\ ************

        0 constant %0		\ -- f%0
$3F800000 constant %1		\ -- f%1
$40490FDB constant %pi		\ -- f%pi
$3FC90FDB constant %pi/2	\ -- f%pi/2
$3F490FDB constant %pi/4	\ -- f%pi/4
$3C8EFA35 constant %pi/180	\ -- f%pi/180
$42652EE0 constant %180/pi	\ -- f%180/pi


\ *********************
\ FP control operations
\ *********************

code finit	\ -- ; resets FPU
  finit
  next,
end-code  inline

finit

code cw@	\ -- cw ; get NDP control word
  lea	ebp, -8 [ebp]		\ save TOS, reserve one more cell
  mov	4 [ebp], ebx
  fstcw 0 [ebp]			\ get control word to memory
  sub	ebx, ebx
  mov	bx, 00 [ebp]		\ and then to TOS
  lea	ebp, 4 [ebp]		\ release spare cell
  fnext,
end-code

code cw!	\ cw -- ; set NDP control word
  lea   ebp, -4 [ebp]		\ must load CW to memory
  mov	0 [ebp], ebx		\ new CW is in TOS
  fldcw 0 [ebp]
  mov	ebx, 4 [ebp]		\ restore TOS
  lea	ebp, 8 [ebp]
  fnext,
end-code

code sw@	\ -- sw ; get NDP status word
  lea	ebp, -4 [ebp]		\ save TOS, reserve one more cell
  mov	0 [ebp], ebx
  fstsw ax			\ get status word
  sub 	ebx, ebx		\ clear result
  mov	bx, ax
  next,				\ so that debugger isn't triggered repeatedly
end-code

code fclex	\ -- ; clear exceptions
  fclex				\ clear exceptions
  next,
end-code


\ *******************
\ FP Stack operations
\ *******************

: fdup		\ f -- f f
( *G Floating point equivalent of DUP )
  dup
;

: fswap		\ f1 f2 -- f2 f1
( *G Floating point equivalent of SWAP )
  swap
;

: fdrop		\ f --
( *G Floating point equivalent of DROP )
  drop
;

: fover		\ f1 f2 -- f1 f2 f1
( *G Floating point equivalent of OVER )
  over
;

: frot		\ f1 f2 f3 -- f2 f3 f1
( *G Floating point equivalent of ROT )
  rot
;

: fpick		\ x0..xn n -- x0..xn f
( *G Floating point equivalent of PICK. )
  pick
;

((
: froll		\ fn..f0 n -- fn-1..f0 fn
  roll
;
))

code ndepth     \ -- n ; depth of NDP stack
( *G Returns on the Forth data stack the number of items on the )
( ** FPUs's internal working stack. )
  fstsw ax                              \ NDP status
  sub   ebp, # 4                        \ preserve TOS
  mov   0 [ebp], ebx
  shr   ax, # #11                       \ line up the bits
  and   al, # $07                       \ only want bottom three
  mov   ebx, # 8                        \ 8 registers on NDP,
  sub   bl, al                          \ convert to depth
  cmp   bl, # 8                         \ convert 8 to 0
  z, if,  sub bl, bl  endif,
  fnext,
end-code


\ *******************************
\ Memory operations SF@ SF! F@ F!
\ *******************************

: sf@		\ addr -- f
( *G Fetches the contents of addr. )
  @
;

: sf!		\ f addr --
( *G Stores f at addr. )
  !
;

: f@		\ addr -- f
( *G Fetches the contents of addr. )
  @
;

: f!		\ f addr --
( *G Stores f at addr. )
  !
;

code f+!	\ f addr -- ; add f to data at addr
( *G Add F to the data at ADDR. )
  fld   fword 0 [ebp]
  fld   fword 0 [ebx]		\ get from addr
  faddp st(1), st               \ do the actual addition
  fstp  fword 0 [ebx]		\ put back at addr
  mov   ebx, 4 [ebp]		\ restore TOS
  lea	ebp, 8 [ebp]
  fnext,
end-code

code f-!	\ f addr -- ; sub f from data at addr
( *G Subtract F from the data at ADDR. )
  fld   fword 0 [ebp]
  fld   fword 0 [ebx]		\ get from addr
  fsubrp st(1), st              \ do the actual addition
  fstp  fword 0 [ebx]		\ put back at addr
  mov   ebx, 4 [ebp]		\ restore TOS
  lea	ebp, 8 [ebp]
  fnext,
end-code

: f,            \ f --
( *G lays a default float into the dictionary, reserving )
( ** FPCELL bytes )
  here  FPCELL allot  f!  ;

: falign        \ --
( *G Aligns the dictionary to accept a default float. )
  align
;

: faligned      \ addr -- addr'
( *G Aligns the address to accept a default float. )
  aligned
;

: float+        \ addr -- addr'
( *G Increments addr by FPCELL, the size of a default float. )
  FPCELL +  ;

: floats        \ n1 -- n2
( *G Returns n2, the size of n1 default floats. )
  FPCELL *
;


\ *****************
\ FP defining words
\ *****************

: fvariable     \ -- ; -- addr
( *G Use in the form: FVARIABLE <name> to create a variable )
( ** that will hold a default floating point number. )
  variable
;

: farray	\ n -- ; i -- addr
( *G Use in the form: n FARRAY <name> to create a variable    )
( ** that will hold a default floating point number. When the )
( ** array name is executed, the index i is used to retun the )
( ** address of the i'th 0 zero-based element in the array.   )
( ** For example, 5 FARRAY TEST will set up 5 array elements  )
( ** each containing 0, and then f n TEST F! will store f in  )
( ** the nth element, and n TEST F@ will fetch it.            )
  create
    0 max  floats allot&erase
  does>
    swap FPCELL * +                     \ make address
;

: fconstant     \ f -- ; -- f
( *G Use in the form: <float> FCONSTANT <name> to create a constant )
( ** that will return a floating point number. )
  constant
;

[-short-branches
: fvalue	\ f -- ; ??? -- ???
( *G Use in the form: <float> FVALUE <name> to create a floating )
( ** point version of VALUE that will return a floating point number )
( ** by default, and that can accept the operators TO, ADDR, )
( ** ADD, SUB, and SIZEOF. )
  create  immediate
    f,
  does>
    state @ if
      [compile] literal
      case  OperatorType @off
        0 of  postpone f@  endof	\ Default, fetch
        1 of  postpone f!  endof	\ store, ->, to
        2 of  ( noop )  endof		\ addr
\					\ inc
\					\ dec
        5 of  postpone f+!  endof	\ add
\					\ zero
        7 of  postpone f-!  endof	\ sub
        8 of  postpone drop  postpone FPCELL  endof	\ sizeof
\					\ set
             bad-method
      endcase
    else
      case  OperatorType @off
        0 of  f@  endof			\ default, fetch
        1 of  f!  endof			\ store, ->, to
        2 of  ( noop )  endof		\ addr
\        3 of  incr  endof		\ inc
\        4 of  decr  endof		\ dec
        5 of  f+!  endof		\ add
\        6 of  off  endof		\ zero
        7 of  f-!  endof		\ sub
        8 of  drop FPCELL  endof	\ sizeof
\        9 of  on  endof		\ set
             bad-method
      endcase
    endif
;
short-branches]


\ *******************************************
\ FP four basic functions + - * /  and others
\ *******************************************

code f+         \ f1 f2 -- f1+f2
( *G Floating point add. )
  >FPU >FPU
  faddp st(1), st               \ do the actual addition
  FPU>
  fnext,
end-code inline

code f-         \ f1 f2 -- f1-f2
( *G Floating point subtract. )
  >FPU >FPU
  fsubrp st(1), st              \ do the actual calc.
  FPU>
  fnext,
end-code inline

code f*         \ f1 f2 -- f1*f2
( *G Floating point multiply. )
  >FPU >FPU
  fmulp st(1), st               \ do the actual calc.
  FPU>
  fnext,
end-code inline

code f/         \ f1 f2 -- f1/f2
( *G Floating point divide. )
  >FPU >FPU
  fdivrp st(1), st		\ do the actual calc.
  FPU>
  fnext,
end-code inline

code fmod       \ f1 f2 -- f3
( *G Floating point modulus. Returns f3 the remainder after repeatedly )
( ** subtracting f2 from f1. Often used to force arguments to lie in the )
( ** range: 0 <= arg < f2 )
  >FPU >FPU
  fprem                         \ take modulus, rem in ST, f2 in ST(1)
  fxch
  fstp  st                      \ discard f2
  FPU>
  fnext,
end-code inline

code fsqrt      \ f -- sqrt(f)
( *G Floating point square root. )
  >FPU
  fsqrt
  FPU>
  fnext,
end-code inline

code 1/f        \ f -- 1/f
( *G Floating point reciprocal. )
  >FPU
  fld1                          \ push one on ap stack
  fdivrp st(1), st              \ do the actual division
  FPU>
  fnext,
end-code inline

code fabs       \ f -- |f|
( *G Floating point absolute. )
  >FPU
  fabs
  FPU>
  fnext,
end-code inline

code fnegate    \ f -- -f
( *G Floating point negate. )
  >FPU
  fchs
  FPU>
  fnext,
end-code inline

code f2*        \ f -- f*2
( *G Floating point multiply by two. )
  >FPU
  fld1
  fadd  st, st
  fmulp st(1), st
  FPU>
  fnext,
end-code inline

code f2/        \ f -- f/2
( *G Floating point divide by two. )
  >FPU
  fld1
  fadd  st, st
  fxch
  fdivrp st(1), st
  FPU>
  fnext,
end-code inline


\ ************************
\ Integer to FP conversion
\ ************************

code s>f        \ n -- f
( *G Converts a single integer to a float. )
  mov	-4 [ebp], ebx		\ put TOS in deep stack
  fild  dword -4 [ebp]		\ load 32 bit integer to FPU
  fstp  fword -4 [ebp]		\ store 32 bit float from FPU
  mov	ebx, -4 [ebp]		\ put back in TOS
  fnext,
end-code

code d>f        \ d -- f
( *G Converts a double integer to a float. )
\ The double number needs to be in memory in strict Intel order
  mov   eax, 0 [ebp]    	\ get low portion of double int, now in EBX:EAX
  sub   ebp, # 4        	\ make space
  mov   0 [ebp], eax    	\ stash double
  mov   4 [ebp], ebx
  fild  qword 0 [ebp]   	\ load 64 bit integer to FPU
  fstp  fword 0 [ebp]
  mov   ebx, 0 [ebp]    	\ restore TOS
  lea	ebp, 8 [ebp]		\ discard double and extras
  fnext,
end-code

\ fixed SFP007
code f>s        \ f -- n ; convert float to integer
\ *G Converts a float to a single integer.
\ ** Note that F>S truncates the number towards zero
\ ** according to the ANS specification.
  fstcw word -2 [ebp]		\ save/preserve CW
  fstcw word -4 [ebp]
  or    word -2 [ebp], # $0C00	\ force truncation
  fldcw word -2 [ebp]
  mov	-8 [ebp], ebx		\ float must be in memory
  fld	fword -8 [ebp]
  fistp dword -8 [ebp]
  fldcw word -4 [ebp]
  mov	ebx, -8 [ebp]
  fnext,
end-code

\ fixed SFP007
code f>d        \ f -- d ; convert float to double integer
\ *G Converts a float to a double integer.
\ ** Note that F>D truncates the number towards zero
\ ** according to the ANS specification.
  fstcw word -2 [ebp]		\ save/preserve CW
  fstcw word -4 [ebp]
  or    word -2 [ebp], # $0C00	\ force truncation
  fldcw word -2 [ebp]
  mov	-8 [ebp], ebx		\ put f in memory
  fld   fword -8 [ebp]		\ load f into FPU
  fistp qword #-12 [ebp]	\ store integer from FPU
  fldcw word -4 [ebp]
  mov   ebx, -8 [ebp]		\ get dh as new TOS
  mov   eax, #-12 [ebp]		\ get dl to be new NOS
  mov   -4 [ebp], eax
  sub	ebp, # 4
  fnext,
end-code

\ SFP008...
code fr>s	\ f -- n ; convert float to integer
\ *G Converts a float to a single integer using the current
\ ** rounding mode.
  mov	-4 [ebp], ebx		\ float must be in memory
  fld	fword -4 [ebp]
  fistp dword -4 [ebp]
  mov	ebx, -4 [ebp]
  fnext,
end-code

code fr>d	\ F: f -- d ; convert float to double integer
\ *G Converts a float to a double integer using the current rounding mode.
  mov	-4 [ebp], ebx		\ put f in memory
  fld   fword -4 [ebp]		\ load f into FPU
  fistp qword -8 [ebp]		\ store integer from FPU
  mov   ebx, -4 [ebp]		\ get dh as new TOS
  mov   eax, -8 [ebp]		\ get dl to be new NOS
  mov   -4 [ebp], eax
  sub	ebp, # 4
  fnext,
end-code
\ ...SFP008


\ **************
\ FP comparisons
\ **************

code fcmp1      \ f -- status ; test the fp#, returning status result
  mov   -4 [ebp], ebx
  fld	fword -4 [ebp]
  ftst				\ test FTOS
  fstsw ax                      \ save status result
  fstp  st                      \ drop FPUstack
  and   ax, # $4100             \ mask out status bits
  movzx ebx, ax                 \ copy status to TOS
  fnext,
end-code

code fcmp2      \ f1 f2 -- status ; test the top 2 floats, returning status
  mov	-4 [ebp], ebx		\ put f2 into memory
  fld	fword -4 [ebp]		\ f2
  fld   fword 0 [ebp]		\ f1
  fcompp                        \ compare and discard
  fstsw ax                      \ save status result
  and   ax, # $4100             \ mask out status bits
  movzx ebx, ax                 \ copy status to TOS
  add	ebp, # 4
  fnext,
end-code

: f0<           \ f1 -- t/f ; less than zero?
( *G Floating point 0<. N.B. result is on the Forth integer data stack. )
  fcmp1 $0100 =
;

: f0=           \ f1 -- t/f ; equal zero?
( *G Floating point 0=. N.B. result is on the Forth integer data stack. )
  fcmp1 $4000 =
;

: f0<>          \ f1 -- t/f ; not equal zero?
( *G Floating point 0<>. N.B. result is on the Forth integer data stack. )
  fcmp1 $4000 <>
;

: f0>           \ f1 -- t/f ; greater than zero?
( *G Floating point 0>. N.B. result is on the Forth integer data stack. )
  fcmp1 0=
;

: f<            \ f1 f2 -- t/f ; one less than the other?
( *G Floating point <. N.B. result is on the Forth integer data stack. )
  fcmp2 $0100 =
;

: f=            \ f1 f2 -- t/f ; equal each other?
( *G Floating point =. N.B. result is on the Forth integer data stack. )
  fcmp2 $4000 =
;

: f<>           \ f1 f2 -- t/f ; one not equal the other?
( *G Floating point <>. N.B. result is on the Forth integer data stack. )
  fcmp2 $4000 <>
;

: f>            \ f1 f2 -- t/f ; one less than the other?
( *G Floating point >. N.B. result is on the Forth integer data stack. )
  fcmp2 0=
;

: f<=           \ f1 f2 -- t/f ; one less or equal the other?
( *G Floating point <=. N.B. result is on the Forth integer data stack. )
  fcmp2
  dup $0100 =  swap $4000 = or
  0<>
;

: f>=           \ f1 f2 -- t/f ; one greater or equal the other?
( *G Floating point >=. N.B. result is on the Forth integer data stack. )
  fcmp2
  dup 0=  swap $4000 = or
  0<>
;

[-short-branches
: f~            \ f1 f2 f3 -- flag
( *G Approximation function. If f3 is positive, flag is true if abs[f1-f2] )
( ** is less than f3. If f3 is zero, flag is true if the f2 is exactly equal )
( ** to f1. If f3 is negative, flag is true if abs[f1-f2] less than )
( ** abs[f3*abs[f1+f2]]. )
  fdup f0> if                           \ if f3 > 0
    frot frot f- fabs			\ -- f3 |f1-f2|
    f>                                  \ -- t/f
  else
    fdup f0= if                         \ if f3 = 0
      fdrop  f=				\ f1=f2?
    else                        	\ if f3 < 0
      frot frot  fover fover		\ -- f3 f1 f2 f1 f2
      f- fabs                           \ -- f3 f1 f2 |f1-f2|
      frot frot                         \ -- f3 |f1-f2| f1 f2
      fabs fswap fabs                   \ -- f3 |f1-f2| |f1| |f2|
      f+ frot fabs f*                   \ -- |f1-f2| |f3|*(|f1|+|f2|)
      f<                                \ -- t/f
    endif
  endif
;
short-branches]


\ ******************************
\ Words dependent on FP compares
\ ******************************

: ?fnegate      \ f1 f2 -- f3
( *G Floating point NEGATE. )
  f0<
  if  fnegate  endif
;

: fmax          \ f1 f2 -- f3
( *G Floating point MAX. )
  fover fover f<
  if  fswap  endif
  fdrop
;

: fmin          \ f1 f2 -- f3
( *G Floating point MIN. )
  fover fover f>
  if  fswap  endif
  fdrop
;


\ ******************
\ FP logs and powers
\ ******************

code flog       \ f -- log(f)
( *G Floating point log base 10. )
  >FPU
  fldlg2
  fxch
  fyl2x
  FPU>
  fnext,
end-code

code fln        \ f -- ln(f)
( *G Floating point log base e. )
  >FPU
  fldln2
  fxch
  fyl2x
  FPU>
  fnext,
end-code

decimal
Lbl: 2**z       \ N: f -- 2^f
		\ This routine has to cope with range limitations
  push	edx
  sub	ebp, # #fpframe			\ make transient frame

\ make F absolute, and preserve previous sign
  sub	edx, edx			\ 0 for positive, +ve assumed
  ftst					\ test it
  fstsw	ax				\ save status, -ve if C3 C2 C0 = 001
  and 	ax, # $4100			\ extract C3,C2,C0
  cmp	ax, # $0100			\ -ve?
  z, if,
    dec   edx			        \ mark it so
    fabs				\ make absolute
  endif,

\ round FTOS down, preserving original FTOS
  fstcw word #cw [ebp]                  \ for later use
  fstcw word #rc [ebp]
  and   word #rc [ebp], # $f3ff         \ change the current control; was f3ff
  or    word #rc [ebp], # $0400         \ word to round down
  fldcw word #rc [ebp]
  fld   st                              \ copy number, and round one
  frndint                               \ N: -- f int
  fldcw word #cw [ebp]                  \ restore original control word

\ generate fraction from F and INT
  fxch					\ N: -- int f
  fld	 st(1)				\ N: -- int f int
  fsubrp st(1), st			\ N: -- int frac
  fchs					\ *** BUG in assembler FSUB, FSUBRP

\ generate 2^FRAC, preserving INT
  fld1
  fchs
  fxch					\ N: -- int -1 frac
  fscale				\ N: -- int -1 frac/2
  f2xm1					\ N: -- int -1 2^(frac/2)-1
  fxch
  fchs
  faddp st(1), st			\ N: -- int 2^(frac/2)
  fmul	st, st				\ N: -- int 2^frac

\ check sign of original F, applying as needed
  or	edx, edx
  nz, if,
    fld1
    fdivrp st(1), st			\ N: -- int 2^-frac
    fxch
    fchs
    fxch				\ N: -- -int 2^-frac
  endif,				\ N: -- int' 2^frac'

\ Apply the integer power to get 2^x, and clean up
  fscale				\ N: -- int' 2^(frac'+int')
  fxch
  fstp st				\ N: -- 2^(frac'+int'

  add   ebp, # #fpframe			\ remove transient frame
  pop	edx
  ret
end-code

code fexp       \ f -- e^f ; was called FE^X
( *G Floating point e^f. )
  >FPU
  fldl2e
  fmulp st(1), st
  call  2**z
  FPU>
  fnext,
end-code

: fexpm1	\ F: f -- (e^f)-1 ; 12.6.2.1516
\ *G Floating point log base (e^f)-1.
  fexp %1 f-
;

code falog      \ f -- 10^f ; was called f10^f, new name: ans
( *G Floating point anti-log base 10. )
  >FPU
  fldl2t
  fmulp st(1), st
  call  2**z
  FPU>
  fnext,
end-code

code (f**)	\ f1 f2 -- f1^f2 ; was called fx^y
\ *G Floating point returns f1 raised to the power f2.
\ ** No error checking is performed.
  >FPU >FPU
  fyl2x
  call  2**z
  FPU>
  fnext,
end-code

: f**		\ f1 f2 -- f1^f2 ; was called fx^y
\ *G Floating point: returns f1 raised to the power f2.
\ ** If f1<=0e0, 0e0 is returned.
  fover f0> if
    (f**)
  else
    fdrop fdrop %0
  endif
;

((
code f**	\ F: f1 f2 -- f1^f2 ; SFP002
\ ( *G Floating point returns f1 raised to the power f2. N.B. no error checking )
\ ( ** is performed in this version. If floating point execeptions are masked,  )
\ ( ** which is the default condition, the system will return a NaN for f1<0.   )
  fxch
  fyl2x
  call  2**z
  fnext,
end-code
))

code 2**	\ f -- 2^f
( *G Floating point: returns 2^F. )
  >FPU
  call  2**z
  FPU>
  fnext,
end-code


\ ********
\ Rounding
\ ********

code [fround]   \ f mode -- f' ; round floating point value
  sub   ebp, # #fpframe         \ allow space for results
  fstcw #cw [ebp]               \ get control word
  mov   cx, #cw [ebp]           \ and preserve it
  and   word #cw [ebp], # $F3FF \ strip rounding control ; SFP005
  or    word #cw [ebp], bx      \ or in new mode
  fldcw #cw [ebp]               \ put new rounding control back
  fld   fword #fpframe [ebp]	\ load old NOS
  frndint                       \ rounding operation
  fstp  fword #fpframe [ebp]	\ get result
  mov   ebx, #fpframe [ebp]	\   which will be new TOS
  mov   #cw [ebp], cx           \ restore previous control word
  fldcw #cw [ebp]
  add   ebp, # #fpframe 4 +     \ discard frame and old NOS
  fnext,
end-code

: fround        \ f1 -- f1'
( *G Round the number to nearest or even. )
  0 [fround]
;

: fint          \ f1 -- f1'
( *G Chop the number towards zero. )
  $0C00 [fround]
;

: floor         \ f1 -- f1'
( *G Floored round towards -infinity. )
  $0400 [fround]
;

: roundup       \ f1 -- f1'
( *G Round towards +infinity. )
  $0800 [fround]
;

: rounded	\ -- ; set NDP to round to nearest
( *G Set NDP to round to nearest for all operations other than )
( ** FINT FLOOR and ROUNDUP. )
  cw@ $F3FF and ( $0 or ) cw!
;
rounded

: floored	\ -- ; set NDP to floor
( *G Set NDP to round to floor for all operations other than )
( ** FROUND FINT and ROUNDUP. )
  cw@ $F3FF and $400 or cw!
;

: roundedup	\ -- ; set NDP to round up
( *G Set NDP to round up for all operations other than )
( ** FROUND FINT and FLOOR. )
  cw@ $F3FF and $800 or cw!
;

: truncated	\ -- ; set NDP to chop to 0
( *G Set NDP to chop to 0 for all operations other than )
( ** FROUND FLOOR and ROUNDUP. )
  cw@ $F3FF and $C00 or cw!
;


\ ****************************
\ Number conversion primitives
\ ****************************

\ *****************************
\ Floating point numeric output
\ *****************************

((
>BCD and >ASCII
===============
The following print control words are destructive, both to the
86 stack and the 87.  Too many numbers on the 87 stack will
result in an 87 stack overflow, which is not trapped out.
The 86 stack and the text buffer are serially written to and
then read from, altering the stack size accordingly, and finally
reset to point to the next free location after the 86 has placed
the 'real' floating point number at S0.
All conversions are messy, going through i) ascii to bcd, ii)
bcd to integer, iii) integer to floating, and iv) integer to
exponent.  Implementation of these steps is necessary however,
and attention to detail is required.
The result will be accurate to 2 digits in the 15th decimal
place.

The print command makes extensive use of the text buffer to hold
both the value and its exponent, and intermediate results from
the bcd to ascii conversion.  It is destructive of ANYTHING that
was in the text buffer.

>BCD
====
This subroutine converts the floating point number on the 87
stack to a bcd string and power of ten value in the 8086 text
buffer area.
Unnormal, denormal, and pseudo zeros are not trapped out, but
do not seem to significantly affect the conversion.
The full 18 digit capability of the (8087) data register is used
but the final ascii string will only be 17 digits long.  The
generated ascii string will always have a leading "+" or "-".
The decimal place will always be placed after the first digit,
although the input conversion routines allow free formatting.
It is assumed that the text buffer is longer than 35 bytes, and
that the 8086 stack has at least five free entries.
Do not alter any of the next few routines, they are quite
complex!!!

The number is converted within the following areas

PAD      = ascii_string ( 10 bytes )
PAD + 20 = scaling factor for power_of_ten
PAD + 24 = power_of_ten exponent value by LOG10
PAD + 28 = power_of_ten value
PAD + 34 = power_of_two
PAD + 42 = fraction or bcd string area ( 10 bytes )
PAD + 52 = not used
PAD + 60 = integer or real number flag

TOS is used to carry the sign from one routine to the other
))

hex
create ^1e17  00 c, 00 c, 00 c, C5 c, 2E c, BC c, A2 c, B1 c,  37 c, 40 c,
decimal

code >bcd               \ f pad -- flag
                        \ PAD is in TOS (EBX), and is a valid address
  fld	fword 0 [ebp]			\ load NDP from NOS

  cld                                   \ increment all further instructions
  sub   ebp, # #fpframe                 \ build frame

\ Fill PAD with 20 bytes of ASCII `0'
  mov   edx, edi                        \ preserve EDI (Forth reg)
  mov   al, # char 0
  mov   ecx, # 20
  mov   edi, ebx                        \ es:di==pad
  rep stosb                             \ initialise the pad area for type; al=48, ecx=20
  mov   edi, edx                        \ restore EDI

  xor   eax, eax                        \ setting `real/int' and `power_10' to zero
  mov   dword 60 [ebx], eax             \ assume real unless told otherwise; ds:di==pad
  mov   dword 28 [ebx], eax             \ zero the power of ten value

\ Test and save the sign of the top item of the FP stack
  ftst                                  \ determine sign of top 87 stack word
  fstsw word #sw [ebp]                  \ #sw [BP] = sw
  and   word #sw [ebp], # $4100         \ 04100h == C3 and C0
  test  word #sw [ebp], # $0100         \ 0100h == look at C0, set if <0.0
  nz, if,
    mov   edx, # 1                      \ if -ve,
  else,
    mov   edx, # 0                      \ if +ve, corresponding values used
  endif,

  test  word #sw [ebp], # $4000         \ skip the power code routine if; ebx==sw
  jne   l$3                             \ the number was zero, C3 is FZ bit
  fabs                                  \ work only with positive numbers
  fxtract                               \ separate the significand/exponent
  fstp  tbyte 42 [ebx]                  \ save the fraction temporarily; di==pad
  fist  dword 34 [ebx]                  \ save the exponent temporarily, then
  fldlg2                                \  calculate the proper power of ten.
  fmulp st(1), st                       \  exponent * lg2 base 10
  fistp dword 28 [ebx]                  \ store the power_of_ten value
  mov   dword eax, 28 [ebx]             \ immediately retrieve it again,
  sub   eax, # 17                       \ and calculate the scaling factor to
					\ get the number to fit into bcd limit
					\ of sign+17 characters
  mov   dword 20 [ebx], eax             \ set initial power_of_ten,
\  mov   ecx, eax
  ja    l$6                             \ jump if eax u> 17d
  fild  dword 34 [ebx]                  \ load power_of_two, and
  fld   tbyte 42 [ebx]                  \ load the fraction
  fscale                                \ then set up the number
  fxch                                  \ fswap
  fstp  st                              \ fdrop
  fld   st                              \ fdup
  jmp   l$5
l$3:                                    \ branch island for -128..+127 bytes
  mov   dword 60 [ebx], # 1             \ force result to integer
  jmp   l$4
l$5:

\ determine whether result can be displayed as an integer, or
\ must be displayed in FP format.
  frndint                               \ convert FTOS to integer
  fcomp  st(1)                          \ compare with the real version
  fstsw  word #sw [ebp]                 \ the equality of the number to; ebx==sw
  test   word #sw [ebp], # $4000        \ its integer equivalent
  nz, if,
    mov  dword 60 [ebx], # 1            \ yes, an integer between 1 and 1e18
    jmp  l$4                            \ go to end of proc. via l$11 to l$10
  endif,
  mov dword 60 [ebx], # 0               \ no, non-integer...need to scale to range
  fstp  st                              \ fdrop
l$6:            \ this label is only for numbers needing FP format display
                                        \ EAX holds the scaled exponent
  fild  dword 20 [ebx]                  \ load power_of ten value
  fchs                                  \ negate FTOS
  fldl2t                                \ load log 10 base 2 onto 87 stack
  fmulp st(1), st                       \ power_of_ten * log 10 base 2
  call 2**z                             \ calculate scaling factor and
  fxtract                               \ get scaled up number, power 2
  fld tbyte 42 [ebx]                    \ load the fraction part of the number
  fmulp st(1), st
\ and combine to within the range allowed by the bcd format,
\ to within an order of magnitude.
  jmp   l$2
l$4:                                    \ this jump dest is to get round the branch size
  jmp   l$10                            \ ...ditto...
l$2:
  fild  dword 34 [ebx]                  \ load the power_of_two value
  faddp st(2), st                       \ combine the powers of two, to form
  fscale                                \ the full value of the number without
  fxch                                  \ affecting the value of the exponent.
  fstp  st
  frndint                       	\ rounding operation
  fld	tbyte ptr ^1e17			\ form 10^17
  fcomp st(1)                           \ test the adjusted number to see if
                                        \ the error is one order of magintude
  fstsw word #sw [ebp]                  \ too large to fit into bcd field; ebx==sw
  test  word #sw [ebp], # $4100         \ if c0 = 0 then number is >= to
  nz, if,                               \ to the lower boundary
  else,
    dec  dword 28 [ebx]                 \ adjust power_of_ten value
  endif,
  frndint                               \ form integer value of mantissa
l$10:           \ all numbers pass this way
  fbstp 42 [ebx]                        \ store the integer mantissa in pad

  mov   ebx, edx                        \ flag required by next word

  add   ebp, # #fpframe 4 +		\ release frame and NOS

  fnext,
end-code

hex

code >ascii     \ pad flag -- length
\ Save the Forth registers
  push  esi
  push  edi

  mov   edi, 0 [ebp]                    \ DS:DI=pad
  add   ebp, # 4

  mov   edx, ebx                        \ EDX=flag

  push  ebp                             \ preserve data stack pointer!

  mov   ebp, edi                        \ ebp==pad, PAD must be in stack segment!
  mov   esi, # $08                      \ initial bcd size for index count
  mov   ecx, # $0F04                    \ shift count and mask
  mov   ebx, # 01                       \ initial field size for ascii string
  cld                                   \ autoincrement upwards in memory
  mov   al, # char +                    \ set "sign" field to "+"
  test  dl, # 01                        \ look for a negative number, and if so
  jz    l$1                             \  then change the
    mov   al, # char -                  \  "sign" field to "-"
l$1:
  stosb                                 \ bump sign character out to pad
  and   dl, # $07F                      \ turn off sign bit
l$2:            \ now remove surplus zeros from field
  mov   byte ah, $02A [ebp] [esi]       \ retrieve a bcd byte from PAD
  mov   al, ah                          \ copy the byte's value
  shr   al, cl                          \ get high order digit, and test the
  and   al, ch                          \ high order bit to see if leading
  jnz   l$4                             \ zero found, if not, print number
  mov   al, ah                          \ get bcd byte again
  and   al, ch                          \ test the low order digit, and see
  jnz   l$5                             \ if that was zero, if not, exit
  dec   esi                             \ dec the bcd index correspondingly
  jns   l$2                             \ if the significand has all zeros
  mov   al, # char 0                    \ then only print an initial zero
  stosb                                 \ bump string out to pad
  inc   ebx                             \ and exit the session
  jmp l$6
l$3:
  mov   byte ah, $02A [ebp] [esi]       \ retrieve a bcd byte from PAD
  mov   al, ah
  shr   al, cl                          \ get high order digit
l$4:
  add   al, # char 0                    \ convert to ascii by bumping 48
  stosb                                 \ bump string out to pad
  mov   al, ah                          \ get low order digit
  and   al, ch
  inc   ebx                             \ increase the field size counter
l$5:
  add   al, # char 0                    \ convert to ascii by bumping 48
  stosb                                 \ bump string out to pad
  inc   ebx                             \ increase the field size counter
  dec   esi                             \ try next bcd byte value
  jns l$3

l$6:            \ conversion is now complete
  pop   ebp                             \ Restore the Forth registers
  pop   edi
  pop   esi

  fnext,
end-code

: >exp		\ f -- f n ; EDA010
\ Return decimal exponent as integer n
  fdup f0= if
    0
  else
    fdup fabs flog floor f>s
  endif
;

: f>ascii	\ f buff -- length
  cw@ >r  rounded
  over f0=				\ set decimal exponent
  if  0  else  over fabs flog floor f>s  endif
  over #24 + !
  dup >r  >bcd  r> swap >ascii
  r> cw!
;


\ ---------------------
\ valid number checking
\ ---------------------

((
These words assume that the value to be tested has been placed
in a 4-byte real whose address is passed to them.
))

: fsign?	\ addr -- flag ; true if negative
  @ 0<
;

: infinity?	\ addr -- flag ; true if infinite
  @ dup $7F800000 and $7F800000 =
  swap  $007FFFFF and 0= and
;

: NaN?		\ addr -- flag ; true if not a number
  @ dup $7F800000 and $7F800000 =
  swap  $007FFFFF and 0<> and
;
((
: infinity?	\ addr -- flag ; true if infinite
  @ dup $7F000000 and $7F000000 =
  swap  $00800000 = and
;

: NaN?		\ addr -- flag ; true if not a number
  @ dup $7F000000 and $7F000000 =
  swap  $00800000 <> and
;
))

\ ****************
\ Compiler support
\ ****************

decimal

: flit		\ -- f ; picks up inline literal
  r> dup f@  float+ >r
;

defer fliteral	\ f -- ; -- f
( *G Compiles a float as a literal into the current definition. )
( ** At execution time, a float is returned. For example, )
( ** [ %PI F2* ] FLITERAL will compile 2PI as a floating point )
( ** literal. Note that FLITERAL is immediate. )
  immediate

((
: (fliteral)	\ f -- ; -- f
  compile flit  f,			\ for HfpGL32, a float is one cell
;
))
assign literal to-do fliteral

: #fliteral              \ f1..fn n --
  abs #literal
;


\ ***************
\ FP trigonometry - 387 and up required
\ ***************

hex

code ftan       \ f -- tan(f)
( *G Floating point tangent. )
  >FPU
  fptan
  fstp  st
  FPU>
  fnext,
end-code

code fatan      \ f -- atan(f)
( *G Floating point arctangent. )
  >FPU
  fld1
  fpatan
  FPU>
  fnext,
end-code

code fsin       \ f -- sin(f)
( *G Floating point sine. )
  >FPU
  fsin
  FPU>
  fnext,
end-code

code fasin      \ f -- asin(f)
( *G Floating point arcsine. )
  >FPU
  fld   st                      \ x x
  fld1                          \ x 1 x
  fxch
  fld   st                      \ x 1 x x
  fmulp st(1), st               \ x 1 (x^2)
  fsubp st(1), st               \ x (1-(x^2))
  fsqrt                         \ x (1-(x^2))
  fxch
  fdivrp st(1), st
  fld1
  fpatan                        \ atan (x/(1-(x^2)))
  FPU>
  fnext,
end-code

code fcos       \ f -- cos(f)
( *G Floating point cosine. )
  >FPU
  fcos
  FPU>
  fnext,
end-code

code facos      \ f -- acos(f)
( *G Floating point arctangent. )
  >FPU
  fld   st                      \ x x
  fld1                          \ x 1 x
  fxch
  fld   st                      \ x 1 x x
  fmulp st(1), st               \ x 1 (x^2)
  fsubp st(1), st               \ x (1-(x^2))
  fsqrt                         \ x (1-(x^2))
  fdivrp st(1), st
  fld1
  fpatan                        \ atan (x/(1-(x^2)))
  FPU>
  fnext,
end-code

code fsincos    \ f -- sin(f) cos(f)
( *G Returns sine and cosine values of f. )
  >FPU
  fsincos
  fxch
  FPU> FPU>
  fnext,
end-code

: fatan2        \ f1 f2 -- atan(f1/f2)
( *G Floating point arctangent with prior division. )
  f/ fatan
;

: deg>rad	\ fdeg -- frad
( *G Converts a value in degrees to radians. )
\  %pi f*  #180 s>f f/
  %pi/180 f*
;

: rad>deg               \ -- ;
( *G Converts a value in radians to degrees. )
\  #180 s>f f*  %pi f/
  %180/pi f*
;

code freduce    \ f1 -- f2 ; reduce value to range 0..2pi
( *G Reduce f1 to be in the range 0 <= f2 < 2PI. )
  >FPU
  begin,
    fldpi                       \ constant of 2pi
    fadd   st, st
    fxch
    fprem                       \ reduce
    fstsw  ax                   \ get status
    fxch			\ remove 2pi
    fstp   st
    and    ax, # $0400          \ inspect C2 bit
  z, until,                     \ until C2 clear
  FPU>
  fnext,
end-code


\ ******************************************************
\ Extra trig functions - COSEC SEC COTAN and hyberbolics
\ ******************************************************

: fcosec        \ f -- cosec(f)
( *G Floating point cosecant. )
  fsin 1/f  ;

: fsec          \ f -- sec(f)
( *G Floating point secant. )
  fcos 1/f  ;

: fcotan        \ f -- cot(f)
( *G Floating point cotangent. )
  ftan 1/f  ;

: fsinh         \ f -- sinh(f) ; (e^x - 1/e^x)/2
( *G Floating point hyberbolic sine. )
  fexp  fdup 1/f f-  f2/  ;

: fcosh         \ f -- cosh(f) ; (e^x + 1/e^x)/2
( *G Floating point hyberbolic cosine. )
  fexp  fdup 1/f f+  f2/  ;

: ftanh         \ f -- tanh(f) ; (e^x - 1/e^x)/(e^x + 1/e^x)
( *G Floating point hyberbolic tangent. )
  fexp fdup 1/f  fover fover f-		\ GEK012
  frot frot f+  f/
;

: fasinh        \ f -- asinh(f) ; ln(f+sqrt(1+f*f))
( *G Floating point hyberbolic arcsine. )
  fdup
  fdup f*  %1 f+  fsqrt
  f+  fln  ;

: facosh        \ f -- acosh(f) ; ln(f+sqrt(f*f-1))
( *G Floating point hyberbolic arccosine. )
  fdup
  fdup f*  %1 f-  fsqrt
  f+  fln  ;

: fatanh        \ f -- atanh(f) ; ln((1+f)/(1-f))/2
( *G Floating point hyberbolic arctangent. )
  %1 fover f+  %1 frot f-  f/
  fln  f2/  ;


\ *****************
\ Number conversion
\ *****************

((
Converting from ascii to floating point is not so hard as before
since there are some FORTH words for some of the jobs.  It still
consists of four basic steps: i) obtain the significand and its
decimal maginitude, ii) get the exponent, iii) convert ascii to
integer, and then iv) correct the power value.

The input format takes the following form:

[-] [digit(s)] [.] [digit(s)] [E|e] [+|-] [digit(s)]
))

decimal

create powers10	\ -- addr ; table of powers of 10
                    1, swap , ,		\ in Intel order (low, high)
		   10, swap , ,		\ 10^1
		  100, swap , ,		\ 10^2
		 1000, swap , ,		\ 10^3
		10000, swap , ,		\ 10^4
	       100000, swap , ,		\ 10^5
	      1,000000 swap , ,		\ 10^6
	     10,000000 swap , ,		\ 10^7
	    100,000000 swap , ,		\ 10^8
	   1000,000000 swap , ,		\ 10^9
	  10000,000000 swap , ,		\ 10^10
	 100000,000000 swap , ,		\ 10^11
	1,000000000000 swap , ,		\ 10^12
       10,000000000000 swap , ,		\ 10^13
      100,000000000000 swap , ,		\ 10^14
     1000,000000000000 swap , ,		\ 10^15
    10000,000000000000 swap , ,		\ 10^16
   100000,000000000000 swap , ,		\ 10^17
  1000000,000000000000 swap , ,		\ 10^18

code get10^n	\ n -- f ; 0<=n<=18
  mov	eax, # powers10
  fild  qword 0 [eax] [ebx*8]	\ load integer from table
  fstp  fword -4 [ebp]
  mov	ebx, -4 [ebp]
  fnext,
end-code

: (10**n)	\ +n -- f
  dup #18 > if
    #10 s>f  swap s>f f**  exit
  endif
  get10^n
;

: 10**n		\ n -- ; -- f
  dup abs (10**n)  swap 0<
  if  1/f  endif
;

: exp-char?	\ char -- flag ; true if character is exponent indicator
  case  upc
    [char] D of  true  endof
    [char] E of  true  endof
                 drop false
  end-case
;

[-short-branches
: >FLOAT	\ c-addr u -- f true | 0
( *G Try to convert the string at c-addr/u to a floating point number.  )
( ** If conversion is successful, flag is returned true, and a floating )
( ** number is returned on the float stack, otherwise flag is returned  )
( ** false and the float stack is unchanged.                            )
  { c-addr u | dlo dhi u' fsign flag nshift obase -- flag }

  u 0= IF false exit THEN               \ must have at least one significand digit

  base @ -> obase  decimal		\ preserve and force base

  false -> flag
  0 -> nshift

  c-addr c@ [char] + = if               \ Ignore preceeding +
    c-addr 1 chars + -> c-addr
    u 1 chars - -> u
  then

  c-addr c@ [char] - =                  \ check for preceeding -
  dup -> fsign IF
    c-addr 1+ -> c-addr
    u 1- -> u
  THEN

  \ convert mantissa (first set of digits)
  0 0 c-addr u >number -> u' -> c-addr -> dhi -> dlo
  u' 0> IF
    \ convert optional number after decimal point
    c-addr c@ fp-char @ = IF
      dlo dhi c-addr 1+ u' 1- dup -> nshift >number
      dup nshift - -> nshift
      -> u' -> c-addr -> dhi -> dlo
    THEN
    \ convert exponent
    u' 0 > IF
      c-addr c@ exp-char? IF
        u' 1 > if
          c-addr 1+ u' 1- (integer?)
        else
          0 1
        then
        IF
          nshift + -> nshift
          true -> flag
        THEN
      THEN
    ELSE
      true -> flag
    THEN
  ELSE
    u u' <> -> flag
  THEN

  \ convert double precision int to float
  flag IF
    cw@ >r  rounded
    dlo dhi d>f
\    10 s>f nshift s>f f** f*   \ apply exponent
    nshift 10**n f*
    fsign IF  fnegate  THEN
    r> cw!
  THEN

  flag  obase base !
;
short-branches]


\ --------------------------
\ f. sub words - places, etc
\ --------------------------

decimal

variable #places
  7 #places !

: precision     \ -- u
\ *G Returns the number of significant digits used by *\fo{F.}
\ ** *\fo{FE.} and *\fo{FS.}.
  #places @
;

: set-precision \ u --
\ *G Sets the number of significant digits used by *\fo{F.}
\ ** *\fo{FE.} and *\fo{FS.}.
  0 max  #fdigits min  #places !
;

8 set-precision

: places        \ u --
\ *G Sets the number of significant digits used by *\fo{F.}
\ ** *\fo{FE.} and *\fo{FS.}.
\ ** The ANS version of this word is *\fo{SET-PRECISION}, which is preferable
\ ** in new code.
  set-precision
;

((
: real.         \ c-addr flag -- ; show the significand
  over c@ [char] - = or if		\ decide whether to show sign
    dup c@ emit
  endif
  dup 1 + c@ emit
  fp-char @ emit
  dup 2+  precision 1- 0 max type
  dup 24 + @				\ SFP005
  if
    ." e"
    24 + @ .				\ SFP005
  else
    drop  space
  endif
;
))
: real.         \ c-addr flag -- ; show the significand
  over c@ [char] - = or if		\ decide whether to show sign
    dup c@ emit
  endif
  dup 1 + c@ emit
  fp-char @ emit
  dup 2+  precision 1- 0 max type
  dup 28 + @				\ SFP005
  if
    ." e"
    28 + w@s .				\ SFP005
  else
    drop  space
  endif
;

\ EDA011...
: represent	\ f c-addr u -- n flag1 flag2
( *G Assume that the floating number is of the form +/-0.xxxxEyy. )
( ** Round the significand xxxxx to u significant digits and place its )
( ** representation at c-addr. If u is zero round the fractional )
( ** significand to a whole number. flag2 is true if the results are )
( ** valid. n is the signed integer version of yy and flag1 is true if )
( ** f is negative. In this implementation all errors are handled by )
( ** exceptions, and so flag2 is always true, except for NaNs and )
( ** Infinites. The number of characters at c-addr is the greater of )
( ** u or MAX-FLOAT-DIGITS. )
  { caddr u | checkit[ #10 ] sign exp -- n flag1 flag2 }
  caddr u #fdigits max blank		\ set to spaces for text messages
  fdup checkit[ !			\ save for test
  checkit[ NaN? if
    fdrop checkit[ fsign? dup
    if    [ s" -NaN" #fdigits min ] sliteral
    else  [ s" +NaN" #fdigits min ] sliteral
    then  tuck caddr swap cmove 	\ set o/p to "NAN"
    swap 0 exit				\ return length sign 0
  endif
  checkit[ infinity? if
    fdrop checkit[ fsign? dup
    if    [ s" -Inf" #fdigits min ] sliteral
    else  [ s" +Inf" #fdigits min ] sliteral
    then  tuck caddr swap cmove 	\ set o/p to "INF"
    swap 0 exit				\ return length sign 0
  endif
  caddr u
  2dup #fdigits max [char] 0 fill	\ prefill with '0' chars
  #fdigits min  2>r
  >exp 1+ to exp
  r@ exp - 10**n  f*
  fround  fdup f0< to sign
  fabs f>d  2dup <# #s #>
  dup r@ - addr exp +!
  2r>  rot min  1 max  cmove
  d0= if  1 false  else  exp sign  endif
  true					\ exceptions handled by CPU and CATCH/THROW
;
\ ...EDA011

: fe.		\ f -- ; display in engineering format
( *G Display f in engineering notation, in which the exponent is always a power )
( ** of three, and the significand is always in the range 1.xxx to 999.xxx. )
  { f -- }
  addr f NaN? if
    addr f fsign?
    if  ." -NaN"  else  ." +NaN"  endif
    exit
  endif
  addr f infinity? if
    addr f fsign?
    if  ." -Inf"  else  ." +Inf"  endif
    exit
  endif

  base @ decimal
  f pad f>ascii	 pad swap		\ -- pad len ; generate ASCII text	SFP009
  precision 1+ min			\ allowing for sign and PRECISION
\ Calculate modified exponent
\  pad 24 + @ s>d 3 fm/mod 3 * >r	\ -- pad len step
  pad 28 + w@s s>d 3 fm/mod 3 * >r	\ -- pad len step
\ Display significand
  2+ 0 do
    over c@ emit  1 /string		\ -- pad' len' ; before decimal point
  loop
  fp-char @ emit
  0 max dup
  if  type  else  2drop  endif		\ -- ; after decimal point
\ display exponent
  r>
  [char] E emit
  dup 0< if  ." -"  else  ." +"  endif
  abs .
  base !
;

: fs.		\ f -- ; display in scientific format
( *G Display f in scientific notation, with one digit before the )
( ** decimal point. )
  { f -- }
  addr f NaN? if
    addr f fsign?
    if  ." -NaN"  else  ." +NaN"  endif
    exit
  endif
  addr f infinity? if
    addr f fsign?
    if  ." -INF"  else  ." +INF"  endif
    exit
  endif

  base @ decimal
  f pad f>ascii drop			\ generate ASCII text
  pad true real.			\ sign always displayed
  base !
;

[-short-branches
: f.            \ f -- ; display number in FTOS
( *G Display f as a float in fixed point notation with a trailing space. )
  { f -- }
  addr f NaN? if
    addr f fsign?
    if  ." -NaN"  else  ." +NaN"  endif
    exit
  endif
  addr f infinity? if
    addr f fsign?
    if  ." -INF"  else  ." +INF"  endif
    exit
  endif

  base @ decimal
  f pad f>ascii
  pad 60 + c@ if
    pad swap type space
  else
    drop  pad false real.
  endif
  base !
;
short-branches]

: f?		\ addr -- ; displays contents of FVARIABLE
( *G Displays the contents of the given FVARIABLE. )
  f@ f. ;


\ *********************
\ Environment Variables
\ *********************

also environment definitions

true		constant	FLOATING
true		constant	FLOATING-EXT
0		constant	FLOATING-STACK

pad 3 cells $FF fill

FPCELL #10 = [if]  $00007FFE  pad 2 cells + !  [then]
FPCELL 8   = [if]  $7FEFFFFF  pad 1 cells + !  [then]
FPCELL 4   = [if]  $7EFFFFFF  pad           !  [then]

pad f@		fconstant	MAX-FLOAT

#fdigits	constant	MAX-FLOAT-DIGITS	( not ANS )

previous definitions


\ *********************************************
\ Patch the floating point pack into the system
\ *********************************************

defer old-number?
  action-of number? to-do old-number?

: fnumber?  	\ c-addr -- 0 | n 1 | d 2 | f -1
( *G Behaves like the integer version of NUMBER? except that if )
( ** integer conversion fails, and BASE is decimal, a floating )
( ** point conversion is attempted. If conversion is successful, )
( ** the floating point number is left on the float stack and )
( ** the result code is -1. )
  dup >r  old-number? dup 0= if		\ test for integer first
    base @ #10 = if			\ not integer, floats must be in DECIMAL
      drop  r@ count >float		\ >FLOAT returns well formed flag
\      -1 and				\ good float returns -1,
    then
  then
  r> drop
;

assign fnumber? to-do number?
assign #fliteral to-do c#fliteral

: .NSW		\ sw -- ; display NDP status word
  ." NDP SW = " .word space
;

: .FSysPrompt	\ -- ; adds floating point stack depth display
  .SysPrompt
  ndepth if
    ."  N:-" ndepth .
  endif
  sw@ dup $0040 and if
    cr ." NDP Stack Fault: " dup .NSW  fclex
  endif
  dup $001D and if	\ ignore precision and denormals
    cr ." NDP Potential Exception: " dup .NSW  fclex
  endif
  drop
;
assign .FSysPrompt to-do .prompt

: reals		\ -- ; turn FP system on
( *G Enables the floating point package for number conversion. )
  assign fnumber? to-do number?
  assign .FSysPrompt to-do .prompt
;

: integers	\ -- ; turn FP system off
( *G Disables the floating point package for number conversion. )
  [ also system ]
  assign wininteger? to-do number?
  [ previous ]
  assign .SysPrompt to-do .prompt
;


\ ********************
\ PFW2.x compatibility
\ ********************

: (f#)		\ addr -- fn ior ; counted string to FP number
  fnumber?
  case
    0  of  0  endof
    1  of  s>f FpSystem  endof
    2  of  d>f FpSystem  endof
  end-case
;

: f#		\ -- f ; or compiles it [ state smart ]
( *G Used in the form "F# <number>", the <number> string is converted	)
( ** and promoted if required to a floating point number. If the system	)
( ** is compiling the float is compiled. If <number> cannot be		)
( ** an error occurs.							)
  bl word (f#) ?undef
  state @ if
    [compile] fliteral
  endif
; immediate


\ *****************
\ Debugging support
\ *****************

( *S Debugging facilities )
( *P Debugging floating point code is often difficult, as failures )
( ** can occur because of the necessary approximations involved )
( ** in floating point operations. )

( *P If you set the constant [FPDEBUG] true when compiling HfpGL32.FTH, )
( ** a debug build will be constructed. The state of the FPU will be )
( ** checked after each word. If a floating point exception )
( ** has been generated, a diagnostic is issued, and the system )
( ** aborts. Set this only when testing, as it slows down the normal )
( ** operation of floating point words. )

( *P The debugger works by intercepting the end of each code definition )
( ** which is finished by FNEXT, rather than the normal NEXT, or RET. )
( ** See the source code in LIB\HfpGL32.FTH for more details. )

((
NDP status word:
bit	Function
================
15	Busy
14	C3
13..11	TOP
10	C2
9	C1
8	C0
7	Set on Unmasked Exception
6	Set on stack fault (bit 9 set for overflow, clear for underflow)
5	Set on Precision Exception
4	Set on Underflow
3	Set on Overflow
2	Set on Zero Divide
1	Set on Denormalised Operand
0	Set on Invalid Operation
))

[fpdebug] [if]
: .NDP		\ -- ; what to do if exception found
  sw@ dup .NSW
\  dup $20 and if ." Precision " endif
  dup $10 and if ." Underflow " endif
  dup $08 and if ." Overflow " endif
  dup $04 and if ." ZeroDivide " endif
\  dup $02 and if ." Denormalised " endif
  dup $01 and if ." Invalid " endif
  drop
  fclex				\ clear exceptions
  hex .rs  decimal		\ Return stack trace
  finit abort			\ reset stack and restart
;

: (fpcheck)	\ -- ; action of debugger
  sw@ $0040 and if
    cr ." NDP Stack Fault: " .NDP
  endif
  sw@ $001D and if	\ ignore precision and denormals
    cr ." NDP Exception: " .NDP
  endif
;

: +fpcheck	 \ -- ; enable FP checking
( *G Enables the floating point debugger if it has been compiled. )
  assign (fpcheck) to-do fpcheck
;

: -fpcheck	 \ -- ; disable FP checking
( *G Disbles the floating point debugger if it has been compiled. )
  assign noop to-do fpcheck
;
-fpcheck
[then]


\ =======================================================================
( *> ###                                                                )
\ =======================================================================

((
code tf@        \ addr -- f
( *G Fetches the 80 bit extended float to an HfpGL32 float )
  fld   tbyte 0 [ebx]
  fstp  fword -4 [ebp]
  mov   ebx, -4 [ebp]
  fnext,
end-code inline

code tf!        \ f addr
( *G Stores the HfpGL32 float as an 80 bit extended float )
( ** number at addr. )
  fld   fword 0 [ebp]		\ get HfpGL32 float from NOS
  fstp  tbyte 0 [ebx]		\ store it
  mov   ebx, 4 [ebp]		\ restore TOS
  lea	ebp, 8 [ebp]
  fnext,
end-code inline

: tf,           \ f --
( *G Lays an 80 bit extended float into the dictionary, reserving )
( ** 10 bytes )
  here  #10 allot  tf!          \ allocate 10 bytes and stash number here
;

: tfalign       \ --
( *G Aligns the dictionary to accept an 80 bit extended float. )
; immediate

: tfaligned     \ addr -- addr'
( *G Aligns the address to accept an 80 bit extended float. )
; immediate

: tfloat+       \ addr -- addr'
( *G Increments addr by the size of an 80 bit extended float. )
  #10 +
;

: tfloats       \ n1 -- n2
( *G Returns n2, the size of n1 80 bit extended floats. )
  #10 *
;
))

((
\ test code

decimal

finit rounded

#256 buffer: foo

: t		\ -- ; f: f --
  foo >bcd .
  foo #80 dump
;

: tr		\ --
  cr  #1000 0 do
    cr
    i 5 bounds do
      i s>f  0.1e0 f+  fs.  out @ #15 + #-16 and >pos
    loop
  5 +loop
  cr
;

finit rounded
))

