\ CIAOCORE.FTH - Ciao main code

((
Copyright (c) 2000-2002, 2018
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====
Validate combinations of virtual and static for all methods

Modify static nested class to store vtable if there is one.
probably involves another parameter in LIST to tell whether
the THIS index points to the instance data or the vtable cell
during compile.

IPTR: for instance pointer members is not done.

Change history
==============
20180508 SFP004 v5 and NDCS changes.
20010109 SFP003 Updated code for compilation without DOCGEN
20001102 SFP002 Renamed some words to avoid name clashes with VFX kernel,
		including STRUCT{ ... } from STRUCT ... END-STRUCT
		Fixed PROCESS1STTOKEN.
20001030 SJC001 First release
))

variable start-ciao
  here start-ciao !


\ =======
\ *! ciao
\ *T CIAO - C Inspired Active Objects
\ =======
( *P CIAO is an OOP package modelled on C++ for VFX Forth.              )
( ** CIAO is designed to provide easy interfacing to host operating     )
( ** system structures that are built around a C++ model.               )

\ *P The source code for CIAO is in the *\i{Lib\oop\Ciao} directory,
\ ** as are several example class files. To rebuild CIAO, compile
\ ** the file *\i{ciao.bld}.

\ ============================
\ *S Token and Parsing Helpers
\ ============================
( *P Various utilities and factors useful for parsing text.             )

S" MAX-CHAR" environment? 0= throw 1 chars +
  buffer: token-buffer
( *G A Memory buffer used to hold the result of the last token parse.   )
( ** The size of this buffer comes from the environement variable       )
( ** MAX-CHAR and is MAX-CHAR + 1 characters in length since the string )
( ** is stored as a counted string.                                     )

: new-word	\ char -- $
( *G This is a replacement for WORD which places the output in the      )
( ** token buffer.                                                      )
  word count dup token-buffer c!
  token-buffer 1 chars + swap move
  token-buffer
;

: peek-token	\ -- c-addr u
( *G Copy the next token into the TOKEN-BUFFER without permanent change )
( ** to the input specification uses SAVE-INPUT and RESTORE-INPUT.    )
( ** Returns TOKEN-BUFFER as a c-addr u pair.                           )
  save-input
  bl new-word >r
  restore-input drop
  r> count
;

: drop-token    \ --
( *G Throw away the next token *without* corrupting TOKEN-BUFFER.       )
  bl word drop
;

: ciao-token    \ -- c-addr u
( *G Grab the next space delimited token and return c-addr u. Fills     )
( ** TOKEN-BUFFER.                                                      )
  bl new-word count
;

: bracketed?    \ c-addr u -- flag
( *G Is the string C-ADDR U bracketed?                                  )
  dup 3 <
  if  2drop false  exit  then
  over c@ [char] ( <>
  if  2drop false  exit  then
  + 1 chars - c@ [char] ) =
;


\ =======================================================================
( **                                                                    )
( *S The THIS Stack                                                     )
( **                                                                    )
( *P The heart of this OOP implementation is the concept of "THIS".     )
( ** Just like C++ "THIS" returns the currently active object instance  )
( ** pointer. Instance data is accessed via this pointer as are the     )
( ** "virtual" methods.                                                 )
( **                                                                    )
( *P THIS is kept in a form of stack.                                   )
( **                                                                    )
\ =======================================================================

: >this		\ val --
( *G Set THIS to VAL. Preservation is taken care of in the compiler.  )
  currobj !
;

: this          \ -- instance-pointer
( *G Return the current instance pointer. Only valid within a method    )
( ** declaration.                                                       )
  currobj @
;


\ =======================================================================
( **                                                                    )
( *S CIAO Constants and Internal Data Stores                            )
( **                                                                    )
\ =======================================================================

$DEADCAFE       constant        MAGIC_CLASS

0               constant        SCOPE_PUBLIC
1               constant        SCOPE_PROTECTED
2               constant        SCOPE_PRIVATE

1               constant        DEF_STATIC
2               constant        DEF_VIRTUAL

$00             constant        TYPE_DATA
$01             constant        TYPE_STATICDATA

$10             constant        TYPE_CODE
$11             constant        TYPE_STATICCODE
$12             constant        TYPE_VIRTUALCODE

$20             constant        TYPE_CLASS
$24             constant        TYPE_CLASSPTR

SCOPE_PUBLIC Value CurrentScope	\ -- n
\ *G When defining a derived class this holds the scoping type employed.

0 Value CurrentClass	\ -- n
\ *G When defining a class this points to its CLASS structure.

0 Value DefFlags	\ -- n
( *G The declaration flags to be employed by the next method or data    )
( ** member defined in the current class. Records information from      )
( ** control definitions such as VIRTUAL and STATIC.                    )

0 Value CurrentDefClass	\ -- class
( *G During compilation of a code method, this value holds a pointer    )
( ** to the associated CLASS structure.                                 )

0 Value CurrentDefXT	\ -- xt
( *G During compilation of a code method, this value holds a pointer to )
( ** the XT of the definition. See the CIAO-COLON hook for details.     )

0 Value CurrentDefList	\ -- list
( *G During compilation of a code method, this value holds a pointer to )
( ** the internal method list to be used. The list will either be the   )
( ** classes public, protected or private chain depending on the scope  )
( ** at the time of the method declaration prototype.                   )

variable class-base-mem	\ -- addr
( *G This variable holds the value of HERE after the building of CIAO.  )
( ** It is used to sanity check the values passed to the instance       )
( ** destruction definition DELETE. Any passed value between this       )
( ** variables value and the current HERE is in dictionary space and    )
( ** must be a static instance which cannot be DELETEd.                 )


\ =======================================================================
( **                                                                    )
( *S Search Order Utilities                                             )
( **                                                                    )
\ =======================================================================

: NSEARCH-WORDLIST      \ WIDN .. WID1 N C-ADDR U -- XT FLAG | 0
( *G A most useful definition. FIND takes a counted string but searches )
( ** the whole search-order. SEARCH-WORDLIST takes a C-ADDR U pair but  )
( ** only searches one wordlist. This definition combines the two, and  )
( ** looks through a number of wordlists for a name described by a       )
( ** C-ADDR U pair. Usually used in association with GET-ORDER to       )
( ** provide a more useful version of FIND.                             )
  2>R
  BEGIN
    DUP
  WHILE
    1- SWAP
    2R@ ROT SEARCH-WORDLIST ?DUP IF
      2R> 2DROP 2>R
      0 ?DO DROP LOOP
      2R> EXIT
    THEN
  REPEAT
  DROP 2R> 2DROP 0
;

: (FindClass)   \ c-addr u -- ptr | THROW
( *G Run through the current search-order lookin for the name supplied. )
( ** If the name is found then a >BODY @ is employed on the XT to look  )
( ** for the MAGIC_CLASS identifier. Never called directly, this        )
( ** definition is run from FINDCLASS via CATCH to protect against the  )
( ** times when the token is found but is not a class. Due to the       )
( ** exception handling abilities of CATCH under VFX, this operation    )
( ** should be safe no matter what XT it is employed against.           )
  2>r get-order 2r> nSearch-WordList 0= if abort then
  >body dup @ MAGIC_CLASS <>
  if  -1 throw  then
;

: FindClass     \ c-addr u -- ptr | ABORTs
( *G Invoke FINDCLASS via CATCH. Will look for the token supplied and )
( ** if found will ensure it really is a class definition. ABORTs with  )
( ** text if anything goes wrong.                                       )
\ cr ." fc "  2dup type
  ['] (FindClass) catch
  abort" Failed to locate class"
;

\ =======================================================================
( **                                                                    )
( *S Method Lists                                                       )
( **                                                                    )
( *P The Method Lists hold all the required compiler information for    )
( ** each method within a class. In CIAO, methods don't ever actually   )
( ** exist as regular Forth words. Instead the act of defining a class  )
( ** builds the method lists. Each class has three of these, one for    )
( ** each valid scope public/protected and private.                   )
( **                                                                    )
( *N The Format of a Method List.                                       )
( **                                                                    )
( *E    | Link | Type | Param1 | Param2 | Name Len | Name Text |        )
( **    | CELL | CELL |  CELL  |  CELL  |   CHAR   |  n chars  |        )
( **                                                                    )
( *D Link       Pointer to start of previous list entry or 0 for top  )
( *D Type       The type of the method see types below                )
( *D Param1     Parameter 1, varies depending on TYPE.                  )
( *D Param2     Parameter 2, varies depending on TYPE.                  )
( *D NameLen    Length of method name.                                  )
( *D NameText   The text for the method name.                           )
( **                                                                    )
( *N TYPE_DATA                                                          )
( *P  Describes an instance data buffer, PARAM1 is the base offset from )
( **  THIS.                                                             )
( **                                                                    )
( *N TYPE_STATICDATA                                                    )
( *P  Describes a static data buffer. A static data buffer is placed    )
( **  within the global dictionary rather than being offset from THIS.  )
( **  The net result is that all instances of the owning class and any  )
( **  derived classes share the same location for this data element.    )
( **  PARAM1 is the address in global space of the buffer start.        )
( **                                                                    )
( *N TYPE_CODE                                                          )
( *P  The default code method type. PARAM1 points to a CELL in global   )
( **  dataspace which will contain the XT of the method body as soon as )
( **  it becomes available. A CODE method cannot be re-defined or       )
( **  rewritten and it's behaviour is inherited by any derived class.   )
( **                                                                    )
( *N TYPE_STATICCODE                                                    )
( *P  The second type of code method. It behaves in a similar fashion to)
( **  TYPE_CODE except the instance pointer THIS is not valid within    )
( **  the method body. These means that a static member has no access to)
( **  any other member of the class which is non-static. A static member)
( **  can also be invoked from a colon definition or the interpreter by )
( **  using the "named scope override" syntax, which does not require an)
( **  instance pointer. Primarily used to store "normal" functions in a )
( **  restricted namespace. Ie "do <something> in the name of           )
( **  <some class>"                                                     )
( **                                                                    )
( *N TYPE_VIRTUALCODE                                                   )
( *P  One of the most useful syntactic additions to C++ was the virtual )
( **  method. A virtual method can best be described as a method in a   )
( **  base class which you expect to have to modify or replace in a     )
( **  derived class. PARAM1 holds a 0 based index into a table of XTs   )
( **  called a "vtable". Each class has a vtable which in the case of a )
( **  derived class is initially inherited from the superclass. A       )
( **  derived class can either omit its function body and thus inherit)
( **  the behaviour of the superclass or it can define its own body   )
( **  which can also optionally elect to invoke the superclass's body   )
( **  by using the named scope override syntax. Therefore a virtual     )
( **  method can be either modified or replaced within the context of a )
( **  derived class. A particularly useful feature of the usefulness of )
( **  virtual methods can be seen later.                                )
( **                                                                    )
( *N TYPE_CLASS                                                         )
( *P  This type of method specifies a static instance of another class  )
( **  as being a part of the current. PARAM1 specifies the class type   )
( **  whilst PARAM2 specifies the offset from THIS for the instance     )
( **  pointer of the contained class.                                   )
( **                                                                    )
( *N TYPE_CLASSPTR                                                      )
( *P  A special form of data store which holds a pointer to a class     )
( **  instance. PARAM1 specifies the class type and PARAM2 the offset   )
( **  from THIS to a single cell. This cell will hold an instance       )
( **  pointer which can be dynamically assigned.                        )
( **                                                                    )
( *N The definitions which deal with lists are:                         )
( **                                                                    )
\ =======================================================================

: list_link     \ *entry -- *link
( *G Modify a pointer to a list head to point to the link field.        )
;

: list_type     \ *entry -- *type
( *G Modify a pointer to a list head to point to the type field.        )
  1 cells +
;

: list_param1   \ *entry -- *param1
( *G Modify a pointer to a list head to point to the param1 field.      )
  2 cells +
;

: list_param2   \ *entry -- *param2
( *G Modify a pointer to a list head to point to the param2 field.      )
  3 cells +
;

: list_namelen  \ *entry -- *namelen
( *G Modify a pointer to a list head to point to the namelen field.     )
  4 cells +
;

: list_name     \ *entry -- *name
( *G Modify a pointer to a list head to point to the name field.        )
  4 cells + 1 chars +
;

: .list-type    \ n --
( *G Given contents of a list entry's type field print it's name as an  )
( ** ascii text string.                                                 )
  case
        TYPE_DATA               of      ." IData"       endof
        TYPE_STATICDATA         of      ." SData"       endof
        TYPE_CODE               of      ." ICode"       endof
        TYPE_STATICCODE         of      ." SCode"       endof
        TYPE_VIRTUALCODE        of      ." VCode"       endof
        TYPE_CLASS              of      ." Class"       endof
        TYPE_CLASSPTR           of      ." *Clss"       endof
        drop ." Dunno"
  end-case
;

: .list-entry   \ *entry --
( *G Supplied with a pointer to a list entry this definition will print )
( ** its contents in human readable form.                              )
  cr ."     List Entry At $" dup .dword
  ."  Link $" dup list_link @ .dword
  ."  Type " dup list_type @ .list-type
  ."  Par1 $" dup list_param1 @ .dword
  ."  Par2 $" dup list_param2 @ .dword
  ."  Name '" dup list_name swap list_namelen c@ type ." '"
;

: .list         \ *head --
( *G Supplied with the address of a variable which points to a list     )
( ** entry this definition will walk backwards through the linked list  )
( ** performing .LIST-ENTRY on each in turn.                            )
  @
  begin
    ?dup
  while
    dup .list-entry
    list_link @
  repeat
;

: +LIST         \ Type Param1 Param2 c-addr u *list-head --
( *G Using the first 5 parameters lay a list-entry structure in the     )
( ** dictionary and add it to the end of the list whose anchor address  )
( ** is at the address pointed to by *LIST-HEAD.                        )
  here over @ , swap !          \ lay link and update
  2>r rot ,                     \ type
  swap ,                        \ param1
  ,                             \ param2
  2r> dup c,                    \ namelen
  bounds ?do i c@ c, loop       \ name
;

\ =======================================================================
( **                                                                    )
( *S Operator List                                                      )
( **                                                                    )
( *P Each class has a linked list called the Operator Chain. This list  )
( ** contains the mapping of operator-id number against class method    )
( ** list entry from above.                                           )
( **                                                                    )
( *P An operator structure entry consists of three fields, the link,    )
( ** the operator id number and a method-list pointer.                  )
( **                                                                    )
\ =======================================================================

: oplist_link             ;
( *G Given a pointer to an operator structure return pointer to link    )

: oplist_op#    1 cells + ;
( *G Given a pointer to an operator structure return pointer to op#     )

: oplist_list   2 cells + ;
( *G Given a pointer to an operator structure return pointer to *list   )

: .op#                  \ n --
( *G Where possible print human readable description for operator id N  )
  case
    0 of  ." FetchOP"  endof
    1 of  ." StoreOP"  endof
    2 of  ." AddrOP "  endof
    drop  ." <unkwn>"
  end-case
;

: .oplist-entry \ *entry
( *G Display an operator structure in human readable form.              )
  cr ."    Operator " dup oplist_op# @ .op#
  ." , Link " dup oplist_link @ .dword
  ." , *list = $" oplist_list @ .dword
;

: .oplist       \ *head --
( *G Given a pointer to the head of an operator chain from a class,     )
( ** call .OPLIST-ENTRY for each member.                                )
  @
  begin
    ?dup
  while
    dup .oplist-entry
    oplist_link @
  repeat
;

: +OPLIST       \ op# *list-entry *head --
( *G Add record to the operator chain anchored at *HEAD               )
  here over @ , swap !          \ lay link and update
  swap , ,
;

\ =======================================================================
( **                                                                    )
( *S The CLASS structure                                                )
( **                                                                    )
( *P All classes defined have the same structure:                       )
( **                                                                    )
( *E Size       Navigation Word         Useage                          )
( ** ------------------------------------------------------------------ )
( ** CELL       class_magic             A magic 32 bit number used      )
( **                                    to signify a CLASS structure.   )
( ** CELL       class_super             Pointer to parent class for     )
( **                                    a derived class object.         )
( ** CELL       class_private           Method list anchor for PRIVATE  )
( **                                    definitions.                    )
( ** CELL       class_protected         Method list anchor for PROTECTED)
( **                                    definitions.                    )
( ** CELL       class_public            Method list anchor for PUBLIC   )
( **                                    definitions.                    )
( ** CELL       class_opchain           Anchor for the operator chain   )
( **                                    for this class.                 )
( ** CELL       class_sizeidata         Size of Instance data required. )
( ** CELL       class_#vtable           Number of entries in the        )
( **                                    virtual method table.           )
( ** CELL       class_pvtable           Pointer to the virtual method   )
( **                                    table.                          )
\ =======================================================================
\
\ Structure of a class
\
\       CELL    Class Magic
\       CELL    Pointer to SuperClass
\       CELL    Pointer to Private List
\       CELL    Pointer to Protected List
\       CELL    Pointer to Public List
\       CELL    Operator Chain
\       CELL    Size Instance Data
\       CELL    #vtable
\       CELL    pVtable -+
\                        |
\       ....    VTable  <-
\

: class_magic                           ;	\ identifies a class
: class_super           1 cells +       ;	\ pointer to superclass
: class_private         2 cells +       ;	\ anchor for private list
: class_protected       3 cells +       ;	\ anchor for protected list
: class_public          4 cells +       ;	\ anchor for public list
: class_opchain         5 cells +       ;	\ anchor for operator list
: class_sizeidata       6 cells +       ;	\ size of instance data
: class_#vtable         7 cells +       ;	\ size of vtable
: class_pvtable         8 cells +       ;	\ pointer to vtable

: .class	\ "name" --
( *G Display as much information about the class "name" as possible     )
( ** in a human readable form.                                          )
  ' execute >r
  cr ." Class:"
  cr ."   Magic                 $" r@ class_magic @ .dword
  cr ."   Super Class Pointer   $" r@ class_super @ .dword
  cr ."   Private List          " r@ class_private .list
  cr ."   Protected List        " r@ class_protected .list
  cr ."   Public List           " r@ class_public .list
  cr ."   Operator Chain        " r@ class_opchain .oplist
  cr ."   Size IData            $" r@ class_sizeidata @ .dword
  cr ."   Size VTable (idx)     $" r@ class_#vtable @ .dword
  cr ."   Pointer to VTable     $" r@ class_pvtable @ .dword
  r> drop
;


\ =======================================================================
( **                                                                    )
( *S Method Searching                                                   )
( **                                                                    )
( *P Definitions used to find a given method within a class.            )
( **                                                                    )
\ =======================================================================

: FindMethodInClass	\ c-addr u *class -- ptr SCOPE | -1
( *G Given a string containing the method name and a pointer to a class )
( ** structure, this definition attempts to get the method list entry   )
( ** for that method. On success a pointer to the method list structure )
( ** is returned as well as the SCOPE indicator, if the method does     )
( ** not exist in the specified class a -1 is returned.                 )

  dup >r class_public @ 	\ c-addr u *public -- ; R: *class --
  begin
    ?dup
  while
    dup list_namelen
    2over               	\ c-addr u ** $ c-addr u --
    rot count           	\ c-addr u ** c-addr u c-addr1 u1 --
    icompare 0=
    if  nip nip  r> drop  SCOPE_PUBLIC exit  then
    @
  repeat

  r@ class_protected @  	\ c-addr u *prot -- ; R: *class --
  begin
    ?dup
  while
    dup list_namelen
    2over               	\ c-addr u ** $ c-addr u --
    rot count           	\ c-addr u ** c-addr u c-addr1 u1 --
    icompare 0=
    if  nip nip  r> drop  SCOPE_PROTECTED exit  then
    @
  repeat

  r> class_private @    	\ c-addr u *prot --
  begin
    ?dup
  while
    dup list_namelen
    2over               	\ c-addr u ** $ c-addr u --
    rot count           	\ c-addr u ** c-addr u c-addr1 u1 --
    icompare 0=
    if  nip nip  SCOPE_PRIVATE exit  then
    @
  repeat

  2drop -1
;


\ =======================================================================
( **                                                                    )
( *S Default Method Actions                                             )
( **                                                                    )
( *P Any code method has a default action assigned when it is prototyped)
( ** as a debugging aid. Invoking a method for which you have defined no)
( ** code will give a polite message via ABORT"                         )
( **                                                                    )
\ =======================================================================

: vcrash	\ ?? --
( *G Default action for prototyped virtual methods.                     )
  true abort" Uninitialised virtual method invoked."
;

: scrash        \ ?? --
( *G Default action for prototyped static methods.                      )
  true abort" Uninitialised static method invoked."
;

: icrash        \ ?? --
( *G Default action for prototyped instance methods.                    )
  true abort" Uninitialised instance method invoked."
;


\ =======================================================================
( **                                                                    )
( *S Method Scope Specification                                         )
( **                                                                    )
( *P During class definition the scope can be altered. These definitions)
( ** are used to control/handle scoping.                                )
( **                                                                    )
\ =======================================================================

: public:       \ --
( *G During CLASS definition set the current scope to public.           )
  SCOPE_PUBLIC to CurrentScope
;

: protected:    \ --
( *G During CLASS definition set the current scope to protected.        )
  SCOPE_PROTECTED to CurrentScope
;

: private:      \ --
( *G During CLASS definition set the current scope to private.          )
  SCOPE_PRIVATE to CurrentScope
;

: GetCurrentList        \ -- *list-head
( *G Return the method list pointer for the current scope.              )
  CurrentClass CurrentScope
  case
    SCOPE_PUBLIC of     class_public    endof
    SCOPE_PROTECTED of  class_protected endof
    SCOPE_PRIVATE of    class_private   endof
    drop TRUE abort" Internal Error #120"
  end-case
;


\ =======================================================================
( **                                                                    )
( *S Name Format Checking                                               )
( **                                                                    )
( *P In order to reserve characters to provide the syntax for typecasts,)
( ** scope overrides and method definition certain characters are       )
( ** illegal for method and class names.                                )
( **                                                                    )
( *P Brackets are illegal, since they are used to perform typecasts.    )
( *P Colon is an illegal character since it is used for scope overrides.)
( *P Period dot is illegal since it is used for compound invokations. )
( *P Star * is illegal since it declares an instance pointer.         )
( **                                                                    )
\ =======================================================================

: ?validname    \ c-addr u --
( *G Check the name string supplied is valid for either a class or      )
( ** name. Causes an ABORT" on failiure.                                )
  bounds ?do
    i c@
    case
      [char] ( of drop-token true abort" Not a valid class/method name" endof
      [char] ) of drop-token true abort" Not a valid class/method name" endof
      [char] : of drop-token true abort" Not a valid class/method name" endof
      [char] . of drop-token true abort" Not a valid class/method name" endof
      [char] * of drop-token true abort" Not a valid class/method name" endof
      drop
    end-case
  loop
;


\ =======================================================================
( **                                                                    )
( *S Method Type Overrides                                              )
( **                                                                    )
( *P Methods and instance variables defined in a class can have       )
( ** various attributes, these are controlled by simple indicator       )
( ** words.                                                             )
( **                                                                    )
\ =======================================================================

: static        \ --
( *G Modify the global DEFFLAGS to include the static type.             )
  DefFlags DEF_STATIC or to DefFlags
;

: virtual       \ --
( *G Modify the global DEFFLAGS to include the virtual type.            )
  DefFlags DEF_VIRTUAL or to DefFlags
;

: post-def      \ --
( *G Clear the global DEFFLAGS, called after a member definition to     )
( ** reset ready for the next member.                                   )
  0 to DefFlags
;


\ =======================================================================
( **                                                                    )
( *S Data Method Prototyping                                            )
( **                                                                    )
( *P These routines are used within a CLASS or STRUCT{ definition to    )
( ** define data members.                                               )
( **                                                                    )
\ =======================================================================

: buff:         \ size "name" --
( *G Define a data member called "name" of SIZE bytes. By default       )
( ** instance specific data is created, if the member was modified by   )
( ** the STATIC keyword, then global space is allocated. STATIC data    )
( ** members share the same memory location for all instances of the    )
( ** class and any derived classes.                                     )

  peek-token ?validname

  DefFlags DEF_STATIC and if
        here swap allot
        TYPE_STATICDATA swap
  else
        TYPE_DATA swap
        CurrentClass class_sizeidata dup @ \ type size pidata idx --
        -rot +!                         \ type idx --
  then

  0 ciao-token

  GetCurrentList +List

  post-def
;

: cell:         \ "name" --
( *G A shortcut for a BUFF: of one cell.                                )
  1 cells buff:
;

: char:         \ "name" --
( *G A shortcut for a BUFF: of one char.                                )
  1 chars buff:
;


\ =======================================================================
( **                                                                    )
( *S Code Method Prototyping                                            )
( **                                                                    )
( *P These routines are used within a CLASS or STRUCT{ definition to    )
( ** define methods.                                                    )
( **                                                                    )
\ =======================================================================

: static-meth:  \ "name" --
( *G The action invoked by METH: when the STATIC modifier was present.  )
( ** Static members have no access to THIS or instance data and like    )
( ** static data members are shared between all instances of the owning )
( ** class and any derived classes.                                     )

  here ['] scrash ,
  TYPE_STATICCODE swap 0 ciao-token GetCurrentList +List
  post-def
;

: virtual-meth: \ "name" --
( *G The action invoked by METH: when the VIRTUAL modifier is in force. )
( ** Virtual methods can be given a code definition for a class and     )
( ** later modified in a derived class.                                 )

  1 CurrentClass class_#vtable dup @ >r +!
  TYPE_VIRTUALCODE r> 0 ciao-token GetCurrentList +List
  post-def
;

: instance-meth:	\ "name" --
( *G The default action of METH: creates a method associated with that  )
( ** class.                                                             )

  here ['] icrash ,
  TYPE_CODE swap 0 ciao-token GetCurrentList +List
  post-def
;

: meth:         \ "name" --
( *G Create a code member method. Dispatches to one of the above      )
( ** definitions depending on any applied modifiers.                    )
  peek-token ?validname
  DefFlags DEF_STATIC  and if static-meth: exit then
  DefFlags DEF_VIRTUAL and if virtual-meth: exit then
  instance-meth:
;


\ =======================================================================
( **                                                                    )
( *S Class Method Prototyping                                           )
( **                                                                    )
( *P These routines are used within a CLASS or STRUCT{ definition to    )
( ** define members which are in turn classes.                          )
( **                                                                    )
\ =======================================================================

: inst:         \ *class "name" --
( *G Embed an instance of the supplied class under the given "name".    )

  peek-token ?validname

  dup class_sizeidata @                 \ *class *req-size --
  CurrentClass class_sizeidata dup @    \ *class *req-size *sz idx --
  -rot +!                               \ *class idx
  TYPE_CLASS -rot ciao-token
  GetCurrentList +List

  post-def
;

: iptr:         \ *class "name" --
( *G Create a typed pointer for the given class inside the current one. )
( ** NOT IMPLEMENTED YET!                                               )

  peek-token ?validname
  post-def
  true abort" IPTR: not implemented."
;


\ =======================================================================
( **                                                                    )
( *S Operator Association                                               )
( **                                                                    )
( *P This code is used to associate an operator with a given method in  )
( ** a class or structure.                                              )
( **                                                                    )
\ =======================================================================

: FindClassOperator     \ op# *class -- *list-entry true | false
( *G Given an operator ID and a class pointer attempt to locate the     )
( ** method list entry associated with that id.                         )

  0 >r
  class_opchain @
  begin
    ?dup
  while
    2dup                \ op# p op# p --
    oplist_op# @ = if r> drop dup oplist_list @ >r then
    oplist_link @
  repeat
  0=                    \ if I asked for a numbered operator
  if
    r> ?dup             \ either pass or fail depending on find
  else
    r@ 0= abort" Failed to find operator for class"
    r> true
  then
;

: AddOperatorToClass    \ op# *list-entry *class --
( *G Routine used to bind a method-list entry to an operator id for     )
( ** the given class.                                                   )
  class_opchain +OPLIST
;

: oper:         \ "name" --
( *G Attempt to assign the method "name" as the action of the currently )
( ** active operator in the current class.                              )

  OperatorType @off
  ciao-token CurrentClass FindMethodInClass -1 =
  abort" Method not in class for OPER:"
  \ op# *list-entry --
  CurrentClass AddOperatorToClass
;


\ =======================================================================
( **                                                                    )
( *S CLASS Definition                                                   )
( **                                                                    )
( *P Code to create CLASS definitions.                                  )
( **                                                                    )
\ =======================================================================

: derived?      \ "text" -- true | "" -- false
( *G A look-ahead parsing definition used as a factor in CLASS to see   )
( ** if the class name is followed by a " : [<scope>] <name> " string   )
( ** for defining derived classes.                                    )
  peek-token S" :" icompare 0= dup if drop-token then
;

: derived-scope \ c-addr u -- scope flag
( *G Another parsing definition used by CLASS, after passing the        )
( ** DERIVED? test the next token is checked for a scope setting. If    )
( ** the next token is one of "public, protected or private" then the   )
( ** scope id is returned and a true flag indicating the next token     )
( ** has been used, otherwise SCOPE_PUBLIC is assumed and a false return)
( ** flag tells CLASS that this token is the actual base class name.    )

\ flg = TRUE if token was valid
  2dup S" public"    icompare 0= if 2drop SCOPE_PUBLIC    true exit then
  2dup S" protected" icompare 0= if 2drop SCOPE_PROTECTED true exit then
       S" private"   icompare 0= if       SCOPE_PRIVATE   true exit then
  SCOPE_PUBLIC false
;

: class         \ "name [ : [ <scope> ] <super> ]" -- ; Exec: -- ptr
( *G Begin a new class definition called NAME. If the new class is      )
( ** derived from a base class, the method lists are copied depending   )
( ** on C++ scoping rules, as is the Instance data size, operator chain )
( ** and vtable size.                                                   )
( ** When a CLASS definition is invoked it can do one of two things:    )
( ** If invoked within another CLASS definition it will invoke the class)
( ** member instance creation See previous section "Class Method       )
( ** Prototyping", at any other time a pointer to the class structure  )
( ** is returned.                                                       )

  peek-token ?validname
  create immediate
        here to CurrentClass
        MAGIC_CLASS     ,
        0               ,
        0               ,
        0               ,
        0               ,
        0               ,
        0               ,
        0               ,
        0               ,
        derived?
        if
                peek-token derived-scope if drop-token then
                ciao-token FindClass dup >r
                CurrentClass class_super !
                \ derived-scope -- ; R: superclass --

                \ Based on scope, copy the list pointers from superclass
                \ to self.

                case
                  SCOPE_PUBLIC          of

                                        r@ class_protected @
                                        CurrentClass class_protected !

                                        r@ class_public @
                                        CurrentClass class_public !

                                        endof

                  SCOPE_PROTECTED       of

                                        r@ class_protected @
                                        CurrentClass class_protected !

                                        r@ class_public @
                                        CurrentClass class_protected !

                                        endof

                  SCOPE_PRIVATE         of

                                        r@ class_protected @
                                        CurrentClass class_private !

                                        r@ class_public @
                                        CurrentClass class_private !

                                        endof

                  drop true abort" Internal Error #1"
                end-case

                \ Copy IDATA size
                r@ class_sizeidata @ CurrentClass class_sizeidata !

                \ Copy Operator CHain
                r@ class_opchain @ CurrentClass class_opchain !

                \ Copy VTable Index (size)
                r> class_#vtable @ CurrentClass class_#vtable !

        then

        \ set default scope

        private:

  does>

        CurrentClass if                 \ if defining a class when invoked then
                                        \ don't return address but perform the
                                        \ nested instance creation. Look for a
                                        \ * character as a separate token
                                        \ to indicate a pointer.

                peek-token
                S" *" icompare 0=
                if
                        drop-token iptr:
                else
                        inst:
                then


        then
;

: end-class     \ --
( *G Finish the definition of the current CLASS. If this is a derived   )
( ** class, the vtable from the parent is copied. After that any new    )
( ** vtable entries have the default crash vector attached.             )

  \ Patch current class vtable pointer
  here CurrentClass class_pvtable !

  \ Copy VTable from SuperClass (if there is one)
  CurrentClass class_super @
  ?dup if
        dup class_pvtable @
        swap class_#vtable @            \ ptr count --
        tuck cells                      \ count ptr size --
        bounds ?do
          i @ ,
        cell +loop                      \ count --
  else
        0
  then

  \ Fill the rest of the vtable with the default crash vector

  CurrentClass class_#vtable @ swap -   \ #todo --
  0 ?do ['] vcrash , loop

  \ Unset current class pointer (definition of class is now complete)
  0 to CurrentClass
;


\ =======================================================================
( **                                                                    )
( *S STRUCTures - A new slant on CLASS                                  )
( **                                                                    )
( *P Under CIAO like C++ a structure is a class. The only technical   )
( ** difference is that by default, members of a STRUCT{ are public.    )
( **                                                                    )
( *P STRUCT{ is provided for more syntactic reasons than technical. It  )
( ** is expected that STRUCT{ is used to declare structures consisting  )
( ** entirely of public data mapped in a contiguous manner so it can be )
( ** be used with operating system supplied structure pointers.         )
( **                                                                    )
\ =======================================================================

: struct{       \ "name [ : <scope> <super>" -- ; Exec: -- ptr
( *G Begin a new STRUCTure definition. Used in the same fashion as      )
( ** CLASS.                                                             )
  class  public:
;

: }		\ --
( *G Terminate a STRUCTure definition. See also END-CLASS.              )
  CurrentClass 0= abort" Invalid use of }"
  end-class
;


\ =======================================================================
( **                                                                    )
( *S Colon and SemiColon Override                                       )
( **                                                                    )
( *P CIAO overrides the standard Forth : and ; definitions to allow     )
( ** for method definitions. After a class has been defined it is       )
( ** necessary to write the actual code for any methods prototyped      )
( ** within it. CIAO like C++ takes a method name as being in the form  )
( *E <class>::<method>                                                  )
( **                                                                    )
\ =======================================================================

: ciao-colon    \ "name" -- | "name" -- "name"
( *G The new action of : when CIAO is installed. Pre-parses the name    )
( ** to see if it contains a double colon. If not then the original     )
( ** Forth : is called. If a double colon is found the assumption is    )
( ** that this definition is a method. The first portion before the ::)
( ** is taken to be a class name and the second portion the member name.)
( ** The class is looked up and its *class pointer stored in a global, )
( ** then the method name is looked up within that class and it's       )
( ** method list entry is also stored. Compilation is then triggered by )
( ** :NONAME and the XT of this definition kept for use in ;            )

  #0 to CurrentDefClass
  #0 to CurrentDefXT
  #0 to CurrentDefList

  peek-token S" ::" search not if 2drop : exit then

  nip >r
  ciao-token            \ c-addr u -- ; R: offset:: --
  2dup r@ -             \ c-addr u class-addr class-u -- ; R: offset:: --
  r> 2 - >r
  2swap                 \ class-addr class-u c-addr u -- ; R: len --
  r@ - + r>             \ class-addr class-u method-addr method-u --
  2swap

  \ method-addr method-u class-addr class-u --

  FindClass dup to CurrentDefClass
  FindMethodInClass -1 = abort" Failed to find method in class#1"
  to CurrentDefList

  :noname to CurrentDefXT
;

: ciao-semicolon        \ --
( *G The CIAO over-riding action of ; to handle the closing of method   )
( ** definitions. If the current definition was not a method-def then   )
( ** only the original ; is invoked. Otherwise ; is invoked and the     )
( ** XT of the method is patched into the class structure depending     )
( ** on the method type.                                                )
  postpone ;
  CurrentDefXT if
    case  CurrentDefList list_type @
      TYPE_CODE of
        CurrentDefXT CurrentDefList list_param1 @ !
      endof
      TYPE_STATICCODE of
        CurrentDefXT CurrentDefList list_param1 @ !
      endof
      TYPE_VIRTUALCODE of
        CurrentDefXT
        CurrentDefList list_param1 @ cells
        CurrentDefClass class_pvtable @ +
        !
      endof
          drop true abort" Impossible condition in modified ;"
    end-case
    0 to CurrentDefXT
  then
; immediate

: :             \ "name" --
( *G The actual overload of Forth's :                                   )
  ciao-colon
;

: ;             \ --
( *G The actual overload of the base Forth ;                            )
  postpone ciao-semicolon
; immediate


\ =======================================================================
( **                                                                    )
( *S OOP Compiler/Interpreter Extension Core Part 1 - EVALUATE BUFFER   )
( **                                                                    )
( *P This code is used by the method compiler extension in the Forth    )
( ** interpreter loop. It is a number of definitions used to            )
( ** create strings to pass to EVALUATE.                                )
( **                                                                    )
\ =======================================================================

1024 buffer: CompileBuffer
( *G The buffer used to build strings for EVALUATE.                     )

: ResetCompileBuffer    \ --
( *G Resets the COMPILEBUFFER for a new string.                         )
  0 CompileBuffer c!
;

: ciao-evaluate \ c-addr u --
( *G All EVALUATEs for CIAO come through here.                          )
  evaluate
;

: EvaluateCompileBuffer \ --
( *G Pass the COMPILEBUFFER string to CIAO-EVALUATE.                    )
  CompileBuffer count ciao-evaluate
;

: $+RCB         \ c-addr u --
( *G Append the string in C-ADDR U to the COMPILEBUFFER.                )
  dup CompileBuffer dup count + >r      \ c-addr u u cb -- ; R: dp --
  +!
  r> swap move
;

: n+RCB         \ n --
( *G Append the ascii representation of the number N to the             )
( ** COMPILEBUFFER. The string added is in the form "$<hex> "           )
  S" $" $+RCB
  base @ >r hex s>d <# #S #> r> base ! $+RCB
  S"  " $+RCB
;

\ =======================================================================
( **                                                                    )
( *S OOP Compiler/Interpreter Extension Core Part 2 - Method Compile    )
( **                                                                    )
( *P These definitions handle the compilation of method-list entries    )
( ** depending on type.                                                 )
( **                                                                    )
\ =======================================================================

: CompileMethod_DATA            \ *list-entry --
( *G Compiles code for an Instance Data member. Generates a pointer by  )
( ** laying code for "THIS <offset> +".                                 )
  S" this " $+RCB
  list_param1 @ n+RCB
  S" +" $+RCB
;

: CompileMethod_STATICDATA      \ *list-entry --
( *G Compiles code for a static data member. This is simply a literal   )
( ** address of the global space.                                       )
  list_param1 @ n+RCB
;

: CompileMethod_CODE            \ *list-entry --
( *G Compiles code for an instance code member. If the member has       )
( ** already been bound to a definition then the definition XT is compiled )
( ** along with execute i.e. " $<XT> execute " is compiled otherwise a  )
( ** pointer to where the XT will be stored is compiled with a @ execute. )
  dup list_param1 @ @ ['] icrash <> if
    list_param1 @ @ n+RCB
    S" execute" $+RCB
  else
    list_param1 @ n+RCB
    S" @ execute" $+RCB
  then
;

: CompileMethod_STATICCODE      \ *list-entry --
( *G Compiles code for a static code member. This is a literal address of )
( ** where the XT will be and a fetch-execute.                          )
  list_param1 @ n+RCB
  S" @ execute" $+RCB
;

: CompileMethod_VIRTUALCODE     \ *list-entry --
( *G Compiles code for a virtual method. It compiles the following code. )
( *E $<virtual-method-index>            \ the index into the vtable     )
( ** cells                              \ convert to vtable offset      )
( ** this                               \ Get current instance pointer  )
( ** cell- @                            \ Fetch objects vtable pointer  )
( ** + @                                \ extract XT from vtable        )
( ** execute                            \ and run it                    )
  list_param1 @ n+RCB
  S" cells this cell- @ + @ execute" $+RCB
;

: CompileMethod_CLASS           \ *list-entry --
( *G Compile code for a class instance member. This is almost identical )
( ** to instance data.                                                  )
  S" this " $+RCB
  list_param2 @ n+RCB
  S" +" $+RCB
;

: CompileMethod_CLASSPTR        \ *list-entry --
  S" this " $+RCB
  list_param2 @ n+RCB
  S" +" $+RCB
;

: OperatorProcess               \ *class --
( *G Compile code for any current operatortype.                         )
  ResetCompileBuffer
  OperatorType @off swap FindClassOperator if
    state @
    if  S" currobj @ >r" ciao-evaluate  then
    case  dup list_type @
      TYPE_CODE of
        S" >this " $+RCB  CompileMethod_CODE
      endof
        2drop TRUE abort" Impossible operator type."
    end-case
    EvaluateCompileBuffer
    state @
    if  S" r> currobj !" ciao-evaluate  then
  then
;

: MethodTokenCompileFromList    \ *entry --
( *G The global factor for this section. Given a method-list-entry it   )
( ** will dispatch to one of the COMPILEMETHOD_xxx definitions depending )
( ** on type.								)
  0 >r
  ResetCompileBuffer
  dup >r list_type @
  case
    TYPE_DATA of        r> CompileMethod_DATA           endof
    TYPE_STATICDATA of  r> CompileMethod_STATICDATA     endof
    TYPE_CODE of        r> CompileMethod_CODE           endof
    TYPE_STATICCODE of  r> CompileMethod_STATICCODE     endof
    TYPE_VIRTUALCODE of r> CompileMethod_VIRTUALCODE    endof
    TYPE_CLASS of       r@ CompileMethod_CLASS
                        r> r> drop list_param1 @ >r     endof
    TYPE_CLASSPTR of    r> CompileMethod_CLASSPTR       endof
    drop true abort" Internal Error #234"
  end-case
  EvaluateCompileBuffer
  r> ?dup if OperatorProcess then
;


\ =======================================================================
( **                                                                    )
( *S OOP Compiler/Interpreter Extension Core Part 3 - Single Token Check)
( **                                                                    )
( *P The single-token check is used at the beginning of the Forths token)
( ** interpret before the normal FIND. This is used to provide some     )
( ** special overrides to single Forth tokens. Namely:                  )
( **                                                                    )
( *P 1. If defining a code method, other members of the current class   )
( ** can be invoked simply by name. Therefore when defining a method any)
( ** single token needs to be looked up in the method table before      )
( ** checking the normal Forth dictionary.                              )
( **                                                                    )
( *P 2. A token can be preceeded by :: which enforces that the name is  )
( ** searched for in global name space the Forth dictionary regardless.)
( ** This is usually used to get you out of rule #1. If for instance you)
( ** are defining a method for a class which has a member called DUMP,  )
( ** simply entering "DUMP" will compile a reference to that member, if )
( ** you actually want to use the Forth DUMP you would type "::DUMP"    )
( **                                                                    )
( *P 3. A token can consist of a class-name and method name separated   )
( ** by a double-colon. NOT IMPLEMENTED YET!. This should compile a     )
( ** reference to a STATIC member of a class.                           )
( **                                                                    )
\ =======================================================================

\ Can be:
\       A scope override        <scope>::<name>
\       A global scope override ::<name>
\ if compiling a method
\ can also be:
\       A method name           <method-name>

: single-token  \ c-addr u -- flag
( *G Does the single-token-check operation and returns TRUE if the      )
( ** token has been processed. If the token should be passed on to      )
( ** the normal Forth lookup FALSE is returned.                         )

  \ If I'm defining a method then search current class for token
  CurrentDefXT if
    2dup CurrentDefClass FindMethodInClass
    -1 <> if
      nip nip
      MethodTokenCompileFromList
      true exit
    then
  then

  \ Check for global scope-override (can I find token without :: on front?)
  dup 2 > if
    over S" ::" tuck compare 0= if
      2>r
      get-order 2r@ 2 /string nSearch-WordList
      ?dup if
        0< state @ and
        if compile-it else execute then	\ SFP004
\        if compile, else execute then
        2r> 2drop true exit
      then
      2r>
    then
  then

  \ Look for named scope override
  2dup S" ::" search nip nip if
    2drop           \ chuck string away, I'm not doing this yet
    abort" Cannot handle scope overrides yet"
  then

  \ Oh well, a single token with nothing special about it, just pass it on
  2drop 0
;


\ =======================================================================
( **                                                                    )
( *S OOP Compiler/Interpreter Extension Core Part 4 - Compounds         )
( **                                                                    )
\ =======================================================================

: Process1stToken       \ c-addr u -- ap 0 | code-to-throw
( *G Used to handle the first part of a dot-notation compound. The      )
( ** first 'token' needs to ultimately lay the code generate THIS and   )
( ** needs to locate the correct class structure pointer that begins    )
( ** this invocation called the address-provider.                     )
( ** How the first token is actually translated depends on a number     )
( ** of rules:                                                          )
( **                                                                    )
( *P 1. If defining a method, the first token may be a class or class   )
( ** pointer member of the current class.                               )
( **                                                                    )
( *P 2. If the first token is surrounded by brackets it's a typecast.   )
( ** the name in brackets identifies the address provider whilst the    )
( ** instance pointer is assumed to already be top of the data stack.   )
( **                                                                    )
( *P 3. Next it could be that the first token is a named scope override.)
( ** as in rule 3 for single-token This behaviour is NOT IMPLEMENTED YET! )
( ** This is generally used for a virtual method in a derived class to  )
( ** invoke the action of the virtual method associated with it's parent.)
( **                                                                    )
( *P 4. Finally the token can either be a name in global space or a     )
( ** name defined as a LOCAL.                                           )

  ResetCompileBuffer

  CurrentDefXT if
    2dup CurrentDefClass FindMethodInClass -1 <> if
      \ found method in class of current method under definition
      nip nip         \ *list-entry --
      \ if the method entry is NOT a class or class pointer then bail
      dup  list_type @
      case
        TYPE_CLASS    of list_param2 @ n+RCB S" this + >this" $+RCB   endof
        TYPE_CLASSPTR of list_param2 @ n+RCB S" this + @ >this" $+RCB endof
          2drop
	  true abort" 1st token in local compound is not a class/cptr"
      end-case
      EvaluateCompileBuffer
      CurrentDefClass 0 exit
    then
  then

  2dup bracketed? if
    1 /string 1- FindClass 0            \ address-provider 0 --
    state @ 0=
    if  rot n+RCB  then
    S" >this" $+RCB
    EvaluateCompileBuffer
    exit
  then

  2dup S" ::" search if
    \ c-addr u f-addr fu --
    nip over <>                  \ c-addr u at-start?
    abort" Cannot use named scope override for 1st token"
    2 /string
  else
    2drop
  then

  2>r get-order 2r> nSearch-WordList 0=
\ if
\ ShowErrorLine ShowSourceOnError true
  abort" Failed to locate global instance."
\ endif

  \ instance-xt --

  dup here > if                 \ a "local" name
    >body dup cell+ @       	\ *local ap --
    0 rot                   	\ ap 0 *local --
    S" [ " $+RCB @ n+RCB S" s_lvarray, ] >this" $+RCB EvaluateCompileBuffer
  else
    >body dup @ 0 rot     	\ ap 0 pfa
    2 cells + n+RCB S" >this" $+RCB EvaluateCompileBuffer
  then
;

: ciao-hook     \ c-addr u -- flag
( *G This code receives a token that has fallen through the Forth       )
( ** FIND and NUMBER? cycle. If the string does not contain a dot       )
( ** separator it is not a compound statement and false is returned.    )
( ** Otherwise the string is split into a heap allocated token buffer   )
( ** using dot as the delimiter and the following steps taken:          )
( **                                                                    )
( *P 1. If in compile state code is laid to preserve the current THIS. )
( **                                                                    )
( *P 2. The first token is passed to PROCESS1STTOKEN above to obtain    )
( ** the instance pointer and address provider.                         )
( **                                                                    )
( *P 3. NOT IMPLEMENTED YET. The middle tokens should be processed.     )
( ** any of these tokens MUST represent a Class instance or class       )
( ** instance pointer as a member of the current address-provider. Each )
( ** middle token should compile/execute code to modify THIS and the    )
( ** address provider each time.                                        )
( **                                                                    )
( *P 4. The Final token is processed according to special rules. It must )
( ** exist in the name space of the current address provider. If we are )
( ** defining a method for the current address provider all three scope )
( ** lists are valid, otherwise the method must be in PUBLIC name space. )
( ** If successfully located according to scope rules the method entry  )
( ** is passed to METHODTOKENCOMPILEFROMLIST see earlier to compile   )
( ** the invokation code for that method and any applicable operator. )
( **                                                                    )
( *P 5. The code to restore the saved value of THIS is compiled.        )

  0 0 0 0 locals| *token-basemem *token-mem #tokens address-provider |

  2dup S" ." search nip nip 0=		\ NOT A COMPOUND?
  if  2drop 0 exit  then

  \ It's a compound (oh shit!)
  \
  \ There are 3 parts to the compound.
  \     1)      The first token
  \     2)      The middle token(s)
  \     3)      The last token.
  \
  \ The first token is the difficult one, it can be:
  \
  \     1)      (If compiling a method) It can be a class member or
  \             a class pointer member.
  \
  \     2)      It can be a typecast for TOS
  \
  \     3)      It can be a global instance name
  \
  \     4)      It can be a global instance name with global scope override
  \
  \     NB/ I am not support named scope overrides for compound statements!!
  \
  \     For case 1:     treat as a middle token. Self is the address provider
  \                     and holds the new THIS
  \     For case 2:     The typecast name gives the address provider and TOS
  \                     will become the new THIS
  \     For case 3:     This and the address provider come from the global.
  \     For case 4:     (as case 3)
  \
  \ Middle tokens are allways either classes or class pointers which are
  \ members of the current address provider.
  \
  \ The Last token is treated as single method token (see above)
  \

  \ Tokenise the input stream

  dup 1 chars + allocate abort" Failed to allocate parsememory for compound"
  dup to *token-basemem to *token-mem 1 to #tokens
  *token-mem over 1 chars + erase

  *token-mem *token-mem 1 chars + 2swap         \ base cp c-addr u --
  bounds
  ?do
    i c@ dup [char] . = if
      drop                      \ lose character
      nip dup 1 chars +         \ pointer becomes new base, new ptr is base+1
      #tokens 1+ to #tokens     \ inc num tokens
    else
      over c!                   \ store character
      1 chars +                 \ inc pointer
      over 1 swap c+!           \ inc count byte
    then
  loop
  2drop

  \ Okay, #tokens is the number of tokens in the stream and *token-memory
  \ points to an array of counted strings (no alignment) for the tokens in
  \ order.

  state @ if  S" currobj @ >r" ciao-evaluate  then

  \ TOKEN#1 Processing:
  *token-mem count Process1stToken dup ?throw
  to address-provider
  *token-mem count + to *token-mem

  \ MIDDLE TOKEN Processing:
  #tokens 2 - 0 ?do
          true abort" not processing middle tokens"
  loop

  \ LAST TOKEN Processing:
  *token-mem count address-provider FindMethodInClass dup -1 =
  abort" Failed to find method in class#3"
  CurrentDefXT if drop TRUE else SCOPE_PUBLIC = then
  not abort" method out of scope"
  MethodTokenCompileFromList

  state @
  if  S" r> currobj !" ciao-evaluate  then

  \ delete the token memory
  *token-basemem free drop

  true

  OperatorType @off abort" ISE: Unprocessed operator in CIAO-HOOK"
;


\ *********************************
\ *S Installing CIAO into VFX Forth
\ *********************************
\ *P CIO can be installed into VFX v4 or v5 onwards.

\ ===========
\ *N VFX v4.x
\ ===========
\ THIS CODE BEHAVES IN THE SAME MANNER AS ME MOVING CLASSHOOK TO THE
\ END OF THE INTERPRETER LOOP

VFX5? 0= [if]

: ciao-classhook	\ caddr len -- flag
\ *G The action of *\fo{CLASSHOOK} when CIAO is installed.
  depth >r  single-token  r> depth <> abort" damn"
;

: ciao-undefined	\ caddr len --
\ *G The action of *\fo{UNDEFINED} when CIAO is installed.
  ciao-hook 0=
  if  undefined  then
;

: +ciao		\ --
\ *G Install CIAO's system hooks.
  assign ciao-classhook to-do classhook
  assign ciao-undefined to-do cundefined
;
+ciao

: -ciao		\ --
\ *G Uninstall CIAO's system hooks.
  assign (classhook) to-do classhook
  assign undefined to-do cundefined
;

[then]


\ ===========
\ *N VFX v5.1 onwards
\ ===========

VFX5? [if]

errdef err_postCiao  "Cannot POSTPONE aa.bb.cc for CIAO"

: postCiaoText	\ --
  err_postCiao throw
;

' noop  ' noop  ' postCiaoText  RecType: r:CiaoHook	\ -- struct
\ *G Contains the three actions for dot parsers.

: rec-CiaoHook	\ caddr u -- r:CiaoHook | r:fail
\ *G The parser part of the floating point recogniser.
\ cr ." r-ch " 2dup type  .s
  single-token if  r:CiaoHook  else  r:fail  then
\ .s
;

' noop  ' noop  ' postCiaoText  RecType: r:CiaoUndef	\ -- struct
\ *G Contains the three actions for dot parsers.

: rec-CiaoUndef	\ caddr u -- r:CiaoUndef | r:fail
\ *G The parser part of the floating point recogniser.
\ cr ." r-cu "  2dup type  .s
  ciao-hook if  r:CiaoUndef  else  r:fail  then
\ .s
;

: +ciao		\ --
\ *G Switch on the CIAO parser.
  ['] rec-CiaoHook forth-recognizer +stack
  ['] rec-CiaoUndef forth-recognizer +stack-bot
;
+ciao

: -ciao		\ --
\ *G Switch off the CIAO parser.
  ['] rec-CiaoHook forth-recognizer -stack
  ['] rec-CiaoUndef forth-recognizer -stack
;

[then]


\ =======================================================================
( **                                                                    )
( *S Instance Creation Primitives                                       )
( **                                                                    )
\ =======================================================================

-sin

: dynamicnew	\ *class -- this
( *G The runtime code called for instances created via DNEW. Allocate   )
( ** a block of heap memory large enough for the 2 control cells        )
( ** *class pointer and vtable pointer followed by the instance data. )
( ** The THIS pointer returned is the beginning of the instance data.   )
  dup class_sizeidata @ 2 cells + allocate abort" Failed new() allocation"
  >r
  dup r@ !
  class_pvtable @ r@ 1 cells + !
  r> 2 cells +
;

: staticnew     \ *class "name" -- ; Exec-child: -- this
( *G Create a new instance of class called "name" in the dictionary.    )
( ** The instance is a child of CREATE which has a body containing      )
( ** the *CLASS value, a pointer to the instance vtable then the        )
( ** instance data. At runtime a THIS instance pointer is returned, this)
( ** is 2 cells on from the PFA IE past the vtable.                   )
( ** Forms the action of NEW when invoked outside of a : definition.    )
  create immediate
    dup ,
    dup class_pvtable @ ,
    class_sizeidata @ allot&erase
  does>
    dup 2 cells +
    state @
    if  postpone literal  else  swap  then
    @ OperatorProcess
;

: localnew2     \ *class frame-add --
( *G I apologise unreservedly for this trick. This definition performs  )
( ** the compile-time tail of local-new. Since LOCALNEW performs a      )
( ** create..does> I cannot add any compile time tail so I tick this    )
( ** definition and push it on the return stack! Sorry ;-          )
  >real-dict
  over class_pvtable @ over cell+       \ *class frame-add *vtable fadd[1] --
  swap  postpone literal  postpone ->  s_lvarg,        \ lv!,
  swap  postpone literal  postpone ->  s_lvarg,        \ lv!,
;

: localnew      \ *class "name" --
( *G This definition forms the action of NEW when making a local method.)
( ** It hacks into VFX locals to create a new entry on the local frame  )
( ** and lays the code necessary in LOCALNEW2 to setup the two        )
( ** control cells AT RUNTIME. The net result is a very fast and useable)
( ** named local instance. Implementors beware, this is the most system )
( ** specified piece of code imaginable.                                )
  ?init-locals
  >temp-dict
  ['] localnew2 >r
  create immediate
    dup class_sizeidata @ 2 cells + fradjust           \ *class ptr --
    dup 2 cells + ,
    over ,
  does>
    dup @ s_lvarray,
    cell + @ OperatorProcess
; immediate

+sin


\ =======================================================================
( **                                                                    )
( *S Instance Creation                                                  )
( **                                                                    )
\ =======================================================================

-sin
: dnew          \ *class -- this
( *G State smart definition to create a new class instance on the heap. )
( ** returns a pointer to the instance which can be stored.             )
( ** A heap allocated instance pointer can be held for as long as       )
( ** required and does not go "out of scope" until explicitly removed   )
( ** with DELETE. Use this type to create a dynamic instance which you  )
( ** can safely return from a method/colon definition. Note that unlike )
( ** in C++ you must use DNEW in CIAO for heap allocation.		)
  state @ if
    postpone literal postpone dynamicnew
  else
    dynamicnew
  then
; immediate

: new           \ *class "name" --
( *G A state smart definition to create a named instance of a class.    )
( ** When used within a : definition the object is created in a locals  )
( ** frame one is created if required. When invoked outside of a      )
( ** definition the instance space is ALLOTed from the dictionary.      )
( ** The instance goes out of scope and is implicitly DELETEd when     )
( ** local at the same time as the name goes out of scope. For a static)
( ** instance, i.e. one in the dictionary, it remains in scope for the  )
( ** lifetime of the application until BYE whereas for a local        )
( ** instance it is DELETEd and goes out of scope at the end of the     )
( ** definition. Therefore please note that returning a pointer to a    )
( ** local instance *will* break your code - you must DNEW instead. You cannot )
( ** ever explicitly DELETE a named instance. In future you will be     )
( ** allowed to attempt it and the effect will be to call the destructor)
( ** method, as will happen when scope is lost anyway.                  )
  state @ if
    postpone localnew
  else
    staticnew
  then
; immediate

: delete        \ this --
( *G DELETE is used to release the memory of a dynamic instance created )
( ** via DNEW. Memory release for static and local instances is         )
( ** automagic. Any valid destructor is called prior to releasing the   )
( ** memory back to the free-heap. In future performing DELETE on a     )
( ** static or local will simply invoke the destructor.                 )
  dup class-base-mem @ here within
  abort" Cannot perform delete operation on static class."
  2 cells -
  free drop
;
+sin

\ =======================================================================

here class-base-mem !

\ =======================================================================
( **                                                                    )
( *S AutoVar - An example of a Class                                    )
( **                                                                    )
( *P This example shows how to create and use a class called AUTOVAR.   )
( ** This class contains one private data member and two public methods.)
( ** one method initialises the data store, the other will return the   )
( ** contents of that store and post-increment it.                      )
( **                                                                    )
( *H Defining the class                                                 )
( **                                                                    )
( *E class AutoVar           \ begin a new class definition             )
( ** private:                                                           )
( **         cell:   m_data  \ Where the count will be stored           )
( ** public:                                                            )
( **         meth:   read++  \ The method to read and increment         )
( **         meth:   set     \ The method to initialise the count       )
( ** end-class               \ end definition                           )
( **                                                                    )
( *H Coding the Methods                                                 )
( **                                                                    )
( *E : AutoVar::read++       \ -- n ; Method                            )
( **   1 m_data dup @ -rot +!        \ read and increment data store    )
( ** ;                                                                  )
( **                                                                    )
( ** : AutoVar::set          \ n -- ; Method                            )
( **   m_data !                      \ write to data store              )
( ** ;                                                                  )
( **                                                                    )
( *H Test Code                                                          )
( **                                                                    )
( *P Here are three test routines which each take an initial value      )
( ** for an AutoVar type and then run the read++ method 10 times        )
( ** writing the result. The first case uses a static instance of       )
( ** AutoVar, the second case uses a local instance, and finally the    )
( ** third case shows how to use a heap allocated instance and how      )
( ** to typecast an object pointer. Note that in CIAO there are         )
( ** separate words for static/local instances with NEW and heap        )
( ** allocation with DNEW.						)
( **                                                                    )
( *E AutoVar new Foo         \ create a static instance of Autovar      )
( **                                                                    )
( ** : test1                 \ n -- ; Test static instance FOO          )
( **   foo.set                       \ init with supplied index         )
( **   10 0 do                                                          )
( **     cr foo.read++ .             \ read and increment 10 times!     )
( **   loop                                                             )
( ** ;                                                                  )
( **                                                                    )
( ** : test2                 \ n -- ; Same thing, local class tho'      )
( **                                                                    )
( **   AutoVar new Foobar            \ create local instance of AUTOVAR )
( **                                                                    )
( **   foobar.set                                                       )
( **   10 0 do                                                          )
( **     cr foobar.read++ .                                             )
( **   loop                                                             )
( ** ;                                                                  )
( **                                                                    )
( ** : test3                 \ n -- ; Third time, heap allocated instance )
( **                                                                    )
( **   AutoVar dnew                  \ create instance and store pointer)
( **                                                                    )
\ **   tuck (AutoVar).set            \ use type cast on heap pointer    )
( **   10 0 do                                                          )
\ **     cr dup (AutoVar).read++ .                                      )
( **   loop                                                             )
( **                                                                    )
( **   delete                        \ destroy heap instance            )
( ** ;                                                                  )
( *S AutoVar2 - Another Example                                         )
( **                                                                    )
( *P AUTOVAR2 is a class derived from AUTOVAR to extend its             )
( ** functionality. AutoVar2 has no publically accessable methods but   )
( ** uses operators to perform the read and initialise.                 )
( **                                                                    )
( *E Class AutoVar2 : private Autovar   \ Create new class, inherit from)
( **                                    \ Autovar with all methods in   )
( **                                    \ private scope.                )
( **      oper: read++                  \ Default operator performs     )
( **                                    \ read++ method                 )
( **   to oper: set                     \ TO operator performs set method )
( ** end-class                                                          )
( **                                                                    )
( *P The test procedures could now look like:                           )
( **                                                                    )
( *E AutoVar2 new Foo         \ create a static instance                )
( **                                                                    )
( ** : test1                 \ n -- ; Test static instance FOO          )
( **   to Foo                        \ init with supplied index         )
( **   10 0 do                                                          )
( **     cr foo .                    \ read and increment 10 times!     )
( **   loop                                                             )
( ** ;                                                                  )
( **                                                                    )
( ** : test2                 \ n -- ; Same thing, local class tho'      )
( **                                                                    )
( **   AutoVar2 new Foobar            \ create local instance           )
( **                                                                    )
( **   to Foobar                                                        )
( **   10 0 do                                                          )
( **     cr foobar .                                                    )
( **   loop                                                             )
( ** ;                                                                  )

\ \ =======================================================================
\ ( **                                                                    )
\ ( *S That's All Folks                                                   )
\ ( **                                                                    )
\ ( *P Now you know more than you ever wanted to about this implementation)
\ ( ** More examples on creating and using classes can be found by looking)
\ ( ** at the class library sources.                                      )
\ ( **                                                                    )
\ \ =======================================================================

cr ." CIAO Implementation took "
   class-base-mem @ start-ciao @ - . ." bytes"
cr

\ -----------------------------------------------------------------------------

