\ Class.fth - intelligent objects

((
Copyright (c) 1999, 2000, 2005, 2008, 2010, 2011, 2018
Construction Computer Software Pty Ltd.
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441

net: tech-support@mpeforth.com
web: www.mpeforth.com

To do
=====

Change History
==============
20180508 SFP001 v5 and NDCS changes.
))

\ *T Class System Requirements

VARIABLE Debugclass Debugclass OFF
\ *G Variable to debug class compile

: +DEBUGCLASS	\ ---
\ *G Compile class and instance cfa's into image. Make class code easier
\ ** to debug,
  Debugclass ON  ;

: -DEBUGCLASS	\ --
\ *G Compile class and instance references as literals. Turn off class
\ ** code being easier to debug.
   Debugclass OFF  ;

[undefined] ptr [if]
cell field-type ptr
[then]

: OFFSET-ALIGN	( offset --- offset' )
\ Align towards +/- infinity
  DUP ABS ALIGNED SWAP ?NEGATE  ;

\ *S Byte operators
: cincr		\ addr --
  1 swap c+!  ;

: cdecr		\ addr --
  -1 swap c+!  ;

: coff         \ addr --
  0 swap c!  ;

: con		\ addr --
  -1 swap c!  ;

\ *S Word operators
: wincr		\ addr --
  1 swap w+!  ;

: wdecr		\ addr --
  -1 swap w+!  ;

: woff		\ addr --
  0 swap w!  ;

: won		\ addr --
  -1 swap w!  ;

\ *S Double Cell operators
: 2incr		\ addr -- ; increments a number in high:low order
  dup 2@ 1, d+ rot 2!  ;

: 2decr		\ addr -- ; decrements a number in high:low order
  dup 2@ 1, d- rot 2!  ;

: 2+!		\ d addr -- ; updates a number in high:low order
  >r  r@ 2@ d+  r> 2!  ;

: 2-!		\ d addr --
  >r  r@ 2@  2swap d-  r> 2!  ;

: 2off		\ addr --
  dup off  cell+ off  ;

: 2on		\ addr  --
  dup on  cell+ on  ;

\ *> DEV-OBJECTS
\ *T CLASS STRUCTURE
\ *S Data structures defined by TYPE:
\ *P Structure definitions, fields, objects and so on all use a common data
\ ** structure that is generated by the defining words.
\ ** Structures are associated with a word (the address provider) that can
\ ** provide the starting address of the structure implementation. By supplying
\ ** the cfa of NOOP, no address is provided, and so the structure is purely a
\ ** template.  For templates (address provider = 0 or NOOP), an offset may
\ ** also be defined. NOOP and 0 are the default address provider and
\ ** offset of templates.

\ *S TYPE: definitions
\ *C    header                          standard VFX layout
\ *C 0) jsr do_type                     5 bytes
\ *C 1) cfa of address provider         4 bytes
\ *C 2) initial offset                  4 bytes (0 for class)
\ *C 3) link to last field defined      4 bytes
\ *C 4) type size (final offset)        4 bytes
\ *C 5) Magic number                    4 bytes
\ *C 6) anchor of instance method chain 4 bytes
\ *C 7) anchor of type method chain     4 bytes
\ *C 8) link to previous type defined   4 bytes
\ *C 9) private wordlist                ? bytes


\ *S MAKE-INST definitions
\ *C    header                          standard VFX layout
\ *C 0) jsr do_inst                     5 bytes
\ *C 1) cfa of address provider         4 bytes
\ *C 2) offset from start of type       4 bytes
\ *C 3) link to last instance of type   4 bytes
\ *C 4) size of instance data           4 bytes
\ *C 5) 0                               4 bytes
\ *C 6) pointer to TYPE/CLASS           4 bytes
\ *C 7) data if static

  struct /type-head
       ptr provider                 \ 1) cfa of address provider         4 bytes
       int off-start                \ 2) start offset of 1st field 4 bytes......
                                    \    (normally 0)
       ptr field-link               \ 3) link to last field defined      4 bytes
       int type-size                \ 4) structure size (final offset)   4 bytes
       int type-magicID             \ 5) Magic ID for for a TYPE:        4 bytes
       ptr inst-chain               \ 6) link to last instance method    4 bytes
       ptr type-chain               \ 7) link to last class method       4 bytes
       int type-prev                \ 8) link to last structure defined  4 bytes
       0 field type-wordlist        \ 9) private wordlist for members    x bytes
  end-struct

\ *S INSTANCE STRUCTURE

  struct /inst-head
       5 cells +                    \ address provider, offset, field link, size
       ptr ^type    ***                     \ pointer to defining TYPE structure
       int inst-bpe ***                            \ instance bytes per elelment
  end-struct

\ *S METHOD CHAIN STRUCTURE

  struct /method        \ -- size ; methods for a type are linked by this chain
       int methodid                                         \ method indentifier
       int mlink                                       \ link to previous method
       int mflags                        \ flag, set if INST structure operation
       int mxt
  end-struct

\ *S Instance Primatives
: COMP-LOCAL-PROVIDER	( inst --- )
\ *G Compile local instance
   ?COMP  off-start @ s_lvarray,  ;

: ?COMPLIT	( x -- | x )
\ *G Compile literal if in compile state or leave on pstack if executing.
\ ** State smart literal.
  state @
  IF  postpone literal  THEN
\  IF [COMPILE] LITERAL THEN
;

: ?COMPLIT+	( n -- | n' )
\ *G Compile as LIT + if in compile state
  state @ IF
    ?DUP
    IF  postpone literal  postpone +  THEN
\    IF  ?COMPLIT COMPILE +  THEN
  ELSE
    +
  THEN
;

: XT-COMP?	( xt --- )
\ *G Compile xt if in compile state or execute
\ ** State smart xt. Checks immediate flag if compile state.
   { %xt --- }
    state @ IF
         %xt immediate?
         IF   %xt EXECUTE
         ELSE
           %xt compile-it
\           %xt compile,
         THEN
    ELSE %xt EXECUTE
    THEN
;

: ?COMP/EXEC	( xt --- )
\ *G Compile or execute depending on state
   { %xt --- }
   %xt ['] noop <>
   IF
        state @
        IF   %xt compile-it
\        IF   %xt compile,
        ELSE %xt execute
        THEN
   THEN
;

\ *S CLASS GLOBAL VARIABLES
\ *[
VARIABLE with-inst	\ -- addr ; hold overridden provider
VARIABLE last-inst      \ -- addr ; holds ctrl field of last inst
VARIABLE TypeFather     \ -- addr ; current typedef being defined
VARIABLE struct-chain   \ -- addr ; holds pointer to last structure defined
VARIABLE ptr-chain      \ -- addr ; holds pointer to last pointer defined

VARIABLE #extras #extras OFF
\ *G Holds number of extra bytes needed by arrays during compilation of a
\ ** TYPE: defintion.

VARIABLE #bpe #bpe OFF  \ -- addr ;
\ *G Holds number of bytes per element for arrays during compilation
\ ** of a TYPE: defintion.

VARIABLE <type/inst> <type/inst> OFF  \ -- addr
\ *G Holds 4 for TYPE/CLASS methods, 0 for INST

HERE CONSTANT TYPE:-ID	\ -- x ; magic number for a type
HERE CONSTANT INST:-ID  \ -- x ; magic number for an instance
\ *]

\ *N CLASS GLOBAL VARIABLE HELPER WORDS
: STRUCT-CHILDXT	\ addr -- addr' ; select type or instance chain
  inst-chain  <type/inst> @ +  ;

: STRUCTURE-METHOD      \ struct -- struct
\ *G Marks last method as operating on the structure itself
  DUP struct-childxt @ mflags ON  ;	\ set flags in last method

VARIABLE curinst	\ -- addr
\ *G Holds current instance being executed in does part of make-inst

: INST-ALLOT&ERASE	( inst --- )
\ *G Allot static instance data space and erase
   type-size @ ALLOT&ERASE  ;

: INST-ALLOT&BLANK      ( inst --- )
  type-size @ HERE OVER ALLOT SWAP BLANK  ;

: CLASS		\ --
\ *G Select class methods in a TYPE: definition.
  cell <type/inst> !  ;

: INST		\ --
\ *G Select instance methods in a TYPE: definition.
   0 <type/inst> !  ;

: #BYTES	\ n --
\ *G Set instance build "size" global variable in a TYPE: definition.
  0 max #extras !  ;

: #BYTES-PER-ELM	\ n --
\ *G Set instance build "bytes per element" global variable in a TYPE:
\ ** definition.
   #0 max #bpe !  ;

\ *S CLASS STRUCTURE ACCES DEFINITIONS

: INST>TYPE	( inst --- type )
\ *G Return the type of an instance
  ^type @  ;

: INST-SIZE	\ ^inst-head -- size
  INST>TYPE type-size @  ;

: TYPE-ID@	\ inst-head --- id
  type-magicid @  ;

: IS-TYPE?	\ addr -- flag
\ *G True if address is pointing to a TYPE: definition
  TYPE-ID@ TYPE:-ID =  ;

: IS-INST?	\ addr -- flag
\ *G True if address is pointing to an INST: definition.
  TYPE-ID@ INST:-ID =  ;


\ *S INSTANCE COMPILE TIME BUILD DEFINITIONS

: INST-LOCAL?	\ inst --- flag
   \ *G Return true if instance provider is for a local.
  [ ALSO SYSTEM ]
   provider @  ['] lv =
  [ PREVIOUS ]
;

((
\ \ *N Instance Run time debug code

:  XT->STATIC                                                     ( xt --- addr )
   \ Debug objects runtime code. Convert instance xt to static data addr.
   discard-sinline
   >body /inst-head +
;

: XT-SIZE@                                                       ( xt --- size )
   \ Debug objects runtime code. Convert instance xt to size.
   discard-sinline
   >body type-size @
;

: XT-BPE@                                                         ( xt --- bpe )
   \ Debug objects runtime code. Convert instance xt to bytes per element.
   discard-sinline
   >body inst-bpe @
;


: XT-PROVIDER+                                                   ( xt --- addr )
   \ Debug objects runtime code. Convert instance xt to address plus offset
   discard-sinline
   >body off-start @ +
;

: XT-OFFSET                                                    ( xt --- offset )
   \ Debug objects runtime code. Convert instance xt to offset
   discard-sinline
   >body off-start @
;
))

\ *S INSTANCE COMPILE TIME BUILD DEFINITIONS

: STATICPROVIDER	\ i*x inst -- j*y
\ Address provider for static data.
   /inst-head + ?complit  ;

: SIZEPROVIDER		\ i*x inst -- j*y
\ *G Address provider for size.
  type-size @ ?complit  ;		\ compile size as a literal

: INST-BPEPROVIDER	\ i*x inst -- j*y
\ *G Address provider for BPE.
   inst-bpe @ ?complit  ;

: OFFSETPROVIDER	\ i*x inst -- j*y
\ *G Address provider for offset.
  off-start @ ?COMPLIT  ;

: +PROVIDER             \ i*x inst -- j*y
\ *G Address provider for offset addition.
  off-start @ ?COMPLIT+  ;

: INST-PROVIDER		\ i*x inst -- j*y inst
\ *G Compile/execute instance provider code
   [ ALSO SYSTEM ]
   >R
   CASE
   R@ provider @                                      \ inspect address provider
   ['] lv    OF  R@ COMP-LOCAL-PROVIDER ENDOF                            \ local
   ['] +     OF  R@ +PROVIDER           ENDOF                    \ field in type
   0         OF  R@ STATICPROVIDER      ENDOF                      \ static data
   ['] noop  OF  ENDOF                                       \ no action to take
             XT-COMP?
   END-CASE
   R>
  [ PREVIOUS ]
;

: .provider	\ inst --
   cr  CASE  provider @                               \ inspect address provider
     ['] lv    OF  ." lv " ENDOF                            \ local
     ['] +     OF  ." +  "           ENDOF                    \ field in type
     0         OF  ." static "      ENDOF                      \ static data
     ['] noop  OF  ." noop "  ENDOF                          \ no action to take
             ." xt " .dword
   END-CASE
;

: INST-ADDR		\ i*x inst -- j*y
\ *G Compile/execute instance provider code
  INST-PROVIDER DROP  ;

: INST-PROVIDER-BPE	\ inst ---
\ Compile or execute addr and bytes per element of object code
  INST-PROVIDER inst-bpe @ ?COMPLIT  ;

: INST-COUNT	\ inst ---
\ Compile or execute count of object code
  INST-PROVIDER SIZEPROVIDER  ;		\ compile size as a literal

: INST-SIZE-ADDR  	\ inst --- typesize instaddr
\ Compile or execute size of object and addres  of instance
  INST-COUNT ASSIGN SWAP ?COMP/EXEC
;

: INST-COPY	\ inst ---
\ Compile or execute copy object code
  INST-COUNT  ASSIGN MOVEX ?COMP/EXEC  ;

: INST-BLANK	\ inst ---
\ Compile or execute blank object code
  INST-COUNT  #32 ?COMPLIT  ASSIGN FILL ?COMP/EXEC  ;

: INST-ERASE	\ inst ---
\ Compile or execute erase object code
  INST-COUNT  #0 ?COMPLIT  ASSIGN FILL ?COMP/EXEC  ;

: INST-ARRAY	\ inst ---
\ Compile or execute count and bytes per element of object code
  INST-PROVIDER
  DUP >R type-size @ ?COMPLIT
  R> inst-bpe @ ?complit
;

: EXECUTE-LAST-INST	( --- )
\ Execute last defined instance, last-inst global var must be set.
  last-inst @ BODY> EXECUTE  ;

\ *> DEV-OBJECTS
\ *T Object Service Definitions
\ *S General
\ *P Object Service Definitions provide a service to an object. The idea of an
\ ** specific object service definition is to avoid creating a new language
\ ** using the object methods. Use the FORTH language and apply some discipline
\ ** on parameter stack inputs and outputs. Document the service requirement
\ ** and object effect. Announce and guaratee the service. Using standard
\ ** naming conventions developers can use standard tools to get access to the
\ ** service definitions.
\ *P The stack input of the definition can only be one of the following three
\ ** formats to qualify as an object service definition.
\ *B 1. object  --- flag or object
\ *B 2. object> --- flag or void
\ *B 3. vobject> --- flag or void
\
\ *N Service definition type 1
\ *P The definition copies the object to own data space, performs the service by
\ ** operating on the data space and returns a result flag or modified object.
\
\ *N Service definition type 2
\ *P The definition performs the service by operating on the client data
\ ** space. A result flag or nothing is returned. The client can use the
\ ** modified object data space.
\
\ *N Service definition type 3
\ *P The definition performs the service by operating on the client virtual data
\ ** space. A result flag or nothing is returned. The client can use the
\ ** modified object data space. The service defintion does not destroy the
\ ** client virtual data space, the client must make and destroy the virtual
\ ** data space.
\
\ *N Input Parameterstack comments for Object Service Definitions
\ *B The stack comment must indicate the TYPE: instance.
\ *B Object Pointers must have a trailing ">".
\ *B Virtual object instances must have a leading "v" and a trailing ">".
\ *H The Elements of Simple Design
\ *C 1.Passes its tests
\ *C 2.Minimizes duplication
\ *C 3.Maximizes clarity
\ *C 4.Has fewer elements
\ *> DEV-OBJECTS
\ #2 VFX AND 220 SOURCE COMPATIBILITY CODE
\ ----------------------------------------
\ *S Objects
\ *C Public
\ *C Private
\ *C Virtual
\ *C Protected

\ *S COMPILE TIME ABORT MESSAGE

: COMP-$ABORT	( addr cnt --- )
\ *G Abort message for compile errors
   >ABORT$                                           \ fill abort message buffer
   LATEST-NFA COUNT >ABORT-FUNC$  S"  - " +ABORTFUNC$
   pocket count +ABORTFUNC$
   Abort-Code ON                                           \ set gbl abort value
   True THROW                                          \ return to abort frame
;

: ?TYPE		( addr --  )
\ *G Error if address not pointing to a TYPE: definition.
   is-type? 0= IF S" Not a TYPE: " COMP-$ABORT THEN
;

: ?INST		( addr --  )
\ *G Error if address is not pointing to an INST: definition.
   is-inst? 0= IF S" Not an INST: " COMP-$ABORT THEN
;

: OBJ-CRASH	( obj --- )
\ Abort on this method
   { %inst --- }
   CurrSourceName >ABORT-FNAME$
   LATEST-NFA COUNT >ABORT-FUNC$   S"  - " +ABORTFUNC$
   %inst BODY> XT>NFA COUNT +ABORTFUNC$ S"  : " +ABORTFUNC$
   S"  << " +ABORTFUNC$ Line# N>$ +ABORTFUNC$
   S"  - "  +ABORTFUNC$ Source -LEADING -TRAILING +ABORTFUNC$
   S"  >>" +ABORTFUNC$
   \
   %inst IS-INST?
   IF
        %inst INST>TYPE BODY> XT>NFA COUNT +ABORTFUNC$
   THEN
   S" This method not valid for this object." >ABORT$
   ACTION-OF ABORT ASSIGN (ABORT) =
   IF
        CONSOLE-DISP-ABORT
   THEN
   ABORT
;

\ *S EXECUTE INSTANCE METHOD CHAIN

: TYPE-METHOD-EXECUTE	( type> methodid --- flag )
\ *G Execute type method. Return true if method existed for type else
\ ** false.
   { %type> %method | %chain> --- }
\ cr ." T-M-E " %method .
   %type> inst-chain @ <to> %chain>             \ get instance method chain head
   BEGIN                                        \ -- INST METHOD-ID METHODSTRUCT
      %chain>
   WHILE                                       \ -- inst method-id methodstruct
      %method %chain> methodid @ =                         \ found method ids ??
      IF
           %chain> mxt @ execute
           True
           EXIT
      THEN
      %chain> mlink @ <to> %chain>
   REPEAT
   False
;

: EXECUTE-TYPE-METHOD	( inst type methodid --- flag )
\ Attempt to execute method for inst. Return true if successful.
   {: %inst %type %method | %chain> %debug? -- flag :}
\ cr ." E-T-M " %method .
   %type inst-chain @ <to> %chain>              \ get instance method chain head
   BEGIN                                        \ -- INST METHOD-ID METHODSTRUCT
      %chain>
   WHILE                                       \ -- inst method-id methodstruct
\ %chain> methodid @ .
      %method %chain> methodid @ =                         \ found method ids ??
      IF
\ cr %chain> mflags @ .
           %chain> mflags @                       \ instance structure-method ??
           IF \ yes!
                %inst INST-LOCAL?                            \ local instance ??
                IF   \ yes! debuging not required
                     Debugclass @OFF
                ELSE \ no! dict entry allow debug
                     Debugclass @
                THEN
                <to> %debug?
                %inst         \ always pass the instance thru a structure-method
                %chain> mxt @ execute           \ it's immediate, always execute
                %debug? Debugclass !
           ELSE \ no!
\ ." [ETM " .s   %inst .provider
                %inst inst-addr             \ generate the address provider code
\ .s
                %chain> mxt @ ?comp/exec     \ get method xt, compile or execute
\ .s ." ETM]"
           THEN
           True
           EXIT
      THEN
      %chain> mlink @ <to> %chain>
   REPEAT
   False
;

: EXECUTE-INST-METHOD	( inst methodid --- )
\ Execute instance method
   { %inst %method | %chain> %type --- }
\ cr ." E-I-M-1 " %method .
   %inst %inst INST>TYPE %method EXECUTE-TYPE-METHOD 0=
   IF
\ cr ." E-I-M-2 " %method .
        %inst IS-INST?                                      \ normal instance ??
        IF  \ yes! report method not applicable
             %inst OBJ-CRASH
        ELSE \ no!
             %inst TYPE-ID@ <to> %type
             %type IS-TYPE?                              \ id point to a type ??
             IF
                  %type inst-chain @ <to> %chain>   \ get inst method chain head
                  BEGIN \ -- INST METHOD-ID METHODSTRUCT
                       %chain>
                  WHILE \ -- inst method-id methodstruct
                       %method %chain> methodid @ =         \ found method id ??
                       IF \ yes!
                            %inst %inst INST>TYPE
                            OP# <ptr@> EXECUTE-TYPE-METHOD
                            IF
                                 %chain> mflags @     \ inst structure-method ??
                                 IF \ yes!
                                      %type  \ pass type thru a structure-method
                                      %chain> mxt @ execute
                                 ELSE \ no!
                                      %chain> mxt @             \ get method xt,
                                      ?comp/exec            \ compile or execute
                                 THEN
                                 EXIT
                            ELSE
                                 %inst OBJ-CRASH
                            THEN
                       THEN
                       %chain> mlink @ <to> %chain>
                  REPEAT
                  %type OBJ-CRASH
             ELSE
                  %inst OBJ-CRASH
             THEN
        THEN
   THEN
;

: EX-STATIC-METHOD	( inst --- )
   { %inst --- }
   %inst curinst !
   %inst op# <static> EXECUTE-INST-METHOD                \ get static method id
;

: MAKE-INST-DOES	( inst --- )
   { %inst --- }
   [ ALSO SYSTEM ]
   %inst curinst !
   %inst OperatorType @off EXECUTE-INST-METHOD      \ get & clr method indicator
   [ PREVIOUS ]
;

: EX-INST-METHOD-CHAIN	( type_addr --- )
   { %type --- }
\  cr ." E-I-M-C " OperatorType @ .
  [ ALSO SYSTEM ]
    %type
    OperatorType @off                           \ get and clear method indicator
    %type inst-chain @                          \ get instance method chain head
    begin                                       \ -- inst method-id methodstruct
      dup
     while                                      \ -- inst method-id methodstruct
      2dup methodid @ =                                       \ check method ids
      if nip >r                 \ -- inst ; discard methodid, save method struct
         r@ mflags @                         \ if instance structure data method
         if
            r> mxt @ execute                    \ it's immediate, always execute
         else
              inst-addr                     \ generate the address provider code
            r> mxt @ ?comp/exec              \ get method xt, compile or execute
         endif
         exit
      endif
      mlink @
    repeat
   %type OBJ-CRASH
   [ PREVIOUS ]
;

\ *S COMPLEX TYPE DEFINITIONS
\ *N TWIST COMPLEX TYPE

: TYPE-<TWIST>	( addr type> --- )
  { %addr %type> | %member --- }
  %type> <to> %member
  BEGIN
    %member field-link @ ?dup
  WHILE
       <to> %member
      %addr %member off-start @ +                  \ calculate address in buffer
      %member ^type @  OP# <twist> TYPE-METHOD-EXECUTE 0=
      IF
           DROP
      THEN
  REPEAT
;

: TYPE-CLASS-[TWIST]	( type> --- )
  ?COMP
  ?COMPLIT ASSIGN TYPE-<twist> xt-comp?
;

: TYPE-INST-[TWIST]	( obj --- )
   ?COMP                                                   \ compilation only !!
   INST-PROVIDER                                              \ get data address
   DUP IS-INST?                                         \ a instance of type: ??
   IF   \ yes!
        INST>TYPE ?COMPLIT
   ELSE \ no! must be a pointer, laydown type cast
        ?COMPLIT
   THEN
   ASSIGN TYPE-<twist> XT-COMP?
;

\ *N INIT COMPLEX TYPE

: TYPE-<INIT>                                                 ( addr type> --- )
  { %addr %type> | %member --- }
  %addr %type> OP# <init> TYPE-METHOD-EXECUTE 0=        \ type: method exists ??
  IF   \ no! go down type: member chain
       DROP                                                \ drop buffer address
       %type> <to> %member
       BEGIN
            %member field-link @ ?dup
       WHILE
            <to> %member
            %addr %member off-start @ +            \ calculate address in buffer
            %member INST>TYPE OP# <init> TYPE-METHOD-EXECUTE 0=
            IF
                 DROP                                      \ drop buffer address
            THEN
       REPEAT
  THEN
;

: TYPE-CLASS-[INIT]                                                  ( type> --- )
   \
   ?COMP                                                   \ compilation only !!
   ?COMPLIT ASSIGN TYPE-<init> xt-comp?
;

: TYPE-INST-[INIT]                                                     ( obj --- )
   \
   ?COMP                                                   \ compilation only !!
   INST-PROVIDER                                              \ get data address
   DUP IS-INST?                                         \ a instance of type: ??
   IF   \ yes!
        INST>TYPE ?COMPLIT
   ELSE \ no! must be a pointer, laydown type cast
        ?COMPLIT
   THEN
   ASSIGN TYPE-<init> XT-COMP?
;

\ *H Class version number
:NONAME                                                                 ( --- )
   #2765375	 #1212695859
   CCSExtensions BODY> >NAME NFA>CTRL CTRL>LINK
   #5 CELLS - TUCK ! CELL+
   #18753 OVER !     CELL+
   SWAP OVER !       CELL+
   #976893739 OVER ! CELL+
   #2962993   SWAP !
  ; EXECUTE

  \ *> ####
    HIDE-**