\ ClassVFX.fth - intelligent objects

((
Copyright (c) 1999, 2000, 2005, 2008, 2010, 2011, 2012
Construction Computer Software Pty Ltd.
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441

net: tech-support@mpeforth.com
web: www.mpeforth.com

To do
=====
Add SUPER and SELF.
Add late binding (ideal for message handling in winprocs)

Change History
==============
20190603 SFP022 Add trapped PTR! PTR@ PTR-OFF to ptr-template
20190504 SFP021 Refactored :M and MRUNS
20120718 SFP020 Added CURR-TYPE-SIZE.
20110407 SFP019 Documentation update
20101001 WJB018 v4.4 conversion
20080814 SFP017 Source tidy up
20000816 SFP016 Removed redundant code
		Moved odds and sods to CLASSHELP.FTH
		Made PFW/VFX version
20000522 SFP015 Change LV, to S_LVARRAY, to avoid compilation issues.
                Add DISCARD-SINLINE when a local instance is made.
                Avoid problems with optimised DOES>.
28 Oct 1999     WJB014  Merge CCS and MPE versions of CLASSVFX.FTH
01 Sep 1999     WJB013  Compilation of dot notation structures in comp-struct?
                        First, middle and last
                        First :- can be a local, colon, create or type def
                        Middle :- can be a local, colon, create or type def but
                                   added to what is on the stack already.
                        Last :- type def, actual operation on this member
18 Aug 1999     WJB012  add structures vocabulary for members
                        convert to one style of under and over lining
                        add CCS folding levels to headers
                        (comp-struct) add sizeof exception in "first."
17 Mar 1999     WJB011  add curinst-size for structure methods
                        add curinst-offset for structure methods
                        add curinst-addr for structure methods
                        add curinst-inst for structure methods
                        set and reset CurrentTypeDef in does of make-inst
24 Feb 1999     WJB010  Correct spelling of /inst-head in debug cfa list
                        Incorporate into CCS harness and drop newstruc.fth
23 Feb 1999     SFP009  Added different compilation for debugging
 8 Feb 1999     SFP008  Incorporated CCS changes to MAKE-INST
 5 Feb 1999     SFP007  Debug of first pass finished
24 Jan 1999     SFP006  Totally rewritten
))


\ ===========
\ *! classvfx
\ *T ClassVfx OOP
\ ===========
\ *P There are two sets of documentation for the *\i{ClassVFX}
\ ** system. There is a chapter in the main VFX Forth manual,
\ ** and there is a full PDF manual in the *\i{Manual} subdirectory
\ ** of *\i{Lib\OOP\ClassVFX\}.

[undefined] ERR_NULL_POINTER [if]
ErrDef ERR_NULL_POINTER		"Null pointer error"
[then]

\ ***************
\ *S Introduction
\ ***************
\ *P The source code is in the directory *\i{Lib\OOP\ClassVFX}.
\ ** The file *\i{MakeClassVfx.bld} is compiled to produce the
\ ** production version of the code. *\i{TestClassVfx.fth}
\ ** contains test code.

( *P ClassVFX was developed over a number of years in collaboration with )
( ** Construction Computer Software of Cape Town, South Africa. We gratefully )
( ** acknowledge their collaboration and permission to release it. ClassVFX is )
( ** heavily used in their construction industry planning software, which is )
( ** one of the largest Forth applications ever written. Modifications to )
( ** ClassVFX will only be released after the agreement of CCS. )

( *P ClassVFX is a halfway house between a full object oriented system and an )
( ** intelligent structures system. Types, or classes, can be defined with )
( ** single inheritance. Method names have to be predefined using )
( *C   OPERATOR: <method-name> )
\ *P Field, or data member, names are private, but are accessible
\ ** using a dot notation. There are no equivalents of *\fo{SUPER}
\ ** and *\fo{SELF}. There is no late binding.

\ *P In this documentation *\b{type}s and *\b{class}es are synonymous.
\ ** *\b{Object}s are *\b{instance}s of a *\b{type}. Objects have a
\ ** default action if no method is specified. Usually the default
\ ** action is to fetch the contents of the object, but in a few
\ ** cases the default action is to return an address.

\ *P Types/Classes can have both class and instance methods. The
\ ** default method for a type is to create an instance. If a
\ ** type is used inside a colon definition a local variable
\ ** version is created and destroyed at run time.

\ *P Operators, or methods, must be declared as above before use.


\ *************************
\ *S How to use TYPE: words
\ *************************
( *P *\fo{TYPE:} definitions may be used in four ways: )
\ *(
\ *B as an abstract template which is used with a base address on
\ ** the stack. In this case *\fo{Point} is a type (class).
\ *C   Point.x or x
\ *B to define an instance of a structure in the dictionary, e.g.
\ *C   Point: MyPoint
\ *B to define a local variable inside a colon definition, but
\ ** outside any other local variable defining mechanism. If
\ ** another locals defining mechanism such as the ANS
\ ** *\fo{LOCALS| ... |} mechanism or the MPE { ... } mechanism
\ ** has been used the use of *\fo{Point: foo} inside a colon
\ ** definition will simply add *\fo{FOO} to the  existing local
\ ** frame.
\ *B to define a field inside another *\fo{TYPE:} definition.
\ *)

( *E operator: <method1> )
( ** operator: <method2> )
( ** operator: <method3> )
( ** )
( ** type: line: )
( **   point: start )
( **   point: end )
( ** )
( **   :m <method1>     ... ;m )
( **   :m <method2>     ... ;m )
( **   mruns <method3>  <some-word> )
( **   :m <xxx>         a b c d  ;m  structure-method )
( ** )
( ** end-type )
( ** )
( ** Line: MyLine )
( **   1 2 to Myline.start )
( **   5 to Myline.end.y )

( *P At runtime, the method operates on the address of the data. Because of )
( ** this, a method which requires the address of the instance structure has )
( ** to be marked by the word *\fo{STRUCTURE-METHOD} which causes the compiler to )
( ** generate the address of the instance structure, *\b{not} the type structure. )

\ *P ClassVFX allows both *\fo{CLASS} and *\fo{INST}ance methods
\ ** to be defined for a type. *\fo{INST}ance methods, the default,
\ ** operate on the address of the data item. *\fo{CLASS} methods
\ ** operate on the address of the type data structure. As described
\ ** above, *\fo{STRUCTURE-METHOD}s operate on the instance data
\ ** structure.

\ *P Single inheritance can be defined using *\fo{SUPERCLASS <type>}
\ ** or *\fo{INHERITS <type>} before any field or method is defined.
( *E TYPE: <type>  SUPERCLASS <supertype> )
( **   ...  )
( ** END-TYPE )

\ *P At run-time, methods are provided with the address of the
\ ** required data. CLASS/TYPE methods receive the address of the
\ ** TYPE/CLASS data structure, INSTance methods receive the address
\ ** of the data item. INSTance methods that require the address
\ ** of the instance data structure must be marked by
\ ** *\fo{STRUCTURE-METHOD}. Methods may be defined as nameless
\ ** words:
( *C   :M <method-name> ...  ;M )
( *P or as the action of a method: )
( *C   MRUNS <method-name> <action-name> )

( *P The code below is taken from the definition of the default type. )
( *E  class     \ define methods for the type )
( **     :m default    make-inst  ;m )
( **     :m sizeof     type-size @ ?complit  ;m )
( **     :m addr       ?complit  ;m )
( **  inst      \ define methods for the instance )
( **     mruns default noop )
( **     :m sizeof     type-size @ ?complit  ;m  structure-method )
( **     mruns addr    noop )
( **     :m offsetof   off-start @ ?complit  ;m  structure-method )
( **     :m +offsetof  off-start @ ?complit+  ;m  structure-method )

\ *P To use the nested field system, the Forth system has been
\ ** modified to accept compound names in which the elements of
\ ** the structure are separated by the `.' character. This
\ ** feature is enabled and disabled by the words *\fo{+STRUCTURES}
\ ** and *\fo{-STRUCTURES}.

\ *******************
\ *S Predefined types
\ *******************
\ *P
( *E char:         byte - 8 bit variable )
( ** word:         word - 16 bit variable )
( ** int:          long - 32 bit variable and synonyms )
( **   dword: )
( **   long:  )
( **   ptr:   )
( ** xlong:        longlong - 64 bit variable )
( ** bytes:        byte array: size specified by n BYTES BYTES: )
( ** cstring:      counted string: size specified by BYTES before CSTRING: )
( ** zstring:      zero term. string: size specified by BYTES before ZSTRING: )
( ** field:        byte array, only ADDR operator, size specified by BYTES )

\ *******************************
( *S Predefined methods/operators )
\ *******************************
( *P Note that not all predefined types support all methods. )
( *E  0 operator default            usually a fetch operation )
( **  1 operator ->                 store operator )
( **  1 operator to                   "	           )
( **  2 operator addr               address operator )
( **  3 operator inc                increment by one )
( **  4 operator dec                decrement by one )
( **  5 operator add                n add to         )
( **  6 operator zero               set to 0         )
( **  7 operator sub                subtract from    )
( **  8 operator sizeof             size             )
( **  9 operator set                set to -1        )
( ** 10 operator offsetof           offset in object )
( ** 11 operator +offsetof          add offset in object  )
( ** 12 OPERATOR FETCH              get contents          )
( ** 13 OPERATOR ADDR\CNT           address under count   )
( ** 14 OPERATOR TWIST              change endian of the data type )
( ** 15 OPERATOR CONSTRUCT          build an instance of this type )
( ** 15 OPERATOR MAKE               build an instance of this type )
( ** op# ADDR OPERATOR ADDROF
( ** OPERATOR: <=>    type_addr_y <=> <type_x> --- set typedef_x = typedef_y )
( **   op# <=> OPERATOR <copy>                                   )
( ** OPERATOR: <blank>              blank object for object size )
( ** OPERATOR: <erase>              fill obj with null for object size )
( ** OPERATOR: <COUNT>                                           )
( ** OPERATOR: <make>                                            )
( ** OPERATOR: <destroy>                                         )
( ** OPERATOR: <INIT>                                            )
( ** OPERATOR: <fetch>                                           )


\ ********************
\ *S Example structure
\ ********************
( *E TYPE: POINT:  \ --
\ ** \ Defines a type called POINT: with the following fields )
( **   PROVIDER: NOOP        \ defines the address provider, defaults to NOOP )
( **   0 OFFSET:             \ defines the initial offset, defaults to 0      )
( **   INT: Y                                                                 )
( **   INT: X                                                                 )
( **   10 BYTES FIELD: FOO                                                    )
( **                         \ fetch operation                                )
( **   :m default                                                             )
( **      2@                                                                  )
( **   ;m                                                                     )
( **   mruns to      2!                                                       )
( **   ...                                                                    )
( **                                                                          )
( ** END-TYPE                                                                 )


\ ***********************************
\ *S Data structures created by TYPE:
\ ***********************************
\ *P *\fo{TYPE:} definitions, fields, objects and so on all use a
\ ** common data structure that is generated by the defining words.

\ *P These structures are associated with a word (the address
\ ** provider) that can provide the starting address of the
\ ** structure implementation. By supplying the cfa of *\fo{NOOP},
\ ** no address is provided, and so the structure is purely a
\ ** template. For templates, address provider = 0 or *\fo{NOOP},
\ ** an offset may also be defined. *\fo{NOOP} and 0 are the
\ ** default address provider and offset of templates.                                                                 )

\ *P A similar structure is used for instances of a *\fo{TYPE:}.
\ ** These are created by the word *\fo{MAKE-INST}.

\ ====================
( *N TYPE: definitions )
\ ====================
\ *P The following structure is created by *\fo{TYPE:}
( *E    header                          standard PFW layout   )
( ** 0  jmp do_type                     5 bytes               )
( ** 1  cfa of address provider         4 bytes               )
( ** 2  initial offset                  4 bytes  0 for class  )
( ** 3  link to last field defined      4 bytes               )
( ** 4  type size - final offset        4 bytes               )
( ** 5  Magic number                    4 bytes               )
( ** 6  anchor of instance method chain 4 bytes               )
( ** 7  anchor of type method chain     4 bytes               )
( ** 8  link to previous type defined   4 bytes               )
( ** 9  private wordlist                ? bytes               )

\ ========================
( *N MAKE-INST definitions )
\ ========================
\ *P The following structure is created by *\fo{MAKE-INST}
( *E    header                          standard PFW layout  )
( ** 0  jmp do_inst                     5 bytes              )
( ** 1  cfa of address provider         4 bytes              )
( ** 2  offset from start of type       4 bytes              )
( ** 3  link to last instance of type   4 bytes              )
( ** 4  size of instance data           4 bytes              )
( ** 5  0                               4 bytes              )
( ** 6  pointer to TYPE/CLASS           4 bytes              )
( ** 7  data if static                                       )


\ ***************************
( *S Local variable instances )
\ ***************************
( *P When an instance is defined inside a colon definition, an uninitialised )
( ** local variable/array is built. Several instances can be built. Normally )
( ** the size of all local variables is rounded up to a cell boundary by the )
( ** compiler

\ ( Especially )
\ ( ** when building Windows structures, it may be desirable to define a structure )
\ ( ** which is known to be on a four byte boundary and is rounded up at the end )
\ ( ** of the instance. This is done using the {{ ... }} notation. )
\
\ ( *E : foo )
\ ( **   point: x      \ local instance )
\ ( **   point: y      \ local instance )
\ ( **   byte: ctemp   \ local instance )
\ ( **   {{            \ force cell alignment )
\ ( **     Rect: Baah )
\ ( **     Rect: Lamb )
\ ( **   }}            \ end of alignment )
\ ( **   ... )
\ ( ** ;  )

  -sindoes
  [FORTH]
  decimal
here


\ *******************
\ *S Defining methods
\ *******************

\ SFP021...

: method,	\ struct "<method>" -- struct ^xt
\ *G Given a TYPE structure, lay an entry in one of the method chains.
  ?exec align                           \ barf if not interpreting
  here >r                               \ save start address of structure
  [compile] op# ,                       \ lay selector
  dup struct-childxt @ ,                \ update method chain
  r> over struct-childxt !
  0 ,                                   \ lay 0 flags
  here 0 ,                              \ lay dummy xt of new definition
;

: :M		\ struct "<method>" -- struct  ; :M <operator> <actions ...> ;M
\ *G defines the start of a method. The method/operator name must
\ ** follow.
  method,  :noname swap !		\ *** SFP *** VFX specific ; SFP012
;

: ;M		\ struct -- struct ; SFP012
\ *G marks the end of a method definition.
  postpone ;
; IMMEDIATE

: MRUNS         \ struct "<method>" "<word>" -- struct  ; MRUNS <operator> <word>
\ *G Defines a method which runs a previously defined word.
  method,  ' swap !			\ insert xt of word to run
;

\ ...SFP021

\ SFP015...
2 cells buffer: o_state	\ -- addr
\ First cell holds true when optimising.
\ Second cell holds true when tokenising.

: [o_state	\ -- ; start unoptimised sequence
  optimising? dup o_state !
  if  [compile] [o/f]  unoptimised  endif
  sin? dup o_state cell+ !
  if  -sin  endif
; immediate

: o_state]      \ -- ; end unoptimised sequence
  o_state @
  if  optimised  endif
  o_state cell+ @
  if  +sin  else  -sin  endif
; immediate
\ ...SFP015

0 VALUE doinst	\ -- addr
\ Returns address of instance code.

[undefined] <typedef> [if]                                              \ WJB014
VARIABLE <TYPEDEF>	\ -- addr ; true indicates we inside TYPE: ... END-TYPE
  <TYPEDEF> OFF
[endif]

: InstChildComp,	\ xt --
  >body MAKE-INST-DOES  ;

\ *******************************
\ *S Create Instance of an object
\ *******************************

: CREATE-INST	\ "<name>" -- ; -- addr
\ *G  From VFX Forth v4.4, this is a synonym for *\fo{CREATE}.
\ ** When compiling on previous VFX versions instances needed to
\ ** be immediate.
  OperatorType OFF
  CREATE  ;

: make-inst	\ class -- ; i*x -- j*y ; build instance of type
\ *G Builds an instance of a *\fo{TYPE:}. This word has serious
\ ** carnal knowledge of the internal workings of VFX Forth.
\ ** Don't call us for help!
 [ ALSO SYSTEM ]
  state @ if				\ if compiling, MUST be LOCAL instance
    ?init-locals  discard-sinline       \ LOCAL ; SFP015
    >temp-dict
  then
  CREATE-INST                           \ create new definition
    here >r                             \ mark address of new instance
    r@ last-inst !                      \ mark last instance built
    /inst-head allot&erase              \ reserve structure space
    r@ ^type !                          \ set type/class pointer
    #bpe @OFF R@ inst-bpe !             \ set bytes per element
    INST:-ID  R@ type-magicid !         \ set INST indicator
    state @ if                          \ LOCAL instance defined
      ['] lv r@ provider !              \ set local variable provider action
      Lvcount @ OFFSET-ALIGN Lvcount !  \ align local start before offset
      r@ inst-size #extras @off +
      dup OFFSET-ALIGN fradjust r@ off-start !	\ update local var frame size
      r> type-size !                    \ set field size
      ['] >real-dict >r                 \ executes after DOES>
    else                                \ STATIC or TYPEDEF
      <typedef> @ if                    \ TYPEDEF
        dup field-link @                \ structure being defined's link
        r@ field-link !                 \ is new link, cos this is a field
        r@ over field-link !            \ and then update the structure's link
        dup type-size @                 \ structure being defined's size so far
        r@ off-start !                  \   is the offset of this field
        r@ inst-size ( aligned )        \ size of instance
        #extras @off +                  \  + user data size (arrays)
        2dup swap type-size +!          \  is added to current structure size
        r@ type-size !                  \  and is field size
        ['] + r> provider !             \ so just add the offset at run time
      else                              \ no! i am a static instance
        with-inst @ 0= if               \ yes! do standard static instance
          0 r@ provider !               \ mark provider action as static
          r@ inst-size                  \ base data size
          #extras @off +                \ + user data size ( arrays)
          r@ type-size !                \ set size of data
          R@ EX-STATIC-METHOD
        else                            \ not default addres provider
          with-inst @
          r@ provider !                 \ mark provider action as static
          r@ inst-size                  \ base data size
          #extras @off +                \ + user data size (arrays)
          r@ type-size !                \ set size of data
        then
        r> drop
      endif
    endif
    with-inst off                       \ provider override is a one-shot
    ['] InstChildComp, set-compiler
    [o_state                            \ start unoptimised sequence
  interp>
    [ here 7 -  to doinst ]             \ kluge to patch DOINST *** CARNAL ***
    o_state]                            \ end unoptimised sequence
    MAKE-INST-DOES
   [ PREVIOUS ]
;


\ *****************************
\ *S Defining TYPE: and friends
\ *****************************

create type-template	\ -- addr
\ *G The type chain from which others are derived.
  here  /type-head allot&erase          \ return address for method use
  CLASS                                 \ define methods for the type
         mruns <default> make-inst
         mruns <sizeof>  sizeprovider
         mruns <addr>    ?complit
         mruns <self>    ?complit
  INST                                  \ define methods for the instance
         mruns <offsetof>  offsetprovider   structure-method
         mruns +offsetof +provider        structure-method
         mruns <sizeof>  sizeprovider     structure-method
         mruns <default> inst-addr        structure-method
         mruns <ptr>     inst-addr        structure-method
         mruns <&ptr>    inst-addr        structure-method
         mruns <addr>    inst-addr        structure-method
         mruns <copy>    inst-copy        structure-method
         mruns <blank>   inst-blank       structure-method
         mruns <erase>   inst-erase       structure-method
         mruns <init>    inst-erase       structure-method
         mruns <zero>    inst-erase       structure-method
         mruns <count>   inst-count       structure-method
         mruns <self>    ?COMPLIT         structure-method
         mruns <#bpe>    inst-bpeprovider structure-method

         :m <cfa> BODY> ?COMPLIT ;M
         mruns <static>  INST-ALLOT&ERASE STRUCTURE-METHOD
  drop

: PTR@  ( obj> --- addr )
   { %obj> --- }
   %obj> 0= IF ERR_NULL_POINTER ERROR THEN
   %obj> @
;

: PTR!  ( addr obj> --- )
   { %addr %obj> --- }
   %obj> 0= IF ERR_NULL_POINTER ERROR THEN
   %addr %obj> !
;

: PTR-OFF ( obj> --- )
   { %obj> --- }
   %obj> 0= IF ERR_NULL_POINTER ERROR THEN
   %obj> OFF
;

CREATE ptr-template	\ -- addr
\ *G The ptr chain from which others are derived.
   HERE  /type-head allot&erase         \ return address for method use
   CLASS                                \ define methods for the type
        :m <default>  make-inst  ;m
   INST                                 \ define methods for the instance
        :m <default>  inst-addr postpone PTR@ ;M  STRUCTURE-METHOD
        :m <ptr@>     inst-addr postpone PTR@ ;M  STRUCTURE-METHOD
        :m <ptr!>     inst-addr postpone PTR! ;M  STRUCTURE-METHOD
        :m <ptr0>     inst-addr postpone PTR-OFF ;M STRUCTURE-METHOD
        :m <static>   INST-ALLOT&ERASE     ;M  STRUCTURE-METHOD
        :m <&ptr>     inst-addr            ;M  STRUCTURE-METHOD
        \ Methods not allowed on pointers
        :m <to>       OBJ-CRASH ;M STRUCTURE-METHOD
        :m <zero>     OBJ-CRASH ;M STRUCTURE-METHOD
        :m <ptr>      OBJ-CRASH ;M STRUCTURE-METHOD
        :m <sizeof>   OBJ-CRASH ;M STRUCTURE-METHOD
        drop

\ ==================
\ *N TYPE definition
\ ==================

0 value dotype	\ -- addr ; runtime address for child of TYPE:
0 value doptr	\ -- addr ; runtime address for child of PTR:

variable prev-current	\ -- addr ; used by TYPE: and friends to restore data
variable prev-last      \ -- addr ; used by TYPE: and friends to restore data
variable CurTypeDef     \ -- addr ; used by type: to set current type

: type:-runtime	\ type-struct --
\ *G The run-time action of children of *\fo{TYPE:}.
  dup typefather !
  OperatorType @off			\ get and clear method indicator
  over type-chain @                     \ get class method chain head
  begin                                 \ -- type method-id methodstruct
    dup
   while                                \ -- type method-id methodstruct
    2dup methodid @ = if                \ check method ids
      nip  mxt @ execute                \ discard methodid, do method xt
      exit
    endif
    mlink @
  repeat
  typefather OFF
  1 abort" TYPE Method not available for this type"
;

\ SFP020...
: ?in-typedef	\ -- ; error if not in a TYPE: definition.
  <typedef> @ 0= abort" You must be inside a TYPE: definition"
;

: CURR-TYPE-SIZE	\ -- u
\ *G Use between TYPE: <name> and END-TYPE to return the current
\ ** size of the type.
  ?in-typedef  CurTypeDef @ type-size @
;
\ ...SFP020

: TypeChildComp,	\ xt --
\ *G Compile a child of *\fo{TYPE:}.
  >body type:-runtime  ;

: type:		\ -- struct ; --
\ *G Start a new *\fo{TYPE:} definition.
   \ Compilation pre init of type deLinition
   \ ---------------------------------------
   #extras OFF                          \ clr size gbl variable
   #bpe OFF                             \ clr bytes per element gbl variable
   \ Create new type definition
   \ --------------------------
    CREATE-INST                         \ create new definition
    here DUP CurTypeDef !               \ return address for field usage
    ['] noop ,                          \ 1) dummy provider
    0 ,                                 \ 2) then offset
    0 ,  0 ,                            \ 3) link, then 4) initial size=offset
    TYPE:-ID ,                          \ 5) mark this as a TYPE:
    type-template inst-chain @ ,        \ 6) instance method chain
    type-template type-chain @ ,        \ 7) class method chain
    struct-chain @ ,                    \ 8) link all structures
    #threads @  1 #threads !  wordlist >r  #threads !      \ 9) private wordlist
    dup struct-chain !
    <typedef> on                        \ mark start of TYPE: <name> .. END-TYPE
    inst                                \ default to instance methods
    current @ prev-current !            \ save Forth state data
    last @ prev-last !
    r> current !                        \ private wordlist is new CURRENT voc
    also                                \ make space for private word list
    ['] TypeChildComp, set-compiler     \ compiles children of TYPE:
    [o_state                            \ start unoptimised sequence
  interp>                               \ -- type ; things we can do with a type
    [ here 7 -  to dotype ]		\ *** CARNAL ***
    o_state]
    type:-runtime
;


: ptr:-runtime	\ pfa --
  dup typefather !
  OperatorType @off                     \ get and clear method indicator
  over type-chain @                     \ get class method chain head
  begin                                 \ -- type method-id methodstruct
    dup
   while                                \ -- type method-id methodstruct
    2dup methodid @ = if                \ check method ids
      nip  mxt @ execute                \ discard methodid, do method xt
      exit
    endif
    mlink @
  repeat
  typefather OFF
  1 abort" TYPE Method not available for this type"
;

: PtrChildComp,	\ xt --
  >body ptr:-runtime  ;

: PTR:		\ -- struct ; --
\ *G Make a new structure defining word.
   \ Pre init of type definition
   \ ---------------------------
   #extras OFF                          \ clr size gbl variable
   #bpe OFF                             \ clr bytes per element gbl variable
   \ Create new type definition
   \ --------------------------
  CREATE-INST                           \ create new definition
    here DUP CurTypeDef !               \ return address for field usage
    ['] noop ,                          \ 1) dummy provider
    0 ,                                 \ 2) then offset
    0 ,  0 ,                            \ 3) link, then 4) initial size=offset
    TYPE:-ID ,                          \ 5) mark this as a TYPE:
    ptr-template inst-chain @ ,         \ 6) instance method chain
    ptr-template type-chain @ ,         \ 7) class method chain
    ptr-chain @ ,                       \ 8) link all structures
    #threads @  1 #threads !  wordlist >r  #threads !      \ 9) private wordlist
    dup ptr-chain !
    <typedef> on                        \ mark start of TYPE: <name> .. END-TYPE
    inst                                \ default to instance methods
    current @ prev-current !            \ save Forth state data
    last @ prev-last !
    r> current !                        \ private wordlist is new CURRENT voc
    also                                \ make space for private word list
    ['] ptrChildComp, set-compiler
    [o_state                            \ start unoptimised sequence
  INTERP>                               \ -- type ; things we can do with a type
    [ here 7 -  to doptr ]		\ *** CARNAL ****
    o_state]
    ptr:-runtime
;

: end-type	\ struct --
\ *G Finish off a *\fo{TYPE:} definition
  previous                              \ remove private word list
  prev-current @ current !              \ restore Forth state data
  prev-last @ last !
  drop  <typedef> off  cleanup-locals
  inst
  CurTypeDef OFF ;

: EXTEND-TYPE	\ "<type>" -- struct ; EXTEND-TYPE <type> ... END-TYPE
\ *G Extend the given *\fo{TYPE:} definition.
   \ Pre init of type definition
   \ ---------------------------
   #extras OFF                          \ clr size gbl variable
   #bpe OFF                             \ clr bytes per element gbl variable
   \ Extend type definition
   \ ----------------------
  ?exec                                 \ error if not interpreting
  ' dup >does dotype <> IF              \ is it a TYPE: definition
    dup >does doptr <> IF
      True abort" Not a child of TYPE:"	\ barf if not
    THEN
  THEN
  >body  DUP CurTypeDef !               \ return structure address
  current @ prev-current !
  last @ prev-last !
  dup type-wordlist current !
  also                                  \ make space in for private word list
  dup type-wordlist context !           \ set search order to private word list
;

: SUPERCLASS	\ struct "<type>" -- struct
\ *G Use this inside a *\fo{TYPE:} definition before defining any
\ ** data or methods. The current type will inherit the data and
\ ** methods of the superclass.
  ?exec  <typedef> @ 0=                 \ must be used in definition of a type
  abort" must be used between TYPE: and END-TYPE"
  dup off-start @  over type-size @ or  \ check that no fields are defined
  abort" Subclass cannot have previous offset or previous member"
  ' dup >does dotype <>                 \ get superclass type, and check
  abort" Superclass must be defined by TYPE"
  >body  over /type-head cell - move    \ copy superclass data, first 6 fields
;                                       \ do not overwrite type chain

: INHERITS	\ struct "type" -- struct
\ *G A synonym for *\fo{SUPERCLASS}.
   superclass ;

: provider:     \ struct "name" -- struct ; <name> is address provider
\ *G Sets a different address provider.
  ' DUP with-inst ! over provider !
;

: with:         \ -- ; WITH: <some-provider> LINE: <myline>
\ *G Used before declaring an instance to override the default address
\ ** provider.
  ?exec
  ' DUP RemoveSin with-inst !
  ;

: SKIPPED       \ struct size -- struct
\ *G Increase overall size of struct by size. *\fo{SKIPPED} can
\ ** be used to jump over items from a previous instance.
   over type-size +!
;

: OFFSET:	\ struct offset -- struct
\ *G Define the offset of a *\fo{TYPE:} as starting at a value
\ ** other than zero. Must be used before any data is defined.
  2dup swap  off-start !                \ initial offset to offset and size
  skipped
;

: TypeCast:	\ -- ; TYPECAST: <inst> <type>
( *G Forces a previously defined instance to be a pointer to a type/class. )
  ' >body dup ?inst
  ' >body dup ?type
  swap type-magicid !
; immediate

SYNONYM PointsTo: TypeCast:	\ -- ; synonym for TYPECAST:
( *G Forces a previously defined instance to be a pointer to a type/class. )

: type-self	\ -- type
\ *G Used in TYPE: <name> ... END-TYPE to refer to the type/class
\ ** being defined.
   ?EXEC  CurTypeDef @  ;

: EXECUTE-MEMBER-METHOD	\ struct-inst member-inst methodid ---
\ *G Attempt to execute method for inst. Return true if successful.
   { %struct-inst %member-inst %method | %chain> %type --- }
   [ ALSO SYSTEM ]
   %member-inst INST>TYPE <to> %type
   %type inst-chain @ <to> %chain>	\ get instance method chain head
   BEGIN                                \ -- INST METHOD-ID METHODSTRUCT
      %chain>
   WHILE                                \ -- inst method-id methodstruct
      %method %chain> methodid @ =      \ found method ids ??
      IF
         %struct-inst inst-addr
         %chain> mflags @               \ instance structure-method ??
         IF \ yes!
            %member-inst %chain> mxt @ execute	\ it's immediate, execute
         ELSE \ no!
            %member-inst inst-addr      \ generate the address provider code
            %chain> mxt @ ?comp/exec    \ get method xt, compile or execute
         THEN
         EXIT
      THEN
      %chain> mlink @ <to> %chain>
   REPEAT
   [ PREVIOUS ]
;

: EXECUTE-PTR-MEMBER-METHOD	\ member-inst methodid ---
\ *G Attempt to execute method for inst. Return true if successful.
   { %member-inst %method | %chain> %type --- }
   [ ALSO SYSTEM ]
   %member-inst INST>TYPE <to> %type
   %type inst-chain @ <to> %chain>	\ get instance method chain head
   BEGIN                                \ -- INST METHOD-ID METHODSTRUCT
      %chain>
   WHILE                                \ -- inst method-id methodstruct
      %method %chain> methodid @ =      \ found method ids ??
      IF
         POSTPONE R@                    \ get pointer address
         %chain> mflags @               \ instance structure-method ??
         IF \ yes!
            %member-inst %chain> mxt @ execute	\ it's immediate, execute
         ELSE \ no!
            %member-inst inst-addr      \ generate the address provider code
            %chain> mxt @ ?comp/exec    \ get method xt, compile or execute
         THEN
         EXIT
      THEN
      %chain> mlink @ <to> %chain>
   REPEAT
   [ PREVIOUS ]
;

: EXECUTE-MEMBERS	\ inst method --
\ *G Apply the given method to all members of the instance of a type/class.
  { %inst %method | %next %count --- }
  [ ALSO SYSTEM ]
  %inst IS-INST?
   IF
      %inst INST>TYPE <to> %next
      BEGIN
         %next field-link @ ?dup
      WHILE
         <to> %next
         %inst %next %method EXECUTE-MEMBER-METHOD
      REPEAT
   ELSE
      %inst IS-TYPE?                    \ id point to a type ??
      IF   \ yes!
         1 <to> %count
         %inst <to> %next
         BEGIN
            %next field-link @ ?dup
         WHILE
            <to> %next
            %count 1 =
            IF
               POSTPONE >R              \ save pointer addr
            THEN
            %next %method EXECUTE-PTR-MEMBER-METHOD
            <inc> %count
         REPEAT
         %count  1 >
         IF
            POSTPONE R>                 \ loose pointer addr
            POSTPONE DROP
         THEN
      ELSE \ no! can not compile this
         %inst OBJ-CRASH
      THEN
   THEN
   [ PREVIOUS ]
;

: TWIST-STRUCTURE	\ inst --
\ *G Twist structure method
  OP# <twist> EXECUTE-MEMBERS  ;

: INIT-STRUCTURE	\ inst --
\ *G Init structure method
  OP# <init> EXECUTE-MEMBERS  ;

+sindoes

[forth]

cr ." Implementation takes " here swap - decimal . ." bytes"


\ ======
\ *> ###
\ ======
