\ DotPsVfx.fth - Dot notation parser for CLASSVFX system

((
Copyright (c) 1999, 2000, 2010, 2011, 2019
Construction Computer Software
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441

net: tech-support@mpeforth.com
web: www.mpeforth.com

20190504 SFP001 Refactored.
))


[FORTH]

\ ===========
\ *> classvfx
\ ===========

\ **********************
\ *S Dot notation parser
\ **********************
( *P In order to deal with structures and fields without having to backtrack the )
( ** input stream or the execution order, an additional stage is added to the    )
( ** Forth parser to allow phrases of the forms:                                 )
( *C   inst.field                                                                )
( *C   inst.field.field                                                          )
( *C   type.field                                                                )
( *C   type.field.field                                                          )
( *P to be parsed, where each item is separated by a dot character. The first item  )
( ** must be an instance of a type or a type. If it is an instance, the address is  )
( ** provided, otherwise the base address is assumed to be on the stack. Any items  )
( ** between the first and last item add their offsets to the address, and the last )
( ** item performs the usual operation of the field as defined by an operator. For  )
( ** example:                                                                    )

( *E   type: point:                                                              )
( **     int: x0                                                                 )
( **     int: y0                                                                 )
( **                                                                             )
( **     :m <op1> ... ;m                                                         )
( **     :m <op2> ... ;m                                                         )
( **   end-type                                                                  )
( **                                                                             )
( **   point: Mypoint                                                            )
( **     5 to MyPoint.x0                                                         )

( *P We might define a line as joining two points:                               )
( *E   type: line:                                                               )
( **     point: p1                                                               )
( **     point: p2                                                               )
( **     ...                                                                     )
( **   end-type                                                                  )
( **                                                                             )
( **   line: MyLine                                                              )
( **     5 to MyLine.p2.y0                                                       )

: must-be-inst-throw	\ xt --
\ *G THROW because the xt is not of an instance of a type.
  XT>NFA COUNT >ABORT-FUNC$
  S" Use Instance of type only !" COMP-$ABORT
;

: class-ise-throw	\ --
\ *G THROW because we have misconstructed a CLASS.
  S" CLASS Internal Software Error" COMP-$ABORT
;

: COMP-1ST-STRUCT  **	\ operator cfa flag --
\ *G Compile the first part of a dotted phrase.
\ ** Instances of TYPE: or POINTER: are the only valid cfa's.
   {: %operator %cfa %immflag | %pfa %id --- :}
   %cfa >body <to> %pfa
   %pfa is-inst? 0= if                                \ instance of type: ??
     %pfa INST>TYPE INFORTH? 0=             \ addr at ^type within forth ??
     if  %cfa must-be-inst-throw  then
     %pfa INST>TYPE is-type? 0=           \ pointer points to valid type: ??
     if  %cfa must-be-inst-throw  then
   then

  %operator op# <offsetof> <>
  %operator op# <sizeof>   <> AND
  IF
       %cfa >body TYPE-ID@ dup  <to> %id  0=
       if  class-ise-throw  then
       %id INST:-ID =
       if  OP# <addr>  else  OP# <ptr@>  then
       OperatorType !
       %cfa execute
  THEN
;

: COMP-MIDDLE-STRUCT **	( operator cfa flag --- )
\ *G Compile the middle portions of a dotted phrase.
\ ** Instances of TYPE: are the only valid cfas.
  {: %operator %cfa %immflag -- :}
  %cfa >body is-inst? 0=
  if  %cfa must-be-inst-throw  then

  %operator op# <sizeof> <> if
    %operator op# <offsetof> =
    if  %operator  else  op# +offsetof  then
    OperatorType !  %cfa execute
  THEN
;

: COMP-LAST-STRUCT  **	\ operator cfa flag --
\ Compile the last part of a dotted phrase.
  {: %operator %cfa %immflag | %pfa --- :}
  %cfa >body <to> %pfa
  %pfa is-type? %pfa is-inst? OR 0=           \ type: def or instance of type: ??
  if  must-be-inst-throw  then
  %operator OperatorType !
  %pfa is-inst?
  IF
    %cfa execute
  ELSE
    %cfa >BODY EX-INST-METHOD-CHAIN
  THEN
;

DEFER dbgSET-TYPEVOC  **	( pfa --- )
ASSIGN DROP TO-DO dbgSET-TYPEVOC
VARIABLE dbgVar **

: SET-TYPEVOC  **	\ xt --
\ set top of search order to TYPE: private wordlist
   { %cfa | %pfa --- }
   %cfa >body <to> %pfa
   %pfa IS-TYPE?                                         \ if a TYPE: definition
   IF   \ yes! type: defintion
        %pfa type-wordlist context !
        #1 dbgVar ! %pfa dbgSET-TYPEVOC
        exit
   THEN
   %pfa IS-INST?                               \ is xt an instance of a type: ??
   IF \ yes! instance of type
        %pfa INST>TYPE type-wordlist context !
        #2 dbgVar ! %pfa dbgSET-TYPEVOC
        exit
   THEN
   %pfa INST>TYPE INFORTH?
   IF   \ yes!
        %pfa INST>TYPE IS-TYPE?                \ pointer point to valid type: ??
        IF  \ yes!
             #3 dbgVar ! %pfa dbgSET-TYPEVOC
             %pfa TYPE-ID@ type-wordlist context !
             exit
        THEN
   THEN
   S" Set Search Order for TYPE: members" >ABORTCAPTION$
   %cfa XT>NFA COUNT >ABORT-FUNC$
   S" not a TYPE: or child of TYPE:"  COMP-$ABORT
  ;

DEFER dbg(comp-struct)	**	( pfa --- )
  ASSIGN DROP TO-DO dbg(comp-struct)

: TYPE-INST-MEMBER?	**	( type inst --- flag )
   { %type %inst | %flag %next --- }
   <zero> %flag
   %type field-link @ <to> %next
   BEGIN
      %next
   WHILE
        %next %inst =
        IF
             <zero> %next
             <set> %flag
        ELSE
             %next field-link @ <to> %next
        THEN
   REPEAT
   %flag
;

: must-be-type/inst-throw	\ --
\ error when item is not a type or an instance of a type.
  OperatorType OFF
  S" Use Instance of type or type!" COMP-$ABORT
;

: must-be-valid-member-throw	\ --
\ error when itme is not a valid member.
  S" Not a valid member!" COMP-$ABORT
;

: (comp-struct)	**	\ -- flag ; process aaa.bbb.ccc from i/p stream ;\ WJB004
\ The stream is treated as first.middles.last
\ If the operator is ADDR or OFFSETOF, change the last operation to
\ be performed, and then do the first operation.
\ The first operator in the dot sequence may be a colon definition that
\ provides the address for the rest of the sequence.
\ Flag is returned true for a valid dot sequence.
  {: | %operator %immflag %cfa %error %pfa %type -- flag :}
  zBuffer: %zbuf

  <zero> %error
  <zero> %type
  discard-sinline
  BL parse-leading
  pocket count <to> %zbuf
  [CHAR] . word find dup 0=		\ does the word exist?
  if  exit  endif			\ no, get out
  <to> %immflag				\ yes! get the immediate status of the word
  <to> %cfa				\ ...get the cfa of the word
  %immflag 0=				\ must be immediate
  if  false exit  endif

  %cfa >BODY <to> %pfa			\ get the body of the word
  %cfa Set-TypeVoc			\ Open the wordlist, process it
  OperatorType @OFF <to> %operator
  %pfa IS-TYPE?				\ a type: definition ??
  IF					\ yes, base address is on stack
    %pfa <to> %type
    #1 dbgVar ! %pfa dbg(comp-struct)
  ELSE \ no! must be instance or pointer ??
    %pfa IS-INST?                          \ a instance of type: ??
    IF   \ yes!
      %pfa INST>TYPE <to> %type
      #2 dbgVar ! %pfa dbg(comp-struct)
    ELSE \ no! must be a child of ptr:
      %pfa TYPE-ID@ <to> %type
      #3 dbgVar ! %pfa dbg(comp-struct)
    THEN
    %operator op# <addr> =
    IF   OP# +offsetof >R
    ELSE  %operator >R
    THEN
    %operator %cfa %immflag COMP-1ST-STRUCT
    R> <to> %operator
    <count> %zbuf pocket place
  THEN

  \ Now process the middles with the +OFFSETOF method
  BEGIN
    [CHAR] . WORD FIND
    <to> %immflag                 \ get the immediate status of the word
    <to> %cfa                               \ ...get the cfa of the word
    %immflag 0=                                                 \ found ??
    if  must-be-type/inst-throw  exit  then
    %cfa >BODY <to> %pfa
    %cfa SET-TYPEVOC               \ open required private wordlist
    #4 dbgVar ! %pfa dbg(comp-struct)
    >in @ #tib @ <>                   \ yes! end of input stream ??
    IF
      %pfa IS-INST? 0=                     \ a instance of type: ??
      if  %cfa must-be-inst-throw  exit  then
      %type %pfa TYPE-INST-MEMBER? 0=
      if  must-be-valid-member-throw  exit then
      %operator %cfa %immflag comp-middle-struct
      %operator op# <offsetof> =
      IF
        op# +offsetof <to> %operator
      THEN
      %pfa INST>TYPE <to> %type
      #5 dbgVar ! %pfa dbg(comp-struct)
      False
    ELSE
      %pfa IS-INST?                        \ a instance of type...
      %pfa IS-TYPE? OR               \ ...or a type: definition ??
      0=
      if  must-be-type/inst-throw exit  then
      %pfa IS-INST?                  \ a instance of type: ??
      IF \ yes!
        %type %pfa TYPE-INST-MEMBER? 0=
        if  must-be-valid-member-throw  exit  then
      THEN
      %operator %cfa %immflag comp-last-struct
      #6 dbgVar ! %pfa dbg(comp-struct)
      True True              \ return true flag and exit loop
    THEN
  UNTIL
;

\ =======================
\ *N Compiling for VFX v4
\ =======================
\ *P VFX v4 provides a hook in the interpreter loop especially for
\ ** object package parsers.

vfx5? 0= [if]

: comp-struct?  **	\ -- flag
\ The stream is treated as first.middles.last
\ If the operator is ADDR or OFFSETOF, change the last operation to
\ be performed, and then do the first operation.
  save-input bl word find >r drop restore-input drop
  r> if 0 exit then
  save-input
  bl word number? dup >r
  0 ?do  drop  loop
  restore-input drop
  r> if 0 exit then
  also
  ['] (comp-struct) CATCH DUP Abort-Code !
  previous
  IF   \ yes! compile error
    ACTION-OF DISP-ABORT ASSIGN NOOP =                      \ in a console ??
    if  Abort-Code @OFF THROW  then
    DISP-ABORT  False
  THEN
;

: StructHook  **	\ c-addr u -- flag
\ Return true if compiled properly
  pocket place                                           \ copy away from HERE
  save-input n>r
  >in off  blk off
  pocket count to-source -1 to source-id
  comp-struct?
  nr> restore-input drop
;

: +structures	runword	\ --
\ *G Switch on the structure compiler.
  assign StructHook to-do ClassHook
;

: -structures   runword	\ --
\ *G Switch off the structure compiler.
  assign (classhook) to-do ClassHook
;

[then]


\ =======================
\ *N Compiling for VFX v5
\ =======================
\ *P VFX v5 uses recognisers for all parsers. Installing a dot notation
\ ** parser is something of a kluge as the minimum has been done to make
\ ** code work without a total rewrite of the parsing code.

vfx5? [if]

: dotNotation?	\ -- flag
\ *G Return true if the text at *\fo{POCKET} appears to be a
\ ** well-formed dot notation string
  pocket count [char] . scan nip 0=	\ must contain a .
  if  0 exit  then

  save-input				\ must not be an existing word
  bl word find >r drop
  restore-input drop
  r> if 0 exit then

  save-input				\ must not be a number
  bl word number? dup >r
  0 ?do  drop  loop
  restore-input drop
  r> 0=
;

2variable DotPS$	\ -- addr
\ *G Temporarily holds text string address and length.

: isDotText?	\ c-addr u -- flag
\ *G Return true if the text appears to be a well-formed dot
\ ** notation string.
  2dup DotPS$ 2!  pocket place
  save-input n>r
  >in off  blk off
  pocket count to-source -1 to source-id
  dotNotation?
  nr> restore-input drop
;

: doDotText	\ --
\ cr ." d-dt " .s
  DotPS$ 2@ pocket place
  save-input n>r
  >in off  blk off
  pocket count to-source -1 to source-id
\ .s
  also
  ['] (comp-struct) CATCH DUP Abort-Code !	\ -- flag ior
  previous
\ .s
  nr> restore-input drop
  swap 0= or
  IF   \ yes! compile error
    ACTION-OF DISP-ABORT ASSIGN NOOP =	\ in a console ??
    if  Abort-Code @OFF THROW  then
    DISP-ABORT  False
  THEN
\ .s
;

: postDotText	\ --
  s" Cannot POSTPONE aa.bb.cc" comp-$abort
;

' doDotText  ' doDotText  ' postDotText  RecType: r:classVFX	\ -- struct
\ *G Contains the three actions for dot parsers.

: rec-classVFX	\ caddr u -- r:float | r:fail
\ *G The parser part of the floating point recogniser.
\ cr ." r-cv" .s
  isDotText? if  r:classVFX  else  r:fail  then
\ .s
;

: +structures	runword	\ --
\ *G Switch on the structure compiler.
  ['] rec-classVFX forth-recognizer +stack-bot
;

: -structures   runword	\ --
\ *G Switch off the structure compiler.
  ['] rec-classVFX forth-recognizer -stack
;

[then]

\ ======
\ *> ###
\ ======

  HIDE-**
