((
$Date: 2008-05-10 11:52:33 +0200 (sáb 10 de may de 2008) $
$Revision: 520 $
$Author: astrorafael $
$HeadURL$

yet another Forth objects extension
written by Anton Ertl 1996-2000
public domain; NO WARRANTY
))
  

\ #####################################
\ *! objects
\ *R @node{Object Oriented Programming}
\ *T Object Oriented Programming
\ #####################################


\ ***************
\ *S Introduction
\ ***************

\ *P This module is a porting of GForth's *\f{'objects.fs'} object oriented
\ ** package. As such, it maintains all the features found in the original
\ ** package. However, many of the internal words have been hidden inside the
\ ** module and a 'persistant' extension have been made. Please consult the
\ ** GForth User's manual for an introduction on how to use this package.  

\ =====================
\ *N Summary of changes
\ =====================

\ *( 1
\ *B *\b{Modifying the Gforth's catch}.  VFX Forth uses a separate exception
\ ** stack, so >ep and ep> are used, instead or >r and >r.
\ *B *\b{Deleted some helper words}. Words like *\fo{-rot}, *\fo{?dup-if} or
\ ** *\fo{\g} have been deleted.
\ *B *\b{Added word}: *\fo{[parent']} to help catching exceptions in methods
\ ** when calling the parent method for a given selector.
\ *B *\b{User defined exception}: *\fo{Method-Excp} instead of *\fo{abort"}
\ ** for missing selector implementation.
\ *B *\b{Thread-safe implementation of the 'this' pointer}. *\fo{this} is no
\ ** longer a *\fo{VALUE} but a *\fo{USER} variable.
\ *B *\b{Added persistance support for turnkey file generation}
\ *B *\fo{Construct}, a non documented VFX word has been redefined and
\ ** adopted as the constructor. No side effects have been observed.  
\ *)  


\ =============
\ *N Exceptions
\ =============
  
\ *P Added  Anton's new word [parent'] to get the xt of the parent method.
\ ** This allows catching exceptions within methods.

\ *P  instead of:
\ *C      this [parent]  foomethod
\ *P  to capture an exception, write:
\ *C      this [parent'] foomethod catch

\ =====================
\ *N Adding Persistance
\ =====================

\ *P A noteworthy feature absence is the lack of persistance support for
\ ** saved images or turnkey applications. Indeed, the class/interface maps
\ ** are stored in the heap, so a new turnkey application will not have them
\ ** and will crash.

\ *P To alleviate this, a *\fo{persistant} word has been defined.
\ ** It is automatically invoked at *\fo{end-interface}. However, it must
\ ** be manually invoked after *\fo{end-methods} or *\fo{end-class}, depending
\ ** on your class declaration style. If you want to generate turnkey images,
\ ** you should use it in each and every class you define.
\ ** If you always stick to the same declaration style in your project
\ ** you can make it easy with a redefinition, for instance:
\ *C   : end-class    end-class persistant ;


\ ***************
\ +S Design notes
\ ***************

  
\ ==================
\ +N Data Structures
\ ==================  

\ *P These data structures are interbally defined using the ported
\ ** *\fo{GForth-Struct} module.  
  
\ -----------------------------  
\ +H *\fo{INTERFACE%} structure
\ -----------------------------

\ +P The *\fo{interface-map} field is a double cell containing the starting
\ +* address and size of the class map (the "virtual methods jump table").
\ +* *\fo{interface-map-offset} is the difference between the real xt's
\ +* jump table and the interfaces implementation table (see figure). It is
\ +* zero for classes not implementing interfaces and also for interfaces
\ +* themselves.
\ +* The  *\fo{interface-offset} is a negative number (0 for classes) that
\ +* serves both as unique interface identifier and also as offset into the
\ +* class-map.
  
\ *P The *\fo{interface%} structure is allocated in the dictionary. 
  
\ +E  +----------------------+
\ +*  |  ( start address )   |
\ +*  + interface-map   -----+
\ +*  |   ( size )           |           
\ +*  +----------------------+
\ +*  | interface-map-offset |
\ +*  +----------------------+
\ +*  | interface-offset     |
\ +*  +----------------------+
  
\ ------------------------
\ +H *\fo{CLASS%} structure
\ -------------------------

\ +P The *\fo{class%} structure extends the *\fo{interface%} structure
\ +* with new fields specific to classes. Among these, the
\ +* *\fo{class-inst-size} contains the alignment and size of class instances
\ ** (aka objects). This structure is also allocated in the dictionary.  
  
\ +E  +----------------------+
\ +*  |  ( start address )   |
\ +*  + interface-map    ----+
\ +*  |   ( size )           |          
\ +*  +----------------------+
\ +*  | interface-map-offset |
\ +*  +----------------------+
\ +*  | interface-offset     |
\ +*  +----------------------+
\ +*  | class-parent         |
\ +*  +----------------------+
\ +*  | class-wordlist       |
\ +*  +----------------------+
\ +*  |  (align)             |
\ +*  + class-inst-size   ---+
\ +*  |  (size)              |
\ +*  +----------------------+

\ ----------------------  
\ +H Class/Interface map
\ ---------------------- 

\ *P This is where the real XT's of methods are stored, Both interfaces
\ +* and classes have this table. The trick to accomodate classes that
\ +* implement one or several interfaces is to extend the table at the front
\ +* with negative offsets. Class implementing an interface whose id is -12
\ +* for instance, will have this table padded with NULL pointers up to the
\ +* third entry.  
  
\ +E     +------>  +--------------------+
\ +*     |         | interface-map -4*n |  <-- interface-map start address
\ +*     |         +--------------------+      /\
\ +*     |         //                  //      |
\ +*     |         +--------------------+      |
\ +*     |         | interface-map -8   |   interface-map-offset
\ +*     |         +--------------------+      |
\ +*     |         | interface-map -4   |      |
\ +* interface-map +--------------------+      V
\ +*    size       | class% ptr         |  <--- object-map
\ +*     |         +--------------------+
\ +*     |         | method xt 4        |
\ +*     |         +--------------------+
\ +*     |         | method xt 8        |
\ +*     |         +--------------------+
\ +*     |         //                  //
\ +*     |         +--------------------+
\ +*     |         | method xt 8*n      |
\ +*     +------>  +--------------------+

\ +P In the original implementation, these class/interface maps are allocated
\ +* in the heap, so when saving a system, al XTs and temporary memory
\ +* references are lost.  
  
\ ---------------------
\ +H Selector structure
\ ---------------------

\ *P Selectors are structures compiled in the dictionary having the following
\ ** elements:
  
\ +E  +----------------------+
\ +*  | selector-offset      |
\ +*  +----------------------+
\ +*  | selector-interface   |
\ +*  +----------------------+

\ *P *\fo{selector-offset} is the possitive offset within the interface-map
\ ** that retrieves the real XT to execute. *\fo{selector-interface} is
\ ** the interface unique id that this selector belongs to. As we know,
\ ** this is alos the offset to retrieve an interface-map for a given
\ ** class-map.  
  
    
\ =====================
\ +N Adding Persistance
\ =====================

\ +P There is a slight memory leak when using *\fo{persistant} in class
\ +* implementing interfaces because we cannot guarantee that interface-maps
\ +* are in the heap.There was a free() crash detected by libc when trying
\ +* to do so. This memory leak is not present in the saved image, of course.
  
  
\ ***********
\ *S Glossary
\ ***********

MODULE objects

  REQUIRES GForth-Struct
  

\ =========================
\ +N Support words
\ =========================
  
: perform ( ... addr -- ... )     
\ +G Execute xt stored at address *\i{addr}.
   @ execute
;

\ =========================
\ +N Heap memory management
\ =========================

: save-mem	( addr1 u -- addr2 u ) \ gforth
\ +G copy a memory block into a newly allocated region in the heap
   swap >r
   dup ProtAlloc
   swap 2dup r> -rot move
;

: resize ( a-addr1 u -- a-addr2 ior ) \ gforth
\ +G Resizes a heap memory region *\i{a-addr1 u}. If *\i{a-addr1} is NULL
\ +* then alocate a new region.   
   over
   if
      resize				\ standard resize
   else
      nip ProtAlloc
   then
;

: extend-mem	( addr1 u1 u -- addr addr2 u2 )
\ +G Extend memory block allocated from the heap *\i{addr1 u} by *\i{u}
\ +* addressing units.
\ +* The (possibly reallocated) piece is *\i{addr2 u2},
\ +* the extension is at *\i{addr} (= addr2+u1)
   over >r + dup >r resize  #-59 ?throw	\ addr2 ; R: u1 u2
   r> over r> + -rot			\ addr2+u1 addr2 u2
;


\ =======================
\ +N Main Data structures
\ =======================

\ +P The *\fo{object%} structure holds all data value for class instances and
\ +* its first data member is always a pointer to the class-map.
\ +[
struct%
   cell% field% object-map		\ a heap memory region
end-struct% object%

struct%
   cell% 2* field% interface-map	\ a heap memory region (addr1 u1)
   cell%    field% interface-map-offset \ aus
   \ difference between where interface-map points and where
   \ object-map points (0 for non-classes)
   cell%    field% interface-offset \ aus
   \ offset of interface-map pointer in class-map (0 for classes)
end-struct% interface%

interface%
   cell%    field% class-parent
   cell%    field% class-wordlist \ inst-vars and other protected words
   cell% 2* field% class-inst-size 
end-struct% class%

struct%
   cell% field% selector-offset    \ the offset within the (interface) map
   cell% field% selector-interface \ the interface offset or id.
end-struct% selector%
\ +]


\ ========================
\ +N Selectors and methods
\ ========================

Variable current-interface ( -- addr ) \ objects- objects
\ +G This variable contains the class or interface currently being
\ ** defined.
   
ErrDef Method-Excp "No method defined for this object/selector combination"
\ *G

: no-method ( -- )
\ +G The default actio when there is no method (throws *\fo{Method-Excp}).   
   Method-Excp throw
;

: do-class-method ( -- )
\ +G Run time code for a direct class selector.   
 does> ( ... obj selector% -- ... )
   selector-offset @			\ -- obj offset 
   over object-map @ +			\ -- obj xtp
   perform
;


: do-interface-method ( -- )
\ +G Run time code for an interface selector.      
 does> ( ... obj selector% -- ... )
   2dup selector-interface @	        \ -- obj selector% obj ifz-offset
   swap object-map @ + @		\ -- obj selector% ifz-map
   swap selector-offset @ +		\ -- obj xtp
   perform
;


: method ( xt "name" -- ) 
\ +G This word lays out a *\fo{selector%} structure on the dictionary,
\ +* extend class/interface map on the heap by 1 cell and
\ +* chooses the run time behaviour (either class or interface) depending
\ +* on the value of *\fo{interface-offset}   
    create
    current-interface @ interface-map 2@ \ -- xt map-addr map-size
    dup current-interface @ interface-map-offset @ - , \ store selector-offset
    1 cells extend-mem current-interface @ interface-map 2! ! ( -- )
    current-interface @ interface-offset @ dup ,
    ( 0<> ) if
	do-interface-method
    else
	do-class-method
    then ;

: selector ( "name" -- ) 
\ *G Create selector *\i{"name"} for the current class and its
\ ** descendents; you can set a method for the selector in the
\ ** current class with *\fo{overrides}.
   ['] no-method method
;

: interface-override! ( xt sel-xt interface-map -- )
\ +G xt is the new method for the selector sel-xt in interface-map
   swap >body ( xt map selector-body )
   selector-offset @ + !
;


: class->map ( class -- map ) \ objects- objects
\ +G *\i{map} is the pointer to *\i{class}'s method map; it
\ +* points to the place in the map to which the selector offsets
\ +* refer (i.e., where *\fo{object-map}s point to).
   dup interface-map 2@ drop swap interface-map-offset @ +
;

: unique-interface-map ( class-map offset -- )
\ +G if the interface at offset in class map is the same as its parent,
\ +* copy it to make it unique; used for implementing a copy-on-write policy.
   over @ class-parent @ class->map  \  class-map offset parent-map
   over + @ >r		          \  class-map offset ; R: parent-ifz-map
   + dup @			  \  class-ifz-map-ptr class-ifz-map 
   dup r> =
   if					\  class-ifz-map-ptr class-ifz-map
      dup @ interface-map 2@ nip save-mem drop	
      swap !
   else
      2drop
   then
;

: class-override! ( xt sel-xt class-map -- ) 
\ +G *\fo{xt} is the new method for the selector *\i{sel-xt} in
\ +* *\i{class-map}.
   over >body ( xt sel-xt class-map selector-body )
   selector-interface @ ( xt sel-xt class-map offset )
   ?dup if \ the selector is for an interface
      2dup unique-interface-map
      + @
   then
   interface-override!
;

: overrides ( xt "selector" -- ) \ objects- objects
\ *G replace default method for *\i{"selector"} in the current class
\ ** with *\i{xt}. *\fo{overrides} must not be used during an
\ ** interface definition.
   ' current-interface @ class->map class-override!
;

EXPORT method
EXPORT selector
EXPORT overrides
EXPORT Method-Excp

\ =====================
\ *N Adding persistance
\ =====================

: allot&move				  \ addr1 u1 -- addr2
\ +G Copy a chunk of memory *\i{addr1 u1} to a new *\fo{ALLOT}ed region of
\ +* the dictionary,
\ +* Return the new starting address.
   here dup >r over allot		
   swap move
   r> 
;


: ?allot&move				  \ addr1 u1 -- addr2
\ +G Copy a chunk of memory *\i{addr1 u1} to a new *\fo{ALLOT}ed region of
\ +* the dictionary *\b{if and ony if}
\ +* *\i{addr1 u1} is a heap memory region.  Also releases the heap memory.  
\ +* The heuristic chosen for VFX Forth is that the
\ +* heap memory is always greater then *\fo{HERE} at any time.
\ +* Return the new starting address in any case.      
   over here < if drop exit then
   over >R allot&move
   R> ProtFree 
;


: persist-interface-map		  \ old-ifce-map -- new-ifce-map
\ +G Persist a single interface map into the dictionary memory.
   dup
   @ interface-map 2@ nip		\ get the interface map size
   ?allot&move				\ and reallocate
;


: persist-interface-maps		  \ class --
\ +G Persist all interface maps for *\i{class} that implement interfaces.
\ +* Does nothing if *\i{class} do not implement interfaces.   
   dup interface-map-offset @ swap interface-map 2@ drop
   swap bounds ?do
      i @ ?dup if persist-interface-map i ! then
   1 cells +loop
;

: persistant				  \ --
\ *G Copy the last defined interface map from the heap
\ ** back to the dictionary so that it can be safely saved in turnkey
\ ** applications.
   current-interface @ interface-map 2@ dup >R
   ?allot&move
   R> current-interface @ interface-map 2!
   current-interface @ persist-interface-maps
;      

EXPORT persistant

\ ============= 
\ *N Interfaces
\ ============= 


Variable last-interface-offset
\ *G Every interface gets a different offset; the latest one is stored here.

0 last-interface-offset !

: interface ( -- ) 
\ *G Start an interface definition.
   interface% %allot >r			\ allocates data structure
   r@ current-interface !
   current-interface 1 cells save-mem	\ new interface map in the heap
   r@ interface-map 2!			\ and store it into interface%
   -1 cells last-interface-offset +!
   last-interface-offset @ r@ interface-offset !
   0 r> interface-map-offset !
;


: end-interface-noname ( -- interface ) \ objects- objects
\ +G End an interface definition. The resulting interface structure is
\ +* *\i{interface}.
   current-interface @
;

: end-interface ( "name" -- ) 
\ *G End an interface definition and give it a *\i{"name"}.
   end-interface-noname constant persistant
;

EXPORT interface
EXPORT end-interface

\ =====================
\ *N Vsibility control
\ =====================
 
variable public-wordlist

: protected ( -- ) \ objects- objects
\ *G Set the compilation wordlist to the current class's wordlist.
   current-interface @ class-wordlist @
   dup get-current <>
   if \ we are not protected already
      get-current public-wordlist !
   then
   set-current
;

: public ( -- ) \ objects- objects
\ *G Restore the compilation wordlist that was in effect before the
\ ** last *\fo{protected} that actually changed the compilation
\ ** wordlist.
   current-interface @ class-wordlist @ get-current =
   if \ we are protected
      public-wordlist @ set-current
   then
;

EXPORT protected
EXPORT public

\ ==========
\ *N Classes
\ ==========
 
: add-class-order ( n1 class -- wid1 ... widn n+n1 )
\ +G   
   dup >r class-parent @
   ?dup if
      recurse \ first add the search order for the parent class
   then
   r> class-wordlist @ swap 1+
;

: class>order ( class -- ) 
\ +G Add *\i{class}'s wordlists to the head of the search-order.
   >r get-order r> add-class-order set-order
;


: methods ( class -- ) 
\ *G Makes *\i{class} the current class. This is intended to be
\ ** used for defining methods to override selectors; you cannot
\ ** define new fields or selectors.
   dup current-interface ! class>order
;

: class ( parent-class -- align offset ) 
\ *G Start a new class definition as a child of
\ ** *\i{parent-class}. *\i{align offset} are for use by
\ ** *\fo{field%} etc.
   class% %allot >r
   dup interface-map 2@ save-mem r@ interface-map 2! \ copy ifz-map from parent
   dup interface-map-offset @ r@ interface-map-offset ! \ copy map-offset from parent
   r@ dup class->map !			\ point back to class in class-map
   0 r@ interface-offset !		
   dup r@ class-parent !		\ set this parent
   wordlist r@ class-wordlist !		\ create and set private wordlist
   r> methods				\ and search order and current-interfac
   class-inst-size 2@ 			\ begin with parent inst-size
;

: remove-class-order ( wid1 ... widn n+n1 class -- n1 )
\ +G   
   \ note: no checks, whether the wordlists are correct
   begin
      >r nip 1-
      r> class-parent @ dup 0=
   until
   drop
;

: class-previous ( class -- ) 
\ +G Drop *\i{class}'s wordlists from the search order. No
\ +* checking is made whether *\i{class}'s wordlists are actually
\ +* on the search order.
   >r get-order r> remove-class-order set-order
;


: end-methods ( -- ) 
\ *G Switch back from defining methods of a class to normal mode
\ ** (currently this just restores the old search order).
   current-interface @ class-previous
;

: end-class-noname ( align offset -- class ) 
\ +G End a class definition. The resulting class is *\i{class}.
   public end-methods
   current-interface @ class-inst-size 2!
   end-interface-noname
;


: end-class ( align offset "name" -- ) 
\ *G End a class definition and give it a name. 
   end-class-noname constant
;

EXPORT methods
EXPORT end-methods
EXPORT class
EXPORT end-class

\ ==================================== 
\ *N Classes that implement interfaces
\ ====================================

: front-extend-mem ( addr1 u1 u -- addr addr2 u2 )
\ +G Extend memory block allocated from the heap by u aus, with the
\ +* old stuff coming at the end
   2dup + dup >r ProtAlloc ( addr1 u1 u addr2 ; R: u2 )
   dup >r + >r over r> rot move ( addr1 ; R: u2 addr2 )
   ProtFree
   r> dup r>
;
    
: implementation ( interface -- ) 
\ *G The current class implements *\i{interface}. I.e., you can
\ ** use all selectors of the interface in the current class and its
\ ** descendents.
   dup interface-offset @ ( interface offset )
   current-interface @ interface-map-offset @ negate over - dup 0>
   if \ the interface does not fit in the present class-map
      >r current-interface @ interface-map 2@
      r@ front-extend-mem
      current-interface @ interface-map 2!
      r@ erase
      dup negate current-interface @ interface-map-offset !
      r>
   then ( interface offset n )
   drop >r
   interface-map 2@ save-mem drop ( map )
   current-interface @ dup interface-map 2@ drop
   swap interface-map-offset @ + r> + !
;

EXPORT implementation

\ ====================================
\ *N this/self, instance variables etc.
\ ===================================== 

 
cell +User thisp
\ +G Where the *\fo{this} pointer is really stored.
 
: this ( -- object )
\ *G Thread-safe implementation of 'this'    
   thisp @
;


: to-this ( object -- )
\ *G Thread-safe implementatio to set a new 'this' pointer       
   thisp !
;

  
: enterm ( -- ; run-time: object -- )
\ +G method prologue; *\i{object} becomes new *\fo{this}.
   POSTPONE this
   POSTPONE >r
   POSTPONE to-this
;
    
: m: ( -- xt colon-sys; run-time: object -- )
\ *G Start an anonymous method definition; *\i{object} becomes new *\fo{this}.
\ ** Has to be ended with *\fo{;m}.
   :noname enterm
;

: :m ( "name" -- xt; run-time: object -- ) 
\ *G Start a named method definition; *\i{object} becomes new
\ ** *\fo{this}.  Has to be ended with *\fo{;m}.
   : enterm
;

: exitm ( -- ) \ objects- objects
\ *G *\fo{exit} from a method; restore old *\fo{this}.
   POSTPONE r>
   POSTPONE to-this
   POSTPONE exit
; immediate

: ;m ( colon-sys --; run-time: -- ) 
\ *G End a method definition; restore old *\fo{this}.
   POSTPONE r>
   POSTPONE to-this
   POSTPONE ;
; immediate

\ the following is a bit roundabout; this is caused by the standard
\ disallowing to change the compilation wordlist between CREATE and
\ DOES> (see RFI 3)

: inst-something ( align1 size1 align size xt "name" -- align2 size2 )
   \ xt ( -- ) typically is for a DOES>-word
   get-current >r
   current-interface @ class-wordlist @ set-current
   >r create-field r> execute
   r> set-current
;

: do-inst-var ( -- )
 does> \ name execution: ( -- addr )
   ( addr1 ) @ this +
;

: inst-var ( align1 offset1 align size "name" -- align2 offset2 )
\ *G Define a private variable with a given *\i{align and size} and name.  
   ['] do-inst-var inst-something
;

: do-inst-value ( -- )
does> \ name execution: ( -- w )
  ( addr1 ) @ this + @
;

: inst-value ( align1 offset1 "name" -- align2 offset2 ) 
\ *G Define a private *\fo{VALUE} with a given name.  
   cell% ['] do-inst-value inst-something
;

: <to-inst> ( w xt -- ) 
\ +G store *\i{w} into the field *\i{xt} in *\fo{this} object.
   >body @ this + !
;

: [to-inst] ( compile-time: "name" -- ; run-time: w -- ) 
\ *G Store *\i{w} into field *\i{name} in *\fo{this} object.
   ' >body @ POSTPONE literal
   POSTPONE this
   POSTPONE +
   POSTPONE !
; immediate

EXPORT this
EXPORT m:
EXPORT :m
EXPORT ;m
EXPORT exitm
EXPORT inst-var
EXPORT inst-value
EXPORT [to-inst]

\ =====================
\ *N Exception handling
\ =====================

\ +P VFX Forth uses a couple of words to set/restore the exception stack
\ +* not being necessarily the return stack.

0 [if]
: save-this				  \ -- n
   this >ep 1 
;

: restore-this				  \ n --
   drop ep> to-this 
;

' save-this ' restore-this extends-catch

[else]
   
: catch
   this >ep catch ep> to-this
;

EXPORT catch

[then]


\ ======================
\ *N Class binding stuff
\ ======================

: <bind> ( class selector-xt -- xt ) 
   \ ** @var{xt} is the method for the selector @var{selector-xt} in
   \ ** @var{class}.
   >body swap class->map over selector-interface @
   ?dup if
      + @
   then
   swap selector-offset @ + @
;

: bind' ( "class" "selector" -- xt ) \ objects- objects
   \ ** @var{xt} is the method for @var{selector} in @var{class}.
   ' execute ' <bind>
;

: bind ( ... "class" "selector" -- ... ) \ objects- objects
   \ ** Execute the method for @var{selector} in @var{class}.
   bind' execute
;

: [bind] ( compile-time: "class" "selector" -- ; run-time: ... object -- ... ) 
   \ ** Compile the method for @var{selector} in @var{class}.
   bind' compile,
; immediate

: current' ( "selector" -- xt ) \ objects- objects
   \ ** @var{xt} is the method for @var{selector} in the current class.
   current-interface @ ' <bind>
;

: [current] ( compile-time: "selector" -- ; run-time: ... object -- ... ) 
    \ ** Compile the method for @var{selector} in the current class.
   current' compile,
; immediate

 
: [parent] ( compile-time: "selector" -- ; run-time: ... object -- ... ) 
    \ ** Compile the method for @var{selector} in the parent of the
    \ ** current class.
   current-interface @ class-parent @ ' <bind> compile,
; immediate

\ ** By Anton Ertl in a posting to GForth mailing list, 2008-Jan-10    
: [parent'] ( compilation: "selector" -- ; run-time: -- xt )
   current-interface @ class-parent @ ' <bind> POSTPONE literal
; immediate

EXPORT bind
EXPORT [bind]
EXPORT [parent]
EXPORT [parent']
EXPORT [current]

\ ==============================
\ *N The *\fo{OBJECT} root class
\ ==============================

\ +P Because *\fo{OBJECT} has no parent class, we have to build it by hand
\ +* (instead of with *\fo{class})

class% %allot current-interface !
current-interface 1 cells save-mem current-interface @ interface-map 2!
0 current-interface @ interface-map-offset !
0 current-interface @ interface-offset !
0 current-interface @ class-parent !
wordlist current-interface @ class-wordlist !
object%
current-interface @ class>order

' drop ( object -- )
method construct ( ... object -- ) 
\ *G Initialize the data fields of *\i{object}. The method for the
\ ** root class *\fo{object} just does nothing.

:noname ( object -- )
   ." object:" dup . ." class:" object-map @ @ .
;
method print ( object -- ) 
\ *G Print the object. The method for the root class *\fo{object} prints
\ ** the address of the object and the address of its class structure.

selector equal ( object1 object2 -- flag )
\ *G Default object comparison. Should be overriden by subclasses.

end-class object ( -- class ) \ objects- objects
\ *G the ancestor of all classes.

persistant 				\ make 'object' class persistant

EXPORT object
EXPORT construct			\ silently overrides VFX's 'construct'
EXPORT print
EXPORT equal

\ =======================
\ *N Constructing objects
\ =======================

: init-object ( ... class object -- ) 
\ +G Initialize a chunk of memory (*\i{object}) to an object of
\ +* class *\i{class}; then performs *\fo{construct}.
   swap class->map over object-map ! ( ... object )
   construct
;

: xt-new ( ... class xt -- object ) 
\ +G Make a new object, using either the xt of *\fo{%allot} or *\fo{%alloc} to
\ +* get memory.
   over class-inst-size 2@ rot execute
   dup >r init-object r>
;

: dict-new ( ... class -- object ) 
\ *G *\fo{allot} and initialize an object of class *\i{class} in
\ ** the dictionary.
   ['] %allot xt-new
;

: heap-new ( ... class -- object ) 
\ **  *\fo{allocate} and initialize an object of class *\i{class}.
   ['] %alloc xt-new
;

EXPORT dict-new
EXPORT heap-new

\ =================
\ +N Debugging aids
\ =================

: .interface				\ ifce --
\ +G Print the relevant fields of an *\fo{interface%} structure.   
   >R
   cr
   r@ interface-map 2@ swap   ." interface-map (addr)    = " .LWORD cr
   ." interface-map (size)    = " . cr 
   r@ interface-map-offset @  ." interface-map-offset    = " . cr
   r@ interface-offset @      ." interface-offset        = " . cr
   r> drop
;

: .interfaces				  \ class --
\ +G Print all the interface maps and structures that a class implements.
   dup interface-map-offset @ swap interface-map 2@ drop swap bounds
   ?do
      i @  if
      ." points to interface map = " i @ .lword
      ."  whose interface structure is " i @ @ .lword cr	 
      then
      
   4 +loop
;

: .class				  \ class --
\ +G Print the relevant fields of an *\fo{class%} structure.      
   >r
   r@ .interface
   r@ class-parent   @ ." class-parent            = " .LWORD cr
   r@ class-wordlist @ ." class-wordlist          = " .lword cr
   r@ class-inst-size 2@ swap ." class-inst-size (align) = " . cr
   ." class-inst-size (size)  = " . cr
   r@ .interfaces
   r> drop
;


END-MODULE
