\ TSEPOS 559 9 25 \
\ Object-oriented extensions from Neon/Yerk/Mops

\ Converted and optimised for VFX Forth for Window by
\ MicroProcessor Engineering
\ 24 April 1998

\ Version 1.0, 4 Feb 1997
\ Andrew McKewan

\ 98/11/04 SJC000       Changed name of WIDTH to IDX-WIDTH to stop
\                       collision with existing system variable in
\                       Kernel.
\ 98/11/04 SJC001       Disabled "is redefined" warning around the
\                       redefinition of CATCH.
\ 98/11/04 SJC002       Changes not marked. Removed use of CODE
\                       within system. Change :M and ;M to use ANS
\                       style not carnal knownledge (IE Used :NONAME)
\

\ ======
\ *! oop
\ *T Neon-style Object Oriented Programming
\ ======

\ *P This OOP package can be found in the *\i{Lib\oop\Neon\v1.x}
\ ** directory of the VFX Forth distribution.

( *P This Objected oriented package for VFX Forth for Windows is derived )
( ** from an ANS OOPs package which has its origins in Neon, an object  )
( ** oriented language derived from Forth for the Macintosh. It was     )
( ** converted to an ANS package by Andrew McKewan, and ported to       )
( ** VFX Forth and optimised and extended by MPE. The text below is      )
( ** based on that provided by Andrew McKewan, but has been modified    )
( ** where the code has been modified for VFX Forth for        )
( ** Windows.                                                           )

( *P Extensions to the original implementation have been made so that   )
( ** the code can be used inside tasks and winprocs.                    )

( *S Why do this?                                                       )

( *P When I first began programming in Forth for Windows NT, I became   )
( ** aware of the huge amount of complexity in the environment. In      )
( ** looking for a way to tame this complexity, I studied the           )
( ** object-oriented Forth design in Yerk. Yerk is the Macintosh Forth  )
( ** system that was formerly marketed as a commercial product under    )
( ** the name Neon. It implemented an environment that allowed you to   )
( ** write object-oriented programs for the Macintosh.                  )

( *P While much of Yerk was Macintosh-specific, the underlying          )
( ** class/object/message ideas were quite general. What I hope to      )
( ** accomplish here is to provide any ANS Forth System the ability to  )
( ** use the object-oriented syntax and programming style in these      )
( ** platform-specific systems. In doing so, I have sacrificed some     )
( ** performance and a few of the features.                             )

( *S Object-oriented concepts                                           )

( *P The object-oriented model closely follows Smalltalk. I will first  )
( ** describe the names used in this model: Objects, Classes, Messages, )
( ** Methods, Selectors, Instance Variables and Inheritance.            )

( *P Objects are the entities that are used to build programs. Objects  )
( ** contain private data that is not accessible from outside the       )
( ** object. The only way to communicate with an object is by sending   )
( ** it a message.                                                      )

( *P A Message consists of a selector and arguments. When an            )
( ** object receives the message, it executes a corresponding method.   )
( ** The arguments and results of this method are passed on the Forth   )
( ** stack.                                                             )

( *P A Class is a template for creating objects. Classes describe the   )
( ** instance variables and methods for the object. Once a class is     )
( ** defined, you can make many objects from that same class. Each      )
( ** object has its own copy of the instance variables, but share the   )
( ** method code.                                                       )

( *P Instance variables are the private data belonging to an object.    )
( ** Instance variables can be accessed in the methods of the object,   )
( ** but are not visible outside the object. Instance variables are     )
( ** themselves objects with their own private data and public methods. )

( *P Methods are the code that is executed in response to a message.    )
( ** They are similar to normal colon definitions but use a special     )
( ** syntax using the words :M and ;M. You can put any Forth code       )
( ** inside a method including sending messages to other objects.       )

( *P Inheritance allows you to define a class as a subclass of another  )
( ** class called the superclass. This new class "inherits" all of the  )
( ** instance variables and methods from the superclass. You can then   )
( ** add instance variables and methods to the new class. This can      )
( ** greatly decrease the amount of code you have to write if you       )
( ** design the class hierarchy carefully.                              )

( *S How to define a class                                              )

( *P This example of a Point class illustrates the basic syntax used to )
( ** define a class:                                                    )

( *E :Class Point  <Super Object                                        )
( **   Var x                                                            )
( **   Var y                                                            )
( **                                                                    )
( **   :M Get:   -- x y   Get: x  Get: y  ;M                          )
( **   :M Put:   x y --   Put: y  Put: x  ;M                          )
( **   :M Print:   --                                                 )
( **      Get: self  SWAP  ." x = " .   ." y = " .                      )
( **   ;M                                                               )
( **   :M ClassInit:   1 Put: x  2 Put: y  ;M                           )
( ** ;Class                                                             )

( *P The class Point inherits from the class Object. Object is the root )
( ** of all classes and defines some common behaviour such as getting  )
( ** the address of an object or getting its class but does not have   )
( ** any instance variables. All classes must inherit from a superclass.)

( *P Next we define two instance variables, x and y. Both of these are  )
( ** instances of class Var. Var is a basic cell-sized class similar to )
( ** a Forth variable. It has methods Get: and Put: to fetch and store  )
( ** its data.                                                          )

( *P The Get: and Put: methods of class Point access its data as a pair )
( ** of integers. They are implemented by sending Get: and Put:         )
( ** messages to the instance variables. Print: prints out the x and y  )
( ** coordinates.                                                       )

( *P ClassInit: is a special initialisation method. Whenever an object  )
( ** is created, the system sends it a ClassInit: message. This allows  )
( ** the object to perform any initialisation functions. Here we        )
( ** initialise the variables x and y to a preset value. Whenever a     )
( ** point is created, it will be initialised to these values. This is  )
( ** similar to a constructor in C++.                                   )

( *P Not all classes need a ClassInit: method. If a class does not      )
( ** define the ClassInit: method, there is one in class Object that    )
( ** does nothing.                                                      )

( *S Creating an instance of a class                                    )

( *P Now we have defined the Point class, let's create a point:         )

( *E Point myPoint                                                      )

( *P As you can see, Point is a defining word. It creates a Forth       )
( ** definition called myPoint. Let's see what it contains:             )

( *E Print: myPoint                                                     )

( *P This should print the text "x = 1 y = 2" on the screen. You can    )
( ** see that the new point has been initialised with the ClassInit:    )
( ** message.                                                           )

( *P Now we can modify myPoint and we should see the new value:         )

( *E 3 4 Put: myPoint                                                   )
( ** Print: myPoint                                                     )

( *P Notice that in the definition of Point, we created two instance    )
( ** variables of class Var. The object defining words are "class smart")
( **  and will create instance variables if used inside a class and     )
( ** global objects if used outside of a class.                         )

( *S Sending a message to yourself                                      )

( *P In the definition of Print: we used the phrase Get: self. Here we  )
( ** are sending the Get: message to ourselves. Self is a name that     )
( ** refers to the current object. The compiler will compile a call to  )
( ** Point's Get: method. Similarly, we could have defined ClassInit:   )
( ** like this:                                                         )

( *E :M ClassInit:   1 2 Put: self  ;M                                  )

( *P This is a common factoring technique in Forth and is equally       )
( ** applicable here.                                                   )

( *S Creating a subclass                                                )

( *P Let's say we wanted an object like myPoint, but one that printed   )
( ** itself in a different format.                                      )

( *E :Class NewPoint  <Super Point                                      )
( **   :M Print:   --   Get: self  SWAP  0 .R ." @" .  ;M             )
( ** ;Class                                                             )

( *P A subclass inherits all of the instance variables of its           )
( ** superclass, and can add new instance variables and methods of its  )
( ** own, or override methods defined in the superclass. Now lets try   )
( ** it out:                                                            )

( *E NewPoint myNewPoint                                                )
( ** Print: myNewPoint                                                  )

( *P This will print "1@2" which is the Smalltalk way of printing       )
( ** points. We have changed the Print: method but have inherited all   )
( ** of the other behaviours of a Point.                                )

( *S Sending a message to your superclass                               )

( *P In some cases, we do not want to replace a method but just add     )
( ** something to it. Here's a class that always prints its value on a  )
( ** new line:                                                          )

( *E :Class CrPoint  <Super NewPoint                                    )
( **   :M Print:   --   CR  Print: super  ;M                          )
( ** ;Class                                                             )
( ** CrPoint myCrPoint                                                  )
( ** Print: myCrPoint                                                   )

( *P When we use the phrase *\fo{Print: super} we are telling the      )
( ** compiler to send the print message that was defined in our         )
( ** superclass.                                                        )

( *S Windows messages and integers                                      )

( *P When implementing winprocs and programming Windows, it is often    )
( ** useful to be able to treat the Windows message such as             )
( ** *\fo{WM_CHAR} as a method selector. The VFX Forth for Windows      )
( ** implementation allows any Windows message or integer to be used as )
( ** a selector.                                                        )

( *E :M WM_CHAR ... ;M                                                    )
( ** :M $55AA ... ;M                                                      )

( *P A pseudo selector WM: has been defined that treats an item on the  )
( ** stack as a selector. Inside a winproc or other code, use WM: to    )
( ** pass the selector to the object.                                   )

( *E <message/int> WM: object                                           )
( ** or                                                                 )
( ** <message/int> WM: [ object ]                                       )

( *P At a lower level, useful for passing Windows messages to an object )
( ** without knowing the message until runtime, use MSG>OBJ which       )
( ** requires a selector and an object reference. If the method is      )
( ** applicable to the object, the appropriate method is executed and   )
( ** true is returned, otherwise false is returned.                     )

( *E <message> <object> MSG>OBJ                                         )
( **  \ i*x message ^object -- j*x true                                 )
( **  \ i*x message ^object -- i*x false                                )

( *S Indexed instance variables                                         )

( *P Class Point had two named instance variables, "x" and "y."  The    )
( ** type and number of named instance variables is fixed when the      )
( ** class is defined. Objects may also contain indexed instance        )
( ** variables. These are accessed via a zero-based index. Each object  )
( ** may define a different number of indexed index variables. The size )
( ** of each variable is defined in the class header by the word        )
( ** <Indexed.                                                          )

( *E :Class Array  <Super Object  CELL <Indexed                         )
( **   :M At:   index -- value   At  ;M                             )
( **   :M To:   value index --   To  ;M                             )
( ** ;Class                                                             )

( *P We have declared that an Array will have indexed instance variables)
( ** that are each CELL bytes wide. To define an array, put the number  )
( ** of elements before the class name:                                 )

( *E 10 Array myArray                                                   )

( *P This will define an Array with 10 elements, numbered from 0 to 9.  )
( ** We can access the array data with the At: and To: methods:         )

( *E 4 At: myArray .                                                    )
( ** 64 2 To: myArray                                                   )

( *P Indexed instance variables allow the creation of arrays, lists and )
( ** other collections.                                                 )

( *S Early vs. late binding                                             )

( *P In these examples, you may have been thinking, "all of this message)
( ** sending must be taking a lot of time." In order to execute a       )
( ** method, an object must look up the message in its class, and then  )
( ** its superclass, until it is found.                                 )

( *P But if the class of the object is known at compile time, the       )
( ** compiler does the lookup then and compiles the execution token of  )
( ** the method. This is called "early binding." There is still some    )
( ** overhead with calling a method, but it is quite small. In all of   )
( ** the code we have seen so far, the compiler will do early binding.  )

( *P There are cases when you do want the lookup to occur at runtime.   )
( ** This is called "late binding." An example of this is when you have )
( ** a Forth variable that will contain a pointer to an object, yet the )
( ** class of the object is not known until runtime. The syntax for     )
( ** this is:                                                           )

( *E VARIABLE objPtr   myPoint objPtr !                                 )
( ** Print: [ objPtr @ ]                                                )

( *P The expression within the brackets must produce an object address. )
( ** The compiler recognised the brackets and will do the message       )
( ** lookup at runtime.                                                 )

( *P Don't worry, "[" or "]"  have not been redefined. When a message  )
( ** selector recognises the left bracket, it uses PARSE and EVALUATE   )
( ** to compile the intermediate code and then compiles a late-bound    )
( ** message send. This also works in interpret state.                 )

( *S Class binding                                                      )

( *P Class binding is an optimisation that allows us to get the         )
( ** performance of early binding when we have object pointers or       )
( ** objects that are passed on the stack. If we use a selector with a  )
( ** class name, the compiler will early bind the method, assuming that )
( ** an object of that class is on the stack. So if we write a word to  )
( ** print a point like this,                                           )

( *E : .Point   aPoint --    Print: Point ;                           )
( ** objPtr @ .Point                                                    )

( *P it will early bind the call. If you pass anything other that a     )
( ** Point, you will not get the expected result It will print the     )
( ** first two cells of the object, no matter what they are. This is   )
( ** an optimisation technique that should be used with care until a    )
( ** program is fully debugged.                                         )

( *S Creating objects on the heap                                       )

( *P If a system has dynamic memory allocation, the programmer may want )
( ** to create objects on the heap at runtime. This may be the case,    )
( ** for instance, if the programmer does not know how many objects     )
( ** will be created by the application.                                )

( *P The syntax for creating an object on the heap is:                  )

( *E Heap> Point objPtr !                                               )

( *P Heap> will return the address of the new point, which can be kept  )
( ** on the stack or stored in a variable. To release the point and     )
( ** free its memory, we use:                                           )

( *E objPtr @ Release                                                   )

( *P Before the memory is freed, the object will receive a Release:     )
( ** message. It can then do any cleanup necessary like releasing      )
( ** other instance variables. This is similar to a C++ destructor.    )

( *S Implementation                                                     )

( *P The address of the current object is stored in the user variable   )
( ** CURROBJ, and the contents are returned by ^BASE. This means that   )
( ** the only time you can use ^BASE is inside a method. Whenever a     )
( ** method is called, ^BASE is saved and loaded with the address of    )
( ** the object being sent the message. When the method exits, ^BASE is )
( ** restored.                                                          )

( *S Class structure                                                    )

( *P All offsets and sizes are in Forth cells.                          )

( *E Offset	Size cells	Name	Description                     )
( ** 0	8	MFA	Method dictionary 8-way hashed list           )
( ** 8	1	IFA	Linked-list of instance variables               )
( ** 9	1	DFA	Data length of named instance variables         )
( ** 10	1	XFA	Width of indexed instance variables             )
( ** 11	1	SFA	Superclass pointer                              )
( ** 12	1	TAG	Class tag field                                 )
( ** 13	User defined    User-defined field                              )

( *P The first 8 cells are an 8-way hashed list of methods. Three bits  )
( ** from the method selector are used to determine which list the      )
( ** method may be in. This cuts down search time for late-bound        )
( ** messages.                                                          )

( *P The IFA field is a linked list of named instance variables. The    )
( ** last two entries in this list are always "self" and "super."       )

( *P The DFA field contains the length of the named instance variables  )
( ** for an object.                                                     )

( *P The XFA field actually serves a dual role. For classes with        )
( ** indexed instance variables it contains the width of each element.  )
( ** For non-indexed classes this field is usually zero. A special      )
( ** value of -1 is a flag for general classes see below.             )

( *P The TAG field contains a special value that helps the compiler     )
( ** determine if a structure really represents a class.                )

( *P The USR field is not used by the compiler but is reserved for a    )
( ** programmer's use. In the future I may extend this concept of       )
( ** "class variables" to allow adding to the class structure. This     )
( ** field is used in a Windows implementation to store a list of       )
( ** window messages the class will respond to.                         )

( *S Object structure                                                   )

( *P The first field of a global or heap-based object is a pointer to   )
( ** the object's class. This allows us to do late binding. Normally,   )
( ** the class field is not stored for an instance variable. This saves )
( ** space and is not usually needed because the compiler knows the     )
( ** class of the instance variable and the instance variable is not    )
( ** visible outside the class definition. For indexed classes, the     )
( ** class pointer is always stored because the class contains          )
( ** information needed to locate the indexed data. Also, the           )
( ** programmer may mark a class as "general" so that the class pointer )
( ** is always stored. This is needed in cases where the object sends   )
( ** itself late-bound messages i.e. msg: [ self ].                   )

( *E Offset	Size cells	Description                             )
( ** 0	        1	        Pointer to object's class               )
( ** 1	        DFA	        Named instance variable data            )
( ** DFA+1	1	        Number of indexed instance variables    )
( ** DFA+2	?	        Indexed instance variables if indexed )

( *P When an object executes, it returns the address of the first named )
( ** instance variable. This is what we refer to when we mean the       )
( ** "object address." This field contains the named instance variable  )
( ** data. Since instance variables are themselves objects, this        )
( ** structure can be nested indefinitely.                              )

( *P Objects with indexed instance variables have two more fields. The  )
( ** indexed header contains the number of indexed instance variables.  )
( ** The width of the indexed variables is stored in the class          )
( ** structure, which is why we must always store a class pointer for   )
( ** indexed objects.                                                   )

( *P Following the indexed header is the indexed data. The size of this )
( ** area is the product of the indexed width and the number of         )
( ** elements. There are primitives defined to access this data area.   )

( *S Instance variable structure                                        )

( *P The link field points to the next instance variable in the class.  )
( ** The head of this list is the IFA field in the class. When a new    )
( ** class is created, all the class fields are copied from the         )
( ** superclass and so the new class starts with all of the instance    )
( ** variables and methods from the superclass.                         )

( *E Offset cells	Size cells	Name	Description             )
( ** 0	                1	        Link	points to link of next  )
( **                                            ivar in chain           )
( ** 1	                1	        Name	hash value of name      )
( ** 2	                1	        Class	pointer to class        )
( ** 3	                1	        Offset	offset in object to     )
( **                                            start of ivar data      )
( ** 4	                1	        #elem	number of elements      )
( **                                            indexed ivars only    )

( *P The name field is a hash value computed from the name of the       )
( ** instance variable. This could be stored as a string with a space   )
( ** and compile-time penalty. But with a good 32-bit hash function     )
( ** collisions are not common. In any event, the compiler will abort   )
( ** if you use a name that collides with a previous name. You can      )
( ** rename your instance variable or improve the hash function.        )

( *P Following the name is a pointer to the class of the instance       )
( ** variable. The compiler will always early-bind messages sent to     )
( ** instance variables.                                                )

( *P The offset field contains the offset of this instance variable     )
( ** within the object. When sending a message to an object, this       )
( ** offset is added to the current object address.                     )

( *P If the instance variable is indexed, the number of elements is     )
( ** stored next. This field is not used for non-indexed classes.       )

( *P Unlike objects, instance variables are not names in the Forth      )
( ** dictionary. Correspondingly, you cannot execute them to get their  )
( ** address. You can only send them messages. If you need an address,  )
( ** you can use the Addr: method defined in class Object.              )

( *S Method structure                                                   )

( *P Methods are stored in an 8-way linked-list from the MFA field. A   )
( ** 32-bit selector identifies each method, which is the parameter     )
( ** field address of the message selector.                             )

( *E Offset cells	Size cells	Description                     )
( ** 0	                1	        Link to next method             )
( ** 1	                1	        Selector                        )
( ** 2	                1	        Method execution token          )

( *P Methods are defined using the words :M and ;M which act like the   )
( ** familiar words : and ; except that they compile coded routines     )
( ** that manage the current object pointer as well as handling the     )
( ** procedure call. VFX Forth's use of coded routines makes the method  )
( ** call overhead negligible.                                          )

( *S Selectors are special words                                        )

( *P In the Yerk implementation, the interpreter was changed by        )
( ** vectoring FIND so that it automatically recognised words ending   )
( ** in ":" as a message to an object. It computed a hash value from    )
( ** the message name and used this as the selector. This kept the      )
( ** dictionary small.                                                  )

( *P In ANS Forth, there is no way to modify the interpreter short of  )
( ** writing a new one. It has also been argued whether or not this is )
( ** a "good thing" anyway.                                             )

( *P In this implementation, message selectors are immediate Forth      )
( ** words. They are created automatically the first time they are used )
( ** in a method definition. Since they are unique words, we use the    )
( ** parameter field of the word as the selector.                       )

( *P When the selector executes it compiles or executes code to send a  )
( ** message to the object that follows. If used inside a class, it     )
( ** first looks to see if the word is one of the named instance        )
( ** variables. If not, it sees if it is a valid object. Lastly it sees )
( ** if it is a class name and does class binding.                      )

( *P Yerk also allowed sending messages to values and local variables   )
( ** and automatically compiled late-bound calls. In ANS Forth, we      )
( ** cannot tell anything about these words from their execution token, )
( ** so this feature is not implemented. We can achieve the same effect )
( ** by using explicit late binding:                                    )

( *E Message: [ aValue ]                                                )

( *S Object initialisation                                              )

( *P When an object is created, it must be initialised. The memory for  )
( ** the object is cleared to zero and the class pointer and indexed    )
( ** header are set up. Then each of the named instance variables is    )
( ** initialised.                                                       )

( *P This is done with the recursive word ITRAV. It takes the address   )
( ** of an instance variable structure and an offset and follows the    )
( ** chain, initialising each of the named instance variables in the    )
( ** class and sending it a ClassInit: message. As it goes it           )
( ** recursively initialises that instance variable's instance          )
( ** variables, and so on.                                              )

( *P Finally, the object is sent a ClassInit: message. This same        )
( ** process is followed when an object is created from the heap.       )

( *S Example classes                                                    )

( *P Some simple classes have been implemented to serve as a basis for  )
( ** your own class library. These classes have similar names and       )
( ** methods to the predefined classes in Yerk and Mops. The code for   )
( ** the class implementation and sample classes is available from the  )
( ** Forth Interest Group FIG FTP site:                               )

( *P ftp://ftp.forth.org/pub/Forth/ANS/CLASS01.ZIP                      )
( *P http://www.forth.org                                               )

( *S Conclusions                                                        )

( *P For me, the primary benefit of using objects is in managing        )
( ** complexity. Objects are little bundles of data that understand and )
( ** act on messages sent by other parts of the program. By keeping the )
( ** implementation details inside the object, it appears simpler to    )
( ** the rest of the program. Inheritance can help reduce the amount of )
( ** code you have to write. If a mature class library is available,    )
( ** you can often find needed functionality already there.             )

( *P If the Forth community could agree on a object-oriented model, we  )
( ** could begin to assemble an object-oriented Forth library similar   )
( ** to the Forth Scientific Library project headed by Skip Carter,     )
( ** code and tools that all Forth programmers can share. That project  )
( ** had not been possible before the ANS standardisation of            )
( ** floating-point in Forth.                                           )
( *> ###                                                                )

DECIMAL

[forth]

vocabulary oop
also oop definitions

\ =====================================================================
\ Misc words. You may already have some of these.

have reserve 0= [if]
: RESERVE  ( n -- )  HERE OVER ERASE ALLOT ; doNotSin
[then]

have link 0= [if]
\ Build link to list head at addr
: LINK  ( addr -- )   HERE  OVER @ ,  SWAP ! ;
[then]

\ =====================================================================
\ Class Structure.

0 VALUE ^Class          \ pointer to class being defined
0 VALUE newObject       \ object being created
0 VALUE (ClassInit:)    \ selector for ClassInit: message

\ The following offsets refer to the ^Class, or Pfa of the class.
\ The first 8 cells are the hashed method dictionary.
struct classSize
  8 cells field HMD     \ the hashed method dictionary (HMD)
  cell    field IFA     \ ivar dict Latest field
  cell    field DFA     \ datalen of named ivars
  cell    field XFA     \ width of indexed area, <= 0 if not indexed
  cell    field SFA     \ superclass ptr field
  cell    field TAG     \ class tag field
end-struct


CREATE classTag         \ contents of tag field for valid class

: ?isClass  ( pfa -- f )  \ is this a valid class?
        TAG @ classTag = ;

: ?isObj  ( pfa -- )  \ is this a valid object?
        @ DUP IF ?isClass THEN ;

: classAllot  ( n -- )  \ Allot space in the current class
        ^Class DFA +! ;

: classAlign  ( -- )  \ Align class data size (optional)
        ^Class DFA @  ALIGNED  ^Class DFA ! ;

: @width  ( ^class -- elWidth )  \ return the indexed element width for a class
        XFA @  0 MAX ;

\ Error if not compiling a new class definition
: ?Class   ^Class   0= ABORT" Not inside a class definition" ;


\ =====================================================================
\ Objects have a pointer to their class stored in the first cell of
\ their pfa. When they execute, they return the address of the cell
\ following the class pointer, which is location of the first named
\ instance variable.
\
\ Object structure:     | ^class | named ivars |
\
\ If an object is indexed, an indexed header appears after the data area.
\ This header consists of a cell containing the number of elements.
\ The indexed data follows this header.
\
\ Indexed object:       | ^class | named ivars | #elems | indexed ivars |
\ =====================================================================

: (Obj)  ( -- )  CREATE  DOES> CELL+ ;


: >obj  ( xt -- ^obj )  \ get the object address from the execution token
        >BODY CELL+ ;

: >class  ( ^obj -- ^class )  \ get the class of an object
        CELL - @ ;


\ =====================================================================
\ Methods are stored in an 8-way linked-list from the MFA field.
\ Each method is identified by a 32-bit selector which is the parameter
\ field of the selector. Offsets are in cells.
\
\ Method Structure:
\       0   link to next method
\       1   selector
\       2   method execution token (called mcfa below)
\
\ =====================================================================

\ Find the top of the method link for a given selector.
\ The "2/ 2/" below is to get a better distribution if the selectors
\ are aligned values.

: MFA           \ SelID ^Class -- SelID MFA
       OVER 2/ 2/ 7 AND CELLS + ;

\ Search through a linked-list of methods for the given selector.

: ((FINDM))  ( SelID MFA -- mcfa true | false )
       BEGIN   @ DUP
       WHILE   2DUP CELL+ @ = IF  2 CELLS + ( mfca )  NIP TRUE  EXIT THEN
       REPEAT  NIP ;

: (FINDM)       \ SelID ^Class -- xt ; find method in a class
        MFA ((FINDM)) IF  @ EXIT  THEN
        TRUE ABORT" Message not understood by class" ;

: FIND-METHOD   ( SelID ^obj -- ^obj xt )   \ find method in object's class
        TUCK >class (FINDM) ;


\ ======================================================================
\ Methods are compiled in the same way as colon definitions, except that
\ the entry and exit code is different. On entry, the object address is
\ on the data stack. The previously current object is preserved on the
\ return stack by NESTM, and the provided object become the current
\ object. UNNESTM restores the previous current object.
\ ======================================================================
\ The current object address is stored in the user variable
\ CURROBJ, which can be retrieved using the code word ^BASE.
\ The object is only valid inside a method definition. When we call a
\ method, we save the old object pointer and set it to the current object.
\ When the method returns, we restore the object pointer.
\ ======================================================================


: nestm         \ obj -- ; R: -- obj addr ; performs the run-time action of : but saves CURROBJ as well
  r> currobj @ >r >r currobj !
;

: unnestm       \ -- ; R: obj addr -- ; return from method call
  r> r> currobj ! >r
;

: +objlit       \ -- ^obj ; return an object address from an in-line literal + current object
  r@ @ currobj @ +
  r> cell+ >r
;

also forth definitions

: ^base         \ -- ^obj ; return the current object
  currobj @
;

previous definitions


: EXECUTE-METHOD        \ ^obj xt -- ; execute method, saving object pointer
  execute  ;

: EXECUTE-IVAR          \ xt offset -- ; execute ivar method at given offset
  ^base +  swap execute  ;


\ Wrap catch so that it preserves the current object
HAVE CATCH [IF]
also forth definitions
also system -warnings previous                                                  \ SJC001
: CATCH  ( -- n )  currobj @ >R   CATCH   R> currobj !  ;
also system +warnings previous                                                  \ SJC001
previous definitions
[THEN]


\ =====================================================================
\ For late-bound method calls, we compile code to look up the method
\ at runtime and execute it. The syntax is:
\
\       Selector: [ object ]
\
\ The code between [ and ] must return an object reference. If the method
\ is not found in the class of the object, a runtime error occurs.
\ Because we use PARSE, [ and ] must be on the same source line.
\ =====================================================================

: (Defer)  ( ^obj SelId -- )  OVER >class (FINDM) EXECUTE-METHOD ;
: Defer,   ( SelId -- )  POSTPONE LITERAL  POSTPONE (Defer) ;

: Defered  ( SelId -- )         \ Compile or execute a defered message send
        >R  [CHAR] ] PARSE  EVALUATE  R>
        STATE @ IF  Defer,  ELSE  (Defer)  THEN ;

\ True if string is "[" to start defered message send
: ?isParen  ( str -- f )   1+ C@ [CHAR] [ = ;


\ =====================================================================
\ Hash function for instance variable names. The "32 OR" is for
\ case-insensitive names. The compiler will warn you if you have
\ a hash collision.

: HASH  ( addr len -- n )
        TUCK BOUNDS ?DO  5 LSHIFT  I C@ 32 OR XOR  LOOP
        DUP 0< IF EXIT THEN  INVERT ;

: hash>  ( -- n )  BL WORD COUNT HASH ;


\ =====================================================================
\ Instance variable consists of five cell-sized fields. The fifth field
\ is used for indexed ivars only. Offsets are in cells.
\
\     Offset   Name      Description
\     ------   ----      ---------------------------------------
\        0     link      points to link of next ivar in chain
\        1     name      32-bit hash value of name
\        2     class     pointer to class
\        3     offset    offset in object to start of ivar data
\        4     #elem     number of elements (indexed ivars only)
\
\ In the stack diagrams, "ivar" refers to the starting address of this
\ structure.  The IFA field of a class points to the first ivar.
\ =====================================================================

: iclass     ( ivar -- 'class )   2 CELLS +   ;
: @IvarOffs  ( ivar -- offset )   3 CELLS + @ ;
: @IvarElems ( ivar -- #elems )   4 CELLS + @ ;


\ =====================================================================
\ Build SUPER and SELF pseudo ivars. These are always the last
\ two ivars in a class. When we define a class, we will patch the
\ class fields to the appropriate class and superclass.

CREATE ^Self
        0 ,             \ link
        hash> self ,    \ name
        0 ,             \ class
        0 ,             \ offset

CREATE ^Super
        ^Self ,         \ link
        hash> super ,   \ name
        0 ,             \ class
        0 ,             \ offset


\ Create a dummy class that "object" inherits from.

CREATE Meta     classSize RESERVE

^Super   Meta IFA !     \ latest ivar
classTag Meta TAG !     \ class tag


\ =====================================================================
\ Determine if next word is an instance var.
\ Return pointer to class field in ivar structure.

: ivarFind  ( str -- str false | ^iclass true )
        ^Class
        IF      DUP COUNT HASH ^Class IFA ((FINDM))
                DUP IF  ROT DROP  THEN
        ELSE    FALSE
        THEN ;

\ send ClassInit: message to ivar on stack
: InitIvar  ( ivar offset -- )
        OVER @IvarOffs + newObject +   ( ivar addr )
        (ClassInit:) ROT iclass @ (FINDM) EXECUTE-METHOD ;

: ClassInit  ( -- )  \ send ClassInit: to newObject
        newObject (ClassInit:)
        newObject >class (FINDM) EXECUTE-METHOD ;


\ =====================================================================
\ ITRAV traverses the tree of nested ivar definitions in a class,
\ building necessary class pointers and indexed area headers.

[-short-branches
: ITRAV   ( ivar offset -- )  >R  ( ivar -- )
        BEGIN   DUP ^Super <>
        WHILE   DUP iclass @ IFA @
                OVER @IvarOffs R@ + RECURSE  ( initialize ivar's ivars )

                DUP iclass @  ( get ivar class )
                DUP XFA @  ( needs class pointer? )
                IF      OVER @IvarOffs R@ + newObject +
                        ( ivar ^Class ivarAddr -- )
                        2DUP CELL - !                   \ store class pointer
                        OVER @width  ( indexed? )
                        IF      SWAP DFA @ +            \ addr of indexed area
                                OVER @IvarElems         \ #elems
                                SWAP !                  \ upper array limit
                        ELSE    2DROP
                        THEN
                ELSE    DROP
                THEN
                DUP R@ InitIvar         \ send ClassInit:
                @                       \ next ivar in chain
        REPEAT  R> 2DROP ;  doNotSin
short-branches]


\ =====================================================================
\ Compile an instance variable dictionary entry.

also forth definitions
: <Var  ( #elems ^Class | ^Class -- )
        BL WORD ivarFind ABORT" Duplicate instance variable"
        COUNT HASH                              \ get hash value of name
        ALIGN ^Class IFA LINK ,                 \ link & name
        DUP ,                                   \ class
        DUP XFA @ IF  CELL classAllot  THEN     \ if indexed, allow for class ptr
        ^Class DFA @ ,                          \ offset to ivar data
        DUP @width DUP
        IF  ROT DUP ,  * CELL+  THEN            \ #elems, cell for idx-hdr
        SWAP DFA @ + classAllot ; doNotSin	\ Account for named ivar lengths
previous definitions


\ =====================================================================
\ Build an instance of a class. If we are inside a class definition,
\ build an instance variable. Otherwise build a global object.

\ Compile the indexed data header into an object
: IDX-HDR   ( #elems ^Class | ^Class -- indlen )
        @width DUP IF  OVER , ( limit )  *  THEN ;

: (Build)   ( #elems ^Class | ^Class -- )
        ^Class
        IF      <Var    \ build an ivar if we are inside a class
        ELSE
                (Obj)                   \ create object
                DUP >R ,                \ store class pointer
                HERE TO newObject       \ remember current object
                R@ DFA @ RESERVE        \ allot space for ivars
                R@ IDX-HDR RESERVE      \ allot space for indexed data
                R> IFA @ 0 ITRAV        \ init instance variables
                ClassInit               \ send CLASSINIT: message
        THEN ;


\ =====================================================================
\ Build a class header with its superclass pointer.

also forth definitions
: :Class  ( -- )
        CREATE  0 TO ^Class
        DOES>  (Build) ;

: <Super  ( -- )
        HERE TO ^Class                  \ save current class
        classSize ALLOT                 \ reserve rest of class data
        ' >BODY                         \ pfa of superclass
        DUP ^Class classSize MOVE       \ copy class data
        DUP ^Class SFA !                \ store pointer to superclass
        ^Super iclass !                 \ store superclass in SUPER
        ^Class ^Self iclass ! ;         \ store my class in SELF

: ;Class  ( -- )
        classAlign              \ align class data size (optional)
        0 ^Super iclass !       \ clear out super and self class pointers
        0 ^Self  iclass !
        0 TO ^Class ;           \ clear class compiling flag
previous definitions

\ =====================================================================
\ Object Compiler. We rely on being able to classify the type of
\ object from it's execution token. There is no general way to
\ do this in ANS forth for builtin types such as VALUEs. So we
\ only allow message sends to objects and classes. In Yerk, the
\ following will send a late-bound message to an object pointer:
\
\       0 VALUE theObject  ' myObject TO theObject
\
\       Message: theObject
\
\ Here we will have to use the following syntax (which does the same
\ this and is also allowed in Yerk):
\
\       Message: [ theObject ]
\
\ Key to instantiation actions
\ 1 = objType           defined as an object
\ 2 = classType         as a class
\ 5 = parenType         open paren for defer group
\
\ =====================================================================

\ ( str -- xt tokenID )  Determine type of token referenced by string.
: refToken
        DUP ?isParen       IF  3 EXIT  THEN
        FIND 0= ABORT" undefined object"
        DUP >BODY ?isObj   IF  1 EXIT  THEN
        DUP >BODY ?isClass IF  2 EXIT  THEN
        TRUE ABORT" Invalid object type" ;

: (ivarRef)   ( xt offset -- )          \ compile ivar reference
        ?dup                            \ can optimise if offset=0
        if
          postpone +objlit ,            \ will be added to CURROBJ, offset is inline
        else
          postpone ^base                \ offset=0, just get CURROBJ
        endif
        compile,  ;

((
: ivarRef   ( selID ^iclass -- )        \ compile ivar reference
        CELL+ FIND-METHOD  SWAP @  ( xt offset )  ?DUP
        IF      (ivarRef)
        ELSE    COMPILE,  ( optimize for offset zero )
        THEN ;
))
: ivarRef   ( selID ^iclass -- )        \ compile ivar reference
        CELL+ FIND-METHOD  SWAP @  ( xt offset )
        (ivarRef)  ;

: callMethod  ( xt -- )   \ compile method call
        compile,  ;

: (objRef)   ( SelID objCfa -- )        \ compile object reference
        >obj FIND-METHOD SWAP ( xt ^obj )
        POSTPONE LITERAL  callMethod ;

\ ( selID $str -- )  Build a reference to an object or vector
: objRef
        refToken CASE
          1 ( object ) OF  (objRef)                     ENDOF
          2 ( class  ) OF  >BODY (FINDM) callMethod     ENDOF
          3 ( paren  ) OF  DROP Defered                 ENDOF
        ENDCASE ;

\ ( selID $str -- )  Execute using token in stream
: runRef
        refToken CASE
          1 ( object ) OF  >obj FIND-METHOD             ENDOF
          2 ( class  ) OF  >BODY (FINDM)                ENDOF
          3 ( paren  ) OF  DROP Defered EXIT            ENDOF
        ENDCASE  EXECUTE-METHOD ;


\ =====================================================================
\ Selectors are immediate words that send a message to the object
\ that follows. The Yerk requirement that selectors end in ":" is
\ enforced here but not otherwise required by the implementation.

\ This is the message compiler invoked by using a selector.
: comp-message   ( SelID -- )
        BL WORD  STATE @
        IF      ivarFind        \ instance variable?
                IF    ivarRef   \ ivar reference
                ELSE   objRef   \ compile object/vector reference
                THEN
        ELSE    runRef          \ run state - execute object/vector ref
        THEN ;

: ?isSel  ( str -- flag )  \ true if word at addr is a selector xxx:
        DUP DUP C@ CHARS +  C@ [CHAR] :  =  SWAP C@ 1 > AND ;

: ?Selector ( -- )  \ Verify that following word is valid selector
        >IN @  BL WORD ?isSel 0= ABORT" Not a selector"  >IN ! ;

also forth definitions
\ Create a selector that sends a message when executed.
: Selector  ( n -- )  ?Selector
        CREATE IMMEDIATE  DOES> comp-message ;

\ If the selector already exists, just return the existing selector,
\ otherwise create a new selector.
: getSelect     \ -- n
  >IN @  BL WORD FIND                   \ does it exist?
  IF
    >BODY NIP                           \ y: already defined
  ELSE
    case  integer?                      \ n: try integer or Windows constant
      0  of  >in !  Selector  here  endof
      1  of  nip  endof
      2  of  2drop  >in !  Selector  here  endof
             1 abort"  GETSELECT: invalid response from INTEGER?"
    endcase
  THEN
;

Selector ClassInit:    getSelect ClassInit: TO (ClassInit:)

\ Allow a selector to be an integer or a Windows constant. In this case
\ the syntax is:
\   <int/lit> WM: object
\ or
\   <int/lit> WM: [ object ]
: WM:           \ SelID --
  state @
  if
    here 2 cells - @ ['] lit <>         \ preceded by a literal?
    abort" WM: must be preceded by a selector constant"
    here cell - @                       \ retrieve it
    -2 cells allot                      \ remove from compilation stream
  endif
  comp-message
;  immediate
previous definitions

: msg>obj       \ i*x SelID ^obj -- j*y true | i*x false ; execute message in object
                \ returns true if executed, false if message not found
  tuck >class MFA ((findm))             \ is message acceptable?
  if
    @ execute-method  true
  else
    drop  0
  endif
;


\ =====================================================================
\ Build a methods dictionary entry. :M starts a method definition
\ by adding to the class method list and starting the compiler with
\ :NONAME. ;M ends a method and saves the method xt.

also forth definitions
: :M    \ -- addr xt
        ?Class
        getSelect
        DUP ^Class MFA ((FINDM))                \ is method already defined?
        IF
                ^Class U> ABORT" Method redefined in same class"
        THEN
        ALIGN                           \ align method
        ^Class MFA LINK                 \ link into method chain
        ( SelID ) ,                     \ name is selector's hashed value
        HERE 0 ,                        \ save location for method xt

        :noname
          ['] nestm compile,
;

: ;M    \ addr xt -- ; ends a method definition
        ?Class
        ['] unnestm compile,
        postpone ;
        SWAP !                          \ save xt
; IMMEDIATE
previous definitions

\ =====================================================================
\ Build a new object on the heap for class. Use: Heap> className
\ gets heap, and returns ptr. Throws an error if not enough memory

HAVE ALLOCATE [IF]

also forth definitions
: ?MEMERR  ( ior -- )  ABORT" Memory allocation error" ;

Selector Release:       \ sent to an object before it is freed
previous definitions

: allocObj  ( size class -- )  \ allocate object and store in newObject
        OVER CELL+              \ allow for class ptr
        ALLOCATE ?MEMERR        \ ( size class addr -- )
        DUP CELL+ TO newObject  \ object address
        !                       \ create the class ptr
        newObject SWAP ERASE ;  \ clear to zero

: (heapObj)  ( #elems class | class -- ^obj )
        >R  ( save class on return stack )
        R@ DFA @  ( ivar data size )
        R@ @width ?DUP
        IF      \ indexed object, add size of indexed area
                \ ( #elems size width -- )
                2 PICK * + ( indexed data )  CELL+ ( idxHdr )
                R@ allocObj
                newObject R@ DFA @ + !  ( store #elems in idxHdr )
        ELSE
                R@ allocObj     \ non-indexed object
        THEN
        R> IFA @ 0 ITRAV        \ initialize instance variables
        ClassInit               \ send ClassInit: message to new object
        newObject  ;            \ return object address

also forth definitions
: HEAP>   ( -- addr )
        ' >BODY  DUP ?isClass 0= ABORT" Not a class"
        STATE @
        IF      POSTPONE LITERAL  POSTPONE (heapObj)
        ELSE    (heapObj)
        THEN ; IMMEDIATE

: RELEASE  ( ^obj -- )          \ free heap object
        Release: [ DUP ]        \ send Release: message to object
        CELL - FREE ?MEMERR ;   \ deallocate it
previous definitions
[THEN]


\ =====================================================================
\ Support for indexed instance variables. When object of these classes
\ are defined, the number of elements should be on the stack.

also forth definitions

\ Set a class and its subclasses to indexed

: <Indexed  ( width -- )  ?Class  ^Class XFA !  ;

\ For some classes, we always want the class pointer to be stored whenever
\ the object is an instance variable, so that the object can receive late-
\ bound messages. This is already the case for indexed classes. Here we
\ fake it out by storing -1 into the indexed field. This is the reason
\ for the "0 MAX" in the definition of @width.

: <General  ( -- )  -1 <Indexed ;

\ self returns the same address as ^base, but is used for late bound calls
\ to the object, i.e.   Selector: [ self ]

: self  ( -- )  ?Class
        ^Class XFA @ 0= ABORT" Class must be <General or <Indexed"
        POSTPONE ^base ; IMMEDIATE

\ =====================================================================
\ Indexed primitives. These should be in code for best performance.

: idxBase  ( -- addr )          \ get base of idx data area
        ^base  DUP >class DFA @ +  CELL+ ;

: limit    ( -- n )             \ get idx limit (#elems)
        ^base  DUP >class DFA @ +  @ ;

: idx-width    ( -- n )         \ width of an idx element                       \ SJC000
        ^base >class XFA @ ;

: ^elem   ( index -- addr )     \ get addr of idx element
        idx-width *  idxBase + ;                                                \ SJC000

\ Fast access to byte and cell arrays.
: At1  ( index -- char )   idxBase + C@ ;
: At4  ( index -- cell )   CELLS idxBase + @ ;

: To1  ( char index -- )   idxBase + C! ;
: To4  ( cell index -- )   CELLS idxBase + ! ;

: ++1  ( char index -- )   idxBase + C+! ;
: ++4  ( cell index -- )   CELLS idxBase + +! ;

\ Compute total length of object.
\ The length does not include class pointer.
: objlen  ( -- objlen )
        ^base >class DUP DFA @  ( non-indexed data )
        SWAP @width ?DUP
        IF  idxBase CELL - @ ( #elems ) * +  CELL+  THEN ;
previous definitions

\ =====================================================================
\ Runtime indexed range checking. Use +range and -range to turn range
\ checking on and off.

: ?range  ( index -- index )  \ range check
        DUP idxBase CELL - @ ( #elems )  U< IF EXIT THEN
        TRUE ABORT" Index out of range" ;

0 VALUE (?idx)          \ execution vector for range checking

also forth definitions

: ?idx   (?idx) EXECUTE ;

: +range  ['] ?range TO (?idx) ;  +range
: -range  ['] NOOP   TO (?idx) ;

\ =====================================================================
\ primitives for cell-sized objects.

: M@  ( -- n )  POSTPONE ^base  POSTPONE @ ; IMMEDIATE
: M!  ( n -- )  POSTPONE ^base  POSTPONE ! ; IMMEDIATE

\ =====================================================================
\ Define base class "object" from which all other classes inherit
\ Some of the common indexed methods are defined here.

:Class Object  <Super Meta

        :M Class:   ^base >class  ;M    \ non-IX - leave class ptr
        :M Addr:    ^base         ;M    \ get object address

        \ ( -- len )  Return total length of object
        :M  Length:    objlen      ;M

HAVE DUMP [IF]
        :M  Dump:     ^base objlen DUMP  ;M
[THEN]
        :M  Print:     ." Object@" ^base U.  ;M

        :M ClassInit:   ;M      \ null method for object init
        :M Release:     ;M      \ null method for object release
\        :m wm_char      ;m      \ Test that Windows constant is accepted
\        :m 55           ;m      \ Test that constant is accepted
;Class

\ Bytes is used as the allocation primitive for basic classes
\ It creates an object of class Object that is n bytes long.
\ You can get the address by sending it an addr: message.

: bytes  ( n -- )
        ?Class  ['] Object >BODY <Var  classAllot ;

[forth]

\ =====================================================================
\ Load primitive classes and test routines.

cr .(         Variable class )  include %lib%/oop/var.fth
cr .(         Array class    )  include %lib%/oop/array.fth

