\ OVLVFX.FTH - VFX Forth Overlay Manager

((
Copyright (c) 1998-2003
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


Requires: VFX Forth 4.20 or later

Think of an overlay as a self-relocating Forth based DLL.

Format of an OverLay
====================

An overlay consists of:
      +----------------------------------------+
      |                 HEADER                 |
      +----------------------------------------+
      |          n RELOCATION records          |
      +----------------------------------------+
      |       n RELOCATION owner records       |
      +----------------------------------------+
      |               CODE BLOCK               |
      +----------------------------------------+
      |             WIDLINK FIXUPS             |
      +----------------------------------------+
      |             VOCLINK FIXUPS             |
      +----------------------------------------+
      |            WORDLIST FIXUPS             |
      +----------------------------------------+

The sub sections are defined by later structure definitions.

To do
=====
Clean HASEXT for '\' or use kernel version.

RES-LINK chain

Add dependency chain in overlay of later overlays dependent on this one.
Could be nasty?

Change history
==============
20120715 SFP025 Removed WID>LIMITS for WID-THREADS.
20030127 SFP024 Added unique build number
20020314 SFP023 UNMARKED - Major overhaul to cope with relocation errors
		caused by code sequences matching overlay addresses.
20020307 SFP022 Provided facility to load overlays in the dictionary.
20020305 SFP021 Cleaned up error handling.
		Added more error checking to BUILDOVERLAY.
		File names are taken from the source using GETPATHSPEC.
		Added relocation counting.
20020304 SFP020 Added CHECKDICT to MAKEOVERLAY.
20011210 SFP019 Changed relocation memory strategy on reading to reduce
                number of ALLOCATES - was in excess of 100,000!
		Cleaned THROW handlers.
20011122 SFP018 Permitted relocation of end address of overlay.
20011119 SFP017 Changed ALLOCATE and FREE to CODE-ALLOCATE and CODE-FREE.
                DEFERred CODE-ALLOCATE and CODE-FREE to permit use of
		other memory managers.
		DEFERred OVL-INIT-COMPILE which sets the initial state
		of the compiler for each overlay build.
20011107 SFP016 Preserved binary to avoid clean-up problems.
20011015 SFP015 Moved relocation value to data area and removed it from the relocation section.
20010802 SFP014 Reduced size of relocation to 1 byte.
20010730 WJB013 (WriteRelocDiff) change error condition to throw,
		catch frame in MakeOverlay.
		Delete SysErrDef on 256 modulo, not required anymore.
		Remove check for error in CheckRuleInfo.
20010730 WJB012 Include execution of CleanAfterMake in (WriteRelocDiff) before execution of throw
                in overlay reloction errors. MARKER xt now saved in VALUE MARKERXT.
20010702 SFP011 Removed ERRDEF code now in the kernel and replaced
		use of ERRDEF with SYSERRDEF.
20010629 SFP010 Prevented modulo 256 problem.
20010616 SFP009 Fixed absolute reference to other overlays.
20010615 SFP008 Fixed CODE-ALLOCATE for W2000.
20010614 SFP007 Added RELEASEALLOVERLAYS.
20010418 SFP006 Expanded macros in the output file name for file creation.
		Overlay header contains the OUTPUT name, not the SOURCE name.
20000928 SJC005 Source Inline macros are removed after each compilation
20000926 SJC004 (Change not marked - code moved)
                I've yanked the overlay information structure from this
                file and place in the KERNEL.FTH file - sorry but I need
                it to complete INFORTH? and IP>NFA from the Kernel.
20000830 SFP003 Forced compilation state when building the overlay
20000826 SFP002 Added more documentation about gotchas. Added more
		error checks.
20000812 SFP001 Removed OVERLAYFILES vocabulary and replaced it by
                chaining the overlays anchored at OVL-LINK. This
	        change permits tools to be loaded without affecting
                HERE, so enabling the assembler to be an overlay.
))

\ ==========
\ *> library
\ *S Binary Overlays
\ ==========

( *N Introduction )
( *P Binary overlays are pieces of the dictionary that have been compiled )
( ** and saved with relocation information. They can be reloaded )
( ** as needed and released on demand. )
( ** Binary overlays are useful when you want to ship tools that are only )
( ** needed during development, or if you have a large application whose )
( ** memory footprint you want to reduce by only loading parts of the )
( ** application when needed. )

\ *P The binary overlay utility is not part of the kernel, but can be
\ ** compiled from LIB\OVLVFX.FTH. As of build 3.40.0808, there has been
\ ** major change in the way overlays are constructed. This change
\ ** removes many restrictions that were present in earlier builds.
\ ** To use the new overlay handler, all overlays must be rebuilt.

( *N Using overlays )
( *P An overlay is generated by MAKEOVERLAY )
( *E   MAKEOVERLAY <sourcename> <overlayname>  )
( *P the file <sourcename> is compiled twice. Relocation information is )
( ** extracted and saved to the overlay along with the raw binary )
( ** information. If any previously loaded overlays are needed by this )
( ** overlay, their names are saved in the overlay and they will be )
( ** automatically reloaded if necessary. After the overlay has been )
( ** generated, the overlay code is removed. Overlays can be tested by )
( ** compiling <sourcename> conventionally, and then finally generating )
( ** the overlay when you are satisfied with it. MAKEOVERLAY preserves )
( ** and links all vocabularies including SOURCEFILES. )
( ** Overlay files are saved by MAKEOVERLAY in the current directory. )
( ** The compiler imposes the following initial condition before the )
( ** overlay file is compiled: )
( *E   DECIMAL -SHORT-BRANCHES +SIN +SINDOES )

\ *P MAKEOVERLAY releases all previously loaded overlays. As a consequence,
\ ** if the overlay to be compiled requires other overlays, you must load
\ ** them explicitly by specifying them as dependencies before using
\ ** MAKEOVERLAY. A dependency list is defined by the word [DEPENDENCIES
\ ** followed by a list of overlay file names as required by LOADOVERLAY below.
\ ** The list is termininated by DEPENDENCIES]. Use in the form:
\ *E [dependencies
\ **   primovl secovl ...
\ ** dependencies]
\ ** makeoverlay MyOvL
\ *P This will cause MAKEOVERLAY to load the dependent overlays PRIMOVL.OVX
\ ** and SECOVL.OVX and so on.

( *P When an overlay is reloaded by LOADOVERLAY )
( *E   LOADOVERLAY <overlayname>   )
( *P the binary code and relocation information are loaded. If the )
( ** overlay file references other overlays, these are loaded before )
( ** the relocated binary code is installed. Overlay code is loaded into )
( ** memory allocated from the Windows heap, and are linked in reverse
( ** load order, so that the last loaded is found first. )
( ** The result of this is that the overlays are always loaded in )
( ** dependency order, and releasing a "leaf" overlay will not affect )
( ** the dependencies of other previously loaded overlays. )
( *P Although overlay files are saved by MAKEOVERLAY in the current )
( ** directory, LOADOVERLAY will look first in the current directory )
( ** and then in the directory from which the application was loaded. )
( ** This allows all overlays and the main executable to reside in the )
( ** same directory regardless of the current directory, but maintains )
( ** convenience during development. )

( *P An overlay can be released by the use of RELEASEOVERLAY.
( *E   RELEASEOVERLAY <overlayname>   )
\ *P All loaded overlays can be released by RELEASEALLOVERLAYS
\ *E   ReleaseAllOverlays

( *N Load and Release actions )
( *P A word can be set to excute whenever the overlay is loaded from file )
( ** or released. These words permit the overlay to allocate and free resources )
( ** such as memory buffers. )
( *E   ' <load-action> SetOvlLoadHook   )
( **   ' <release-action> SetOvlReleaseHook   )
( *P Note that these settings should be in the overlay load file. The )
( ** stack effect of <load-action> and <release-action> must be neutral, )
( ** i.e. take nothing and return nothing [ -- ]. )

( *N File name conventions )
( *P From VFX Forth v3.4 onwards, the naming conventions have been changed. )
( *B The internal overlay name is always the output file name )
( ** after any default extension name has been added by MAKEOVERLAY. )
( *B LOADOVERLAY checks the internal overlay name after adding the )
( ** default file extension. )

( *P The binary overlay files have a ".OVX" extension. The word MAKEOVERLAY )
( ** creates the overlay for you as follows:
( *E   MAKEOVERLAY <sourcename> <overlayname>  )
( *P If the source file name )
( ** does not have an extension, the rules of INCLUDED will be followed, )
( ** checking for files with extensions ".BLD" ".FTH" ".F" ".CTL" ".SEQ" )
( ** in that order. If the destination file name does not have an extension )
( ** ".OVX" will be used. If the destination file name is not provided, )
( ** the source file name is used with a ".OVX" extension. Thus, just )
( ** typing MAKEOVERLAY FOO will compile FOO.FTH to create FOO.OVX. )
( ** The overlay name held by the system is the output file specification )
( ** as given or created by MAKEOVERLAY, converted to upper case. This is )
( ** important when reloading the overlay. )

( *P If no extension is provided for LOADOVERLAY, a ".OVX" extension )
( ** will be added to the file name. Thus LOADOVERLAY FOO will check )
( ** if an overlay called FOO.OVX has been loaded, and will load from file )
( ** FOO.OVX. Similarly, LOADOVERLAY FOO.OVX will check )
( ** if an overlay called FOO.OVX has been loaded, and will load from file )
( ** FOO.OVX. )

( *N Version control )
( *P Each overlay contains VFX Forth information, and overlays cannot )
( ** be loaded by a version of VFX Forth other than the one that built it. )
( ** A user defined version string can be added to the version control )
( ** information using SETOVLVER, which takes the address of a counted )
( ** string. The format of the string is entirely user defined, the overlay )
( ** handler simply checks the strings for identity. )

\ *P Note that this version of LIB\OVLVFX.FTH requires VFX Forth build
\ ** 3.40.0808 of 15 March 2002 or later.

( *N Restrictions )
( *P The following system state is preserved and restored by the overlay )
( ** handler. )
( *E Overlays needed by the current overlay )
( ** Vocabularies and vocabulary link )
( ** Wordlists and wordlist link )
( ** Libraries )
( ** Imported functions )
( *P If you generate other system-wide chains, these will NOT be preserved. )
( ** To preserve them, modify the code in LIB\OVLVFX.FTH using the )
( ** xxxIMPORTLINK words as a model. Future versions of this code may )
( ** support a chain of chains model, but this will require that ALL such )
( ** chains are anchored in the VFX Forth kernel/application before any )
( ** overlays are either generated or reloaded. )
( *P N.B. If you modify this code, please pass it back to MPE so that it can )
( ** be incorporated in later builds. This will reduce your maintenance )
( ** work, our technical support load, and you will benefit from the work )
( ** of others. )

( *N Gotchas )
( *H Bad or random data )
( *P The overlay is produced by comparing two versions of the binary at )
( ** different addresses, and generating relocation information from )
( ** any differences. If a relocation value does not correspond to )
( ** another overlay or the VFX Forth kernel, the build of the overlay )
( ** will cause an error. Such errors can be caused by anything that )
( ** inadvertently changes the data or code generation of the two )
( ** versions being compared. )

( *H Uninitialised buffers )
( *P If data space in the dictionary is not initialised at compile time, )
( ** it may contain random data. Compare:
( *E   <size> BUFFER: <name>               \ safe    )
( **   CREATE <name>   <size> ALLOT        \ unsafe  )
( **   CREATE <name>   <size> ALLOT&ERASE  \ safe    )

( *H Different initial conditions )
( *P The initial conditions of directives that affect code generation )
( ** must be the same for each build. At least the following directives )
( ** should be considered: )
( *E   +SHORT-BRANCHES  -SHORT-BRANCHES   branch code size )
( **   +SIN             -SIN              source inlining )
( **   +SIN-DOES        -SIN-DOES         DOES> clause inlining )
( *P Similarly the starting codition of BASE should also be considered. )
( ** The compiler imposes the following initial condition before the )
( ** overlay file is compiled: )
( *E   DECIMAL -SHORT-BRANCHES +SIN +SINDOES )

\ *H Search order issues
\ *P When compiling an overlay strict control of the initial search order
\ ** is often necessary, especially because of redefinitions. We recommend
\ ** that overlays are constructed from a build file which ensures that
\ ** other required overlays are installed.

\ *P A sign of bad search order control is that the overlay can be correctly
\ ** built with the source inliner turned off, but will not build with it on.

\ *H Long file names
\ *P You cannot use file names with spaces, even though GETPATHSPEC is used to
\ ** input the file names, because the file names are internally used as
\ ** Forth word names.

\ *H Code conflicts with address
\ *P There are occasions when a four-byte code sequence matches an address
\ ** in another overlay, causing false relocation data to be generated.
\ ** The result will be code that is corrupt after loading.

\ *P This situation has been drastically improved by the overhaul of
\ ** 14 March 2002, but the warning has been left in until we are confident
\ ** that all situations have been covered.

-sin
+short-branches


\ *****************
\ Debugging control
\ *****************

only forth definitions

[undefined] [D [if]
variable debugging	\ -- addr ; set true to enable debug code
  debugging off

: D]		\ -- ; end of [D ... [D] ... D] clause
  S" key? drop" evaluate		\ force console to flush
; immediate

: [D]		\ -- ; middle of [D ... [D] ... D] clause
  1 PassOver
; immediate

: [D		\ -- ; code between [D and D] is compiled if debugging is enabled
  debugging @ 0=
  if  postpone [D]  then
; immediate

also Compilation? definitions
: [D		1+ ;

: D]		1- ;

: [D]		1- dup if 1+ then ;
previous definitions
[then]

[D
: ?throw	\ flag n --
  swap if  .rs throw else drop  endif
;
D]


\ ****************************************
\ Allocating memory for overlay dictionary
\ ****************************************

[defined] Target_386_Windows [if]
Extern: void * PASCAL VirtualAlloc( void *, dword, dword, dword );
Extern: BOOL PASCAL VirtualFree( void *, dword, dword );

defer code-allocate	\ size -- ptr ior
:noname		\ size -- ptr ior
  NULL
  swap
  MEM_COMMIT ( MEM_RESERVE or )		\ SFP008
  PAGE_EXECUTE_READWRITE
  VirtualAlloc dup NULL =
; to-do code-allocate

defer code-free	\ ptr -- ior
:noname		\ ptr -- ior
  0 MEM_RELEASE VirtualFree 0=
; to-do code-free
[else]
: code-allocate		allocate  ;	\ size -- ptr ior
: code-free		free  ;		\ ptr -- ior
[then]


\ **********************
\ Visibility control 1/2
\ **********************

((
defer show-reloc-addr		\ for debugging only
  assign drop to-do show-reloc-addr
))
vocabulary OverlayVoc
also OverLayVoc definitions


\ ************************
\ Structures and Variables
\ ************************

$5555AAAB constant ovl_ver#	\ -- n ; increment for each version change of this file
0 value pass#			\ -- n ; pass counter where n=0, 1, ...

(( SFP014 SFP015
In order to reduce the size on disc of the relocation data:
a) the relocation type is written as a byte
b) the relocation offset and owner are written as is
c) the target address is converted to an offset and written
   back into the code image.
))
struct ORELOC	\ -- len ; OVL file relocation record
        cell    field   ORELOC.next		\ internal use only
        cell    field   ORELOC.type		\ RELOCATION from here
        cell    field   ORELOC.offset		\ relative to start of CODE section
        cell    field   ORELOC.owner		\ owner ID, here so we can use 2@ and 2!
        cell    field   ORELOC.TargetAddress	\ relative to start of OWNER, put at OFFSET
end-struct

struct  RESOV	\ -- len ; internal overlay reference information
		\ used both for read and write
        cell    field   RESOV.next
        cell    field   RESOV.flag	\ set if referenced
        cell    field   RESOV.end-dp	\ end first so we can use 2@ and 2!
        cell    field   RESOV.start-dp	\ start
\ from here on is the OVL file owner (needed overlay) record
	cell	field   RESOV.owner	\ owner id for file
        1 chars field   RESOV.namelen	\ count byte n
        0       field   RESOV.name	\ n bytes of text
end-struct

$80000000 constant RFLAG_USED	\ flag set when used in MAKEOVERLAY

2 constant /VDrows	\ -- n ; number of rows in /VocData structure below
			\ N.B. only the first two rows are saved to the OVL file

struct /VocData	\ -- len ; vocabulary data in OVL file
  int VD.link		\ link to previous structure
  int VD.wid		\ WID at compile time
  int VD.owner		\ owner ID of overlay, unique within overlay file
  int VD.offset		\ offset into owner overlay
  int VD.#threads	\ n - number of threads in voc
  0 field VD.data	\ followed by 2 rows of n cells
			\   row 0  top nfa	  restored to voc when loaded
			\   row 1  bottom lfa	  patched to 0 before write
\ below removed from build 3.40.0808 onwards
			\   row 2  (blfa) copy 2  restored after use
			\   row 3  (blfa) copy 1  restored after use
			\   row 4  original llfa  row 1 is modified at save
			\   row 5  original nfa   row 0 is modified at save
end-struct

Struct /VL	\ -- len ; WID-LINK and VOC-LINK data in OVL file
  int VL.next		\ link, not written
  int VL.HighVocLink	\ NFA of last word defined
  int VL.LowVocLink	\ LFA of first word defined
end-struct

Struct /LIB	\ -- len ; LIB-LINK data in OVL file
  int LIB.next		\ link, not written
  int LIB.High		\ top in overlay
  int LIB.Low		\ bottom link in overlay
\  int LIB.Copy2		\ contents of copy2
\  int LIB.Copy1		\ contents of copy1
end-struct

Struct /IMPORT	\ -- len ; LIB-LINK data in OVL file
  int IMPORT.next	\ link, not written
  int IMPORT.High	\ top in overlay
  int IMPORT.Low	\ bottom link in overlay
\  int IMPORT.Copy2	\ contents of copy2
\  int IMPORT.Copy1	\ contents of copy1
end-struct

max_path buffer: SourceName	\ -- addr ; overlay build file buffer
max_path buffer: OutputName	\ -- addr ; overlay output file name
  S" default.ovx" OutputName place

variable LoadedOverlayChain	\ -- addr ; anchors list of loaded overlays
  LoadedOverlayChain off
variable WidDataChain	\ -- addr ; anchors WidData chain
  WidDataChain off
variable WidLinkChain	\ -- addr ; anchors WidLink chain
  WidLinkChain off
variable VocLinkChain	\ -- addr ; anchors VocLink chain
  VocLinkChain off
variable oreloc-chain	\ -- addr ; anchors relocation chain
  oreloc-chain off
variable LibChain	\ -- addr ; anchors library chain
  LibChain off
variable ImportChain	\ -- addr ; anchors imported function chain
  ImportChain off

variable #bytes		\ -- addr ; hold number of bytes read/written

0 value fp		\ -- handle ; overlay file handle

0 value curr-dp		\ -- addr ; base of current overlay
0 value initial-dp	\ -- addr ; start of first copy
0 value ovlen		\ -- len ; length of overlay
0 value img1		\ -- addr ; start of store of first copy
0 value second-dp	\ -- addr ; start of second copy
0 value len2		\ -- len ; length of overlay
0 value img2		\ -- addr ; start of store of second copy
0 value ovCompiled	\ -- addr ; address at which overlay was compiled
0 value ovReloc		\ -- offset ; relocation offset to apply to -1
0 value OvBuffer	\ -- addr ; address of loaded code
0 value KernelBase	\ -- addr ; base address of Forth kernel
0 value User_Ver$	\ -- addr ; address of user version string

\ variable LoadHook1	\ -- addr ; holds user's load time action
\ variable LoadHook2	\ -- addr ; holds user's load time action
\ variable UnloadHook1	\ -- addr ; holds user's unload time action
\ variable UnloadHook2	\ -- addr ; holds user's unload time action

defer CleanAfterMake	\ -- ; clean up and restore ; SFP014
defer SaveBeforeMake	\ -- ; save data to be restored ; SFP016


\ ***********
\ Error codes
\ ***********

\ These replace the ABORT" strings
#-600 NextSysError !
\ SysErrDef ovl_OutOfMem	"OVL: can't allocate enough memory"
\ SysErrDef ovl_BadFree	"OVL: can't free memory"
SysErrDef ovl_OutOfVMem	"OVL: can't allocate enough virtual memory"
SysErrDef ovl_BadVFree	"OVL: can't free virtual memory"
SysErrDef ovl_BadWid	"OVL: wid has no VD structure"
SysErrDef ovl_BadOpen	"OVL: Failed to open overlay"
SysErrDef ovl_BadHeader	"OVL: Invalid overlay header"
SysErrDef ovl_BadRead	"OVL: File read error"
SysErrDef ovl_BadWrite	"OVL: file write error"
SysErrDef ovl_NotLoaded	"OVL: overlay not loaded"
SysErrDef ovl_BadRelocType "OVL: Invalid relocation type"
SysErrDef ovl_BadRelocOwner "OVL: Invalid relocation owner"
SysErrDef ovl_BadOpName	"OVL: missing or invalid output file name"
SysErrDef ovl_BadRelocVal "OVL: Invalid relocation value, probably caused by ALLOT without ERASE"
SysErrDef ovl_NoPreload	"OVL: All needed overlays must be loaded before MAKEOVERLAY is used "
\ SysErrDef ovl_NotMod256 "OVL: Overlay length must NOT be modulo 256, insert n ALLOT and remake "	\ del WJB013
SysErrDef ovl_Unstable  "OVL: System *may* not be stable"
SysErrDef ovl_BadLengths "OVL: Overlay lengths differ, check initial code generation conditions"
SysErrDef ovl_ChainActive "OVL: ISE: Chain still active"		\ SFP021
SysErrDef ovl_BadBuild	"OVL: Overlay build causes stack depth change"	\ SFP021
SysErrDef ovl_ExcessRelocs "OVL: Too many relocations, increase MAX#RELOCS in LIB\OVLVFX.FTH"	\ SFP021
SysErrDef ovl_BadVer	"OVL: Overlay built on a different version - cannot be loaded"


\ ***************
\ Debugging tools
\ ***************

[D
: rel-dump      \ addr len --
                { | ss -- }
  over -> ss
  ?dup if
    bounds                              \ get limits
    do
      cr  i ss - .lword  space  i 8 + i	\ print address
      do  space i c@ .byte  loop        \ first eight bytes
      space  i $10 + i 8 +              \ gap
      do  space i c@ .byte  loop        \ second eight bytes
      2 spaces  i $10 + i               \ gap
      do  i c@ .ascii  loop             \ ascii o/p
      halt? ?leave                      \ quit or halt
    $10 +loop                           \ until done
    cr                                  \ restore base
  else
    drop
  then
;
D]


\ *************
\ Misc. support
\ *************

: test-mask	\ mask addr -- flag
  @ and
;

: set-mask	\ mask addr --
  tuck @  or  swap !
;

: clear-mask	\ mask addr --
  tuck @  swap invert and  swap !
;

: ovl_throw	\ n -- ; special case of throw
  ?dup
  if  .throw  -1 throw  endif
;

: ovl_?throw	\ flag n -- ; special case of ?THROW ; SFP019
  swap
  if  ovl_throw  else  drop  endif
;

: ovl_alloc	\ n -- addr ; allocate overlay memory
  code-allocate ovl_OutofVMem ovl_?throw	\ SFP019
;

: ovl_free	\ addr -- ; free overlay memory
  ?dup if
    code-free ovl_BadVFree ovl_?throw	\ SFP019
  endif
;

: +code		\ offset -- code ; convert load offset to address
  OvBuffer +
;

: CopyImg	\ addr len -- addr' ; get memory and copy it
  dup >r  ovl_alloc tuck  r> move
;

: in-copy2?	\ addr -- flag ; check if addr is in the second copy
  second-dp dup len2 + 1+ within		\ SFP018
;

: in-copy1?	\ addr -- flag ; check if addr is in the first copy
  initial-dp dup ovlen + 1+ within
;

: inImg1	\ addr -- addr' ; convert memory image address to image address
  initial-dp - img1 +
;

: inImg2	\ addr -- addr' ; convert memory image address to image address
  second-dp - img2 +
;


: FreeChain	\ var -- ; free all chain memory
  dup @
  begin
    ?dup
   while
    dup @  swap ovl_free
  repeat
  off
;

: ?sayFreeChain	\ var --
  dup @ if
\    cr  dup body> >name .name ."  still active: "
    FreeChain
\    ." cleared"
    ovl_chainActive throw		\ SFP021
  else
    drop
  endif
;

: ?Chains	\ --
  LoadedOverlayChain ?sayFreeChain
  WidDataChain ?sayFreeChain
  WidLinkChain ?sayFreeChain
  VocLinkChain ?sayFreeChain
  LibChain ?sayFreeChain
  ImportChain ?sayFreeChain
  oreloc-chain ?sayFreeChain
;

: ResetChains	\ --
  LoadedOverlayChain off
  WidDataChain off
  WidLinkChain off
  VocLinkChain off
  LibChain off
  ImportChain off
  oreloc-chain off
;


\ *************
\ File handling
\ *************

: >ovl		\ addr len -- ; write buffer to output
  dup #bytes +!
  fp write-file ovl_BadWrite ovl_?throw	\ SFP019
;

: b>ovl		{ b | ch -- ; write 8 bit byte to the output ; SFP014 }
  b addr ch c!				\ portable
  addr ch 1 >ovl
;

: n>ovl		{ n -- ; write a 32 bit number to the output }
  addr n cell >ovl
;

: $>ovl		\ $src -- ; write counted string to file
  dup 1 chars >ovl			\ write count
  count >ovl				\ write text body
;

: ovl>		\ addr len -- ; read data from overlay
  dup #bytes +!
  fp read-file ovl_BadRead ovl_?throw	\ SFP019
  drop
;

: ovl>$		\ $dest -- ; read counted string from overlay
  dup 1 chars ovl>			\ read count byte
  count ovl>				\ read text body
;

: padding	\ -- n ; number of padding bytes to a 16 byte boundary
  #bytes @ #15 + #-16 and  #bytes @ -
;

: CreateOverlay	\ c-addr u -- ; create overlay, and write header
  { | buff[ #16 ] name[ MAX_PATH CELL+ ] -- }

\ SFP006..
  name[ max_path cell+ erase		\ expand macros in the output file name
  name[ ExpandMacro
\ ..SFP006
  r/w bin Create-File drop -> fp	\ make it

  #bytes off				\ zero #bytes processed
  version# $>ovl			\ markers at start of binary file
  cpu_target $>ovl
  host_target $>ovl
  User_ver$ if
    User_ver$ $>ovl
  else
    0 buff[ c!
    buff[ 1 >ovl
  endif

  buff[ #16 erase		        \ write 0 padding bytes
  buff[ padding >ovl
;

: Check$	\ $fixed $buff -- ; check string in file
  dup ovl>$				\ read string from file
  >r count r> count compare
  ovl_Badheader ovl_?throw
;

: OpenOverlay	\ c-addr u -- ; open requested overlay for read only
  { c-addr u | name[ MAX_PATH CELL+ ] -- }
  name[ MAX_PATH CELL+ erase

  c-addr u r/o open-file if		\ if can't open file in current dir
    drop
    Forth-Buff name[ $move		\ concatenate EXE dir and file
    c-addr u name[ append
    name[ count r/o open-file		\ open the file
    ovl_BadOpen ovl_?throw		\ SFP019
  endif
  -> fp

  #bytes off
  version# name[ Check$			\ read and test strings
  cpu_target name[ Check$
  host_target name[ Check$
  User_ver$ if
    User_Ver$ name[ Check$
  else
    name[ 1 ovl>  name[ c@
    ovl_Badheader ovl_?throw
  endif

  name[ padding ovl>		        \ read padding bytes
;

: CloseOverlay	\ -- ; close overlay
[D  cr #bytes @ . ." bytes read/written"  D]
  fp close-file drop
;

also forth definitions
defer ovl-init-compile	\ -- ; set initial state ; SFP017
:noname
  decimal optimised -short-branches +sin +sindoes
; to-do ovl-init-compile
previous definitions

((
Struct /OVI	\ -- len ; overlay information block
		\ MUST match BUILDOVERLAY
  int OVI.link		\ link to previously loaded overlay (initialised to 0)
  int OVI.magic		\ $5555AAxx
  int OVI.len		\ length of overlay
  int OVI.end		\ end of overlay - must precede start
  int OVI.start		\ start of overlay
  int OVI.flags		\ bit 0 = 1 = loaded into dictionary
  int OVI.loadhook	\ xt of load action
  int OVI.unloadhook	\ xt of unload action
  int OVI.ID		\ overlay unique ID
  0 field OVI.name	\ name follows as a counted and zero terminated string
end-struct
))
/OVI #36 <> [if]
  cr ." /OVI structure mismatch with kernel" ABORT
[then]

\ Flag field definitions
$0001 constant dictmask		\ bit 0 = 1 = loaded into dictionary

[-sin
: BuildOverlay	\ -- addr len ; compile the overlay build file, return start and length
		\ MUST match the /OVI structure
  align here				\ align and return start of overlay
  dup -> curr-dp			\ set as current overlay
  0 ,					\ link
  ovl_ver# ,				\ marker and version control
  0 ,					\ length
  0 ,					\ end
  0 ,					\ start
  0 ,					\ flags
  ['] noop ,				\ load action
  ['] noop ,				\ unload action
  Ovl-ID @ ,				\ overlay ID number ; SFP024
  OutputName dup c@ 1+ >r		\ add output file name ; SFP006
  here r@ move
  here count upper
  r> allot  align

\ decimal -short-branches +sin +sindoes	\ force compilation state ; SFP003
  ovl-init-compile			\ SFP017
  depth >r				\ SFP021
  SourceName count included		\ do build
  depth r> <> ovl_BadBuild ?throw	\ check for balanced stack ; SFP021
  0 ,					\ unrelocatable item for safety ; SFP018
  align					\ align dictionary after build
  here over - $0FF and 0=		\ avoid modulo 256 problem ; SFP010
  if  0 ,  endif
  here over -				\ -- addr len
  2dup swap ovi.len !			\ set length in header

  inc pass#
;
sin]

: ForceExt	\ extaddr extlen $addr1 u1 -- $addr1 u2 ; add extension
  2 pick over + >r  over >r		\ stash final addr/len
  + swap move				\ copy extension to end
  r> r>
;

: AddExt	\ extaddr extlen $addr1 u1 -- $addr1 u2 ; add extension if not present
  2dup Extension? if			\ if extension present
    drop 2swap 2drop			\ return original string
  else
    ForceExt				\ add extension
  endif
;


\ ****************
\ Overlay handling
\ ****************

: OvlLoaded?	\ c-addr u -- *ovi true | 0 0 ; true if loaded
  ovl-link @
  begin
    dup
   while				\ -- caddr u *ovi/0
    >r
    2dup r@ OVI.name count str= if
      2drop  r> true  exit		\ -- *ovi true
    endif
    r> OVI.link @
  repeat
  nip nip dup				\ -- 0 0
;

\ Scan RESOV chain to check for reference inside another overlay.
\ If found, mark that overlay with the RFLAG_USED bit and return pointer to
\ RESOV structure. Otherwise return 0 for kernel or -1 for self
: FindOwner	\ addr -- offset *RESOV | offset 0 | offset -1
  dup in-Copy2?				\ in this overlay
  if  second-dp -  -1  exit  then
[D cr ." FindOwner in overlay" D]
  LoadedOverLayChain @
  begin
    ?dup
  while					\ -- addr *RESOV
    2dup RESOV.end-dp 2@ within if	\ if addr is the range of this overlay
      RFLAG_USED over RESOV.flag set-mask	\ set mask in flags
      tuck RESOV.start-dp @ - swap	\ convert addr to offset
[D ."  found " D]
      exit				\ and get out
    then
    @
  repeat
[D ."  defaulted " D]
  KernelBase -  0			\ assume in kernel if not in overlay
[D ." done" D]
;

: addr>off	\ addr -- offset | -1 ; leave 0 alone
  dup if
    dup in-copy2? if
      second-dp -
    else
      drop -1
    then
  endif
;

: WriteNeededOverlays	\ --
[D cr ." WriteNeededOverlays ... " D]
  LoadedOverlayChain @
  begin
    ?dup
  while
    dup RESOV.flag @ RFLAG_USED and if	\ used?
      RFLAG_USED over RESOV.flag clear-mask
      dup n>ovl				\ write *RESOV as index
      dup RESOV.namelen dup c@ 1 chars + >ovl	\ write name
    then
    dup RESOV.next @  swap ovl_free
  repeat
  LoadedOverlayChain off
  0 n>ovl				\ terminator
[D ." done" D]
;

[D
: .showowner    \ *resov --
  cr ."   Relocation owner "
    dup RESOV.owner @ .dword ."  is "
    dup RESOV.namelen $.
  cr ."   Range: "
    dup RESOV.end-dp 2@
    swap .dword  ."  to " .dword
  drop
  key? drop
;

: .showowners   \ --
cr ." Loaded overlays"
  LoadedOverLayChain
  begin
    @ ?dup
  while
    dup .showowner
  repeat
;
D]

: AddRESOV	\ *resov --
  dup RESOV.namelen c@ 1+  RESOV +	\ length of required structure
  dup ovl_alloc dup >r 			\ -- src len dest ; get space for it
  swap move				\ copy to new buffer
  r> LoadedOverlayChain Addlink		\ add to chain
;

: (InitLoadedOverlay)	{ *ovi | t[ RESOV MAX_PATH + ] -- }	\ SFP001
  t[ RESOV MAX_PATH + erase
  *ovi t[ RESOV.owner !			\ set ID for this overlay, unique to build
  *ovi ovi.end 2@ t[ RESOV.end-dp 2!	\ start & end
  *ovi ovi.name t[ RESOV.namelen over c@ 1 chars + move	\ copy name
  t[ AddRESOV				\ make new RESOV entry
[D cr ." Overlay " t[ RESOV.namelen count type ."  detected" D]
;

: InitLoadedOverlays	\ -- ; SFP001
  ovl-link @
  begin
    dup
   while
    dup (InitLoadedOverlay)
    OVI.link @
  repeat
  drop
[D .showowners D]
;

defer ReloadOverlay	\ c-addr u -- start end

-short-branches
: ReadNeededOverlays	{ | t[ RESOV MAX_PATH + ] -- }
[D  cr ." READ: Needed Overlays"  D]
  begin
    t[ RESOV MAX_PATH + erase
    t[ RESOV.owner cell ovl>		\ overlay owner ID
    t[ RESOV.owner @
   while
    t[ RESOV.namelen 1 chars ovl>	\ read count byte
    t[ RESOV.namelen count ovl>        	\ read text
    t[ RESOV.namelen count OvlLoaded? if
      OVI.end 2@			\ get where loaded ; SFP001
    else
      drop
      t[ RESOV.namelen count ReloadOverlay \ bring in dependency
      dup -1 = if  drop throw  endif	\ error if failed
    endif
    t[ RESOV.end-dp 2!
    t[ AddRESOV				\ add to loaded overlay chain
  repeat
;
+short-branches


\ *************
\ Code handling
\ *************

: WriteCode	\ addr len --
[D cr ." WriteCode ... " D]
  over n>ovl				\ compilation address
  dup n>ovl				\ length
  >ovl					\ code
[D ." done" D]
;

: ReadCode	\ -- ; Read CODE Section
[D  cr ." Read Code"  D]
  addr ovcompiled cell ovl>		\ compilation address
  addr ovlen cell ovl>			\ code length
  ovlen ovl_alloc -> OvBuffer		\ buffer for code
  OvBuffer ovlen ovl>			\ read code
  ovbuffer ovi.magic @ ovl_ver# <>	\ check version number
    ovl_BadVer ovl_?throw
;

variable ovl_in_dict	\ -- addr ; true to load overlays in the dictionary ; SFP022

: ApplyCode	\ -- ; place the code where it is needed
[D  cr ." copying code"  D]
\ SFP022...
  ovl_in_dict @ if
    align here  ovlen allot  align	\ -- addr ; allot space in dictionary
  else
    ovlen ovl_alloc			\ -- addr ; get space from heap
  endif
\ ...SFP022
  OvBuffer over ovlen cmove		\ copy code
  OvBuffer ovl_free			\ release memory
  -> OvBuffer				\ reset buffer pointer
  ovCompiled OvBuffer - -> ovReloc	\ calculate relocation
[D
cr ." Code placed at   " OvBuffer .dword   ." , length " ovlen .dword
cr ." Code compiled at " OvCompiled .dword ." , offset " ovReloc .dword
ovBuffer ovlen rel-dump
D]
;


\ *******************
\ Relocation handling
\ *******************

#125000 constant max#relocs	\ -- n ; maximum number of relocations permitted
variable #relocs		\ -- addr ; counts number of relocations

: CountRelocs	\ --
  1 #relocs +!
  #relocs @ max#relocs > ovl_ExcessRelocs ?throw
;


\ -------------------
\ Writing relocations
\ -------------------

\ variable #relocErrs  0 #relocErrs !

: HandleBadReloc	{ diff index -- }
  cr ." Reloc val " diff .dword
  cr ." Ovl len   " ovlen .dword
  cr ." Base 1    " initial-dp .dword
  cr ." Base 2    " second-dp .dword
  cr ." Offset    " index .dword
  cr ." Image 1   " img1 .dword
  cr ." Image 2   " img2 .dword
  cr ." Address 1 " img1 index +  dup .dword  space  @ .dword
  cr ." Address 2 " img2 index +  dup .dword  space  @ .dword
  img1 index + $80 -  $100 dump
  img2 index + $80 -  $100 dump

  CleanAfterMake
  ovl_BadRelocVal ovl_throw		\ comment out if foolish ; WJB013 SFP019
  -1					\ bad reloc type
;

: FindOvlRef	\ target -- flag
  ovl-link @
  begin
    dup
  while					\ -- target *ovi
    2dup ovi.end 2@ within		\ if addr is the range of this overlay
    if  2drop  true exit  endif		\ return true
    ovi.link @
  repeat
  nip
;

: FindAbsOvlRef	\ index -- flag
  img2 + @ FindOvlRef
;

: FindRelOvlRef	\ index -- flag
  dup second-dp +			\ compilation address
  swap img2 + @ +			\ plus offset
  cell+					\ plus cell
  FindOvlRef
;

-short-branches
: (WriteRelocDiff)	{ index | reloc? targ oft owner -- step ; data at index }
  0 -> reloc?
  case  img2 index + @ img1 index + @ -	\ compare cells in two versions

  \ check references within the overlay or to the kernel
    second-dp initial-dp - of		\ Make an absolute relocation for offset I
      img2 index + @ -> targ		\ target address
      index -> oft                	\ offset into initial image
      1 -> reloc?			\ relocation type
    endof
    initial-dp second-dp - of		\ Make a relative relocation for offset I
      second-dp index +			\ compilation address
      img2 index + @ + cell+ -> targ	\ target address = address + data + 4
      index -> oft			\ Remember offset into initial image
      2 -> reloc?			\ relocation type
    endof
\ check references to other overlays
    index FindAbsOvlRef ?of		\ in an overlay
      drop				\ discard diff
      img2 index + @ -> targ		\ target address
      index -> oft                	\ offset into initial image
      1 -> reloc?			\ relocation type
    endof
    index FindRelOvlRef ?of
      drop				\ discard diff
      second-dp index +			\ compilation address
      img2 index + @ + cell+ -> targ	\ target address = address + data + 4
      index -> oft			\ Remember offset into initial image
      2 -> reloc?			\ relocation type
    endof

\ not found, treat as bad
      HandleBadReloc -> reloc?
\      1 #relocErrs +!  #relocerrs @ #40 > -100 ?throw
  endcase

\ Here, if RELOC? is non-zero I had a relocation.
\           RELOC?  is      1) Absolute
\                           2) Relative
\			    3) Unknown
\           TARG    is the absolute target address, or offset when owner found
\           OFT     is offset into initial image for relocation
\           OWNER   is either 0 for KERNEL or pointer to RESOV for overlay
  reloc? if
    targ FindOwner -> owner -> targ	\ find owner and offset

    reloc? -1 <> if
      targ img2 index + !		\ SFP015 ; Put TARG back into the image so that we don't need
					\ to write it into the relocation table.
      reloc? b>ovl			\ relocation type ; SFP014
      oft    n>ovl			\ offset into overlay
      owner  n>ovl			\ Relocate Against
\      CountRelocs			\ count them ; SFP021
    endif
    cell				\ step to next char
  else
    1					\ step to next byte
  then
;
+short-branches

: WriteReloc	\ -- ; write relocation data
[D cr ." WriteReloc ... " D]
  #relocs off				\ SFP021
  ovlen cell 1- -  0 ?do		\ last byte to test is at end-3
    img1 i + c@				\ compare BYTES in two versions
    img2 i + c@ <> if
      i (WriteRelocDiff)		\ bytes are different
    else
      1					\ same, step to next byte
    endif
  +loop					\ step to next cell/byte

  0 b>ovl				\ to terminate list ; SFP014
  2 0 do  0 n>ovl  loop			\ SFP015
\  CountRelocs
[D ." done" D]
;

[D
: .showreloc    \ *ORELOC --
  cr ."   "
  dup
  case  ORELOC.type @
    1 of  ." Abs"  endof
    2 of  ." Rel"  endof
  endcase
  ."  CodeOffset $"
  dup ORELOC.offset @ .dword
  ."  Against Owner Idx:Offset "
  dup ORELOC.owner @ .dword ." :"  dup ORELOC.TargetAddress @ .dword
  drop
;

: .showrelocs   \ --
cr ." Relocations"
  oreloc-chain
  begin
    @ ?dup
  while
    dup .showreloc
    ORELOC.next
  repeat
;
D]


\ -------------------
\ Reading relocations
\ -------------------

\ SFP019...
0 value RelocBase		\ -- addr ; base of relocation memory
0 value RelocOffset		\ -- n ; offset into relocation memory
0 value RelocSize		\ -- n ; size of relocation memory
max#relocs ORELOC *
  constant InitRelocSize	\ -- n ; initial size of relocation memory

: InitRelocMem	\ -- ; initialise relocation memory
  0 -> RelocBase  0 -> RelocOffset  0 -> RelocSize
  InitRelocSize ovl_alloc -> RelocBase
  InitRelocSize -> RelocSize
;

: TermRelocMem	\ -- ; release relocation memory
  RelocBase ovl_free
  0 -> RelocBase  0 -> RelocOffset  0 -> RelocSize
;

: AddRelocMem	\ addr len -- buffer ; add to relocation memory
  dup RelocOffset +  RelocSize u> if	\ if it will overrun
    RelocSize 2* ovl_alloc >r		\ allocate twice as much
    RelocBase r@ RelocOffset cmove	\ copy previous data
    RelocBase ovl_free			\ release previous memory
    r> -> RelocBase			\ set new pointer
    RelocSize 2* -> RelocSize		\ update size
  endif
  tuck					\ -- len addr len
  RelocBase RelocOffset + dup >r	\ -- len addr len dest ; R: -- dest
  swap cmove				\ -- len ; R: -- dest ; copy data
  RelocOffset + -> RelocOffset		\ update offset
  r>					\ return dest
;

: AddReloc      \ *ORELOC -- ; add a relocation entry
  ORELOC AddRelocMem			\ copy to reloc buffer
  oreloc-chain AddLink			\ link to chain
;
\ ...SFP019

: ReadReloc	\ -- ; Read Relocation data
		{ | ch[ cell ] r[ ORELOC ] -- }
[D  cr ." Read Reloc"  D]
  InitRelocMem				\ set up reloc memory ; SFP019
  r[ ORELOC erase
  begin
    ch[ 1 ovl>  ch[ c@ r[ ORELOC.type !	\ portable ; SFP014
    r[ ORELOC.offset  ORELOC 3 cells - ovl>	\ rest of record ; SFP014 SFP015
    r[ ORELOC.type @
   while
    r[ AddReloc
  repeat
;

: FindOvlAddr	\ offset owner -- address
  LoadedOverlayChain
  begin
    @ ?dup
   while	\ -- offset owner *RESOV
    2dup RESOV.owner @ = if
      nip  RESOV.start-dp @ +  exit
    endif
  repeat
  ovl_BadRelocOwner ovl_throw		\ SFP019
;

: FindAbs	\ offset owner -- address
  case
     0 of  KernelBase +  endof
    -1 of  +code  endof
           FindOvlAddr
  end-case
;

-short-branches
: (ApplyReloc)	{ ^rr -- ; apply relocation record }
\ ^rr ORELOC dump
  case  ^rr ORELOC.type @
    0 of	\ None
    endof
    1 of	\ Abs
      ^rr ORELOC.offset @ +code @	\ get relative target address ; SFP015
        ^rr ORELOC.TargetAddress !	\ and set into structure ; SFP015
      ^rr ORELOC.owner 2@ FindAbs
      ^rr ORELOC.offset @ +code  ( dup show-reloc-addr )  !
    endof
    2 of	\ Rel
      ^rr ORELOC.offset @ +code @	\ get relative target address ; SFP015
        ^rr ORELOC.TargetAddress !	\ and set into structure ; SFP015
      ^rr ORELOC.owner 2@ FindAbs	\ abs address
      ^rr ORELOC.offset @ +code cell+ -	\ make relative
      ^rr ORELOC.offset @ +code  ( dup show-reloc-addr )  !
    endof
\          ^rr ORELOC dump
\          cr ." Relocation type " .
          ovl_BadRelocType ovl_throw	\ SFP019
  endcase
;
+short-branches

: ApplyReloc	\ --
[D  cr ." ApplyReloc" .showrelocs  D]
  oreloc-chain @
  begin
    ?dup
  while
    dup (ApplyReloc)
    ORELOC.next @			\ SFP019
  repeat
  oreloc-chain off
[D  ovBuffer ovlen dump  D]
  TermRelocMem				\ release relocation memory
;


\ ******************************
\ Wordlist & Vocabulary handling
\ ******************************

\ Does the wordlist supplied contain definitions beyond DP?
\ Used to see if an overlay has created definitions within WID
: wid-used?	\ wid -- flag
  wid-threads bounds ?do
    i @ in-copy2? if
      true  unloop  exit
    then
  cell +loop
  false
;

: Find.VD	\ wid -- ^vd ; error if not found
\ ***WID*** ???
  WidDataChain
  begin
    @ ?dup				\ exit on 0, not found
   while				\ -- wid ^vd
    2dup VD.wid @ =			\ check
    if  nip  exit  endif
    @					\ link is ALWAYS first
  repeat
  ovl_BadWid ovl_throw			\ SFP019
;

: >lowest       \ thread -- lfa | 0 ; return lowest lfa in overlay
  dup @ in-copy2? if			\ any words in copy2?
    begin
      @ ctrl>link			\ yes, get next one
      dup @ in-copy2? 0=                \      points to copy2?
    until                               \
  else
    drop 0                              \ no, return 0
  endif
;

: SaveWidData	{ wid | ^vd #thr -- ; wid is used by overlay }
[D  cr wid ip>nfa .name ." ... "  D]
  wid >#threads @ /VDrows * cells  /VocData +	\ length of required data
  ovl_alloc -> ^vd			\ -- wid ^vd ; reserve space
\ ."  at address " ^vd .dword
  ^vd WidDataChain AddLink		\ update WidDataChain
  wid ^vd VD.wid !			\ for searching later
  wid >threads  ^vd VD.data		\ -- threads table
  wid >#threads @ dup ^vd Vd.#threads !	\ save number of threads
  cells  dup -> #thr  0 ?do		\ -- threads table
    over i + >lowest ?dup if		\ if thread in this overlay, find first
      swap				\ -- threads llfa table
      over addr>off  over #thr + !	\ put llfa2 in row 1 - offset
      swap second-dp -			\ turn into offset
      dup img1 + off  img2 + off	\ zero (llfa2) and (llfa1) ; -- threads table
      over i + @ addr>off  over !	\ put top nfa in row 0 - offset
    endif
    cell+				\ step to next table entry in row 0
  cell +loop
  2drop
[D  ." done"  D]
;
+short-branches

: SaveWids	\ -- ; generate WIDDATA chain
[D cr ." Wids" wids D]
  Wid-Link
  begin
    @ ?dup
   while
    dup wid-used?
    if  dup SaveWidData  endif
    >wid-link
  repeat
[D cr ." done" D]
;

: RestoreWidData	{ ^vd | #thr -- ; needed by MARKER }
[D  cr ." Restoring wordlist " ^vd vd.wid @ ip>nfa .name  D]
;

: RestoreWids	\ -- ; restore previously generated data
  WidDataChain @			\ walk down the wid chain
  begin
    ?dup
  while
    dup RestoreWidData			\ restore data
    dup @  swap ovl_free		\ get next, release memory
  repeat
  WidDataChain off
;

: WriteWidData	\ ^vd -- ; written from previously generated table
  >r
  r@ VD.wid @ FindOwner			\ find OWNER and offset
  r@ VD.owner !  r@ VD.offset !
  r@  /VocData r@ VD.#threads @ cells 2* + >ovl	\ save header + first two rows
  r> drop
;

: WriteWids	\ -- ; write previously generated data
[D cr ." Wids ... " D]
  WidDataChain				\ walk down the wid chain
  begin
    @ ?dup
  while
    dup WriteWidData
  repeat

  /VocData cell / 0			\ null record to terminate
  do  0 n>ovl  loop
[D ." done" D]
;

[D
: .showWidData	\ *VD --
  >r
  cr ." The Wid owned by $" r@ VD.owner @ .dword
  ."  at offset $" r@ VD.offset @ .dword
  ."  needs patching with "
  r@ VD.data  r@ VD.#threads @ 2* cells  dump
  r> drop
;

\ ***WID*** ???
: .showVDs	\ --
  WidDataChain				\ walk down the wid chain
  begin
    @ ?dup
  while
    dup .ShowWidData
  repeat
;
D]

: ReadWids	{ | ^vd t[ /VocData ] -- ; Read WID data }
[D  cr ." READ: Wids"  D]
  begin
    t[ /VocData ovl>			\ read header
    t[ VD.#threads @			\ terminated by #threads=0
  while
    /VocData  t[ VD.#threads @ cells 2*  +	\ header + first two rows
    ovl_alloc -> ^vd				\ need memory
    t[ ^vd /VocData move			\ copy header
    ^vd VD.data  t[ VD.#threads @ cells 2* ovl>	\ read threads
    ^vd WidDataChain AddLink			\ add to chain
  repeat
;

: ApplyWid	{ ^vd | wid #thr -- ; apply fix ups to voc/wordlist }
  ^vd VD.owner 2@ FindAbs -> wid	\ find vocabulary address
  ^vd Vd.#threads @ cells -> #thr	\ offset into row 1
  ^vd Vd.data  wid >threads		\ source and dest
  ^vd VD.#threads @ 0 ?do		\ -- src dest
    over @  dup -1 <>  swap 0<> and if	\ if new top NFA valid (row 0)
      dup @ >r				\ get current NFA
      over @ +code  over !		\ reset wid thread
      ^vd VD.owner @ -1 <> if		\ if wordlist is not in overlay
        over #thr + @ +code  r> swap !	\ patch lowest LFA in overlay
      else
        r> drop
      endif
    endif
    cell+ swap cell+ swap		\ step to next thread
  loop					\ -- src' dest'
  2drop
;

: ApplyWids	\ -- ; apply wordlist fixups
[D  cr ." ApplyWids"  D]
  WidDataChain @
  begin
    ?dup
   while
    dup ApplyWid
    dup @  swap ovl_free
  repeat
  WidDataChain off
;

: SaveVocLink	\ --
[D cr ." SaveVocLink ... " D]
  voc-link @ in-copy2? if		\ voc has been defined
    Voc-link
    begin
      @ >voc-link			\ get next voc (in Copy2)
      dup @ in-Copy2? 0=		\ if it points outside, we're done
    until
    second-dp -				\ convert to offset
    dup img1 + off  img2 + off		\ zero the chains in the images
  endif
[D ." done" D]
;

: RestoreVocLink	\ --
;

: SaveWidLink	\ --
[D cr ." SaveWidLink ... " D]
  Wid-link @ in-copy2? if		\ wid has been defined
    Wid-link
    begin
      @ >wid-link			\ get next wid (in Copy2)
      dup @ in-Copy2? 0=		\ if it points outside, we're done
    until
    second-dp -				\ convert to offset
    dup img1 + off  img2 + off		\ zero the chains in the images
  endif
[D ." done" D]
;

: RestoreWidLink	\ --
;


\ --------------
\ WID-LINK chain
\ --------------

: WriteWidLink	\ --
[D cr ." WidLinks" D]
  wid-link @ in-copy2? if		\ wid has been defined
    wid-link @ addr>off n>ovl		\ write top wid
    wid-link
    begin
      @ >wid-link			\ get next wid (in Copy2)
      dup @ in-Copy2? 0=		\ if it points outside, we're done
    until
    addr>off n>ovl			\ write low wid link
  else
    -1 n>ovl  -1 n>ovl			\ write dummies
  endif
;

: ReadWidLink	\ -- ; Read WidLink change record
[D  cr ." ReadWidLink"  D]
  /VL ovl_alloc 			\ get space
  dup /VL cell /string ovl>		\ read it, but not the link
  WidLinkChain AddLink			\ add to chain
;

: (ApplyWidLink)	{ ^vl -- ; apply widlink chain }
  ^vl VL.HighVocLink @ -1 <> if
    wid-link @  ^vl VL.LowVocLink @ +code !
    ^vl VL.HighVocLink @ +code wid-link !
  endif
;

: ApplyWidLink	\ -- ; apply widlink chain
[D  cr ." ApplyWidLink "  D]
  WidLinkChain @
  begin
    ?dup
   while
    dup (ApplyWidLink)
    dup @  swap ovl_free
  repeat
  WidLinkChain off
;


\ --------------
\ VOC-LINK chain
\ --------------

: WriteVocLink	\ --
[D cr ." VocLinks" D]
  voc-link @ in-copy2? if		\ voc has been defined
    voc-link @ addr>off n>ovl		\ write top wid
    Voc-link
    begin
      @ >voc-link			\ get next voc (in Copy2)
      dup @ in-Copy2? 0=		\ if it points outside, we're done
    until
    addr>off n>ovl			\ write low voc link
  else
    -1 n>ovl  -1 n>ovl			\ write dummies
  endif
;

: ReadVocLink	\ -- ; Read VocLink change record
[D  cr ." ReadVocLink"  D]
  /VL ovl_alloc 			\ get space
  dup /VL cell /string ovl>		\ read it, but not the link
  VocLinkChain AddLink			\ add to chain
;

: (ApplyVocLink)	{ ^vl -- ; apply voclink chain }
  ^vl VL.HighVocLink @ -1 <> if
    voc-link @  ^vl VL.LowVocLink @ +code !
    ^vl VL.HighVocLink @ +code voc-link !
  endif
;

: ApplyVocLink	\ -- ; apply voclink chain
[D  cr ." ApplyVocLink "  D]
  VocLinkChain @
  begin
    ?dup
   while
    dup (ApplyVocLink)
    dup @  swap ovl_free
  repeat
  VocLinkChain off
;


\ *************************
\ LIBRARY and EXTERN chains
\ *************************

: SaveLibLink	\ --
  lib-link @ in-copy2? if		\ lib has been defined
    /LIB ovl_alloc >r
    lib-link dup @ r@ LIB.high !
    begin
      @ >liblink			\ get next lib (in Copy2)
      dup @ in-Copy2? 0=		\ if it points outside, we're done
    until
    dup r@ LIB.low !			\ save lib low link address
    second-dp -
    dup img1 + off  img2 + off		\ zero the chains in the images
    r> LibChain AddLink
  endif
;

: RestoreLibLink	\ --
  LibChain @ ?dup if
    ovl_free
    LibChain off
  endif
;

: WriteLibLink	\ --
  LibChain @ ?dup if			\ lib has been defined
    dup LIB.high @ addr>off n>ovl	\ write top lib
    LIB.low @ addr>off n>ovl		\ write low link
  else
    -1 n>ovl  -1 n>ovl			\ write dummies
  endif
;

: ReadLibLink	\ -- ; Read LIB-LINK change record
  /LIB ovl_alloc 			\ get space
  dup  cell+ 2 cells ovl>		\ read saved data only
  LIBChain AddLink			\ add to chain
;

: ApplyLibLink	\ -- ; apply voclink chain
  LibChain @
  begin
    ?dup
   while
    dup LIB.high @ -1 <> if
      lib-link @  over LIB.low @ +code !
      dup LIB.high @ +code  lib-link !
    endif
    dup @  swap ovl_free
  repeat
  LibChain off
;

: SaveImportLink	\ --
  import-func-link @ in-copy2? if	\ import has been defined
    /IMPORT ovl_alloc >r
    import-func-link dup @ r@ IMPORT.high !
    begin
      @					\ get next lib (in Copy2)
      dup @ in-Copy2? 0=		\ if it points outside, we're done
    until
    dup r@ IMPORT.low !			\ save lib low link address
    second-dp -
    dup img1 + off  img2 + off		\ zero the chains in the images
    r> ImportChain AddLink
  endif
;

: RestoreImportLink	\ --
  ImportChain @ ?dup if
    ovl_free
    ImportChain off
  endif
;

: WriteImportLink	\ --
  ImportChain @ ?dup if			\ lib has been defined
    dup IMPORT.high @ addr>off n>ovl	\ write top lib
    IMPORT.low @ addr>off n>ovl		\ write low link
  else
    -1 n>ovl  -1 n>ovl			\ write dummies
  endif
;

: ReadImportLink	\ -- ; Read LIB-LINK change record
  /IMPORT ovl_alloc 			\ get space
  dup cell+ 2 cells ovl>		\ read saved data only
  ImportChain AddLink			\ add to chain
;

: ApplyImportLink	\ -- ; apply import link chain
  ImportChain @
  begin
    ?dup
   while
    dup IMPORT.high @ -1 <> if
      import-func-link @  over IMPORT.low @ +code !
      dup IMPORT.high @ +code  import-func-link !
    endif
    dup @  swap ovl_free
  repeat
  ImportChain off
;


\ **********************
\ Load and Release Hooks
\ **********************

: SetOvlLoadHook	\ xt --
  curr-dp OVI.loadhook !
;

: SetOvlUnloadHook	\ xt --
  curr-dp OVI.unloadhook !
;


\ ****************
\ Chain management
\ ****************
\ So that recovery after an error is possible

: GenerateChainData	\ -- ; generate info that must be restored
  SaveWids				\ generate wordlist info
  SaveVocLink				\ generate VocLink chain info
  SaveWidLink				\ generate WidLink chain info
  SaveLibLink				\ libraries
  SaveImportLink			\ imported functions
;

: RestoreChainData	\ -- restore chain data
  RestoreWidLink			\ so that MARKER can work
  RestoreVocLink			\ so that MARKER can work
  RestoreWids				\ so that MARKER can work
  RestoreLibLink			\ so that MARKER can work
  RestoreImportLink			\ so that MARKER can work
;

: SetMarker		\ -- ; make a marker for the cleanup
  s" anew $$EliminateOverlay$$" evaluate
;

: ExecMarker	\ -- ; execute clean up marker
  s" $$EliminateOverlay$$" evaluate
;


\ ******************
\ Consistency checks
\ ******************

\ variable remembered_Ovl

: remember-rule-info    \ --
\  ovl-link @ remembered_Ovl !
;

: CheckRuleInfo		\ -- ; check for overlay conditions failing
\  ovl-link @ remembered_Ovl @ <> if
\    CleanAfterMake
\    ovl_NoPreload ovl_throw
\  endif
  len2 ovlen <> if
    CleanAfterMake
    Ovl_BadLengths ovl_throw
  endif
;


\ ******************
\ Main Overlay words
\ ******************

: .overlays	\ -- ; display loaded overlays
  ovl-link @ 0= if
    cr ." No overlays loaded"  exit
  endif
  ovl-link
  begin
    @ ?dup
   while
    cr
    dup OVI.start @ .dword ." .." dup ovi.end @ .dword  ."   "
    dup OVI.name count type
    OVI.link
  repeat
;

\ ---------
\ Releasing
\ ---------

: UnlinkOVI	\ *ovi -- ; SFP022
  [ also system ]
  dup OVI.unloadhook @ execute		\ execute the unload hook
  dup ['] forth-kernel-end here within if	\ in the dictionary?
    ovi.start @				\ get overlay start address
    dup here trim-dictionary		\ the essence of FORGET
    dp !				\ and reset HERE
  else
    dup OVI.end 2@ trim-dictionary	\ unhook the overlay region
    dup OVI.flags @ dictmask and	\ if not loaded into dictionary
    if  drop  else  ovl_free  endif	\ release overlay memory
  endif
  [ previous ]
;

[-sin
: ReleaseOvl	\ *ovi --
  ovl-link @
  begin				\ -- *ovi *next
    dup OVI.link @			\ -- *ovi *next link
    over UnLinkOVI
    -rot = if				\ -- link
      ovl-link !
      exit
    then
  again
;
sin]

: ReleaseAllOvls	\ --
  begin
    ovl-link @
   while
     ovl-link @ ReleaseOvl
  repeat
;

: $ReleaseOvl	\ c-addr u -- ior ; 0 for success
  OvlLoaded? if
    ReleaseOvl 0
   else
    drop ovl_NotLoaded
  then
;

prune:		\ start end -- start end ; when child of MARKER is executed
  ovl_in_dict @
  if  ReleaseAllOvls  endif
;


\ -------
\ Loading
\ -------

: ($LoadOverLay) \ c-addr u -- start end ; THROWS on Error
  { | #fn fn[ max_path ] #ovn ovn[ max_path ] -- }

  fn[ max_path erase
  ovn[ max_path erase

  -> #fn  fn[ #fn move			\ copy string
  fn[ #fn upper				\ must be upper case
  fn[ ovn[ #fn move			\ copy name		       }
  #fn -> #ovn				\ and length		       } moved SFP006
  S" .OVX" fn[ #fn AddExt -> #fn drop	\ add extension if none there  }

  fn[ #fn OvlLoaded? if
    [ also system ] bverboseinclude?  [ previous ] if
      cr ." Overlay " fn[ #fn type ."  already loaded"
    endif
    ovi.end 2@  exit			\ SFP001
  else
    drop
  endif
  [ also system ] bverboseinclude? [ previous ] if
    cr ." Loading overlay " fn[ #fn type
  endif

  ovl_in_dict @ if			\ check for overlay loading during MAKEOVERLAY
    pass# cells allot			\ to produce different load address in both pases.
  endif

  fn[ #fn OpenOverlay			\ Open overlay ; SFP006
  ReadReloc				\ Read relocation information - MUST be before binary
  ReadCode				\ the binary, not yet relocated
  ReadWidLink				\ Wid-Link information
  ReadVocLink				\ Voc-Link information
  ReadWids				\ Vocabulary information
  ReadLibLink				\ libraries
  ReadImportLink			\ imported functions
  ReadNeededOverlays		        \ only loaded if not present
  CloseOverLay				\ done with overlay file

  ApplyCode				\ Move code where needed
  ApplyReloc				\ Relocate it
  ApplyWids				\ Install wordlist chains
  ApplyWidLink				\ Relink the WID-LINK chain
  ApplyVocLink				\ Relink the VOC-LINK chain
  ApplyLibLink				\ Relink the LIB-LINK chain
  ApplyImportLink			\ Relink the IMPORT-FUNC-LINK chain
  LoadedOverlayChain FreeChain		\ release overlay chain

  Init-Libs				\ initialise any libraries
  Init-Imports				\ initialise any imported functions

  OvBuffer dup ovlen +			\ -- start end
  over OVI.loadhook @ execute		\ execute load hook
  over >r				\ save start ; SFP001
  2dup r@ OVI.end 2!			\ set limits of overlay in overlay itself ; SFP001
  ovl_in_dict @ if			\ check for overlay loading during MAKEOVERLAY
    dictmask r@ ovi.flags set-mask
  endif

  ovl-link @ r@ OVI.link !		\ update overlay link ; SFP001
  r> ovl-link !				\ SFP001
;

: $LoadOverLay  \ c-addr u -- start end
  0 getModuleHandle -> Kernelbase	\ get base address of Forth kernel
\ save chains and vars
  LoadedOverlayChain @off >r
  WidDataChain @off >r
  WidLinkChain @off >r
  VocLinkChain @off >r
  LibChain @off >r
  ImportChain @off >r
  oreloc-chain @off >r
  #bytes @off >r
  fp >r  	-1 -> fp
  ovlen >r      0 -> ovlen
  ovCompiled >r	0 -> ovCompiled
  ovReloc >r	0 -> ovReloc
  OvBuffer >r	0 -> OvBuffer
  RelocBase >r  0 -> RelocBase		\ SFP019
  RelocOffset >r  0 -> RelocOffset	\ SFP019
  RelocSize >r  0 -> RelocSize		\ SFP019

\ perform the base function, and trap errors
  ['] ($LoadOverLay) catch ?dup if
    nip nip
    dup .throw  -1			\ return throw code and -1 ; SFP021
  endif
\ -- start end

\ restore the vars and chains
  ?Chains

  r> -> RelocSize			\ SFP019
  r> -> RelocOffset			\ SFP019
  r> -> RelocBase			\ SFP019
  r> -> OvBuffer
  r> -> ovReloc
  r> -> ovCompiled
  r> -> ovlen
  r> -> fp
  r> #bytes !
  r> oreloc-chain !
  r> ImportChain !
  r> LibChain !
  r> VocLinkChain !
  r> WidLinkChain !
  r> WidDataChain !
  r> LoadedOverlayChain !
;
assign $LoadOverlay to-do ReloadOverlay


\ ------------
\ Dependencies - overlays that must be preloaded
\ ------------

0 value $depends	\ -- c-addr ; pointer to dependency list
0 value #depends	\ -- u ; length of dependency list

: [dependencies	\ -- ; read the dependency list
  { | buff[ $100 chars ] -- }
  0 -> $depends  0 -> #depends
  $10000 ovl_alloc -> $depends
  begin
    bl get-word count			\ get file name or DEPENDENCIES]
    2dup buff[ place			\ make an upper case copy
    buff[ count 2dup upper		\ check for end of list
    s" DEPENDENCIES]" compare
   while				\ -- addr len
    tuck $depends #depends + place	\ add to list as counted string
    1+ add #depends			\ update length
  repeat
  2drop
  0 $depends #depends + !		\ terminate list
;

: GetDependents	\ -- ; get all overlays in the list
  $depends if				\ -- c-addr
    $depends
    begin
      dup count dup			\ while non-zero string
     while				\ -- c-addr c-addr' len'
       $loadoverlay 2drop		\ load the overlay
       count +				\ step over string
    repeat
    3drop
  endif
;

: TermDependents	\ -- ; clean up dependencies
  $depends ovl_free
  0 -> $depends  0 -> #depends
;


\ ------
\ Making
\ ------

expose-module IDE
: NextOverlayID	\ --
\ Read the overlay ID and increment it
\  szConfigName zcount Ini.Open if  exit  endif
  IniFile$ count Ini.Open if  exit  then
  S" Overlay" Ini.Section
  S" OvlID"   0 Ini.ReadInt
  1+ dup 0<
  if  drop  1  endif
  Ovl-ID !
  S" OvlID"   Ovl-ID @ Ini.WriteInt
  Ini.Close
;
previous


: (MakeOverlay)	\ "src-name" "dest-name" --
  GetPathSpec SourceName place		\ get build file name

  GetPathSpec ?dup if			\ get output file name ; SFP021
    OutputName place			\ stash for reporting
    S" .OVX" OutputName count AddExt	\ Add .OVX if no existing extension
    2dup swap 1 chars - c!		\ hack to update count byte
  else
    drop
    SourceName OutputName $move		\ recycle input name
    S" .OVX" OutputName count ForceExt	\ Add .OVX regardless of existing extension
    2dup swap 1 chars - c!		\ hack to update count byte
  endif
  CreateOverlay				\ create the file

  remember-rule-info                    \ remember things which shouldn't change

  fence @ HERE RemoveSinInRange		\ delete existing source inline information

  SetMarker				\ for later clean up
  GetDependents				\ required overlays
  NextOverlayID				\ guarantee unique ID
  BuildOverlay -> ovlen  -> initial-dp	\ First build, remember where 1st build starts
  initial-dp here RemoveSinInRange
  initial-dp ovlen CopyImg -> img1

  SetMarker				\ executes previous one
  GetDependents				\ required overlays
  InitLoadedOverlays			\ initialise RESOV chain
  pass# cells allot			\ second copy at fixed offset from first

  BuildOverlay -> len2  -> second-dp	\ Second build goes here
  second-dp here RemoveSinInRange
  second-dp len2 CopyImg -> img2

  SaveBeforeMake			\ prevents clean-up problems ; SFP016

  GenerateChainData			\ make data chains

  CheckRuleInfo				\ Consistency check

  WriteReloc				\ write relocation info
  img2 len2 WriteCode			\ write binary code - MUST be after relocation data
  WriteWidLink				\ write WID-LINK info
  WriteVocLink				\ write VOC-LINK info
  WriteWids				\ write out the vocabulary records
  WriteLibLink				\ libraries
  WriteImportLink			\ imported functions
  WriteNeededOverlays			\ write overlay list

  CleanAfterMake			\ restore chains, close, run marker

  [ also system ] bverboseinclude?  [ previous ] if
    cr #bytes @ . ." bytes written to " OutputName count type
  endif
;

: MakeOverlay	\ "src-name" "dest-name" --
  ReleaseAllOvls			\ must start with clean slate
  CheckDict				\ barf if dictionary bad ; SFP020
  0 getModuleHandle -> Kernelbase	\ get base address of Forth kernel
  0 -> pass#
  ResetChains

  ovl_in_dict @ >r  ovl_in_dict on	\ overlays must be loaded in the dictionary
  warnings? >r  -warnings		\ turn off redefine warnings
  sin? >r  sindoes? >r			\ save SIN state ; SFP019

  ['] (MakeOverLay) catch

  r> if +sindoes else -sindoes endif	\ restore source inliner state ; SFP006
  r> if +sin  else -sin endif
  r> if +warnings else -warnings endif	\ restore warning state
  r> ovl_in_dict !

  ?Chains
  CheckDict				\ barf if dictionary bad ; SFP020

  ovl_throw				\ uses ior returned by CATCH above
;


\ -----------
\ Cleaning up
\ -----------

: (SaveBeforeMake)	\ -- ; save data needed by below ; SFP016
;
assign (SaveBeforeMake) to-do SaveBeforeMake

: (CleanAfterMake)	\ -- ; clean up and restore ; SFP014
  CheckDict
  img1 ovl_free
  img2 ovl_free
  TermDependents

  RestoreChainData  CloseOverlay	\ restore data
  ExecMarker				\ run marker
;
assign (CleanAfterMake) to-do CleanAfterMake



\ **********************
\ Visibility control 2/2
\ **********************

\ ===================
\ *N Overlay glossary
\ ===================

also Forth definitions

(( for DOCGEN
defer ovl-init-compile	\ -- ; set initial state
\ *G A DEFERred word to set the initial compilation state for
\ ** both compilations of the overlay source code. The default
\ ** condition is:
\ *E   decimal optimised -short-branches +sin +sindoes
\ *P Do not rely on this word being present in future releases.
\ ** It is only present for experimental use with very large
\ ** overlays.
:noname
  decimal optimised -short-branches +sin +sindoes
; to-do ovl-init-compile
))

: [dependencies	\ -- ; set up dependency list
\ *G This word is used before MAKEOVERLAY below to define a list
\ ** of overlays required by the overlay to be made. It is followed
\ ** by a list of overlay file names as required by LOADOVERLAY below.
\ ** The list is termininated by DEPENDENCIES]. Use in the form:
\ *E [dependencies
\ **   primovl secovl ...
\ ** dependencies]
  [dependencies
;

: $MakeOverlay	\ c-addr1 u1 c-addr2 u2 --
( *G Use the first string as the source file name and the second string )
( ** as the overlay name. This word constructs a MAKEOVERLAY string and )
( ** EVALUATES it. $MAKEOVERLAY is provided for the construction of )
( ** higher level overlay management functions. )
		{ | eval[ max_path ] -- }
  S" MAKEOVERLAY " eval[ place
  2swap eval[ append
  S"  " eval[ append
  eval[ append
  state @off >r
  eval[ count evaluate
  r> state !
;

: MakeOverlay	\ "src" ["dest"] -- ; MAKEOVERLAY <buildfile> <overlay>
( *G Creates an overlay by loading an input file, which can itself load )
( ** other files, and producing an output file. If the source file name )
( ** does not have an extension, the rules of INCLUDED will be followed, )
( ** checking for files with extensions ".BLD" ".FTH" ".F" ".CTL" ".SEQ" )
( ** in that order. If the destination file name does not have an extension )
( ** ".OVX" will be used. If the destination file name is not provided, )
( ** the source file name is used with a ".OVX" extension. Thus, just )
( ** typing MAKEOVERLAY FOO will compile FOO.FTH to create FOO.OVX. )
( ** The overlay name held by the system is the output specification )
( ** as given. This is important when reloading the overlay. )
( ** The compiler imposes the following initial condition before the )
( ** overlay file is compiled: )
( *E   DECIMAL -SHORT-BRANCHES +SIN +SINDOES )
  MakeOverlay
;

: SetOvlLoadHook	\ xt -- ; ' <load-action> SETOVLOADHOOK
( *G This word sets the action to be performed whenever the overlay )
( ** is loaded from the file. This action is NOT called by LOADOVERLAY )
( ** if the overlay is already loaded. SETOVLLOADHOOK must be included )
( ** in the overlay load file. )
  SetOvlLoadHook
;

: SetOvlReleaseHook	\ xt -- ; ' <release-action> SETOVLRELEASEHOOK
( *G This word sets the action to be performed when the overlay )
( ** is released. SETOVLRELEASEHOOK must be included )
( ** in the overlay load file. )
  SetOvlUnloadHook
;

: SetOvlVer	\ c-addr --
( *G Sets the address of a counted string added to the version control )
( ** information. All overlay loads will be checked against this string. )
( ** SETOVLVER must be used before MAKEOVERLAY. The string can be reset )
( ** at any time by 0 SETOVLVER. )
  to User_ver$
;

: $OvlLoaded?	\ c-addr u -- start true | 0 0
( *G Converts the string to upper case and tests whether or not )
( ** the overlay has been loaded, returning its start address )
( ** in memory and true if loaded, or two zeros if not loaded. )
( ** See MAKEOVERLAY for a discussion of overlay names. )
  2dup upper OvlLoaded?
;

: $LoadOverlay	\ c-addr u -- start|ior end|-1
( *G Uses the given string as an overlay name, and reloads the )
( ** the overlay if not already loaded. If the overlay name does )
( ** not have an extension, ".OVX" will be used. Any other required )
( ** overlays will be loaded before the requested overlay. The start )
( ** and end+1 address of the overlay code after installation are )
( ** returned. $LOADOVERLAY is provided for the construction of )
( ** higher level overlay mangement functions. )
( ** On error, the start and end values are replaced by ior and -1. )
  $LoadOverlay
;

: LoadOverLay   \ "name" -- ; LOADOVERLAY <name>
( *G Load an overlay whose name follows in the input stream. )
( ** See $LOADVERLAY for more details. )
  GetPathSpec $LoadOverLay 2drop	\ SFP021
;

: .overlays	\ -- ; display loaded overlays
( *G Shows the names of the the loaded overlays. )
  .overlays
;

: lo	        \ "name" -- ; LO <name>
( *G A synonym for LOADOVERLAY. )
( ** See $LOADVERLAY for more details. )
  LoadOverLay ;

: mo		\ "src" ["dest"] -- ; MO <buildfile> <overlay>
( *G A synonym for MAKEOVERLAY. )
  makeoverlay ;

: $ReleaseOverlay	\ c-addr u -- ior
( *G Release the overlay of the given name, returning a non-zero code if )
( ** the overlay was not loaded. The name is converted to upper case before )
( ** the comparison is performed. $RELEASEOVERLAY is provided for the )
( ** construction of higher level overlay mangement functions. )
\ ** If the overlay was loaded when OVL_IN_DICT was set FALSE (the default),
\ ** overlays loaded after the specified one will also be removed.
\ ** If the overlay was loaded when OVL_IN_DICT was set TRUE, the overlay
\ ** is in the 'kernel' area of the dictionary, and any code compiled or loaded after
\ ** the overlay will also be removed. Overlays dependent on this one will
\ ** be removed.
  2dup upper  $ReleaseOvl
;

: ReleaseOverlay	\ "text" -- ; RELEASEOVERLAY <name>
( *G Uses $RELEASEOVERLAY to release the overlay whose name follows. )
\ ** See $RELEASEOVERLAY for more details.
  GetPathSpec $ReleaseOverlay drop	\ SFP021
;

: ro		\ "text" -- ; RO <name>
( *G A synonym for RELEASEOVERLAY. )
\ ** See $RELEASEOVERLAY for more details.
  ReleaseOverlay ;

: ReleaseAllOverlays	\ --
\ *G Releases and unhooks all overlays. Executed automatically by
\ ** the Exit chain.
  ReleaseAllOvls
;
' ReleaseAllOvls AtExit

: ovl_in_dict	\ -- addr ; true to load overlays in dictionary ; SFP022
\ *G Set this variable to TRUE to load overlays at the end of the
\ ** dictionary, rather than in memory allocated from the heap.
\ ** This is only required in special circumstances.
\ ** After overlays have been built, restore OVL_IN_DICT
\ ** to FALSE.
  ovl_in_dict
;

previous definitions

previous definitions

+sin -short-branches


\ ======
\ *> ###
\ ======



