\ SWITCHES for VFX Forth

((
Copyright:
----------
Copyright (c) 1999, 2001 by:
MicroProcessor Engineering
133 Hill lane
Southampton SO15 5AF
England
tel: +44 (0)2380 631441
net: tech-support@mpeforth.com
web: http://www.mpeforth.com

Change history
==============
20010103 SFP002 Modified SWITCH] to avoid source inline problems
19990928 SFP001 Original release

Documentation
=============
Derived from an article in Forth Dimensions vol. XX no. 3
by Rick vanNorman.

The SWITCH structure is built as follows:
HEAD
cell	link to previous switch
cell	anchor for item chain
cell	default action

ITEM
cell	link to previous item
cell	id for this item
cell	xt to run
))

( *! switches     )
( *T Switch chains )
( *S Introduction )
( *P Switch chains provide a mechanism for generating extensible chains similar  )
( ** to the CASE ... OF ... ENDOF ... ENDCASE control structure, except that )
( ** the user may extend the chain at any time. These chains are of particular use )
( ** when defining winprocs whose action may need to be adjusted or extended after )
( ** the chain itself has been defined. )

( *P The code may have been precompiled, but its source code may be found in )
( ** the file SWITCHES.FTH in the LIB directory. )

( *S An example )
( *P The following example shows how to define a simple chain that translates numbers )
( ** to text. At a leter date, translations in Italian are added. )

( *P Define some words which will be executed by the chain. )
( *E : one   ." one"  ; )
( ** : two   ." two"  ; )
( ** : three ." three" ; )
( ** : many  . ." more" ; )

( *P The following definition defines a switch called NUMBERS which )
( ** executes ONE when 1 is the selector, TWO if 2 is the selector, )
( ** or MANY if any other number is the selector. Note that MANY must )
( ** consume the selector. The word RUNS associates a word with the )
( ** given selector. )
( *E [switch numbers many )
( **   1 runs one )
( **   2 runs two )
( ** switch] )
( ** cr 1 numbers )
( ** cr 5 numbers )

( *P The next piece of code extends the NUMBERS switch chain, and )
( ** demonstrates the use of RUN: to define an action without giving )
( ** it a name. )
( *E [+switch numbers )
( **   3 runs three )
( **   4 run: ." four" ; )
( **   5 run: ." five" ; )
( ** switch]
( ** cr 1 numbers )
( ** cr 5 numbers )
( ** cr 8 numbers )

( *P The following portion of this example demonstrates how selectors )
( ** are overridden by the last action defined. Although an action )
( ** has already been defined for selectors 1 and 2, if another action )
( ** is defined, it will be found before the old ones, and so the action )
( ** will be performed. )
( *E [+switch numbers	)
( **   1 run: ." uno" ;	)
( **   2 run: ." due" ;	)
( ** switch]		)
( ** cr 1 numbers	)
( ** cr 2 numbers	)
( ** cr 3 numbers	)
( ** cr 5 numbers	)
( ** cr 8 numbers	)

( *S Switches glossary )

[undefined] link, [if]
: link,		\ addr -- ; update chain anchored at addr
( *G Given the address of a variable anchoring a chain, lay )
( ** a link and update the chain with HERE. )
  here  over @ ,  swap !
;
[then]

variable switch-link	\ -- addr ; links all switches

struct /SwitchHead	\ -- size
  cell field sw_link		\ link to previous switch
  cell field sw_itemlink	\ anchor for item chain
  cell field sw_defxt	        \ default action
end-struct

struct /SwitchItem	\ -- size
  cell field itm_link	        \ link to previous item
  cell field itm_id	        \ id for this item
  cell field itm_xt	        \ xt to run
end-struct

((
: switcher	\ i*x id switchhead -- j*x
  dup sw_defxt @ >r		\ save default action
  sw_itemlink @
  begin				\ -- id item
    ?dup
   while			\ -- id item
    2dup itm_id @ =		\ check ids match
    if
      nip itm_xt @ execute
      r> drop  exit
    endif
    itm_link @
  repeat
  r> execute
;
))
code switch	\ i*x id switchhead -- j*x
( *G Given an id and the head of a switch chain, SWITCH )
( ** will perform the action of the given id if it is found, )
( ** otherwise it will perform the default action, passing id )
( ** to that action. )
  push	0 sw_defxt [ebx]		\ save default action
  mov	ecx, 0 sw_itemlink [ebx]	\ get head of chain
  mov	eax, 0 [ebp]			\ get id
  begin,
   ncxz, while,
    cmp	  eax, 0 itm_id [ecx]		\ if ids match
    z, if,
      lea    esp, cell [esp]		\   discard default action
      mov    ebx, 4 [ebp]		\   restore TOS
      lea    ebp, 8 [ebp]
      jmp    dword 0 itm_xt [ecx]	\   execute itm_xt
    endif,
    mov	  ecx, 0 itm_link [ecx]		\ get next item
  repeat,
  mov	ebx, eax			\ put id in TOS
  lea   ebp, 4 [ebp]			\ adjust stack
  ret				        \ executes default action
end-code

: SwitchHead,	\ -- addr ; builds a switch chain head
  here				\ address of head
  switch-link link,		\ links all switches
  0 ,				\ no items added yet
  ' ,				\ set default action
;

: ItemHead,	\ head id -- head item
  here >r			\ address of item head
  over sw_itemlink link,	\ link to previous item
  ,				\ id
  0 ,				\ dummy xt
  r>
;

: [switch	\ "default" -- head ; i*x id -- j*x
( *G Builds a new named switch list with a default action. Use in )
( ** the form: [SWITCH <name> <default.action> where <default.action> )
( ** must consume the selector id. )
  create
    SwitchHead,
  does>
    switch
;

: [+switch	\ "head" -- head ; to extend an existing switch
( *G Used in the form [+SWITCH <switch> to extend an existing )
( ** switch chain. )
  ' >body
;

: run:		\ head id -- head ; add nameless action to switch
( *G Used in the form  <id> RUN: <words> ; to define a nameless )
( ** action in a switch chain. )
  itemhead, :noname swap itm_xt !
;

: runs		\ head id "word" -- head
( *G Used in the form  <id> RUNS <word> to define a named )
( ** action in a switch chain. )
  itemhead,  ' swap itm_xt !
;

: switch]	\ head -- ; finishes a switch chain or extension
( *G Used to finish a [SWITCH <name> <default> or [+SWITCH <name> )
( ** chain definition. )
\  drop
  body> dup >code-len off  >code-gen off
;

: .switches	\ -- ; lists defined switches
( *G Lists all the defined switch chains. )
  switch-link
  begin
    @ ?dup
   while
    dup ip>nfa .name
    sw_link
  repeat
;

(( \ test code
: one   ." one"  ;
: two   ." two"  ;
: three ." three" ;

: many  . ." more" ;

[switch numbers many
  1 runs one
  2 runs two
switch]

cr 1 numbers
cr 5 numbers

[+switch numbers
  3 runs three
  4 run: ." four" ;
  5 run: ." five" ;
switch]

cr 1 numbers
cr 5 numbers
cr 8 numbers

[+switch numbers
  1 run: ." uno" ;
  2 run: ." due" ;
switch]

cr 1 numbers
cr 2 numbers
cr 3 numbers
cr 5 numbers
cr 8 numbers
))

