\ VFX Forth v5.1+ hardware floating point support for x64 using NDP stack only

((
Copyright:
----------
Copyright (c) 1998-2009, 2011, 2012, 2014, 2015, 2016, 2018, 2019, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441

net: tech-support@mpeforth.com
web: http://www.mpeforth.com

Requires VFX Forth v5.1 or greater.

To do
=====
Remove all uses of PAD.

According to Forth-94:

    12.6.1.2143 REPRESENT     FLOATING

    ( c-addr u -- n flag1 flag2 )  (F: r -- )   or
      ( r c-addr u -- n flag1 flag2 )

    At c-addr, place the character-string external representation
    of the significand of the floating-point number r.  ...
    The character string shall consist of the u most significant
    digits ...  The significand is rounded to u digits following the
    round to nearest rule ...

The above states the number must be *rounded* to u digits, so

    1234.56E0  PAD 5 REPRESENT
    PAD 5 DUMP
    0CDF:2352  31 32 33 34 36   12346 ok

giving the resulting string  '12346'.

Unfortunately some Forths have REPRESENT truncate and so
give  '12345'.  This is non-compliant and will cause incorrect
number display in applications that use REPRESENT.


Change history
==============
20200508 SFP050 x64 conversion.
20200408 SFP049 Corrected POSTPONE action for FP literals.
20190326 SFP048 Changes for VFX v5.1 with recognisers.
20180628 SFP047 NUMBER? changes for VFX v5.
20160719 HAA046 Fixed disassembly of FLIT.
20151106 ROT045 Added stack juggler extensions.
20141023 GEK044 Fixed FTANH.
20140829 EDA043 More signed zero updates - note that FSIGN
		has changed.
20140531 EDA042 Fix signed zero presentation - mistyped by SFP.
20140318 EDA041 Fix signed zero presentation.
20120609 EDA040 Update FPOUT routines to v3.9 to improve ruggedness.
20110524 MPE039 Major overhaul of FP output routines.
20100210 SFP038 Clarified number conversion documentation.
20091221 SFP037 Reverted to native FATAN2 using FPATAN.
20091024 SFP036 Added FLNP1.
20090928 SFP035 Treated FP-CHAR as a byte array.
20090506 SFP035 Made >FLOAT more Forth200x compliant.
20090429 SFP034 Updated for tokeniser and disassembler.
20090427 SFP033 Corrected (RLITERAL) for v4.4/4.4.
20090410 MPE032 Added FTRUNC and deprecated FINT.
20090318 RAG031 Made FATAN2 return principal angle.
20080922 SFP030 Corrected trig functions.
20080922 SFP029 Corrected rounding errors when generating exponent.
20071026 SFP028 Forced DECIMAL in >FLOAT.
20070103 EDA027 Updated REPRESENT.
20060831 SFP026 Optimiser improvements.
20060830 EDA025 Update REPRESENT.
20060814 SFP024 Updated for v4 optimiser.
20040701 SFP023	Unmarked. Added FP optimiser.
		Restored checked version of F** as the FSL requires it.
20021023 SFP022 Added FR>S and FR>D after much discussion on comp.lang.forth.
20021007 SFP021 Corrected F>S and F>D.
SFP020	20 Mar 2002	Corrected F>ASCII and >FLOAT
SFP019	19 Feb 2002	Corrected REPRESENT and reduced >BCD in size.
			Corrected exponent handling in numeric output.
SFP018	13 Mar 2001	Added F# for PFW2.x compatibility
SFP017	2 Jan 2001	Improved rounding in >BCD.
			Renamed (FLITERAL) to (RLITERAL) to avoid name conflict.
			Added REALS and INTEGERS for user control of number input.
SFP016	27 Sept 2000	For performance, removed error checking from F** as this
			is consistent with the rest of the package.
SFP015	25 Sept 2000	Changed >BCD to preserve temp value as 10 byte real.
			This stops 1E400 being displayed as 9.9999e399 etc.
SFP014	23 Sept 2000	Corrected 2**z to preserve EDX
SFP013	20 Sept 2000	Added FPSYSTEM value to define FP type in use
SFP012	19 Sept 2000	Corrected display for NaNs and Infs
SFP011	24 July 2000	Converted from HFP387
			Corrected FMOD
SFP010	20 July 2000	Inlined some code definitions.
			Renamed popFPU and pushFPU to FPU> and >FPU
SFP009	10 Oct 1999	Corrected rounding control in I/O operations.
			Made 2**Z general, Z may be negative.
			Clipped F** ( F1^F2 ) for F1<=0.
			Enhanced prompt for floats.
			Added FP debugger.
SFP008	9 Oct 1999	Added FVALUE
SFP007	8 Oct 1999	Fixed F~
SFP006  7 Oct 1999      Renamed FULL-NUMBER? to FNUMBER?
			Changed FULL-NUMBER? to check for integers first
			Enhanced the documentation section.
			Changed number output to use FP-CHAR for point separator
SFP005  28 Sept 1999	FST cannot use 80 bit real to memory!
SFP004  15 Sept 1999	Tidy ups and code improvements (unmarked)
SJC005  99/03/08 	Modified FNUMBER?, can only parse float number if BASE is
                	DECIMAL according to ANS
SJC004  99/03/02 	Fixed FPICK after SFP002.
SFP003  11 Feb 1999     Modified for VFX Forth
SFP002  5 Aug 1998      Improved FOVER, FPICK and miscellaneous words
SFP001  1 Aug 1998      Modified for X8632 cross compiler
))

\ ========
\ *! floatndp64
\ *T NDP Floating Point
\ ========

\ **************
\ *S Introduction
\ ***************
\ *P This floating point package uses the FPU/x87 instruction set. The
\ ** source code can be found in the file *\i{Lib/x64/Ndpx64.fth}.
\ ** The other floating point pack, *\i{Lib/x64/FPSSE64S.fth} is the
\ ** default floating point package for modern systems and uses the
\ ** SSE instruction set.
\ ** VFX Forth v5.1+ with recognisers is required.

\ =================================
\ *N Ndpx64.fth - coprocessor stack
\ =================================
\ *P In *\i{Ndpx64.fth} floating point numbers are kept in the
\ ** floating point unit's internal stack only. This code is
\ ** significantly faster than when using an external stack, but
\ ** is limited to the use of 8 floats on the NDP stack, including
\ ** any working temporary numbers.

\ *P By default, *\i{Ndpx64.fth} defines floating point stack
\ ** items and literals to be in 80 bit extended real format.
\ ** This may be of advantage compared to the 64 bit IEEE format.

\ **********************
\ *S Radians and Degrees
\ **********************
\ *P Please note that the trig functions are calculated in
\ ** radians, for calculations in degrees use *\fo{DEG>RAD} beforehand,
\ ** *\fo{RAD>DEG} afterwards.

\ ************************************
\ *S Number formats, ANS and Forth200x
\ ************************************
\ *P The ANS Forth standard specifies that floating point numbers must
\ ** be entered in the form 1.234e5 and must contain a point '.' and
\ ** 'e' or 'E', and that double integers are terminated by a point '.'.

\ *P This situation prevents the use of the standard conversion
\ ** words in international applications because of the
\ ** interchangable use of the '.' and ',' characters in numbers.
\ ** Because of this, VFX Forth uses two four-byte arrays,
\ ** *\fo{FP-CHAR} and *\fo{DP-CHAR}, to hold the characters used
\ ** as the floating point and double integer indicator characters.
\ ** By default, *\fo{FP-CHAR} is initialised to '.' and *\fo{DP-CHAR}
\ ** is initialised to to ',' and '.'. For strict ANS and
\ ** Forth-2012 compliance, you should set them as follows:

\ *E \ ANS standard setting
\ **   char . dp-char !
\ **   char . fp-char !
\ ** : ans-floats    \ -- ; for strict ANS compliance
\ **   [char] . dp-char !
\ **   [char] . fp-char !
\ ** ;
\ ** \ MPE defaults
\ **   char , dp-char !
\ **   char . dp-char 1+ c!
\ **   char . fp-char !
\ ** : mpe-floats    \ -- ; for existing and most legacy code
\ **   [char] , dp-char !
\ **   [char] . dp-char 1+ c!
\ **   [char] . fp-char !
\ ** ;
\ ** \ Legacy defaults, including ProForth
\ **   char , dp-char !
\ **   char . fp-char !
\ ** : legacy-floats \ -- ; for legacy code
\ **   [char] , dp-char !
\ **   [char] . fp-char !
\ ** ;

\ *P You can of course set these arrays to hold any values which
\ ** suit your application's language and locale. Note that integer
\ ** conversion is always attempted before floating point
\ ** conversion. This means that if the *\fo{FP-CHAR} and *\fo{DP-CHAR}
\ ** arrays contain the same character, floating point numbers must
\ ** contain 'e' or 'E'. If the arrays are all different, a number
\ ** containing the *\fo{FP-CHAR} will be successfully converted
\ ** as a floating point number, even if it does not contain 'e'
\ ** or 'E'.

\ *P As of January 2007, recommendations made to the Forth200x
\ ** standards effort have been adopted by MPE for *\fo{REPRESENT}.
\ ** The impact of these changes is that the minimum buffer size
\ ** used for *\fo{REPRESENT} should be at least *\fo{#FDIGITS}
\ ** characters, normally 18 bytes. For details of the proposal,
\ ** see:
\ *C   Examples/usenet/Ed/Represent_20.txt
\ *C   Examples/usenet/Ed/Represent_30.txt

\ ****************************
\ *S Floating point exceptions
\ ****************************
\ *P Exception handling is determined by the operating system. On current
\ ** Windows platforms, floating point exceptions are not generated by
\ ** the NDP. This can be changed by altering the bottom 6 bits of the
\ ** NDP control word using *\fo{CW@} and *\fo{CW!}.

\ *P By default, the system prompt will report exception status, and clear
\ ** the pending exception status. Exception status reporting does not mean that
\ ** the Windows exception handler has been triggered, it only means that
\ ** the status flag has been set.

\ ************************************
\ *S Standards compliance, F>S and F>D
\ ************************************
\ *P After much discussion on the comp.lang.forth newsgroup, a
\ ** consensus was reached that *\fo{F>D} and *\fo{F>S} must
\ ** truncate to zero. This is also the behaviour required by the
\ ** Forth Scientific Library (FSL). Historically, MPE floating
\ ** point packs permit the integer rounding mode to be set by
\ ** the user. In order to support both camps, VFX Forth now
\ ** behaves as follows:
\ *(
\ *B *\fo{F>D} and *\fo{F>S} truncate to zero,
\ *B *\fo{FR>D} and *\fo{FR>S} follow the current rounding mode.
\ *)


\ ****************
\ *S Configuration
\ ****************

[defined] integers [if]  integers  [then]	\ unhook previous FP pack

[undefined] FpSystem [if]
0 value FpSystem
\ *G The value FPSYSTEM defines which floating point pack is installed and
\ ** active. Each floating point pack defines its own type as follows:
\ *(
\ *B 0 constant NoFPSystem
\ *B 1 constant HFP387System (also 64 bit)
\ *B 2 constant NDP387System (also 64 bit)
\ *B 3 constant OpenGL32System (obsolete)
\ *B 4 constant SSE64System (default)
\ *)
\ *P When *\i{FPSystem} changes, the following files that use *\i{FPSystem}
\ ** are affected:
\ *C   Extern*.fth  kernel64.fth  Tokeniser.fth
[then]

\ 0 constant NoFPSystem
\ 1 constant HFP387System
  2 constant NDP387System
\ 3 constant OpenGL32System
NDP387System -> FpSystem

\ Define size of items for default memory access 4, 8, or 10 bytes
#10 constant FPCELL	\ -- n
\ *G Defines the size of literals and floating point numbers in memory
\ ** and on floating point stacks in memory
FPCELL negate constant -FPCELL

#18 constant #fdigits	\ -- u
\ *G Returns the largest number of usable digits available from
\ ** *\fo{REPRESENT}. Equivalent to the environment variable
\ ** *\fo{MAX-FLOAT-DIGITS}.

false constant [fpdebug] immediate
( *G Set this true when compiling *\i{NDPx64.FTH}, and a debug )
( ** build will be constructed. In this state, the state of the )
( ** FPU is checked after each word. If a floating point exception )
( ** has been generated, a diagnostic is issued, and the system )
( ** aborts. Set this only when testing. )
\ ** Note that the NDP optimiser may well cause this to be
\ ** ignored.

[fpdebug] [if]
  defer fpcheck				\ see later for real action
\ *G A DEFERred word called at the end of *\fo{CODE} routines when
\ ** *\fo{[FPDEBUG]} is non-zero.
  assign noop to-do fpcheck
[then]

variable signed-zero	\ -- addr
\ *G Set non-zero to display signed-zero.
  signed-zero off

1 constant FPext?	\ -- flag
\ *G Set non-zero to compile FP extensions.


\ *********************************
\ Variable and frames for FP stacks
\ *********************************

\ Memory copies of the status word, control word, and rounding control word
\ are held on a stack frame built as an offset on the data stack pointed
\ to by RBP.
0 constant #sw          \ offset of status word
2 constant #cw          \ offset of control word
4 constant #rc          \ offset of rounding control word
6 constant #spare       \ dummy to bring up to 8 bytes

8 constant #fpframe     \ size of frame required for all four items


\ *******************
\ *S Assembler macros
\ *******************

also asm-access definitions

: fword		\ --
  tbyte  ;

: fnext,	\ -- ; can be changed for debugging
\ *G The equivalent of *\fo{NEXT,} for floating point routines.
\ ** If *\fo{[FPDEBUG]} is non-zero, a call to *\fo{FPCHECK} is
\ ** assembled.
[fpdebug] [if]
  fwait				\ forces last instruction flush
  call 	[] ['] fpcheck >body	\ FPCHECK is DEFERred
[then]
  ret
;

previous definitions


\ ********************
\ *S Optimiser support
\ ********************

0 value FPsin?	\ -- flag
\ *G Returns non-zero if tokenising is permitted for words
\ ** containing floating point code sequences. By default,
\ ** FP tokenising is disabled.

: [+FPsin	\ -- x
\ *G Start a *\fo{[+FPSIN ... FPSIN]} section in which new FP
\ ** code can be tokenised.
  ?exec  FPsin?  -1 to FPsin?
; immediate

: [-FPsin	\ -- x
\ *G Start a *\fo{[-FPSIN ... FPSIN]} section in which new FP
\ ** code cannot be tokenised.
  ?exec  FPsin?  0 to FPsin?
; immediate

: FPsin]	\ x --
\ *G End a *\fo{[+/-FPSIN ... FPSIN]} section. The previous
\ ** FP stokeniser state is preserved.
  ?exec  to FPsin?
; immediate

: ?noFPsin	\ --
\ Discard tokeniser information if required.
  FPsin? 0=
  if  discard-sinline  endif
;

: fseq:		\ -- ; FSEQ: <name> ... ;FSEQ
\ *G Start an assembler sequence which is compiled for *\fo{<name>}.
  [ also h-c ]
  oc:  postpone drop  also asm-access
  [ previous ]
;

: ;fseq		\ -- ; FSEQ: <name> ... ;FSEQ
\ *G Ends an assembler sequence started by *\fo{FSEQ:}.
  [ also h-c ]
  postpone use-a  postpone reset-opt  postpone ?noFPsin  postpone ;
  previous
  [ previous ]
; immediate


\ ***************
\ *S FP constants
\ ***************

code %0         \ F: -- f#(0)
( *G Floating point 0.0 )
  fldz
  fnext,
end-code
fseq: %0	fldz  ;fseq

code %1         \ F: -- f#(1)
( *G Floating point 1.0 )
  fld1
  fnext,
end-code
fseq: %1	fld1  ;fseq

code %pi        \ F: -- f#(pi)
( *G Floating point PI )
  fldpi
  fnext,
end-code
fseq: %pi	fldpi  ;fseq

code %pi/2      \ F: -- f#(pi/2)
( *G Floating point PI/2 )
  fld1
  fadd  st, st
  fldpi
  fdivrp st(1), st
  fnext,
end-code

code %pi/4      \ F: -- f#(pi/4)
( *G Floating point PI/4 )
  fld1                          \ 1.0
  fadd  st, st                  \ 2.0
  fadd  st, st                  \ 4.0
  fldpi
  fdivrp st(1), st
  fnext,
end-code

code %lg2e	\ F: -- log2(e)
\ *G Returns log (base 2) of e.
  fldl2e  fnext,  end-code
fseq: %lg2e  fldl2e  ;fseq


\ *********************
\ *S FP control operations
\ *********************

code finit	\ F: -- ; resets FPU
\ *G Reset the floating point unit and NDP stack.
  finit
  next,
end-code inline

finit

code cw@	\ -- cw ; get NDP control word
\ *G Return the floating point unit Control Word.
  lea	rbp, -2 cells [rbp]	\ save TOS, reserve one more cell
  mov	cell [rbp], rbx
  fstcw 0 [rbp]			\ get control word to memory
  xor	rbx, rbx
  mov	bx,  0 [rbp]		\ and then to TOS
  lea	rbp, cell [rbp]		\ release spare cell
  fnext,
end-code

code cw!	\ cw -- ; set NDP control word
\ *G Set the floating point unit Control Word.
  lea   rbp, -1 cells [rbp]	\ must load CW to memory
  mov	0 [rbp], rbx		\ new CW is in TOS
  fldcw word 0 [rbp]
  mov	rbx, cell [rbp]		\ restore TOS
  lea	rbp, 2 cells [rbp]
  fnext,
end-code

code sw@	\ -- sw ; get NDP status word
\ *G Return the floating point unit Status Word.
  lea	rbp, -1 cells [rbp]	\ save TOS, reserve one more cell
  mov	0 [rbp], rbx
  fstsw ax			\ get status word
  sub 	rbx, rbx		\ clear result
  mov	bx, ax
  next,				\ so that debugger isn't triggered repeatedly
end-code

code fclex	\ -- ; clear exceptions
\ *G Clear any pending floating point exceptions.
  fclex				\ clear exceptions
  next,
end-code inline


\ *******************
\ *S FP Stack operations
\ *******************

code fdup       \ F: f -- f f
\ *G Floating point equivalent of *\fo{DUP}.
  fld	st			\ duplicate TOS
  fnext,
end-code
also h-c  oc: fdup
  drop  ?noFPsin ( c_shuffle )  a[ fld st ]a use-a  reset-opt
;  previous

code fswap      \ F: f1 f2 -- f2 f1
\ *G Floating point equivalent of *\fo{SWAP}.
  fxch
  fnext,
end-code \ inline
fseq: fswap	fxch  ;fseq

code fdrop      \ F: f --
\ *G Floating point equivalent of *\fo{DROP}.
  fstp  st
  fnext,
end-code \ inline
fseq: fdrop	fstp st  ;fseq

code fover      \ F: f1 f2 -- f1 f2 f1
\ *G Floating point equivalent of *\fo{OVER}.
  fld   st(1)
  fnext,
end-code \ inline
fseq: fover	fld st(1)  ;fseq

code frot       \ F: f1 f2 f3 -- f2 f3 f1
\ *G Floating point equivalent of *\fo{ROT}.
  fxch  st(1)
  fxch  st(2)
  fnext,
end-code
fseq: frot	fxch st(1)  fxch st(2)  ;fseq

lbl: (fpick)
  fld  st	ret		\ 0
  fld  st(1)	ret		\ 1
  fld  st(2)	ret		\ 2
  fld  st(3)	ret		\ 3
  fld  st(4)	ret		\ 4
  fld  st(5)	ret		\ 5
  fld  st(6)	ret		\ 6
  fld  st(7)	ret		\ 7
end-code

\ rewrite with scaled-index LEA
code fpick      \ n -- ; F: -- f
\ *G Floating point equivalent of *\fo{PICK}. Note that because the
\ ** pick index is an integer, it is on the normal Forth integer
\ ** data stack, and the result, being a floating point number,
\ ** is on the floating point stack.
  imul  rax, rbx, # 3		\ 3 bytes per instructions in table
  mov   rbx, 0 [rbp]		\ restore TOS
  add   rax, # (fpick)		\ base of table
  add   rbp, # cell		\ update PSP
  jmp   rax
end-code

((
\ From Andrew Haley on comp.lang.Forth
\ This has a side effect on PAD, unfortunately.  However, it's portable
\ and multi-tasking safe.
: froll		\ n -- ; F: fn..f0 -- fn-1..f0 fn
  ?dup 0= if  exit  then
  >r  pad r@ floats + f!
  r@ 1- recurse
  pad r> floats + f@
  fswap
;
))

code ndepth     \ -- n ; depth of NDP stack
\ *G Returns on the Forth data stack the number of items on the
\ ** FPUs's internal working stack.
  fstsw ax                      \ NDP status
  sub   rbp, # cell             \ preserve TOS
  mov   0 [rbp], rbx
  shr   ax, # #11               \ line up the bits
  and   al, # $07               \ only want bottom three
  mov   rbx, # 8                \ 8 registers on NDP,
  sub   bl, al                  \ convert to depth
  cmp   bl, # 8                 \ convert 8 to 0
  z, if,  sub bl, bl  endif,
  fnext,
end-code

: fdepth        \ -- #f
\ *G Floating point equivalent of *\fo{DEPTH}. The result is reurned
\ ** on the Forth data stack, *\b{not} the float stack.
  ndepth
;


\ *************************************
\ *S Memory operations SF@ SF! DF@ DF! etc
\ *************************************

code f@         \ addr -- ; F: -- f
\ *G Places the contents of addr on the float stack. The size
\ ** of the item fetched was defined by *\fo{FPCELL} at compile time.
  fld   fword 0 [rbx]           \ load from the address
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code sf@        \ addr -- ; F: -- f
\ *G Places the 32 bit float at addr on the float stack.
  fld   float 0 [rbx]           \ load from the address
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code df@        \ addr -- ; F: -- f
\ *G Places the 64 bit double float at addr on the float stack.
  fld   double 0 [rbx]          \ load from the address
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code tf@        \ addr -- ; F: -- f
\ *G Places the 80 bit extended float at addr on the float stack.
  fld   tbyte 0 [rbx]		\ load from the address
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code f!         \ addr -- ; F: f --
\ *G Stores the top of the float stack as an *\fo{FPCELL} sized
\ ** number at addr.
  fstp  fword 0 [rbx]
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code sf!        \ addr -- ; F: f --
\ *G Stores the top of the float stack as an 32 bit float
\ ** number at addr.
  fstp  float 0 [rbx]
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code df!        \ addr -- ; F: f --
\ *G Stores the top of the float stack as an 64 bit double float
\ ** number at addr.
  fstp  double 0 [rbx]
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code tf!        \ addr -- ; F: f --
\ *G Stores the top of the float stack as an 80 bit extended float
\ ** number at addr.
  fstp  tbyte 0 [rbx]
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code f+!	\ F: f -- ; addr -- ; add f to data at addr
\ *G Add F to the data at ADDR.
  fld   fword 0 [rbx]		\ get from addr
  faddp st(1), st               \ do the actual addition
  fstp  fword 0 [rbx]		\ put back at addr
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code f-!	\ F: f -- ; addr -- ; sub f from data at addr
\ *G Subtract F from the data at ADDR.
  fld   fword 0 [rbx]		\ get from addr
  fsubrp st(1), st		\ do the actual subtraction
  fstp  fword 0 [rbx]		\ put back at addr
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code sf+!	\ F: f -- ; addr -- ; add f to data at addr
\ *G Add F to the 32 bit float at ADDR. )
\ ** NDP387.FTH only.
  fadd  float 0 [rbx]		\ add from addr
  fstp  float 0 [rbx]		\ put back at addr
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code sf-!	\ F: f -- ; addr -- ; sub f from data at addr
\ *G Subtract F from the 32 bit float at ADDR. \
\ ** Ndpx64.fth only.
  fsubr float 0 [rbx]		\ subtract from addr
  fstp  float 0 [rbx]		\ put back at addr
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code df+!	\ F: f -- ; addr -- ; add f to data at addr
\ *G Add F to the 64 bit float at ADDR.
\ ** Ndpx64.fth only.
  fadd  double 0 [rbx]		\ add from addr
  fstp  double 0 [rbx]		\ put back at addr
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code df-!	\ F: f -- ; addr -- ; sub f from data at addr
\ *G Subtract F from the 64 bit float at ADDR.
\ ** Ndpx64.fth only.
  fsubr double 0 [rbx]		\ subtract from addr
  fstp  double 0 [rbx]		\ put back at addr
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code tf+!	\ F: f -- ; addr -- ; add f to data at addr
\ *G Add F to the 80 bit float at ADDR.
\ ** Ndpx64.fth only.
  fld   tbyte 0 [rbx]		\ get from addr
  faddp st(1), st               \ do the actual addition
  fstp  tbyte 0 [rbx]		\ put back at addr
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code tf-!	\ F: f -- ; addr -- ; sub f from data at addr
\ *G Subtract F from the 80 bit float at ADDR.
\ ** Ndpx64.fth only.
  fld   tbyte 0 [rbx]		\ get from addr
  fsubrp st(1), st		\ do the actual subtraction
  fstp  tbyte 0 [rbx]		\ put back at addr
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

code f@+        \ addr -- addr' ; F: -- f
\ *G Places the contents of addr on the float stack and increments
\ ** the address. The size of the item fetched and the increment
\ ** is defined by *\fo{FPCELL}.
\ ** Ndpx64.fth only.
  fld   fword 0 [rbx]           \ load from the address
  add   rbx, # FPCELL		\ update address
  fnext,
end-code

code sf@+	\ addr -- addr' ; F: -- f
\ *G Places the 32 bit float at addr on the float stack, and
\ ** increments addr by 4.
\ ** Ndpx64.fth only.
  fld   float 0 [rbx]           \ load from the address
  add   rbx, # cell		\ update address
  fnext,
end-code

code df@+	\ addr -- addr' ; F: -- f
\ *G Places the 64 bit float at addr on the float stack, and
\ ** increments addr by 8.
\ ** Ndpx64.fth only.
  fld   double 0 [rbx]		\ load from the address
  add   rbx, # 8		\ update address
  fnext,
end-code

code tf@+	\ addr -- addr' ; F: -- f
\ *G Places the 80 bit float at addr on the float stack, and
\ ** increments addr by 10.
\ ** Ndpx64.fth only.
  fld   tbyte 0 [rbx]		\ load from the address
  add   rbx, # #10		\ update address
  fnext,
end-code

code f!+        \ addr -- addr' ; F: f --
\ *G Stores the top of the float stack as an *\fo{FPCELL} sized number
\ ** at addr, and updates addr appropriately.
\ ** Ndpx64.fth only.
  fstp  fword 0 [rbx]
  add	rbx, # FPCELL
  fnext,
end-code

code sf!+	\ addr -- addr' ; F: f --
\ *G Stores the top of the float stack as a 32 bit float
\ ** at addr, and updates addr appropriately.
\ ** Ndpx64.fth only.
  fstp  float 0 [rbx]
  add	rbx, # 4
  fnext,
end-code

code df!+	\ addr -- addr' ; F: f --
\ *G Stores the top of the float stack as a 64 bit float
\ ** at addr, and updates addr appropriately.
\ ** Ndpx64.fth only.
  fstp  double 0 [rbx]
  add	rbx, # 8
  fnext,
end-code

code tf!+	\ addr -- addr' ; F: f --
\ *G Stores the top of the float stack as an 80 bit float
\ ** at addr, and updates addr appropriately.
\ ** Ndpx64.fth only.
  fstp  tbyte 0 [rbx]
  add	rbx, # #10
  fnext,
end-code

also h-c definitions

: lay-asm:      \ addr -- addr' ; compile a string as an an assembler macro, putting xt at addr
  :noname !xt++                         \ generate : portion, put xt at addr, step on
  !csp					\ because we have consumed the xt
  also asm-access                       \ enable assmbler vocabulary
  "xxx" evaluate                        \ use the string
  previous
  postpone ;
; doNotSin

-3 constant SFSIZE	\ 32 bit float
-2 constant DFSIZE	\ 64 bit float
-1 constant TFSIZE	\ 80 bit float

TFSIZE constant defFPsize

: +FPsize	\ fpsize --
\ lay assembler indicator for required size
  case
    SFSIZE of  a[ float ]a   endof
    DFSIZE of  a[ double ]a   endof
    TFSIZE of  a[ tbyte ]a   endof
              size-error
  endcase
; doNotSin

: (fpmem)	{ size opc -- ; compile code for FP fetch operations }
  1 min-depth(ws)

\ deal with reg+lit
  nREGID opt?  RFROMnREGID opt? or if
    pop(ws) drop			\ discard TOS
    seq-backup				\ discard LEA
    opc execute  size +FPsize		\ FLD size n [reg]
    seqflags @ seqprev @ +n[reg]  use-a
    reset-opt  exit
  endif

\ Deal with TOS=dreg
  tos-res dreg? if			\ FLD size n [reg]
    ?rem-litplus >r                     \ remove LIT + code
    opc execute  size +FPsize  r> pop(ws) +n[reg]  use-a
    reset-opt  exit
  endif

\ Deal with TOS=reg
  tos-res reg? if			\ FLD size 0 [reg]
    opc execute  size +FPsize  pop(ws) +0[reg]  use-a
    reset-opt  exit
  endif

\ Deal with TOS=lit ; SFP026
  tos-res lit? if			\ FLD size [] lit
    opc execute  size +FPSIZE  pop(ws) []lit  use-a
    reset-opt  exit
  endif

\ Deal with the general case
  GetNewReg ?dup 0= if
    c_shuffle  size opc recurse  exit
  endif
  >r  pop(ws) r@ P0sdmove,		\ address to reg
  opc execute  size +FPsize  0 r> +n[reg]  use-a
  reset-opt
;

: FPmem		\ size -- ; -- ; xsize FPMEM <name> "instruction"
  create
    ,                                   \ size
    here  0 ,  lay-asm: drop            \ opc
  does>
    dup @  swap cell+ @ (fpmem)
;

SFSIZE    FPmem c_sf@  "FLD "
DFSIZE    FPmem c_df@  "FLD "
TFSIZE    FPmem c_tf@  "FLD "
defFPsize FPmem c_f@   "FLD "
SFSIZE    FPmem c_sf!  "FSTP "
DFSIZE    FPmem c_df!  "FSTP "
TFSIZE    FPmem c_tf!  "FSTP "
defFPsize FPmem c_f!   "FSTP "

oc: sf@   drop c_sf@  ;
oc: df@   drop c_df@  ;
oc: tf@   drop c_tf@  ;
oc: f@    drop c_f@  ;
oc: sf!   drop c_sf!  ;
oc: df!   drop c_df!  ;
oc: tf!   drop c_tf!  ;
oc: f!    drop c_f!  ;

: fpldopst	{ res size opc opcp -- }
\ Generate FLD FOPP FSTP or FOP mem FSTP
  size TFSIZE = if
    a[ FLD ]a  size +FPsize  res l>a  use-a
    opcp execute  a[ st(1), st ]a  use-a
    a[ FSTP ]a  size +FPsize  res l>a  use-a
  else
    opc execute  size +FPsize  res l>a  use-a
    a[ FSTP ]a  size +FPsize  res l>a  use-a
  endif
;

: (fpmemop)	{ size opc opcp -- ; compile code for FP fetch/store operations }
  1 min-depth(ws)

\ deal with reg+lit
  nREGID opt?  RFROMnREGID opt? or if
    pop(ws) drop			\ discard TOS
    seq-backup				\ discard LEA
    seqflags @ seqprev @ MakeIndMem	\ convert to memref
    size opc opcp fpldopst
    reset-opt  exit
  endif

\ Deal with TOS=dreg
  tos-res dreg? if			\ FLD size n [reg]
    ?rem-litplus pop(ws) MakeIndMem     \ remove LIT + code, convert to mem ref
    size opc opcp fpldopst
    reset-opt  exit
  endif

\ Deal with TOS=reg
  tos-res reg? if			\ FLD size 0 [reg]
    0 pop(ws) MakeIndMem size opc opcp fpldopst
    reset-opt  exit
  endif

\ Deal with TOS=lit ; SFP026
  tos-res lit? if			\ FLD size [] lit
    pop(ws) res-value @ MakeAbsMem size opc opcp fpldopst
    reset-opt  exit
  endif

\ Deal with the general case
  GetNewReg ?dup 0= if
    c_shuffle  size opc recurse  exit
  endif
  >r  pop(ws) r@ P0sdmove,		\ address to reg
  0 r> MakeIndMem size opc opcp fpldopst
  reset-opt
;

: FPmemop		\ size -- ; -- ; xsize FPMEMOP <name> "FOP" "FOPP"
  create
    ,                                   \ size
    here  0 , 0 ,		        \ two xts
    lay-asm: lay-asm: drop   		\ opc opcp
  does>
    dup @ swap cell+			\ -- size addr
    dup @ swap cell+			\ -- size opc addr'
    @					\ -- size opc opcp
    (fpmemop)
;

SFSIZE    FPmemop c_sf+!  "FADD "   "FADDP "
DFSIZE    FPmemop c_df+!  "FADD "   "FADDP "
TFSIZE    FPmemop c_tf+!  "bad  "   "FADDP "
defFPsize FPmemop c_f+!   "FADD "   "FADDP"
SFSIZE    FPmemop c_sf-!  "FSUBR "  "FSUBRP "
DFSIZE    FPmemop c_df-!  "FSUBR "  "FSUBRP "
TFSIZE    FPmemop c_tf-!  "bad "    "FSUBRP "
defFPsize FPmemop c_f-!   "FSUBR "  "FSUBRP "

oc: sf+!  drop c_sf+!  ;
oc: df+!  drop c_df+!  ;
oc: tf+!  drop c_tf+!  ;
oc: f+!   drop c_f+!  ;
oc: sf-!  drop c_sf-!  ;
oc: df-!  drop c_df-!  ;
oc: tf-!  drop c_tf-!  ;
oc: f-!   drop c_f-!  ;

: tos>reg	\ -- ; force TOS into register if not modifiable
  1 min-depth(ws)

  tos-res reg/lit-s32? 0= if
    GetNewReg ?dup if			\ -- oldres newres ; get new register
      pop(ws) over P0sdmove,  push(ws)  exit
    endif
  endif

  tos-res count=1? 0= if		\ if not modifiable
    GetNewReg ?dup if			\ -- oldres newres ; get new register
      pop(ws) over P0sdmove,  push(ws)
    endif
  endif
;

oc: f@+		drop  tos>reg c_dup c_f@   FPCELL c_lit  c_+  ;
oc: sf@+	drop  tos>reg c_dup c_sf@  4 c_lit  c_+  ;
oc: df@+	drop  tos>reg c_dup c_df@  8 c_lit  c_+  ;
oc: tf@+	drop  tos>reg c_dup c_tf@  #10 c_lit  c_+  ;
oc: f!+		drop  tos>reg c_dup c_f!   FPCELL c_lit  c_+  ;
oc: sf!+	drop  tos>reg c_dup c_sf!  4 c_lit  c_+  ;
oc: df!+	drop  tos>reg c_dup c_df!  8 c_lit  c_+  ;
oc: tf!+	drop  tos>reg c_dup c_tf!  #10 c_lit  c_+  ;

previous definitions


\ ************************
\ *S Dictionary operations
\ ************************

: tf,           \ F: f --
\ *G Lays an 80 bit extended float into the dictionary, reserving
\ ** 10 bytes
  here  #10 allot  tf!          \ allocate 10 bytes and stash number here
;

: df,           \ F: f --
\ *G Lays an 64 bit double float into the dictionary, reserving
\ ** 8 bytes )
  here  8 allot  df!            \ allocate 8 bytes and stash number here
;

: sf,           \ F: f --
\ *G Lays a 32 bit float into the dictionary, reserving
\ ** 4 bytes
  here  4 allot  sf!            \ allocate 4 bytes and stash number here
;

: f,            \ F: f --
\ *G lays a default float into the dictionary, reserving
\ ** *\fo{FPCELL} bytes
  here  FPCELL allot  f!  ;

: falign        \ --
\ *G Aligns the dictionary to accept a default float.
; immediate

: faligned      \ addr -- addr'
\ *G Aligns the address to accept a default float.
;

: float+        \ addr -- addr'
\ *G Increments addr by *\fo{FPCELL}, the size of a default float.
  FPCELL +  ;

: floats        \ n1 -- n2
\ *G Returns n2, the size of n1 default floats.
  FPCELL *
;

: sfalign       \ --
\ *G Aligns the dictionary to accept a 32 bit float.
; immediate

: sfaligned     \ addr -- addr'
\ *G Aligns the address to accept a 32 bit float.
; immediate

: sfloat+       \ addr -- addr'
\ *G Increments addr by the size of a 32 bit float.
  4 +
;

: sfloats       \ n1 -- n2
\ *G Returns n2, the size of n1 32 bit floats.
  4 *
;

: dfalign       \ --
\ *G Aligns the dictionary to accept a 64 bit double float.
; immediate

: dfaligned     \ addr -- addr'
\ *G Aligns the address to accept a 64 bit float.
; immediate

: dfloat+       \ addr -- addr'
\ *G Increments addr by the size of a 64 bit double float.
  8 +
;

: dfloats       \ n1 -- n2
\ *G Returns n2, the size of n1 64 bit double floats.
  8 *
;

: tfalign       \ --
\ *G Aligns the dictionary to accept an 80 bit extended float.
; immediate

: tfaligned     \ addr -- addr'
\ *G Aligns the address to accept an 80 bit extended float.
; immediate

: tfloat+       \ addr -- addr'
\ *G Increments addr by the size of an 80 bit extended float.
  #10 +
;

: tfloats       \ n1 -- n2
\ *G Returns n2, the size of n1 80 bit extended floats.
  #10 *
;


\ *****************
\ *S FP defining words
\ *****************

: fvariable     \ F: -- ; -- addr
\ *G Use in the form: *\fo{FVARIABLE <name>} to create a variable
\ ** that will hold a default floating point number.
  FPCELL buffer:
;

: farray                \ n -- ; i -- addr
\ *G Use in the form: *\fo{n FARRAY <name>} to create a variable
\ ** that will hold a default floating point number. When the
\ ** array name is executed, the index i is used to retun the
\ ** address of the i'th 0 zero-based element in the array.
\ ** For example, *\fo{5 FARRAY TEST} will set up 5 array elements
\ ** each containing 0, and then *\fo{f n TEST F!} will store f in
\ ** the nth element, and *\fo{n TEST F@} will fetch it.
  create
    0 max  floats allot&erase
  does>
    swap FPCELL * +                     \ make address
;

: fconstant     \ F: f -- ; F: -- f
\ *G Use in the form: *\fo{<float> FCONSTANT <name>} to create a constant
\ ** that will return a floating point number.
  create
    f,
  does>
    f@
;

: fvalCompChild	\ xt --
\ Given the xt of an FVALUE, compile the required action.
  >body postpone literal
  case  OperatorType @off
    0 of  postpone f@  endof		\ Default, fetch
    1 of  postpone f!  endof		\ store, ->, to
    2 of  ( noop )  endof		\ addr
\					\ inc
\					\ dec
    5 of  postpone f+!  endof		\ add
\					\ zero
    7 of  postpone f-!  endof		\ sub
    8 of  postpone drop  postpone FPCELL  endof	\ sizeof
\					\ set
      bad-method
  endcase
; doNotSin

: fvalue	\ F: f -- ; ??? -- ???
\ *G Use in the form: *\fo{<float> FVALUE <name>} to create a
\ ** floating point version of *\fo{VALUE} that will return a
\ ** floating point number by default, and that can accept the
\ ** operators *\fo{TO}, *\fo{ADDR}, *\fo{ADD}, *\fo{SUB}, and
\ ** *\fo{SIZEOF}. )
  create
    f,  ['] fvalCompChild set-compiler
  interp>
    case  OperatorType @off
      0 of  f@  endof			\ default, fetch
      1 of  f!  endof			\ store, ->, to
      2 of  ( noop )  endof		\ addr
\                         		\ inc
\					\ dec
      5 of  f+!  endof			\ add
\                        		\ zero
      7 of  f-!  endof			\ sub
      8 of  drop FPCELL  endof		\ sizeof
\					\ set
         bad-method
    endcase
;


\ **************************************
\ *S Basic functions + - * /  and others
\ **************************************

code f+         \ f1 f2 -- f1+f2
\ *G Floating point add.
  faddp st(1), st               \ do the actual addition
  fnext,
end-code
fseq: f+	faddp st(1), st  ;fseq

code f-         \ f1 f2 -- f1-f2
\ *G Floating point subtract.
  fsubp st(1), st               \ do the actual calc.
  fnext,
end-code
fseq: f-	fsubp st(1), st  ;fseq

code f*         \ f1 f2 -- f1*f2
\ *G Floating point multiply.
  fmulp st(1), st               \ do the actual calc.
  fnext,
end-code
fseq: f*	fmulp st(1), st  ;fseq

code f/         \ f1 f2 -- f1/f2
\ *G Floating point divide.
  fdivp st(1), st               \ do the actual calc.
  fnext,
end-code
fseq: f/	fdivp st(1), st  ;fseq

code fmod       \ F: f1 f2 -- f3
\ *G Floating point modulus. Returns f3 the remainder after repeatedly
\ ** subtracting f2 from f1. Often used to force arguments to lie in the
\ ** range: 0 <= arg < f2.
  fxch
  fprem                         \ take modulus, rem in ST, f2 in ST(1)
  fxch
  fstp  st                      \ discard f2
  fnext,
end-code
fseq: fmod	fxch fprem fxch  fstp st  ;fseq

code fsqrt      \ F: f -- sqrt(f)
\ *G Floating point square root.
  fsqrt
  fnext,
end-code
fseq: fsqrt	fsqrt  ;fseq

code 1/f        \ F: f -- 1/f
\ *G Floating point reciprocal.
  fld1                          \ push one on ap stack
  fdivrp st(1), st              \ do the actual division
  fnext,
end-code
fseq: 1/f	fld1  fdivrp st(1), st  ;fseq

code fabs       \ F: f -- |f|
\ *G Floating point absolute.
  fabs
  fnext,
end-code
fseq: fabs	fabs  ;fseq

code fnegate    \ F: f -- -f
\ *G Floating point negate.
  fchs
  fnext,
end-code
fseq: fnegate	fchs  ;fseq

code f2*        \ F: f -- f*2
\ *G Floating point multiply by two.
  fld1
  fadd  st, st
  fmulp st(1), st
  fnext,
end-code inline

code f2/        \ F: f -- f/2
\ *G Floating point divide by two.
  fld1
  fadd  st, st
  fxch
  fdivrp st(1), st
  fnext,
end-code inline


\ ************************
\ *S Integer to FP conversion
\ ************************

code s>f        \ n -- ; F: -- f
\ *G Converts a single integer to a float.
  mov   rax, 0 [rbp]	\ swap TOS and NOS, so that TOS is in memory
  mov   0 [rbp], rbx
  mov   rbx, rax
  fild  dword 0 [rbp]   \ load 32 bit integer to FPU
  add   rbp, # cell     \ drop n from data stack
  fnext,
end-code inline

code d>f        \ d -- ; F: -- f
\ *G Converts a double integer to a float.
\ The double number needs to be in memory in strict Intel order
  mov   rax, 0 [rbp]    \ get low portion of double int, now in EBX:EAX
  sub   rbp, # cell     \ make space
  mov   0 [rbp], rax    \ stash double
  mov   cell [rbp], rbx
  fild  qword 0 [rbp]   \ load 64 bit integer to FPU
  mov   rbx, 2 cells [rbp]    \ restore TOS
  add   rbp, # 3 cells  \ discard 64 bit integer and extras
  fnext,
end-code

\ fixed SFP021
code f>s        \ F: f -- ; -- n ; convert float to integer
\ *G Converts a float to a single integer.
\ ** Note that *\fo{F>S} truncates the number towards zero
\ ** according to the ANS specification.
\ ** See *\fo{FR>S} below.
  sub   rbp, # 2 cells        	\ reserve space on stack
  fstcw word 0 [rbp]		\ save/preserve CW
  fstcw word 2 [rbp]
  or    word 0 [rbp], # $0C00	\ force truncation
  fldcw word 0 [rbp]
  fistp qword cell [rbp]   	\ store integer from FPU
  fldcw word 2 [rbp]
  mov   rax, cell [rbp]    	\ get new TOS
  mov   cell [rbp], rbx    	\ save old TOS
  add   rbp, # cell		\ remove control word info
  mov   rbx, rax        	\ set new TOS
  fnext,
end-code

: f>d        \ F: f -- ; -- d ; convert float to double integer
\ *G Converts a float to a double integer.
\ ** Note that *\fo{F>D} truncates the number towards zero
\ ** according to the ANS specification.
\ ** See also *\fo{FR>D} below.
  f>s s>d  ;

\ SFP022...
code fr>s	\ F: f -- ; -- n ; convert float to integer
\ *G Converts a float to a single integer using the current rounding mode.
  sub   rbp, # cell        	\ reserve space on stack
  fistp qword 0 [rbp]   	\ store integer from FPU
  mov   rax, 0 [rbp]    	\ get new TOS
  mov   0 [rbp], rbx    	\ save old TOS
  mov   rbx, rax        	\ set new TOS
  fnext,
end-code

: fr>d	\ F: f -- ; -- d ; convert float to double integer
\ *G Converts a float to a double integer using the current rounding mode.
  fr>s s>d  ;
\ ...SFP022


\ *****************
\ *S FP comparisons
\ *****************

code fcmp1      \ F: f -- ; -- status
\ test the fp#, returning status result
  sub   rbp, # cell             \ save TOS
  mov   0 [rbp], rbx
  ftst                          \ test FTOS
  fstsw ax                      \ save status result
  fstp  st                      \ drop FPUstack
  and   ax, # $4100             \ mask out status bits
  movzx rbx, ax                 \ copy status to TOS
  fnext,
end-code

code fcmp2      \ F: f1 f2 -- ; -- status
\ test the top 2 floats, returning status
  sub   rbp, # cell             \ save TOS
  mov   0 [rbp], rbx
  fxch
  fcompp                        \ compare and discard
  fstsw ax                      \ save status result
  and   ax, # $4100             \ mask out status bits
  movzx rbx, ax                 \ copy status to TOS
  fnext,
end-code

: f0<           \ F: f1 -- ; -- t/f ; less than zero?
( *G Floating point *\fo{0<}. N.B. result is on the Forth integer data stack. )
  fcmp1 $0100 =
;

: f0=           \ F: f1 -- ; -- t/f ; equal zero?
( *G Floating point *\fo{0=}. N.B. result is on the Forth integer data stack. )
  fcmp1 $4000 =
;

: f0<>          \ F: f1 -- ; -- t/f ; not equal zero?
( *G Floating point *\fo{0<>}. N.B. result is on the Forth integer data stack. )
  fcmp1 $4000 <>
;

: f0>           \ F: f1 -- ; -- t/f ; greater than zero?
( *G Floating point *\fo{0>}. N.B. result is on the Forth integer data stack. )
  fcmp1 0=
;

: f<            \ F: f1 f2 -- ; -- t/f ; one less than the other?
( *G Floating point *\fo{<}. N.B. result is on the Forth integer data stack. )
  fcmp2 $0100 =
;

: f=            \ F: f1 f2 -- ; -- t/f ; equal each other?
( *G Floating point *\fo{=}. N.B. result is on the Forth integer data stack. )
  fcmp2 $4000 =
;

: f<>           \ F: f1 f2 -- ; -- t/f ; one not equal the other?
( *G Floating point *\fo{<>}. N.B. result is on the Forth integer data stack. )
  fcmp2 $4000 <>
;

: f>            \ F: f1 f2 -- ; -- t/f ; one less than the other?
( *G Floating point *\fo{>}. N.B. result is on the Forth integer data stack. )
  fcmp2 0=
;

: f<=           \ F: f1 f2 -- ; -- t/f ; one less or equal the other?
( *G Floating point *\fo{<=}. N.B. result is on the Forth integer data stack. )
  fcmp2
  dup $0100 =  swap $4000 = or
  0<>
;

: f>=           \ F: f1 f2 -- ; -- t/f ; one greater or equal the other?
( *G Floating point *\fo{>=}. N.B. result is on the Forth integer data stack. )
  fcmp2
  dup 0=  swap $4000 = or
  0<>
;

\ EDA041...
code fsignbit	\ F: f -- ; -- sign
\ *G Return the sign bit of the floating point number. This is
\ ** not the same as *\fo{f0<} for f=+/-0e0.
  sub  rbp, # cell
  mov  0 [rbp], rbx
  sub  rbx, rbx
  fxam   fstsw ax   fstp  st
  and ax, # $0200
  nz, if,  dec rbx  then,
  fnext,
end-code
\ ...EDA041

: fsign         \ F: r1 -- ; -- sign
\ *G Get the sign of floating point *\i{r1}. The sign is zero
\ ** for positive numbers and -1 for negative numbers.
  fdup f0= >r  fsignbit  r>
  if  signed-zero @ 0<> and  then
;

: f~            \ F: f1 f2 f3 -- ; -- flag
\ *G Approximation function. If f3 is positive, flag is true if abs[f1-f2]
\ ** to f1. If f3 is negative, flag is true if abs[f1-f2] less than
\ ** abs[f3*abs[f1+f2]].
  fdup f0> if                           \ if f3 > 0
    frot frot  f- fabs			\ -- f3 |f1-f2|
    f>  exit                            \ -- t/f ; SFP007
  then
  fdup f0= if				\ if f3 = 0
    fdrop  fdup fsign >r
    fover fsign >r
    f= r> r> xor 0= and  exit
  then
  frot frot  fover fover		\ -- f3 f1 f2 f1 f2
  f- fabs				\ -- f3 f1 f2 |f1-f2|
  frot frot				\ -- f3 |f1-f2| f1 f2
  fabs fswap fabs			\ -- f3 |f1-f2| |f1| |f2|
  f+ frot fabs f*			\ -- |f1-f2| |f3|*(|f1|+|f2|)
  f<					\ -- t/f
;


\ ******************************
\ *S Words dependent on FP compares
\ ******************************

: ?fnegate      \ F: f1 f2 -- f3
( *G Floating point *\fo{NEGATE}. )
  f0<
  if  fnegate  endif
;

: fmax          \ F: f1 f2 -- f3
( *G Floating point *\fo{MAX}. )
  fover fover f<
  if  fswap  endif
  fdrop
;

: fmin          \ F: f1 f2 -- f3
( *G Floating point *\fo{MIN}. )
  fover fover f>
  if  fswap  endif
  fdrop
;


\ ******************
\ *S FP logs and powers
\ ******************

code flog       \ F: f -- log(f)
( *G Floating point log base 10. )
  fldlg2
  fxch
  fyl2x
  fnext,
end-code

code fln        \ F: f -- ln(f)
( *G Floating point log base e. )
  fldln2
  fxch
  fyl2x
  fnext,
end-code

decimal
Lbl: 2**z       \ F: f -- 2^f ; SFP009
		\ This routine has to cope with range limitations
  push	rdx
  sub 	rbp, # #fpframe			\ make transient frame

\ make F absolute, and preserve previous sign
  sub	rdx, rdx			\ 0 for positive, +ve assumed
  ftst					\ test it
  fstsw	ax				\ save status, -ve if C3 C2 C0 = 001
  and 	ax, # $4100			\ extract C3,C2,C0
  cmp	ax, # $0100			\ -ve?
  z, if,
    dec   rdx			        \ mark it so
    fabs				\ make absolute
  endif,

\ round FTOS down, preserving original FTOS
  fstcw word #cw [rbp]                  \ for later use
  fstcw word #rc [rbp]
  and   word #rc [rbp], # $f3ff         \ change the current control; was f3ff
  or    word #rc [rbp], # $0400         \ word to round down
  fldcw word #rc [rbp]
  fld   st                              \ copy number, and round one
  frndint                               \ N: -- f int
  fldcw word #cw [rbp]                  \ restore original control word

\ generate fraction from F and INT
  fxch					\ N: -- int f
  fld	 st(1)				\ N: -- int f int
  fsubrp st(1), st			\ N: -- int frac
  fchs					\ *** BUG in assembler FSUB, FSUBRP

\ generate 2^FRAC, preserving INT
  fld1
  fchs
  fxch					\ N: -- int -1 frac
  fscale				\ N: -- int -1 frac/2
  f2xm1					\ N: -- int -1 2^(frac/2)-1
  fxch
  fchs
  faddp st(1), st			\ N: -- int 2^(frac/2)
  fmul	st, st				\ N: -- int 2^frac

\ check sign of original F, applying as needed
  or	rdx, rdx
  nz, if,
    fld1
    fdivrp st(1), st			\ N: -- int 2^-frac
    fxch
    fchs
    fxch				\ N: -- -int 2^-frac
  endif,				\ N: -- int' 2^frac'

\ Apply the integer power to get 2^x, and clean up
  fscale				\ N: -- int' 2^(frac'+int')
  fxch
  fstp st				\ N: -- 2^(frac'+int'

  add   rbp, # #fpframe                 \ remove transient frame
  pop	rdx

  ret
end-code

code 2**	\ F: f -- 2^f
( *G Floating point: returns 2^F. )
  call  2**z
  fnext,
end-code
fseq: 2**	call 2**z  ;fseq

code fexp       \ F: f -- e^f ; was called FE^X
( *G Floating point e^f. )
  fldl2e
  fmulp st(1), st
  call  2**z
  fnext,
end-code

: fexpm1	\ F: f -- (e^f)-1 ; 12.6.2.1516
\ *G Floating point log base (e^f)-1.
  %lg2e f* fdup %1 f< if
    [asm  f2xm1  asm]
  else
    2** %1 f-
  endif
;

code flnp1	\ F: f1 -- f2
\ *G The output *\i{f2} is the natural logarithm of the input
\ ** plus one. An ambiguous condition exists if *\i{f1} is less
\ ** than or equal to negative one.
  fld1
  faddp  st(1), st
  fldln2
  fxch
  fyl2x
  ret
end-code

code falog      \ F: f -- 10^f ; was called f10^f, new name: ans
( *G Floating point anti-log base 10. )
  fldl2t
  fmulp st(1), st
  call  2**z
  fnext,
end-code

code (f**)	\ F: f1 f2 -- f1^f2
\ *G Floating point returns f1 raised to the power f2.
\ ** No error checking is performed. If floating point
\ ** execeptions are masked, which is the default condition,
\ ** the system will return a NaN for f1<0.
  fxch
  fyl2x
  call  2**z
  fnext,
end-code
fseq: (f**)	fxch fyl2x  call 2**z  ;fseq

: f**		\ F: f1 f2 -- f1^f2
\ *G Floating point: returns f1 raised to the power f2.
\ ** If f1<=0e0, 0e0 is returned. This behaviour is required
\ ** by the Forth Scientific Library.
  fover f0> if
    (f**)  exit
  endif
  fdrop fdrop %0
;


\ ***********
\ *S Rounding
\ ***********
\ *P The default rounding configuration is round to nearest.

code [fround]   \ mode -- ; F: f -- f' ; round floating point value
  sub   rbp, # #fpframe         \ allow space for results
  fstcw #cw [rbp]               \ get control word
  mov   cx, #cw [rbp]           \ and preserve it
  and   word #cw [rbp], # $F3FF \ strip rounding control
  or    word #cw [rbp], bx      \ or in new mode
  fldcw #cw [rbp]               \ put new rounding control back
  frndint                       \ rounding operation
  mov   #cw [rbp], cx           \ restore previous control word
  fldcw #cw [rbp]
  add   rbp, # #fpframe         \ discard control word
  mov   rbx, 0 [rbp]            \ restore TOS
  add   rbp, # cell
  fnext,
end-code

: fround        \ F: f1 -- f1'
\ *G Round the number to nearest or even.
  0 [fround]  ;

: ftrunc	\ F: f1 -- f1'
\ *G Round the number towards zero, returning an integer result
\ ** on the FP stack.
  $0C00 [fround]  ;

: fint          \ F: f1 -- f1'
\ *G A synonym for *\fo{FTRUNC}. *\fo{FINT} will be removed in a
\ ** future release.
  $0C00 [fround]  ;

: floor         \ F: f1 -- f1'
\ *G Floored round towards -infinity.
  $0400 [fround]  ;

: roundup       \ F: f1 -- f1'
\ *G Round towards +infinity.
  $0800 [fround]  ;

: rounded	\ -- ; set NDP to round to nearest
\ *G Set NDP to round to nearest for all operations other than
\ ** *\fo{FINT}, *\fo{FLOOR} and *\fo{ROUNDUP}.
  cw@ $F3FF and ( $0 or ) cw!  ;
rounded

: floored	\ -- ; set NDP to floor
\ *G Set NDP to round to floor for all operations other than
\ ** *\fo{FROUND}, *\fo{FINT} and *\fo{ROUNDUP}.
  cw@ $F3FF and $400 or cw!  ;

: roundedup	\ -- ; set NDP to round up
\ *G Set NDP to round up for all operations other than
\ ** *\fo{FROUND}, *\fo{FINT} and *\fo{FLOOR}.
  cw@ $F3FF and $800 or cw!  ;

: truncated	\ -- ; set NDP to chop to 0
\ *G Set NDP to chop to 0 for all operations other than
\ ** *\fo{FROUND}, *\fo{FLOOR} and *\fo{ROUNDUP}.
  cw@ $F3FF and $C00 or cw!  ;


\ ****************
\ Compiler support
\ ****************

decimal

code flit       \ F: -- f ; takes floating point number inline
\ *G Followed in line by a floating point number (*\fo{FPCELL} bytes)
\ ** returning this number when executed.
  pop   rdx                     \ get return address
  fld   fword 0 [rdx]           \ load new inline value
  add   rdx, # FPCELL
  jmp   rdx
end-code

[+switch dasm-switch	\ fix up the disassembler
  ' flit run:  base @ decimal
  	       DisAddr dup
               cr ." ( FPlit: " f@ f. ." )"
               FPCELL +  set-DisAddr
               base !  ;
switch]

defer fliteral	\ F: f -- ; F: -- f
\ *G Compiles a float as a literal into the current definition.
\ ** At execution time, a float is returned. For example,
\ ** *\fo{[ %PI F2* ] FLITERAL} will compile 2PI as a floating point
\ ** literal. Note that *\fo{FLITERAL} is immediate, whereas *\fo{(RLITERAL)}
\ ** below is not.
  immediate

: (rliteral)	\ F: f -- ; F: -- f
\ *G Compiles a float as a literal into the current definition.
\ ** At execution time, a float is returned. This is the default
\ ** action of *\fo{FLITERAL} above.
  fdup %0 f=
  if  fdrop  postpone %0  exit  endif
  fdup %1 f=
  if  fdrop  postpone %1  exit  endif
\  postpone flit  f,  discard-sinline	\ SFP033
  [ also h-c ]
  GenTokens? if				\ SFP034...
    fdup genFlitToken  0 -> GenTokens?
    postpone flit  f,  1 -> GenTokens?
  else
    postpone flit  f,
  endif					\ ...SFP034
  [ previous ]
;
assign (rliteral) to-do fliteral

: #fliteral              \ n -- ; F: real --
  drop  postpone fliteral
;


\ ******************
\ *S FP trigonometry
\ ******************

hex

code ftan       \ F: f -- tan(f)
( *G Floating point tangent. )
  fptan
  fstp  st
  fnext,
end-code

code fatan      \ F: f -- atan(f)
( *G Floating point arctangent. )
  fld1
  fpatan
  fnext,
end-code

code fsin       \ F: f -- sin(f)
( *G Floating point sine. )
  fsin
  fnext,
end-code

\ SFP030...
code fasin      \ F: f -- asin(f)
\ *G Floating point arcsine.
  fld    st
  fld    st
  fmulp  st(1), st
  fld1
  fsubrp st(1), st
  fsqrt
  fpatan
  fnext,
end-code
\ ...SFP030

code fcos       \ F: f -- cos(f)
( *G Floating point cosine. )
  fcos
  fnext,
end-code

\ SFP030...
code facos      \ F: f -- acos(f)
\ *G Floating point arctangent.
  fld    st
  fld    st
  fmulp  st(1), st
  fld1
  fsubrp st(1), st
  fsqrt
  fxch   st(1)
  fpatan
  fnext,
end-code
\ ...SFP030

code fsincos    \ F: f -- sin(f) cos(f)
\ *G Returns sine and cosine values of *\i{f}.
  fsincos
  fnext,
end-code inline

code fatan2	\ F: f1 f2 -- atan(f1/f2)
 \ *G Floating point arctangent with prior division.
  fpatan
  fnext,
end-code

((
\ SFP037...
: (FATAN2a)                  \ F: y x -- phi
  \ *G *\fo{FATAN2} factor when x=0. Formula:
  \ *C    pi/2*sign(y)
    fdrop fsign 1 and s>f %pi/2 f*
;

: (FATAN2b)                  \ F: y x -- phi
  \ *G *\fo{FATAN2} factor when x<0. Formula:
  \ *C    tan(y/x)+PI*sign(y)
   fover fswap f/ fatan fswap fsign 1 and s>f %pi f* f+
;

: FATAN2                  \ F: y x -- phi
  \ *G Floating point arc tangent of (y/x) with quadrant disambiguation.
  \ ** Range {-PI ... PI}.
   fdup f0= if  (fatan2a) exit  then	\ extreme case for x = 0
   fdup f0< if  (fatan2b) exit  then	\ 2nd and 3rd quadrants
   f/ fatan				\ 1st and 4th quadrants
;
\ ...SFP037
))

: deg>rad	\ F: fdeg -- frad
( *G Converts a value in degrees to radians. )
  %pi f*  #180 s>f f/
;

: rad>deg               \ -- ;
( *G Converts a value in radians to degrees. )
  #180 s>f f*  %pi f/
;

code freduce    \ F: f1 -- f2 ; reduce value to range 0..2pi
( *G Reduce f1 to be in the range 0 <= f2 < 2PI. )
  begin,
    fldpi                       \ constant of 2pi
    fadd   st, st
    fxch
    fprem                       \ reduce
    fstsw  ax                   \ get status
    fxch			\ remove 2pi
    fstp   st
    and    ax, # $0400          \ inspect C2 bit
  z, until,                     \ until C2 clear
  fnext,
end-code


\ ******************************************************
\ Extra trig functions - COSEC SEC COTAN and hyberbolics
\ ******************************************************

: fcosec        \ F: f -- cosec(f)
( *G Floating point cosecant. )
  fsin 1/f  ;

: fsec          \ F: f -- sec(f)
( *G Floating point secant. )
  fcos 1/f  ;

: fcotan        \ f: f -- cot(f)
( *G Floating point cotangent. )
  ftan 1/f  ;

: fsinh         \ F: f -- sinh(f) ; (e^x - 1/e^x)/2
( *G Floating point hyberbolic sine. )
  fexp  fdup 1/f f-  f2/  ;

: fcosh         \ F: f -- cosh(f) ; (e^x + 1/e^x)/2
( *G Floating point hyberbolic cosine. )
  fexp  fdup 1/f f+  f2/  ;

: ftanh         \ F: f -- tanh(f) ; (e^x - 1/e^x)/(e^x + 1/e^x)
( *G Floating point hyberbolic tangent. )
  fexp fdup 1/f  fover fover f-		\ GEK044
  frot frot f+  f/
;

: fasinh        \ F: f -- asinh(f) ; ln(f+sqrt(1+f*f))
( *G Floating point hyberbolic arcsine. )
  fdup
  fdup f*  %1 f+  fsqrt
  f+  fln  ;

: facosh        \ F: f -- acosh(f) ; ln(f+sqrt(f*f-1))
( *G Floating point hyberbolic arccosine. )
  fdup
  fdup f*  %1 f-  fsqrt
  f+  fln  ;

: fatanh        \ F: f -- atanh(f) ; ln((1+f)/(1-f))/2
( *G Floating point hyberbolic arctangent. )
  %1 fover f+  %1 frot f-  f/
  fln  f2/  ;


\ ********************
\ *S Number conversion
\ ********************

((
Converting from ASCII to floating point is not so hard as before
since there are some Forth words for some of the jobs.  It still
consists of four basic steps: i) obtain the significand and its
decimal maginitude, ii) get the exponent, iii) convert ascii to
integer, and then iv) correct the power value.

The input format takes the following form:

[-] [digit(s)] [.] [digit(s)] [E|e] [+|-] [digit(s)]
))

decimal

create powers10	\ -- addr ; table of powers of 10
                    1: ,		\ in Intel order (low, high)
		   10: ,		\ 10^1
		  100: ,		\ 10^2
		 1000: ,		\ 10^3
		10000: ,		\ 10^4
	       100000: ,		\ 10^5
	      1:000000 ,		\ 10^6
	     10:000000 ,		\ 10^7
	    100:000000 ,		\ 10^8
	   1000:000000 ,		\ 10^9
	  10000:000000 ,		\ 10^10
	 100000:000000 ,		\ 10^11
	1:000000000000 ,		\ 10^12
       10:000000000000 ,		\ 10^13
      100:000000000000 ,		\ 10^14
     1000:000000000000 ,		\ 10^15
    10000:000000000000 ,		\ 10^16
   100000:000000000000 ,		\ 10^17
  1000000:000000000000 ,		\ 10^18

code get10^n	\ n -- ; -- f ; 0<=n<=18
  mov	rax, # powers10
  fild  qword 0 [rax] [rbx*8]	\ load integer from table
  mov	rbx, 0 [rbp]		\ restore TOS
  lea	rbp, cell [rbp]
  fnext,
end-code

: (10**n)	\ +n -- ; -- f
  dup #18 > if
    #10 s>f  s>f f**  exit
  endif
  get10^n
;

: 10**n		\ n -- ; -- f
\ *G Generate a floating point value 10 to the power n, where
\ ** n is an integer.
  dup abs (10**n)  0<
  if  1/f  endif
;

: exp-char?	\ char -- flag ; true if character is exponent indicator
  case  upc
    [char] D of  true  endof
    [char] E of  true  endof
                 drop false
  end-case
;

[-short-branches
: (>FLOAT)	\ c-addr u -- flag ; F: -- f | --
\ *G Try to convert the string at c-addr/u to a floating point number.
\ ** If conversion is successful, flag is returned true, and a floating
\ ** number is returned on the float stack, otherwise flag is returned
\ ** false and the float stack is unchanged.
  { c-addr u | dlo dhi u' fsgn flag nshift obase -- flag }

  u 0= IF false exit THEN               \ must have at least one significand digit

  base @ -> obase  decimal		\ preserve and force base ; SFP028

  false -> flag
  0 -> nshift

  c-addr c@ [char] + = if               \ Ignore preceeding +
    c-addr 1 chars + -> c-addr
    u 1 chars - -> u
  then

  c-addr c@ [char] - =                  \ check for preceeding -
  dup -> fsgn  IF
    c-addr 1+ -> c-addr
    u 1- -> u
  THEN

  \ convert mantissa (first set of digits)
  0 0 c-addr u >number -> u' -> c-addr -> dhi -> dlo

  u' 0> IF
    \ convert optional number after decimal point
    c-addr c@ fp-char isSep? IF		\ SFP035
      dlo dhi c-addr 1+ u' 1- dup -> nshift >number
      dup nshift - -> nshift
      -> u' -> c-addr -> dhi -> dlo
    THEN
    \ convert exponent
    u' 0> IF
      c-addr c@ exp-char? IF
        u' 1 > if
          c-addr 1+ u' 1- (integer?)
        else
          0 1
        then
        IF
          nshift + -> nshift
          true -> flag
        THEN
      THEN
    ELSE
      true -> flag
    THEN
  ELSE
    u u' <> -> flag
  THEN

  \ convert double precision int to float
  flag IF
    cw@ >r  rounded
    dlo dhi d>f
\    10 s>f nshift s>f f** f*   \ apply exponent
    nshift 10**n f*
    fsgn IF  fnegate  THEN
    r> cw!
  THEN

  flag  obase base !			\ SFP028
;
short-branches]

: >FLOAT	\ c-addr u -- flag ; F: -- f | --
\ *G Try to convert the string at c-addr/u to a floating point number.
\ ** If conversion is successful, flag is returned true, and a floating
\ ** number is returned on the float stack, otherwise flag is returned
\ ** false and the float stack is unchanged.
\ ** Leading and trailing white space are removed before processing.
\ ** If the resulting string is of zero length, true is returned
\ ** with a floating point zero. Yes, this is what the standard
\ ** requires. The previous behaviour without this special case
\ ** is available as *\fo{(>FLOAT)} above.
  -white dup if
    (>float)
  else
    2drop true  %0
  endif
;

\ ************
\ *S FP output
\ ************
\ *P A significant portion of the output code is taken from
\ ** *\i{FPOUT v3.7} by Ed. See
\ *C   http://dxforth.webhop.org/
\ *P or one of its mirrors.

decimal

2variable fdp		\ -- addr ; trailing decimal point, # digits
  1 FDP !
variable fechar		\ -- char ; exponent output character
  char E fechar c!
variable fedigits	\ -- addr ; min exponent digits
  2 fedigits !
variable #places	\ -- addr ; # significant digits
  8 #places !

#fdigits constant max-precision
#fdigits constant mc#

: precision     \ -- u
\ *G Returns the number of significant digits used by *\fo{F.}
\ ** *\fo{FE.} and *\fo{FS.}.
  #places @  ;

: set-precision \ u --
\ *G Sets the number of significant digits used by *\fo{F.}
\ ** *\fo{FE.} and *\fo{FS.}.
  0 max  #fdigits min  #places !  ;

8 set-precision

: places        \ u --
\ *G Sets the number of significant digits used by *\fo{F.}
\ ** *\fo{FE.} and *\fo{FS.}.
\ ** The ANS version of this word is *\fo{SET-PRECISION}, which
\ ** should be used in new code.
  set-precision  ;

\ The next three words assume that the value to be tested has
\ been placed in a 10-byte real whose address is passed to them.

\ SFP012...
: fsign?	\ addr -- flag ; true if negative
  8 + w@ $8000 and 0<>  ;

: infinity?	\ addr -- flag ; true if infinite
  dup 8 + w@ $7FFF and $7FFF =
  swap 2@ $80000000 $00000000 d= and
;

: NaN?		\ addr -- flag ; true if not a number
  dup 8 + w@ $7FFF and $7FFF =
  swap 2@ $80000000 $00000000 d= 0= and
;
\ ...SFP012

: BadFloat?	\ F: f -- ; -- caddr u true | false
\ *G If the float is a NaN or Infinite, return a string such as
\ ** "+NaN" and true, otherwise just return false (0).
  { | checkit[ #10 ] -- }
  checkit[ tf!				\ save for test
  checkit[ NaN? if
    checkit[ fsign?
    if  s" -NaN"  else  s" +NaN"  endif
    true  exit
  endif
  checkit[ infinity? if
    checkit[ fsign?
    if  s" -Inf"  else  s" +Inf"  endif
    true  exit
  endif
  false
;

: >exp		\ -- n ; F: f -- f ; EDA025 SFP029
\ Return decimal exponent as integer n
  { | temp[ 8 ] -- }
  fdup f0= if
    0
  else
    fdup fabs flog
    temp[ df!  temp[ df@		\ prevent rounding errors after FLOG
    floor f>s
  endif
;

\ EDA027...
: represent	\ c-addr len -- n flag1 flag2 ; F: f --
\ *G Assume that the floating number is of the form +/-0.xxxxEyy.
\ ** Round the significand xxxxx to *\i{len} significant digits
\ ** and place its representation at *\i{c-addr}. If *\i{len} is
\ ** zero round the fractional significand to a whole number.
\ ** If *\i{len} is negative the fractional significand is rounded to zero.
\ ** *\i{Flag2} is true if the results are valid. *\i{N} is the
\ ** signed integer version of yy and *\i{flag1} is true if
\ ** *\i{f} is negative. In this implementation all errors are
\ ** handled by exceptions, and so *\i{flag2} is always true
\ ** except for NaNs and Infinites.
\ ** The number of characters placed at *\i{c-addr} is the
\ ** greater of *\i{len} or *\fo{MAX-FLOAT-DIGITS}. For a Nan or
\ ** Infinite, a three character non-numeric string is returned.
\ SFP012...
  { caddr u | sign exp -- n flag1 flag2 }
  caddr u #fdigits max blank		\ set to spaces for text messages
  fdup fsignbit to sign			\ EDA042
  fdup BadFloat? if
   over c@ [char] - = >r		\ sign of result
   1 /string				\ strip sign
   caddr swap cmove			\ set o/p to "NaN" or "Inf"
   3 r> 0				\ -- 3 sign 0
   fdrop exit
  endif
\ ...SFP012
  caddr u
  2dup #fdigits max [char] 0 fill	\ prefill with '0' chars
\  dup 0< if				\ del EDA041
\    2drop  1 f0< true  exit
\  endif
  #fdigits min  2>r
  >exp 1+ to exp
  r@ exp - 10**n  f*
  fround				\ EDA042
  fabs f>d  2dup <# #s #>
  dup r@ - addr exp +!
  2r>  rot min  1 max  cmove
  d0= if				\ EDA041
    1 sign  signed-zero @ 0<> and
  else
    exp  sign
  endif
\  d0= if  1 false  else  exp sign  endif
  true					\ exceptions handled by CPU and CATCH/THROW
;
\ ...EDA027

also system definitions

mc# +user fbuf	\ -- addr ; buffer for output from REPRESENT
cell +user uex#		\ exponent
cell +user usn#		\ sign
cell +user uef#		\ exponent factor  1=FS. 3=FE.
cell +user upl#		\ +n  places right of decimal point
			\ -1  compact display

\ get exponent
: (f1) ( F: r -- r ) ( -- exp )
  >exp 1+  ;
((
\ get exponent
: (f1) ( F: r -- r ) ( -- exp )
  FDUP [UNDEFINED] FLOG [IF]
    fbuf MAX-PRECISION REPRESENT NIP AND
  [ELSE]
    FABS FLOG FLOOR F>D D>S 1+
  [THEN] ;
))

\ apply exponent factor
: (f2) ( exp -- offset exp2 )
  S>D  uef# @ FM/MOD  uef# @ *  ;

\ float to character string
: (f3) ( F: r -- ) ( places -- c-addr u flag )
  DUP upl# !  0< IF
    PRECISION
  ELSE
    (f1)  uef# @ 0> IF  1- (f2) DROP 1+  THEN
    upl# @ +				\ EDA040
  THEN
  MAX-PRECISION MIN			\ EDA040
  fbuf SWAP REPRESENT >R
  usn# !  uex# !  fbuf mc# -TRAILING  R> <# ;

\ insert exponent
: (f4) ( exp -- )
  DUP  ABS S>D  upl# @ 0< 0=  DUP >R
  IF  fedigits @ 1 MAX 1 ?DO # LOOP  THEN	\ EDA040
  #S 2DROP  DUP SIGN  0< 0=  R> AND
  IF  [CHAR] + HOLD  THEN
  FECHAR C@ HOLD ;

\ insert digit and update flag
: (f5) ( char -- )
  HOLD  1 FDP CELL+ ! ;

\ insert string
: (f6) ( c-addr u -- )
  0 MAX  BEGIN  DUP  WHILE  1- 2DUP CHARS + C@ (f5)
  REPEAT 2DROP ;

\ insert '0's
: (f7) ( n -- )
  0 MAX 0 ?DO [CHAR] 0 (f5) LOOP ;

\ insert sign
: (f8) ( -- )
  usn# @ SIGN  0 0 #> ;

\ trim trailing '0's
: (f9) ( c-addr u1 -- c-addr u2 )
  upl# @ 0< IF
    BEGIN  DUP WHILE  1- 2DUP CHARS +
    C@ [CHAR] 0 -  UNTIL  1+  THEN
  THEN ;

: (fa) ( n -- n n|pl# )
  upl# @ 0< IF  DUP  ELSE  upl# @  THEN ;

\ insert fraction string n places right of dec. point
: (fb) ( c-addr u n -- )
  0 FDP CELL+ !
  >R (f9)  R@ +
  (fa) OVER - (f7)     \ trailing 0's
  (fa) MIN  R@ - (f6)  \ fraction
  R> (fa) MIN (f7)     \ leading 0's
  FDP 2@ OR IF
    [CHAR] . HOLD
  THEN ;

\ split string into integer/fraction parts at n and insert
: (fc) ( c-addr u n -- )
  >R  2DUP R@ MIN 2SWAP R> /STRING  0 (fb) (f6) ;

\ exponent form
[-short-branches
: (fd) ( F: r -- ) ( n factor -- c-addr u )
  uef# !  (f3) IF  uex# @ 1- (f2) (f4) 1+ (fc) (f8)  THEN ;
short-branches]

\ display c-addr u right-justified in field width u2
: (fe) ( c-addr u u2 -- )
  OVER - SPACES TYPE ;

previous definitions

also system

: (FS.)		\ F: f -- ; n -- c-addr u
\ *G Convert float *\i{f} to a string *\i{c-addr/u} in scientific
\ ** notation with *\i{n} places right of the decimal point.
  >r  fdup BadFloat? if
    fdrop  r> drop  exit
  endif
  r> 1 (fd) ;

: FS.R		\ F: r -- ; n u --
\ *G Display float *\i{f} in scientific notation right-justified
\ ** in a field width *\i{u} with *\i{n} places right of the
\ ** decimal point.
  >r >r  fdup BadFloat? if
    fdrop  r> r> 2drop  exit
  endif
  r> (FS.) R> (fe)
;

: FS.		\ F: f --
\ *G Display float *\i{f} in scientific notation, with one
\ ** digit before the decimal point and a trailing space.
  -1 ( precision ) 0 FS.R SPACE ;

: (FE.)		\ F: r -- ; n -- c-addr u
\ *G Convert float *\i{f} to a string c-addr u in engineering
\ ** notation with n places right of the decimal point.
  >r  fdup BadFloat? if
    fdrop  r> drop  exit
  endif
  r> 3 (fd)
;

: FE.R		\ F: r -- ; n u --
\ *G Display float *\i{f} in engineering notation right-justified
\ ** in a field width *\i{u} with *\i{n} places right of the
\ ** decimal point.
  >R (FE.) R> (fe) ;

: FE.		\ F: f --
\ *G Display float *\i{f} in engineering notation, in which the
\ ** exponent is always a power of three, and the significand is
\ ** always in the range 1.xxx to 999.xxx.
  -1 ( precision ) 0 FE.R SPACE ;

[-short-branches
: (F.)		\ F: f -- ; n -- c-addr u
\ *G Convert float *\i{f} to string *\i{c-addr/u} in fixed-point
\ ** notation with *\i{n} places right of the decimal point.
  >r  fdup BadFloat? if
    fdrop  r> drop  exit
  endif
  r>  0 uef# ! (f3) IF
    uex# @  DUP mc# > IF
      fbuf 0 ( dummy ) 0 (fb)
      mc# - (f7) (f6)
    ELSE
      DUP 0> IF
        (fc)
      ELSE
        ABS (fb) 1 (f7)
      THEN
    THEN (f8)
  THEN ;
short-branches]

: F.R		\ F: f -- ; n u --
\ *G Display float *\i{f} in fixed-point notation right-justified
\ ** in a field width *\i{u} with *\i{n} places right of the
\ ** decimal point.
  >R (F.) R> (fe) ;

: F.		\ F: f --
\ *G Display f as a float in fixed point notation with a trailing
\ ** space. The ANS specification says that the display is in
\ ** fixed-point format, but restricted by *\fo{PRECISION}. What
\ ** should 1e308 display? In this implementation 1e308 displays
\ ** a 1 followed by 308 zeros. Several people believe that the
\ ** specification for *\fo{F.} is broken. For a display word
\ ** that always provides sensible output, use *\fo{G.} below.
  -1 ( precision ) 0 F.R SPACE ;

: (G.)		\ F: f -- ; n -- c-addr u
\ ** Convert float *\i{f} to string *\i{c-addr/u} with *\i{n}
\ ** places right of the decimal point. Fixed-point is used if
\ ** the exponent is in the range -4 to 5 otherwise scientific
\ ** notation is used.
  >r  fdup BadFloat? if
    fdrop  r> drop  exit
  endif
  (f1)  -4 1+  5 2 +  WITHIN  R> SWAP
  IF  (F.)  ELSE  (FS.)  THEN
 ;

: G.R		\ F: f -- ; n u --
\ *G Display float *\i{f} right-justified in a field width *\i{u}
\ ** with *\i{n} places right of the decimal point. Fixed-point
\ ** is used if the exponent is in the range -4 to 5 otherwise
\ ** scientific notation is used.
  >R (G.) R> (fe) ;

: G.		\ F: f --
\ *G Display float *\i{f} followed by a space. Floating-point
\ ** is used if the exponent is in the range -4 to 5 otherwise
\ ** use scientific notation. Non-essential zeros and signs are
\ ** removed.
  -1 ( precision ) 0 G.R SPACE ;

previous

: f?		\ addr -- ; displays contents of addr
( *G Displays the contents of the given *\fo{FVARIABLE}. )
  f@ f. ;

[-short-branches
: f.s           \ F: i*f -- i*f
\ *G Display the contents of the floating point stack in a vertical
\ ** format.
  fdepth 0< -400 ?throw
  fdepth if
    cr ." -- FP top of "
    fdepth dup .  #16 min 0	\ limit to 16 items
    ?do  cr i fpick f.  loop
  else
    cr ." Empty FP stack"
  endif
  cr
;

: f.sh		\ F: i*f -- i*f
\ *G Display the contents of the floating point stack in a
\ ** horizontal format.
  fdepth 0< -400 ?throw
  fdepth if
    cr ." F: -- " 0 fdepth 1-
    do  i fpick f.  -1 +loop
  else
    cr ." Empty FP stack"
  endif
;
short-branches]


\ *********************
\ Environment Variables
\ *********************

also environment definitions

true constant FLOATING
true constant FLOATING-EXT
#8   constant FLOATING-STACK

pad 3 cells $FF fill

$00007FFE  pad 2 cells + !

pad f@ fconstant MAX-FLOAT

#fdigits constant MAX-FLOAT-DIGITS	( not ANS )

previous definitions


\ ***************************
\ *S Patch FP into the system
\ ***************************

defer is-old-number?
  action-of isNumber? to-do is-old-number?

: isFnumber?  	\ caddr len -- 0 | n 1 | d 2 | -2 ; F: -- r
\ *G Behaves like the integer version of *\fo{isNumber?} except
\ ** that if  integer conversion fails, and *\fo{BASE} is decimal,
\ ** a floating point conversion is attempted. If conversion is
\ ** successful, the floating point number is left on the float
\ ** stack and the result code is -2.
  2dup 2>r  is-old-number? dup 0= if	\ test for integer first
    base @ #10 = if			\ not integer, floats must be in DECIMAL
      drop  2r@ (>float)		\ >FLOAT returns well formed flag
      -2 and				\ good float returns -2,
    then
  then
  2r> 2drop
; doNotSin

: Fnumber?  	\ caddr -- 0 | n 1 | d 2 | -2 ; F: -- r
\ *G As *\fo{isFnumber?} above, but takes a counted string.
  count isFnumber?
;

: post-float	\ f: f -- ; --
\ *G POSTPONE a floating point number. The word being defined will
\ ** itself compile the given floating point number.
  (rliteral)  ['] (rliteral) compile,
;

' noop  ' (rliteral)  ' post-float  RecType: r:NDPfloat	\ -- struct
\ *G Contains the interpret, compile and postpone actions for
\ ** floating point literals.

: rec-NDPfloat	\ caddr u -- r:float | r:fail ; F: -- [f]
\ *G The parser part of the floating point recogniser.
  (>float) if  r:NDPfloat  else  r:fail  then
;

: .NSW		\ sw -- ; display NDP status word
  ." NDP SW = " .word space
;

: .FSysPrompt	\ --
\ *G Adds floating point stack depth display.
  .SysPrompt
  fdepth if
    ."  F:-" fdepth .
  endif
  sw@ dup $0040 and if
    cr ." NDP Stack Fault: " dup .NSW  fclex
  endif
  dup $001D and if			\ ignore precision and denormals
    cr ." NDP Potential Exception: " dup .NSW  fclex
  endif
  drop
;

\ assign #fliteral to-do c#fliteral
\ assign .FSysPrompt to-do .prompt

: reals		\ -- ; turn FP system on
\ *G Enables the floating point package for number conversion.
  assign #fliteral to-do c#fliteral
  assign .FSysPrompt to-do .prompt
  ['] rec-NDPfloat forth-recognizer +stack-bot
;
reals

: integers	\ -- ; turn FP system off
\ *G Disables the floating point package for number conversion.
  assign (fliteral) to-do c#fliteral
  assign .SysPrompt to-do .prompt
  ['] rec-NDPfloat forth-recognizer -stack
;


\ ***********************
\ *S PFW2.x compatibility
\ ***********************

: (f#)		\ addr -- fn ior ; counted string to FP number
  fnumber?
  case
    0  of  0  endof
    1  of  s>f FpSystem  endof
    2  of  d>f FpSystem  endof
  end-case
;

: f#		\ -- f ; or compiles it [ state smart ]
\ *G Used in the form "*\fo{F# <number>}", the <number> string is
\ ** converted and promoted if required to a floating point number.
\ ** If the system is compiling the float is compiled. If <number>
\ ** cannot be converted an error occurs.							)
  bl word (f#) ?undef
  state @
  if  [compile] fliteral  endif
; immediate


\ ********************
\ *S Debugging support
\ ********************
\ *P Debugging floating point code is often difficult, as failures
\ ** can occur because of the necessary approximations involved
\ ** in floating point operations.

\ *P If you set the constant *\fo{[FPDEBUG]} true when compiling
\ ** *\i{Ndp387.fth}, a debug build will be constructed. The
\ ** state of the FPU will be checked after each word. If a
\ ** floating point exception has been generated, a diagnostic
\ ** is issued, and the system aborts. Set this only when testing,
\ ** as it slows down the normal operation of floating point words.

\ *P The debugger works by intercepting the end of each code
\ ** definition which is finished by *\fo{FNEXT,} rather than the
\ ** normal *\fo{NEXT,} or *\fo{RET.}
\ ** See the source code in *\i{Lib/x64/Ndpx64.fth for more details.

((
NDP status word:
bit	Function
================
15	Busy
14	C3
13..11	TOP
10	C2
9	C1
8	C0
7	Set on Unmasked Exception
6	Set on stack fault (bit 9 set for overflow, clear for underflow)
5	Set on Precision Exception
4	Set on Underflow
3	Set on Overflow
2	Set on Zero Divide
1	Set on Denormalised Operand
0	Set on Invalid Operation
))

[fpdebug] [if]
: .NDP		\ -- ; what to do if exception found
  sw@ dup .NSW
\  dup $20 and if ." Precision " endif
  dup $10 and if ." Underflow " endif
  dup $08 and if ." Overflow " endif
  dup $04 and if ." ZeroDivide " endif
\  dup $02 and if ." Denormalised " endif
  dup $01 and if ." Invalid " endif
  drop
  fclex				\ clear exceptions
  hex .rs  decimal		\ Return stack trace
  finit abort			\ reset stack and restart
;

: (fpcheck)	\ -- ; action of debugger
  sw@ $0040 and if
\    cr  r@ .dword space  r@ ip>nfa .name space
    cr ." NDP Stack Fault: " .NDP
  endif
  sw@ $001D and if	\ ignore precision and denormals
\    cr  r@ .dword space  r@ ip>nfa .name space
    cr ." NDP Exception: " .NDP
  endif
;

: +fpcheck	 \ -- ; enable FP checking
\ *G Enables the floating point debugger if it has been compiled.
  assign (fpcheck) to-do fpcheck
;

: -fpcheck	 \ -- ; disable FP checking
\ *G Disbles the floating point debugger if it has been compiled.
  assign noop to-do fpcheck
;
+fpcheck
[then]


\ *************
\ *S Extensions
\ *************

FPext? [if]

\ ======================
\ *N F.P. stack jugglers
\ ======================
\ *P Due to the Mac's usage of fp for all graphic related things, F.P.
\ ** stack jugglers similar to those for the data stack are handy. We
\ ** deal with F.P. pairs as used for points, sizes and ranges and F.P.
\ ** quads for rectangles and colours.
\ ** *\fo{F2SWAP F2OVER F2DROP F4DUP FTUCK FNIP} do what you expect ...

code F2DUP	\ F: r1 r2 -- r1 r2 r1 r2
\ *G Floating point equivalent of *\fo{2DUP}.
 fld	st(1)				\ r1 r2 -- r1 r2 r1
 fld	st(1)				\ r1 r2 r1 -- r1 r2 r1 r2
 fnext,
end-code

code F2SWAP	\ F: r1 r2 r3 r4 -- r3 r4 r1 r2
\ *G Floating point equivalent of *\fo{2SWAP}.
 fxch	st(2)				\ r1 r2 r3 r4 -- r1 r4 r3 r2
 fxch	st(1)				\ r1 r4 r3 r2 -- r1 r4 r2 r3
 fxch	st(3)				\ r1 r4 r2 r3 -- r3 r4 r2 r1
 fxch	st(1)				\ r3 r4 r2 r1 -- r3 r4 r1 r2
 fnext,
end-code

code F2OVER	\ F: r1 r2 r3 r4 -- r1 r2 r3 r4 r1 r2
\ *G Floating point equivalent of *\fo{2OVER}.
 fld	st(3)				\ r1 r2 r3 r4 -- r1 r2 r3 r4 r1
 fld	st(3)				\ r1 r2 r3 r4 r1 -- r1 r2 r3 r4 r1 r2
 fnext,
end-code

code F2DROP	\ F: r1 r2 --
\ *G Floating point equivalent of *\fo{2DROP}.
 fstp	st(0)				\ r1 r2 -- r1
 fstp	st(0)				\ r1 --
 fnext,
end-code

code F4DUP	\ F: r1 r2 r3 r4 -- r1 r2 r3 r4 r1 r2 r3 r4
\ *G Floating point equivalent of *\fo{4DUP}.
 fld	st(3)			\ r1 r2 r3 r4 -- r1 r2 r3 r4 r1
 fld	st(3)			\ r1 r2 r3 r4 r1 -- r1 r2 r3 r4 r1 r2
 fld	st(3)			\ r1 r2 r3 r4 r1 r2 -- r1 r2 r3 r4 r1 r2 r3
 fld	st(3)			\ r1 r2 r3 r4 r1 r2 r3 -- r1 r2 r3 r4 r1 r2 r3 r4
 fnext,
end-code

code FTUCK	\ F: r1 r2 -- r2 r1 r2
\ *G Floating point equivalent of *\fo{TUCK}.
 fxch	st(1)				\ r1 r2 -- r2 r1
 fld	st(1)				\ r2 r1 -- r2 r1 r2
 fnext,
end-code

code FNIP	\ F: r1 r2 -- r2
\ *G Floating point equivalent of *\fo{NIP}.
 fxch	st(1)				\ r1 r2 -- r2 r1
 fstp	st(0)				\ r2 r1 -- r2
 fnext,
end-code

[then]


\ ===========
\ Linux tests
\ ===========

((
[defined] Target_386_Linux [if]

Library: libm.so.6

Extern: double cos( double );

[then]
))

\ ======
\ *> ###
\ ======

finit reals
decimal


((
\ test code

decimal

finit rounded

#256 buffer: foo

: t		\ -- ; f: f --
  foo >bcd .
  foo #80 dump
;

: tr		\ --
  cr  #1000 0 do
    cr
    i 5 bounds do
      i s>f 0.1e0 f+ fe.  out @ #15 + #-16 and >pos
    loop
  5 +loop
  cr
;

finit rounded
))

0 [IF]

CR .( Loading demo words... ) CR
CR .( TEST1  formatted, n decimal places )
CR .( TEST2  compact & right-justified )
CR .( TEST3  display FS. )
CR .( TEST4  display F. )
CR .( TEST5  display G. )
CR .( TEST6  display 8087 non-numbers ) CR
CR .( 'n PLACES' sets decimal places for TEST1. )
CR .( SET-PRECISION sets maximum significant )
CR .( digits displayable. )
CR CR

[UNDEFINED] F, [IF]
: F, ( r -- )  FALIGN HERE 1 FLOATS ALLOT F! ;
[THEN]

CREATE f-array  \ floating-point numbers array

FALIGN HERE
1.23456E-16  F,
1.23456E-11  F,
1.23456E-7   F,
1.23456E-6   F,
1.23456E-5   F,
1.23456E-4   F,
1.23456E-3   F,
1.23456E-2   F,
1.23456E-1   F,
0.E0         F,
1.23456E+0   F,
1.23456E+1   F,
1.23456E+2   F,
1.23456E+3   F,
1.23456E+4   F,
1.23456E+5   F,
1.23456E+6   F,
1.23456E+7   F,
1.23456E+11  F,
1.23456E+16  F,
HERE SWAP -  1 FLOATS /  CONSTANT #numbers

: do-it ( xt -- )
  #numbers 0 DO
    f-array FALIGNED I FLOATS +
    OVER >R  F@  CR  R> EXECUTE
  LOOP DROP ;

2VARIABLE (dw)
: d.w ( -- dec.places width )  (dw) 2@ ;
: PLACES ( places -- ) d.w SWAP DROP (dw) 2! ;
: WIDTH  ( width -- )  d.w DROP SWAP (dw) 2! ;

5 PLACES  19 WIDTH

: (t1) ( r -- )
  FDUP d.w FS.R  FDUP d.w F.R  FDUP d.w G.R  d.w FE.R ;

: TEST1 ( -- )
  CR ." TEST1   right-justified, formatted ("
  d.w DROP 0 .R ."  decimal places)" CR
  ['] (t1) do-it  CR ;

: (t2) ( r -- )
  FDUP -1 d.w NIP FS.R  FDUP -1 d.w NIP F.R
  FDUP -1 d.w NIP G.R        -1 d.w NIP FE.R ;

: TEST2 ( -- )
  CR ." TEST2   right-justified, compact" CR
  ['] (t2) do-it  CR ;

: TEST3 ( -- )
  CR ." TEST3   FS."
  CR ['] FS. do-it  CR ;

: TEST4 ( -- )
  CR ." TEST4   F."
  CR ['] F. do-it  CR ;

: TEST5 ( -- )
  CR ." TEST5   G."
  CR ['] G. do-it  CR ;

: TEST6 ( -- )
  PRECISION >R  1 SET-PRECISION
  CR ." TEST6   8087 non-numbers  PRECISION = 1" CR
  CR 1.E0 0.E0 F/  FDUP G.
  CR FNEGATE            G.
  CR 0.E0 0.E0 F/  FDUP G.
  CR FNEGATE            G.
  CR
  R> SET-PRECISION  fclex  ;

[THEN]

