\ xchar reference implementation: UTF-8 (and ISO-LATIN-1)

\ This code is provided under the terms of the GNU LESSER
\ GENERAL PUBLIC LICENSE version 3.
\ The original code was provided by Bernd Paysan as part of the
\ Minos port to VFX Forth.

\ environmental dependency: characters are stored as bytes
\ environmental dependency: lower case words accepted

\ =========
\ *> xchars
\ *T Extended character words and UTF-8
\ =========
\ *P Forth defines graphic and control characters from the ASCII
\ ** character set.  ASCII is only appropriate for the English
\ ** language.  Most western languages however fit somewhat into the
\ ** Forth frame, since a single byte is sufficient to encode the few
\ ** special characters in each (though not always the same encoding
\ ** can be used; latin-1 is widely used, though).  For other
\ ** languages, different char sets have to be used, several of them
\ ** variable-width.  Most prominent representative of the variable-
\ ** width character sets is UTF-8, which has gained quite some
\ ** popularity recently.  Since ANS Forth specifies ASCII encoding for
\ ** characters, only ASCII-compatible encodings may be used.
\ ** Fortunately, being ASCII compatible has so many benefits that most
\ ** encodings actually are ASCII compatible.  The characters beyond
\ ** the ASCII encoding are called "extended characters".

\ *P Fixed width representations like UTF-16 have also been proposed,
\ ** but those did not catch on, especially since they are not
\ ** sufficiently ASCII compatible.  Furthermore, UTF-16 also became a
\ ** variable-width representation when UCS32 was introduced.

\ *P The xchar wordset does not solve problems that come from using
\ ** multiple different encodings and switching or converting between
\ ** them. A system implementing the xchar proposal should support at
\ ** least UTF-8 as widespread and common encoding.

\ *P All words dealing with strings should handle xchars when the xchar
\ ** wordset is present.  This includes dictionary definitions, so the
\ ** dictionary entries should not use bit 8 for other purposes.  White
\ ** space parsing does not have to treat code points greater hex 20 as
\ ** white space.


\ ************
\ Dependencies
\ ************

decimal

[undefined] safe/string [if]
: safe/string ( c-addr u n -- c-addr' u' )
\ *G protect /string against overflows.
    dup negate >r  dup 0> IF
	/string dup r> u>= IF  + 0  THEN
    ELSE
	/string dup r> u< IF  + 1+ -1  THEN
    THEN ;
[then]

-77 Constant mal-xchar


\ *********
\ Main code
\ *********

base @ hex

$80 Value maxascii

: xc-size ( xc -- n )
\ *G u is the number of pchars used to encode xchar in memory..
    dup      maxascii u< IF  drop 1  EXIT  THEN \ special case ASCII
    $800  2 >r
    BEGIN  2dup u>=  WHILE  5 lshift r> 1+ >r  dup 0= UNTIL  THEN
    2drop r> ;

: xc@+ ( xcaddr -- xcaddr' u )
\ *G Fetches the xchar at xc-addr1.  Xc-addr2 points to the first
\ ** memory location after the retrieved xchar.
    count  dup maxascii u< IF  EXIT  THEN  \ special case ASCII
    7F and  40 >r
    BEGIN   dup r@ and  WHILE  r@ xor
	    6 lshift r> 5 lshift >r >r count
\	    dup C0 and 80 <> abort" malformed character"
	    3F and r> or
    REPEAT  r> drop ;

: xc!+ ( xc xcaddr -- xcaddr' )
\ *G Stores the xchar at xc-addr1. Xc-addr2 points to the first
\ ** memory location after the stored xchar.
    over maxascii u< IF  tuck c! char+  EXIT  THEN \ special case ASCII
    >r 0 swap  3F
    BEGIN  2dup u>  WHILE
	    2/ >r  dup 3F and $80 or swap 6 rshift r>
    REPEAT  7F xor 2* or  r>
    BEGIN   over 80 u< 0= WHILE  tuck c! char+  REPEAT  nip ;

: xc, ( xchar -- )
\ *G Append the encoding of xchar to the dictionary.
  here xc!+ dp !  ;

: xc!+? ( xc xcaddr u -- xcaddr' u' flag )
\ *G Stores the xchar into the string buffer specified by
\ ** xc-addr1 u1. Xc-addr2 u2 is the remaining string buffer.
\ ** If the xchar did fit into the buffer, flag is true,
\ ** otherwise flag is false, and xc-addr2 u2 equal xc-addr1 u1.
\ ** XC!+? is safe for buffer overflows.
    >r over xc-size r@ over u< IF ( xc xc-addr1 len r: u1 )
	\ not enough space
	drop nip r> false
    ELSE
	>r xc!+ r> r> swap - true
    THEN ;


\ scan to next/previous character

: xchar+ ( xcaddr -- xcaddr' )
\ *G Adds the size of the xchar stored at xc-addr1 to this
\ ** address, giving xc-addr2.
  xc@+ drop ;
: xchar- ( xcaddr -- xcaddr' )
\ *G Goes backward from xc-addr1 until it finds an xchar so
\ ** that the size of this xchar added to xc-addr2 gives xc-addr1.
\ ** There is an ambiguous condition when the encoding does not
\ ** permit reliable backward stepping through the text.
    BEGIN  1 chars - dup c@ $C0 and maxascii <>  UNTIL ;

: +x/string ( xc-addr1 u1 -- xc-addr2 u2 )
\ *G Step forward by one xchar in the buffer defined by
\ ** xc-addr1 u1. Xc-addr2 u2 is the remaining buffer after
\ ** stepping over the first xchar in the buffer.
    over dup xchar+ swap - safe/string ;
: x\string- ( xc-addr1 u1 -- xc-addr2 u2 )
\ *G Search for the penultimate xchar in the string xc-addr1 u1.
\ ** The string xc-addr1 u2 contains all xchars of xc-addr u1,
\ ** but the last. Unlike XCHAR-, X\STRING- can be implemented
\ ** in encodings where xchar boundaries can only reliably
\ ** detected when scanning in the forward direction.
    over dup xchar- swap - safe/string ;

: x-size ( xcaddr u1 -- u2 )
\ *G U2 is the number of pchars used to encode the first xchar
\ ** stored in the string xc-addr u1.  To calculate the size of
\ ** the xchar, only the bytes inside the buffer may be accessed.
\ ** An ambiguous condition exists if the xchar is incomplete or
\ ** malformed.
    0=        IF drop 0 exit  THEN
    \ length of UTF-8 char starting at u8-addr (accesses only u8-addr)
    c@
    dup $80 u< IF drop 1 exit THEN
    dup $c0 u< IF mal-xchar throw THEN \ really is a malformed character
    dup $e0 u< IF drop 2 exit THEN
    dup $f0 u< IF drop 3 exit THEN
    dup $f8 u< IF drop 4 exit THEN
    dup $fc u< IF drop 5 exit THEN
    dup $fe u< IF drop 6 exit THEN
    mal-xchar throw ; \ also malformed character

: -trailing-garbage ( xcaddr u1 -- xcaddr u2 )
\ *G Examine the last xchar in the string xc-addr u1 - if the
\ ** encoding is correct and it represents a full xchar, u2
\ ** equals u1, otherwise, u2 represents the string without the
\ ** last (garbled) xchar.  -TRAILING-GARBAGE does not change
\ ** this garbled xchar.
    2dup + dup xchar- ( addr u1 end1 end2 )
    2dup dup 2dup - x-size + = IF	\ last character ok
	2drop
    ELSE
	nip nip over -
    THEN ;

\ utf key and emit

: xkey ( -- xc )
\ *G Reads an xchar from the terminal. This will discard all
\ ** input events up to the completion of the xchar.
    key dup maxascii u< IF  EXIT  THEN  \ special case ASCII
    $7F and  $40 >r
    BEGIN  dup r@ and  WHILE  r@ xor
	    6 lshift r> 5 lshift >r >r key
\	    dup $C0 and $80 <> mal-xchar and throw
	    $3F and r> or
    REPEAT  r> drop ;

: xemit ( xc -- )
\ *G Prints an xchar on the terminal.
    dup maxascii u< IF  emit  EXIT  THEN \ special case ASCII
    0 swap  $3F
    BEGIN  2dup u>  WHILE
	    2/ >r  dup $3F and $80 or swap 6 rshift r>
    REPEAT  $7F xor 2* or
    BEGIN   dup $80 u< 0= WHILE emit  REPEAT  drop ;

[undefined] holds [if]
: holds ( addr u -- )
\ *G Add a string to the pictured numeric output buffer.
    BEGIN  dup  WHILE  1- 2dup + c@ hold  REPEAT  2drop ;
[then]

Create xholdbuf 8 allot

: xhold ( xchar -- )
\ *G Add an xchar to the pictured numeric output buffer.
  xholdbuf tuck xc!+ over - holds ;

\ utf size

\ uses wcwidth ( xc -- n )

: wc, ( n low high -- )  1+ , , , ;

Create wc-table \ derived from wcwidth source code, for UCS32
  0 $0300  $0357 wc,
  0 $035D  $036F wc,
  0 $0483  $0486 wc,
  0 $0488  $0489 wc,
  0 $0591  $05A1 wc,
  0 $05A3  $05B9 wc,
  0 $05BB  $05BD wc,
  0 $05BF  $05BF wc,
  0 $05C1  $05C2 wc,
  0 $05C4  $05C4 wc,
  0 $0600  $0603 wc,
  0 $0610  $0615 wc,
  0 $064B  $0658 wc,
  0 $0670  $0670 wc,
  0 $06D6  $06E4 wc,
  0 $06E7  $06E8 wc,
  0 $06EA  $06ED wc,
  0 $070F  $070F wc,
  0 $0711  $0711 wc,
  0 $0730  $074A wc,
  0 $07A6  $07B0 wc,
  0 $0901  $0902 wc,
  0 $093C  $093C wc,
  0 $0941  $0948 wc,
  0 $094D  $094D wc,
  0 $0951  $0954 wc,
  0 $0962  $0963 wc,
  0 $0981  $0981 wc,
  0 $09BC  $09BC wc,
  0 $09C1  $09C4 wc,
  0 $09CD  $09CD wc,
  0 $09E2  $09E3 wc,
  0 $0A01  $0A02 wc,
  0 $0A3C  $0A3C wc,
  0 $0A41  $0A42 wc,
  0 $0A47  $0A48 wc,
  0 $0A4B  $0A4D wc,
  0 $0A70  $0A71 wc,
  0 $0A81  $0A82 wc,
  0 $0ABC  $0ABC wc,
  0 $0AC1  $0AC5 wc,
  0 $0AC7  $0AC8 wc,
  0 $0ACD  $0ACD wc,
  0 $0AE2  $0AE3 wc,
  0 $0B01  $0B01 wc,
  0 $0B3C  $0B3C wc,
  0 $0B3F  $0B3F wc,
  0 $0B41  $0B43 wc,
  0 $0B4D  $0B4D wc,
  0 $0B56  $0B56 wc,
  0 $0B82  $0B82 wc,
  0 $0BC0  $0BC0 wc,
  0 $0BCD  $0BCD wc,
  0 $0C3E  $0C40 wc,
  0 $0C46  $0C48 wc,
  0 $0C4A  $0C4D wc,
  0 $0C55  $0C56 wc,
  0 $0CBC  $0CBC wc,
  0 $0CBF  $0CBF wc,
  0 $0CC6  $0CC6 wc,
  0 $0CCC  $0CCD wc,
  0 $0D41  $0D43 wc,
  0 $0D4D  $0D4D wc,
  0 $0DCA  $0DCA wc,
  0 $0DD2  $0DD4 wc,
  0 $0DD6  $0DD6 wc,
  0 $0E31  $0E31 wc,
  0 $0E34  $0E3A wc,
  0 $0E47  $0E4E wc,
  0 $0EB1  $0EB1 wc,
  0 $0EB4  $0EB9 wc,
  0 $0EBB  $0EBC wc,
  0 $0EC8  $0ECD wc,
  0 $0F18  $0F19 wc,
  0 $0F35  $0F35 wc,
  0 $0F37  $0F37 wc,
  0 $0F39  $0F39 wc,
  0 $0F71  $0F7E wc,
  0 $0F80  $0F84 wc,
  0 $0F86  $0F87 wc,
  0 $0F90  $0F97 wc,
  0 $0F99  $0FBC wc,
  0 $0FC6  $0FC6 wc,
  0 $102D  $1030 wc,
  0 $1032  $1032 wc,
  0 $1036  $1037 wc,
  0 $1039  $1039 wc,
  0 $1058  $1059 wc,
  1 $0000  $1100 wc,
  2 $1100  $115f wc,
  0 $1160  $11FF wc,
  0 $1712  $1714 wc,
  0 $1732  $1734 wc,
  0 $1752  $1753 wc,
  0 $1772  $1773 wc,
  0 $17B4  $17B5 wc,
  0 $17B7  $17BD wc,
  0 $17C6  $17C6 wc,
  0 $17C9  $17D3 wc,
  0 $17DD  $17DD wc,
  0 $180B  $180D wc,
  0 $18A9  $18A9 wc,
  0 $1920  $1922 wc,
  0 $1927  $1928 wc,
  0 $1932  $1932 wc,
  0 $1939  $193B wc,
  0 $200B  $200F wc,
  0 $202A  $202E wc,
  0 $2060  $2063 wc,
  0 $206A  $206F wc,
  0 $20D0  $20EA wc,
  2 $2329  $232A wc,
  0 $302A  $302F wc,
  2 $2E80  $303E wc,
  0 $3099  $309A wc,
  2 $3040  $A4CF wc,
  2 $AC00  $D7A3 wc,
  2 $F900  $FAFF wc,
  0 $FB1E  $FB1E wc,
  0 $FE00  $FE0F wc,
  0 $FE20  $FE23 wc,
  2 $FE30  $FE6F wc,
  0 $FEFF  $FEFF wc,
  2 $FF00  $FF60 wc,
  2 $FFE0  $FFE6 wc,
  0 $FFF9  $FFFB wc,
  0 $1D167 $1D169 wc,
  0 $1D173 $1D182 wc,
  0 $1D185 $1D18B wc,
  0 $1D1AA $1D1AD wc,
  2 $20000 $2FFFD wc,
  2 $30000 $3FFFD wc,
  0 $E0001 $E0001 wc,
  0 $E0020 $E007F wc,
  0 $E0100 $E01EF wc,
here wc-table - Constant #wc-table

\ inefficient table walk:

: xc-width ( xchar -- n )
\ *G N is the number of monospace ASCII characters that take
\ ** the same space to display as the xchar; i.e. xchar width
\ ** is always an integer multiple of the width of an ASCII
\ ** character.
  wc-table #wc-table over + swap ?DO
    dup I 2@ within
    IF  drop I 2 cells + @  UNLOOP EXIT  THEN
  3 cells +LOOP
  drop 1 ;

: x-width ( xcaddr u -- n )
\ *G N is the number of monospace ASCII characters that take
\ ** the same space to display as the the xchar string xc-addr u;
\ ** assuming a monospaced display font, i.e. xchar width is
\ ** always an integer multiple of the width of an ASCII character.
    0 rot rot over + swap ?DO
	I xc@+ swap >r xc-width +
    r> I - +LOOP ;

: char  ( "name" -- xc )
\ *G *\fo{CHAR} is redefined when XCHARs are present. Skip
\ ** leading space delimiters.  Parse name delimited by a space.
\ ** Put the value of its first xchar onto the stack.
  bl word count drop xc@+ nip ;
: [char] ( "name" -- rt:xc )
\ *G *\fo{[CHAR]} is redefined when XCHARs are present.
\ ** Compile the first xchar of the next token in the input stream
\ ** as a literal. Usually used to avoid magic numbers in the source
\ ** code.
  char postpone Literal ; immediate

\ switching encoding is only recommended at startup
\ only two encodings are supported: UTF-8 and ISO-LATIN-1

$080 Constant utf-8		\ -- x
\ *G Used with *\fo{maxascii} to say that the UTF-8 character
\ ** set is being used.
$100 Constant iso-latin-1	\ -- x
\ *G Used with *\fo{maxascii} to say that the Latin-1 character
\ ** set is being used.

base !

[defined] char@ [IF]
  ' xc@+ IS char@
[THEN]


\ ======
\ *> ###
\ ======

decimal
