\ XREF.FTH - cross references for VFX Forth

((
Copyright (c) 2001, 2003, 2004, 2005, 2007, 2008, 2009, 2014, 2017

MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

20190830 SFP017 Cleared DP(X) at startup.
20170221 SFP016 Update for tooltips.
20140106 SFP015 Exposed INIT-XREF and TERM-XREF.
20140105 SFP014 Removed redundant code.
20090427 SFP013 Corrected ADD-LIT(X) for floats.
		Replaced EVALUATE with ASSESS.
20080125 SFP012 Added code for non-GUI systems.
		Updated formatting in documentation.
20070815 SFP011 Reserved 256 byte header in new files only.
20050329 SFP010 Corrected .XREFS-UNUSED.
20040716 SFP009 Added warning box and resizing XREF memory.
20031015 SFP008 Added WALKXREF and WALKDECOMP.
20030805 SFP007 Used LATEST-XT.
20030314 SFP006	Added protection against overflow.
SFP005	14 June 2001	Inhibited XREF from console and EVALUATE,
			and of definitions in transient areas.
			Added macro expansion to file handling.
			Replaced all uses of ABORT" by ?THROW.
SFP004	24 Nov 1999	Enhanced. Added literal handling.
			Handled floats using EVALUATE.
			Added inline string display.
			Added extra documentation.
SFP003  22 Jul 1999     Converted to PFW/VFX
SFP002  18 Nov 1998     Added XREF-ALL and XREF-UNUSED
SFP001  15 Nov 1998     Original release
))

\ ==========
\ *> library
\ *S Building cross references
\ ==========
\ *N Introduction

\ *P Cross reference information helps you to manage your source code.
\ ** When *\i{LIB\XREF.FTH} is loaded you can use *\fo{XREF <name>} to
\ ** find out in which other words *\fo{<name>} is used. You can also
\ ** find out which words you  defined but did not use. *\fo{XREF}
\ ** is precompiled in the Studio version  of VFX Forth but not in
\ ** the base version.

\ *P The compiler generates cross references by building a chain of fields
\ ** including *\fo{LOCATE} format (link:32, xt:32, line#:32) in a
\ ** separate area of memory. Links and pointers are relative to the start
\ ** of the *\fo{XREF} memory area.

\ *P Two chains are maintained. The first produces a chain of
\ ** where a word is used, so that the user can find out where
\ ** (say) *\fo{DUP} is used. The second produces  a chain of
\ ** which words and literals are called in order. This is the
\ ** basis of decompilation and debugging.

\ *N Initialisation
\ *P *\fo{XREF} is initialised by the switch *\fo{+XREFS} and is
\ ** terminated by *\fo{-XREFS}.  You must use *\fo{+XREFS} to
\ ** turn on the production of cross reference information.

( *P By default 1Mb of cross reference memory is allocated from the heap. )
( ** If you need more than this for a very large application, use the phrase )
( ** *\fo{<n> XREF-KB} to set the size of the cross reference memory, where )
( ** <n> is in kilobytes. )

( *N Decompilation and SHOW )
( *P Because the VFX code generator optimises so heavily, there is no )
( ** direct relationship between the binary code and the source code. )
( ** Consequently *\fo{DIS} and *\fo{DASM} use disassembly and special cases, but )
( ** cannot produce a good approximation to the original source code. )

( *P The cross reference information includes a decompilation chain. )
( ** When you use *\fo{SHOW <name>} the cross reference information is used to )
( ** produce a machine decompilation. This includes none of the comments )
( ** from the original source code, and is machine formatted. )

( *N Extending SHOW )
( *P The decompilation produced by *\fo{SHOW} is mostly default and )
( ** automatic. However, some words such as string handling )
( ** take in line data which would not be displayed by *\fo{SHOW} without )
( ** special handling. )

( *P *\fo{SHOW} can be extended by adding items to the *\fo{DCC-SWITCH} chain. The )
( ** stack effect of the action is: addrx -- addr ; where addrx is the )
( ** offset of the cross reference packet in the cross reference information )
( ** memory. See the *\fo{/REF[X]} structure in *\i{LIB\XREF.FTH} for details )
( ** of the structure of this data packet. The example below is for a )
( ** word X" which takes an in-line string like S". )

( *E [+switch dcc-switch					)
( **   ' X"       run:  ." X"  [char] " emit  dup .$inline  ;   )
( ** switch]

( *P Note that unlike previous VFX Forth decompilers, *\fo{SHOW} is based )
( ** on cross reference information which references the source word )
( ** without knowledge of what it compiles. The only reasons for special )
( ** cases are control of the decompilation layout and display of )
( ** associated data to reconstruct source code.)

( *N Glossary )

[undefined] [switch [if]
  include %LibraryDir%\switches
[then]

-short-branches

MODULE XREFFER


\ **************
\ Error messages
\ **************

\ SFP006...
-901 #ErrDef err_xref_mem "XREF: overflow - Use <n> XREF-KB to increase"
\ ...SFP006


\ ********************************
\ Structures and XREF table access
\ ********************************
\ The start of an XREF file or memory block contains two slots
\ at the start of the block:
\ 1) HERE when the block was initially created.
\ 2) A pointer to the last XREF header. This is used when
\    reloading the block.

struct /head(x) \ -- n ; word header
\ Every referenced word's >XREF field points to one of these.
  cell field +xrc-link          \ xref chain link, anchors >LINK(X).
  cell field +dcc-head          \ first entry in decompilation chain, >LINK(D)
  cell field +dcc-tail          \ last entry in decompilation chain, >LINK(D)
  cell field +xrc-appXref	\ points to >XREF field in running Forth host ; SFP010
  cell field +xhd-link		\ links all XREF heads ; SFP010
  cell field +xttip		\ address of tooltip text ; SFP016
end-struct

struct /ref(x)	\ -- n ; reference to a word
\ Every reference to a word or literal results in one of these.
  cell field >link(x)           \ xref link field
  cell field >sym(x)            \ symbol (xt) in which this reference occurs
  cell field >line#(x)          \ line# in file being compiled
  cell field >in(x)             \ offset in line
  cell field >xtHere(x)         \ HERE when reference was made
  cell field >type(x)           \ 0 for xts, 1 for literals, 2 doubles,
  				\ -2 for floats on FP stack, -1 for OGL floats on data stack
  cell field >link(d)           \ decompilation link
  3 cells field >xtadded(x)     \ xt being added or literal value (1, 2, or 3 cells)
end-struct

variable /xref  \ -- addr ; xref memory size, default to 2Mb
  $0020:0000 /xref !
variable ^xref  \ -- addr ; holds address of XREF memory
  0 ^xref !
variable dp(x)  \ -- addr ; holds current offset into XREF memory
  0 dp(x) !

: here(x)       \ -- addr ; analogue of HERE
  dp(x) @
;

: ,(x)          \ x -- ; analogue of ,
  ^xref @ dp(x) @ + !
  cell dp(x) +!
;

: w,(x)         \ w -- ; analogue of W,
  ^xref @ dp(x) @ + w!
  2 dp(x) +!
;

: c,(x)         \ w -- ; analogue of W,
  ^xref @ dp(x) @ + c!
  1 dp(x) +!
;

: @(x)          \ offset -- x
  ^xref @ + @
;

: !(x)          \ x offset --
  ^xref @ + !
;

: w@(x)         \ offset -- w
  ^xref @ + w@
;

: w!(x)         \ w offset --
  ^xref @ + w!
;

: c@(x)         \ offset -- b
  ^xref @ + c@
;

: c!(x)         \ b offset --
  ^xref @ + c!
;

: 2@(x)		\ offset -- d
  ^xref @ + 2@
;

: 2!(x)		\ d offset --
  ^xref @ + 2!
;

[defined] Target_386_Windows [if]
: ?xref-mem	\ -- ; check for enough XREF memory ; SFP009
  dp(x) @  /xref @ $100 - u> if		\ overflow test
    hWndMain z" XREF memory full, OK to increase, CANCEL to throw" zSysName
    MB_OKCANCEL MessageBox
    IDOK <> err_xref_mem ?throw
    /xref @ 3 2 */ /xref !
    ^xref @ /xref @ resize if
      hWndMain z" Cannot increase XREF memory" zSysName MB_OK MessageBox drop
      #-406 throw
    endif
    ^xref !
  endif
;
[else]
\ SFP012...
[undefined] y/n? [if]
: y/n?		\ -- flag
\ Issue "Y/N: " and wait for a key and return true if the key
\ is 'Y' or 'y'.
  ."  Y/N: " key upc  dup emit  [char] Y =
;
[then]
: ?xref-mem	\ -- ; check for enough XREF memory ; SFP009
  dp(x) @  /xref @ $100 - u> if		\ overflow test
    cr ." XREF memory full, increase it?" y/n?
    0= err_xref_mem ?throw
    /xref @ 3 2 */ /xref !
    ^xref @ /xref @ resize if
      cr ." Cannot increase XREF memory" #-406 throw
    endif
    ^xref !
  endif
;
\ ...SFP012
[then]

: dump(x)       \ offset len --
( *G Displays the specified contents of the XREF table. Note that the )
( ** given address is an *\b{offset} from the start of the XREF table. )
  swap ^xref @ + swap dump
;

: align(x)	\ -- ; align XREF space
  dp(x) @ aligned dp(x) !
;

: text,(x)	\ caddr len -- ; lay text in XREF space
  #255 umin dup c,(x)  bounds
  ?do  i c@ c,(x)  loop
  align(x)
;


\ *********************
\ XREF table management
\ *********************

\ SFP010...
: (wipeXref)	\ thread# nfa -- flag ; return TRUE to continue
  name> >xref off  drop true
;

: wipeXrefs	\ -- ; wipe all XREF info in host
  ['] (wipeXref) WalkAllWords
;
\ ...SFP010

: init-xref     \ --
\ *G Initialise XREF memory and information if not already set up.
  ^xref @ 0= if                         \ ignore on subsequent runs
    wipeXrefs				\ delete existing XREFs ; SFP010
    /xref @ allocate #-406 ?throw	\ abort" Can't get XREF memory"
    ^xref !  0 dp(x) !
    here ,(x)                           \ place dictionary marker at start
    0 ,(x)				\ anchors XREF headers ; SFP010
    #256 dp(x) !			\ reserve header block ; SFP011
  endif
; doNotSin

: term-xref     \ --
\ *G Free up XREF memory.
  ^xref @ if
    ^xref @off free #-407 ?throw	\ abort" Can't free XREF memory"
    0 dp(x) !				\ SFP017
  endif
; doNotSin
' term-xref AtExit

max_path buffer: XREF$  \ -- addr ; buffer for XREF file name
  c" XREF.XRF" xref$ $move

: save-xref     \ -- ; save XREF memory to file
( *G Save the cross reference memory to disc. Unless the )
( ** file name has been changed by *\fo{XREF: <filename>} the )
( ** file will be called *\i{XREF.XRF}. )
  { | temp[ max_path ] -- }		\ SFP005
  ^xref @ if
    xref$ count temp[ expandmacro	\ expand source macros ; SFP005
    2dup fileExist? if
      2dup delete-file drop
    then
    r/w create-file #-412 ?throw	\ abort" Can't create XREF file"
    >r
    ^xref @ here(x) r@ write-file #-415 ?throw	\ abort" Can't Write XREF file"
    r> close-file #-416 ?throw		\ abort" Can't close XREF file"
  endif
; doNotSin

: install-xrefs	\ -- ; install file's XREF info ; SFP010
\  cr ." Installing XREFs ... "
  cell @(x)
  begin
    dup
   while
\    dup dup +xrc-appXref @(x) @ <> if	\ check for
\      cr ." head = " dup .dword  ."  was " dup +xrc-appXref @(x) @ .dword
\    endif
    dup dup +xrc-appXref @(x) !		\ set header location in host
    +xhd-link @(x)
  repeat
  drop
\  ." done. "
; doNotSin

: load-xref     \ -- ; reload XREF file from disc
\ *G Load the cross reference memory from disc. Unless the
\ ** file name has been changed by *\fo{XREF: <filename>} the
\ ** file used will be *\i{XREF.XRF}.
  { | temp[ max_path ] -- }		\ SFP005
  term-xref init-xref			\ release and get memory
  xref$ count temp[ expandmacro		\ expand source macros ; SFP005
  r/o open-file #-413 ?throw		\ abort" Can't open XREF file"
  >r
  ^xref @ /xref @ r@ read-file #-414 ?throw  dp(x) !	\ abort" Can't read XREF file"  dp(x) !
  r> close-file #-416 ?throw		\ abort" Can't close XREF file"
  install-xrefs				\ set all >XREF fields ; SFP010
; doNotSin

: xref:         \ "filename" -- ; enable XREFs
\ *G Use in the form *\fo{XREF: <filename>} to define the file that
\ ** *\fo{SAVE-XREF} and *\fo{LOAD-XREF} will use.
  bl word xref$ $move
;

: xref-kb       \ n --
\ *G Specifies the size of the cross reference memory in kilobytes.
\ ** By default this is 1024 kb, or 1Mb.
  #1024 * /xref !
;


\ ***************************************
\ Adding a symbol or literal to the table
\ ***************************************

: >xref-head    \ xt -- addr(x) ; make new head if needed
  >xref  dup @ dup if
    nip                                 \ discard ^XREF field
  else
    drop                                \ >xref 0 -- >xref
    dup >r
    here(x) dup rot !                   \ save new header address
    0 ,(x)  0 ,(x)  0 ,(x)
\ N.B. Change this to ignore LOCALS - check with inLocals?
    r> dup inLocals? if			\ locals have no head pointer or link
      drop  0 ,(x)  0 ,(x)
    else
      ,(x)				\ host XREF head ; SFP010
      cell @(x) ,(x)  dup cell !(x)	\ links all non-local heads ; SFP010
    endif
    0 ,(x)				\ tooltip text ; SFP016
  endif
;

: >decomp	\ xt addr(x) -- ; add packet to XT's decomp chain
  >r                                    \ save packet
  >xref-head                            \ get/make defn header (AFTER main block!)
  dup +dcc-head @(x) 0=                 \ if Decomp - chain head has not been set
  if  r@ over +dcc-head !(x)  endif     \   set it up
  dup +dcc-tail @(x) ?dup               \ if previous block exists
  if  r@ swap >link(d) !(x)  endif      \   set decomp link in previous block
  r> swap +dcc-tail !(x)                \ update tail pointer
;

: EndTokenPos	\ -- pos ; position in line of END of token
  >in @
  dup #tib @ u<
  if  1-  endif
  0 max
;

: add-xt(x)     \ xt -- ; add XREF location to symbol chain
  state @ 0=                            \ don't XREF during interpretation
  if  drop  exit  endif
  source-id 0=                          \ don't XREF from console
  if  drop  exit  endif
  source-id -1 =                        \ don't XREF from EVALUATE
  if  drop  exit  endif
  dup inforth? 0=			\ don't XREF transients
  if  drop  exit  endif

  ?xref-mem				\ oveflow test ; SFP006

  dup >xref-head                        \ point to XREF chain anchor
  here(x) >r                            \ start of this block
    dup +xrc-link @(x) ,(x)             \ add and update xref link field
    r@ swap +xrc-link !(x)
  latest-xt dup ,(x)			\ add current symbol
  line# @ ,(x)                          \ add LOCATE field, -1 for keyboard
  endTokenPos ,(x)                      \ offset of last character in text
\ -- xt-to-add xt-defn
  here ,(x)                             \ HERE when xt was added
  0 ,(x)                                \ XT is type 0
  0 ,(x)                                \ no further blocks (yet)
  swap ,(x)                             \ XT of word being added
  0 ,(x)  0 ,(x)			\ filler

  r> >decomp				\ set up decompilation chain
;

: add-lit(x)    \ n1..nn n -- n1..nn n ; copy reference
  state @ 0=                            \ don't XREF during interpretation
  if  exit  endif
  source-id 0=                          \ don't XREF from console
  if  exit  endif
  source-id -1 =                        \ don't XREF from EVALUATE
  if  exit  endif

  ?xref-mem				\ oveflow test ; SFP006

  here(x) >r				\ start of this reference
  0 ,(x)                                \ null XREF link field
  latest-xt  dup >r  ,(x)		\ add current symbol
  line# @ ,(x)                          \ add LOCATE field, -1 for keyboard
  endTokenPos ,(x)                      \ offset of last character in text
  here ,(x)                             \ HERE when xt was added
  dup ,(x)                              \ literals are of type 1/2...
  0 ,(x)                                \ decompilation link
  here(x) >r  0 ,(x)  0 ,(x)  0 ,(x)	\ space for largest literal (10 byte float)
  case					\ inspect n
     1 of  dup r> !(x)  1  endof	\ single
     2 of  2dup r> 2!(x)  2  endof	\ double
    -2 of  r> ^xref @ +			\ float on FP stack
           s" fdup f!" assess		\ SFP013

((
           state off			\ SFP013, next line is executed not compiled
           s" fdup f!" evaluate
           state on			\ SFP013
))
           -2
       endof
    -1 of  dup r> !(x)  -1  endof	\ 32 bit OGL float on data stack
           r> drop
  end-case

  r> r> >decomp
;

: no-xref	\ -- ; force xref info to be cleared at cold
  0 ^xref !
  0 dp(x) !				\ SFP017
  assign drop to-do addxref
  assign noop to-do addXrefLit
;
' no-xref AtCold

: +xrefs        \ -- ; enable XREF
\ *G Initialises the cross reference system if it has not already
\ ** been initialised, and enables production of cross reference
\ ** information.
  init-xref
  assign add-xt(x) to-do addXref
  assign add-lit(x) to-do addXrefLit
;

: -xrefs        \ -- ; disable XREF
\ *G Stops production of cross reference information, which can
\ ** be restarted by *\fo{+XREFS}.
\ ** Cross reference memory is not erased or released. Thus,
\ ** restarting with *\fo{+XREFS} will retain information.
\ ** To release all previous information use *\fo{TERM-XREF}
\ ** before *\fo{+XREFS}.
  assign drop to-do addxref
  assign noop to-do addXrefLit
;
-xrefs

: xrefs?	\ -- flag ; true if XREF table exists
  ^xref @ 0<>
;


\ ******************
\ Display formatting
\ ******************

variable tab-width      \ width of a tab stop
  #16 tab-width !

: tab           \ -- ; step to next tab stop
  out @ c/l tab-width @ - >= if
    cr
  else
    tab-width @  out @ over mod - spaces
  endif
;


\ ********
\ Displays
\ ********

variable last-xref-xt	\ -- addr ; used to stop duplicate displays

: .xref         \ xref -- ; display name of symbol
  >sym(x) @(x)  dup last-xref-xt @ <> if
    dup last-xref-xt !
    dup inForth? if			\ check header is in Forth
      >name tab .name                   \ show name of symbol in which used
    else
      drop tab ." LV/unknown "
    endif
  else
    drop
  then
;

: .xrefs        \ xt -- ; display list of words that use the symbol
  last-xref-xt off
  dup >xref @ if                        \ get head of chains
    cr dup >name .name ." is used by:"  \ display word name     SFP002
    c/l out !
    >xref @ +xrc-link @(x)              \ get XREF chain anchor from head
    begin
      ?dup
     while
      dup .xref
      >link(x) @(x)
    repeat
    cr
  else
    drop
  endif
;

: xref-report   \ -- ; display XREF information
( *G Displays some statistics about cross reference memory )
( ** usage. )
  ^xref @ if
    cr
    ." XREF memory used = " here(x) #1024 / . ." Kb of "
    /xref @ #1024 / . ." Kb allocated"
  endif
;

: WalkXref	\ xt1 xt2 -- ; XREF of XT1 using XT2 to display.
\ *G Used by application tools to walk the XREF chain for XT1.
\ ** The structure offset for each step in the chain is handled
\ ** by XT2 ( offset -- ). Because writing XT2 requires use of
\ ** the internal XREF structure, you must expose the *\fo{XREFFER}
\ ** module: *\fo{EXPOSE-MODULE XREFFER} to get access to the words
\ ** in *\i{Lib\XREF.FTH}.
  >r
  dup >xref @ if                        \ get head of chains
    >xref @ +xrc-link @(x)              \ get XREF chain anchor from head
    begin
      ?dup
     while
      dup r@ execute
      >link(x) @(x)
    repeat
  else
    drop
  endif
  r> drop
;


\ ****************************************************************
\ Using the XREF system to decompile a word - used by the debugger
\ ****************************************************************

variable #indent

: indent        \ -- ; cr and spaces
  cr  #indent @ spaces
;

: ?newline      \ -- ; new line if needed
  out @ c/l >=
  if  indent  endif
;

: at-end        \ -- ; forces new line
  c/l out !
;

2 constant #step

: +indent       \ -- ; change indent level
  #step #indent +!  at-end
;

: -indent       \ -- ; change indent level
  #step negate #indent +!  at-end
;

: .$inline	\ addr -- ; display inline string after xt at addr
  space
  >xthere(x) @(x) 5 + count type
  [char] " emit  space
;

: dcc-default	\ addr(x) xt -- addr(x)
  ?newline >name .name
;

also forth definitions
[switch dcc-switch dcc-default
  ' if       run:  indent ." IF " +indent  ;
  ' else     run:  -indent indent ." ELSE " +indent  ;
  ' endif    run:  -indent indent ." ENDIF " at-end  ;
  ' then     run:  -indent indent ." THEN " at-end  ;
  ' begin    run:  indent ." BEGIN " +indent  ;
  ' while    run:  -indent indent ." WHILE " +indent ;
  ' repeat   run:  -indent indent ." REPEAT " at-end  ;
  ' again    run:  -indent indent ." AGAIN " at-end  ;
  ' ;        run:  -indent indent ." ; " indent  ;
  ' do       run:  indent ." DO " +indent  ;
  ' ?do      run:  indent ." ?DO " +indent  ;
  ' loop     run:  -indent indent ." LOOP " at-end  ;
  ' +loop    run:  -indent indent ." +LOOP " at-end  ;
  ' case     run:  indent ." CASE " +indent  ;
  ' of       run:  indent ." OF " +indent  ;
  ' endof    run:  -indent indent ." ENDOF " at-end  ;
  ' endcase  run:  -indent indent ." ENDCASE " at-end  ;
  ' end-case run:  -indent indent ." END-CASE " at-end  ;
  ' \        runs noop
  ' (        runs noop
  ' ((       runs noop		\ )) for syntax highlighting
  ' S"       run:  ." S"  [char] " emit  dup .$inline  ;
  ' C"       run:  ." C"  [char] " emit  dup .$inline  ;
  ' ."       run:  ." ."  [char] " emit  dup .$inline  ;
  ' abort"   run:  ." ABORT"  [char] " emit  dup .$inline  ;
switch]
previous definitions

: .dcc-xt	\ addr(x) -- ; display XT referenced
  dup
  >xtadded(x) @(x) dcc-switch
  drop
;

: .dcc-ref	\ addr(x) -- ; display decompilation reference
  case  dup >type(x) @(x)
    0  of  .dcc-xt  endof
    1  of  >xtadded(x) @(x) .  endof
    2  of  >xtadded(x) 2@(x) 0 d.r  dp-char @ emit space  endof
\    -2 of  >xtadded(x) ^xref @ + s" f@ f." evaluate  endof
    -2 of  >xtadded(x) ^xref @ + s" f@ f." assess  endof	\ SFP013
           cr ." Unknown decompilation/number type: " .
	   drop
  end-case
;

: (show)        \ xt -- ; show/decompile words used by this XT
( *G Given an XT, produces a machine decompilation of the word )
( ** using the cross reference information. If cross referencing )
( ** is not enabled, no action is taken. )
  ^xref @ 0=
  if  drop  exit  endif
  #step #indent !
  cr ." : "  dup >name .name
  >xref @ ?dup if
    indent
    +dcc-head
    begin
      @(x) ?dup
     while
     dup .dcc-ref
      >link(d)
    repeat
  else
    ." has no XREF data"
  endif
;

: $show         \ $addr --
( *G Given a counted string, it is looked up as a Forth word name )
\ ** and *\fo{(SHOW)} produces a machine decompilation of the word
( ** using the cross reference information. If cross referencing )
( ** is not enabled, no action is taken. )
  ^xref @ 0=
  if  drop  exit  endif
  find
  if  (show)  else  ." is undefined" drop  endif
;

: show          \ -- ; SHOW <name>
( *G The following name is looked up as a Forth word name )
\ ** and *\fo{(SHOW)} produces a machine decompilation of the word
( ** using the cross reference information. If cross referencing )
( ** is not enabled, no action is taken. )
  get-token $show
;

: hasXref?	\ xt -- flag ; true if word has XREF info
( *G produces TRUE if xt has XREF information )
( ** otherwise FALSE is returned. )
  >xref @ 0<>
;

: hasXDecomp?	\ xt -- flag ; true if word has XREF decompilation info
( *G produces TRUE if xt has XREF decompilation information )
( ** otherwise FALSE is returned. )
  >xref @ dup if
    +dcc-head @(x) 0<>
  endif
;

: WalkDecomp	\ xt1 xt2 -- ; DECOMP of XT1 using XT2 to display.
\ *G Used by application tools to walk the decompilation chain for XT1.
\ ** The structure offset for each step in the chain is handled
\ ** by XT2 ( offset -- ). Because writing XT2 requires use of
\ ** the internal XREF structure, you must expose the *\fo{XREFFER}
\ ** module: *\fo{EXPOSE-MODULE XREFFER} to get access to the words
\ ** in *\i{Lib\XREF.FTH}.
  >r
  >xref @ ?dup if
    +dcc-head
    begin
      @(x) ?dup
     while
      dup r@ execute
      >link(d)
    repeat
  endif
  r> drop
;


\ ****************
\ Debugger support
\ ****************

((
struct /ref(x)
  cell field >link(x)           \ xref link field
  cell field >sym(x)            \ symbol (xt) in which this reference occurs
  cell field >line#(x)          \ line# in file being compiled
  cell field >in(x)             \ offset in line
  cell field >xthere(x)         \ HERE when reference was made
  cell field >type(x)           \ 0 for xts, 1 for literals, 2 doubles, -2 for floats
  cell field >link(d)           \ decompilation link
  3 cells field >xtadded(x)     \ xt being added or literal value (1, 2, or 3 cells)
end-struct
))

: FindXrefInfo	\ pc xt -- info | 0 ; finds xref packet corresponding to PC
\ *G Given the current PC and the XT of the word the PC is in,
\ ** *\fo{FindXrefInfo} returns a pointer to an XREF packet if the
\ ** PC is at an exact compilation boundary, otherwise it returns
\ ** zero.
  ^xref @ 0=			\ check that info is available
  if  2drop 0  exit  endif
  dup >xref @ 0=
  if  2drop 0  exit  endif
  >xref @ +dcc-head		\ head of decompilation chain
  begin
    @(x) ?dup
   while			\ -- pc addr(x)
    2dup >xtHere(x) @(x) =
    if  nip  exit  endif
    >link(d)
  repeat
  drop 0
;

: FindXrefNearest	\ pc xt -- info|0
\ *G Given the current PC and the XT of the word the PC is in,
\ ** *\fo{FindXrefNearest} returns a pointer to the Xref packet
\ ** for the address at or less than the PC. If no Xref
\ ** information is available for the word, zero is returned.
  ^xref @ 0=			\ check that info is available
  if  2drop 0  exit  endif
  dup >xref @ 0=
  if  2drop 0  exit  endif
  >xref @ +dcc-head @(x) tuck	\ head of decompilation chain
  begin				\ -- ref(x) pc ref(x)
    dup
   while			\ -- refp(x) pc ref(x)
    2dup >xtHere(x) @(x) u<
    if  2drop  exit  endif
    rot drop			\ -- refp(x) pc ref(x)
    tuck >link(d) @(x)		\ -- ref(x) pc refn(x)
  repeat
  2drop
;

: GetXrefPos	\ info -- startpos len line addr
( *G Given a pointer to an XREF packet, *\fo{GetXrefPos} returns the )
( ** position, name length, line number of the source text in the source file, and )
( ** the value of *\fo{HERE} at the time of compilation. )
  dup >r  >in(x) @(x)
  r@ >type(x) @(x) if			\ nz for literals
    1					\   don't know length, assume 1
  else
    r@ >xtAdded(x) @(x) >name c@	\   get xt, extract name length
  endif
  tuck - 0 max swap			\ -- start len ; clip start
  r@ >line#(x) @(x)  r> >xtHere(x) @(x)	\ -- start len line here
;

: NextXref	\ info1 -- info2
( *G Steps to the next info packet, given the offset of the previous. )
  >link(d) @(x)
;


\ ***************
\ Display support
\ ***************

[undefined] .banner [if]
: .stars	\ n -- ; SFP010
\ Display n asterisks.
  0 ?do  ." *"  loop
;

: .banner	\ caddr len -- ; SFP010
\ Display a bannered line with asterisks above and below.
  tuck dup cr cr .stars  cr type  cr .stars
;
[then]


\ *********************************************
\ USES <name> cross references a specific word
\ XREF-ALL cross references all target words
\ XREF-UNUSED finds all unused words
\ *********************************************

: ?xref-enabled \ -- ; exits caller if XREFfing not turned on
  ^xref @ 0= if
    cr
    cr ." Use +XREFS before this to turn on XREF system"
    cr
    r> drop                             \ exits CALLER
  endif
;

: xref          \ -- ; XREF <name>
( *G Use in the form XREF <name> to display where <name> is used. )
  ?xref-enabled
  get-token find ?undef                 \ error if symbol not found
  .xrefs
;

: uses          \ -- ; synonym for XREF
\ *G A synonym for *\fo{XREF} above.
  xref
;

: .xrefs(1)     \ thread nfa -- flag ; no display if unresolved
  nip name> .xrefs true
;

: xref-all      \ -- ; cross reference all words
( *G Produces a cross reference listing of all the words with )
( ** cross reference information. This information is often too )
( ** long to be directly useful, but can be pasted from the )
( ** console to an editor for sorting, printing, and other )
( ** post-processing. )
  ?xref-enabled
\  cr
\  cr ." **********************************************************"
\  cr ." Cross Reference listing of all words with XREF information"
\  cr ." **********************************************************"
  s" Cross Reference listing of all words with XREF information" .banner
  c/l out !
  ['] .xrefs(1) ( ProcessAllVocs ) WalkAllWords
  cr
;

: .xrefs-unused	\ thread nfa -- flag ; display word if unused ; SFP010
  nip					\ discard thread
  dup 0 @(x) u> if			\ name is in this XREF area
    dup name> >xref @ ?dup 0= if	\ XREF field is zero if word is unused
      tab .name  true exit
    else
      +xrc-link @(x) 0= if
        tab .name  true exit
      endif
    endif
  endif
  drop true
;

: xref-unused   \ -- ; cross reference all words
( *G Produces a cross reference listing of all the unused words with )
( ** cross reference information. This information is often too )
( ** long to be directly useful, but can be pasted from the )
( ** console to an editor for sorting, printing, and other )
( ** post-processing. )
  ?xref-enabled
\  cr
\  cr ." *******************************************************"
\  cr ." The following words are unused inside colon definitions"
\  cr ." *******************************************************"
  s" The following words are unused inside colon definitions" .banner
  c/l out !
  ['] .xrefs-unused ( ProcessAllVocs ) WalkAllWords
  cr
;

\ SFP016...
: ttx-set	\ xt -- ; xt TTX-SET "<text>"
\ *G The quoted string is saved as the tooltip text for the
\ ** word whose *\i{xt} is given, e.g.
\ *C   ' dup ttx-set "x -- x x ; duplicate top item on stack"
  ?xref-enabled  >xref-head
  here(x)  "xxx" text,(x)	\ -- head txt
  swap +xttip !(x)
;

: ttx-get	\ xt -- caddr len
\ *G Given an *\i{xt}, return the tooltip text for the word.
  ?xref-enabled
  >xref-head +xttip @(x) dup if		\ text address in x format
    ^xref @ + count			\ return real address
  else
    drop  s" No tooltip text available"	\ return warning text
  endif
;

: ttx?		\ xt -- flag
\ *G Return true if the word whose *\i{xt} is given has a tooltip.
  ?xref-enabled
  >xref-head +xttip @(x) 0<>		\ text address in x format
;
\ ...SFP016


\ *******
\ Exports
\ *******

export XREF
export USES
export XREF-ALL
export XREF-UNUSED
export +XREFS
export -XREFS
export INIT-XREF
export TERM-XREF
export XREFS?
export XREF-KB
export XREF-REPORT
export XREF:
export SAVE-XREF
export LOAD-XREF
export (SHOW)
export $SHOW
export SHOW

export hasXref?
export hasXDecomp?
export FindXrefInfo
export FindXrefNearest
export GetXrefPos
export NextXref
export dump(x)

export WalkXref
export WalkDecomp

export ttx-set
export ttx-get
export ttx?

END-MODULE

+xrefs
((
\ XREF test code
: t over + swap ?do  i 10 mod 0= if cr endif i . loop ;
: t2  10 20 30 40, 50 60,  ;
))

\ ======
\ *> ###
\ ======


((
Extending XREF to handle overlays
Design notes
============

/HEAD(X)
--------
struct /head(x) \ -- n ; word header
  cell field +xrc-link          \ xref chain link
  cell field +dcc-head          \ first entry in decompilation chain
  cell field +dcc-tail          \ last entry in decompilation chain
end-struct

This structure is pointed to by the XREFINFO field in the word's
name field. The pointer is a relative offset from the start of the
XREF data area. This structure MUST be maintained in the XREF data
for the kernel/overlay containg the word.

/REF(X)
-------
struct /ref(x)
  cell field >link(x)           \ xref link field
\ may be in a different overlay
  cell field >sym(x)            \ symbol (xt) in which this reference occurs
\ in this overlay
  cell field >line#(x)          \ line# in file being compiled
  cell field >in(x)             \ offset in line
  cell field >xthere(x)         \ HERE when reference was made
\ in this overlay
  cell field >type(x)           \ 0 for xts, 1 for literals, 2 doubles,
  				\ -2 for floats on FP stack, -1 for OGL floats on data stack
  cell field >link(d)           \ decompilation link
\ in this overlay
  3 cells field >xtadded(x)     \ xt being added or literal value (1, 2, or 3 cells)
\ may be in a different overlay
end-struct

This structure is laid down during compilation of a source token
and hence must remain with the kernel/overlay being compiled.

The >LINK(X) and >XTADDED(X) fields will need to be extended with an
owner field. The overlay system uses the following notation:
  0     kernel
  -1    this overlay
  addr  base of loaded overlay
We prefer not to use the base address of the overlay as this changes
with system memory usage. Thus overlays need a unique identifier.
  0     kernel
  -1    this overlay
  n     overlay identifier

We can't use any build address as the build address for two overlays
may be the same. We can't use a variable in the kernel as this would mean
saving the kernel again after building any overlay. This pretty much
leaves using a private INI file or registry item to hold this
identifier. This identifier will be an extension to the overlay
header information defined in KERNEL.FTH.

When handling the >LINK(X) field, a value of 0 will be used for the
first reference to another overlay. When 0 is encountered, the
XTADDED(X) field and associated identifier enable the /HEAD(X)
structure for the word to be accessed.

The XTADDED(X) field will be handled as follows:
  kernel     absolute,  ident = 0
  this ovl   relative,  ident = -1
  other ovl  relative,  ident = that overlay's unique id

XREF file names
---------------
The output file name with .XRF added.

If XREFs are enabled during building, the XREF file will be constructed
and saved by MAKEOVERLAY.

If XREFs are enabled at load time, the XREF file will be loaded.

Overlays thus need an XREF indicator and an XREF address field.


Gotchas
=======
The XREF construction words need to be overlay aware. In particular,
when an overlay is being compiled, XREF information must only be
generated for one build. Facilities already exist in the kernel for
this.

The XREF words must also be able to identify not only a preloaded
overlay but also an xt in the overlay being compiled. The VALUEs
INITIAL-DP and SECOND-DP provide this information and at least one
of them will have to be moved to the kernel.

Kernel Changes
==============
Compile time variables so that XREF is overlay aware
  INITIAL-DP and/or SECOND-DP
  Modify /OVI structure

Overlay changes
===============
Unique identifier mechanism.

Identifier and XREF pointer in /OVI structure. N.B. defined in KERNEL.FTH
))

