\	P R I N T   F I L E S  --  original by Greg Guerin

\ This utility prints text files in a pleasant format
\	- output is paginated, with header on each page
\	- tabs are expanded to spaces (arbitrary tab-stop settings)
\	- form-feeds are properly handled in the input stream
\	- changeable paper width (affects formatting)
\	- long input lines are wrapped on output
\	- control characters are made visible, except CR LF HT FF
\   - margins that can be set in inches
\	- Total page count available in header
\
\ To print a file:
\	PrintFile  ( namePtr -- )
\				prints the named file.  Set VolRef# or ioDir first, and
\				you can print files from other folders.
\
\ Margins are set with 5 variables: TopMargin, BottomMargin, 
\ LeftMargin, RightMargin, and HeaderMargin.  Margins are in points,
\ but may be set with inches using the word INCHES that converts
\ from inches to points taking DPL into account.  For example:
\	1.25 inches LeftMargin !
\
\ Other variables
\	PaperFontNum	defaults to Courier = 22, Monaco = 4
\	PaperFontSize	defaults to 10, 9 is good for Monaco
\	PaperQuit		set ON to stop gracefully.  Usually set by a
\					foreground task when (Paper) is in background.
\	PaperHold		set ON to temporarily hold printing.  Usually set by a
\					foreground task when (Paper) is in background, to do
\					things like change ribbons in the middle of a print job.
\	Lines/Page		system variable with number of printed lines per page
\					may be overridden according to margin settings.
\	ShowNumPages	When ON, the total number of pages will be counted
\					before printing and displayed in the header as
\					"Page ## of N" where N is the total.  Default is ON.

\ ***************************************************************************
\ (c) Copyright 2012  by MegaWolf, Inc.
\ All Rights Reserved. 
\ Permission is hereby granted to VFX Forth for Mac customers 
\   to use this source code in their own applications.
\ Sale, Distribution or other use of this source code is prohibited. 
\ ***************************************************************************
\ Last Revision: 11 May 2012  11:32:41  ward@megawolf.com



Only Forth definitions

>M4TH
decimal


vocabulary PaperVocab

Lacking GetFontInfo   Extern: void GetFontInfo ( UInt32 &fInfo  );
Lacking RealFont      Extern: Uint32 Flag RealFont ( UInt32 fontNum, UInt32 size  );
(( 

Extern: void GetFNum ( UInt32 * theName$, UInt32 * ^theID  );

: GET.FONT.NUMBER ( name$ -- id# )
	0 >R rp@			\ scratch area for id# on return stack
	GetFNum
	r> 16 rshift ;
 ))



Lacking POWER10 : POWER10  ( n -- 10^n ) 0 max 1 swap 0 ?DO 10 * LOOP ;

72 CONSTANT POINTS/INCH

: INCHES  ( inchesDouble -- points )
	D>S dpl @ power10 points/inch swap */ ;

variable TopMargin		1.0 inches TopMargin !
variable BottomMargin	1.0 inches BottomMargin !
variable LeftMargin		1.0 inches LeftMargin !
variable RightMargin	1.0 inches RightMargin !
variable HeaderMargin	0.4 inches HeaderMargin !

variable PaperFontNum	22 PaperFontNum !	\ Courier
variable PaperFontSize	10 PaperFontSize !

variable ShowNumPages   ShowNumPages ON		\ slows printing of large documents
variable ShowTime	    ShowTime OFF



also PrintVocab
ALSO PaperVocab Definitions


\ Lowest-level line construction

VARIABLE Column   Column off
  CREATE TabMarks
	z", :...:...:...:...:...:...:...:...:...:...:...:...:...:...:...:...:...:...:...:...:...:"
  TabMarks zstrlen  CONSTANT TabEnd	\ length of TabMarks


ascii :  CONSTANT IsTab				\ marker for tab-stop
ascii >  CONSTANT IsLeftMargin		\ marker for left margin (unused)

: NextColumn  ( -- )	\ advance to next column
	1 Column +!  ;

: +Column  ( n -- n+column )	\ advance n by column offset
	Column @ +  ;

\ --------------------------------------------
\ Words to control actual printing on the page

  ascii \ CONSTANT BreakMark		\ marks end of broken lines
  CREATE  FormFeed  12 c,			\ ASCII FF sequence



variable StartX
variable StartY
variable HeaderStartY
variable Max.Lines/Page		\ calculated from margins and line height
variable Max.Chars/Line		\ calculated from margins and char width
variable PaperLine#			\ broken lines count as one line, like in Editor
variable NumPages
variable CountingOnly

  200 CONSTANT #LineBuf				\ length of both buffers


0 0 0 0 rect PaperRect
0 0 0 0 rect PageRect


: GetPrinterRectangles ( --  )
	aPMPageFormat @  aPMRect PMGetAdjustedPageRect (?PrintError)
	aPMRect +PMRectTop    df@ f>d drop
	aPMRect +PMRectLeft   df@ f>d drop
	aPMRect +PMRectBottom df@ f>d drop  
	aPMRect +PMRectRight  df@ f>d drop
	PageRect !rect

	aPMPageFormat @  aPMRect PMGetAdjustedPaperRect (?PrintError)
	aPMRect +PMRectTop    df@ f>d drop
	aPMRect +PMRectLeft   df@ f>d drop
	aPMRect +PMRectBottom df@ f>d drop  
	aPMRect +PMRectRight  df@ f>d drop
	PaperRect !rect
;


variable PaperLineHeight		12 PaperLineHeight !

: FigureMargins  ( -- )
	get.textsize  get.textfont
		PaperFontSize @ _TextSize PaperFontNum @ textfont
		pad GetFontInfo
		pad w@ pad 2 + w@ + pad 6 + w@s +
			PaperLineHeight !  \ line height = ascent + descent + leading
		ascii A charWidth >R	\ save for max.chars/line computation.
								\ can't use the maxWidth field since some
								\ special chars are wider than alphanumeric
	textfont  _TextSize
	GetPrinterRectangles

	PaperRect     w@s TopMargin    @ + PageRect     w@s max  StartY !
	PaperRect 4 + w@s BottomMargin @ - PageRect 4 + w@s min  StartY @
		- PaperLineHeight @ /  Max.Lines/Page !
	PaperRect 2 + w@s LeftMargin   @ + PageRect 2 + w@s max  StartX !
	PaperRect 6 + w@s RightMargin  @ - PageRect 6 + w@s min  StartX @
		- R> /
		#LineBuf min  Max.Chars/Line !
	PaperRect     w@s HeaderMargin @ + PageRect     w@s max  HeaderStartY !
;


VARIABLE PageBreak   PageBreak off		\ flag indicates page end
VARIABLE BrokenLine  BrokenLine off		\ flags a wrapped line


DEFER ToPrinter  ( addr\count -- )  	\ so it can be vectored
\ ' mfTYPE IS ToPrinter
' .TYPE IS ToPrinter			\ set default behavior


: PrintBlanks  ( n -- )	\ print n blanks on printer
	\ more than about 64 will overwrite into PAD
	0 max ?dup
	IF		\ only positive non-zero values
		here swap		\ use HERE cuz PAD may be in use
		 2dup blank		\ make 'em
		  ToPrinter		\ print 'em
	THEN  ;


: LeftJust  ( addr\count -- )	\ print left-justIFied
	-trailing ToPrinter  		\ pretty simple...
;


: RightJust  ( addr\count -- )	\ print right-justIFied
	-trailing					\ remove trailing blanks
	Max.Chars/Line @  over -	\ padding is LineWidth - textSize
	 PrintBlanks				\ padding forces text to right
	ToPrinter   				\ where it is printed
;

: NewLine  ( -- )		\ advance paper to next line
	CountingOnly @ 0=
	IF
		@pen PaperLineHeight @ + nip StartX @ 
		swap MoveTo	\ move down one line and to left margin
	THEN
	1 line# +! ;


: PrintNewLine  ( addr\count -- )	\ print with leading indent
									\ and advance paper to next line
	CountingOnly @ 0=
	IF
		BrokenLine @ 				\ print buffer accordingly...
		IF
			RightJust
		ELSE
			LeftJust
		THEN
	ELSE
		2drop
	THEN
	NewLine  ;		\ flush printer buffer, move to next line

\ --------------------------------------
\ Control print-line construction buffer


CREATE LineBuild  #LineBuf allot	\ buffer holds line under construction

: ClearBuffer  ( -- )	\ clears line-building buffer to blanks
	LineBuild  #LineBuf blank
	Column off  ;

: PrintLine  ( -- )	\ print & clear LineBuild buffer, advance line
	LineBuild  Column @  PrintNewLine	\ print the whole buffer
	ClearBuffer  ;					\ reflect new buffer state

: CountLine  ( -- )	\ control line counter and page breaks
	\ Line# is automatically incremented in Stream.Print
	Line# @      Max.Lines/Page @ >= 
	PaperLine# @ lines/page @ >=
	or IF	\ force page break
		Line# OFF
		PaperLine# @ lines/page @ - PaperLine# ! \ allow catch-up on future pages
		PageBreak ON
	THEN  ;

: EndOfLine  ( -- )	\ print line and advance paper
	PrintLine		\ print line, advance paper, clear buffer
	CountLine			\ control page-breaks
	BrokenLine OFF  ;	\ assume end of previous line

\ ------------------------------------------------
\ Handle default printing on page, i.e. the header

VARIABLE HeadCount   HeadCount OFF	\ length of header
  CREATE HeadBuffer  #LineBuf allot


\ Centering algorithm assumes mono-spaced printer font
: CenterName  ( file$ -- )	\ center counted file$ in HeadBuffer
	locals|  THEName  |
	HeadCount @ 2/		\ middle of header
	 THEName c@ 2/		\ 1/2 of name
	  -					\ before mid-header
	HeadBuffer +		\ place where name goes
	 THEName 1+ swap	\ from THEName into HeadBuffer
	  THEName c@		\ how many to copy
	   cmove  ;			\ copy name into HeadBuffer



: PageNumber  ( page# -- )	\ put page# into header buffer
	S>D <# #s #>					\ addr\count of number string
	HeadBuffer Max.Chars/Line @ + 4	\ addr\count\^endbuffer\4
	ShowNumPages @ IF 6 + THEN -	\ adjust for Page Of header
	over 1 = IF 2 + THEN			\ adjust for single digit
	over 2 = IF 1+ THEN				\ adjust for double digit 
	swap cmove ;


also system
: BuildHeader  ( file$ -- )	\ build unchanging header parts
	HeadBuffer  #LineBuf blank
	ShowNumPages @ IF	\ put appropriate page # message on right side
		C" Page     of " count   HeadBuffer Max.Chars/Line @ + 15 -  swap cmove
		NumPages @ S>D <# #s #> HeadBuffer Max.Chars/Line @ +  3 -  swap cmove
	ELSE
		C" Page  #    " count HeadBuffer Max.Chars/Line @ + 12 - swap cmove
	THEN
	HeadBuffer GetDate$		\ Get.Date$			 		\ put data on left
	ShowTime @
	IF
		HeadBuffer 9 + 	GetTime$	\ Get.Time$
	THEN
	HeadBuffer Max.Chars/Line @ -trailing	\ ignore trailing blanks
	ShowNumPages @ 0= IF 4 + THEN			\ 
	HeadCount !  drop						\ only save length of header
	CenterName  ;							\ put name in center
previous


: PrintHeader  ( page# -- )	\ print header for current page
	\ does not affect counts or LineBuild buffer
	StartX @ HeaderStartY @ MoveTo 
	NewLine
	NewLine
	PageNumber				\ stick in the page number
	HeadBuffer HeadCount @
	 PrintNewLine			\ 1st line has header
	NewLine 				\ 2nd line is blank
	line# off ; 			\ header lines don't count


\ ----------------------------------------------------------
\ How characters appear on output page

\ How printable characters affect the print buffer...

: LastBL  ( -- offset )	\ offset = where to break line
	Max.Chars/Line @
	dup   16 -  over do		\ descending loop
		LineBuild I + c@		\ looking for blanks
		BL = IF
			drop I leave	\ found one, remember where
		THEN
	 -1 +loop  ;			\ IF none found, use LineWidth


: LineBreak  ( -- )	\ break LineBuild at Column
	LastBL
	 Column @  over -	\ how many chars not printed from end
	locals|  break# breakAt  |
	LineBuild breakAt +
	 pad
	  break#
	   cmove		\ copy from buffer to PAD
	BreakMark 
	 LineBuild breakAt +  c!	\ mark end of broken line
	breakAt 1+ Column !			\ new length of buffer
	EndOfLine					\ print, clear, and advance

	pad			\ copy back from PAD to buffer
	 LineBuild
	  break#
	   cmove
	break#  Column !	\ track chars copied in
	BrokenLine on  ;	\ remember what happened


: LineWrap  ( -- )	\ wrap long input lines
	Column @  Max.Chars/Line @  > IF
		LineBreak
	THEN  ;


: StashByte  ( char -- )	\ put character into print buffer
	LineBuild +Column c!	\ stash the byte
	NextColumn				\ advance to next free slot
	LineWrap  ;				\ handle linewrap IF necessary

\ ----------------------------------------------------------
\ How control characters affect the print buffer and page...

\ dup set-current				\ back to FORTH
also m4th definitions

  VARIABLE PaperQuit   PaperQuit OFF	\ begin: "not a quitter"
  VARIABLE PaperHold   PaperHold OFF	\ begin: not holding

: HoldIt  ( addr -- )  \ hold as long as flag is set
	begin
		dup @	\ until flag becomes 0..
	 while
		pause	\ ..just let other tasks run
	repeat
	drop  ;


: HoldingPattern  ( -- )
	PaperHold Holdit  ;



previous Definitions

 
: (PaperEvents)  ( -- )  \ respond to only some events
	CheckEventQueueForUserCancel	\ 
	IF
		PaperQuit ON	\ allow cancelling during printing
	THEN
	Events @ dup
	$0005024E	\ allow mouse, key, update, abort, command-key, menubar  
	and  Events !
	do.events drop
	Events !  ;


: PaperEvents  ( -- )  	\ speed this up a LOT by giving less time to OS
	Event.Record $0005024E 
	?event  ( no sleeptime - much faster than waitnextevent )
	IF
		(PaperEvents)
	THEN
;

: MaybeQuit  ( -- )	\ quit IF requested
	PaperEvents
	PaperQuit @					\ maybe another task set it..
	 abort" Now quitting..."  ;	\ caught by error handler...


: =CR  ( char -- )	\ force line-end
	MaybeQuit
	1 PaperLine# +!
	drop  
	EndOfLine  ;

: =FF  ( char -- )	\ force line and page-end
	drop  EndOfLine	
	PageBreak ON  ;	\ action deferred into MakePage loop

: =HT  ( char -- )	\ advance to next tab-stop
	drop
	Column @  
	  TabEnd  4 -	\ watch for end !!
	  < IF
		begin
		   \ assumes a tab-stop is last char in tab-marks
		   NextColumn		\ look at next place
		   TabMarks +Column \ corresponding char in tab-marks
		    c@ IsTab =
		until
	else
		BL StashByte	\ beyond known tab-stops
	THEN  ;


: As^  ( char -- )	\ print letter with leading ^
	ascii ^  StashByte
	31 and  ascii @ +  StashByte  ;

: ControlChar  ( char -- )
	dup 
	CASE
		 0 OF drop	ENDOF
		 9 OF =HT	ENDOF
\		10 OF drop	ENDOF
		10 OF =CR	ENDOF
		11 OF drop	ENDOF
		12 OF =ff	ENDOF
		13 OF =CR	ENDOF
		( OTHERWISE) swap As^
	ENDCASE
;
	

\ ----------------------------------------------------------
\ Get byte from file; put byte to print buffer

: PutByte  ( char -- )	\ perform action appropriate to character
	dup  BL <
	IF		ControlChar		\ take action or print it
	ELSE	StashByte		\ always print it
	THEN  ;


VARIABLE FileEmpty	FileEmpty off	\ flag marks EOF


\ necessarily handles EOF on input file,
\ returns FALSE and no char IF EOF, also flagging PageBreak

variable fileBuffer		\ handle containing file contents
	0 fileBuffer !
variable bufferOffset	bufferOffset OFF
variable bufferSize		bufferSize   OFF

create CharBuffer 		0 ,

: GetByte  ( -- [char]\flag )		\ read byte from input file
	HoldingPattern					\ hold IF flag is set
	bufferOffset @ bufferSize @ < dup 
	IF
		fileBuffer @@ bufferOffset @ + c@
		swap
		1 bufferOffset +!
	ELSE
		FileEmpty ON				\ informs other routines
		PageBreak ON				\ forces closing pagebreak
	THEN ;

\ ----------------------------------------------------------
\ Highest level routines print pages from a file


CREATE PrintingCounts
	0 ,  0 ,	\ a 2VARIABLE
	
: SetupPagePort  ( -- )
	PaperFontSize 	@ _TextSize
	PaperLineHeight	@ get.xform.window +wline.height !		\ line.height
	PaperFontNum	@ textfont
	NewLine ;


: MakePage ( page# -- )  \ print complete page
	Line# OFF           		\ every page starts a new line-counter..
	CountingOnly @ 0=
	IF
		OPEN.PAGE					\ 
		SetupPagePort
		PrintHeader					\  ..and has a header
		StartX @ StartY @ MoveTo 
		NewLine
	ELSE
		PrintHeader
		NewLine
	THEN
	BEGIN
         GetByte                	\ from input file
         1 PrintingCounts +!		\ count this byte
         IF  PutByte  THEN			\ to print stream
         PageBreak @
	UNTIL             				\ till page-break
	EndOfLine                     	\ IF EOF before CR, print the last line
	CountingOnly @ 0=
	IF
		CLOSE.PAGE
	THEN
	PageBreak OFF
	MaybeQuit
;



: PaperDone  ( -- )  \ close down papering activity - mod 7/89  JWB
    0 0 PrintingCounts 2!
\    OwnPrint RELEASE
;         \ allow others to use printer



: (PrintText)  ( name$\dialog\textHandle -- )  \ print contents of handle
	3 needed
	0 0 
	{ FILE$ DIALOG? textHandle LAST FIRST }

	textHandle fileBuffer !
	
	PaperFontNum @ 0=	\ if not specified or available, use MONACO
	PaperFontNum @ PaperFontSize @ RealFont 0=
	OR
	IF 					\ this also prevents printing with font 0 (System Font)
		4 PaperFontNum !
	THEN

	PaperQuit OFF		\ don't quit yet...
	FileEmpty OFF   PageBreak OFF   BrokenLine OFF
	Line# OFF
	PrintingCounts OFF	\ the running count is cleared
	ClearBuffer
\	OwnPrint GET		\ get exclusive access to printer



	textHandle GetHandleSize dup bufferSize !
	PrintingCounts 4 + !	\ overall size of file

	file$ -> JobName$

	printing.prep
	discard.activates
	-1 0 0 SetCancelDialog   \ default cancel dialog
	dialog?
	IF
		PrJobDialog			 \ job dialog only for parent
	ELSE
		SetUpPrJob	true 	\ hPrint PrValidate drop 
	THEN

	aPMPrintSettings @  0 >R rp@  PMGetFirstPage  drop  R>  -> first
	aPMPrintSettings @  0 >R rp@  PMGetLastPage   drop  R>  -> last

	FigureMargins

	HeadBuffer  #LineBuf blank
	NumPages OFF
	column off
	bufferOffset OFF
	PaperLine# off
 
	\ first just count pages
	dup 					\ was job dialog canceled?
\	ShowNumPages @ and		\ this is fast enough now, we don't need or want to skip
	IF
		CountingOnly ON

		1							\ NotCancelled?\startpage#
		BEGIN
			dup  MakePage		\ print one page
			1+					\ up the page number
	 		\ dup last >		\ count ALL pages
			FileEmpty @ 	
			\ OR				\
		UNTIL					\ until file ends
		1- NumPages !			\ total# pages
	THEN
	bufferOffset OFF
	FileEmpty OFF
	Line# OFF
	PrintingCounts OFF		 \  the running count is cleared

	IF	\ job dialog wasn't canceled
		CountingOnly ON

		PaperLine# off 
		column off
		['] 2drop is ToPrinter
		file$ BuildHeader		\ build unchanging part of header
		OPEN.DOCUMENT			\ 
		1 
		BEGIN
			dup first >=
			IF
				['] .TYPE IS ToPrinter
				CountingOnly OFF
			ELSE
				['] 2drop IS ToPrinter
			THEN

			dup MakePage	\ print one page
			1+				\ up the page number
		 	dup last >
		 	FileEmpty @ or
		UNTIL	\ until file ends or all requested pages done
		drop				\ ..the page number
	THEN
	Printing.Done	\ calls CLOSE.DOCUMENT
	PaperDone
	0 fileBuffer !
;

also m4th definitions

: PrintText  ( name$\dialog\textHandle -- )
	['] (printtext) catch
	IF
		TerminatePrinting
		PaperDone
		true abort" Printing Aborted"
	THEN
;


previous definitions


variable textBuffer
VARIABLE Doing		Doing off		\ file to read from

: ?FileAbort  ( -- )  \ aborts IF IO-Result < 0
	IO-Result @  0< abort" File error"  ;


: (PrintFile)  ( name$\dialog -- )  \ print entire file
	0 0 locals| FILESIZE REFNUM DIALOG? FILE$ |
	file$  RO  open$		\ open file read-only
	?FileAbort				\ did we succeed?
	Doing !					\ IF so, save open file ID
	Doing @ FileSize? -> filesize
	filesize NewHandle textBuffer !
	MemError MemError?
	textBuffer @ HLock 
	textBuffer @@ filesize Doing @ FileREAD drop ?file.error
\	textBuffer @ HUnlock		\ leave it locked while printing
	Doing @ fileclose
	Doing OFF

	file$ dialog? textBuffer @ printText
	textBuffer @ HUnlock
	textBuffer @ DisposeHandle
	textBuffer OFF
;



>m4th also papervocab

: PrintFile  ( file$ -- )
	true  Busy		\ change cursor to indicate what's up
	true  ['] (PrintFile) catch
	false Busy
	IF
		Doing @  ?dup IF fileclose Doing OFF THEN
		textBuffer @ handle? IF DisposeHandle 0 textBuffer ! THEN
		true abort" Aborted"
	THEN
;


CREATE Print.Types  ascii TEXT ,

: PAPER.AEPRINTFILE  ( vrefnum\dirID\fileName -- )
	locals| fileName dirID vRefNum |
	vRefNum dirID fileName 0 get.cat.info ?dup
	IF
		dup @ 32 + @ 	\ get file type
		swap DisposeHandle
		Print.Types @ =
		IF  fileName vRefNum volref# ! dirID iodir ! 
			PrintFile
			VolDirExpires
		THEN
	THEN ;

' Paper.AEPrintFile IS AEPrintFile


: Remove.Paper  ( loAddr\hiAddr -- )
	['] 3drop IS AEPrintFile ;
On.Forget Remove.Paper


: Paper ( file-name ( -- )		\ expects VolRef# and ioDir to be set correctly
	bl word  PrintFile  ;



: PaperDialog  ( -- )  \ run file-selector dialog
 	ClearReply
\	0 1 Print.Types	reply
\	StandardGetFile  ( fileFilter\numTypes\&typeList\&reply -- )
	0 0 0 NRRecord			\ all file types
	0  c" Print File"  c" PRINT"  0  c" Select a text file for Printing"
	aNavEventProcPtr
	(NavGetFile) drop
	discard.activates  ;	


: Print...  ( -- )
	PaperDialog			\ information is returned in Reply
	NavReplyOK? IF							\ didn't Cancel
		NavReplyVolume  VolRef# !
		NavReplyDir 	 iodir   !
\		NavReplyName  pocket 64 cmove		\ put whole name in POCKET
\		pocket PrintFile
		NavReplyName PrintFile
		VolDirExpires
	THEN
;


\ : Paper" ( file-name" ( -- )	\ expects VolRef# and ioDir to be set correctly
\	ascii " ?word  PrintFile  ;

