\ ExternSysV64.fth - DLL and shared library access for VFX Forth 64 bit

((
Copyright (c) 2001-2008, 2009, 2010, 2011, 2012, 2013, 2014, 2017, 2018, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20200817 SFP030 Reworked FP interface. XMM8 is not cached inside
		EXTERNs. XMM regs are not preserved in OS calls.
20200810 RT_029 Redefined: returnFLOAT parse-params setReturn
                           setArgs FUNCTION: ASCALL:
                Added: returnFP16 param? doubles? make-return
		       parseArgs setDouble set2Doubles
20200722 SFP028 Improved initialisation.
20200620 SFP027 Made default argument size 8.
20200514 SFP026 Converted to SystemV x64 ABI
20181029 SFP025 Converted to Win64.
20170223 JD_024 Added more Windows types.
                Added _Reserved_, [in] for Windows.
20140521 SFP023 Enhanced parsing of *** types.
20131119 SFP022	Made LOCATE go to first line of EXTERN:
20130520 SFP021 Exposed INIT-LIB. Added CLEAR-LIB.
2012xxyy SFP020 Added compatibility words for other Forths.
20120418 SFP019 Added FRAMEWORK and FRAMEWORK: for OSX.
		Allowed library and framework redefinitions
		to be ignored.
20120411 SFP018 Prevented inlining of words containing EXTERNs.
20120312 SFP017 Added EXTERNVAR and ALIASEDEXTERNVAR.
20111104 SFP016 More testing for tabs.
20111021 SFP016 Added #BADEXTERNS #BADLIBS and .BADLIBS.
20111004 SFP015 Added PLONG to Windows types.
20110912 SFP014 Added library mode options.
20110427 SFP013 Added additional pointer parsing.
                Added __IN_OPT, __INOUT_OPT and __OUT_OPT for Windows.
20110415 SFP012 Supports C // and /*...*/ comments in argument
                parsing.
                Added __IN, __INOUT and __OUT for Windows.
                Added ** and ***.
20100223 SFP011 Added protection mechanism to Externs.
20100222 SFP010 Moved PREEXTERN and POSTEXTERN to Forth voc.
20091123 SFP009	Refactored with startup.fth.
20090610 SFP008 Added documentation for FLAG from Marcel Hendrix.
20090609 SFP007 Added .BADEXTERNS.
20090502 SFP006	Added code for Windows protection with floats.
20020323 MPE005 Refactored client specific code in prologue and
                epilogue.
20090319 MPE004 Modified after distribution list discussions.
20090310 MPE003 Refactored.
20090131 MPE002 Tightened up.
		Put most words in separate vocabulary.
20090121 MPE001 Ported from the Windows/Linux version.
))


\ =========
\ *! extern
\ *T Functions in shared libraries
\ =========

((
Licence
=======
This code can be used as the basis of an implementation of the
EXTERN: interface for other Forth systems. Copyright of the source
notation and behaviour is retained by MicroProcessor Engineering Ltd.
until another copyright has been assigned to a third party by
MicroProcessor Engineering Ltd..

Controlled words can be identified by comments starting
  \ *G
below the name line.

If you implement the notation and behaviour and release the source
code, this licence must be included in the source code.

Rationale and Requests
======================
The rationale for releasing this code is to encourage portability
of Forth source code that accesses shared libraries.

The point of retaining copyright at MPE is not to prevent you
changing and improving the system, it is to provide a central
point of management for the notation and its behaviour. In the
long term, copyright of the EXTERN: notation will be assigned
to a third party.

If your code is better, faster, shorter or more functional, we
request that you contribute it back to the maintainer, currently
Stephen Pelc, stephen@mpeforth.com. Similarly ports for currently
unsupported Forth systems will be gratefully received.

VFXisms
=======
DONOTSIN can be removed or be a NOOP.

.LWORD ( l -- ) and .DWORD ( l -- )
Display l as an unsigned 32 bit hex number.

.XWORD ( x -- )
Display x as an unsigned 64 bit hex number.

: -leading      bl skip  ;
: -white        -leading -trailing  ;

The code is based around three words which can be implemented
for any Windows or Unix-style host.
  LoadLibrary      \ zaddr -- handle|0
  FreeLibrary      \ handle -- status
  GetProcAddress   \ handle zaddr -- addr|0


Legacy code
===========
The words L>R and R>L are for compiling library code from other
Forth systems which may use a right-to-left stack order.

Parameters for 64 bit
=====================
left to right, left top/first of C stack, right top of Forth stack
6 integer: RDI, RSI, RDX, RCX, R8, R9
left to right, left top/first of C stack, right top of Forth float stack
8 float  : XMM0 ... XMM7
))

only forth definitions decimal

\ vocabulary externals
\ Where library and imported function definitions live.
vocabulary types	\ --
\ Where the "C" style types go for the *\fo{EXTERN:} notation.
vocabulary Extern.Voc
\ Where the guts of the EXTERN: mechanism live.

also Extern.Voc definitions


\ ***************
\ *S Introduction
\ ***************
\ *P VFX Forth supports calling external API calls in dynamic link
\ ** libraries (DLLs) for Windows and shared libraries in MacOS, Linux and
\ ** other Unix-derived operating systems.
\ ** Various API libraries export functions in a variety of methods
\ ** mostly transparent to programmers in languages such as C,
\ ** Pascal and Fortran. Floating point data is supported for use
\ ** with *\i{Lib\x64\FPSSE64S.fth}.

\ *P Before a library function can be used, the library itself must
\ ** be declared, e.g.
\ *E LIBRARY: /usr/lib/libSystem.B.dylib

\ *P The default calling convention is nearly always applicable.
\ ** The following example shows that definitions can occupy more
\ ** than one line. It also indicates that some token separation may
\ ** be necessary for pointers:
\ *E Library: libc.so.6
\ **
\ ** Extern: int execve(
\ **   const char * path,
\ **   char * const argv[],
\ **   char * const envp[]
\ ** );
\ *P This produces a Forth word *\fo{execve}.
\ *C   execve       \ path argv envp -- int

\ *P The parser used to separate the tokens is not ideal. If you
\ ** have problems with a definition, make sure that *\fo{*}
\ ** tokens are white-space separated. Formal parameter names,
\ ** e.g. *\i{argv} above are ignored. Array indicators, *\i{[]}
\ ** above, are also ignored when part of the names.

\ *P The input types may be followed by a dummy name which is
\ ** discarded. Everything on the source line after the closing
\ ** ')' is discarded.

\ *P From VFX Forth v4.3 onwards, PASCAL is the default calling
\ ** convention in the Windows version. The default for the Linux
\ ** and OS X versions is "C". The default is always used unless
\ ** overridden in the declaration.


\ *********
\ *S Format
\ *********
\ *E EXTERN: <return> [ <callconv> ] <name> '(' <arglist> ')' ';'
\ **
\ ** <return>    := { <type> [ '*' ] |  void }
\ ** <arg>       := { <type> [ '*' ] [ <name> ] }
\ ** <args>      := { [ <arg>, ]* <arg> }
\ ** <arglist>   := { <args> | void }	Note: "void, void" etc. is illegal.
\ ** <callconv>  := { PASCAL | WINAPI | STDCALL | "PASCAL" | "C" }
\ ** <name>      := <any Forth acceptable namestring>
\ ** <type>      := ... (see below, "void" is a valid type)

\ *P Note that during searches <name> is passed to the operating
\ ** system exactly as it is written, i.e. case sensitive. The
\ ** Forth name is case-insensitive.

\ *P As a standard Forth's string length for dictionary names is
\ ** only guaranteed up to 31 characters for portable source code,
\ ** very long API names can cause problems. Therefore the word
\ ** *\fo{AliasedExtern:} allows separate specification of API
\ ** and Forth names (see below). *\fo{AliasedExtern:} also
\ ** solves problems when API functions only differ in case
\ ** or their names conflict with existing Forth word names.

\ **********************
\ *S Calling Conventions
\ **********************
\ *P In the discussion *\b{caller} refers to the Forth system
\ ** (below the application layer and *\b{callee} refers to a
\ ** a function in a DLL or shared library. The *\fo{EXTERN:}
\ ** mechanism supports three calling conventions.
\ *(
\ *B C-Language: *\fo{"C"} *\br{}
\ **  Caller tidies the stack-frame.
\ ** The arguments (parameters) which are passed to the library
\ ** are reordered. This convention can be specified by using
\ ** *\fo{"C"} after the return type specifier and
\ ** before the function name. For Linux and most Unix-derived
\ ** operating systems, this is the default.
\ *B Pascal language: *\fo{"PASCAL"} *\br{}
\ ** Callee removes arguments from the stack frame. This is
\ ** invisible to the programmer at the application layer
\ ** The arguments (parameters) which are passed to the library are
\ ** not reordered. This convention is specified by
\ ** *\fo{"PASCAL"} after the return type specifier and
\ ** before the function name.
\ *B Windows API: *\fo{WINAPI | PASCAL | STDCALL} *\br{}
\ ** In nearly all cases (but *\b{not all}), calls to
\ ** Windows API functions require C style argument reversal and
\ ** the called function cleans up. Specify this convention with
\ ** *\fo{PASCAL}, *\fo{WinAPI} or *\fo{StdCall} after the
\ ** return type specifier and before the function name. For
\ ** Windows, this is the default.
\ *)

\ *P Unless otherwise specified, the Forth system's default
\ ** convention is used. Under Windows this is *\fo{WINAPI} and
\ ** under Linux and other Unices it is *\fo{"C"}.

\ *************************
\ *S Promotion and Demotion
\ *************************
\ *P The system generates code to either promote or demote non-CELL
\ ** sized arguments and return results which can be either signed or
\ ** unsigned. Although Forth is an un-typed language it must deal with
\ ** libraries which do have typed calling conventions. In general
\ ** the use of non-CELL arguments should be avoided but return
\ ** results should be declared in Forth with the same size as the C or
\ ** PASCAL convention documented.

\ ********************
\ *S Argument Reversal
\ ********************
\ *P When calling an external function the parameters are reordered
\ ** if required by the operating system; this is to enable the
\ ** argument list to read left to right in Forth source as well
\ ** as in the C-style operating system documentation.

\ *P Under certain conditions, the order can be reversed. See the
\ ** words *\fo{"C"} and *\fo{"PASCAL"} which define the order for
\ ** the operating system. See *\fo{L>R} and *\fo{R>L} which define
\ ** the Forth stack order with respect to the arguments in the
\ ** prototype.

\ *****************************
\ *S C comments in declarations
\ *****************************
\ *P Very rudimentary support for C comments in declarations is
\ ** provided, but it is good enough for the vast majority of
\ ** declarations.
\ *(
\ *B Comments can be *\fo{// ...} or *\fo{/* ... */},
\ *B Comments must be at the end of the line,
\ *B Comments are treated as extending to the end of the line,
\ *B Comments must not contain the ')' character.
\ *)
\ *P The example below is taken from a *\i{SQLite} interface.
\ *E Extern: "C" int sqlite3_open16(
\ **   const void * filename,  /* Database filename [UTF-16] */
\ **   sqlite3 ** ppDb         /* OUT: SQLite db handle */
\ ** );


\ **********************************
\ *S Controlling external references
\ **********************************

also forth definitions
1 value ExternWarnings?	\ -- n
\ *G Set this true to get warning messages when an external reference
\ ** is redefined.
0 value ExternRedefs?	\ -- n
\ *G If non-zero, redefinitions of existing imports are permitted.
\ ** Zero is the default for VFX Forth so that redefinitions of
\ ** existing imports are ignored.
1 value LibRedefs?	\ -- n
\ *G If non-zero, redefinitions of existing libraries are permitted.
\ ** Non-zero is the default for VFX Forth so that redefinitions of
\ ** existing libraries and OS X frameworks are permitted. When set
\ ** to zero, redefinitions are silently ignored.
1 value InExternals?	\ -- n
\ *G Set this true if following import definitions are to be in
\ ** the *\fo{EXTERNALS} vocabulary, false if they are to go into
\ ** the wordlist specified in *\fo{CURRENT}. Non-Zero is the
\ ** default for VFX Forth.
: InExternals	\ --
\ *G External imports are created in the *\fo{EXTERNALS} vocabulary.
  1 -> InExternals?  ;
: InCurrent	\ --
\ *G External imports are created in the wordlist specified by
\ ** *\fo{CURRENT}.
  0 -> InExternals?  ;
previous definitions

' externals voc>wid constant Externals.wid
\ The wordlist ID of the EXTERNALS vocabulary.


\ ******************
\ *S Library Imports
\ ******************
\ *P In VFX Forth, libraries are held in the *\fo{EXTERNALS}
\ ** vocabulary, which is part of the minimum search order.
\ ** Other Forth systems may use *\fo{FORTH} vocabulary or
\ ** the *\fo{CURRENT} wordlist.

\ *P For turnkey applications, initialisation, release and
\ ** reload of required libraries is handled at start up.

(( for DocGen
variable lib-link	\ -- addr
\ *G Anchors the chain of dynamic/shared libraries.
))
variable lib-mask	\ -- addr
\ *G If non-zero, this value is used as the mode for *\fo{dlopen()}
\ ** calls in Linux and OS X.

struct /libstr	\ -- size
\ *G The structure used by a *\fo{Library:} definition.
\ *[
  int >liblink		\ link to previous library
  int >libaddr		\ library Id/handle/address, depends on O/S
  int >libmask		\ mask for dlopen()
  0 field >libname	\ zero terminated string of library name
end-struct
\ *]

struct /funcstr	\ -- size
\ *G The structure used by an imported function.
  int >funclink		\ link to previous function
  int >funcaddr		\ function address
  0 field >funcname	\ zero terminated string of function name
end-struct

also forth definitions

1 constant RTLD_LAZY
: init-lib	\ libstr --
\ *G Given the address of a library structure, load the library.
  dup >libname LoadLibrary swap >libaddr !  ; doNotSin

: clear-lib	\ libstr --
\ *G Unload the given library and zero its load address.
  dup >libaddr @ FreeLibrary drop  >libaddr off
; doNotSin

: clear-libs    \ --
\ *G Clear all library addresses.
  lib-link
  begin
    @ dup
  while
    0 over >libaddr !
    >liblink
  repeat
  drop
; doNotSin

: init-libs     \ --
\ *G Release and reload the required libraries.
  clear-libs  lib-link
  begin
    @ dup
  while
    dup init-lib  >liblink
  repeat
  drop
;
' init-libs AtCold

: find-libfunction	\ z-addr -- address|0
\ *G Given a zero terminated function name, attempt to find the
\ ** function somewhere within the already active libraries.
  >r  lib-link
  begin					\ -- link ; R: -- z$
    @ dup
  while                 		\ -- struct ; R: -- z$
    dup >libaddr @ r@ GetProcAddress dup \ -- struct addr|0 addr|0 ; R: -- z$
    if  nip  r> drop exit  endif
    drop >liblink
  repeat
  drop  r> drop 0
;

: .Libs         \ --
\ *G Display the list of declared libraries.
  lib-link
  begin
    @ dup
  while
    dup >libname cr .z$
    out @ 31 and 32 swap - spaces
    dup >libaddr @ .lword
    >liblink
  repeat
  drop
;

: #BadLibs	\ -- u
\ *G Return the number of declared libraries that have not yet been
\ ** loaded.
  0 lib-link
  begin
    @ dup
  while
    dup >libaddr @ 0= if
      swap 1+ swap
    then
    >liblink
  repeat
  drop
;

: .BadLibs         \ --
\ *G Display a list of declared libraries that have not yet been
\ ** loaded.
  0 lib-link
  begin
    @ dup
  while
    dup >libaddr @ 0= if
      dup >libname cr .z$
      swap 1+ swap
    then
    >liblink
  repeat
  drop
  cr . ." Unresolved Libraries"
;

: Library:      \ "<name>" -- ; -- loadaddr|0
\ *G Register a new library by name.
\ ** If *\fo{LibRedefs?} is set to zero, redefinitions are silently
\ ** ignored.
\ ** Use in the form:
\ *C   LIBRARY: <name>
\ *P Executing *\fo{<name>} later will return its load address.
\ ** This is useful when checking for libraries that may not be
\ ** present. After definition, the library is the first one
\ ** searched by import declarations.
  LibRedefs? 0= if			\ ignore redefinitions?
    >in @ parse-name search-context	\ already defined?
    if  drop drop exit  endif
    >in !
  endif
  >in @  get-current  InExternals?	\ save >IN, which wordlist?
  if  Externals.wid set-current  then	\ -- >in wid
  create
    set-current  >in !			\ restore wordlist and >IN
    here				\ start of  structure
    lib-link link,			\ link to previous library
    0 ,					\ load address, filled in later
    lib-mask @ ,  lib-mask off		\ mask for dlopen(), 0 = use default
    parse-name z$,			\ lay name as zero terminated string
\    bl word $>z,			\ lay name as zero terminated string
    init-lib
  does>
    >libaddr @
;

: topLib	\ libstr --
\ *G Make the library structure the top/first in the library
\ ** search order.
  lib-link  2dup delLink  AddLink  ;

: firstLib	\ "<name>" --
\ *G Make the library first in the library search order. Use during
\ ** interpretation in the form:
\ *C   FirstLib <name>
\ *P to make the library first in the search order. This is useful
\ ** when you know that there may be several functions of the
\ ** same name in different libraries.
  ' >body topLib  ;

: [firstLib]	\ "<name>" --
\ *G Make the library first in the library search order. Use during
\ ** compilation in the form:
\ *C   [firstLib] <name>
\ *P to make the library first in the search order. This is useful
\ ** when you know that there may be several functions of the
\ ** same name in different libraries.
  ' >body postpone literal  postpone topLib
; immediate

previous definitions


\ ======================
\ *N Mac OS X extensions
\ ======================
\ *P The phrase *\fo{Framework <name.framework>} creates two Forth words,
\ ** one for the library access, the other to make that library top in
\ ** the search order. For example:
\ *C   framework Cocoa.framework
\ *P produces two words
\ *C   Cocoa.framework/Cocoa
\ *C   Cocoa.framework
\ *P The first word is the library definition itself, which
\ ** behaves in the normal VFX Forth way, returning its load
\ ** address or zero if not loaded. The second word forces
\ ** the library to be top/first in the library search order.
\ ** Thanks to Roelf Toxopeus.

\ *P As of OSX 10.7, *\fo{FRAMEWORK} (actually *\b{dlopen()})
\ ** will search for frameworks in all the default Frameworks
\ ** directories:
\ *(
\ *B /Library/Frameworks
\ *B /System/Library/Frameworks
\ *B ~/Library/Frameworks
\ *)

[defined] Target_x64_OSX [if]

also forth definitions

: framework	\ --
\ *G Build the two framework words. See above for more details.
\ ** If *\fo{LibRedefs?} is set to zero, redefinitions are silently
\ ** ignored.
  {: | buff[ #256 ] -- :}
  LibRedefs? 0= if			\ ignore redefinitions?
    >in @ parse-name search-context	\ already defined?
    if  drop drop exit  endif
    >in !
  endif
  >in @
  s" Library: " buff[ place		\ build library: name.framework/name
  parse-name 2dup buff[ append		\ Library: name.framework
  s" /" buff[ append			\ Library: name.framework/
  #10 - buff[ append			\ Library: name.framework/name
  buff[ count evaluate
  >in !
  create
    lib-link @ ,			\ address of last library structure
  does>
    @ toplib
;

previous definitions

[then]


\ *******************
\ *S Function Imports
\ *******************
\ *P Function declarations in shared libraries are compiled into
\ ** the *\fo{EXTERNALS} vocabulary. They form a single linked
\ ** list. When a new function is declared, the list of previously
\ ** declared libraries is scanned to find the function. If the
\ ** function has already been declared, the new definition is
\ ** ignored if *\fo{ExternRedefs?} is set to zero. Otherwise,
\ ** the new definition overrides the old one as is usual
\ ** in Forth.

\ *P In VFX Forth, *\fo{ExternRedefs?} is zero by default.

(( for DocGen
variable import-func-link	\ -- addr
\ *G Anchors the chain of imported functions in shared libraries.
))

((
: resolve-libfunction   \ z-addr -- address|0 ; Windows version
\ +G Look up a definition in loaded libraries and return XT or 0.
\ +* Automatically attempts to find a version with 'A' appended for
\ +* Windows.
  { | tn[ MAX_PATH ] -- }
  [ also system ]
  dup find-libfunction dup		\ look for raw name
  if  nip exit  then
  drop
  tn[ MAX_PATH erase			\ look for name with 'A' appended
  zcount tn[ swap move
  [char] A tn[ zcount + c!
  tn[ find-libfunction
  [ previous ]
;
))
: resolve-libfunction   \ z-addr -- address|0 ; Unix version
  [ also system ]
  find-libfunction			\ look for raw name
  [ previous ]
;

: resolveImport	\ struct --
\ Resolve the imported function whose structure is given.
  dup >funcname resolve-libfunction swap >funcaddr !  ;

also forth definitions
: ExternLinked	\ c-addr u -- address|0
\ *G Given a string, attempt to find the named function in  the
\ ** already active libraries. Returns zero when the function is
\ ** not found.
  {: | temp[ 256 ] -- :}
  255 min temp[ zplace
  temp[ resolve-libfunction
;

: init-imports  \ --
\ *G Initialise Import libraries. *\fo{INIT-IMPORTS} is called by
\ ** the system cold chain.
  import-func-link @
  begin
    dup
  while
    dup resolveImport
    >funclink @
  repeat
  drop
;
' init-imports AtCold
previous definitions


\ ********************************
\ +S Import parsing and generation
\ ********************************

\ ======================
\ +N Data and structures
\ ======================

\ These are constants used as THROW codes with associated text messages.
SysErrdef err_ImpType	"Invalid type in EXTERN: import"
SysErrDef err_CGsize	"Invalid size in EXTERN: argument preparation"
SysErrDef err_RetType	"Invalid return type"
SysErrDef err_RetSize	"Invalid return size"
SysErrDef err_RegParm	"Invalid register for parameter"
SysErrDef err_noFPABI	"No external FP ABI defined"
SysErrDef err_noFPsystem "No Forth FP system defined"

\ argument types - BOOL could be considered a form of UINT, but
\ is O/S specific.
0 constant SINTtype	\ -- n ; default is signed int
1 constant UINTtype	\ -- n ; signed int
2 constant BOOLtype	\ -- n ; C boolean
3 constant FLOATtype	\ -- n ; float/double/ext ....

\ Extended types above $100, mostly for Objective C
$101 constant dPOINTtype	\ -- n ; Point of two doubles ; SFP030

0 constant notReg	\ -- n ; parameter not in register
1 constant IntReg	\ -- n ; parameter in integer/general register
2 constant FPReg	\ -- n ; parameter is in VFP register

struct /argItem	\ -- len
  int ai.Size		\ argument size in bytes, 0=void, 1=byte/char, 2=short, 4=int ...
  int ai.type		\ 0=sint, 1=uint, 2=bool, 3=float ...
  int ai.inReg?		\ true if low 8 bytes in register (may be FPreg). 0=no, 1=IntReg, 2=FPreg
  int ai.Reg		\ register number to be used
  int ai.FrameOff	\ offset on operating system frame to data
  int ai.PSPoff		\ offset on Forth parameter stack (0=TOS, 1=NOS ...)
  int ai.FSPoff		\ offset on Forth float stack (0=FTOS, 1=FNOS
end-struct
#32 /argitem * constant /argList	\ size of the argument list, max 32 items


\ Operating system dependent items. These VALUEs are set to
\ defaults at the start of each external import declaration.

0 value CalleeCleaned?	\ -- flag
\ 0 = "C", 1 = "PASCAL", used for cleanup only. If the calling
\ mechanism preserves the Forth stack pointers, this value is not
\ needed.
6 value /MaxIntRegArgs	\ -- u
\ Max number of integer arguments passed in registers.
8 value /MaxFPRegArgs	\ -- u
\ Max number of integer arguments passed in registers.
#16 value FrameAlignment	\ -- u
\ Set to alignment of stack frame in memory. Must be a power of two.
true value AlignInFrame? \ -- flag
\ True if items on the stack frame are padded for data alignment.
true value REQUIRE_NREV	\ -- flag
\ Set true if arguments are reversed by default (C argument order),
\ i.e. the leftmost argument in the argument declaration list is
\ on the top of the O/S stack.
true value L>R?		\ -- flag
\ Set true when the Forth stack order matches the argument
\ declaration order.
0 value ObjPtr?		\ -- flag
\ True if an object-oriented call is required, i.e. a hidden
\ "this" pointer is used. The location of "this" is compiler and
\ OS dependent.
\   0 = none
\   1 = VC++, pointer in ECX
\   2 = gcc, pointer on top, but below struct pointer (see below)
0 value StructRet?	\ -- flag
\ Set this true if the import returns a structure and requires a
\ hidden structure pointer. The location of the structure pointer
\ is compiler and OS dependent.
\   0 = none
\   1 = VC++
\   2 = gcc
0 value varargs?	\ -- flag
\ Set this true if the parameter list is of variable length.

defer setDefExt	\ --
\ Set the call type defaults in the VALUEs above.
((
: (setDefExt)	\ --
\ set up the defaults for EXTERN: calls.
  0 to FrameAlignment  0 to AlignInFrame?
  TRUE to REQUIRE_NREV  1 to CalleeCleaned?
  TRUE to L>R?
  0 to ObjPtr?  0 to StructRet?  0 to varargs?
;
assign (setDefExt) to-do setDefExt
))
#256 buffer: importName		\ name of DLL function
#256 buffer: externName		\ name of Forth word
variable usesFloats		\ non-zero if FP arguments or return ; SFP006
variable NumArgs		\ number of input arguments
variable NumInts		\ number of arguments on Forth data stack
variable NumFloats		\ number of arguments on Forth Float stack
variable /DataFrame		\ total size of data frame with padding
variable /ForthFrame		\ number of bytes taken from Forth data stack
variable /ForthFPFrame		\ number of bytes taken from Forth float stack
variable NCRN			\ next core register number
variable NSAA			\ next stacked argument address/offset
variable NFRN			\ next floating point register number
/argItem buffer: ReturnRes	\ structure for return data
/argList buffer: ArgList	\ argument data during compilation

0 value LastImport		\ address of current/last import structure

: .ArgType	\ n -- ; display argument type
  case
    SINTtype of  ." SINT"  endof
    UINTtype of  ." UINT"  endof
    BOOLtype of  ." BOOL"  endof
    FLOATtype of  ." Float"  endof
    dPOINTtype of  ." DPOINT_DOUBLE"  endof	\ SFP030
      ." ???"
  endcase
;

: .IntArgReg	\ ^ai -- ; display argument 64 bit int register
  ." , "
  case  ai.reg @
    0 of  ." rdi "  endof
    1 of  ." rsi "  endof
    2 of  ." rdx "  endof
    3 of  ." rcx "  endof
    4 of  ." r8  "  endof
    5 of  ." r9  "  endof
  endcase
;

: .FPArgReg	\ ^ai -- ; display FP register
  ." , "
  case  ai.reg @
    0 of  ." xmm0 "  endof
    1 of  ." xmm1 "  endof
    2 of  ." xmm2 "  endof
    3 of  ." xmm3 "  endof
    4 of  ." xmm4 "  endof
    5 of  ." xmm5 "  endof
    6 of  ." xmm6 "  endof
    7 of  ." xmm7 "  endof
  endcase
;

: .ArgData	\ ^ai --
\ Display the argument data
  cr ." Size: " dup ai.size @ dup 0 .r
  0= if
    drop  ."  (void)"
  else
    ." , Type: " dup ai.Type @ .ArgType
    ." , PSP: " dup ai.PSPoff @ 0 .r
    ." , FSP: " dup ai.FSPoff @ 0 .r
    ." , Frame: " dup ai.FrameOff @ 0 .r
    case  dup ai.inReg? @
      0 of  endof
      1 of  dup .IntArgReg  endof
      2 of  dup .FPArgReg   endof
    endcase
    drop
  endif
;

: ArgList[]	\ u -- addr
\ Return the address of the uth item in argList
  /argItem * argList +  ;

: .ArgList	\ --
\ Display the calculated argument and return data.
  cr ." Arguments"  numArgs @ 0
  ?do  i ArgList[] .ArgData  loop
  cr ." --" cr ." Returns"
  ReturnRes .ArgData
;

0 value ShowExtData?	\ -- flag
\ Set non-zero to see the argument data at compile time.

: ?ShowExtData	\ --
  ShowExtData? if
    cr cr ExternName count type .ArgList
  endif
; doNotSin

: initArgData	\ -- ; initialise ARGLIST and RETURNRES ; SFP027
  ArgList /ArgList erase  ReturnRes /ArgItem erase
;

: CalcSource	\ ^ai --
\ Generate the source Forth data for the given argument.
  dup ai.type @ FLOATtype = if
    /ForthFPFrame @ over ai.FSPoff !	\ where it is on Forth FP stack
    8 /ForthFPframe +!			\ all Forth FPs are 64 bit
    drop  numFloats incr  exit
  then
  dup ai.type @ dPOINTtype = if		\ SFP030
    /ForthFPFrame @ over ai.FSPoff !	\ where it is on Forth FP stack
    #16 /ForthFPframe +!		\ 2 Forth FPs (64 bit)
    drop  numFloats incr  exit
  then
  /ForthFrame @ over ai.PSPoff !	\ where it is on Forth data stack
  ai.size @ aligned /ForthFrame +!
  numInts incr
; doNotSin

: CalcForthOff	\ --
\ Calculate the offsets of data items on the Forth input stack.
\ The Forth order is only affected by the choice of L>R or R>L
\ ordering. By default we use L>R.
  numArgs @ if
    L>R? if
      0 numArgs @ 1-
      do  i ArgList[] CalcSource  -1 +loop
    else
      numArgs @ 0
      do  i ArgList[] CalcSource  loop
    endif
  endif
; doNotSin

: AlignFrame	\ n --
\ Force frame alignment of a item of the given size to an N byte
\ boundary where N is a power of two.
  /DataFrame @				\ offset w.r.t frame after call
  over 1- +  swap negate and		\ -- dp'
  /DataFrame !
; doNotSin

: ?AlignFrame	\ size --
\ Force frame alignment of a item of the given size to an N byte
\ boundary where N is a power of two. Ignore void items.
  ?dup if
    aligned AlignInFrame?
    if  AlignFrame  else  drop  endif
  endif
; doNotSin

: frameDest	\ ^ai --
\ Generate the destination frame data for the given argument.
  dup ai.size @ ?AlignFrame
  /DataFrame @ over ai.FrameOff !
  ai.size @ aligned /DataFrame +!
;

((
struct /argItem	\ -- len
  int ai.Size		\ argument size in bytes, 0=void, 1=byte/char, 2=short, 4=int ...
  int ai.type		\ 0=uint, 1=sint, 2=bool, 3=float
  int ai.inReg?		\ true if low 8 bytes in register (may be FPreg). 0=no, 1=Reg, 2=FPreg
  int ai.Reg		\ register number to be used
  int ai.FrameOff	\ offset on operating system frame to data
  int ai.PSPoff		\ offset on Forth parameter stack (0=TOS, 1=NOS ...)
  int ai.FSPoff		\ offset on Forth float stack (0=FTOS, 1=FNOX, ...)
end-struct
))

: intDest	\ ^ai --
\ We assume that data is no more than 8 bytes.
  NCRN @ /MaxIntRegArgs >=		\ if we have used all registers
  if  frameDest exit  then		\ just put it on the frame
  1 over ai.InReg? !			\ mark as integer register
  NCRN @ over ai.Reg !			\ register number
  1 NCRN +!				\ 1 register used
\  dup ai.Size @ aligned 2 rshift NCRN +!  \ number of registers needed
  drop
;

: fpDest	\ ^ai --
\ We assume that data is 8 bytes.
  NFRN @ /MaxFPRegArgs >=		\ if we have used all registers
  if  frameDest exit  then		\ just put it on the frame
  2 over ai.InReg? !			\ mark as FP register
  NFRN @ over ai.Reg !			\ register number
  1 NFRN +!				\ number of registers needed, Sx/Dx
  drop
;

: CalcDest	\ ^ai --
\ Generate the destination frame data for the given argument.
  dup ai.Type @ FLOATtype =
  if  fpDest  else  intDest  then
; doNotSin

: CalcFrameOff	\ --
\ Calculate the O/S stack offsets. These are from the C-style
\ argument list, and so are unaffected by R>L, which only affects
\ the source offsets on the Forth stack with respect to the
\ declaration list.
  REQUIRE_NREV if
    numArgs @ 0
    ?do  i ArgList[] CalcDest  loop
  else
    numArgs @ if
      0 numArgs @ 1-
      do  i ArgList[] CalcDest  -1 +loop
    endif
  endif
; doNotSin

: initOffsets	\ --
\ Initialise the stack and frame offsets required.
  numInts off  numFloats off
  /DataFrame off			\ C data frame
  /ForthFrame off  /ForthFPFrame off	\ Forth frames
  NCRN off  NSAA off  NFRN off
; doNotSin


\ ******************
\ +S Code generation
\ ******************

\ ========
\ +N Tools
\ ========

: [a		\ -- ; start of assembler sequence
  also asm-access  ;  immediate

: a]            \ -- ; end of assembler sequence
  previous  ;  immediate

: [a]		\ -- ; flush asm sequence
  [ also asm-core ] asm_end asm_start [ previous ]  ;	\ flush assembler


\ ==============================
\ +N Prologue, call and epilogue
\ ==============================
\ The process of making the external call is to:
\ 1) Save Forth system registers on the return (RSP) stack.
\ 2) If required save the FP state.
\ 3) Preserve the Forth TOS (normally cached in RBX) on the
\    deep stack so that the Forth stack is all in memory.
\ 4) Preserve RSP by extending the deep Forth stack.
\ 5) Adjust the return stack (RSP) to operating system requirements
\    with space allocated for the function arguments.
\    This may involve aligment to a 16 byte boundary.
\ 6) Copy the Forth and FP parameters to registers or return
\    stack locations allowing for the calling convention.
\ 7) Call the external function
\ 8) Restore RSP for "C" call convention
\ 9) Save the return data and process TOS for void and float returns
\ 10) Restore the float conditions as required
\ 11) Restore the Forth stack
\ 12) Restore Forth registers

2 cells constant /PSPextend	\ -- u
\ The number of bytes by which the Forth stack is dropped to
\ extend it. The high cell contains the cached TOS. The bottom
\ cell (offset 0) contains the saved RSP.
\ MUST BE at least the size of a LongLong.
/PSPextend cell - constant /PSPtos	\ -- u
\ The offset from the dropped Forth stack to the cell containing
\ TOS.

: FrameDrop	\ -- n
\ The amount by which the frame is dropped before alignment
\ This is a positive number to subtract.
  /DataFrame @ ( FrameAlignment + ) 	\ Unix:
\  /DataFrame @  #32 max  		\ Win: at least 32 for shadow
;

: FrameMask	\ -- mask
\ The mask applied to RSP after FrameDrop has been applied.
  FrameAlignment negate  ;

: -ReturnForth	\ n -- n-x
\ Remove the size of data returned on the Forth data stack.
  ReturnRes ai.Size @ aligned -  ;

: #PSPrestore	\ -- n
\ generate code to restore TOS for void/float returns. This is
\ the number of bytes added to restore the Forth frame.
  /ForthFrame @				\ size of Forth stack arguments
  returnRes ai.size @ if
    case  ReturnRes ai.Type @ 		\ Return type
      SINTtype of  -ReturnForth  endof
      UINTtype of  -ReturnForth  endof
      BOOLtype of  -ReturnForth  endof
    endcase
  endif
  /PSPextend +				\ frame overhead, >= largest return type
;

: #FSPrestore	\ -- n
\ Calculate the number of bytes by which the FSP stack should be
\ adjusted.
  /ForthFPFrame @			\ size of Forth stack arguments
  returnRes ai.size @ if
    ReturnRes ai.Type @ FLOATtype = if	\ Return type
      8 -				\ size of float on FSP stack
    endif
    ReturnRes ai.Type @ dPOINTtype = if	\ Return type ; SFP030
      #16 -				\ size of float on FSP stack
    endif
  endif
;

: ^efn		\ -- addr
\ address holding function address
  LastImport >funcaddr  ;

: prologue,	\ --
\ Lay the call entry prologue
  [a
\ save Forth registers ; MPE005
    push  rsi
    push  rdi
    sub	  r13, # 8			\ save FTOS on R13 stack ; SFP030
    movsd 0 [r13], xmm8

\ preserve RSP and TOS
    lea   rbp, /PSPextend ( 16 ) negate [rbp]	\ save RBX and RSP
    mov   0 [rbp], rsp			\ after call, ESP will be restored
    mov   /PSPtos [rbp], rbx		\ save TOS, easier from memory if double
\ make a frame on RSP and align it to OS convention, 16 byte aligned less 4 bytes
  ]a [a]
  FrameDrop if
  [a
    lea   rax, FrameDrop negate [rsp]	\ compile time constant
    and   rax, # FrameMask		\ 16 byte align
    lea   rsp, 0 [rax]
  a] [a]
  else
  [a
    and   rsp, # FrameMask		\ 16 byte align
  a] [a]
  then
; doNotSin

: generateCall	\ --
\ Lay the code to call the external function, and clean up.
  [a
\ Call external C/PASCAL proc
    cld
    call  ^efn []			\ indirect through pointer
    cld
  a] [a]
  CalleeCleaned? 0= if			\ C style call
    [a  mov   rsp, 0 [rbp]  a] [a]	\ prologue saved ESP, used for C.
  endif
; doNotSin

: removeArgs	\ --
\ Restore the Forth and O/S stacks.
\ Forth float stack restoration
\  ?RestoreFTOS				\ Do we need to reload XMM8 ?/
  #FSPrestore ?dup if
    >r  [a  lea    r13, r> [r13]  a] [a]
  then
\ Forth data stack restoration
  #PSPrestore ?dup if
    >r  [a lea   rbp, r> [rbp] a] [a]	\ adjust Forth PSP stack: 0-4+8=4
  endif
;

: epilogue,	\ --
\ restore Forth registers
  removeArgs
  [a
    movsd xmm8, 0 [r13]			\ FTOS ; SFP030
    add   r13, # 8
    pop   rdi				\ LP
    pop   rsi				\ UP
  a] [a]
; doNotSin


\ ============================
\ +N Copy Forth stack to frame
\ ============================

: IntArgReg32,	\ ^ai --
  case  ai.reg @
    0 of  [a edi, a]  endof
    1 of  [a esi, a]  endof
    2 of  [a edx, a]  endof
    3 of  [a ecx, a]  endof
    4 of  [a r8d, a]  endof
    5 of  [a r9d, a]  endof
  endcase
;

: IntArgReg,	\ ^ai --
  case  ai.reg @
    0 of  [a rdi, a]  endof
    1 of  [a rsi, a]  endof
    2 of  [a rdx, a]  endof
    3 of  [a rcx, a]  endof
    4 of  [a r8, a]  endof
    5 of  [a r9, a]  endof
  endcase
;

: FPArgReg,	\ ^ai --
  case  ai.reg @
    0 of  [a xmm0, a]  endof
    1 of  [a xmm1, a]  endof
    2 of  [a xmm2, a]  endof
    3 of  [a xmm3, a]  endof
    4 of  [a xmm4, a]  endof
    5 of  [a xmm5, a]  endof
    6 of  [a xmm6, a]  endof
    7 of  [a xmm7, a]  endof
  endcase
;


\ =======
\ +H Ints
\ =======

: CopyArgS8	{: ^ai srcoff destoff -- :}
\ Copy a signed 8 bit item from srcoff [RBP] to destoff [ESP] or register
  ^ai ai.inReg? @ if
    [a  movsx  a]  ^ai IntArgReg,  [a byte srcoff [rbp]  a] [a]
  else
    [a  movsx  RAX, BYTE srcoff [RBP]  a] [a]
    [a  mov    destoff [RSP], RAX a] [a]
  then
;

: CopyArgS16	{: ^ai srcoff destoff -- :}
\ Copy a signed 16 bit item from srcoff [RBP] to destoff [ESP] or register
  ^ai ai.inReg? @ if
    [a  movsx  a]  ^ai IntArgReg,  [a  word srcoff [rbp]  a] [a]
  else
    [a  movsx  RAX, word srcoff [RBP]  a] [a]
    [a  mov    destoff [RSP], RAX  a] [a]
  then
;

: CopyArgS32	{: ^ai srcoff destoff -- :}
\ Copy a signed 32 bit item from srcoff [RBP] to destoff [ESP] or register
  ^ai ai.inReg? @ if
    [a  movsxd  a]  ^ai IntArgReg,  [a  dword srcoff [rbp]  a] [a]
  else
    [a  movsxd  RAX, dword srcoff [RBP]  a] [a]
    [a  mov    destoff [RSP], RAX  a] [a]
  then
;

: CopyArgU8	{: ^ai srcoff destoff -- :}
\ Copy an unsigned 8 bit item from srcoff [RBP] to destoff [ESP] or register
  ^ai ai.inReg? @ if
    [a  movzx  a]  ^ai IntArgReg,  [a  byte srcoff [rbp]  a] [a]
  else
    [a  movzx  RAX, BYTE srcoff [RBP]  a] [a]
    [a  mov    destoff [RSP], RAX  a] [a]
  then
;

: CopyArgU16	{: ^ai srcoff destoff -- :}
\ Copy an unsigned 16 bit item from srcoff [RBP] to destoff [ESP] or register
  ^ai ai.inReg? @ if
    [a  movzx  a]  ^ai IntArgReg,  [a  word srcoff [rbp]  a] [a]
  else
    [a  movzx  RAX, word srcoff [RBP]  a] [a]
    [a  mov    destoff [RSP], RAX  a] [a]
  then
;

: CopyArgU32	{: ^ai srcoff destoff -- :}
\ Copy an unsigned 32 bit item from srcoff [RBP] to destoff [ESP] or register
  ^ai ai.inReg? @ if
    [a  movsxd  a]  ^ai IntArgReg32,  [a  dword srcoff [rbp]  a] [a]
  else
    [a  movsxd  EAX, dword srcoff [RBP]  a] [a]
    [a  mov    destoff [RSP], RAX  a] [a]
  then
;

: CopyArgX64	{: pai srcoff destoff -- :}
\ Copy a 64 bit item from srcoff [RBP] to destoff [RSP] or register
  pai ai.inReg? @ if
    [a  mov  a]  pai IntArgReg,  [a  srcoff [rbp]  a] [a]
  else
    [a  mov  RAX, srcoff [RBP]  a] [a]
    [a  mov  destoff [RSP], RAX  a] [a]
  then
;


\ =========
\ +H Floats
\ =========

: FPsrc		\ off -- ; off-8 or XMM8 if off=0
  >r [a r> [R13] a]
;

: copyFPArg32	{: ^ai srcoff destoff -- :}
\ Copy a 32 bit float from the float stack to destoff [RSP] or register
  ^ai ai.inReg? @ if
    [a  cvtsd2ss  a]  ^ai FPArgReg,  srcoff FPsrc [a]
  else
    [a  cvtsd2ss  XMM9, ]a  srcoff FPsrc  [a]
    [a  movss  destoff [RSP],  XMM9  a] [a]
  then
\  dropFTOS
;

: copyFPArg64	{: ^ai srcoff destoff -- :}
\ Copy a 64 bit float from the float stack to destoff [RSP] or register
  ^ai ai.inReg? @ if
    [a movsd a]  ^ai FPArgReg,  srcoff FPsrc [a]
  else
    srcoff 0= if
      [a  movsd  destoff [RSP],  XMM8  a] [a]
    else
      [a  movsd  XMM9, ]a  FPsrc  [a]
      [a  movsd  destoff [RSP],  XMM9  a] [a]
    then
  then
\  dropFTOS
;


\ ==============
\ +H Other types
\ ==============

: (CopyBool)	{: ^ai srcoff destoff -- :}
\ Copy an unsigned 8 bit item from srcoff [RBP] to destoff [ESP] or register
  ^ai ai.inReg? @ if
    [a  movzx  a]  ^ai IntArgReg,  [a  byte srcoff [rbp]  a] [a]
    [a  and  a]  ^ai IntArgReg,  [a  # 1  a] [a]
  else
    [a  movzx  RAX, BYTE srcoff [RBP]  a] [a]
    [a  and    RAX, # 1  a] [a]
    [a  mov    destoff [RSP], RAX  a] [a]
  then
;


\ =================
\ +N General copies
\ =================

: offsets	\ ^ai -- srcoff destoff
\ Extract the base offsets from the structure. The source offset
\ is assumed to be on the PSP/RBP stack and is adjusted to be based
\ from the saved TOS.
  dup ai.PSPoff @ /PSPtos +  swap ai.FrameOff @
;

: FPoffsets	\ ^ai -- srcoff destoff
\ Extract the base offsets from the structure. The source offset
\ is assumed to be on the FSP/R13 stack and is adjusted to be based
\ from the FTOS.
  dup ai.FSPoff @  swap ai.FrameOff @
;

: CopySINT	\ ^ai --
\ Compile code to copy TOS to the frame as an signed INT32/64.
\ with any required promotion.
  case  dup ai.size @
    1 of  dup offsets CopyArgS8  endof
    2 of  dup offsets CopyArgS16  endof
    4 of  dup offsets CopyArgS32  endof
    8 of  dup offsets CopyArgX64  endof
     err_CGsize throw
  endcase
;

: CopyUINT	\ ^ai --
\ Compile code to copy TOS to the frame as a UINT32/64.
\ with any required promotion.
  case  dup ai.size @
    1 of  dup offsets CopyArgU8  endof
    2 of  dup offsets CopyArgU16  endof
    4 of  dup offsets CopyArgU32  endof
    8 of  dup offsets CopyArgX64  endof
     err_CGsize throw
  endcase
;

: CopyBOOL	\ ^ai --
\ Compile code to copy TOS to the frame as a BOOL.
\ with any required promotion/demotion.
  dup offsets (CopyBool)
;

: CopyFLOAT	\ ^ai --
\ Compile code to copy FTOS to the frame as a float or double
  case  dup ai.size @
    4 of  dup FPoffsets copyFPArg32  endof
    8 of  dup FPoffsets copyFPArg64  endof
       err_CGsize throw
  endcase
;

: CopyArg	\ ^ai --
\ Compile code to copy an item from Forth to the calling frame
\ with any required promotion.
  case  dup ai.type @
    UINTtype of  CopyUINT  endof
    SINTtype of  CopySINT  endof
    BOOLtype of  CopyBOOL  endof
    FLOATtype of  CopyFloat  endof
      err_ImpType throw
  endcase
;

: CopyFrame	\ --
\ Copy the arguments from the Forth stack to the O/S stack with
\ any required promotion/demotion.
  L>R? REQUIRE_NREV xor 0= if
    numArgs @ if
      0 numArgs @ 1-
      do  i ArgList[] CopyArg  -1 +loop
    endif
  else
    numArgs @ 0
    ?do  i ArgList[] CopyArg  loop
  endif
;


\ ================
\ +N Return result
\ ================

: restoreTOS	\ --
\ For void and float return values, TOS must be restored.
  [a  mov  rbx,  a]  /PSPtos /ForthFrame @ +  [a [RBP] a] [a]
;

: retLongLong	\ --
\ Return a long long (64 bit).
  [a  mov  rbx, rax  a] [a]		\ most significant  -> TOS
;

: ReturnSINT	\ --
\ Return a signed int.
  case  ReturnRes ai.size @
    1 of  [a  movsx  rbx, al  a] [a]  endof
    2 of  [a  movsx  rbx, ax  a] [a]  endof
    4 of  [a  movsxd  rbx, eax  a] [a]  endof
    8 of  [a  mov  rbx, rax  a] [a]  endof
        err_RetSize throw
  endcase
;

: ReturnUINT	\ --
\ Return an unsigned int.
  case  ReturnRes ai.size @
    1 of  [a  movzx  ebx, al  a] [a]  endof
    2 of  [a  movzx  ebx, ax  a] [a]  endof
    4 of  [a  movsxd  ebx, eax  a] [a]  endof
    8 of  [a  mov  rbx, rax  a] [a]  endof
        err_RetSize throw
  endcase
;

: ReturnBOOL	\ --
\ Return a BOOL as a single byte.
  [a  movzx  rbx, al  a] [a]  ;

: returnFP4	\ -- ; return a float from XMM0 to XMM8
  a[  movss  #FSPrestore [r13], xmm0  ]a [a]
;

: returnFP8	\ -- ; return a float from XMM0 to XMM8
  a[  movsd  #FSPrestore [r13], xmm0  ]a [a]
;

: returnFLOAT	\ -- ; return float/double
\ With the float in XMM0, it's already where we can use it.
\ For a float return, we must still restore RBX correctly.
  case  ReturnRes ai.size @
     4 of  returnFP4  endof
     8 of  returnFP8  endof
        err_RetSize throw
  endcase
  RestoreTOS
;

: return2DOUBLES	\ -- ; SFP030
\ return 2 doubles from XMM0 and XMM1 to fpstack.
  a[  movsd  #FSPrestore [r13], xmm1       ]a [a]
  a[  movsd  #FSPrestore 8 + [r13], xmm0   ]a [a]
;

: returnDPOINT	\ -- ; return float/double
\ With the floats in XMM1 and XMM0, it's already where we can use it.
\ For a float return, we must still restore RBX correctly.
  case  ReturnRes ai.size @
    #16 of  return2DOUBLES  endof
        err_RetSize throw
  endcase
  RestoreTOS
;

: returnData	\ -- ; process return value
\ Do any necessary demotion, or lose return result for VOID.
\ This happens before the epilogue code so that the saved frame
\ is available.
  returnRes ai.size @ if	\ non-void
    case  ReturnRes ai.Type @ 	\ Return type
      SINTtype of  ReturnSINT  endof
      UINTtype of  ReturnUINT  endof
      BOOLtype of  ReturnBOOL  endof
      FLOATtype of  returnFLOAT  endof
      dPOINTtype of  returnDPOINT  endof	\ SFP030 ; also for COMPLEX DOUBLE
        err_RetType throw
    endcase
  else				\ void
    RestoreTOS
  endif
; doNotSin


\ ====================
\ +N Import generation
\ ====================

: comp-extern	\ xt --
\ Compile a child of EXTERN:.
  discard-sinline  doCall,
;

: set-comp-extern	\ xt --
\ Set COMP-EXTERN as the compiler of the given XT.
  >code-len  ['] comp-extern over !  cell+ off
; doNotSin

variable FirstLine#	\ -- addr ; SFP022
\ Holds line number of the first line

: setFirstLine	\ --
  line# @ FirstLine# !
;

: MakeColon	\ -- xt ; SFP022
  line# @ >r  FirstLine# @ line# !	\ use first line number for header
  ExternName count (:)			\ make header and start definition
  r> line# !
  last-colon-xt @			\ return xt we have just made
; doNotSin

: MakeNormal	\ --
\ Make the word that calls a normal imported function.
  MakeColon >r
  prologue,
  copyFrame  generateCall  ( removeArgs )  returnData
  epilogue,
  postpone ;				\ finish definition
  r> set-comp-extern			\ set its compiler.
; doNotSin

defer varXcall	\ x1..xn n funcaddr -- ; Not Yet Implemented
defer nXcall	\ x1..xn n funcaddr -- ; Not Yet Implemented

: MakeVarCall	\ --
\ Make the word that calls a varargs imported function.
  MakeColon >r				\ save xt we have just made ; SFP022
  l>r? if
    ^efn postpone literal  postpone @	\ x1..xn n funcaddr --
    postpone varXcall			\ varXcall ; C style param order
  else
    ^efn postpone literal  postpone @	\ x1..xn n funcaddr --
    postpone swap  postpone nxcall	\ swap nxcall ; PASCAL style param order
  endif
  ReturnRes ai.size @ 0=
  if  postpone drop  endif
  postpone ;				\ finish definition
  r> set-comp-extern			\ set its compiler.
; doNotSin

: MakeImport	\ --
\ Make the word that calls the imported function.
  ?ShowExtData  varargs?
  if  MakeVarCall  else  MakeNormal  endif
; doNotSin

: MakeImpVar	\ -- ; SFP017
\ Make the word that returns a variable's address
  ExternName count (:)			\ make header and start definition
  ^efn postpone literal  postpone @	\ variable address
  postpone ;				\ finish definition
; doNotSin

: ?warnExtRedef	\ -- ; SFP017
\ Warn if redefinitions are to be reported.
  ExternWarnings? if
    cr ." WARNING: Redefinition of Extern: " ExternName $.
  then
; doNotSin

: importEntry,	\ -- ; SFP017
\ lay the import list entry
  align
  here -> LastImport			\ save structure address
  import-func-link link,		\ link to import chain
  0 ,					\ holds the function address
  importName $>z,			\ import name counted + 0 terminated
; doNotSin

: GenerateImport        \ --
\ The following structure is built and must match /funcstr:
\ cell	link to previous import in library
\ cell	holds address of function to XCALL
\ z$	0-terminated ImportName of function (case sensitive)
\ cell	pointer to second cell, for a future inline code generator)
\ colon definition
\ Don't do anything if this function already exists and import
\ redefinitions are to be ignored.
  ExternName count Externals.wid search-wordlist if
    drop  ?warnExtRedef  ExternRedefs? 0= if	\ SFP017
      ExternWarnings?
      if  ."  ignored"  then
      exit
    then
  then
  importEntry,				\ Build import list entry ; SFP017

\ Calculate the stack and frame offsets
  varargs? 0=
  if  initOffsets  CalcForthOff  CalcFrameOff  endif

\ Build the colon definition that interprets the function
  align  ^efn ,			\ points to function address cell of import structure
  InExternals? if
    get-current  Externals.wid set-current
    MakeImport
    set-current
  else
    MakeImport
  endif
  LastImport resolveImport		\ Resolve the import we have just made
;
+short-branches
\ ...SFP004

: genExternVar	\ -- ; SFP017
\ Generate an external variable
  ExternName count Externals.wid search-wordlist if
    drop  ?warnExtRedef  ExternRedefs? 0=
    if  exit  then
  then
  importEntry,				\ Build import list entry
\ Build the colon definition that interprets the function
  align  ^efn ,				\ points to function address cell of import structure
  InExternals? if
    get-current  Externals.wid set-current
    MakeImpVar
    set-current
  else
    MakeImpVar
  endif
  LastImport resolveImport		\ Resolve the import we have just made
;


\ =================
\ Argument handling
\ =================

0 value argSIZE		\ in bytes
0 value argTYPE		\ 0 = integer or bool
0 value argDEFSIGN	\ 0 = unsigned, 1 = signed, for ints < 4 bytes
-1 value argREQSIGN	\ 0 = unsigned, 1 = signed, -1 no request made
0 value argISPOINTER	\ 0 = no, 1 = this is a pointer

: next-item	\ c-addr u char -- 'c-addr 'u c-addr1 u1
\ Return the remaining string and the leftmost item delimited by
\ char. Both strings have leading and trailing spaces removed.
  >r 2dup r> scan			\ ca1 u1 ca2 u2 --
  2swap swap >r				\ ca2 u2 u1 -- ; R: ca1 --
  over - >r				\ ca2 u1 -- ; R: ca1 u1 --
  1 /string -white  2r> -white
; doNotSin

: checkPtr	\ caddr len -- caddr' len' ; SFP013 SFP023
\ The string is a name (no spaces). Check if either or both the
\ leading and trailing characters are '*', and if so mark the
\ argument as a pointer.
  begin
    dup 1 > while			\ -- caddr len
    over c@ [char] * = while
      1 to argISPOINTER  1 /string
  repeat then
  begin
    dup 1 > while
    2dup + 1- c@ [char] * = while	\ check trailing '*'
      1 to argISPOINTER  1-
  repeat then
; doNotSin

-short-branches
: parse-arg     \ c-addr u --
   8 to argSIZE				\ SFP027
   0 to argTYPE				\ default to int
   0 to argDEFSIGN
  -1 to argREQSIGN
   0 to argISPOINTER

  begin
    -leading dup 0>
  while
    \ c-addr u --
    bl next-item checkPtr		\ SFP013
    ['] types voc>wid search-wordlist
    \ -- c-addr-rem u-rem xt flag | c-addr-rem u-rem 0
    dup 0=
    if  ['] noop swap  then		\ Use NOOP if not found
    \ -- c-addr-rem u-rem xt flag
    0= 2 pick 0> and err_impType ?throw	\ bail if token not found and wasn't last!
    execute
  repeat
  2drop

  argREQSIGN -1 =			\ if no sign set,
  if  argDEFSIGN to argREQSIGN  then	\  use default
  argISPOINTER if                       \ if pointer then override to use
    0 to argTYPE			\ treat as int
    0 to argREQSIGN                     \ unsigned cell as data type
    8 to argSIZE			\ use CELL instead?
  then
; doNotSin
+short-branches

: delimiter?    \ char -- flag
\ Return true if the character is a type delimiter
  dup [char] * =			\ pointer
  if  drop  true exit  then
  dup bl =				\ space
  if  drop  true exit  then
  dup 9 =				\ tab ; SFP017
  if  drop  true exit  then
  drop false
;

: splitRet&Name	\ -- c-addr u' c-addr1 u1
\ Get the return type and function name from the input stream.
\   DWORD Pascal  GetLastError( void );
\ ----------------------------^
\ Split the string at the last delimiter, so that caddr/u'
\ contains the return type info and caddr1/u1 contains the
\ function name.
\ ---------------^
\ parse the return type and function name
  [char] ( word count -trailing		\ -- c-addr u ; return type, call conv, name )
\ 2dup type cr key? drop
\ find last delimiter
  2dup + 1 -			\ -- c-addr u ^last
  begin
    dup c@ delimiter? 0=
  while
    1-
  repeat
  1+			\ -- c-addr u c-addr1
\ calculate strings
  dup 3 pick -          \ -- c-addr u c-addr1 u'
  swap rot		\ -- c-addr u' c-addr1 u
  2 pick -		\ -- c-addr u' c-addr1 u-u'
; doNotSin

: setArgData	\ ^ai --
\ Copy the parsed data into the argument data arry or the return
\ result array.
  argSIZE over ai.size !
  argTYPE SINTtype = if			\ deal with integers
    argREQSIGN
    if  SINTtype  else  UINTtype  endif
    swap ai.type !
  else
    argTYPE swap ai.type !
  endif
; doNotSin

: ?refill	\ --
\ Perform REFILL and throw on error.
  refill 0= #-411 ?throw  ;

: parse*/	\ caddr len -- caddr' len --
\ Parse out a C comment of the form "/* ... */", returning the
\ following string. May parse multiple lines.
  begin
    s" */" search 0=			\ look for terminating "*/"
   while
    2drop  ?refill source
  repeat
  2 /string -white			\ step over "/", strip white space
;

: toEOL		\ --
  #tib @ >in !  ;

: stripLeadingComment	\ caddr len -- caddr' len'
\ Remove leading C style comments from the string. The first
\ character is not white space.
  dup 2 <
  if  exit  endif
  over s" //" s=			\ if string starts with //
  if  toEOL  drop 0  exit  endif	\ set length to zero
  over s" /*" s=			\ if string starts with /*
  if  parse*/  toEOL  endif
;

: stripTrailingComment	\ caddr len -- caddr' len'
\ Remove trailing C style comments from the string. Any leading
\ comment has already been removed. All trailing comments are
\ assumed to extend to the end of the line.
  dup 2 <
  if  exit  endif
  2dup s" //" search
  if  nip - -white  toEOL  exit  endif
  2drop  2dup s" /*" search
  if  nip - -white  toEOL  exit  endif
  2drop
;

: (extractArgs)	\ caddr len --
\ Extract arguments between '(' and ')'. Since this may be a
\ multiline operation, this word deals with a single line operation.
  begin
    -leading dup 0>			\ if something to do
  while					\ c-addr u --
    stripLeadingComment dup 0>		\ remove comment, if something left
  while
    [char] , next-item			\ next , separated item
    stripTrailingComment dup if		\ remove comment, if something left
      parse-arg				\ next argument to trailing , or )
      numArgs @ ArgList[] setArgData	\ save data in argument array
      1 numArgs +!
    else				\ discard empty string
      2drop
    endif
  repeat then
  2drop
; doNotSin

-short-branches
: ExtractArgs	\ caddr u --
\ The input string is the return type information. Then parse the
\ rest of the line after the leading '(' character. The function
\ name has already been removed.
  -white parse-arg			\ parse the return type
  ReturnRes setArgData

  0 numArgs !
  begin
    [char] ) parse		        \ look for ')' ; -- caddr len
    'tib @ >in @ 1- 0 max + c@ >r	\ save terminator, may not be ')'
    (extractArgs)
    r> [char] ) <>
   while
    ?refill
  repeat
  numArgs @ 1 = if			\ special case, #args=1 and argsize=0 ; SFP007
    arglist @ 0= if			\ indicates void list
      numArgs off
    endif
  endif
  0 word drop				\ discard rest of line
; doNotSin
+short-branches

: isSingleName	\ caddr len -- ; SFP020
\ Set as name of Forth word and DLL function.
  2dup importName place			\ set name of DLL function
  ExternName place			\ set name of Forth word
; doNotSin

: SingleName	\ "name" -- ; SFP020
\ G Get a name as the name of the Forth word and the DLL function.
  parse-name isSingleName
;


\ =============
\ +N Public API
\ =============

(( for DocGen
: InExternals	\ --
\ *G External imports are created in the *\fo{EXTERNALS} vocabulary.
: InCurrent	\ --
\ *G External imports are created in the wordlist specified by
\ ** *\fo{CURRENT}.
))
also forth definitions

: Extern:	\ "text" --
\ *G Declare an external API reference. See the syntax above.
\ ** The Forth word has the same name as the function in the
\ ** library, but the Forth word name is *\b{not} case-sensitive.
\ ** The length of the function's name may not be longer than a
\ ** Forth word name. For example:
\ *C   Extern: DWORD Pascal GetLastError( void );
  initArgData  setFirstLine		\ SFP022 SFP028
  splitRet&Name
  isSingleName     			\ -- c-addr u ; leaves return type info
  setDefExt				\ set default declaraction type ; SFP006
  ExtractArgs  GenerateImport
;

: AliasedExtern:	\ "forthname" "text" --
\ *G Like *\fo{EXTERN:} but the declared external API reference
\ ** is called by the explicitly specified *\fo{forthname}.
\ ** The Forth word name follows and then the API name.
\ ** Used to avoid name conflicts, e.g.
\ *C AliasedExtern: saccept int accept( HANDLE, void *, unsigned int *);
\ *P which references the Winsock *\fo{accept} function but gives
\ ** it the Forth name *\fo{SACCEPT}. Note that here we use the
\ ** fact that formal parameter names are optional.
  initArgData  setFirstLine		\ SFP022 SFP028
  parse-name ExternName place		\ process Forth name
  splitRet&Name importName place	\ -- c-addr' u'
  setDefExt				\ set default declaration type ; SFP006
  ExtractArgs GenerateImport
;

: LocalExtern:	\ "forthname" "text" --
\ *G As *\fo{AliasedExtern:}, but the import is always built into
\ ** the *\fo{CURRENT} wordlist.
  InExternals?   0 -> InExternals?
  AliasedExtern:
  to InExternals?
;

: extern	\ "text" --
\ *G An alias for *\fo{EXTERN:}.
  extern:  ;

\ SFP017...
: ExternVar	\ "<name>" -- ; ExternVar <name>
\ *G Used in the form
\ *C   ExternVar <name>
\ *P to find a variable in a DLL or shared library. When executed,
\ ** *\fo{<name>} returns its address.
  SingleName  genExternVar		\ generate variable ; SFP020
;

: AliasedExternVar	\ "<forthname>" "<dllname>" --
\ *G Used in the form
\ *C   AliasedExternnVar <forthname> <varname>
\ *P to find a variable in a DLL or shared library. When executed,
\ ** *\fo{<forthname>} returns its address.
  parse-name ExternName place		\ process Forth name
  parse-name importName place		\ variable name
  genExternVar				\ generate variable
;
\ ...SFP017

: +DebugExterns	\ --
  1 to ShowExtData? ;
: -DebugExterns	\ --
  0 to ShowExtData? ;

: .Externs	\ -- ; display EXTERNs
\ *G Display a list of the external API calls.
  0 import-func-link
  begin
    @ dup halt? 0= and
   while				\ -- count item
    cr  dup >funcname .z$
    out @ #31 and #32 swap - spaces
    dup >funcaddr @ .lword
    swap 1+ swap
  repeat
  drop
  cr . ." Externs"
;

: #BadExterns	\ -- u
\ *G Silently return the number of unresolved external API calls.
  0 import-func-link
  begin
    @ dup
   while				\ -- count item
    dup >funcaddr @ 0= if
      swap 1+ swap
    endif
  repeat
  drop
;

: .BadExterns	\ --
\ *G Display a list of any external API calls that have not been
\ ** resolved.
  0 import-func-link
  begin
    @ dup halt? 0= and
   while				\ -- count item
    dup >funcaddr @ 0= if
      cr  dup >funcname .z$
      swap 1+ swap
    endif
  repeat
  drop
  cr . ." Unresolved Externs"
;

: func-pointer	\ xt -- addr
\ *G Given the XT of a word defined by *\fo{EXTERN:} or friends,
\ ** returns the address that contains the run-time address.
  >link cell - @
;

: func-loaded?	\ xt -- addr|0
\ *G Given the XT of a word defined by *\fo{EXTERN:} or friends,
\ ** returns the address of the DLL function in the DLL,
\ ** or 0 if the function has not been loaded/imported yet.
  >link cell - @ @
;

defer func-not-loaded	\ c-addr u --
\ Client special
  assign 2drop to-do func-not-loaded

: all-func-loaded?	\ -- flag
\ Client special
  true >R
  import-func-link @
  begin
    dup
  while
    dup >funcaddr @ 0= if
      dup >funcname zcount func-not-loaded
      r> drop false >r
    then
    @
  repeat
  drop r>
;
previous definitions


\ ******************************
\ *S Pre-Defined parameter types
\ ******************************
\ *P The types known by the system are all found in the vocabulary
\ ** *\fo{TYPES}. You can add new ones at will. Each *\fo{TYPE}
\ ** definition modifies one or more of the following *\fo{VALUE}s.                        )
\ *D argSIZE      Size in bytes of data type.
\ *D argDEFSIGN   Default sign of data type if no override is supplied.
\ *D argREQSIGN   Sign OverRide. This and the previous use 0 = unsigned
\ **              and 1 = signed.
\ *D argISPOINTER 1 if type is a pointer, 0 otherwise

\ *P Each *\fo{TYPES} definition can either set these flags
\ ** directly or can be made up of existing types.

\ *P Note that you should explicitly specify a calling convention
\ ** for every function defined.


\ ======================
\ *N Calling conventions
\ ======================

also types definitions

: "C"           \ --
\ *G Set Calling convention to "C" standard. Arguments are
\ ** reversed, and the caller cleans up the stack.
  0 to CalleeCleaned?  TRUE to Require_NRev
;

: "PASCAL"      \ --
\ *G Set the calling convention to the "PASCAL" standard as used
\ ** by Pascal compilers. Arguments are *\b{not} reversed, and the
\ ** called routine cleans up the stack.
  1 to CalleeCleaned?  FALSE to Require_NRev
;

: WinApi	\ --
\ *G A synonym for *\fo{PASCAL}.
  "C"
;

: VC++		\ --
\ *G Defines the calling convention as being for a C++ member
\ ** function which requires "this" in the ECX register.
\ ** The function must be defined with an explicit this
\ ** pointer (void * this). Because exported VC++ member
\ ** functions can have either "C" or "PASCAL" styles, the this
\ ** pointer must be positioned so that it is leftmost when
\ ** reversed (C/WINAPI/StdCall style) or is rightmost when
\ ** not reversed ("PASCAL" style). See also the later section
\ ** on interfacing to C++ DLLs.
  1 to ObjPtr?
;

: R>L	\ --
\ *G By default, arguments are assumed to be on the Forth stack
\ ** with the top item matching the rightmost argument in the
\ ** declaration so that the Forth parameter order matches that
\ ** in the C-style declaration.
\ ** *\fo{R>L} reverses this.
  FALSE to L>R?
;

: L>R	\ --
\ *G By default, arguments are assumed to be on the Forth stack
\ ** with the top item matching the rightmost argument in the
\ ** declaration so that the Forth parameter order matches that
\ ** in the C-style declaration.
\ ** *\fo{L>R} confirms this.
  TRUE to L>R?
;

previous definitions


\ ==============
\ *N Basic Types
\ ==============

also types definitions

: unsigned      \ --
\ *G Request current parameter as being unsigned.
  0 to argREQSIGN  ;

: signed        \ --
\ *G Request current parameter as being signed.
  1 to argREQSIGN  ;

: int           \ --
\ *G Declare parameter as integer. This is a signed 32 bit quantity
\ ** unless preceeded by *\fo{unsigned}.
  4 to argSIZE  1 to argDEFSIGN  ;

: char          \ --
\ *G Declare parameter as character. This is a signed 8 bit quantity
\ ** unless preceeded by *\fo{unsigned}.
  1 to argSIZE  1 to argDEFSIGN  ;

: void          \ --
\ *G Declare parameter as void. A *\fo{VOID} parameter has no
\ ** size. It is used to declare an empty parameter list, a null
\ ** return type or is combined with *\fo{*} to indicate a generic
\ ** pointer.
  0 to argSIZE  0 to argDEFSIGN  ;

: *		\ --
\ *G Mark current parameter as a pointer.
  1 to argISPOINTER  ;
: **		\ --
\ *G Mark current parameter as a pointer.
  1 to argISPOINTER  ;
: ***		\ --
\ *G Mark current parameter as a pointer.
  1 to argISPOINTER  ;

: const ;	\ --
\ *G Marks next item as *\b{constant} in C terminology. Ignored
\ ** by VFX Forth.

: int32		\ --
\ *G A 32bit signed quantity.
  4 to argSIZE  1 to argDEFSIGN  ;

: int16		\ --
\ *G A 16 bit signed quantity.
   2 to argSIZE  1 to argDEFSIGN  ;

: int8		\ --
\ *G An 8 bit signed quantity.
   1 to argSIZE  1 to argDEFSIGN  ;

: uint32	\ --
\ *G 32bit unsigned quantity.
  unsigned int  ;

: uint16	\ --
\ *G 16bit unsigned quantity.
  unsigned int  2 to argSIZE  ;

: uint8		\ --
\ *G 8bit unsigned quantity.
  1 to argSIZE  0 to argDEFSIGN  ;

: Long		\ --
\ *G A 64 bit signed or unsigned integer. At run-time, the argument
\ ** is taken from the Forth data stack as a normal Forth single
\ integer.
  8 to argSIZE  1 to argDEFSIGN  ;

: LongLong	Long  ;
\ *G A 64 bit signed quantity. See *\fo{Long} above.

: SHORT		\ --
\ *G For most compilers a *\b{short} is a 16 bit signed item,
\ ** unless preceded by *\fo{unsigned}.
  2 to argSIZE  1 to argDEFSIGN  ;

: BYTE		\ --
\ *G An 8 bit unsigned quantity.
  1 to argSIZE  0 to argDEFSIGN  ;

: float		\ --
\ *G 32 bit float.
  4 to argSIZE  0 to argDEFSIGN  FLOATtype to argTYPE  ;

: double	\ --
\ *G 64 bit float.
  8 to argSIZE  0 to argDEFSIGN  FLOATtype to argTYPE  ;

: point_double	\ --
\ *G MPEism for points using double floats.
\ ** FOR OUTPUTS ONLY.
  #16 to argSIZE  0 to argDEFSIGN  dPOINTtype to argTYPE  ;

: complex_double	\ --
\ *G MPEism for complex numbers using double floats.
\ ** FOR OUTPUTS ONLY.
  point_double  ;

: bool1		\ --
\ *G One byte boolean.
  1 to argSIZE  0 to argDEFSIGN  BOOLtype to argTYPE  ;
: bool4		\ --
\ *G Four byte boolean.
  int  ;
: bool8		\ --
\ *G Eight byte boolean.
  int  ;

: ...		\ --
\ *G The parameter list is of unknown size. This is an indicator
\ ** for a C varargs call. Run-time support for this varies between
\ ** operating system implementations of VFX Forth. Test, test,
\ ** test.
  1 to varargs?
;

previous definitions


\ ==============
\ *N Linux Types
\ ==============

[defined] Target_x64_Linux [if]

: (setDefExt)	\ --
\ set up the Linux defaults for EXTERN: calls.
  #16 to FrameAlignment  0 to AlignInFrame?
  TRUE to REQUIRE_NREV  0 to CalleeCleaned?
  TRUE to L>R?
  0 to ObjPtr?  0 to StructRet?  0 to varargs?
;
assign (setDefExt) to-do setDefExt

also types definitions
: OSCALL	"C"  ;
\ *G Used for portable code to avoid three sets of declarations.
\ ** For Windows, this is a synonym for *\fo{PASCAL} and under
\ ** Linux this is a synonym for *\fo{"C"}.
: FILE		uint32  ;
\ *G Always use as *\fo{FILE * stream}.
: DIR		uint32  ;
\ *G Always use as *\fo{DIR * stream}.
: size_t	Long  ;
\ *G Linux type for unsigned INT.
: off_t		Long  ;
\ *G Linux type for unsigned INT.
: int32_t	int32  ;
\ *G Synonym for *\fo{int32}.
: int16_t	int16  ;
\ *G Synonym for *\fo{int16}.
: int8_t	int8  ;
\ *G Synonym for *\fo{int8}.
: uint32_t	uint32  ;
\ *G Synonym for *\fo{uint32}.
: uint16_t	uint16  ;
\ *G Synonym for *\fo{uint16}.
: uint8_t	uint8  ;
\ *G Synonym for *\fo{uint8}.
: time_t	Long  ;
\ *G Number of seconds since midnight UTC of January 1, 1970.
: clock_t	Long  ;
\ *G Processor time in terms of CLOCKS_PER_SEC.
: pid_t		unsigned int  ;
\ *G Process ID.
: uid_t		unsigned int  ;
\ *G User ID.
: mode_t	unsigned short  ;
\ *G File mode.
previous definitions

[then]		\ Target_x64_Linux

\ =================
\ *N Mac OS X Types
\ =================

[defined] Target_x64_OSX [if]

: (setDefExt)	\ --
\ set up the Mac OS X defaults for EXTERN: calls.
  #16 to FrameAlignment  0 to AlignInFrame?
  TRUE to REQUIRE_NREV  0 to CalleeCleaned?
  TRUE to L>R?
  0 to ObjPtr?  0 to StructRet?  0 to varargs?
;
assign (setDefExt) to-do setDefExt

also types definitions
: OSCALL	"C"  ;
\ *G Used for portable code to avoid three sets of declarations.
\ ** For Windows, this is a synonym for *\fo{PASCAL} and under
\ ** OS X this is a synonym for *\fo{"C"}.
: FILE		uint32  ;
\ *G Always use as *\fo{FILE * stream}.
: DIR		uint32  ;
\ *G Always use as *\fo{DIR * stream}.
: size_t	Long  ;
\ *G OS/X type for file size.
: off_t		Long  ;
\ *G OS/X type for file position.
: int32_t	int32  ;
\ *G Synonym for *\fo{int32}.
: int16_t	int16  ;
\ *G Synonym for *\fo{int16}.
: int8_t	int8  ;
\ *G Synonym for *\fo{int8}.
: uint32_t	uint32  ;
\ *G Synonym for *\fo{uint32}.
: uint16_t	uint16  ;
\ *G Synonym for *\fo{uint16}.
: uint8_t	uint8  ;
\ *G Synonym for *\fo{uint8}.
: time_t	Long  ;
\ *G Number of seconds since midnight UTC of January 1, 1970.
: clock_t	Long  ;
\ *G Processor time in terms of CLOCKS_PER_SEC.
: pid_t		unsigned int  ;
\ *G Process ID.
: uid_t		unsigned int  ;
\ *G User ID.
: mode_t	unsigned short  ;
\ *G File mode.
previous definitions

[then]		\ Target_386_OSX


\ **********************
\ *S Compatibility words
\ **********************
\ *P These words are mainly for users converting code from other
\ ** Forth systems. They provide a more "Forth-like" interface that
\ ** does not permit copy/paste from the operating system header files
\ ** or documentation.

: skipTo(	\ -- ; )
\ Step forward to a space-delimited '('. )
  begin
    parse-name dup 0= abort" missing ("	\ error at end of line ; )
    s" (" compare 0=			\ look for '('; )
  until
;

: param? \ caddr len -- flag
\ should ignore S: F: N: R: stack indicators as parameters.
\ leave a true flag when genuine parameter, false otherwise.
        2dup upper
        2dup s" S:" compare dup 0= if nip nip  exit then
   drop 2dup s" F:" compare dup 0= if nip nip  exit then
   drop 2dup s" N:" compare dup 0= if nip nip  exit then
   drop      s" R:" compare dup 0= if nip nip  exit then
   -1 or ;

: parse-params	\ -- #in #out
\ Parse a parameter list of the form:
\   ( a b c d -- e f )
  skipTo(				\ scan up to '('    )
  0 begin				\ -- u ; begin parsing
    parse-name  2dup s" --" compare	\ compare leaves zero flag when equal !
   while				\ -- u caddr len
    2dup param? if                      \ do we have a genuine param?
      s" ..." compare if		\ -- u
        dup 0 >=			\ if not varargs (* not ...  -rt *)
        if  1+  endif			\  count up
      else
        drop -1				\ #in=-1 for varargs  (* is ...  -rt *)
      endif
    else  2drop  then			\ ignore invalid param, try again
  repeat
   2drop                                \ get rid of dupped word+count
   0 begin                              \ begin parsing outputs
    parse-name s" )" compare
   while
     1+
  repeat
;

: doubles?  \ -- f
\ check for fp stackpicture
   >in @ parse-name rot >in ! s" (" compare 0= ;	\ ) for syntax highlighter

: make-return \ #double-rets #int-rets -- type
\ added 3 as double return type for now
   dup  0= if drop 3 *  exit then       \ no int-ret, 3 is doubles return type, allow for void, 1 and 2 doubles
   over 0= if nip       exit then       \ no double-ret, allows 1 int return. 2 ints not used
   cr ." Not both integer and double returns !"
   abort ;

: parseArgs  \ <text> -- #in-ints #in-doubles out-type
\ Parse a parameter list of the form:
\   ( a b c d -- e ) and optional second ( r1 r2 r3 r4 - r5 r6 )
   parse-params      ( #ints #int-rets )
   doubles? 0= if 0 swap exit then
   parse-params      ( #ints #int-rets #doubles #double-rets )
   rot make-return ;

: setVoid	\ struct --
  0 over ai.Size !
  UINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.PSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

((
: setInt32	\ struct --
  4 over ai.Size !
  SINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.PSPoff !			\ offset on Forth parameter stack as if TOS uncached
;
))

: setInt64	\ struct --
  8 over ai.Size !
  SINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.PSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: setDouble	\ struct --
  8 over ai.Size !
  FLOATtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.FSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: set2Doubles	\ struct --
  16 over ai.Size !
  dPOINTtype over ai.Type !
  0 over ai.FrameOff !			\ offset on operating system frame to data
  0 swap ai.FSPoff !			\ offset on Forth parameter stack as if TOS uncached
;

: setReturn	\ type --
\ set the return data
  case
    0 of  ReturnRes setVoid  endof	\ 0 becomes void
    1 of  ReturnRes setInt64  endof	\ all ints are signed 64
    3 of  ReturnRes setDouble   endof	\ all floats are 64 bit
    6 of  ReturnRes set2Doubles endof	\ not yet
      -1 abort" Bad return data		\ no 2 ints or 2 doubles for now
  endcase
;

-short-branches
: setArgs    \ #in-ints #in-doubles out-type --
\ Generate the data handling structures. Varargs are indicated
\ by #in=-1.
   swap >r ( *)
   over 0< if
      setReturn  r> 2drop  1 to varargs?
   else
      setReturn  0 to varargs?
      dup numArgs ! dup NumInts ! 0
      ?do  i ArgList[] setInt64  loop
      numArgs @   ( start index)
      r> ( *) dup numArgs +! NumFloats !
      numArgs @ swap
      ?do  i ArgList[] setDouble  loop
   then
;
+short-branches

\ *P The following section provides shared library imports in the form:
\ *C   function: foo  ( a b c d -- x ) ( F: r1 r2 -- rx )
\ *P where the brackets *\b{must} be space delimited. The first argument
\ ** list represents the arguments seen by Forth as integers. The second
\ ** arguments list is optional, and is used when F.P. doubles are involved.
\ ** Imports use the default calling convention for the operating system.

also forth definitions

: FUNCTION:	\ "<name>" "<parameter list>" --
\ *G Generate a reference to an external function. The Forth name
\ ** is the same as the name of the external function.
\ ** The first parameter list is a Forth integers list plus an optional
\ ** second Forth doubles list. There should always be an integer list,
\ ** the first list, even when not used.
\ ** Use in the form:
\ *C  function: foo1 ( a b c d -- )
\ *C  function: foo2 ( a b c d -- e )
\ *C  function: foo3 ( a b c d -- e ) ( r1 r2 r3 -- )
\ *C  function: foo4 ( a b c d -- ) ( r1 r2 r3 -- r4 )
\ *C  function: foo5 ( a b c d -- ) ( r1 r2 r3 -- r4 r5 )
\ *C  function: foo6 ( -- ) ( r1 r2 r3 -- r4 )
\ *C  function: foo7 ( -- ) ( r1 r2 r3 -- r4 r5 )
\ *C  function: foo7 ( -- a ) ( r1 r2 r3 -- )
\ *P The inclusion of Forth stack indicators like S: F: N: R: is optional.
\ *P They will be ignored while parsing and are only informative on which
\ *P Forth stack you must place the parameters.
\ *P The returned value may be 0, 1 integer or 0, 1, 2 doubles
\ ** corresponding to void, int/long and 64 bit double
  initArgData  setFirstLine		\ SFP022
  singleName				\ name of Forth word and DLL func ; SFP020
  setDefExt				\ set default declaraction type
  parseArgs setArgs			\ the parameter parsing and setup for...
  GenerateImport
;

: ASCALL:	\ "<synonym-name>" "<name>" "<parameter list>" --
\ *G Generate a reference to an external function. The Forth name
\ ** is not the same as the name of the external function.
\ ** Use in the form:
\ *C  ascall: forthname funcname ( a b c d -- e )
  initArgData  setFirstLine		\ SFP022
  parse-name ExternName place		\ set name of Forth word
  parse-name ImportName place		\ set name of DLL function
  setDefExt				\ set default declaraction type
  parseArgs setArgs                     \ the parameter parsing and setup for...
  GenerateImport
;

: GLOBAL:	\ "<name>" --
\ *G Generate a reference to an external variable.
\ ** Use in the form:
\ *C  global: varname
  singleName genExternVar		\ generate variable ; SFP020
;

previous definitions


\ *********
\ Test code
\ *********

((

+debugExterns
extern: double foo( int a, int b, double c, float d, double e);
extern: long foo2( int a, int b );
extern: long foo3( int a, int b, double c, float d, double e);
-debugExterns

dis foo
dis foo2
dis foo3

))

\ ======
\ *> ###
\ ======

decimal

