\ Extern.fth - DLL and shared library access for VFX Forth

((
Copyright (c) 2001, 2003, 2006, 2007, 2008, 2009
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====
Support C // comments in argument parsing.

Change history
==============
20091123 SFP009	Refactored with startup.fth.
20090610 SFP008 Added documentation for FLAG from Marcel Hendrix.
20090609 SFP007 Added .BADEXTERNS.
20090502 SFP006	Added code for Windows protection with floats.
20020323 MPE005 Refactored client specific code in prologue and
                epilogue.
20090319 MPE004 Modified after distribution list discussions.
20090310 MPE003 Refactored.
20090131 MPE002 Tightened up.
		Put most words in separate vocabulary.
20090121 MPE001 Ported from the Windows/Linux version.
))


\ =========
\ *! extern
\ *T Functions in shared libraries
\ =========

((
License
=======
This code can be used as the basis of an implementation of the
EXTERN: interface for other Forth systems. Copyright of the source
notation and behaviour is retained by MPE until another copyright
has been assigned to a third party by MPE.

Controlled words can be identified by comments starting
  \ *G
below the name line.

If you implement the notation and behaviour and release the source
code, this license must be included in the source code.

Rationale and Requests
======================
The rationale for releasing this code is to encourage portability
of Forth source code that accesses shared libraries.

The point of retaining copyright at MPE is not to prevent you
changing and improving the system, it is to provide a central
point of management for the notation and its behaviour. In the
long term, copyright of the EXTERN: notation will be assigned
to a third party.

If your code is better, faster, shorter or more functional, we
request that you contribute it back to the maintainer, currently
Stephen Pelc, stephen@mpeforth.com. Similarly ports for currently
unsupported Forth systems will be gratefully received.

VFXisms
=======
DONOTSIN can be removed or be a NOOP.

.LWORD ( x -- ) and .DWORD ( x -- )
Display x as an unsigned hex number.

: -leading      bl skip  ;
: -white        -leading -trailing  ;

The code is based around three words which can be implemented
for any Windows or Unix-style host.
  LoadLibrary      \ zaddr -- handle|0
  FreeLibrary      \ handle -- status
  GetProcAddress   \ handle zaddr -- addr|0


Legacy code
===========
The words L>R and R>L are for compiling library code rom other
Forth systems which may use a right-to-left stack order.
))

only forth definitions decimal

[undefined] externals [if]
vocabulary externals
\ Where library and imported function definitions live.
[then]
vocabulary types	\ --
\ Where the "C" style types go for the *\fo{EXTERN:} notation.
vocabulary Extern.Voc
\ Where the guts of the EXTERN: mechanism live.

also Extern.Voc definitions


\ ***************
\ *S Introduction
\ ***************
\ *P VFX Forth supports calling external API calls in dynamic link
\ ** libraries (DLLs) for Windows and shared libraries in Linux and
\ ** other Unix-derived operating systems.
\ ** Various API libraries export functions in a variety of methods
\ ** mostly transparent to programmers in languages such as C,
\ ** Pascal and Fortran. Floating point data is supported for use
\ ** with *\i{Lib\x86\Ndp387.fth}.

\ *P Before a library function can be used, the library itself must
\ ** be declared, e.g.
\ *E LIBRARY: Kernel32.dll

\ *P Access to functions in a library is provided by the
\ ** *\fo{EXTERN:} syntax which is similar to a C style function
\ ** prototype, e.g.
\ *E EXTERN: int PASCAL SendMessage(
\ **   HWND hwnd, DWORD mesg, WPARAM wparam, LPARAM lparam
\ ** );
\ *P This can be used to prototype the function *\b{SendMessage}
\ ** from the Microsoft Windows API, and produces a Forth word
\ ** *\fo{SendMessage}.
\ *C   SendMessage  \ hwnd mesg wparam lparam -- int

\ *P For Linux and other Unices, the same notation is used.
\ ** The default calling convention is nearly always applicable.
\ ** The following example shows that definitions can occupy more
\ ** than one line. It also indicates that some token separation may
\ ** be necessary for pointers:
\ *E Library: libc.so.6
\ **
\ ** Extern: int execve(
\ **   const char * path,
\ **   char * const argv[],
\ **   char * const envp[]
\ ** );
\ *P This produces a Forth word *\fo{execve}.
\ *C   execve       \ path argv envp -- int

\ *P The parser used to separate the tokens is not ideal. If you
\ ** have problems with a definition, make sure that *\fo{*}
\ ** tokens are white-space separated. Formal parameter names,
\ ** e.g. *\i{argv} above are ignored. Array indicators, *\i{[]}
\ ** above, are also ignored when part of the names.

\ *P The input types may be followed by a dummy name which is
\ ** discarded. Everything on the source line after the closing
\ ** ')' is discarded.

\ *P From VFX Forth v4.3 onwards, the use of PASCAL is not required
\ ** as PASCAL is the default calling convention in the Windows
\ ** version. The default for the Linux version is "C". The default
\ ** is always used unless overridden in the declaration.


\ *********
\ *S Format
\ *********
\ *E EXTERN: <return> [ <callconv> ] <name> '(' <arglist> ')' ';'
\ **
\ ** <return>    := { <type> [ '*' ] |  void }
\ ** <arg>       := { <type> [ '*' ] [ <name> ] }
\ ** <args>      := { [ <arg>, ]* <arg> }
\ ** <arglist>   := { <args> | void }	Note: "void, void" etc. is illegal.
\ ** <callconv>  := { PASCAL | WINAPI | STDCALL | "PASCAL" | "C" }
\ ** <name>      := <any Forth acceptable namestring>
\ ** <type>      := ... (see below, "void" is a valid type)

\ *P Note that during searches <name> is passed to the operating
\ ** system exactly as it is written, i.e. case sensitive. The
\ ** Forth name is case-insensitive.

\ *P As a standard Forth's string length for dictionary names is
\ ** only guaranteed up to 31 characters for portable source code,
\ ** very long API names can cause problems. Therefore the word
\ ** *\fo{ALIASEDEXTERN:} allows separate specification of API
\ ** and Forth names (see below). *\fo{ALIASEDEXTERN:} also
\ ** solves problems when API functions only differ in case
\ ** or their names conflict with existing Forth word names.

\ **********************
\ *S Calling Conventions
\ **********************
\ *P In the discussion *\b{caller} refers to the Forth system
\ ** (below the application layer and *\b{callee} refers to a
\ ** a function in a DLL or shared library. The *\fo{EXTERN:}
\ ** mechanism supports three calling conventions.
\ *(
\ *B C-Language: *\fo{"C"} *\br{}
\ **  Caller tidies the stack-frame.
\ ** The arguments (parameters) which are passed to the library
\ ** are reordered. This convention can be specified by using
\ ** *\fo{"C"} after the return type specifier and
\ ** before the function name.
\ *B Pascal language: *\fo{"PASCAL"} *\br{}
\ ** Callee removes arguments from the stack frame. This is
\ ** invisible to the programmer at the application layer
\ ** The arguments (parameters) which are passed to the library are
\ ** not reordered. This convention is specified by
\ ** *\fo{"PASCAL"} after the return type specifier and
\ ** before the function name.
\ *B Windows API: *\fo{WINAPI | PASCAL | STDCALL} *\br{}
\ ** In nearly all cases (but *\b{not all}), calls to
\ ** Windows API functions require C style argument reversal and
\ ** the called function cleans up. Specify this convention with
\ ** *\fo{PASCAL}, *\fo{WinAPI} or *\fo{StdCall} after the
\ ** return type specifier and before the function name.
\ *)

\ *P Unless otherwise specified, the Forth system's default
\ ** convention is used. Under Windows this is *\fo{WINAPI} and
\ ** under Linux and other Unices it is *\fo{"C"}.

\ *************************
\ *S Promotion and Demotion
\ *************************
\ *P The system generates code to either promote or demote non-CELL
\ ** sized arguments and return results which can be either signed or
\ ** unsigned. Although Forth is an un-typed language it must deal with
\ ** libraries which do have typed calling conventions. In general
\ ** the use of non-CELL arguments should be avoided but return
\ ** results should be declared in Forth with the same size as the C or
\ ** PASCAL convention documented.

\ ********************
\ *S Argument Reversal
\ ********************
\ *P The default calling convention for the host operating system
\ ** is used. The right-most argument/parameter in the C-style
\ ** prototype is on the top the Forth data stack.
\ ** When calling an external function the parameters are reordered
\ ** if required by the operating system; this is to enable the
\ ** argument list to read left to right in Forth source as well
\ ** as in the C-style operating system documentation.

\ *P Under certain conditions, the order can be reversed. See the
\ ** words *\fo{"C"} and *\fo{"PASCAL"} which define the order for
\ ** the operating system. See *\fo{L>R} and *\fo{R>L} which define
\ ** the Forth stack order with respect to the arguments in the
\ ** prototype.


\ **********************************
\ *S Controlling external references
\ **********************************

also forth definitions
1 value ExternWarnings?	\ -- n
\ *G Set this true to get warning messages when an external reference
\ ** is redefined.
0 value ExternRedefs?	\ -- n
\ *G If non-zero, redefinitions of existing imports are permitted.
\ ** Zero is the default for VFX Forth so that redefinitions of
\ ** existing imports are ignored.
1 value InExternals?	\ -- n
\ *G Set this true if following import definitions are to be in
\ ** the *\fo{EXTERNALS} vocabulary, false if they are to go into
\ ** the wordlist specified in *\fo{CURRENT}. Non-Zero is the
\ ** default for VFX Forth.
: InExternals	\ --
\ *G External imports are created in the *\fo{EXTERNALS} vocabulary.
  1 -> InExternals?  ;
: InCurrent	\ --
\ *G External imports are created in the wordlist specified by
\ ** *\fo{CURRENT}.
  0 -> InExternals?  ;
previous definitions

' externals voc>wid constant Externals.wid
\ The wordlist ID of the EXTERNALS vocabulary.


\ ******************
\ *S Library Imports
\ ******************
\ *P In VFX Forth, libraries are held in the *\fo{EXTERNALS}
\ ** vocabulary, which is part of the minimum search order.
\ ** Other Forth systems may use the *\fo{CURRENT} wordlist.

\ *P For turnkey applications, initialisation, release and
\ ** reload of required libraries is handled at start up.

(( for DocGen
variable lib-link
\ *G Anchors the chain of dynamic/shared libraries.
))

struct /libfn	\ -- size
  int >liblink		\ link to previous library
  int >libaddr		\ library Id/handle/address, depends on O/S
  0 field >libname	\ zero terminated string of library name
end-struct

: clear-libs    \ --
\ +G Clear all library addresses and free DLLs.
  lib-link
  begin
    @ dup
  while
    0 over >libaddr !
    >liblink
  repeat
  drop
;

also forth definitions

: init-lib	\ libfn --
\ Given the address of a library structure, load the library.
  dup >libname LoadLibrary swap >libaddr !
;

: init-libs     \ --
\ *G Release and reload the required libraries.
  clear-libs  lib-link
  begin
    @ dup
  while
    dup init-lib  >liblink
  repeat
  drop
;
' init-libs AtCold

: find-libfunction	\ z-addr -- address|0
\ *G Given a zero terminated name, attempt to find the named function
\ ** somewhere within the already active libraries.
  >r  lib-link
  begin					\ -- link ; R: -- z$
    @ dup
  while                 		\ -- struct ; R: -- z$
    dup >libaddr @ r@ GetProcAddress dup \ -- struct addr|0 addr|0 ; R: -- z$
    if  nip  r> drop exit  endif
    drop >liblink
  repeat
  drop  r> drop 0
;

: .Libs         \ --
\ *G Display the list of declared libraries.
  lib-link
  begin
    @ dup
  while
    dup >libname cr .z$
    out @ 31 and 32 swap - spaces
    dup >libaddr @ .lword
    >liblink
  repeat
  drop
;

((
: CheckExtName	\ caddr len -- flag
\ Return true if the name already exists and should not be
\ redefined.
  get-current search-wordlist if
    drop  ExternWarnings? if
      cr ." WARNING: Redefinition of Extern:/Library: " ExternName $.
      ExternRedefs? 0=
      if  ."  ignored"  endif
    then
    ExternRedefs? 0=
    if  true exit  then
  endif
  0
;
))

: Library:      \ "<name>" -- ; -- loadaddr|0
\ *G Register a new library by name.
\ ** Use in the form:
\ *C   LIBRARY: <name>
\ *P Executing *\fo{<name>} later will return its load address.
\ ** This is useful when checking for libraries that may not be
\ ** present. After definition, the library is the first one
\ ** searched by import declarations.
  >in @  get-current  InExternals?	\ save >IN, which wordlist?
  if  Externals.wid set-current  then	\ -- >in wid
  create
    set-current  >in !			\ restore wordlist and >IN
    here				\ start of  structure
    lib-link link,			\ link to previous library
    0 ,					\ load address, filled in later
    parse-name z$,			\ lay name as zero terminated string
\    bl word $>z,			\ lay name as zero terminated string
    init-lib
  does>
    >libaddr @
;

: topLib	\ libfn --
\ *G Make the library structure the top/first in the library
\ ** search order.
  lib-link  2dup delLink  AddLink
;

: firstLib	\ "<name>" --
\ *G Make the library top of the library search order. Use during
\ ** interpretation in the form:
\ *C   FirstLib <name>
\ *P to make the library top of the search order. This is useful
\ ** when you know that there may be several functions of the
\ ** same name in different libraries.
  ' >body topLib
;

: [firstLib]	\ "<name>" --
\ *G Make the library top of the library search order. Use during
\ ** compilation in the form:
\ *C   [firstLib] <name>
\ *P to make the library top of the search order. This is useful
\ ** when you know that there may be several functions of the
\ ** same name in different libraries.
  ' >body postpone literal  postpone topLib
; immediate
previous definitions


\ *******************
\ *S Function Imports
\ *******************
\ *P Function declarations in shared libraries are compiled into
\ ** the *\fo{EXTERNALS} vocabulary. They form a single linked
\ ** list. When a new function is declared, the list of previously
\ ** declared libraries is scanned to find the function. If the
\ ** function has already been declared, the new definition is
\ ** ignored if *\fo{ExternRedefs?} is set to zero. Otherwise,
\ ** the new definition overrides the old one as is usual
\ ** in Forth.

\ *P In VFX Forth, *\fo{ExternRedefs?} is set by default to zero.

(( for DocGen
variable import-func-link	\ -- addr
\ *G Anchors the chain of imported functions in shared libraries.
))

[defined] Target_386_Windows [if]
: resolve-libfunction   \ z-addr -- address|0
\ +G Look up a definition in loaded libraries and return XT or 0.
\ +* Automatically attempts to find a version with 'A' appended for
\ +* Windows.
  { | tn[ MAX_PATH ] -- }
  [ also system ]
  dup find-libfunction dup		\ look for raw name
  if  nip exit  then
  drop
  tn[ MAX_PATH erase			\ look for name with 'A' appended
  zcount tn[ swap move
  [char] A tn[ zcount + c!
  tn[ find-libfunction
  [ previous ]
;
[else]
: resolve-libfunction   \ z-addr -- address|0
  [ also system ]
  find-libfunction			\ look for raw name
  [ previous ]
;
[then]

: resolveImport	\ struct --
\ Resolve the import whose structure is given.
  dup >libname resolve-libfunction swap >libaddr !
;

also forth definitions
: ExternLinked	\ c-addr u -- address|0
\ *G Given a string, attempt to find the named function in  the
\ ** already active libraries. Returns zero when the function is
\ ** not found.
  { | temp[ 256 ] -- }
  255 min temp[ zplace
  temp[ resolve-libfunction
;

: init-imports  \ --
\ *G Initialise Import libraries. *\fo{INIT-IMPORTS} is called by
\ ** the system cold chain.
  import-func-link @
  begin
    dup
  while
    dup resolveImport
    >liblink @
  repeat
  drop
;
' init-imports ATCOLD
previous definitions


\ ********************************
\ +S Import parsing and generation
\ ********************************

\ ======================
\ +N Data and structures
\ ======================

\ These are constants used as THROW codes with associated text messages.
SysErrdef err_ImpType	"Invalid type in EXTERN: import"
SysErrDef err_CGsize	"Invalid size in EXTERN: argument preparation"
SysErrDef err_RetType	"Invalid return type"
SysErrDef err_RetSize	"Invalid return size"

\ argument types - BOOL could be considered a form of UINT, but
\ is O/S specific.
0 constant SINTtype	\ -- n ; default is signed int
1 constant UINTtype	\ -- n ; signed int
2 constant BOOLtype	\ -- n ; C boolean
3 constant FLOATtype	\ -- n ; float/double/ext ....

struct /argItem	\ -- len
  int ai.Size		\ argument size in bytes, 1=byte/char, 2=short, 4=int ...
  int ai.type		\ 0=uint, 1=sint, 2=bool, 3=float
  int ai.FrameOff	\ offset on operating system frame to data
  int ai.PSPoff		\ offset on Forth parameter stack as if TOS uncached
end-struct
#32 /argitem * constant /argList	\ size of the argument list, max 32 items


\ Operating system dependent items. These VALUEs are set to
\ defaults at the start of each external import declaration.

16 value FrameAlignment	\ -- u
\ Set to alignment of stack frame in memory. Must be a power of two.
4 value FrameBackup	\ -- u
\ If the stack must be aligned after the call, set this to the
\ number of byes used by the call
true value AlignInFrame? \ -- flag
\ True if items on the stack frame are padded for data alignment.
true value REQUIRE_NREV	\ -- flag
\ Set true if arguments are reversed by default (C argument order),
\ i.e. the leftmost argument in the argument declaration list is
\ on the top of the O/S stack.
true value L>R?		\ -- flag
\ Set true when the Forth stack order matches the argument
\ declaration order.
0 value ObjPtr?		\ -- flag
\ True if an object-oriented call is required, i.e. a hidden
\ "this" pointer is used. The location of "this" is compiler and
\ OS dependent.
\   0 = none
\   1 = VC++, pointer in ECX
\   2 = gcc, pointer on top, but below struct pointer (see below)
0 value StructRet?	\ -- flag
\ Set this true if the import returns a structure and requires a
\ hidden structure pointer. The location of the structure pointer
\ is compiler and OS dependent.
\   0 = none
\   1 = VC++
\   2 = gcc

\ 0 value CalleeCleaned?	\ -- flag
\ 0 = "C", 1 = "PASCAL", used for cleanup only. If the calling
\ mechanism preserves the Forth stack pointers, this value is not
\ needed.

defer setDefExt	\ --
\ Set the call type defaults in the VALUEs above.

#256 buffer: importName		\ name of DLL function
#256 buffer: externName		\ name of Forth word
variable usesFloats		\ non-zero if FP arguments or return ; SFP006
variable NumArgs		\ number of input arguments
variable NumInts		\ number of arguments on Forth data stack
variable NumFloats		\ number of arguments on Forth Float stack
variable /DataFrame		\ total size of data frame with padding
variable /ForthFrame		\ number of bytes taken from Forth stack
/argItem buffer: ReturnRes	\ structure for return data
/argList buffer: ArgList	\ argument data during compilation

0 value LastImport		\ address of current/last import structure

: .ArgType	\ n --
  case
    SINTtype of  ." SINT"  endof
    UINTtype of  ." UINT"  endof
    BOOLtype of  ." BOOL"  endof
    FLOATtype of  ." Float"  endof
      ." ???"
  endcase
;

: .ArgData	\ ^ai --
\ Display the argument data
  cr ." Size: " dup ai.size @ dup 0 .r
  0= if
    drop  ."  (void)"
  else
    ." , Type: " dup ai.Type @ .ArgType
    ." , PSP: " dup ai.PSPoff @ 0 .r
    ." , Frame: " ai.FrameOff @ 0 .r
  endif
;

: ArgList[]	\ u -- addr
\ Return the address of the uth item in argList
  /argItem * argList +
;

: .ArgList	\ --
\ Display the calculated argument and return data.
  cr ." Arguments"  numArgs @ 0
  ?do  i ArgList[] .ArgData  loop
  cr ." --" cr ." Returns"
  ReturnRes .ArgData
;

: CalcSource	\ ^ai --
\ Generate the source Forth data for the given argument.
  dup ai.type @ FLOATtype = if
    drop  numFloats incr
  else
    /ForthFrame @ over ai.PSPoff !
    ai.size @ aligned /ForthFrame +!
    numInts incr
  endif
;

: CalcForthOff	\ --
\ Calculate the offsets of data items on the Forth input stack.
\ The Forth order is only affected by the choice of L>R or R>L
\ ordering. By default we use L>R.
  numArgs @ if
    L>R? if
      0 numArgs @ 1-
      do  i ArgList[] CalcSource  -1 +loop
    else
      numArgs @ 0
      do  i ArgList[] CalcSource  loop
    endif
  endif
;

: AlignFrame	\ n --
\ Force frame alignment of a item of the given size to an N byte
\ boundary where N is a power of two.
  /DataFrame @ FrameBackup +		\ offset w.r.t frame after call
  over 1- +  swap negate and		\ -- dp'
  FrameBackup - /DataFrame !
;

: ?AlignFrame	\ size --
\ Force frame alignment of a item of the given size to an N byte
\ boundary where N is a power of two. Ignore void items.
  ?dup if
    aligned AlignInFrame?
    if  AlignFrame  else  drop  endif
  endif
;

: CalcDest	\ ^ai --
\ Generate the destination frame data for the given argument.
  dup ai.size @ ?AlignFrame
  /DataFrame @ over ai.FrameOff !
  ai.size @ aligned /DataFrame +!
;

: CalcFrameOff	\ --
\ Calculate the O/S stack offsets. These are from the C-style
\ argument list, and so are unaffected by R>L, which only affects
\ the source offsets on the Forth stack with respect to the
\ declaration list.
  REQUIRE_NREV if
    numArgs @ 0
    ?do  i ArgList[] CalcDest  loop
  else
    numArgs @ if
      0 numArgs @ 1-
      do  i ArgList[] CalcDest  -1 +loop
    endif
  endif
;

: CalcOffsets	\ --
\ Calculate the stack and frame offsets required.
  numInts off  numFloats off  /DataFrame off  /ForthFrame off
  CalcForthOff  CalcFrameOff
;


\ ******************
\ +S Code generation
\ ******************

\ ========
\ +N Tools
\ ========

: [a		\ -- ; start of assembler sequence
  also asm-access  ;  immediate

: a]            \ -- ; end of assembler sequence
  previous  ;  immediate

: [a]		\ -- ; flush asm sequence
  [ also asm-core ] asm-end asm-start [ previous ]       \ flush assembler
;


\ ==============================
\ +N Prologue, call and epilogue
\ ==============================
\ The process of making the external call is to:
\ 1) Save Forth system registers on the return (ESP) stack.
\ 2) If required save the FP state.
\ 3) Preserve the Forth TOS (normally cached in EBX) on the
\    deep stack so that the Forth stack is all in memory.
\ 4) Preserve ESP by extending the deep Forth stack.
\ 5) Adjust the return stack (ESP) to operating system requirements
\    with space allocated for the function arguments.
\    This may involve aligment to a 16 byte boundary.
\ 6) Copy the Forth and FP parameters to the ESP stack locations
\    allowing for the calling convention.
\ 7) Call the external function
\ 8) Restore ESP for "C" call convention
\ 9) Save the return data and process EBX for void returns
\ 10) Restore the NDP as required
\ 11) Restore the Forth stack
\ 12) Restore Forth registers

8 constant /PSPextend	\ -- u
\ The number of bytes by which the Forth stack is dropped to
\ extend it. The high cell contains the cached TOS.
/PSPextend cell - constant /PSPtos	\ -- u
\ The offset from the dropped Forth stack to the cell containing
\ TOS.


: FrameDrop	\ -- n
\ The amount by which the frame is dropped before alignment
  /DataFrame @ FrameAlignment + negate
;

: FrameMask	\ -- mask
\ The mask applied to ESP after FrameDrop has been applied.
  FrameAlignment negate
;

: -FrameCall	\ -- -n
\ The number of bytes that allow for the call data after the
\ frame has been dropped.
  FrameBackup negate
;

: -ReturnForth	\ n -- n-x
\ Remove the size of data returned on the Forth data stack.
  ReturnRes ai.Size @ aligned -
;

: #PSPrestore	\ -- n
\ generate code to restore TOS for void/float returns.
  /ForthFrame @				\ size of Forth stack arguments
  returnRes ai.size @ if
    case  ReturnRes ai.Type @ 		\ Return type
      SINTtype of  -ReturnForth  endof
      UINTtype of  -ReturnForth  endof
      BOOLtype of  -ReturnForth  endof
    endcase
  endif
  /PSPextend +				\ frame overhead
;

: ^efn		\ -- addr
\ address holding function address
  LastImport >libaddr
;

[defined] Target_386_Windows [if]
#108 constant /fsave	\ -- u
\ space required to save the NDP state
/fsave negate constant -/fsave	\ -- u

also Forth definitions
(( \ moved to STARTUP.FTH ; SFP009
defer preExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls without floating point parameters or return
\ ** items. It is executed at the start of the external
\ ** call before any parameter processing.
  assign noop to-do preExtCall
defer postExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls without floating point parameters or return
\ ** items. It is executed at the end of the external
\ ** call after return data processing.
  assign noop to-do postExtCall
))
\ SFP006...
defer preFPExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls with floating point parameters or return
\ ** items. . It is executed at the start of the external
\ ** call before any parameter processing.
  assign noop to-do preFPExtCall
defer postFPExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls with floating point parameters or return
\ ** items. It is executed at the end of the external
\ ** call after return data processing.
  assign noop to-do postFPExtCall
\ ...SFP006
previous definitions
[then]

: prologue,	\ --
\ Lay the call entry prologue
[defined] Target_386_Windows [if]
  usesFloats @ if			\ SFP006
    [a  call  [] ['] preFPExtCall 5 +  a] [a]	\ pre call hook
  else
    [a  call  [] ['] preExtCall 5 +  a] [a]	\ pre call hook
  endif
[then]
  [a
\ save Forth registers ; MPE005
    push  esi				\ Win: required
    push  edi
\ preserve ESP and TOS
    sub   ebp, # 8			\ save EBX and ESP
    mov   0 [ebp], esp			\ after call,
    mov   4 [ebp], ebx			\ save TOS, easier from memory if double
  a] [a]
  FrameAlignment if
    \ make a frame on ESP and align it to OS convention, 16 byte aligned less 4 bytes
    [a
      lea   eax, FrameDrop [esp]	\ compile time constant
      and   eax, # FrameMask		\ 16 byte align
      lea   esp, FrameBackup [eax]	\ allow for size of call
    a] [a]
  else
    FrameDrop if
    [a
      lea   esp, FrameDrop [esp]	\ compile time constant
    a] [a]
    endif
  endif
; doNotSin


also forth definitions
#16 cells buffer: CallResults
CallResults 0 cells + constant gc0	\ EBP before call
CallResults 1 cells + constant gc1	\ ESP before call
CallResults 2 cells + constant gc2	\ 0 [EBP] before call
CallResults 3 cells + constant gc3	\ EBP after call
CallResults 4 cells + constant gc4	\ ESP after call
CallResults 5 cells + constant gc5	\ 0 [EBP] after call

variable LastWinProc	\ holds last callback

variable XcallFaulted	\ set after first fault
defer XcallHook
  assign noop to-do XcallHook

: resetXCfaults	\ --
  LastCallback off  LastWinproc off  XcallFaulted off  ;
' resetXCfaults AtCold
previous definitions

: generateCall	\ --
\ Lay the code to call the external function, and clean up.
  [a
\ Save data before call
    cmp   [] XcallFaulted , # 0
    z, if,
      mov   [] LastCallback , # 0
      mov   [] gc0 , ebp
      mov   [] gc1 , esp
      mov   eax, 0 [ebp]
      mov   [] gc2 , eax
    endif,
\ Call external C/PASCAL proc
    cld
    call  ^efn []		\ indirect through pointer
    cld
\ Save data after call
    cmp   [] XcallFaulted , # 0
    z, if,
      mov   [] gc3 , ebp
      mov   [] gc4 , esp
      mov   esi, 0 [ebp]
      mov   [] gc5 , esi
    endif,

\ Test return value
    cmp   esp, 0 [ebp]
    nz, if,
\ assume INT return value
      cmp   [] XcallFaulted , # 0	\ if this is the first fault
      z, if,
        mov   [] XcallFaulted , # -1	\ set fault flag
        mov   esi, [] LastCallback	\ copy callback address
        mov   [] LastWinproc , esi
      endif,
      mov   ebx, eax
      add   ebp, # #PSPrestore		\ adjust Forth PSP stack: 0-4+8=4

      pop   edi
      pop   esi
      a] [a]
[defined] Target_386_Windows [if]
      usesFloats @ if			\ SFP006
        [a  call  [] ['] postFPExtCall 5 +  a] [a]	\ post call hook
      else
        [a  call  [] ['] postExtCall 5 +  a] [a]	\ post call hook
      endif
[then]
      [a
      call  [] ['] XcallHook 5 +
      ret
    endif,
\ Restore EBP reg
    mov   esp, 0 [ebp]	\ call must have preserved EBP, works with C/PASCAL
  a] [a]
;

: epilogue,	\ --
\ Restore the Forth and O/S stacks.
  [a
\ Forth data stack restoration
    add   ebp, # #PSPrestore		\ adjust Forth PSP stack: 0-4+8=4
\ restore Forth registers
    pop   edi
    pop   esi
  a] [a]
\ restore NDP stack
[defined] Target_386_Windows [if]
  usesFloats @ if			\ SFP006
    [a  call  [] ['] postFPExtCall 5 +  a] [a]	\ post call hook
  else
    [a  call  [] ['] postExtCall 5 +  a] [a]	\ post call hook
  endif
[then]
; doNotSin


\ ============================
\ +N Copy Forth stack to frame
\ ============================

: CopyS8	\ srcoff destoff --
\ Copy a signed 8 bit item from srcoff [EBP] to destoff [ESP]
  >r >r					\ R: -- destoff srcoff
  [a  movsx  EAX, BYTE a] r> [a [EBP] a] [a]
  [a  mov  a]  r> [a [ESP], EAX a] [a]
;

: CopyS16	\ srcoff destoff --
\ Copy a signed 16 bit item from srcoff [EBP] to destoff [ESP]
  >r >r					\ R: -- destoff srcoff
  [a  movsx  EAX, WORD a] r> [a [EBP] a] [a]
  [a  mov  a]  r> [a [ESP], EAX a] [a]
;

: CopyU8	\ srcoff destoff --
\ Copy an unsigned 8 bit item from srcoff [EBP] to destoff [ESP]
  >r >r					\ R: -- destoff srcoff
  [a  movzx  EAX, BYTE a] r> [a [EBP] a] [a]
  [a  mov  a]  r> [a [ESP], EAX a] [a]
;

: CopyU16	\ srcoff destoff --
\ Copy an unsigned 16 bit item from srcoff [EBP] to destoff [ESP]
  >r >r					\ R: -- destoff srcoff
  [a  movzx  EAX, WORD a] r> [a [EBP] a] [a]
  [a  mov  a]  r> [a [ESP], EAX a] [a]
;

: CopyX32	\ srcoff destoff --
\ Copy a cell from srcoff [EBP] to destoff [ESP]
  >r >r					\ R: -- destoff srcoff
  [a  mov   EAX, a] r> [a [EBP] a] [a]
  [a  mov  a]  r> [a [ESP], EAX a] [a]
;

: CopyX64-F	\ srcoff destoff --
\ Copy a 64 bit double integer in Forth format from srcoff [EBP]
\ to destoff [ESP]
  2dup 4 + CopyX32  swap 4 + swap CopyX32
;

: CopyX64-C	\ srcoff destoff --
\ Copy a 64 bit double integer in C format from srcoff [EBP]
\ to destoff [ESP]
  2dup CopyX32  swap 4 + swap 4 + CopyX32
;

: offsets	\ ^ai -- srcoff destoff
\ Extract the base offsets from the structure. The source offset
\ is assumed to be on the PSP/EBP stack and is adjusted to be based
\ from the saved TOS.
  dup ai.PSPoff @ /PSPtos +  swap ai.FrameOff @
;

: CopySINT	\ ^ai --
\ Compile code to copy TOS to the frame as an signed INT32/64.
\ with any required promotion.
  case  dup ai.size @
    1 of  offsets CopyS8  endof
    2 of  offsets CopyS16  endof
    4 of  offsets CopyX32  endof
    8 of  offsets CopyX64-F  endof
     err_CGsize throw
  endcase
;

: CopyUINT	\ ^ai --
\ Compile code to copy TOS to the frame as a UINT32/64.
\ with any required promotion.
  case  dup ai.size @
    1 of  offsets CopyU8  endof
    2 of  offsets CopyU16  endof
    4 of  offsets CopyX32  endof
    8 of  offsets CopyX64-F  endof
     err_CGsize throw
  endcase
;

: CopyBOOL	\ ^ai --
\ Compile code to copy TOS to the frame as a BOOL.
\ with any required promotion/demotion.
  offsets CopyU8
;

: CopyFLOAT	\ ^ai --
\ Compile code to copy FTOS to the frame as a float or double
  case  dup ai.size @
    4 of  >r  [a fstp float a]  r> ai.FrameOff @ [a [esp] a] [a]  endof
    8 of  >r  [a fstp double a]  r> ai.FrameOff @ [a [esp] a] [a]  endof
    10 of  >r  [a fstp extended a]  r> ai.FrameOff @ [a [esp] a] [a]  endof
       err_CGsize throw
  endcase
;

: CopyArg	\ ^ai --
\ Compile code to copy an item from Forth to the calling frame
\ with any required promotion.
  case  dup ai.type @
    UINTtype of  CopyUINT  endof
    SINTtype of  CopySINT  endof
    BOOLtype of  CopyBOOL  endof
    FLOATtype of  CopyFloat  endof
      err_ImpType throw
  endcase
;

: CopyFrame	\ --
\ Copy the arguments from the Forth stack to the O/S stack with
\ any required promotion/demotion.
  L>R? REQUIRE_NREV xor 0= if
    numArgs @ if
      0 numArgs @ 1-
      do  i ArgList[] CopyArg  -1 +loop
    endif
  else
    numArgs @ 0
    ?do  i ArgList[] CopyArg  loop
  endif
;


\ ================
\ +N Return result
\ ================

: restoreTOS	\ --
\ For void and float return values, TOS must be restored.
  [a  mov  ebx,  a]  /PSPtos /ForthFrame @ +  [a [EBP] a] [a]
;

: ReturnSINT	\ --
\ Return a signed int.
  case  ReturnRes ai.size @
    1 of  [a  movsx  ebx, al  a] [a]  endof
    2 of  [a  movsx  ebx, ax  a] [a]  endof
    4 of  [a  mov  ebx, eax  a] [a]  endof
        err_RetSize throw
  endcase
;

: ReturnUINT	\ --
\ Return an unsigned int.
  case  ReturnRes ai.size @
    1 of  [a  movzx  ebx, al  a] [a]  endof
    2 of  [a  movzx  ebx, ax  a] [a]  endof
    4 of  [a  mov  ebx, eax  a] [a]  endof
        err_RetSize throw
  endcase
;

: ReturnBOOL	\ --
\ Return a BOOL as a single byte.
  [a  movzx  ebx, al  a] [a]
;

: returnFLOAT	\ -- ; return float/double
\ With the float in ST0, it's already where we want it.
\ For a float return, we must still restore EBX correctly.
  RestoreTOS
;

: returnData	\ -- ; process return value
\ Do any necessary demotion, or lose return result for VOID.
\ This happens before the epilogue code so that the saved frame
\ is available.
  returnRes ai.size @ if	\ non-void
    case  ReturnRes ai.Type @ 		\ Return type
      SINTtype of  ReturnSINT  endof
      UINTtype of  ReturnUINT  endof
      BOOLtype of  ReturnBOOL  endof
      FLOATtype of  returnFLOAT  endof
        err_RetType throw
    endcase
  else				\ void
    RestoreTOS
  endif
; doNotSin


\ ====================
\ +N Import generation
\ ====================

: MakeImport	\ --
\ Make the word that calls the imported function.
  ExternName count (:)			\ make header and start definition
  prologue,
  copyFrame  generateCall  returnData
  epilogue,
  postpone ;				\ finish definition
; doNotSin

: GenerateImport        \ --
\ The following structure is built:
\ cell	link to previous import in library
\ cell	holds address to XCALL
\ z$	0-terminated ImportName of function (case sensitive)
\ cell	pointer to second cell, for a future inline code generator)
\ colon definition

\ Don't do anything if this function already exists and import
\ redefinitions are to be ignored.
  ExternName count Externals.wid search-wordlist if
    drop
    ExternWarnings? if
      cr ." WARNING: Redefinition of Extern: " ExternName $. ."  ignored"
    then
    ExternRedefs? 0=
    if  exit  then
  then

\ Build import list entry
  align
  here -> LastImport			\ save structure address
  import-func-link link,		\ link to import chain
  0 ,					\ holds the function address
  importName $>z,			\ import name counted + 0 terminated

\ Calculate the stack and frame offsets
  CalcOffsets

\ Build the colon definition that interprets the function
  align  ^efn ,				\ points to second cell of import structure
  InExternals? if
    get-current  Externals.wid set-current
    MakeImport
    set-current
  else
    MakeImport
  endif
  LastImport resolveImport		\ Resolve the import we have just made
;
+short-branches
\ ...SFP004


\ =================
\ Argument handling
\ =================

0 value argSIZE		\ in bytes
0 value argTYPE		\ 0 = integer or bool
0 value argDEFSIGN	\ 0 = unsigned, 1 = signed, for ints < 4 bytes
-1 value argREQSIGN	\ 0 = unsigned, 1 = signed, -1 no request made
0 value argISPOINTER	\ 0 = no, 1 = this is a pointer
\ 0 value argCalleeCleans	\ 0 = "C", 1 = "PASCAL"

: next-item	\ c-addr u char -- 'c-addr 'u c-addr1 u1
\ Return the remaining string and the leftmost item delimited by
\ char. Both strings have leading and trailing spaces removed.
  >r 2dup r> scan			\ ca1 u1 ca2 u2 --
  2swap swap >r				\ ca2 u2 u1 -- ; R: ca1 --
  over - >r				\ ca2 u1 -- ; R: ca1 u1 --
  1 /string -white  2r> -white
;

-short-branches
: parse-arg     \ c-addr u --
   4 to argSIZE
   0 to argTYPE				\ default to int
   0 to argDEFSIGN
  -1 to argREQSIGN
   0 to argISPOINTER

  begin
    -leading dup 0>
  while
    \ c-addr u --
    bl next-item
    ['] types voc>wid search-wordlist
    \ -- c-addr-rem u-rem xt flag | c-addr-rem u-rem 0
    dup 0=
    if  ['] noop swap  then		\ Use NOOP if not found
    \ -- c-addr-rem u-rem xt flag
    0= 2 pick 0> and err_impType ?throw	\ bail if token not found and wasn't last!
    execute
  repeat
  2drop

  argREQSIGN -1 =			\ if no sign set,
  if  argDEFSIGN to argREQSIGN  then	\  use default
  argISPOINTER if                       \ if pointer then override to use
    0 to argTYPE			\ treat as int
    0 to argREQSIGN                     \ unsigned cell as data type
    4 to argSIZE			\ use CELL instead?
  then
  argTYPE FLOATtype =			\ SFP006
  if  usesFloats on  endif
;
+short-branches

: delimiter?    \ char -- flag
\ Return true if the character is a type delimiter
  dup [char] * =
  if  drop  true exit  then
  dup bl =
  if  drop  true exit  then
  drop false
;

: parseRet&Name	\ c-addr u -- c-addr u' c-addr1 u1
\ Get the return type and function name from the input stream.
\ Split the string at the last delimiter, so that caddr/u'
\ contains the return type info and caddr1/u1 contains the
\ function name.

\ parse the return type and function name
  [char] ( word count -trailing		\ -- c-addr u ; return type, call conv, name )

\ find last delimiter
  2dup + 1 chars -      \ -- c-addr u last
  begin
    dup c@ delimiter? 0=
  while
    1-
  repeat
  1+			\ -- c-addr u c-addr1
\ calculate strings
  dup 3 pick -          \ -- c-addr u c-addr1 u'
  swap rot		\ -- c-addr u' c-addr1 u
  2 pick -		\ -- c-addr u' c-addr1 u-u'
;

: setArgData	\ ^ai --
  argSIZE over ai.size !
  argTYPE 0= if				\ default is int/bool
    argREQSIGN
    if  SINTtype  else  UINTtype  endif
    swap ai.type !
  else
    argTYPE swap ai.type !
  endif
;

-short-branches
: ExtractArgs	\ caddr u --
\ The input string is the return type information. Then parse the
\ rest of the line after the leading '(' character.

  -white parse-arg			\ parse the return type
\  argCalleeCleans to CalleeCleaned?
  ReturnRes setArgData

  0 numArgs !
  begin
    [char] ) parse		        \ -- caddr len
    'tib @ >in @ 1- 0 max + c@ >r	\ terminator
    begin
      -leading dup 0>
    while
      \ c-addr u --
\ Add another while here to check for C // comments.
\ Note that if a // comment is found, the terminator on the
\ return stack MUST NOT be a ')'. Alternatively, fold comment
\ parsing into NEXT-ITEM.
      [char] , next-item parse-arg	\ next argument to trailing , or )
      numArgs @ ArgList[] setArgData	\ save data in argument array
      1 numArgs +!
    repeat
    2drop
    r> [char] ) <>
   while
    refill 0= #-411 ?throw
  repeat
  numArgs @ 1 = if			\ special case, #args=1 and argsize=0 ; SFP007
    arglist @ 0= if			\ indicates void list
      numArgs off
    endif
  endif
  0 word drop				\ discard rest of line
;
+short-branches


\ =============
\ +N Public API
\ =============

(( for DocGen
: InExternals	\ --
\ *G External imports are created in the *\fo{EXTERNALS} vocabulary.
: InCurrent	\ --
\ *G External imports are created in the wordlist specified by
\ ** *\fo{CURRENT}.
))
also forth definitions

: Extern:	\ "text" --
\ *G Declare an external API reference. See the syntax above.
\ ** The Forth word has the same name as the function in the
\ ** library, but the Forth word name is *\b{not} case-sensitive.
\ ** The length of the function's name may not be longer than a
\ ** Forth word name.
  parseRet&Name		     		\ -- c-addr u c-addr1 u1 ; 1 is name
  2dup importName place			\ set name of DLL function
  ExternName place			\ set name of Forth word
  usesFloats off  setDefExt		\ set default declaraction type ; SFP006
  ExtractArgs GenerateImport
;

: AliasedExtern:	\ "forthname" "text" --
\ *G Like *\fo{EXTERN:} but the declared external API reference
\ ** is called by the explicitly specified *\fo{forthname}.
\ ** The Forth word name follows and then the API name.
\ ** Used to avoid name conflicts, e.g.
\ *C AliasedExtern: saccept int accept( HANDLE, void *, unsigned int *);
\ *P which refernces the Winsock *\fo{accept} function but gives
\ ** it the Forth name *\fo{SACCEPT}. Note that here we use the
\ ** fact that formal parameter names are optional.
  parse-name ExternName place		\ process Forth name
  parseRet&Name importName place	\ -- c-addr' u'
  usesFloats off  setDefExt		\ set default declaration type ; SFP006
  ExtractArgs GenerateImport
;

: LocalExtern:	\ "forthname" "text" --
\ *G As *\fo{AliasedExtern:}, but the import is always built into
\ ** the *\fo{CURRENT} wordlist.
  InExternals?   0 -> InExternals?
  AliasedExtern:
  to InExternals?
;

: extern	\ "text" --
\ *G An alias for *\fo{EXTERN:}.
  extern: ;

: .Externs	\ -- ; display EXTERNs
\ *G Display a list of the external API calls.
  0 import-func-link
  begin
    @ dup halt? 0= and
   while				\ -- count item
    cr  dup 2 cells + .z$
    out @ #31 and #32 swap - spaces
    dup cell+ @ .lword
    swap 1+ swap
  repeat
  drop
  cr . ." Externs"
;

: .BadExterns	\ --
\ *G Display a list of any external API calls that have not been
\ ** resolved.
  0 import-func-link
  begin
    @ dup halt? 0= and
   while				\ -- count item
    dup cell+ @ 0= if
      cr  dup 2 cells + .z$
      swap 1+ swap
    endif
  repeat
  drop
  cr . ." Unresolved Externs"
;

: func-loaded?	\ xt -- addr|0
\ *G Given the XT of a word defined by *\fo{EXTERN:} or friends,
\ ** returns the address of the DLL function in the DLL,
\ ** or 0 if the function has not been loaded/imported yet.
  >link cell - @ @
;

defer func-not-loaded	\ c-addr u --
\ Client special
  assign 2drop to-do func-not-loaded

: all-func-loaded?	\ -- flag
\ Client special
  true >R
  import-func-link @
  begin
    dup
  while
    dup cell+ @ 0= if
      dup 2 cells + zcount func-not-loaded
      r> drop false >r
    then
    @
  repeat
  drop r>
;

previous definitions


\ ******************************
\ *S Pre-Defined parameter types
\ ******************************
\ *P The types known by the system are all found in the vocabulary
\ ** *\fo{TYPES}. You can add new ones at will. Each *\fo{TYPE}
\ ** definition modifies one or more of the following *\fo{VALUE}s.                        )
\ *D argSIZE      Size in bytes of data type.
\ *D argDEFSIGN   Default sign of data type if no override is supplied.
\ *D argREQSIGN   Sign OverRide. This and the previous use 0 = unsigned
\ **              and 1 = signed.
\ *D argISPOINTER 1 if type is a pointer, 0 otherwise
\ \ *D argCalleeCleans  Used to set calling convention. 0 = "C" and 1 = PASCAL.

\ *P Each *\fo{TYPES} definition can either set these flags
\ ** directly or can be made up of existing types.

\ *P Note that you should explicitly specify a calling convention
\ ** for every function defined.


\ ======================
\ *N Calling conventions
\ ======================

also types definitions

: "C"           \ --
\ *G Set Calling convention to "C" standard. Arguments are
\ ** reversed, and the caller cleans up the stack.
  ( 0 to argCalleeCleans )  TRUE to Require_NRev
;

: "PASCAL"      \ --
\ *G Set the calling convention to the "PASCAL" standard as used
\ ** by Pascal compilers. Arguments are *\b{not} reversed, and the
\ ** called routine cleans up the stack. Note that this is *\b{not}
\ ** the same as the word *\fo{PASCAL} below.
  ( 1 to argCalleeCleans )  FALSE to Require_NRev
;

: PASCAL        \ --
\ *G Set the calling convention to the Windows PASCAL standard.
\ ** Arguments are reversed in C style, but the called routine
\ ** cleans up the stack. This is the standard Win32 API calling
\ ** convention. N.B. There are exceptions!
\ ** This convention is also called "stdcall" and "winapi" by
\ ** Microsoft, and is commonly used by Fortran programs.
  ( 1 to argCalleeCleans )  TRUE to Require_NRev
;

: WinApi	\ --
\ *G A synonym for PASCAL.
  PASCAL
;

: StdCall	\ --
\ *G A synonym for PASCAL.
  PASCAL
;

: VC++		\ --
\ *G Defines the calling convention as being for a C++ member
\ ** function which requires "this" in the ECX register.
\ ** The function must be defined with an explicit this
\ ** pointer (void * this). Because exported VC++ member
\ ** functions can have either "C" or "PASCAL" styles, the this
\ ** pointer must be positioned so that it is leftmost when
\ ** reversed (C/WINAPI/StdCall style) or is rightmost when
\ ** not reversed ("PASCAL" style). See also the later section
\ ** on interfacing to C++ DLLs.
  1 to ObjPtr?
;

: R>L	\ --
\ *G By default, arguments are assumed to be on the Forth stack
\ ** with the top item matching the rightmost argument in the
\ ** declaration so that the Forth parameter order matches that
\ ** in the C-style declaration.
\ ** *\fo{R>L} reverses this.
  FALSE to L>R?
;

: L>R	\ --
\ *G By default, arguments are assumed to be on the Forth stack
\ ** with the top item matching the rightmost argument in the
\ ** declaration so that the Forth parameter order matches that
\ ** in the C-style declaration.
\ ** *\fo{L>R} confirms this.
  TRUE to L>R?
;

previous definitions


\ ==============
\ *N Basic Types
\ ==============

also types definitions

: unsigned      \ --
\ *G Request current parameter as being unsigned.
  0 to argREQSIGN
;

: signed        \ --
\ *G Request current parameter as being signed.
  1 to argREQSIGN
;

: int           \ --
\ *G Declare parameter as integer. This is a signed 32 bit quantity
\ ** unless preceeded by *\fo{unsigned}.
  4 to argSIZE  1 to argDEFSIGN
;

: char          \ --
\ *G Declare parameter as character. This is a signed 8 bit quantity
\ ** unless preceeded by *\fo{unsigned}.
  1 to argSIZE  1 to argDEFSIGN
;

: void          \ --
\ *G Declare parameter as void. A *\fo{VOID} parameter has no
\ ** size. It is used to declare an empty parameter list, a null
\ ** return type or is combined with *\fo{*} to indicate a generic
\ ** pointer.
  0 to argSIZE  0 to argDEFSIGN
;

: *             \ --
\ *G Mark current parameter as a pointer.
  1 to argISPOINTER
;

: const ;	\ --
\ *G Marks next item as *\b{constant} in C terminology. Ignored
\ ** by VFX Forth.

: int32		\ --
\ *G A 32bit signed quantity.
  4 to argSIZE  1 to argDEFSIGN
;

: int16		\ --
\ *G A 16 bit signed quantity.
   2 to argSIZE  1 to argDEFSIGN
;

: int8		\ --
\ *G An 8 bit signed quantity.
   1 to argSIZE  1 to argDEFSIGN
;

: uint32	\ --
\ *G 32bit unsigned quantity.
  unsigned int  ;

: uint16	\ --
\ *G 16bit unsigned quantity.
  unsigned int  2 to argSIZE  ;

: uint8		\ --
\ *G 8bit unsigned quantity.
  1 to argSIZE  0 to argDEFSIGN
;

: LongLong	\ --
\ *G A 64 bit signed or unsigned integer. At run-time, the argument
\ ** is taken from the Forth data stack as a normal Forth double
\ ** with the top item on the top of the data stack.
  8 to argSIZE  1 to argDEFSIGN
;

: LONG          int  ;
\ *G A 32 bit signed quantity.

: SHORT		\ --
\ *G For most compilers a *\b{short} is a 16 bit signed item,
\ ** unless preceded by *\fo{unsigned}.
  2 to argSIZE  1 to argDEFSIGN
;

: BYTE		\ --
\ *G An 8 bit unsigned quantity.
  1 to argSIZE  0 to argDEFSIGN
;

: float		\ --
\ *G 32 bit float.
  4 to argSIZE  0 to argDEFSIGN  FLOATtype to argTYPE
;

: double	\ --
\ *G 64 bit float.
  8 to argSIZE  0 to argDEFSIGN  FLOATtype to argTYPE
;

previous definitions


\ ================
\ *N Windows Types
\ ================
\ *P The following parameter types are non "C" standard and are used by
\ ** Windows in function declarations. They are all defined in terms
\ ** of existing types.

[defined] Target_386_Windows [IF]

: (setDefExt)	\ --
\ set up the defaults for EXTERN: calls.
  0 to FrameAlignment  0 to FrameBackup  0 to AlignInFrame?
  TRUE to REQUIRE_NREV  ( 1 to CalleeCleaned? )
  TRUE to L>R?
  0 to ObjPtr?  0 to StructRet?
;
assign (setDefExt) to-do setDefExt

also types definitions

: OSCALL	PASCAL  ;
\ *G Used for portable code to avoid three sets of declarations.
\ ** For Windows, this is a synonym for *\fo{PASCAL} and under
\ ** Linux and other Unices this is a synonym for *\fo{"C"}.
: DWORD         unsigned int  ;
\ *G 32 bit unsigned quantity.
: WORD          unsigned int  2 to argSIZE  ;
\ *G 16 bit unsigned quantity.
: HANDLE        void *  ;
\ *G HANDLEs under Windows are effectively pointers.
: HMENU         handle  ;
\ *G A Menu HANDLE.
: HDWP          handle  ;
\ *G A DEFERWINDOWPOS structure Handle.
: HWND          handle  ;
\ *G A Window Handle.
: HDC           handle  ;
\ *G A Device Context Handle.
: HPEN          handle  ;
\ *G A Pen Handle.
: HINSTANCE     handle  ;
\ *G An Instance Handle.
: HBITMAP       handle  ;
\ *G A Bitmap Handle.
: HACCEL        handle  ;
\ *G An Accelerator Table Handle.
: HBRUSH        handle  ;
\ *G A Brush Handle.
: HMODULE       handle  ;
\ *G A module handle.
: HENHMETAFILE  handle  ;
\ *G A Meta File Handle.
: HFONT         handle  ;
\ *G A Font Handle.
: HRESULT       DWORD   ;
\ *G A 32bit Error/Warning code as returned by various COM/OLE calls.
: LPPOINT       void *  ;
\ *G Pointer to a POINT structure.
: LPACCEL       void *  ;
\ *G Pointer to an ACCEL structure.
: LPPAINTSTRUCT void *  ;
\ *G Pointer to a PAINTSTRUCT structure.
: LPSTR         void *  ;
\ *G Pointer to a zero terminated string buffer which may be modified.
: LPCTSTR       void *  ;
\ *G Pointer to a zero terminated string constant.
: LPCSTR        void *  ;
\ *G Another string pointer.
: LPTSTR        void *  ;
\ *G Another string pointer.
: LPDWORD       void *  ;
\ *G Pointer to a 32 bit DWORD.
: LPRECT        void *  ;
\ *G Pointer to a RECT structure.
: LPWNDPROC     void *  ;
\ *G Pointer to a WindowProc function.
: ATOM          word  ;
\ *G An identifier used to represent an atomic string in the OS table.
\ ** See *\b{RegisterClass()} in the Windows API for details.
: WPARAM        dword  ;
\ *G A parameter type which used to be 16 bit but under Win32 is an
\ ** alias for DWORD.
: LPARAM        dword  ;
\ *G Used to mean LONG-PARAMETER (i.e. 32 bits, not 16 as under Win311)
\ ** and is now effectively a DWORD.
: UINT          dword  ;
\ *G Windows type for unsigned INT.
: BOOL          int  ;
\ *G Windows Boolean type. 0 is false and non-zero is true.
: LRESULT       int  ;
\ *G Long-Result, under Win32 this is basically an integer.
: colorref      DWORD  ;
\ *G A packed encoding of a color made up of 8 bits RED, 8 bits GREEN,
\ ** 8 bits BLUE and 8 bits ALPHA.
: SOCKET	dword  ;
\ *G Winsock socket reference.

previous definitions

[THEN]          \ Target_386_Windows


\ ==============
\ *N Linux Types
\ ==============

[defined] Target_386_Linux [if]

: (setDefExt)	\ --
\ set up the Linux defaults for EXTERN: calls.
  0 to FrameAlignment  0 to FrameBackup  0 to AlignInFrame?
  TRUE to REQUIRE_NREV  ( 0 to CalleeCleaned? )
  TRUE to L>R?
  0 to ObjPtr?  0 to StructRet?
;
assign (setDefExt) to-do setDefExt

also types definitions
: OSCALL	"C"  ;
\ *G Used for portable code to avoid three sets of declarations.
\ ** For Windows, this is a synonym for *\fo{PASCAL} and under
\ ** Linux this is a synonym for *\fo{"C"}.
: FILE		uint32  ;
\ *G Always use as *\fo{FILE * stream}.
: DIR		uint32  ;
\ *G Always use as *\fo{DIR * stream}.
: size_t	uint32  ;
\ *G Linux type for unsigned INT.
: off_t	uint32  ;
\ *G Linux type for unsigned INT.
: int32_t	int32  ;
\ *G Synonym for *\fo{int32}.
: int16_t	int16  ;
\ *G Synonym for *\fo{int16}.
: int8_t	int8  ;
\ *G Synonym for *\fo{int8}.
: uint32_t	uint32  ;
\ *G Synonym for *\fo{uint32}.
: uint16_t	uint16  ;
\ *G Synonym for *\fo{uint16}.
: uint8_t	uint8  ;
\ *G Synonym for *\fo{uint8}.
: time_t	uint32  ;
\ *G Number of seconds since midnight UTC of January 1, 1970.
: clock_t	uint32  ;
\ *G Processor time in terms of CLOCKS_PER_SEC.
: pid_t		int32  ;
\ *G Process ID.
: uid_t		uint32  ;
\ *G User ID.
: mode_t	uint32  ;
\ *G File mode.
previous definitions

[then]		\ Target_386_Linux

\ =================
\ *N Mac OS X Types
\ =================

[defined] Target_386_OSX [if]

: (setDefExt)	\ --
\ set up the Mac OS X defaults for EXTERN: calls.
  #16 to FrameAlignment  4 to FrameBackup  1 to AlignInFrame?
  TRUE to REQUIRE_NREV  ( 0 to CalleeCleaned? )
  TRUE to L>R?
  0 to ObjPtr?  0 to StructRet?
;
assign (setDefExt) to-do setDefExt

also types definitions
: OSCALL	"C"  ;
\ *G Used for portable code to avoid three sets of declarations.
\ ** For Windows, this is a synonym for *\fo{PASCAL} and under
\ ** OS X this is a synonym for *\fo{"C"}.
: FILE		uint32  ;
\ *G Always use as *\fo{FILE * stream}.
: DIR		uint32  ;
\ *G Always use as *\fo{DIR * stream}.
: size_t	uint32  ;
\ *G Unix type for unsigned INT.
: off_t	uint32  ;
\ *G Unix type for unsigned INT.
: int32_t	int32  ;
\ *G Synonym for *\fo{int32}.
: int16_t	int16  ;
\ *G Synonym for *\fo{int16}.
: int8_t	int8  ;
\ *G Synonym for *\fo{int8}.
: uint32_t	uint32  ;
\ *G Synonym for *\fo{uint32}.
: uint16_t	uint16  ;
\ *G Synonym for *\fo{uint16}.
: uint8_t	uint8  ;
\ *G Synonym for *\fo{uint8}.
: time_t	uint32  ;
\ *G Number of seconds since midnight UTC of January 1, 1970.
: clock_t	uint32  ;
\ *G Processor time in terms of CLOCKS_PER_SEC.
: pid_t		int32  ;
\ *G Process ID.
: uid_t		uint32  ;
\ *G User ID.
: mode_t	uint32  ;
\ *G File mode.
previous definitions

[then]		\ Target_386_OSX


\ **************************
\ *S Using the Windows hooks
\ **************************
\ *P The hooks *\fo{preExtCall} and *\fo{postExtCall} are
\ ** *\fo{DEFER}red words into which you can plug actions that
\ ** will be run before and after any external call. They are
\ ** principally used:
\ *(
\ *B To save and restore the NDP state when using screen and printer
\ ** drivers that do not obey all the Windows rules.
\ *B To save and restore the NDP state and you want the NDP state
\ ** preserved regardless of any consequences. Although this is safe,
\ ** the system overhead is greater than that of preserving your
\ ** floats in variables or locals as required.
\ *B Installing error handlers that work with nested callbacks.
\ *)

\ *P The hooks *\fo{preFPExtCall} and *\fo{postFPExtCall} are
\ ** compiled into calls with floating point parameters or return
\ ** values. They do not affect the NDP state.

\ *P The example below illustrates both actions.

[defined] Target_386_Windows [if]

also forth definitions
(( for DocGen
defer preExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls. It is executed at the start of the external
\ ** call before any parameter processing.
  assign noop to-do preExtCall
defer postExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls. It is executed at the end of the external
\ ** call after return data processing.
  assign noop to-do postExtCall
defer preFPExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls with floating point parameters or return
\ ** items. It is executed at the start of the external
\ ** call before any parameter processing.
  assign noop to-do preFPExtCall
defer postFPExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls with floating point parameters or return
\ ** items. It is executed at the end of the external
\ ** call after return data processing.
))
variable XcallSaveNDP?	\ -- addr
\ *G Set true when imports must save and restore the NDP state.
\ ** Windows only. From build 2069 onwards, the default behaviour
\ ** for Windows includes saving and restoring the FPU state.
\ ** This can be inhibited by clearing *\fo{XcallSaveNDP?}
\ ** before execution.
variable abort-code	\ -- addr
\ *G Holds error code for higher level routines, especially
\ ** *\fo{RECOVERY} below. Windows versions only.
  abort-code off
defer xcall-fault	\ -- ; handles errors in winprocs
\ *G Used by application code in the *\fo{DEFER}red words
\ ** *\fo{preExtCall} and *\fo{postExtCall} above to install
\ ** user-defined required actions.
  assign noop to-do xcall-fault
previous definitions

\ *[
code PreExtern	\ -- ; R: -- sys
\ Clears the abort code and saves the NDP state if XcallSaveNDP?
\ is set.
  mov   dword ptr abort-code , # 0	\ no previous abort code
  cmp	  [] XcallSaveNDP? , # 0	\ Win: required
  nz, if,
    pop     eax
    lea     esp, -/fsave [esp]
    fsave   0 [esp]
    push    eax
  endif,
  ret
end-code
assign preExtern to-do preExtCall

code PostExtern	\ -- ; R: sys --
\ Restore the NDP state if XcallSaveNDP? is set and test the
\ abort code.
  cmp   [] XcallSaveNDP? , # 0		\ required
  nz, if,
    pop     eax
    frstor  0 [esp]
    lea     esp, /fsave [esp]
    push eax
  endif,
  \ Detecting faults in nested callbacks.
  cmp   dword ptr abort-code , # 0	\ test previous aborting code
  nz, if,
    call  [] ' xcall-fault 5 +		\ execute xcall-fault if set
  endif,
  ret
end-code
assign postExtern to-do postExtCall

code PreFPExtern	\ -- ; R: -- sys ; SFP006
\ Clears the abort code.
  mov   dword ptr abort-code , # 0	\ no previous abort code
  ret
end-code
assign preFPExtern to-do preFPExtCall

code PostFPExtern	\ -- ; R: sys -- ; SFP006
\ Test the abort code.
  cmp   dword ptr abort-code , # 0	\ test previous aborting code
  nz, if,
    call  [] ' xcall-fault 5 +		\ execute xcall-fault if set
  endif,
  ret
end-code
assign postFPExtern to-do postFPExtCall
\ *]

[then]


\ **************************
\ *S Interfacing to C++ DLLs
\ **************************
\ *N Caveats
\ *P These notes were written after testing on Visual C++ v6.0.
\ ** Don't blame us if the rules change!

\ *N Example code
\ *P The example code may be found in the directory EXAMPLES\VC++.
\ ** Because of the inordinate amount of time we spent wandering
\ ** around inside debuggers to get this far, we recommend that
\ ** you adopt a cooperative and investigative attitude when
\ ** requesting technical support on this topic.

\ *N Accessing constructors and destructors
\ *P Example code for accessing the constructor of class is
\ ** provided in *\i{TRYCPP.FTH} which accesses the class DllTest
\ ** in *\i{DLLTEST.CPP}.
\ *P Since C++ is supposed to provide a higher level of abstraction,
\ ** apparently simple operations may generate reams of code. So it
\ ** is with the equivalent of
\ *C   pClass = new SomeClass;
\ *P The actual code generated may/will be a call to a function new
\ ** to generate an object structure (not a single cell) followed
\ ** by passing the return value from new to the class constructor.

\ *P The class constructor (in C++ CDllTest::CDllTest()) is not normally
\ ** exported from C++ without some extra characters being added to
\ ** the name. For example, the reference to it in the example code is:
\ *C   extern: PASCAL void * ??0CDLLTest@@QAE@XZ( void );
\ *P This function is not directly callable because it has
\ ** to be passed the result of the *\fo{new} operator. To solve this
\ ** problem *\i{DLLTest.dll} contains a helper function *\fo{CallNew}
\ ** which is passed the address of the constructor for the class.
\ ** This is redefined as *\fo{NEW} for normal use.
\ *E \ C++ Helpers
\ ** extern: PASCAL void * CallNew( void * );
\ ** extern: PASCAL void CallDelete( void * this);

\ *E \ CDLLTest class specific
\ ** extern: PASCAL void * ??0CDLLTest@@QAE@XZ( void );
\ **
\ ** 0 value CDLLTest	\ -- class|0
\ **
\ ** : InitCDLLTest	\ -- ; initialise the CPP interface
\ **   ['] ??0CDLLTest@@QAE@XZ func-loaded? -> CDLLTest
\ ** ;
\ **
\ ** : New	\ class -- *obj|0
\ **   CallNew
\ ** ;
\ **
\ ** : Delete	\ *obj --
\ **   CallDelete
\ ** ;

\ *P The word *\fo{INITCDLLTEST} gets the address of the constructor for the class,
\ ** and NEW then runs the CallNew function which executes the
\ ** C++ new operator and calls the constructor. Unfortunately, you will
\ ** have to do this for each class that use in the DLL. What is returned
\ ** by CallNew is an object pointer. This is not the object itself, but
\ ** the address of another (undocumented) data structure. It can be used
\ ** as the this pointer for all following member function calls.

\ *P Once you have finished with the onject, you must relase its
\ ** resources using the delete method (the destructor). This is
\ ** implemented in VC++ by passing the object pointer to the
\ ** delete function. This is performed by the CallDelete function
\ ** exported from the DLL. Again, the Forth word DELETE provides
\ ** syntactic sugar by just calling CallDelete.

\ =============================
\ *N Accessing member functions
\ =============================
\ *P A Visual C++ member function exported from a DLL requires
\ ** the "this" pointer in the ECX register. This can be achieved
\ ** using the following form:
\ *E extern: VC++ PASCAL BOOL TestWindow1( void * this, char * ch, int n, int nbyte );
\ *P The function must be defined with an explicit this
\ ** pointer (void * this). Because exported VC++ member
\ ** functions can have either C or "PASCAL" styles, the this
\ ** pointer must be positioned so that it is leftmost when
\ ** reversed (C/PASCAL/WINAPI/StdCall/APIENTRY style) or is
\ ** rightmost when not reversed ("PASCAL" style).

\ *E extern: PASCAL VC++ BOOL GetHello( void * this, char * buff, int len );
\ ** extern: PASCAL VC++ BOOL TestWindow1( void * this, char * ch, int n, int nbyte );
\ ** extern: PASCAL VC++ BOOL TestWindow2( void * this, void * pvoid, int ndword, int nlong );
\ **
\ ** 0 value CDLLTest	\ -- constructor/class
\ **
\ ** : InitCDLLTest	\ --;  Initialise the CPP interface
\ **   ['] ??0CDLLTest@@QAE@XZ func-loaded? to CDLLTest
\ ** ;
\ **
\ ** create Magic#  $AAAA5555 ,	\ -- addr
\ **
\ ** #64 buffer: StringBuff	\ -- ; buffer for GetHello
\ **
\ ** : TestCDLLTest	\ -- ; test CDLLTest interface
\ **   InitCDLLTest  CDLLTest if
\ **     cr ." Initialisation succeeded"
\ **     CDLLTest new ?dup if
\ **       cr ." new succeeded"
\ **       dup StringBuff #64 GetHello drop
\ **       cr ." GetHello returns: " StringBuff .z$
\ **       dup Magic# 4 5 TestWindow1 drop
\ **       dup Magic# #20 #30 TestWindow2 drop
\ **       delete
\ **       cr ." delete done"
\ **     else
\ **       cr ." new failed"
\ **     endif
\ **   else
\ **     cr ." Initialisation failed"
\ **   endif
\ ** ;
\ **

\ *P Please note that the actual code in TRYCPP.FTH may/will be
\ ** different as we extend the facilities. See the source code
\ ** itself!

\ =================================
\ *N Accessing third party C++ DLLs
\ =================================
\ *P Most third party C++ DLLs are provided with C header files which
\ ** define the interfaces. Study of these will provide the information
\ ** you need to determine how to access them.
\ *P For simple C++ classes, the DllTest.dll file can be used
\ ** to provide constructor and destructor access. Note that classes with multiple
\ ** constructors will export these as functions with the same basic
\ ** name differentiated by the name mangling.

\ *P The DLL *\i{Fth2VC60.dll} contains new and delete access for use
\ ** with other DLLs. Note that the third party DLLs must be
\ ** compatible with VC++ v6.0. The example file *\i{EXAMPLES\VC++\USECPP.FTH}
\ ** demonstrates using *\i{Fth2VC60.dll}.

\ *E library: Fth2VC60.dll
\ **
\ ** extern: PASCAL void * FTH2CPPNew( void * constructor);
\ ** extern: PASCAL void FTH2CPPDelete( void * this);
\ **
\ ** : New		\ *class -- *obj|0
\ **   FTH2CPPNew
\ ** ;
\ **
\ ** : Delete	\ *obj --
\ **   FTH2CPPDelete
\ ** ;
\ **

\ *P If you are using an incompatible compiler or DLL, create a
\ ** similar support DLL for that compiler. You can use the source
\ ** code for *\i{Fth2VC60.dll} as an example.


\ ******************
\ *S Changes at v4.3
\ ******************
\ *P The guts of the *\fo{EXTERN:} mechanism have been rewritten
\ ** to provide more features and to support more operating systems.

\ =====================
\ *N Additional C types
\ =====================
\ *P The following C data types are now supported:
\ *(
\ *B Float - a 32 bit floating point item.
\ *B Double - a 64 bit floating point item
\ *B LongLong - a 64 bit integer. These are taken from and
\ ** returned to the Forth data stack as Forth double numbers
\ ** with the high portion topmost on the stack.
\ *)

\ *P Floating point numbers are taken from the NDP floating
\ ** point unit. This is directly compatible with the the Forth
\ ** floating point pack in *\i{Lib\x86\Ndp387.fth}.

\ =========================
\ *N More Operating Systems
\ =========================
\ *P The requirements of newer operating systems, especially
\ ** those for 64-bit operation, are more stringent for things
\ ** like data alignment. Consequently the underlying mechanism
\ ** has changed.

\ ================
\ *N Miscellaneous
\ ================
\ *P These notes are probably only relevant for code that has
\ ** carnal knowledge of the VFX Forth internals.
\ *(
\ *B The *\fo{XCALL} primitive has been removed and is replaced
\ ** by *\fo{NXCALL}.
\ *B The compile time code generation is completely different and
\ ** there is no centralised despatch mechanism.
\ *B Some faciliies provided by *\i{Lib\Win32\SafeCallback.fth}
\ ** are now built in to the Windows system. You *\b{must} use
\ ** the new version!
\ *)

\ *P The word *\fo{NXCALL} is provided for constructing your own
\ ** import mechanisms, but it only deals with single-cell
\ ** arguments and provides no type safety at all. It is used
\ ** internally by VFX Forth in the first stage build of a
\ ** console-mode kernel.

(( for DocGen
code NXCALL     \ i*x addr i -- res
\ *G Calls the operating system function at *\i{addr} with *\i{i}
\ ** arguments, returning the result *\i{res}. As far as the
\ ** operating systems is concerned, *\i{i*x} appear on the CPU
\ ** return stack pointed to by ESP, and the return value is taken
\ ** from EAX. After executing *\fo{NXCALL} the return value *\i{res}
\ ** is the contents of the EAX register.
end-code
))


\ *********
\ Test code
\ *********

((
\ OSX format test code
: OSXDefExt	\ --
\ set up the Mac OS X defaults for EXTERN: calls.
  #16 to FrameAlignment  4 to FrameBackup  1 to AlignInFrame?
  TRUE to REQUIRE_NREV  ( 0 to CalleeCleaned? )
  0 to ObjPtr?  0 to StructRet?
;
: LinDefExt	\ --
\ set up the Linux defaults for EXTERN: calls.
  0 to FrameAlignment  0 to FrameBackup  0 to AlignInFrame?
  TRUE to REQUIRE_NREV  ( 0 to CalleeCleaned? )
  0 to ObjPtr?  0 to StructRet?
;

assign OSXDefExt to-do setDefExt

extern: int nanosleep( void * req, void * rem );

: Sleep	\ ms -- ior
  { | tspec[ 2 cells ] -- }
  #1000 /mod				\ -- rem quotient
  tspec[ 0 + !				\ seconds
  #1000000 * tspec[ cell + !		\ ns
  tspec[ 0 nanosleep
;
assign LinDefExt to-do setDefExt
))

previous definitions


\ ======
\ *> ###
\ ======

decimal
