\ IniParser.fth - create and read INI files

((
Copyright (c) 2008, 2011, 2012, 2013, 2014, 2018, 2019, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To Do
=====

Change History
==============
20200708 SFP014 64 updates.
20191210 SFP013 Removed Community distinction.
20180628 SFO012 Made changes for Community edition.
		Allowed for any number of directory levels.
		Moved general directory code to Win/Lin/MacTools.fth
20140507 SFP011 Corrected -INI-EXEC switch handling.
20140130 SFP010 Added -INI-EXEC switch to use default INI file.
20140101 SFP009 Provide option for old-style INI path.
20120305 SFP008 Unified all operating systems except for defaults.
20110301 SFP007 Fixed off by one error.
20110120 MPE006 Updated for OSX.
20080821 MPE005 Added Ini.WriteMem, Ini.ReadMem, Ini.WriteZStr
		and Ini.ReadZStr.
		Corrected Ini.WriteStr trailing '\' characters.
20080630 MPE004 Corrected Ini.WriteInt because the library uses
		strtol() which is signed.
20080226 MPE003 Added load and save chains.
20080219 MPE002 Added check for library in Ini.Open.
20080110 MPE001 Linux and Windows versions.

This file accesses the iniparser shared library for MacOS and
Linux and the DLL for Windows.
))

also forth definitions  decimal

\ ==========
\ *> tools
\ *S INI files
\ ==========
\ *P If you are upgrading from a system installed before March
\ ** 2012, you may/will need to change how your application
\ ** specifies its INI files.

\ *P The INI file mechanism used by Windows is also available for
\ ** other operating systems. This allows us to use the same
\ ** configuration file mechanism for all operating systems that
\ ** support shared libraries (not in VFX Forth for DOS yet).

\ *P The code accesses a derivative of the iniParser v3.0b shared
\ ** library published by Nicholas Devillard at *\i{http://ndevilla.free.fr/iniparser/},
\ ** where the latest version may be found. Note that the MPE versions
\ ** differ from this version, but are upward compatible. We have
\ ** submitted our changes to the author. A binary copy of the
\ ** library can be found in the *\i{Bin} folder. You are free to
\ ** release this with your applications.

\ *P The full sources for iniParser as used by MPE are in the directories
\ ** *\i{<VFX>\Tools\iniparser3.0b\src} and *\i{<VFX>\Tools\iniparser3.0b\src.win}.
\ ** The relevant shared library (.so or .dll) is in the *\i{Bin} folder
\ ** and is copied to the *\i{Windows\System32} or *\i{/user/lib}
\ ** directory during installation.

\ *P The private profile file mechanism used by VFX Forth for
\ ** Windows before version 4.20 may be found in *\i{<VFX>Lib\Win32\Profile.fth}.
\ ** The old mechanism is not portable between operating systems
\ ** and is much slower. We strongly recommend that you convert
\ ** any existing code that uses the *\fo{PROFILE::xxx} words to
\ ** use the new mechanism.

\ *P The following example shows how INI files are used. A later
\ ** section describes the words in detail.
\ *E : SaveSome	  \ --
\ **   s" %IniDir%\UserIde.ini" Ini.Open 0= if
\ **     S" Options" Ini.Section
\ **     s" FontSet" FontSet? Ini.WriteInt
\ **     s" LogFont" lf[ LOGFONT Ini.WriteMem
\ **     s" Editor"  szEditor zcount Ini.WriteStr
\ **     Ini.Close
\ **   endif
\ ** ;
\ **
\ ** : LoadSome   \ --
\ **   s" %IniDir%\UserIde.ini" Ini.Open 0= if
\ **     S" Options" Ini.Section
\ **     s" FontSet" 0 Ini.ReadInt dup -> FontSet? if
\ **       s" LogFont" lf[ LOGFONT Ini.ReadMem
\ **       lf[ CreateFontIndirect -> hConsoleFont
\ **     endif
\ **     s" Editor" szEditor MAX_PATH zNull zcount Ini.ReadZStr
\ **     Ini.Close
\ **   endif
\ ** ;

\ *P The main things to note are:
\ *(
\ *B The INI file must be opened before use.
\ *B Sections are the parts of the INI file delimited by
\ ** *\fo{[name]}. Sections contain key/value pairs in the form
\ ** *\fo{key=value}. The value portion is represented as text.
\ *B Your code can use a key as a flag to determine how others
\ ** are handled.
\ *B The INI file must be closed, otherwise changes will not
\ ** be written back.
\ *)


\ ***************************
\ *N Shared library interface
\ ***************************

[defined] Target_x64_OSX [if]
Library: libmpeparser64.0.dylib
: IniLib	\ -- addr|0
  libmpeparser64.0.dylib  ;
[then]

[defined] Target_x64_Windows [if]
Library: mpeparser64.dll
: IniLib	\ -- addr|0
  mpeparser64.dll
;
[then]
[defined] Target_x64_Linux [if]
Library: libmpeparser64.so.0
\ *G The library reference.
: IniLib	\ -- addr|0
\ *G Used to determine if the library is available and to isolate the
\ ** host-dependent library name from the rest of the code.
  libmpeparser64.so.0
;
[then]

[defined] Target_386_OSX [if]
Library: libmpeparser.0.dylib
: IniLib	\ -- addr|0
  libmpeparser.0.dylib  ;
[then]

[defined] Target_386_Windows [if]
Library: mpeparser.dll
: IniLib	\ -- addr|0
  mpeparser.dll
;
[then]
[defined] Target_386_Linux [if]
Library: libmpeparser.so.0
\ *G The library reference.
: IniLib	\ -- addr|0
\ *G Used to determine if the library is available and to isolate the
\ ** host-dependent library name from the rest of the code.
  libmpeparser.so.0
;
[then]

[defined] Target_ARM_Linux [if]
Library: libarmmpeparser.so.0
\ *G The library reference.
: IniLib	\ -- addr|0
\ *G Used to determine if the library is available and to isolate the
\ ** host-dependent library name from the rest of the code.
  libarmmpeparser.so.0
;
[then]


\ [defined] Target_386_Linux
\ [defined] Target_ARM_Linux or
\ [if]
Extern: int iniparser_getnsec( void * dict );
\ *G This function returns the number of sections found in a dictionary.
\ ** The test to recognize sections is done on the string stored in the
\ ** dictionary: a section name is given as "section" whereas a key is
\ ** stored as "section:key", thus the test looks for entries that do not
\ ** contain a colon.
\ ** This function returns -1 in case of error.
Extern: char * iniparser_getsecname( void * dict, int n );
\ *G This function locates the n-th section in a dictionary and returns
\ ** its name as a pointer to a string statically allocated inside the
\ ** dictionary. Do not free or modify the returned string!
\ ** This function returns NULL in case of error.
Extern: void iniparser_dump_ini( void * dict, void * file );
\ *G This function dumps a given dictionary into a loadable ini file.
\ ** It is Ok to specify *\i{stderr} or *\i{stdout} as output files.
Extern: void iniparser_dump( void * dict, void * file );
\ *G This function prints out the contents of a dictionary, one element by
\ ** line, onto the provided file pointer. It is OK to specify *\i{stderr}
\ **  or *\i{stdout} as output files. This function is meant for debugging
\ **  purposes mostly.
Extern: char * iniparser_getstring( void * dict, const char * key, char * def );
\ *G This function queries a dictionary for a key. A key as read from an
\ ** ini file is given as "section:key". If the key cannot be found,
\ ** the pointer passed as 'def' is returned.
\ ** The returned char pointer is pointing to a string allocated in
\ ** the dictionary, do not free or modify it.
Extern: int iniparser_getint( void * dict, const char * key, int notfound );
\ *G This function queries a dictionary for a key. A key as read from an
\ ** ini file is given as "section:key". If the key cannot be found,
\ ** the notfound value is returned.
\ ** Supported values for integers include the usual C notation
\ ** so decimal, octal (starting with 0) and hexadecimal (starting with 0x)
\ ** are supported. Examples:
\ *E
\ **  - "42"      ->  42
\ **  - "042"     ->  34 (octal -> decimal)
\ **  - "0x42"    ->  66 (hexa  -> decimal)
\ *P Warning: the conversion may overflow in various ways. Conversion is
\ ** totally outsourced to strtol(), see the associated man page for overflow
\ ** handling.
 \ Extern: double iniparser_getdouble( void * d, char * key, double notfound );
 \ *G This function queries a dictionary for a key. A key as read from an
 \ ** ini file is given as "section:key". If the key cannot be found,
 \ ** the notfound value is returned.
Extern: int iniparser_getboolean( void * dict, const char * key, int notfound );
\ *G This function queries a dictionary for a key. A key as read from an
\ ** ini file is given as "section:key". If the key cannot be found,
\ ** the notfound value is returned.
\ ** A true boolean is found if one of the following is matched:
\ *E  - A string starting with 'y'
\ **  - A string starting with 'Y'
\ **  - A string starting with 't'
\ **  - A string starting with 'T'
\ **  - A string starting with '1'
\ *P A false boolean is found if one of the following is matched:
\ *E  - A string starting with 'n'
\ **  - A string starting with 'N'
\ **  - A string starting with 'f'
\ **  - A string starting with 'F'
\ **  - A string starting with '0'
\ *P  The notfound value returned if no boolean is identified, does not
\ ** necessarily have to be 0 or 1.
Extern: int iniparser_set( void * dict, char * entry, char * val );
\ *G If the given entry can be found in the dictionary, it is modified to
\ ** contain the provided value. If it cannot be found, -1 is returned.
\ ** It is Ok to set val to NULL.
Extern: void iniparser_unset( void * dict, char * entry );
\ *G If the given entry can be found, it is deleted from the dictionary.
Extern: int iniparser_find_entry( void * dict, char * entry );
\ *G Finds out if a given entry exists in the dictionary. Since sections
\ ** are stored as keys with NULL associated values, this is the only way
\ **  of querying for the presence of sections in a dictionary.
Extern: void * iniparser_load( const char * ininame );
\ *G This is the parser for ini files. This function is called,
\ ** providing the name of the file to be read. It returns an ini
\ ** dictionary object that should not be accessed directly, but
\ ** through accessor functions instead. The returned dictionary
\ ** must be freed using *\b{iniparser_freedict()}.
Extern: int iniparser_save( void * d, char * ininame );
\ *G Saves a dictionary object. This is just a wrapper around
\ ** iniparser_dump_ini() to provide insulation between the caller
\ ** and the file system for languages and operating systems which
\ ** do not expose the libc library. The returned error code is 0
\ ** for a successful operation. You still need to call
\ ** *\fo{iniparser_freedict} below.
Extern: void iniparser_freedict( void * dict );
\ *G Free all memory associated to an ini dictionary.
\ ** It is mandatory to call this function before the dictionary object
\ ** gets out of the current context.
\ [then]


\ ********
\ *N Tools
\ ********
\ *P These tools are in the *\fo{SYSTEM} vocabulary and may change
\ ** from version to version. If all you are interested in is using
\ ** the MPE parser interface API, skip this section.

also system definitions

0 value IniSrcFile	\ -- addr
\ *G Holds the currently loaded INI source file pathname as a
\ ** zero-terminated string.
0 value IniDestFile	\ -- addr
\ *G Holds the currently loaded INI destination file pathname as a
\ ** zero-terminated string.
0 value IniSection	\ -- addr
\ *G Holds the current section name as a zero-terminated string.
0 value IniKey		\ -- addr
\ *G Holds the current key as a zero-terminated string.
0 value IniData		\ -- addr
\ *G Holds the current write data as a zero-terminated string,
0 value IniDefault	\ -- addr
\ *G Holds the current default as a zero-terminated string
0 value IniScratch	\ -- addr
\ *G Holds the current scratch buffer for processing quote marks.

0 value IniDict		\ -- addr
\ *G Holds the current dictionary pointer.
0 value IniChanged?	\ -- flag

: IniAlloc	\ ptr -- ior
\ *G Allocate *\fo{/IniBuff} bytes and place the buffer address
\ ** at *\i{ptr}. The first bye of *\i{ptr} is set to zero.
  MAX_PATH allocate
  if  drop 0  else  0 over c!  endif	\ -- ptr 0|buff
  dup rot !  0=
; doNotSin

: IniFree	\ ptr --
\ *G Free buffer memory allocated by us.
  dup @ free drop  off
; doNotSin

: InitIniBuffs	\ -- ior
\ *G Initialise all the buffers and pointers.
  0 -> IniChanged?
  0 -> IniDict
  addr IniSrcFile IniAlloc
  addr IniDestFile IniAlloc or
  addr IniSection IniAlloc or
  addr IniKey IniAlloc or
  addr IniData IniAlloc or
  addr IniDefault IniAlloc or
  addr IniScratch IniAlloc or
; doNotSin

: TermIniBuffs	\ --
\ *G Free all the buffers and clear pointer.
  addr IniSrcFile IniFree
  addr IniDestFile IniFree
  addr IniSection IniFree
  addr IniKey IniFree
  addr IniData IniFree
  addr IniDefault IniFree
  addr IniScratch IniFree
  0 -> IniDict
  0 -> IniChanged?
; doNotSin

: +DoubleQ	\ z$1 -- z$2
\ *G Convert double quote characters to pairs of double quote
\ ** characters.
  IniScratch begin			\ -- zsrc zdest
    over c@
   while
    over c@ [char] " = if
       [char] " over c!  1+
    endif
    over c@ over c!
    swap 1+ swap 1+
  repeat
  0 swap c!  drop IniScratch
; doNotSin

: -DoubleQ	\ z$1 -- z$2
\ *G Convert pairs of double quote characters to single double quote
\ ** characters.
  IniScratch begin			\ -- zsrc zdest
    over c@
   while
    over c@ [char] " = if
      over 1 + c@ [char] " = if
        swap 1+ swap
      endif
    endif
    over c@ over c!
    swap 1+ swap 1+
  repeat
  0 swap c!  drop IniScratch
; doNotSin

: >IniName	\ caddr len dest --
\ *G Copy name to zero terminated string.
  swap MAX_PATH 1- min swap zplace
; doNotSin

((
: zLastChar	\ z$ -- char ; SFP007
  zcount dup
  if  + 1-  else  +  then
  c@
;
))

: >IniString	\ caddr len dest --
\ *G Copy string to zero terminated string. If the last character
\ ** is a '\', add a dummy comment " ; x".
  dup >r
  swap MAX_PATH 4- umin swap zplace
  r@ zcount 1 umax + 1- c@ [char] \ = if	\ SFP007
    s"  ; x" r@ zAppend
  endif
  r> drop
; doNotSin

: WriteIniFile	\ --
\ *G Write the current INI dictionary to the INI file.
  IniDict IniDestFile iniparser_save drop
; doNotSin

: FormIniKey	\ caddr u --
\ *G Form the key string from the current section name and the
\ ** given key. The key string is of the form "*\f{<section:<key>}".
  IniSection IniKey zmove		\ section name
  dup if
    s" :" IniKey zAppend		\ separator
    IniKey zAppend
  else
    2drop
  endif
; doNotSin

: setIniString	\ dict entry val --
\ *G Calls *\b{iniparser_set()} and marks the INI file as changed.
  iniparser_set drop  1 add IniChanged?
; doNotSin

: IniExists	\ caddr len --
\ *G If the file does not exist create an empty one.
  2dup FileExist? if
    2drop
  else
    r/w create-file 0=
    if  close-file  endif
    drop
  endif
; doNotSin

[undefined] czplace [if]
: czplace	\ caddr len dest
\ *G Store the string *\i{caddr/len} as a counted and zero-terminated string
\ ** at *\i{dest}. The strings must not overlap.
  2dup c!  2dup + 1+ 0 swap c!			\ count and terminate bytes
  1+ swap move
; doNotSin
[then]

: nib>hex	\ 4b -- char
\ *G Convert a nibble to a hex character.
  dup 9 >
  if  7 +  endif
  [char] 0 +
; doNotSin

: Mem>Hex	\ caddr len zdest --
\ *G Generate an ASCII hex representation of the memory block as
\ ** a zero terminated string at *\i{zdest}. The length *\i{len}
\ ** of the memory block must be less than 128 bytes.
  -rot  #127 min bounds ?do		\ -- zdest
    i c@ 4 rshift nib>hex  over c! 1+
    i c@ $0F and  nib>hex  over c! 1+
  loop
  0 swap c!
; doNotSin

: Hex>Nib	\ char -- 4b
\ *G Convert a hex character to a nibble.
  dup [char] 9 >
  if  7 -  endif
  [char] 0 -
; doNotSin

: Hex>Mem	\ zsrc caddr len --
\ *G Convert the zero-terminated ASCII HEX string at *\i{zsrc} to
\ ** its memory representation in the buffer *\i{caddr/len}.
  rot zcount 1+ -2 and rot 2* min	\ -- caddr zsrc zlen
  bounds ?do				\ -- caddr
    i c@ Hex>nib 4 lshift
    i 1+ c@ Hex>nib or
    over c!  1+
  2 +loop
  drop
; doNotSin

previous definitions


\ ********************
\ *N Using the library
\ ********************
\ *P The code here is not thread safe, you so may need a semaphore
\ ** from open to close. Some data is held in global variables/buffers.

also system

: Ini.Open	\ caddr len -- ior
\ *G Define and load the Ini file, returning zero on success.
\ ** Sets the destination file to be the same as the loaded file.
\ ** Macros in the file name are expanded.
  IniLib 0=				\ check for library presence
  if  2drop -2 exit  then
  2dup IniExists			\ make empty file if no file
  InitIniBuffs
  if  2drop -1  TermIniBuffs  exit  endif
  Expand IniSrcFile >IniName
  IniSrcFile IniDestFile zmove
  IniSrcFile iniparser_load
  dup -> IniDict  0=
; doNotSin

: Ini.Dest	\ caddr len --
\ *G Set the destination file. This *\b{must} be done after
\ ** *\fo{Ini.Open} and before *\fo{Ini.Close}.
  Expand IniDestFile >IniName
; doNotSin

: Ini.Close	\ --
\ *G If the dictionary has been changed, write it out to the
\ ** destination file.
  IniDict if
    IniChanged?
    if  WriteIniFile  endif
    IniDict iniparser_freedict
  endif
  TermIniBuffs
; doNotSin

: Ini.Section	\ caddr len --
\ *G Set the current section name.
  IniSection >IniName
  IniDict IniSection zNull iniparser_getstring
  zNull = if
    IniDict IniSection NULL setIniString
  endif
; doNotSin

: Ini.Section?	\ caddr u -- flag
\ *G Given a section name, make it current, and return true
\ ** if it exists.
  Ini.Section
  IniDict IniSection iniparser_find_entry
; doNotSin

: Ini.WriteSection	\ caddr u --
\ *G Make the given section current, and write it to the
\ ** dictionary.
  Ini.Section? 0= if
    IniDict IniSection zNull setIniString
  endif
; doNotSin

: Ini.ReadStr	\ c-addr1 u1 c-addr2 u2 c-addr3 u3 --
\ *G Read a string value under key *\i{c-addr1/u1} and return it in
\ ** the result buffer specified by *\i{c-addr2/u2}. If the key
\ ** couldn't be read then the default string *\i{c-addr3/u3} is
\ ** placed in the result buffer.
\ ** Note that the returned string is placed in the result buffer
\ ** as a zero terminated *\b{and} counted string.
  IniDefault >IniName  2>r  FormIniKey
  IniDict IniKey IniDefault iniparser_getstring	\ -- z$ ; R: -- caddr2 u2
  -DoubleQ
  zcount ( -- caddr len )  r> 2- min 0 max r>	\ -- caz uz ca2
  czplace					\ count and terminate bytes
; doNotSin

: Ini.ReadZStr	\ c-addr1 u1 c-addr2 u2 c-addr3 u3 --
\ *G Read a string value under key *\i{c-addr1/u1} and return it in
\ ** the result buffer specified by *\i{c-addr2/u2}. If the key
\ ** couldn't be read then the default string *\i{c-addr3/u3} is
\ ** placed in the result buffer.
\ ** Note that the returned string is placed in the result buffer
\ ** as a zero terminated string.
  IniDefault >IniName  2>r  FormIniKey
  IniDict IniKey IniDefault iniparser_getstring	\ -- z$ ; R: -- caddr2 u2
  -DoubleQ
  zcount ( -- caddr len )  r> 1- min 0 max r>	\ -- caz uz ca2
  zplace
; doNotSin

: Ini.ReadInt	\ c-addr1 u1 default -- value
\ *G Attempt to read an integer value from key *\i{c-addr1/u1}
\ ** and return it. If the key couldn't be read then the default
\ ** is returned.
  >r  FormIniKey
  IniDict IniKey r> iniparser_getint
; doNotSin

: Ini.ReadBool	\ c-addr1 u1 defbool -- bool
\ *G Attempt to read an boolean value from key *\i{c-addr1/u1}
\ ** and return it. If the key couldn't be read then the default
\ ** is returned.
  0<> 1 and >r  FormIniKey
  IniDict IniKey r> iniparser_getboolean
; doNotSin

: Ini.ReadMem	\ c-addr1 u1 c-addr2 u2 --
\ *G Read a memory block with key *\i{c-addr1/u1} and return it in
\ ** the result buffer specified by *\i{c-addr2/u2}. If the key
\ ** couldn't be read then the result buffer is filled with zero
\ ** bytes. *\i{u2} must be less than 128 bytes.
  #127 min  IniDefault over [char] 0 fill	\ fill with '0' chars
  IniDefault over + 0 swap c!			\ zero terminate
  2>r  FormIniKey
  IniDict IniKey IniDefault iniparser_getstring	\ -- z$ ; R: -- caddr2 u2
  2r> Hex>Mem
; doNotSin

: Ini.WriteStr	\ c-addr1 u1 c-addr2 u2 --
\ *G Write a string value attached to a key, into the currently named
\ ** section of the current INI file. *\i{C-addr1/u1} describes
\ ** the name of the key to place the entry under, and *\i{C-addr2/u2}
\ ** the string to write. If the key already exists it is updated.
  IniData >IniString  IniData +DoubleQ IniData zmove
  FormIniKey
  IniDict IniKey IniData setIniString
; doNotSin

: Ini.WriteZStr	\ c-addr1 u1 caddrz --
\ *G Write a zero-terminated string value attached to a key, into the
\ ** current section of the INI file. *\i{C-addr1/u1} describes
\ ** the name of the key to place the entry under, and *\i{caddr\}
\ ** the string to write. If the key already exists it is updated.
  zcount Ini.WriteStr
; doNotSin

: Ini.WriteInt	\ c-addr1 u1 u2 --
\ *G Write a string value corresponding to decimal text for *\i{u2}
\ ** to the key *\i{C-addr1/u1} in the current section of the
\ ** current dictionary.
  base @ >r decimal (.) r> base !
  Ini.WriteStr
; doNotSin

: Ini.WriteBool	\ c-addr1 u1 bool --
\ *G Write a string value corresponding to decimal text for *\i{bool}
\ ** to the key *\i{C-addr1/u1} in the current section of the
\ ** current dictionary.
  0<> 1 and Ini.WriteInt
; doNotSin

: Ini.WriteMem	\ c-addr1 u1 c-addr2 u2 --
\ *G Write a memory block attached to a key, into the currently named
\ ** section of the current INI file. *\i{C-addr1/u1} describes
\ ** the name of the key to place the entry under, and *\i{C-addr2/u2}
\ ** the memory block to write. If the key already exists it is updated.
  IniData Mem>Hex  FormIniKey
  IniDict IniKey IniData setIniString
; doNotSin

: Ini.DeleteKey	\ c-addr1 u1 --
\ *G Delete a key entry completely from the current section.
\ ** *\i{C-addr1/u1} is the name of the key.
  FormIniKey  IniDict IniKey iniparser_unset
;

previous


\ =============================
\ *N Operating system generics
\ =============================

#256 buffer: IniFile$	\ -- addr
\ *G A 256 byte buffer for the expanded INI file path name, which
\ ** may include macros. The path name is stored as a counted string.
\ ** This buffer is initialised at startup from the three
\ ** components below.

#256 buffer: IniDir$	\ -- addr
\ *G A 256 byte buffer for the expanded INI file directory name,
\ ** which may include macros. The path is stored as a counted
\ ** string. This buffer is initialised at startup

#256 buffer: AppSupp$	\ -- addr
\ *G The directory where application support files go, held as
\ ** a counted string This directory must already exist. May
\ ** contain macros.
\ ** You can change this for your own application.
  s" %$AppLocal%" AppSupp$ place

#256 buffer: AppSuppDir$	\ -- addr
\ *G The directory in *\fo{AppSupp$} in which the INI file is
\ ** placed, held as a counted string. The directory will be
\ ** created if it does not already exist. This string may be
\ ** null, or may define one or two levels of directory.
\ ** You can change this for your own application.
  s" MPE\VfxForth" AppSuppDir$ place

64 buffer: AppSuppIni$	\ -- addr
\ *G The name of the INI file in *\fo{AppSuppDir$}, held as a
\ ** counted string. By default, the file is *\i{VfxForth.ini}.
\ ** You can change this for your own application.

256 buffer: PrevIni$	\ -- addr
\ *G Holds the full pathname of a default file copied to the INI
\ ** file if it does not exist. May contain macros.
\ ** You can change this for your own application.

[undefined] GenINI? [if]	\ in IniParser.fth now?
-1 value GenINI?	\ -- flag ; true to generate INI files
\ *G If this *\fo{VALUE} is set true (the default condition) .INI files
\ ** will be generated for VFX Forth and applications when the
\ ** application performs *\fo{BYE}. Such files will also be loaded
\ ** when VFX Forth or an application is executed.
[then]

defer CheckSysIni	\ --
\ *G This word is run at cold start before any INI file is loaded.
\ ** It should provide an INI file if one is needed and does not
\ ** exist.
' noop is CheckSysIni
' CheckSysIni AtCold

1 value IniParserModes	\ -- modes
\ *G If you need only one INI file that could be in the directory
\ ** from which the application is loaded, set this to zero.

: DefIniFile	\ -- caddr len
\ Return the string for the full pathname of the INI file.
  { | temp[ 256 ] -- }
  AppSupp$ count temp[ place
  temp[ +dirSep
  AppSuppDir$ count temp[ append
  temp[ +dirSep
  AppSuppIni$ count temp[ append
  temp[ count >syspad
; doNotSin

: DefIniDir	\ -- caddr len
\ Return the string for the full directory name of the INI file.
  { | temp[ 256 ] -- }
  AppSupp$ count temp[ place
  temp[ +dirSep
  AppSuppDir$ count temp[ append
  temp[ count >syspad
; doNotSin

TextMacro: IniFile
\ *G A text macro that returns the INI file name after macro
\ ** expansion at startup.
TextMacro: IniDir
\ *G A text macro that returns the INI file directory after macro
\ ** expansion at startup.

also substitutions
: setINIdata	\ --
\ Set up the INI file buffers and macros.
  DefIniFile IniFile$ place
  DefIniDir IniDir$ place
  DefIniFile expand IniFile place
  DefIniDir expand IniDir place
; doNotSin

: -ini-exec	\ --
\ *G When used on the command line in lower case, *\fo{-ini-exec}
\ ** causes the INI file to be loaded from *\fo{PrevIni$}, which
\ ** is usually in the executable directory.
;

: usePrevIni	\ -- ; SFP011
\ Set up to use the old INI file and directory.
  PrevIni$ count IniFile$ place
  PrevIni$ count expand IniFile place
  PrevIni$ count IniDir$ place  IniDir$ stripFileName
  IniDir$ count expand IniDir place
; doNotSin

[-short-branches
: (CheckSysIni)	\ --
\ *G Creates the INI file directory if required and sets up the
\ ** INI file macros. This is the default action of *\fo{CheckSysIni}.
  { | temp[ 256 ] -- }
  GenIni? 0=
  if  exit  then
  [ also system ] OS_GetCommandLine [ previous ]
  zcount s" -ini-exec" search nip nip	\ SFP010
  if  0 to IniParserModes  then
  IniParserModes 0=			\ SFP009, SFP011
  if  usePrevIni exit  then
  DefIniFile fileExist? if		\ already exists
    setINIdata  exit
  endif
  AppSuppDir$ c@ if
    AppSupp$ count AppSuppDir$ count	\ base and extended dirs
    makeDirLevels if  exit  then
  then
  PrevIni$ count DefIniFile 1 copy-file drop
  setINIdata
;  ' (CheckSysIni) is CheckSysIni
((
: (CheckSysIni)	\ --
\ *G Creates the INI file directory if required and sets up the
\ ** INI file macros. This is the default action of *\fo{CheckSysIni}.
  { | temp[ 256 ] -- }
  GenIni? 0=
  if  exit  then
  [ also system ] OS_GetCommandLine [ previous ]
  zcount s" -ini-exec" search nip nip	\ SFP010
  if  0 to IniParserModes  then
  IniParserModes 0=			\ SFP009, SFP011
  if  usePrevIni exit  then
  DefIniFile fileExist? if		\ already exists
    setINIdata  exit
  endif
  AppSuppDir$ c@ if
    AppSuppDir$ count [char] \ scan	\ -- caddr len
    dup if				\ has two parts?
      nip				\ -- len
      AppSupp$ count temp[ place	\ base directory
      temp[ +dirSep
      AppSuppDir$ count rot - temp[ append \ first part
      temp[ count forceDir if exit endif
    else
      2drop
    endif
    DefIniDir forceDir if exit endif	\ first/second part
  endif
  PrevIni$ count DefIniFile 1 copy-file drop
  setINIdata
;  ' (CheckSysIni) is CheckSysIni
))
short-branches]
previous


\ =============================
\ *N Operating system specifics
\ =============================
\ *P Setting the INI files has changed. You must now set up four
\ ** strings rather than two, and *\fo{IniFile$} is set at startup,
\ ** and not by you. The defaults are shown below for three
\ ** operating systems.

\ *P These changes were required by changes in the Windows
\ ** security system, the desire to fit in "well" with
\ ** Unix-derived systems, and the requirement for multiple
\ ** application-specific data files. See the *\fo{IniDir}
\ ** macro in particular.

\ ----------
\ *H Windows
\ ----------
\ *P By default, the INI file is *\i{%$AppLocal%\MPE\VfxForth\VfxForth.ini}.
\ ** If the directory or file does not exist, as may happen after
\ ** installation, the directory is created and/or a default
\ ** file is copied.

[defined] Target_386_Windows [if]
\ *[
  s" %$AppLocal%" AppSupp$ place		\ system dir
  s" MPE\VfxForth" AppSuppDir$ place		\ our dir
  s" VfxForth.ini" AppSuppIni$ place		\ file
  s" %load_path%\VfxForth.ini" PrevIni$ place	\ default/previous
\ *]
[then]


\ -----------
\ *H Mac OS X
\ -----------
\ *P By default, the INI file is
\ ** *\i{%$home%/Library/Application Support/VfxForth/VfxForth.ini}.
\ ** If the directory or file does not exist, as may happen after
\ ** installation, the directory is created and/or a default
\ ** file is copied.


[defined] Target_386_OSX [if]
\ *[
  s" %$home%/Library/Application Support" AppSupp$ place
  s" VfxForth" AppSuppDir$ place
  s" VfxForth.ini" AppSuppIni$ place
  s" %$home%/.VfxForth.ini" PrevIni$ place
\ *]
[then]

[defined] Target_x64_OSX [if]
\ *[
  s" %$home%/Library/Application Support" AppSupp$ place
  s" VfxForth" AppSuppDir$ place
  s" VfxForth64.ini" AppSuppIni$ place
  s" %$home%/.VfxForth64.ini" PrevIni$ place
\ *]
[then]


\ --------
\ *H Linux
\ --------
\ *P By default, the INI file is *\i{%$home%/.VfxForth/VfxForth.ini}.
\ ** If the directory or file does not exist, as may happen after
\ ** installation, the directory is created and/or a default
\ ** file is copied.

[defined] Target_386_Linux
[defined] Target_ARM_Linux or
[if]
\ *[
  s" %$home%" AppSupp$ place
  s" .VfxForth" AppSuppDir$ place
  s" VfxForth.ini" AppSuppIni$ place
  s" %$home%/.VfxForth.ini" PrevIni$ place
\ *]
[then]

[defined] Target_x64_Linux
[if]
\ *[
  s" %$home%" AppSupp$ place
  s" .VfxForth" AppSuppDir$ place
  s" VfxForth64.ini" AppSuppIni$ place
  s" %$home%/.VfxForth64.ini" PrevIni$ place
\ *]
[then]

\ -------------------------
\ For all operating systems
\ -------------------------

DefIniFile IniFile$ place
DefIniDir IniDir$ place
also substitutions
DefIniFile expand IniFile place
DefIniDir expand IniDir place
previous


\ *******************************
\ *N System initialisation chains
\ *******************************
\ *P These chains are used for configurations options which are
\ ** preserved when VFX Forth closes down, and are reloaded when
\ ** it starts. Note that because of the position in the cold and
\ ** exit chains, you must be careful that you still have the
\ ** configuration data. For example, if a window is closed before
\ ** configuration save, its position may have to be saved in a
\ ** buffer rather than reading it directly from the window.
\ ** Similarly, when the configuration information is restored, the
\ ** window may/will not be open yet.

\ *P If you want to read and write directly from live information,
\ ** it is more reliable to provide full handlers in your application
\ ** code. However, there will be a time penalty because the INI file
\ ** is repeatedly opened and closed.

variable IniLoadChain	\ -- addr
\ *G The anchor for the initialisation load sequence.
variable IniSaveChain	\ -- addr
\ *G The anchor for the initialisation save sequence.

: AtIniLoad	\ xt --
\ *G The given word is run during the INI load sequence.
  IniLoadChain link, ,  ;

: AtIniSave	\ xt --
\ *G The given word is run during the INI save sequence.
  IniSaveChain link, ,  ;

\ *P The words run must have no net stack action, and behave
\ ** according to the rules of *\fo{ExecChain}.

: LoadSysIni	\ --
\ *G Open the INI file specified by *\fo{iniFile$}, run the
\ ** *\fo{IniLoadChain}, and close the file. Run in the
\ ** cold chain. No action is taken if *\fo{GenINI?} is zero.
  GenINI? if
    IniFile$ count Ini.Open 0= if
      IniLoadChain ExecChain  Ini.Close
    endif
  endif
;
' LoadSysIni AtCold

: SaveSysIni	\ --
\ *G Open the INI file specified by *\fo{iniFile$}, run the
\ ** *\fo{IniSaveChain}, and close/save the file. Run in the
\ ** exit chain. No action is taken if *\fo{GenINI?} is zero.
  GenINI? if
    IniFile$ count Ini.Open 0= if
      IniSaveChain ExecChain  Ini.Close
    endif
  endif
;
' SaveSysIni AtExit

\ CheckSysIni
\ LoadSysIni


\ *****************************************
\ *S Converting from the previous mechanism
\ *****************************************
\ *P By design, there is an almost one to one correspondence
\ ** between the words in the profile and INI mechanisms. There
\ ** are two major differences.
\ *(
\ *B The new code rarely returns error codes as we and you
\ ** nearly always *\fo{DROP}ped them.
\ *B You *\fo{must} use *\fo{Ini.Close} after you have finished
\ ** with the INI file. The data is in memory, and is only
\ ** flushed at close.
\ *)

\ *P For examples of use, see *\fo{LoadUserIde} and *\fo{SaveUserIde}
\ ** in *\i{Studio\XTB.FTH} for the Windows versions.


\ ======
\ *> ###
\ ======
