\ MiscOsx.fth - miscellaneous OS X tools

((
Copyright (c) 2007, 2008, 2009, 2018
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20180629 SFP003 Added more directory handling.
20090829 SFP002 Added some time functions.
20080506 SFP001 Moved TICKS and friends to the kernel.
))

only forth definitions
decimal

\ ===========
\ *> osxtools
\ ===========

\ ******************
\ *S Error variables
\ ******************
\ *P VFX Forth for Mac OS X uses many functions from the *\b{libSystem}
\ ** shared library. The thread local error variables are exposed.

AliasedExtern: errno int * __error( void );
\ *G *\fo{errno} is the well known *\i{errno} C thread local variable
\ ** used by libraries and system calls.
\ ** Can be read by *\fo{@} and written by *\fo{!}



\ *************************
\ *S Environment variables
\ *************************

: ReadEnv	\ naddr1 nlen -- vaddr vlen
\ *G Read the environment variable whose name is given by
\ ** *\i{naddr/nlen} and return the string. If there is no
\ ** such variable *\i{vaddr} is *\fo{zNull} and *\i{vlen} is
\ ** zero.
  { | temp[ #256 ] -- }
  #255 umin temp[ zplace
  temp[ getenv dup
  if  zcount  else  drop zNull 0  endif
;

: WriteEnv	\ vaddr vlen naddr nlen --
\ *G Write the string value *\i{vaddr/vlen} to the environment
\ ** variable named by *\i{vaddr/vlen}.
  { | ntemp[ #256 ] vtemp[ #256 ] -- }
  #255 umin ntemp[ zplace
  #255 umin vtemp[ zplace
  ntemp[ vtemp[ 1 setenv drop
;

: DelEnv	\ naddr nlen --
\ *G Delete the environment variable *\i{naddr/nlen}.
  { | temp[ #256 ] -- }
  #255 umin temp[ zplace
  temp[ unsetenv
;

: EnvMacro:	\ naddr nlen "<var>" -- ; -- caddr
\ *G Create a text macro called *\i{<var>} that queries the
\ ** environment variable named by *\i{naddr/nlen} the returned
\ ** string is a counted string. Use in the form:
\ *C   s" HOME" EnvMacro: $home
\ *P By convention, environment macro names start with a '$'.
   also substitutions definitions
   create
     here 0 , 0 ,			\ name pointer, buffer pointer
     here over ! -rot  $,		\ lay name string
     align here swap cell + !  0 c, #255 allot	\ value buffer
     previous definitions
   does>
     dup @ count ReadEnv #255 min	\ -- pfa caddr len
     rot cell + @ dup >r place  r>	\ -- vaddr
;

s" HOME" EnvMacro: $home
\ *G Text macro for the home directory.


\ ********************
\ *S Critical sections
\ ********************
\ *P Critical sections are implemented using the standard OS X
\ ** semaphore structures and calls.

: makeSemName	\ addr --
\ Generate a unique name for a semaphore. Place the name at addr.
  {: dest -- :}
  base @ >r hex
  s" /sem_" dest zplace			\ /sem_
  dest (u.) dest zAppend		\ /sem_addr
  s" _" dest zAppend			\ /sem_addr_
  getpid (u.) dest zAppend		\ /sem_addr_pid
  r> base !
;

struct /semosx	\ -- len
\ *G  We use a structure consisting of a pointer to the sem_t
\ ** structure followed by a 32 byte zname.
  int smo.*sem
  32 field smo.zname
end-struct

: InitCritSec	\ *smo --
\ *G Initialise a critical section base on a *\fo{/semosx}
\ ** structure. This *\b{must} be done before using the critical
\ ** section.
  dup smo.zname makeSemName
  dup smo.zname  O_CREAT O_EXCL or
  438 1 sem_open4 swap smo.*sem !
;

: TermCritSec	\ *smo --
\ *G Delete the critical section associated with the smaphore.
\ ** This releases internal OS X data and closes the semaphore.
\ ** Nothing should be waiting on the seamphore before
\ ** calling *\fo{TermCritSec}.
  dup smo.zname sem_unlink drop
  smo.*sem @ sem_close drop
;

: [CritSec	\ *smo --
\ *G Wait until the section is available and lock it.
\ ** Does not call *\fo{PAUSE}.
  smo.*sem @ sem_wait drop
;

: CritSec]	\ *smo --
\ *G Unlock the section.
  smo.*sem @ sem_post drop
;

((
: CritSec?	\ sem -- u
 \ *G Returns the section's counter, where non-zero indicates
 \ ** that it is available, or zero when it is locked. Returns
 \ ** zero on error.
  { | temp[ cell ] -- }
  temp[ off
  temp[ sem_getvalue drop
  temp[ @
;
))

\ *P The critical section words use OS X semaphores, which are
\ ** counted semaphores. Thus when using critical sections you
\ ** must be careful to match the use of *\fo{[CritSec} and
\ *\fo{CritSec]}.


\ ********************
\ *S Millisecond timer
\ ********************
\ *P The OS X ticker frequency varies between implementations.
\ ** The code in this section provides simple tools to return
\ ** and handle a millisecond ticker.

Extern: long sysconf( int name );
Extern: int times( void * buf );
3 constant CLK_TCK

struct /tms	\ -- len
\ Time structure uses by times above.
  int tms.utime  /* user time */
  int tms.stime  /* system time */
  int tms_cutime /* user time of children */
  int tms_cstime /* system time of children */
end-struct

10 value /CLK_MS	\ -- ms
\ Default ms per clock tick

: (ticks)	\ -- ms ; return ticks in ms
\ *G Return the system ticker in milliseconds. Treat this as a 32 bit
\ ** unsigned value that wraps around on overflow.
  { | tmp[ /tms ] -- }
  tmp[ times /CLK_MS *
;

: SetTicks	\ --
\ *G Calibrate the OS X ticker and install it as the action of
\ ** *\fo{TICKS}. Performed at start up.
  CLK_TCK sysconf 1 max
  #1000 swap / -> /CLK_MS
  ['] (ticks) to-do ticks
;
' SetTicks AtCold  SetTicks

also system
5 value tickStepMs	\ -- ms
\ *G Minimum interval and granularity used by *\fo{tick-ms}
\ ** below.

: tick-ms	\ ms --
\ *G Waits for at least *\i{ms} milliseconds. Uses *\fo{PAUSE}
\ ** every *\fo{tickStepMs}. This is the default action of *\fo{MS},
\ ** which is *\fo{DEFER}red.
  later  begin
    tickStepMs (OS_ms)
    dup expired
  until
  drop
; assign tick-ms to-do ms
previous


\ ************
\ *S Raw timer
\ ************
\ *P The following timer tools have better resolution but are
\ ** very basic.

Extern: longlong mach_absolute_time( void );

: ns@		\ -- nanoseconds
\ *G Return a system ticker in nanoseconds.
  mach_absolute_time drop ;
: ms@		\ -- milliseconds
\ *G Return a system ticker in milliseconds.
  mach_absolute_time 1000000 um/mod 	\ quot = ms, rem = ns
  swap 500000 >=			\ round up
  if  1+  then
;


\ ****************
\ *S Time handling
\ ****************

struct /timeval \ -- size
\ *S A structure to mimic the *\b{timeval} structure for libc
\ *[
  4 field tv_sec
  4 field tv_nsec
end-struct
\ *]

struct /tm	\ -- size
\ *S A structure to mimic the *\b{tm} structure for libc
  int tm_sec       // 0 to 59 (or 60 for occasional rare leap-seconds)
  int tm_min       // 0 to 59
  int tm_hour      // 0 to 23
  int tm_mday      // 1 to 31
  int tm_mon       // 0 to 11, stupidly 0=January, 11=December
  int tm_year      // year-1900, so 79 means 1979, 103 means 2003
  int tm_wday      // 0 to 6, 0=Sunday, 1=Monday, ..., 6=Saturday
  int tm_yday      // 0 to 365, 0=1st January
  int tm_isdst     // 0 to 1, 1=DST is in effect, 0=it isn't
  int tm_zone      // pointer to time zone string, e.g. "PDT", "EST".
  int tm_gmtoff    // time zone in seconds from GMT; EST=-18000, WET=3600
end-struct

: td>epoch	\ seconds mins hours day month year -- epoch
\ *G Returns the seconds since the start of the epoch. The input
\ ** time is treated as GMT/UTC.
  { | tm[ /tm ] -- }
  tm[ /tm erase
  tm[
  swap 1900 - over tm_year !  swap 1- over tm_mon !  tuck tm_mday !
  tuck tm_hour !  tuck tm_min !  tuck tm_sec !
  z" GMT" over tm_zone !
  mktime
;

: epoch>td	\ epoch -- seconds mins hours day month year
\ *G Converts an epochal second into a GMT/UTC time and date.
  { | epoch[ cell ]  tm[ /tm ] -- }
  epoch[ !  tm[ /tm erase
  epoch[ tm[ gmtime_r
  dup tm_sec @ swap  dup tm_min @ swap  dup tm_hour @ swap
  dup tm_mday @ swap  dup tm_mon @ 1+ swap  tm_year @ 1900 +
;


\ ****************
\ *S Time and date
\ ****************
\ *P These functions rely on the ANS Forth word
\ ** *\fo{TIME&DATE ( -- s m h dd mm yyyy )} and the non-standard
\ ** *\fo{DOW ( -- dow, 0=Sun)} to get the day of the week.

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

create months	\ -- addr
\ *G String containing 3 character text for the months.
  z", JanFebMarAprMayJunJulAugSepOctNovDec"

[-sin

: .dow		\ dow --
\ *G Display day of week.
  3 * days$ 1+ + 3 type
;

: .2r		\ n --
\ *G Display n as a two digit number with leading zeros.
  s>d <# # # #> type
;

: .4r		\ n --
\ *G Display n as a four digit number with leading zeros.
  s>d <# # # # # #> type
;

: .Time&Date	\ s m h dd mm yy --
\ *G Display the system time
\ ** The format is:
\ *C   hh:mm:ss dd Mmm yyyy
  base @ >r  decimal
  dup #100 <				\ correct two digit year clocks
  if  #2000 +  endif
  rot .2r space				\ day - generate date
  swap 1- 3 * months + 3 type space	\ month
  .4r space				\ year
  .2r ." :"			 	\ hours - generate time
  .2r ." :"				\ minutes
  .2r					\ seconds
  r> base !
; doNotSin

: .AnsiDate	\ zone --
\ *G Display the day of week, date and time. If zone is 0
\ ** GMT (system time) is displayed, otherwise local time
\ ** is displayed. The format is:
\ *C   dow, hh:mm:ss dd Mmm yyyy [GMT]
  if
    time&date				\ -- s m h dd mm yy
    dow .dow ." , " .time&date ."      "
  else
    systime&date			\ -- s m h dd mm yy
    sysdow .dow ." , " .time&date ."  GMT "
  endif
;

sin]


\ ************************
\ *S Program launch status
\ ************************
\ *P Programs can be launched in several ways.
\ *(
\ *B From a shell in foreground mode: it has a terminal.
\ *B From a shell or script in background mode, e.g. *\f{vfxosx &}:
\ ** it shares a terminal which is normally only useful for output.
\ *B From the init process or detached by another process. There
\ ** is no terminal at all.
\ *)
\ *P This code allows you to determine how the program was
\ ** launched.

0 value AppPPID		\ -- x
\ *G This application's parent's process ID.
0 value AppPGRP		\ -- x
\ *G This application's process group.
0 value ctPGRP		\ -- x
\ *G The controlling terminal's process group.
0 value AppLaunch	\ -- x
\ *G How the application was launched:
\ *(
\ *B -1 - detached
\ *B 0 - backgrounded
\ *B 1 - foreground
\ *)

: TestLaunch	\ --
\ *G Set the data above to determine how the application was
\ ** launched. Run at program launch.
  0 -> ctPGRP  -1 -> AppLaunch
  getppid -> AppPPID  getpgrp -> AppPGRP
  z" /dev/tty" 0 0 open dup 0<		\ if no terminal, detached
  if  drop  exit  endif
  dup tcgetpgrp dup -> ctPGRP AppPGRP =
  if 1 else 0 then
  -> AppLaunch
  close drop
;  ' TestLaunch AtCold


\ ********************
\ *S Folders and Files
\ ********************

: (changedir)	\ caddr len -- ior ; 0=success
\ *G Expand text macros and change the current working directory.
\ ** Return 0 on success.
  {: | cl[ 1024 chars ] -- ior :}
  cl[ 1023 substituteZ  cl[ chdir
;

: changeDir	\ caddr len --
\ *G Expand text macros and change the current working directory.
\ ** There is no return code.
  (changeDir) drop
;

: workingdir	\ -- caddr len
\ *G Return the text for the current working directory.
  {: | cd[ #1024 ] -- :}
  0 cd[ c!  cd[ #1023 getcwd
  if  cd[   else  znull  endif
  zcount >syspad
;

: dirExists?	\ caddr len -- flag
\ *G Return true if a directory exists. Macros are expanded.
  { | name[ max_path 2+ ] -- }
  expand max_path min name[ zplace
  name[ opendir dup if
    closedir drop  -1
  endif
;

: create-dir	\ caddr len -- ior
\ *G Create a directory, returning zero on success. Macros are expanded.
\ ** Default permissions are used.
  { | name[ max_path 2+ ] -- }
  expand max_path min name[ zplace
  name[ #493 mkdir			\ Decimal 493 = Octal 755
;

: forceDir	\ caddr len -- ior
\ *G Create the directory if it does not exist. Macros are expanded.
  2dup dirExists? if
    2drop 0 exit
  endif
  create-dir
;

\ SFP003...

: +dirSep	\ c$ --
\ *G Add a directory separator to the end of a counted string.
  dir1-char c@ swap addChar
; doNotSin

: prepFileName	\ caddr len --
\ *G Convert '\' and '/' characters in-place as required by the
\ ** operating system.
  bounds ?do
   i c@ dirchar? if  dir1-char @ i c!  then
  loop
; doNotSin

: prepDirName	\ caddr --
\ *G Force the counted string at caddr to end with a '\' character.
  dup count + 1- c@ dir1-char @ <>	\ -- caddr flag ; no terminator
  if  +dirSep  else  drop  then		\ add terminator
; doNotSin

[-short-branches
: makeDirLevels	\ caddr1 len1 caddr2 len -- ior
\ *G The string *\i{caddr1/len1} represents a directory that
\ ** must already exist. *\i{Caddr2/len2} represents additional
\ ** directory levels that may or may not already exist. The
\ ** additional levels are created if they do not exist.
\ ** Successful operation returns 0. For example, to create the
\ ** directory *\f{/Users/stephen/jim/foo/bar}, you could use
\ *C   s" /Users/stephen"  s" jim/foo/bar"  makedirlevels
  {: | exists[ #256 ] maybe[ #256 ] force[ #256 ] -- :}
  #255 umin maybe[ ExpandMacro prepFileName	\ expand macros, force separators
  #255 umin exists[ ExpandMacro prepFileName
  exists[ prepDirName			\ make strings end in dir separator
  maybe[ prepDirName
  maybe[ count  over >r			\ mark start of string
  begin					\ -- caddr len ; R: start --
    dup					\ if string not empty
   while
    dir1-char @ scan dup		\ find next separator
   while				\ -- caddr' len' ; R: start --
    over r@ dup rot swap -		\ -- caddr' len' start lendir
    exists[ count force[ place		\ first part of name
    force[ append			\ add second part
    force[ count forceDir		\ make directory
    if  r> drop 2drop #-425 exit  then	\ exit on error
    1 /string				\ step over '\'
  repeat then
  r> drop 2drop 0
;
short-branches]

\ : tt  s" D:\"  s" jim\foo\bar" makedirlevels  ;

\ ...SFP003

((
/* flags for copyfile */

#define COPYFILE_ACL	    (1<<0)
#define COPYFILE_STAT	    (1<<1)
#define COPYFILE_XATTR	    (1<<2)
#define COPYFILE_DATA	    (1<<3)

#define COPYFILE_SECURITY   (COPYFILE_STAT | COPYFILE_ACL)
#define COPYFILE_METADATA   (COPYFILE_SECURITY | COPYFILE_XATTR)
#define COPYFILE_ALL	    (COPYFILE_METADATA | COPYFILE_DATA)

#define	COPYFILE_RECURSIVE	(1<<15)	/* Descend into hierarchies */
#define COPYFILE_CHECK		(1<<16) /* return flags for xattr or acls if set */
#define COPYFILE_EXCL		(1<<17) /* fail if destination exists */
#define COPYFILE_NOFOLLOW_SRC	(1<<18) /* don't follow if source is a symlink */
#define COPYFILE_NOFOLLOW_DST	(1<<19) /* don't follow if dst is a symlink */
#define COPYFILE_MOVE		(1<<20) /* unlink src after copy */
#define COPYFILE_UNLINK		(1<<21) /* unlink dst before copy */
#define COPYFILE_NOFOLLOW	(COPYFILE_NOFOLLOW_SRC | COPYFILE_NOFOLLOW_DST)

#define COPYFILE_PACK		(1<<22)
#define COPYFILE_UNPACK		(1<<23)

#define COPYFILE_VERBOSE	(1<<30)

#define	COPYFILE_RECURSE_ERROR	0
#define	COPYFILE_RECURSE_FILE	1
#define	COPYFILE_RECURSE_DIR	2
#define	COPYFILE_RECURSE_DIR_CLEANUP	3
#define	COPYFILE_COPY_DATA	4
#define	COPYFILE_COPY_XATTR	5

#define	COPYFILE_START		1
#define	COPYFILE_FINISH		2
#define	COPYFILE_ERR		3
#define	COPYFILE_PROGRESS	4

#define	COPYFILE_CONTINUE	0
#define	COPYFILE_SKIP	1
#define	COPYFILE_QUIT	2
))

: copy-file	\ src slen dest dlen failexist -- ior
\ *G Copy a file. Macros are expanded. If *\i{failexist} is
\ ** non-zero, existing files are not overwritten.
  { | fail? src[ max_path 2 + ] dest[ max_path 2+ ] -- }
  -> fail?
  expand max_path min dest[ zplace
  expand max_path min src[ zplace
  src[ dest[ NULL $0F fail?
  if  $0002:0000 or  endif
  copyfile
;


\ ======
\ *> ###
\ ======

decimal
