\ FIFO.FTH - Queues for VFX Forth

((
Copyright (c) 1997, 2000, 2007, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20160901 SFP001 Added RESETFIFO.
))


\ ========
\ *> tools
\ *S First-In First-Out Queues
\ ========
\ *P VFX Forth contains a set of words for managing character
\ ** (8 bit ) queues. These queues are allocated from the system
\ ** heap.

MODULE FIFO

STRUCT FIFO	\ -- len
\ *G A structure which defines the internal format of the fifo. To
\ ** create a new FIFO you must create an instance of the structure.
\ ** The instance pointer (address of structure) is then used as an
\ ** identifier for subsequent operations.
  CELL field fifo_size
  CELL field fifo_memptr
  CELL field fifo_tail
  CELL field fifo_head
  CELL field fifo_used
END-STRUCT
\ Do not change the size of this structure! Some code, e.g. Richedit
\ devices, assume the size of this structure.

: InitialiseFIFO        \ *FIFO size -- ior
\ *G Initialise a FIFO with a maximum buffer 'size' bytes long.
\ ** the IOR is 0 for success and non-zero for memory allocation failed.
  swap >r
  dup r@ fifo_size !
  CELL+ allocate if
    drop -1
  else
    dup r@ fifo_memptr !
    dup r@ fifo_head !
        r@ fifo_tail !
    0   r@ fifo_used !
    0
  then
  r> drop
;

: FreeFIFO	\ *FIFO -- ior
\ *G Destroy FIFO. Memory is released back into the heap.
  dup fifo_size @ over fifo_used !
  fifo_memptr @ free
;

: resetFIFO	\ *fifo --
\ *G Discard any characters in the FIFO.
  dup fifo_memptr @ swap	\ -- addr *fifo
  2dup fifo_head !
  tuck  fifo_tail !
  0 swap fifo_used !
;

: FIFO?         \ *FIFO -- n ; Return # bytes used in fifo
\ *G Return the number of storage bytes in use within a fifo.
  fifo_used @
;

: bump-head	\ *fifo -- *fifo ; Move HEAD on one position
  >r
  r@ fifo_head @ 1+     \ newhead --
  r@ fifo_memptr @
  r@ fifo_size @ +      \ newhead last
  over                  \ newhead last newhead
  <= if
    r@ fifo_size @ -
  then
  r@ fifo_head !
  r> drop
;

: bump-tail     \ *fifo -- *fifo ; Move TAIL on one position
  >r
  r@ fifo_tail @ 1+     \ newhead --
  r@ fifo_memptr @
  r@ fifo_size @ +      \ newhead last
  over                  \ newhead last newhead
  <= if
    r@ fifo_size @ -
  then
  r@ fifo_tail !
  r> drop
;

: >FIFO(b)      \ BYTE *FIFO -- ior
( *G Add a byte to the FIFO queue. IOR is 0 for success. )
  dup fifo_used  @
  over fifo_size @
  = if
    2drop -1            \ no room in FIFO
  else
    dup fifo_head @     \ -- byte *fifo *head
    rot swap c!         \ -- *fifo
    dup bump-head
    fifo_used 1 swap +!
    0
  then
; doNotSin

: FIFO>(b)	\ *FIFO -- byte ior
( *G Remove next byte from a FIFO. IOR is 0 for success. )
  dup fifo_used @ if
    dup fifo_tail @ c@		\ -- *FIFO byte
    over bump-tail
    swap fifo_used -1 swap +!
    0
  else
    -1
  then
; doNotSin

\ ------------------------------------------------------------------------------

EXPORT FIFO
EXPORT fifo_size
EXPORT fifo_memptr
EXPORT fifo_tail
EXPORT fifo_head
EXPORT fifo_used

EXPORT InitialiseFIFO
EXPORT FreeFIFO
EXPORT resetFIFO
EXPORT FIFO?
EXPORT >FIFO(b)
EXPORT FIFO>(b)

END-MODULE


\ ========
( *> ### )
\ ========

