\ GenTools - generic VFX tools for all systems

((
Copyright (c) 2019
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *> tools
\ ===========

\ ===========================
\ *S C Language Style Helpers
\ ===========================
\ *P VFX Forth supports a few "helper" definitions to aid the
\ ** parsing of "C" header files.
\ ** See *\i{VFXBase/GenTools.fth}.

\ *P These are especially helpful in the parsing of Windows
\ ** resource scripts which are based on the Microsoft RC Language
\ ** for C, and for cutting and pasting from C header files.

: #define       \ <spaces"NAME"> <eol"value-def"> -- ; Exec: -- value
\ *G A simple version of C's #define preprocessor command.
\ ** Any text between the definition name and the end of the line is
\ ** *\fo{EVALUATE}d when *\fo{<NAME>} is invoked.
  create immediate
    0 word count -trailing		\ -- c-addr len
    here  over 1+ allot  place
  does>
    count evaluate
;

: //            \ --
\ *G An implementation of the C++ single line comment.
  postpone \
; immediate

: /*		\ --
\ *G A simple implementation of the C "/* ... */" comment.
  s" */" ParseUntil
; immediate

also system definitions
: next-enum	\ -- caddr len
\ +G A version of *\fo{PARSE-NAME} that works across multiple lines.
\ +* If a word cannot be obtained, the input stream is *\fo{REFILL}ed.
  begin
    [char] , parse dup 0=		\ scan input, no token?
  while					\ -- caddr len
    2drop  refill 0= #-411 ?throw	\ discard null, reload i/p
  repeat
;

: enum-term?	\ caddr len -- caddr len' flag
\ +G Return true if the last character is '}' or last two are '};'.
  2dup + 1- c@ [char] } =		\ trailing }?
  if  1- 1  exit  endif
  dup 2 >= if
   2dup + 2- s" };" s=
   if  2- 2 exit  endif
  endif
  0
;

[-short-branches
: process-enum	\ u caddr len -- u'
\ +G Process one element of an enum.
  {: | temp[ 256 ] -- :}
  -white dup 2 >= if			\ check for C comments
    over s" //" s=
    if  2drop exit  endif
    over s" /*" s=
    if  2drop exit  endif
  endif
  dup 0= if				\ check if empty
    ( cr ." empty" ) 2drop  exit
  endif
  rot >r
  [char] = csplit 2swap dup if		\ string contains '='
    1 /string -white			\ step over '='
    isInteger? 1 <> #-13 ?throw		\ convert string
    r> drop >r				\ replace current value
  else
    2drop
  endif
\  s" EQU " temp[ place			\ build "EQU <name>"
  s" CONSTANT " temp[ place			\ build "EQU <name>"
  -white temp[ append
  r@ temp[ count evaluate		\ build constant
  r> 1+					\ -- u'
;
short-branches?

: (enum)	\ --
\ +G Process an enum once the name has been processed.
   0  begin
     next-enum enum-term? 0=		\ -- u caddr len ; terminator?
    while
     process-enum
   repeat
   process-enum  drop
;
previous definitions

vocabulary NamedEnums

also system
: enum		\ --
\ *G Process an enum of the form:
\ *C   enum <name> { a, b, c=10, d };
\ *P <name> is placed in the *\fo{NamedEnums) vocabulary.
\ ** The elements appear as Forth constants in the *\fo{CURRENT)
\ ** wordlist/vocabulary.
\ ** The definition may extend over many lines. C comments may
\ ** occur after the ',' separator, e.g.
\ *C  JIM = 25, // comment about this line
  [char] { parse-word ['] NamedEnums voc>wid $create-in
  (enum)
;

: enum{		\ --
\ *G Process an enum of the form:
\ *C   enum{ a, b, c=10, d };
  (enum)
;
previous

: .NamedEnums	\ --
\ *G List all the named ENUMs.
  also NamedEnums  words  previous
;


\ *****************
\ *S Stack guarding
\ *****************
\ *P These words preserve the stack pointers, put dummy items on
\ ** the stack, and restore the stack later. Use these words inside
\ ** a colon definition for protection of (say) a text interpreter.
\ ** See *\i{VFXBase/GenTools.fth}.
\ *C   [GuardSP ... GuardSP]

8 cells constant /guard                                                   \ -- n
\ *G The amount by which the stack is guarded.

: [GuardSP \ -- ; R: -- oldS0 oldSP
\ *G Reserves guard space on the data stack and resets the data
\ ** stack pointers as if the stack was empty.
  ?comp
  postpone S0 postpone @ postpone >R
  postpone SP@ postpone >R                     \ save current stack pointers !!!
  postpone sp@ postpone /guard postpone -
  postpone dup postpone s0 postpone ! postpone sp!   \ drop stack *** CARNAL ***
; immediate

: GuardSP] \ -- ; R: oldS0 oldSP --
\ *G Restore the stack condition saved by *\fo{[GuardSP}.
  postpone R> postpone SP!                            \ reset stack pointers !!!
  postpone R> postpone S0 postpone !
; immediate


\ *************************
\ *S Transient word regions
\ *************************
\ *P There are occasions when we want to run words only once or
\ ** twice in order to initialise existing data. These words
\ ** occupy space. The code here allows you to build words in
\ ** a transient region allocated from the heap. When the region
\ ** is released, all the words in it are removed from the
\ ** dictionary and the memory is freed. The regions are nestable.
\ ** See *\i{VFXBase/GenTools.fth}.
\ *E [tr
\ **   : foo ....  ;
\ **   foo
\ ** tr]

ErrDef bad_transientArea	"Not in transient area"

64 kb value /TA	\ -- len
\ Size of a transient area.

0 value anchorTR	\ -- addr|0
\ *G Anchors the linked list of transient areas.

struct /transient	\ -- len
\ *G Defines the head of the transient region.
  int tr.link			\ previous transient area in chain
  int tr.prevDP			\ previous DP
  int tr.idata			\ IDATA state
  int tr.sp			\ SP on entry
  int tr.s0			\ S0 on entry
end-struct

: [TR		\ --
\ *G Allocate a new transient area
  sp@ >r				\ stack pointer
  anchorTR  /TA protAlloc		\ get space, THROW on error
  tuck tr.link !  			\ save previous transient area
  dp @ over tr.prevDP !			\ save previous DP
  idata? over tr.idata !		\ IDATA state
  r> over tr.sp !			\ preserve old stack pointer
  S0 @ over tr.s0 !			\ preserve old S0
  dup -> anchorTR			\ save new transient area
  /transient + DP !			\ set new DP (HERE).
  0 -> idata?				\ no IDATA
  0 0 0 0  0 0 0 0			\ guard cells on data stack
  sp@ s0 !				\ reset S0
;

: TR]		\ --
\ *G End the current transient area and remove it
  anchorTR				\ -- ta ; last TA
  dup 0= bad_transientArea ?throw	\ check transient area
  dup /transient + here trim-dictionary	\ forget transient words
  dup tr.idata @ -> idata?		\ restore IDATA state
  dup tr.prevDP @ dp !			\ restore HERE
  dup tr.sp @ >r			\ previous SP
  dup tr.s0 @ S0 !
  dup tr.link @  swap protFree -> anchorTR  \ unlink and free
  r> sp!				\ restore stack
;


\ ************************************
\ *S Eliminating compilation surprises
\ ************************************
\ *P There are four things that you can do with a word:
\ *( 1
\ *B Execute the interpretation behaviour of a word. This is
\ ** what happens when you type it on the console. You can pass
\ ** the xt to *\fo{EXECUTE}.
\ *B Execute the compilation behaviour of a word.
\ *B Compile the interpretation behaviour of the word.
\ *B Compile the compilation behaviour of the word.
\ *)
\ *P The four words here provide these functions explicitly.
\ ** See *\i{VFXBase/GenTools.fth}.

: exec-interp	\ xt --
\ *G Execute the interpretation action as if on the console.
   execute  ;

: exec-comp	\ xt --
\ *G Execute the compilation behaviour of the word.
  dup ndcs? if
    >code-len @ execute
  else
    compile,
  then
;

: comp-interp	\ xt --
\ *G Compile the interpretation behaviour of a word.
  postpone literal  postpone execute  ;

: comp-comp	\ xt --
\ *G Compile the compilation behaviour of a word.
  dup ndcs? if			\ NDCS/immediate
    >code-len @ docall,			\ lay as call
  else				\ not NDCS
    clit,  ['] compile, compile,
  then
;

((
: (ndcs,)	\ i*x xt -- j*x
\ *G Like *\fo{(COMPILE,)} but executes the NDCS action for a word
\ ** and may parse and/or have a stack effect during compilation.
  >code-len @ execute			\ execute NDCS/IMMEDIATE word
;
: compile-word	\ i*x xt -- j*x
\ *G Process an XT for compilation.
  dup ndcs?				\ NDCS or normal
  if  ndcs,  else  compile,  then
;
))

\ ======
\ *> ###
\ ======

decimal
