\ MODULES.FTH - private wordlists with visibility control

\ ((
\ Copyright (c) 2003, 2004
\ MicroProcessor Engineering
\ 133 Hill Lane
\ Southampton SO15 5AF
\ England
\
\ tel: +44 (0)23 8063 1441
\ net: mpe@mpeforth.com
\      tech-support@mpeforth.com
\ web: www.mpeforth.com
\
\ From North America, our telephone and fax numbers are:
\   011 44 23 8063 1441
\   011 44 23 8033 9691
\
\
\ To do
\ =====
\
\ Change history
\ ==============
\ ))

decimal

\ ======
\ *> wid
\ *S Source Code Modules
\ ======
\ *P Apart from wordlists and vocabularies, VFX Forth provides
\ ** 'source modules'.
\ ** A *\fo{MODULE} is a section of source code which handles a given
\ ** task. Rather than having all the factored 'sub-words' built into
\ ** the public dictionary, a module exists in its own private
\ ** wordlist and only provides visible access to those words which
\ ** have been deliberately *\fo{EXPORTED} by the author. This method helps to
\ ** improve the maintainability of large source projects both for
\ ** single programmers and for group efforts.

\ *P When using this system, the implication is that a function exported
\ ** by the author will be maintained and not change its meaning or
\ ** implementation in an 'invisible' manner. Unexported words may
\ ** change at any time.

\ *P For example, a module written by one person for use by another may
\ ** require a sub-word to lay a string in the dictionary. If initially
\ ** this word takes a counted string and builds a 0 terminated one in
\ ** the dictionary it is possible that other sources will use this
\ ** function for their own use. If at a later date the author of the
\ ** original module needs to store strings in unicode format due to a
\ ** change in the overall architecture of the module all other
\ ** unauthorised uses of the sub-word will break through no fault
\ ** of the original author. By hiding the mechanics of an API in a
\ ** module this breakage cannot happen.

( +N Constants and Globals                                              )
( +*                                                                    )
( +D widMODULES         A word-list-identifier which describes the      )
( +*                    wordlist to be used for MODULE definitions.     )
( +*                    This is only used for the MODULE names. Each    )
( +*                    individual module has a private wordlist into   )
( +*                    which the contents are compiled.                )

also system definitions

WORDLIST constant widModules	\ -- wid ; links all modules

0 value sModule	\ -- addr ; holds current module structure

: FindModule	\ "<name>" -- xt
\ Find the module represented by the next source item.
\ An error occurs if it is not found.
  get-token count widModules search-wordlist ?undef
;

: FindInModule	\ "<name>" -- xt
\ Find the word represented by the next source item in the
\ current module being defined. An error occurs if it is
\ not found.
  get-token count get-current search-wordlist ?undef
;

previous definitions

also system

\ ====================
\ *N Module definition
\ ====================

: Module        \ <"name"> -- old-current
\ *G Begin the definition of a new source module. Modules can be nested
\ ** and the *\fo{EXPORT}s from any module are placed in the current user
\ ** definitions vocabulary.
  get-current                   \ Preserve original CURRENT till end of MODULE
  widModules set-current        \ Create definition in MODULES wordlist
  wordlist                      \ Create private wordlist for module
  create                        \ Create MODULE entry by name.
    here -> sModule		\ mark current module data
    dup ,                       \ The data for a MODULE entry is the
                                \ private word-list pointer
    ['] noop ,			\ initialisation action
    ['] noop ,			\ termination action
    >r
    get-order r@ swap 1+ set-order \ Add private wordlist to search order ....
    r> set-current              \ and make it the definition wordlist
;

: End-Module    \ old-current --
\ *G Mark the end of the current module under definition.
  get-order nip 1- set-order    \ Remove private wordlist from search order
  set-current                   \ restore original definition wordlist CURRENT
  0 -> sModule
;

: EXPORT	\ old-current -- old-current ; EXPORT <name>
\ *G Export a module definition into the user's definition wordlist.
\ ** The dictionary header for the word is relinked from the wordlist
\ ** in which it was defined to the user's definition wordlist.
\ *C   EXPORT <name>
  FindInModule drop
  LastNameFound over MoveNameToWid
  0= abort" Can't move this word"
;

: Set-Init-Module	\ xt --
\ *G St the initialisation action of a module, which can be triggered
\ ** by *\fo{INIT-MODULE <name>}.
\ ** Must be executed within a module definition, and the xt must
\ ** have no stack effect ( -- ).
\ *C   ' <action> SET-INIT-MODULE
  sModule cell + !
;

: Set-Term-Module	\ xt --
\ *G Set the termination action of a module, which can be triggered
\ ** by *\fo{TERM-MODULE <name>}.
\ ** Must be executed within a module definition, and the xt must
\ ** have no stack effect ( -- ).
\ *C   ' <action> SET-TERM-MODULE
  sModule 2 cells + !
;


\ ====================
\ *N Module management
\ ====================

: INIT-MODULE	\ "<name>" -- ; INIT-MODULE <name>
\ *G Calls the initialiser of the module whose name follows.
  FindModule >body cell +		\ address holding initialiser
  state @ if
    postpone literal  postpone @  postpone execute
  else
    @ execute
  endif
; immediate

: TERM-MODULE	\ "<name>" -- ; TERM-MODULE <name>
\ *G Calls the terminator of the module whose name follows.
  FindModule >body 2 cells +		\ address holding terminator
  state @ if
    postpone literal  postpone @  postpone execute
  else
    @ execute
  endif
; immediate

: REQUIRES      \ "<name>" -- ; REQUIRES <name>
\ *G Specifies by name a module which is required in order to compile
\ ** the current source code. If the required module is not present
\ ** compilation is *\fo{ABORT}ed.
\ *C  requires MyModule
  FindModule drop
;

: EXPOSE-MODULE         \ -- ; EXPOSE-MODULE <name>
( *G This word will add the private word-list of the module <name> to   )
( ** the search order. It is a debugging aid and should only be used as )
( ** such. Using this method to get at a module's internal definitions  )
( ** defeats the purpose of the module mechanism.                       )
\ *C  expose-module MyModule
  FindModule >body @ >r			\ private wordlist pointer
  get-order r> swap 1+ set-order	\ add private wordlist to search order
;

previous

\ ====================
\ *N An Example Module
\ ====================
( *P The code below defines a module with one public word. The module   )
( ** itself doesn't actually do anything of consequence but it does     )
( ** show the definition syntax.                                        )
( **                                                                    )
( *P After compilation the only publically available words will be the  )
( ** two exported at the bottom of the module. All other definitions    )
( ** will be hidden and can only be accessed after an *\fo{EXPOSE-MODULE}     )
( ** is executed. In this way the actual implementation of the API can  )
( ** be isolated, only the author needs to worry about it.              )
( *E    MODULE  counter                                                 )
( **                                                                    )
( **    variable counter                                         )
( **                                                                    )
( **    : incr-counter          \ --                                    )
( **      1 counter +!                                                  )
( **    ;                                                               )
( **                                                                    )
( **    : get-counter           \ -- val                                )
( **      counter @                                                     )
( **    ;                                                               )
( **                                                                    )
( **    : set-counter           \ val --                                )
( **      counter !                                                     )
( **    ;                                                               )
( **                                                                    )
( **    : CounterInitialise     \ --                                    )
( **      0 set-counter                                                 )
( **    ;                                                               )
( **                                                                    )
( **    : Counter@++            \ -- value                              )
( **      get-counter incr-counter                                      )
( **    ;                                                               )
( **                                                                    )
\ **    ' CounterInitialise SET-INIT-MODULE
\ **    ' CounterInitialise SET-TERM-MODULE
\ **
( **    EXPORT  CounterInitialise     \ Public word to init             )
( **    EXPORT  Counter@++            \ Fetch value and incr.           )
( **                                                                    )
( **    END-MODULE                                                      )


\ ======
\ *> ###
\ ======

